"""
Modul mit Observer
"""

try:
    from PyQt5.QtCore import *
    import time
    from collections import OrderedDict
    from gui.view import View
    from gui.model import Model
except ImportError as ex:
    print("Module missing ("+str(__file__)+"): " + ex.name + "\nPlease install.")

    quit()

class Observer(QObject):

    """
    OBSERVER

    - Push-Observer der View fuer das Model
    - Wenn das Model Informationen ausgeben will, wird die View nicht direkt sondern über diese Klasse angesteuert.

    """

    # Signale
    signalPrint = pyqtSignal(str, str)
    signalProgressShow = pyqtSignal(bool, bool)
    signalProgressShowText = pyqtSignal(bool)
    signalProgressHide = pyqtSignal()
    signalProgressSetWert = pyqtSignal(int)
    signalProgressSetMinMax = pyqtSignal(int, int)
    signalProgressSetText = pyqtSignal(str)
    signalStatusbarSetText = pyqtSignal(str)
    signalWorkflowSet = pyqtSignal(str, str, int, int)
    signalInputComboMulti = pyqtSignal(str, OrderedDict)
    signalUpdateMyworms = pyqtSignal()
    signalUpdateSysMon = pyqtSignal()
    signalUpdateIngestTable = pyqtSignal()
    signalProgressErrWarnInfo = pyqtSignal(int,int,int)
    signalBatch = pyqtSignal(int,int,str)

    # Konstruktor
    def __init__(self, view):
        """
        Args:
            view (gui.view.View): View
        """

        QObject.__init__(self)

        # Instanzvariablen
        self.__view = view # View

        # Anti Spam
        self.__fAntiSpamCooldown = 0.5
        self.__timer_status = time.time()
        self.__timer_sysMon = time.time()
        self.__timer_ingestTable = time.time()
        self.__timer_progressValue = time.time()
        self.__timer_progressText = time.time()
        self.__timer_myWorms = time.time()
        self.__timer_errWarnInfo = time.time()

        # Verbinde die View mit mir
        self.__view.sys_connectToObserver(self)

    def inputComboMulti(self, sTitel, odicCombos):
        """
        Benutzer soll eine MultiCombo Eingabe machen

        Args:
            sTitel (str): Titel
            odicCombos (odic): Combos
        """
        self.signalInputComboMulti.emit(sTitel, odicCombos)

    def printConsole(self, sText, sAppendix="\n"):
        """
        Ausdruck in GUI-Konsole

        Args:
            text (str): Der auszudruckende Text
            app (str): Appendix
        """

        self.signalPrint.emit(sText, sAppendix)

    def showProgress(self, **kwargs):
        """
        Zeige Progress

        Args:
            **kwargs:
                abort (Bool): Abort Schalter an
                pulse (Bool): Pulsieren an
        """
        bAbort = True
        bPulse = False
        if "abort" in kwargs:
            bAbort = kwargs["abort"]
        if "pulse" in kwargs:
            bPulse = kwargs["pulse"]
        self.signalProgressShow.emit(bAbort, bPulse)

    def hideProgress(self):
        """
        Progressbar verstecken
        """
        self.signalProgressHide.emit()

    def setProgressbarMinMax(self,min=0,max=0):
        """
        Progressbar min/max einstellen

        Args:
            min (int):
            max (int):
        """
        self.signalProgressSetMinMax.emit(min, max)

    def setProgressbarWert(self, wert):
        """
        Porgressbar Wert setzen
        Args:
            wert (int): Wert

        """
        if time.time() - self.__timer_progressValue > self.__fAntiSpamCooldown:  # Anti Spam
            self.signalProgressSetWert.emit(wert)
            self.__timer_progressValue = time.time()

    def setProgressbarTextAnzeigen(self, bAnzeige):
        """
        Textanzeige beim Progressbar

        Args:
            bAnzeige (bool): Text anzeigen ja/nein

        """
        self.signalProgressShowText.emit(bAnzeige)

    def setProgressbarText(self, sText):
        """
        Text im Progressbar anzeigen

        Args:
            sText (str): Text

        """
        if time.time() - self.__timer_progressText > self.__fAntiSpamCooldown:  # Anti Spam
            self.signalProgressSetText.emit(sText)
            self.__timer_progressText = time.time()


    def setWorkflow(self, sTop, sSub, iStepAkt, iStepMax):
        """
        Aktuellen Status im WorkflowMultisheetTesting Bereich anzeigen

        Args:
            sText (str): Text
            iStepAkt (int): Aktueller Schritt
            iStepMax (int): Alle Schritte

        """
        self.signalWorkflowSet.emit(sTop, sSub, iStepAkt, iStepMax)

    def updateMyworms(self):
        """
        Update My WoRMS
        """
        if time.time() - self.__timer_myWorms > self.__fAntiSpamCooldown: # Anti Spam
            self.signalUpdateMyworms.emit()
            self.__timer_myWorms= time.time()

    def updateIngestTable(self):
        """
        Update IngestTable
        """
        if time.time() - self.__timer_ingestTable > self.__fAntiSpamCooldown: # Anti Spam
            self.signalUpdateIngestTable.emit()
            self.__timer_ingestTable = time.time()

    def updateSysmon(self):
        """
        Update SysMon
        """
        if time.time() - self.__timer_sysMon > self.__fAntiSpamCooldown: # Anti Spam
            self.signalUpdateSysMon.emit()
            self.__timer_sysMon = time.time()

    def setProgressErrWarnInfo(self, iErr, iWarn, iInfo):
        """
        Update Progress Error Warn Info
        """
        if time.time() - self.__timer_errWarnInfo > self.__fAntiSpamCooldown:  # Anti Spam
            self.signalProgressErrWarnInfo.emit(iErr, iWarn, iInfo)
            self.__timer_errWarnInfo = time.time()

    def setStatus(self, sText):
        """
        Setze Statustext

        Args:
            sText (str): Text
        """
        if time.time() - self.__timer_status > self.__fAntiSpamCooldown: # Anti spamming
            self.signalStatusbarSetText.emit(sText)
            self.__timer_status = time.time()

    def setBatch(self, iNr, iMax, sInfo):
        """
        Update Batch
        """
        self.signalBatch.emit(iNr, iMax, sInfo)
