"""
Modul mit Hilfsklasse für die Datensatzselektion
"""

# coding=utf-8
import psycopg2, re, sys, datetime
from tools.printing import SmartPrinter
from tools.coding import Result
from tools.text import PrettyText
from collections import OrderedDict

class Selector:
    """
    CONTAINER FUER SELECTION VON SUBSETS AUS DATENBANKBESTAND
    """

    def __init__(self, model):
        """
        Args:
            model (model.Model): Model
            dicChoiceAll (dic): Auswahl fuer Subselektion
        """

        self.__prnt = model.ui_getSmartprinter()
        self.__model = model
        self.__lstErr = []
        self.__lstSelects = ("date_start", "date_end", "region", "type", "dataset", "ingest", "gear", "taxcover")  # Alle Kategorien der Selectoren für die Subsetauswahl
        self.__odicConfig = OrderedDict()
        self.__dicConfigSpecial = {}  # Werte: date_start_min, date_end_max
        self.__res = None
        self.resetResult()

    def __str__(self):
        sOut = "* Selector Configuration\n"
        sOut += self.infoConfig()
        return sOut

    def infoInput(self, dicInput):
        lstConfig = [["selector", "entry", "active"]]
        for sSelectorName, dicEntries in dicInput.items():
            for sEntryName, bActiv in dicEntries.items():
                lstConfig.append([str(sSelectorName), str(sEntryName), ("yes" if bActiv else "no")])
        return PrettyText.table(lstConfig)

    def infoConfig(self):
        lstConfig = [["selector", "entry", "effective", "active"]]
        for sSelectorName, dicEntries in self.__odicConfig.items():
            for sEntryName, (bEffective, bActive) in dicEntries.items():
                lstConfig.append([str(sSelectorName), str(sEntryName), ("yes" if bEffective else "no"), ("yes" if bActive else "no")])
        if self.__dicConfigSpecial["date_start_min"] is not None:
            lstConfig.append(["date_start_min", str(self.__dicConfigSpecial["date_start_min"]), "-", "-"])
        if self.__dicConfigSpecial["date_end_max"] is not None:
            lstConfig.append(["date_end_max", str(self.__dicConfigSpecial["date_end_max"]), "-", "-"])
        return PrettyText.table(lstConfig)

    def infoConfigShort(self):
        # Selection
        lstSelectors = ["date_start", "date_end", "region", "type", "dataset", "gear", "taxcover", "ingest"]
        lstConfig = [["selector", "no.", "choosen"]]
        for sSelectorName in lstSelectors:
            lstSelectorEntrys = [str(sEntryName) for sEntryName, (bEffective, bActive) in self.__odicConfig[sSelectorName].items() if (bActive and bEffective)]
            lstConfig.append([sSelectorName, str(len(lstSelectorEntrys)) if not sSelectorName in ("date_start", "date_end") else "", PrettyText.lst2strC(lstSelectorEntrys)])
        return PrettyText.table(lstConfig)

    def infoConfigShort2(self):
        # Selection
        sOut = ""
        lstSelectors = ["date_start", "date_end", "region", "type", "dataset", "gear", "taxcover", "ingest"]
        for sSelectorName in lstSelectors:
            lstSelectorEntrys = [str(sEntryName) for sEntryName, (bEffective, bActive) in self.__odicConfig[sSelectorName].items() if (bActive and bEffective)]
            if sSelectorName in ["date_start", "date_end"]:
                sOut += "- "+sSelectorName + ": "+PrettyText.lst2strC(lstSelectorEntrys) + "\n"
            else:
                sOut += "- "+sSelectorName + " (" + (str(len(lstSelectorEntrys)) if not sSelectorName in ("date_start", "date_end") else "") + "): "
                sOut += PrettyText.lst2str(lstSelectorEntrys, "|")
                sOut += "\n"
        return sOut

    # ------------------------------------------------------------------------------------------------------------------

    def resetResult(self):
        self.__res = Result("Selector Configuration")

    def isErr(self):
        return not self.__res.isOk()

    def getResult(self):
        return self.__res

    def isOk(self):
        return self.__res.isOk()

    # ------------------------------------------------------------------------------------------------------------------
    # Eingabe
    # ------------------------------------------------------------------------------------------------------------------

    def deactivateAllEntries(self, sSelector):
        """
        Deaktiviere alle Einträge eines Selectors

        Args:
            sSelector (str): Selector
        """

        if sSelector in self.__odicConfig and sSelector not in ("date_start", "date_end"):
            for sEntryName, bEntryActive in self.__odicConfig[sSelector].items():
                try:
                    bConfigEffective = self.__odicConfig[sSelector][sEntryName][0]
                except:
                    bConfigEffective = False
                self.__odicConfig[sSelector][sEntryName] = (bConfigEffective, False)

    def update(self, dicInput):
        """
        Update: Konfiguration (anhand der Benutzereingaben)
        """
        pd = self.__prnt.debug

        # Nichts zu tun

        pd("* Selector gets new user input\n"+self.infoInput(dicInput))
        if dicInput == {}:
            pd("user input is empty... nothing to do ...")
            return

        # Aktivität der Konfiguration updaten
        self.__updateConfigActivity(dicInput)

        # Effektivität der Konfiguration updaten
        self.__updateConfigEffectivity()

        pd("* Selector is updated\n"+self.infoConfig())

    def reset(self):
        """
        Reset Selector to Default
        """
        pd = self.__prnt.debug

        self.resetResult()
        self.__odicConfig = self.__buildConfigDefault()
        self.__dicConfigSpecial["date_start_min"] = None
        self.__dicConfigSpecial["date_end_max"] = None

        pd("* Selector is resetted\n"+self.infoConfig())

    # ------------------------------------------------------------------------------------------------------------------
    # Hole Ergebnisse
    # ------------------------------------------------------------------------------------------------------------------

    def getSpecial(self, sName):
        """
        Liefert "besondere" Kongfigurationswerte
        Args:
            sName (str): Name des Wertes (date_start_min, date_end_max)
        Returns:
            Wert
        """

        return self.__dicConfigSpecial[sName] if (sName in self.__dicConfigSpecial) else None

    def getEntries(self, sSelectorName):
        """
        Liefert die Dictionry (Name: (effektiv, aktiv) ) aller Einträge eines speziellen Selectors

        Args:
            sSelectorName (str): Selectorname

        Returns:
            (dic): Einträge
        """
        return self.__odicConfig[sSelectorName]


    def getActivEntries(self, sSelectorName):
        """
        Liefert die Liste aller aktivierten Einträge eines speziellen Selectors

        Args:
            sSelectorName (str): Selectorname

        Returns:
            (list): Einträge
        """
        return [sEntryName for sEntryName, (bEffective, bActive) in self.__odicConfig[sSelectorName].items() if bActive]

    def getEffectiveEntries(self, sSelectorName):
        """
        Liefert die Liste aller effektiven Einträge eines speziellen Selectors

        Args:
            sSelectorName (str): Selectorname

        Returns:
            (list): Einträge
        """
        return [sEntryName for sEntryName, (bEffective, bActive) in self.__odicConfig[sSelectorName].items() if bEffective]

    def getEffectiveActiveEntries(self, sSelectorName):
        """
        Liefert die Liste aller effektiven und aktiven Einträge eines speziellen Selectors

        Args:
            sSelectorName (str): Selectorname

        Returns:
            (list): Einträge
        """
        return [sEntryName for sEntryName, (bEffective, bActive) in self.__odicConfig[sSelectorName].items() if bEffective and bActive]

    # ------------------------------------------------------------------------------------------------------------------
    # Berechne Ergebnisse
    # ------------------------------------------------------------------------------------------------------------------

    def __buildConfigDefault(self):
        """
        Baue: Startkonfiguration

        Returns:
            OrderedDict: Grundkonfiguration (Selectorname: (Eintragsname: (effektiv, aktiv)))
        """

        # DB
        sDbSchema = self.__model.subsystem_getDb().account_active_getSchema()
        lstQueries = [
            ("date_start",
             "SELECT MIN(date_start) FROM " + sDbSchema + ".sample"),
            ("date_end",
             "SELECT MAX(date_start) FROM " + sDbSchema + ".sample"),
            ("region",
             "SELECT reg.name FROM " + sDbSchema + ".region reg WHERE id IN (SELECT DISTINCT region_id FROM " + sDbSchema + ".region2sample) GROUP BY reg.name, reg.code ORDER BY reg.code"),
            ("type",
             "SELECT DISTINCT ing.workflow FROM " + sDbSchema + ".ingest ing ORDER BY ing.workflow"),
            ("dataset",
             "SELECT DISTINCT dat.name FROM " + sDbSchema + ".subset sub JOIN " + sDbSchema + ".dataset dat ON dat.id = sub.dataset_id ORDER BY dat.name"),
            ("ingest",
             "SELECT DISTINCT CAST(smp.ingest_id AS VARCHAR) AS id FROM " + sDbSchema + ".sample smp ORDER BY id"),
            ("gear",
             "SELECT DISTINCT gear.category FROM " + sDbSchema + ".sample smp JOIN " + sDbSchema + ".gear gear ON gear.id = smp.gear_id ORDER BY gear.category"),
            ("taxcover",
             "SELECT DISTINCT cov.name FROM " + sDbSchema + ".subset JOIN " + sDbSchema + ".taxcover cov ON cov.id = subset.taxcover_id ORDER BY cov.name")
        ]
        try:
            odicErg = OrderedDict()

            resCursor = self.__model.subsystem_getDb().conn_getConnection()
            if resCursor.isErr():
                raise NameError("Error getting Connection to Database in Selector")
            conn = resCursor.getData("conn")
            cursor = resCursor.getData("cursor")

            for sName, sSql in lstQueries:
                cursor.execute(sSql)
                lstErg = cursor.fetchall()
                # Ergebnisselemente sind Tupel mit einem zweiten leeren Eintrag -> Bereinigen
                #lstErgclean = sorted([tp[0] for tp in lstErg])
                lstErgclean = [tp[0] for tp in lstErg]
                odicErgClean = OrderedDict()
                if sName == "region":
                    # Region: LEFTOVER vor alle anderen Eintraege stellen
                    sLeftover_name = self.__model.subsystem_getRegions().conf_getLeftover_name()
                    if sLeftover_name in lstErgclean:
                        lstErgclean.remove(sLeftover_name)
                        lstErgclean.insert(0, sLeftover_name)
                for elem in lstErgclean:
                    if elem is not None:
                        if sName in ("date_start", "date_end"):
                            # Datum
                            if elem is not None:
                                odicErgClean[elem] = (True, True)
                                # Sezte separat: date_start_min, date_end_max
                                if sName == "date_start":
                                    self.__dicConfigSpecial["date_start_min"] = elem
                                if sName == "date_end":
                                    self.__dicConfigSpecial["date_end_max"] = elem
                        else:
                            # Combo
                            odicErgClean[elem] = (True, True)
                odicErg[sName] = odicErgClean

            return odicErg

        except Exception as ex:
            exc_type, exc_obj, exc_tb = sys.exc_info()
            iLine = exc_tb.tb_lineno
            self.__res.addErr("Unexpected error in Selector: " + str(ex) + " in code line " + str(iLine))
            return {}

    def __updateConfigActivity(self, dicInput):
        """
        Update: Aktivitäten in der Konfiguration
        """

        for sSelectorName, dicEntries in dicInput.items():
            for sEntryName, bEntryActive in dicEntries.items():
                if sSelectorName in ("date_start", "date_end"):
                    # Tweak Datum wird anders eingetragen als Comboauswahlelement
                    self.__odicConfig.pop(sSelectorName, None)
                    self.__odicConfig[sSelectorName] = {sEntryName: (True, True)}
                else:
                    # Comboelement
                    try:
                        bConfigEffective = self.__odicConfig[sSelectorName][sEntryName][0]
                    except:
                        bConfigEffective = False
                    self.__odicConfig[sSelectorName][sEntryName] = (bConfigEffective, bEntryActive)

    def __updateConfigEffectivity(self):
        """
        Update: Effektivitäten in der Konfiguration
        Die Aktivitäten aller anderen Selectoren definieren die Effektivität eines Eintrags
        """

        pd = self.__prnt.debug

        # Hilfsmethode: Liste -> SQL-String
        def lst2str(lst):
            if len(lst) == 0:
                return "\'\'"
            sOut = ""
            for sName in lst:
                sOut += "\'"+sName+"\',"
            sOut = sOut[:-1]
            return sOut

        # Db
        sDbSchema = self.__model.subsystem_getDb().account_active_getSchema()
        pd("* Update Selector Effectivity")
        try:
            resCursor = self.__model.subsystem_getDb().conn_getConnection()
            if resCursor.isErr():
                raise NameError("Error getting Connection to Db in Selector")
            conn = resCursor.getData("conn")
            cursor = resCursor.getData("cursor")

            # Loop: alle Selectornamen
            for sSelectorName in self.__lstSelects:

                # Ergebnis
                if sSelectorName == "date_start":
                    sSql = "SELECT MIN(smp.date_start) "
                elif sSelectorName == "date_end":
                    sSql = "SELECT MAX(smp.date_start) "
                elif sSelectorName == "region":
                    sSql = "SELECT DISTINCT reg.name "
                elif sSelectorName == "type":
                    sSql = "SELECT DISTINCT ing.workflow "
                elif sSelectorName == "dataset":
                    sSql = "SELECT DISTINCT dat.name "
                elif sSelectorName == "gear":
                    sSql = "SELECT DISTINCT gea.category "
                elif sSelectorName == "ingest":
                    sSql = "SELECT DISTINCT CAST(ing.id AS VARCHAR) "
                elif sSelectorName == "taxcover":
                    sSql = "SELECT DISTINCT cov.name "
                else:
                    return None

                # Daten
                sSql += "FROM " + sDbSchema + ".sample smp " + \
                "JOIN " + sDbSchema + ".subsample subsmp ON smp.id = subsmp.sample_id " + \
                "JOIN " + sDbSchema + ".subset sub ON subsmp.id = sub.subsample_id " + \
                "JOIN " + sDbSchema + ".station sta ON sta.id = smp.station_id " + \
                "JOIN " + sDbSchema + ".cruise cru ON cru.id = sta.cruise_id " + \
                "JOIN " + sDbSchema + ".ingest ing ON ing.id = smp.ingest_id " + \
                "JOIN " + sDbSchema + ".dataset dat ON dat.id = sub.dataset_id " + \
                "JOIN " + sDbSchema + ".gear gea ON gea.id = smp.gear_id " + \
                "JOIN " + sDbSchema + ".taxcover cov ON cov.id = sub.taxcover_id " + \
                "JOIN " + sDbSchema + ".region2sample reg2smp ON reg2smp.sample_id = smp.id " + \
                "JOIN " + sDbSchema + ".region reg ON reg2smp.region_id = reg.id " + \
                "WHERE true "

                # Einschränkung (benuzte nur aktive Auswahlen der anderen Selectoren)
                if sSelectorName not in ("date_start", "date_end"):
                    if len(self.getActivEntries("date_start")) != 0 and len(self.getActivEntries("date_end")) != 0:
                        sSql += "AND smp.date_start BETWEEN '" + self.getActivEntries("date_start")[0].isoformat() + "' AND '" + self.getActivEntries("date_end")[0].isoformat() + "' "
                if sSelectorName != "region":
                    sSql += "AND reg.name IN (" + lst2str(self.getActivEntries("region")) + ") "
                if sSelectorName != "type":
                    sSql += "AND ing.workflow IN (" + lst2str(self.getActivEntries("type")) + ") "
                if sSelectorName != "dataset":
                    sSql += "AND dat.name IN (" + lst2str(self.getActivEntries("dataset")) + ") "
                if sSelectorName != "gear":
                    sSql += "AND gea.category IN (" + lst2str(self.getActivEntries("gear")) + ") "
                if sSelectorName != "taxcover":
                    sSql += "AND cov.name IN (" + lst2str(self.getActivEntries("taxcover")) + ") "
                if sSelectorName != "ingest":
                    sSql += "AND CAST(ing.id AS VARCHAR) IN (" + lst2str(self.getActivEntries("ingest")) + ") "

                pd("---"+sSelectorName+"\n"+sSql)

                cursor.execute(sSql)
                lstErg = cursor.fetchall()
                lstErgClean = [i[0] for i in lstErg if i is not None and i[0] is not None]  # Entferne None's und de-tuple das Ergebnis

                # Passe Config an
                if sSelectorName == "date_start":
                    # Spezial: date_start_min
                    self.__dicConfigSpecial["date_start_min"] = lstErgClean[0] if (len(lstErgClean) != 0) else None
                elif sSelectorName == "date_end":
                    # Spezial: date_end:max
                    self.__dicConfigSpecial["date_end_max"] = lstErgClean[0] if (len(lstErgClean) != 0) else None
                else:
                    # Rest: Combo
                    entries = self.__odicConfig[sSelectorName]
                    for sEntry, (bEntryEffective, bEntryActive) in entries.items():
                        if sEntry in lstErgClean:
                            bEffective = True
                        else:
                            bEffective = False
                        self.__odicConfig[sSelectorName][sEntry] = (bEffective, bEntryActive)

        except Exception as ex:
            exc_type, exc_obj, exc_tb = sys.exc_info()
            iLine = exc_tb.tb_lineno
            self.__res.addErr("Unexpected error", add=str(ex) + " in code line " + str(iLine), kat="Internal", sys="Selector")


