"""
Batch
"""
# coding=utf-8

import configparser
import datetime
import os
import pickle
import time
from collections import OrderedDict
from pathlib import Path

from tools.coding import Result, WorkingObject, PrettyText
from tools.fileio import FileTools
from tools.parallel import Stoppable


class BatchBot(Stoppable, WorkingObject):
    """
    Batch
    """

    # Konstruktor
    def __init__(self, model):
        Stoppable.__init__(self)
        WorkingObject.__init__(self)

        self.__model = model
        self.__smartprinter = model.ui_getSmartprinter()

        self.__dicConfigs = {}
        self.__dicConfigs_bak = {}
        self.__bBackup = False

        self.__sFileConfigs = "batchbot.dat"

        resLoad = self.load()
        if resLoad.isErr():
            print(resLoad.getErrWarnInfoString(short=True))
            resSave = self.save()
            if resSave.isErr():
                print(resSave.getErrWarnInfoString(short=True))
                self.setWorking(False)

        # Abort: Childs
        self.addAbortChild(self.__model.subsystem_getImporter())

    # ---

    def __str__(self):
        return self.info()

    def info(self):
        sOut = "*** Status of BatchBot\n\n"
        if not self.isReady():
            sOut += "System is not ready"
        else:
            if self.__bBackup:
                sOut += "--- Actual version\n"
            sOut += "* Configurations ("+str(len(self.__dicConfigs.keys()))+"x)\n\n"
            sOut += PrettyText.lst2strR(self.__dicConfigs.keys())
            sOut += "\n\n"
            if self.__bBackup:
                sOut += "--- Backup version\n"
                sOut += "* Configurations (" + str(len(self.__dicConfigs_bak.keys())) + "x)\n\n"
                sOut += PrettyText.lst2strR(self.__dicConfigs_bak.keys())

        return sOut

    # ---

    def backup(self):
        self.__bBackup = True
        self.__dicConfigs_bak = self.__dicConfigs.copy()

    def backup_del(self):
        self.__bBackup = False
        self.__dicConfigs_bak = {}

    def restore(self):
        if self.__bBackup:
            self.__dicConfigs = self.__dicConfigs_bak.copy()

    def isBackup(self):
        return self.__bBackup

    def load(self):
        res = Result("Load BatchBot Configurations")
        sFile = self.__model.conf_getUserDataDir() + "/" + self.__sFileConfigs
        try:
            with open(sFile, 'rb') as handle:
                date, self.__dicConfigs = pickle.load(handle)
                res.addData("file", sFile)
                res.addData("date", date)
        except Exception as ex:
            self.__dicConfigs = {}
            res.addErr("Problems loading BatchBot configurations -> Made a new empty version", add=str(ex), kat="I/O", sys="BatchBot")
        return res

    def save(self):
        res = Result("Save BatchBot configurations")
        sFile = Path(self.__model.conf_getUserDataDir(), self.__sFileConfigs)
        try:
            date = str(datetime.datetime.now().strftime("%Y-%m-%d %H:%M"))
            data = (date, self.__dicConfigs)
            with open(sFile, 'wb') as handle:
                pickle.dump(data, handle, protocol=pickle.HIGHEST_PROTOCOL)
                res.addData("file", sFile)
                res.addData("date", date)
        except Exception as ex:
            res.addErr("Problems BatchBot configurations", add=str(ex), kat="I/O", sys="BatchBot")
        return res

    # ---

    def getConfig_first(self):
        lstConfigs = self.getConfig_all()
        if lstConfigs is not None and len(lstConfigs) > 0:
            return lstConfigs[0]
        else:
            return None

    def getConfig_last(self):
        lstConfigs = self.getConfig_all()
        if lstConfigs is not None and len(lstConfigs) > 0:
            return lstConfigs[-1]
        else:
            return None

    def getConfig_all(self):
        try:
            lst = list(self.__dicConfigs.keys())
            lst.sort()
            return lst
        except:
            return None

    def getContent(self, sName):
        if sName in self.__dicConfigs:
            return self.__dicConfigs[sName]
        else:
            return None

    def setContent(self, sName, sContent):
        if sName in self.__dicConfigs:
            self.__dicConfigs[sName] = sContent

    def renameConfig(self, sName_old, sName_new):
        if sName_old not in self.__dicConfigs:
            return None
        if sName_old is None:
            return None
        sName_new_uniq = PrettyText.versionize(sName_new, list(self.__dicConfigs.keys()))
        try:
            self.__dicConfigs[sName_new_uniq] = self.__dicConfigs.pop(sName_old)
            return sName_new_uniq
        except Exception as ex:
            print(ex)
            return None

    def delConfig(self, sName):
        try:
            self.__dicConfigs.pop(sName)
        except Exception as ex:
            pass

    def newConfig(self, sName=None, sContent=None):
        if sName is None:
            sName = PrettyText.versionize("new",  list(self.__dicConfigs.keys()))
        if sContent is None:
            sContent = self.__getContent_template()
        self.__dicConfigs[sName] = sContent
        return sName

    def __getContent_template(self):
        # Workflowliste
        lstWfInfo = []
        for sWfShort in self.__model.subsystem_getMapper().workflow_getAll():
            sWfLong = self.__model.subsystem_getMapper().workflow_short2long(sWfShort)
            lstWfInfo.append(sWfShort + " (" + sWfLong + ")")
        sWfListe = PrettyText.lst2strC(lstWfInfo)
        lstWfCnfInfo = []
        for sWfCnfShort in self.__model.subsystem_getMapper().workflow_converter_import_getAll():
            sWfCnfLong = self.__model.subsystem_getMapper().workflow_converter_import_short2long(sWfCnfShort)
            lstWfCnfInfo.append(sWfCnfShort + " (" + sWfCnfLong + ")")
        sWfCnfListe = PrettyText.lst2strC(lstWfCnfInfo)
        # Template
        sOut = \
"""# ********************************************************
# FORMAT
# ********************************************************
# [Section]: Name of the Batch using the same Workflow
# wf:        Workflows: 
#            direct: """ + sWfListe + """      
#            conversion: """ + sWfCnfListe + """
# folder:    Absolute path to folder
# files:     List of files to include spearated by space
#            If no files are give the whole folder with all .xlsx files is processed
# exclude:   List of files to exclude
#
# Tipp:      When you want to insert returns between files the indent of 
#            filenames has to by one tab more right than the first file   
# ********************************************************

[Arktis]
wf: arc 
folder: /Users/paul/git/awi/data/arctic_ckraan

[Antarktis]
wf: ant
folder: /Users/paul/git/awi/data/antarctic_kteschke/normal
exclude: so_euphausidae_atkinson_krillbase.xlsx so_euphausidae_falkpetersen_1999.xlsx so_euphausidae_siegel.xlsx

[Northsea]
wf: nds 
folder: /Users/paul/git/awi/data/north_sea_jdannheim 
files: ben_mon_1969_2000.xlsx
        mpa_rn_infauna_only.xlsx
        beo_fino.xlsx
"""
        return sOut

    # ---

    def startBatch(self, sName, sLog_dir=None):
        self.abort(False)

        # Helper
        def checkLogDirs():
            # Checking LogDir
            if not os.path.exists(sLog_dir):
                try:
                    pn("Directory for logs not found: " + str(sLog_dir))
                    os.makedirs(sLog_dir)
                    pn("... now its made.")
                except Exception as ex:
                    res.addErr("Create folder for logs", add=str(ex), kat="Init", sys="BatchBot")
            # Checking LogDir Err
            if not os.path.exists(sDir_log_err):
                try:
                    pn("Directory for error logs not found: " + str(sDir_log_err))
                    os.makedirs(sDir_log_err)
                    pn("... now its made.")
                except Exception as ex:
                    res.addErr("Create folder for error logs", add=str(ex), kat="Init", sys="BatchBot")
            # Checking LogDir Err
            if not os.path.exists(sDir_log_ok):
                try:
                    pn("Directory for normal logs not found: " + str(sDir_log_ok))
                    os.makedirs(sDir_log_ok)
                    pn("... now its made.")
                except Exception as ex:
                    res.addErr("Create folder for normal logs", add=str(ex), kat="Init", sys="BatchBot")

        pn = self.__smartprinter.normal
        pv = self.__smartprinter.verbose

        progress = self.__smartprinter.progress
        workflow = self.__smartprinter.workflow
        batch = self.__smartprinter.batch
        pn = self.__smartprinter.normal

        workflow()
        batch()
        progress()

        res = Result("BatchBot")
        pn("* BatchBot")

        mapper = self.__model.subsystem_getMapper()
        if sLog_dir is None:
            sLog_dir = self.__model.conf_getWorkingDir()
        # Test
        if sLog_dir is None:
            res.addErr("Missing batch log dir", kat="Init", sys="BatchBot")
        if sName is None:
            res.addErr("Missing config", kat="Init", sys="BatchBot")
        if self.getContent(sName) is None:
            res.addErr("Unknown config", kat="Init", sys="BatchBot")
        if res.isErr():
            return res
        sBatch_content = self.getContent(sName)

        # Kommentare entfernen
        sBatch_tmp = sBatch_content
        sBatch_content = ""
        for sRow in sBatch_tmp.split("\n"):
            sRow_tmp = sRow.strip()
            if len(sRow_tmp) > 0 and sRow_tmp[0] != "#":
                sBatch_content += sRow + "\n"

        iZeitStart = time.time()

        # Log Dirs
        sTimeStamp = PrettyText.createTimestamp()
        sDir_log_timeStamp = Path(sLog_dir, sTimeStamp)
        sDir_log_err = Path(sDir_log_timeStamp, sDir_log_timeStamp, "err")
        sDir_log_ok = Path(sDir_log_timeStamp, sDir_log_timeStamp, "ok")
        checkLogDirs()

        # Reading Batchfile
        lstBatch = []
        if res.isOk():
            # *** Batch Datei auswerten
            config = configparser.ConfigParser(allow_no_value=True)
            try:
                config.read_string(sBatch_content)
            except Exception as ex:
                res.addErr("Consistency problem in configuration", add=str(ex), kat="Init", sys="BatchBot")

        if res.isOk():
            for sSection in config.sections():
                sWf = config[sSection]["wf"] if "wf" in  config[sSection] else None
                sFolder = config[sSection]["folder"] if "folder" in  config[sSection] else None
                sFiles = config[sSection]["files"] if "files" in  config[sSection] else None

                lstFiles = None
                if sFiles is not None and sFiles != "*":
                    sFiles = sFiles.replace("\n", " ")
                    sFiles = sFiles.replace("\t", " ")
                    if sFiles != "":
                        lstFiles_tmp = sFiles.split(" ")
                        lstFiles = [sFile.strip() for sFile in lstFiles_tmp if sFile.strip() != ""]
                        lstFiles_tmp = lstFiles
                        lstFiles = []
                        for sFile in lstFiles_tmp:
                            if sFile.endswith(".xlsx"):
                                lstFiles.append(sFile)
                            else:
                                lstFiles.append(sFile+".xlsx")
                sExclude = config[sSection]["exclude"] if "exclude" in  config[sSection] else None
                lstExclude = None
                if sExclude is not None and sExclude != "*":
                    sExclude = sExclude.replace("\n", " ")
                    sExclude = sExclude.replace("\t", " ")
                    if sExclude != "":
                        lstExclude_tmp = sExclude.split(" ")
                        lstExclude = [sExclude.strip() for sExclude in lstExclude_tmp if sExclude.strip() != ""]
                        lstExclude_tmp = lstExclude
                        lstExclude = []
                        for sFile in lstExclude_tmp:
                            if sFile.endswith(".xlsx"):
                                lstExclude.append(sFile)
                            else:
                                lstExclude.append(sFile + ".xlsx")

                # *** Testen
                pv("* Checking Batch Configuration")
                # --- Info
                lstBatRow_info = []
                lstBatRow_info.append("entry: " + str(sSection))
                lstBatRow_info.append("workflow: " + str(sWf))
                lstBatRow_info.append("folder: " + str(sFolder))
                if lstFiles is not None and len(lstFiles) != 0:
                    lstBatRow_info.append("files: {" + PrettyText.lst2strC(lstFiles)+"}")
                if lstExclude is not None and len(lstExclude) != 0:
                    lstBatRow_info.append("exclude: {" + PrettyText.lst2strC(lstExclude)+"}")
                sBatRow_info = PrettyText.lst2strC(lstBatRow_info)
                # --- Namen
                if sSection in [None, ""]:
                    res.addErr("Entry name not set", add=sBatRow_info, kat="Init", sys="BatchBot")
                if sWf in [None, ""]:
                    res.addErr("Workflow not set", add=sBatRow_info, kat="Init", sys="BatchBot")
                if sFolder in [None, ""]:
                    res.addErr("Folder not set", add=sBatRow_info, kat="Init", sys="BatchBot")
                if (sWf not in mapper.workflow_getAll()) and (sWf not in mapper.workflow_converter_import_getAll()):
                    res.addErr("Workflow not found", add=sBatRow_info, kat="Init", sys="BatchBot")
                # --- Ordner/Dateien
                if not FileTools.isPathDir(sFolder):
                    res.addErr("Folder not found", add=sBatRow_info, kat="Init", sys="BatchBot")
                if res.isErr():
                    return res
                if lstFiles is not None and len(lstFiles) != 0:
                    for sFile in lstFiles:
                        sPathFile = Path(sFolder, sFile)
                        if not FileTools.isPathFile(sPathFile):
                            res.addErr("File not found", add=sBatRow_info, kat="Init", sys="BatchBot")
                if lstExclude is not None and len(lstExclude) != 0:
                    for sExclude in lstExclude:
                        sPathFile = Path(sFolder, sExclude)
                        if not FileTools.isPathFile(sPathFile):
                            res.addInfo("Exclude not found", add=sBatRow_info, kat="Init", sys="BatchBot")
                # *** Finale Dateiliste
                lstFilesFinal = []
                lstFilesFound = []
                lstFilesFound_excel = []
                if lstExclude is None:
                    lstExclude = []
                if lstFiles is not None:
                    # File Liste ist gegeben
                    for sFile in lstFiles:
                        if sFile not in lstExclude:
                            lstFilesFinal.append(sFile)
                else:
                    # Alle Files in Folder nehmen
                    lstFilesFound = os.listdir(sFolder)
                    for sFile in lstFilesFound:
                        if sFile.lower().endswith(".xlsx"):
                            lstFilesFound_excel.append(sFile)
                    for sFile in lstFilesFound_excel:
                        if sFile not in lstExclude:
                            lstFilesFinal.append(sFile)
                if len(lstFilesFinal) == 0:
                    res.addWarn("Final file list will be empty", add=sBatRow_info, kat="Init", sys="BatchBot")

                # *** Speichern
                lstBatch.append((sSection, sWf, sFolder, lstFiles, lstExclude, lstFilesFound_excel, lstFilesFinal))

        if res.isOk():
            # *** Batch Datei ausführen
            iNr_total = 0
            iNr_total_max = 0
            odicErr_ingest = OrderedDict()
            odicOk_ingest = OrderedDict()
            lstErr_logWrite = []

            # Log
            sLog_overview = ""
            lstOverview = [["Name", "files"]]
            for tpBatch in lstBatch:
                sSection = tpBatch[0]
                sWf = tpBatch[1]
                sFolder = tpBatch[2]
                lstFiles = tpBatch[3]
                lstExclude = tpBatch[4]
                lstFilesFound_excel = tpBatch[5]
                lstFilesFinal = tpBatch[6]
                lstFilesFinal = tpBatch[6]
                iNr_total_max += len(lstFilesFinal)
                lstOverview.append((str(sSection), str(len(lstFilesFinal))))

            # *** Loop Sections
            for tpBatch in lstBatch:
                if not self.isAbort():
                    sSection = tpBatch[0]
                    sWf = tpBatch[1]
                    sFolder = tpBatch[2]
                    lstFiles = tpBatch[3]
                    lstExclude = tpBatch[4]
                    lstFilesFound_excel = tpBatch[5]
                    lstFilesFinal = tpBatch[6]
                    iNr_section_max = len(lstFilesFinal)

                    # Ingest
                    for iNr_section, sFile in enumerate(lstFilesFinal):
                        if not self.isAbort():
                            # Info Progress
                            if iNr_total_max != 0:
                                iInfo_Proz = int(((iNr_total+1) / iNr_total_max) * 100)
                            else:
                                iInfo_Proz = 0
                            sInfo_Nr_total = str(iNr_total+1)+"/"+str(iNr_total_max)
                            sInfo_Nr_section = str(iNr_section+1)+"/"+str(iNr_section_max)
                            #sInfo_Nr = " ("+sInfo_Nr_section+" "+str(iInfo_Proz)+"% "+sInfo_Nr_total+") "
                            sInfo_Nr = " ["+str(iInfo_Proz)+"% "+sInfo_Nr_total+"] "
                            pn("* BatchBot: "+str(sSection) + sInfo_Nr + str(sFile), fmt=False)
                            iNr_total += 1
                            iErr_ingest = 0
                            sErr_ingest = ""
                            for sSection_tmp, lstErr in odicErr_ingest.items():
                                iErr_ingest += len(lstErr)
                            if iErr_ingest > 0:
                                sErr_ingest = " [err "+str(iErr_ingest)+"x]"
                            sBatchInfo = "BatchBot: "+str(sSection) + sInfo_Nr + str(sFile)+ sErr_ingest
                            batch(iNr_total, iNr_total_max, sBatchInfo)

                            # Import: Ingest
                            sPathFile = Path(sFolder, sFile)
                            resIng = self.__model.subsystem_getImporter().startImport(sPathFile, ingest=True, lookup=True, ingest_wf=sWf, ingest_lup_search_single=False, tamer=True)
                            if self.isAbort():
                                res.addInfo("Aborted by user ...", kat="Abort", sys="BatchBot")
                                break
                            # Debug
                            # resIng = Result()
                            # if random.choice([True,False]):
                            #     resIng.addErr("Fehler")
                            # resIng.addData("log", "Log")
                            # time.sleep(0.5)

                            # Logging
                            sLog = resIng.getData("log")
                            checkLogDirs()
                            if resIng.isErr():
                                if sSection not in odicErr_ingest:
                                    odicErr_ingest[sSection] = []
                                odicErr_ingest[sSection].append(sFile)
                                sDir_choose = sDir_log_err
                            else:
                                if sSection not in odicOk_ingest:
                                    odicOk_ingest[sSection] = []
                                odicOk_ingest[sSection].append(sFile)
                                sDir_choose = sDir_log_ok
                            sFile_noType = FileTools.getFileWithoutType(sFile)
                            sFile_log = Path(sDir_choose, sFile_noType + ".txt")
                            FileTools.prepareSaveFile(sFile_log)
                            # Save
                            try:
                                with open(sFile_log, "w") as f:
                                    f.write(str(sLog))
                            except Exception as ex:
                                lstErr_logWrite.append(str(sSection) + sInfo_Nr + str(sFile)+": "+str(ex))
                            # Update Ingest Table
                            self.__model.subsystem_getConnector().status_reload(system="ingest") # Reload ingesttables
                            self.__smartprinter.updateIngestTable()

            # Log
            iErr_ingest = 0
            iErr_ingest_proz = 0
            for sSection, lstErr in odicErr_ingest.items():
                iErr_ingest += len(lstErr)
            if iNr_total_max != 0:
                iErr_ingest_proz = str(int((iErr_ingest / iNr_total_max) * 100))
            iOk_ingest = 0
            iOk_ingest_proz = 0
            for sSection, lstOk in odicOk_ingest.items():
                iOk_ingest += len(lstOk)
            if iNr_total_max != 0:
                iOk_ingest_proz = str(int((iOk_ingest / iNr_total_max) * 100))
            iSekTotal = time.time() - iZeitStart
            iStd = int(iSekTotal // 3600)
            iMin = int((iSekTotal % 3600) // 60)
            iSek = int(iSekTotal % 60)
            sLog = "*** Log: Batch\n"
            if self.isAbort():
                sLog += "\n<Aborted by user>\n\n"
            sLog += "Duration: " + str(iStd) + "h" + str(iMin) + "m" + str(iSek) + "s\n"
            sLog += "Files: " + str(iNr_total_max) + "\n"
            sLog += "Successfull ingestions: " + str(iOk_ingest) + "/"+ str(iNr_total_max) + " " + str(iOk_ingest_proz) +"%\n"
            sLog += "Failed ingestions: " + str(iErr_ingest) + "/"+ str(iNr_total_max) + " " + str(iErr_ingest_proz) +"%\n"
            sLog += "\n"
            if len(lstOverview) > 0:
                sLog += "--- Sections\n"
                sLog += PrettyText.table(lstOverview)
                sLog += "\n\n"
            if iOk_ingest > 0:
                sLog += "--- Successful ingestions (" + str(iOk_ingest) + "x)\n"
                lstTable = [["Section", "files"]]
                for sSection, lstok in odicOk_ingest.items():
                    for sOk in lstok:
                        lstTable.append((sSection, sOk))
                sLog += PrettyText.table(lstTable)
                sLog += "\n\n"
            if iErr_ingest > 0:
                lstTable = [["Section", "files"]]
                for sSection, lstErr in odicErr_ingest.items():
                    for sErr in lstErr:
                        lstTable.append((sSection, sErr))
                sLog += "--- Failed ingestions (" + str(iErr_ingest) + "x)\n"
                sLog += PrettyText.table(lstTable)
                sLog += "\n\n"
            if len(lstErr_logWrite) > 0:
                lstTable = [["Section", "files"]]
                sLog += "--- Failed log writes (" + str(len(lstErr_logWrite)) + "x)\n"
                for sErr_logWrite in lstErr_logWrite:
                    lstTable.append((sSection, sErr_logWrite))
                sLog += PrettyText.table(lstTable)
                sLog += "\n\n"

            # Save
            sFile_batch = Path(sDir_log_timeStamp, "log.txt")
            FileTools.prepareSaveFile(sFile_batch)
            try:
                with open(sFile_batch, "w") as f:
                    f.write(str(sLog))
            except Exception as ex:
                pass

            # Antwort
            res.addData("timestamp", sTimeStamp)
            res.addData("err", odicErr_ingest)
            res.addData("dirlog", sDir_log_timeStamp)
            res.addData("log", sLog)
            res.addData("files", iNr_total_max)

        progress()
        workflow()
        batch()

        return res

