# coding=utf-8

import openpyxl
from collections import Counter, OrderedDict
import os, sys, csv, getpass, platform, datetime, time, copy, re
import openpyxl.styles.numbers
import warnings
from tools.coding import Result, WorkingObject
from tools.text import PrettyText
from tools.parallel import Stoppable
from tools.printing import SmartPrinter
from tools.misc import JsonConverter
from tools.geocruncher import GeoCruncher

# -----------------------------------------
# Director
# -----------------------------------------

class Connector(Stoppable, WorkingObject):

    # Konstruktor
    def __init__(self, model):
        Stoppable.__init__(self)
        WorkingObject.__init__(self)
        self.__model = model
        self.__sp = model.ui_getSmartprinter()

        # Systeme
        self.__odicSystems = OrderedDict([
            ("db", "Database"),
            ("schema", "DbSchema"),
            ("ingest", "Ingestion"),
            ("lookup", "Lookup"),
            ("wizard", "TaxonWizard"),
            ("region", "FAO Regions"),
            ("converter", "Converter"),
            ("actionlog", "ActionLog"),
            ("datapatrol", "DataPatrol")
        ])

        self.__sys_db = self.__model.subsystem_getDb()
        self.__sys_schema =  self.__model.subsystem_getSchema()
        self.__sys_mapper = self.__model.subsystem_getMapper()
        self.__sys_regions = self.__model.subsystem_getRegions()
        self.__sys_converter = self.__model.subsystem_getConverter()

        # Progress
        self.__iProgress_sys_total = 0
        self.__iProgress_sys_now = 0

        # Start/Status
        self.__dicStatus = {}
        self.__dicStart = {}

        # Connection
        self.__bDb_reachable = False
        self.__bSchema_existing = False
        self.__bDm_compatabile = False

        self.__bConnectionFailed = False
        self.__bConnected = False

        # Init
        self.start_reset()
        self.status_reset()

    def __str__(self):
        return self.info()
    # ----

    # Progress Counter
    def progress_reset(self, iNr=None):
        self.__iProgress_sys_now = 1
        if iNr is None:
            iNr = len(self.__odicSystems) * 2
        self.__iProgress_sys_total = iNr
    def progress_next(self):
        self.__iProgress_sys_now += 1
        if self.__iProgress_sys_now > self.__iProgress_sys_now:
            self.__iProgress_sys_now = self.__iProgress_sys_total
    def progress_get(self):
        return self.__iProgress_sys_now, self.__iProgress_sys_total

    # Info
    def info(self):
        sOut = "*** Status of Connector\n\n"

        sOut += "------------------------------------\n"
        sOut += "Overview\n"
        sOut += "------------------------------------\n\n"

        if self.isReady():
            sOut += "All systems are ready ...\n\n"
        else:
            sOut += "Some systems have problems ...\n\n"
        # Systems
        lstTable = [["System", "Start", "Status"]]
        lstProblems = []
        for sSystem_key,sSystem_long in self.__odicSystems.items():
            bOk_start = self.start_isOk(sSystem_key)
            bOk_status = self.status_isOk(sSystem_key)
            lstTable.append([sSystem_long, ("ok" if bOk_start else "err"),("ok" if bOk_status else "err")])
            if not bOk_start:
                resSys = self.start_getLog(sSystem_key)
                lstProblems.append((sSystem_long, resSys))
            if not bOk_status:
                resSys = self.status_getLog(sSystem_key)
                lstProblems.append((sSystem_long, resSys))
        sOut += PrettyText.table(lstTable) + "\n\n"
        # Problems
        if len(lstProblems) > 0:
            sOut += "------------------------------------\n"
            sOut += "Problems\n"
            sOut += "------------------------------------\n\n"
            for sSystem_long, resSys in lstProblems:
                 sOut += "* "+str(sSystem_long) +"\n"
                 sOut += resSys.getErrWarnInfoString(short=False) + "\n"

        #return sOut
        sOut += "------------------------------------\n"
        sOut += "Details\n"
        sOut += "------------------------------------\n\n"

        sOut += "* Connection\n"
        sOut += "Connection failed: " + ("yes" if self.db_isConnectionFailed() else "no") + "\n"
        sOut += "Server reachable: " + ("yes" if self.db_isServerReachable() else "no") + "\n"
        sOut += "Server datamodel compatible: " + ("yes" if self.db_isDmCompatible() else "no") + "\n"
        sOut += "Connected: " + ("yes" if self.db_isConnected() else "no") + "\n"
        return sOut

    # ---

    def isReady(self):
        return self.isOk() and self.db_isConnected()

    def isOk(self):
        return self.start_isOk() and self.status_isOk()

    def db_isServerReachable(self):
        return self.__bDb_reachable

    def db_isSchemaExisting(self):
        return self.__bSchema_existing

    def db_isDmCompatible(self):
        return self.__bDm_compatabile

    def db_isConnectionFailed(self):
        return self.__bConnectionFailed

    def db_isConnected(self):
        return self.__bConnected

    def lookup_isCacheEmpty(self):
        return self.__model.subsystem_getLookup().isCacheEmpty()

    def db_isSchemaOk(self):
        return self.start_getLog("schema").isOk()

    # ---

    def ingest_getIngDepNeed(self):
        return self.status_getData("ingest", "ing_dep_need")

    def ingest_getIngDepBase(self):
        return self.status_getData("ingest", "ing_dep_base")

    # ---

    def disconnect(self):
        self.__bReady = False
        self.__bConnectionFailed = False
        self.__bConnected = False
        self.__model.subsystem_getConverter().setReady_submodule("bsh", False)

    def reconnect(self, **kwargs):
        """
        kwargs:
            start (str/list): Liste der neuzustartenden Systeme
            status (str/list): Liste der neueinzulesenden Status
        """
        lstSystem_restart = kwargs.get("start")
        if type(lstSystem_restart) is str:
            lstSystem_restart = [lstSystem_restart]
        lstSystem_status = kwargs.get("status")
        if type(lstSystem_status) is str:
            lstSystem_status = [lstSystem_status]

        res = Result("(Re-)connect to Server")
        pv = self.__sp.verbose
        workflow = self.__sp.workflow
        progress = self.__sp.progress

        # Progress
        progress("on-pulse")
        self.progress_reset()

        pv("* Reconnect to Server")
        # Systems: Restart
        if res.isOk():
            resRestart = self.start(system=lstSystem_restart)
            if resRestart.isErr():
                res.addResult(resRestart, msg_only=True)

        # Systems: Reload Status
        if res.isOk():
            resStatus = self.status_reload(system=lstSystem_status)
            if resStatus.isErr():
                res.addResult(resStatus, msg_only=True)

        # Status: Connection ok/failed
        if self.isOk():
            self.__bConnected = True
            self.__bConnectionFailed = False
        else:
            self.__bConnected = False
            self.__bConnectionFailed = True

        # Progress
        progress("off")
        return res

    # ---

    # Restart: Reset
    def start_reset(self, sSystem=None):
        def newRestart():
            self.__dicStart[sSystem] = Result("Status (start): " + sSystem)
        if sSystem is not None and sSystem in self.__dicStatus:
            newRestart()
        else:
            self.__bSchema_existing = False
            self.__bDb_reachable = False
            self.__bDm_compatabile = False
            self.__dicStatus = {}
            for sSystem in self.__odicSystems:
                newRestart()

    # Restart
    def start(self, **kwargs):
        """
        kwargs:
            system (str/list): Liste der neuyustartenden Systeme
        """
        pv = self.__sp.verbose
        progress = self.__sp.progress
        lstSystem_subset = kwargs.get("system")
        if type(lstSystem_subset) is str:
            lstSystem_subset = [lstSystem_subset]
        res = Result("Restart")
        if lstSystem_subset is not None:
            lstSystem = lstSystem_subset
        else:
            lstSystem = list(self.__dicStatus)
        pv("* Restarting Subsystems")
        progress("on-pulse")
        for sSys in lstSystem:
            iNr_now, iNr_total = self.progress_get()
            # Progress
            progress(iNr_now, iNr_total, iNr_total, abs=True, err=res.getErrNo(), warn=res.getWarnNo(),info=res.getInfoNo())
            self.progress_next()
            if sSys is not None and sSys in self.__dicStatus:
                resSys = self.__start(sSys)
                if resSys.isErr():
                    res.addResult(resSys)
        return res

    def start_getLog(self, sSystem=None):
        if sSystem is not None and sSystem in self.__dicStatus:
            return self.__dicStart[sSystem]
        else:
            res = Result("Start: All Systems")
            for sSystem in self.__odicSystems:
                resSys = self.__dicStart[sSystem]
                if resSys.isErr():
                    res.addResult(resSys, msg_only=True)
            return res

    def start_isOk(self, sSystem=None):
        if sSystem is not None and sSystem in self.__dicStatus:
            return self.__dicStart[sSystem].isOk()
        else:
            for sSystem in self.__odicSystems:
                if self.__dicStart[sSystem].isErr():
                    return False
            return True

    # ---

    # Status Informationen
    def status_reset(self, sSystem=None):
        def newStatus():
            self.__dicStatus[sSystem] = {}
            self.__dicStatus[sSystem]["log"] = Result("Status (running): " + sSystem)
            self.__dicStatus[sSystem]["data"] = {}
        if sSystem is not None and sSystem in self.__dicStatus:
            newStatus()
        else:
            self.__dicStatus = {}
            for sSystem in self.__odicSystems:
                newStatus()

    def status_isOk(self, sSystem=None):
        if sSystem is not None and sSystem in self.__dicStatus:
            return self.__dicStatus[sSystem]["log"].isOk()
        else:
            for sSystem in self.__odicSystems:
                if self.__dicStatus[sSystem]["log"].isErr():
                    return False
            return True

    def status_getLog(self, sSystem=None):
        if sSystem is not None and sSystem in self.__dicStatus:
            return self.__dicStatus[sSystem]["log"]
        else:
            res = Result("Status: All Systems")
            for sSystem in self.__odicSystems:
                resSys = self.__dicStatus[sSystem]["log"]
                if resSys.isErr():
                    res.addResult(resSys, msg_only=True)
            return res

    def status_getData(self, sSystem, sName=None):
        if (sSystem is not None) and (sSystem in self.__dicStatus) and ("data" in self.__dicStatus[sSystem]):
            if sName is None:
                return self.__dicStatus[sSystem].get("data")
            else:
                return self.__dicStatus[sSystem].get("data").get(sName)
        else:
            return None

    def status_reload(self, **kwargs):
        """
        kwargs:
            system (str/list): Liste der neuyustartenden Systeme
        """
        res = Result("Reload status")
        pv = self.__sp.verbose
        progress = self.__sp.progress
        lstSystem_subset = kwargs.get("system")
        if type(lstSystem_subset) is str:
            lstSystem_subset = [lstSystem_subset]
        if lstSystem_subset is not None:
            lstSystem = lstSystem_subset
        else:
            lstSystem = list(self.__dicStatus)
        pv("* Reloding status of subsystems")
        for sSys in lstSystem:
            # Progress
            iNr_now, iNr_total = self.progress_get()
            self.progress_next()
            progress(iNr_now, iNr_total, iNr_total, abs=True, err=res.getErrNo(), warn=res.getWarnNo(),info=res.getInfoNo())
            if sSys is not None and sSys in self.__dicStatus:
                resSys = self.__status_reload(sSys)
                if resSys.isErr():
                    res.addResult(resSys)
        return res

    def __status_reload(self, sSystem):
        pv = self.__sp.verbose
        progress = self.__sp.progress
        workflow = self.__sp.workflow
        res = Result("Reload status: "+sSystem)
        map = self.__sys_mapper
        sSystem_long =self.__odicSystems.get(sSystem)

        if sSystem == "lookup":
            # *** Lookup
            pv("Status: "+str(sSystem_long))
            workflow(top="Connector", sub="Get status: "+str(sSystem_long), fmt=False)
            for sTable in map.lookup_getDbTables():
                if sTable != "taxon":
                    resLupStatus = self.__model.subsystem_getDb().lookup_getStatus(sTable)
                    if resLupStatus.isErr():
                        res.addResult(resLupStatus, msg_only=True)
                    else:
                        self.__dicStatus[sSystem]["data"][sTable] = resLupStatus
        elif sSystem == "db":
            # *** Server Versionen
            pv("Status: "+str(sSystem_long))
            workflow(top="Connector", sub="Get status: "+str(sSystem_long), fmt=False)
            resServer_pg_version = self.__model.subsystem_getDb().version_getVersionConnectedPostgresServer()
            if resServer_pg_version.isErr():
                res.addResult(resServer_pg_version, msg_only=True)
            else:
                sVersionPgServer = resServer_pg_version.getData("version")
                self.__dicStatus[sSystem]["data"]["pg_version"] = sVersionPgServer
            resServer_dm_version = self.__model.subsystem_getDb().version_getVersionConnectedDatamodel()
            if resServer_dm_version.isErr():
                res.addResult(resServer_dm_version, msg_only=True)
            else:
                sVersionSchema = resServer_dm_version.getData("version")
                self.__dicStatus[sSystem]["data"]["dm_version"] = sVersionSchema
        elif sSystem == "ingest":
            # *** Ingest
            # Tables
            pv("Status: "+str(sSystem_long))
            workflow(top="Connector", sub="Get status: "+str(sSystem_long), fmt=False)
            dicIng_cnt = {}
            for sTable in map.ingest_getDbTables():
                resIng = self.__sys_db.table_getCountEntries(sTable)
                if resIng.isErr():
                    res.addResult(resIng, msg_only=True)
                else:
                    iCount = resIng.getData("count")
                    dicIng_cnt[sTable] = iCount
            self.__dicStatus[sSystem]["data"]["table_counts"] = dicIng_cnt
            # Overview
            resIngOverview = self.__sys_db.ingest_getOverview()
            if resIngOverview.isErr():
                res.addResult(resIngOverview, msg_only=True)
            else:
                lstIngestOverview = resIngOverview.getData("content")
                self.__dicStatus[sSystem]["data"]["overview"] = lstIngestOverview
            # Dependencies
            resIngDep = self.__model.subsystem_getDb().ingest_getIngestDependencies()
            if resIngDep.isErr():
                res.addResult(resIngDep, msg_only=True)
            else:
                dicIngDepBase = resIngDep.getData("base")
                dicIngDepNeed = resIngDep.getData("need")
                self.__dicStatus[sSystem]["data"]["ing_dep_base"] = dicIngDepBase
                self.__dicStatus[sSystem]["data"]["ing_dep_need"] = dicIngDepNeed
        elif sSystem == "wizard":
            # *** Wizard
            pv("Status: "+str(sSystem_long))
            workflow(top="Connector", sub="Get status: "+str(sSystem_long), fmt=False)
            # Admin Table Counts
            dicIng_cnt = {}
            for sTable in map.taxon_getDbTables():
                resIng = self.__sys_db.table_getCountEntries(sTable)
                if resIng.isErr():
                    res.addResult(resIng, msg_only=True)
                else:
                    iCount = resIng.getData("count")
                    dicIng_cnt[sTable] = iCount
            self.__dicStatus[sSystem]["data"]["table_counts"] = dicIng_cnt
            # Base gold
            resTaxBase_gold = self.__model.subsystem_getDb().lookup_getGoldLookups("wormsbase")
            if resTaxBase_gold.isErr():
                res.addResult(resTaxBase_gold, msg_only=True)
            else:
                iCountGold = resTaxBase_gold.getData("count")
                self.__dicStatus[sSystem]["data"]["wormsbase_gold"] = iCountGold
            # Taxon List: Count
            resTax = self.__sys_db.table_getCountEntries("taxon")
            if resTax.isErr():
                res.addResult(resTax, msg_only=True)
            else:
                iCount = resTax.getData("count")
                self.__dicStatus[sSystem]["data"]["taxon_count"] = iCount
            # TaxonWizard: Last Update
            resLastTaxListUpdate = self.__model.subsystem_getDb().taxonwizard_getLastTaxonListUpdateInfo()
            if resLastTaxListUpdate.isErr():
                res.addResult(resLastTaxListUpdate, msg_only=True)
            else:
                self.__dicStatus[sSystem]["data"]["taxonlist_lastupdate"] = resLastTaxListUpdate
        elif sSystem == "datapatrol":
            # *** Data Patrol
            pv("Status: " + str(sSystem_long))
            workflow(top="Connector", sub="Get status: " + str(sSystem_long), fmt=False)
            resLastBiotaControl = self.__model.subsystem_getDb().dataPatrol_getLastBiotaControl()
            if resLastBiotaControl.isErr():
                res.addResult(resLastBiotaControl, msg_only=True)
            else:
                self.__dicStatus[sSystem]["data"]["biotacontrol"] = resLastBiotaControl
        elif sSystem == "actionlog":
            # *** Data Patrol
            pv("Status: " + str(sSystem_long))
            workflow(top="Connector", sub="Get status: " + str(sSystem_long), fmt=False)
            resActionlog = self.__model.subsystem_getDb().actionlog_getOverview()
            if resActionlog.isErr():
                res.addResult(resActionlog, msg_only=True)
            else:
                lstActionlog = resActionlog.getData("content")
                self.__dicStatus[sSystem]["data"]["table"] = lstActionlog

        # *** Log fuer jedes System
        self.__dicStatus[sSystem]["log"] = res
        return res

    # ---

    def __start(self, sSystem):
        pv = self.__sp.verbose
        workflow = self.__sp.workflow
        progress = self.__sp.progress
        res = Result("Status restart: "+sSystem)
        map = self.__sys_mapper

        sSystem_long =self.__odicSystems.get(sSystem)
        if sSystem == "db":
            pv("Restart: " + str(sSystem_long))
            # *** Db
            self.__bConnected = False
            self.__bConnectionFailed = False
            self.__bDb_reachable = False
            self.__bSchema_existing = False
            self.__bDm_compatabile = False
            workflow(top="Connector", sub="Check connection", fmt=False)
            resServer =  self.__model.subsystem_getDb().conn_getConnection()
            bServerReachable = resServer.isOk()
            resSchemaExisting =  self.__model.subsystem_getDb().check_isSchemaExisting()
            bSchemaExisting = resSchemaExisting.getData("exist")
            bSchemaExisting = True if bSchemaExisting is True else False
            resSchemaComp =  self.__model.subsystem_getDb().check_checkServerCompatibility()
            bSchemaComp = resSchemaComp.isOk()
            self.__bDb_reachable = bServerReachable
            self.__bSchema_existing = bSchemaExisting
            self.__bDm_compatabile = bSchemaComp
            res.addResult(resServer)
            res.addResult(resSchemaExisting)
            res.addResult(resSchemaComp)
        if sSystem == "schema":
            # *** Schema
            pv("Restart: " + str(sSystem_long))
            workflow(top="Connector", sub="Get Db structure", fmt=False)
            res = self.__model.conf_getDbSchema().update()
        if sSystem == "converter":
            # *** Converter
            pv("Restart: " + str(sSystem_long))
            workflow(top="Connector", sub="Restart: "+str(sSystem_long), fmt=False)
            res = self.__model.subsystem_getConverter().init()
        if sSystem == "lookup":
            # *** Lookup
            pv("Restart: " + str(sSystem_long))
            workflow(top="Connector", sub="Restart: "+str(sSystem_long), fmt=False)
            res = self.__model.subsystem_getLookup().reset()
        if sSystem == "region":
            # *** FAO Regionen
            pv("Restart: " + str(sSystem_long))
            workflow(top="Connector", sub="Restart: "+str(sSystem_long), fmt=False)
            res = self.__model.subsystem_getRegions().reload()

        # *** Log Start fuer jedes System
        self.__dicStart[sSystem] = res
        return res
