# coding=utf-8

import openpyxl
from collections import Counter, OrderedDict
import os, sys, csv, getpass, platform, datetime, time, copy, re
import openpyxl.styles.numbers
import warnings
from tools.coding import Result, WorkingObject, DataTools
from tools.excel import ExcelTools, ExcelNavigator
from tools.text import PrettyText
from tools.parallel import Stoppable
from tools.printing import SmartPrinter
from tools.misc import JsonConverter
from tools.geocruncher import GeoCruncher
from pathlib import Path

# -------------------------------------------------------------
# Converter: Antarctic
# -------------------------------------------------------------

class ConverterAntarctic(Stoppable, WorkingObject):

    # Konstruktor
    def __init__(self, model):
        Stoppable.__init__(self)
        WorkingObject.__init__(self)

        self.__model = model
        self.__smartprinter = model.ui_getSmartprinter()
        self.__map = model.subsystem_getMapper()

    # Info
    def info(self, bHtml):
        self.__sWf_conv = "ant_mat"
        sWfConfLong = self.__map.workflow_converter_import_short2long(self.__sWf_conv)
        sWfConfTargetShort = self.__map.workflow_converter_import_getTarget(self.__sWf_conv)
        sWfConfTargetLong = self.__map.workflow_short2long(sWfConfTargetShort)
        sOut = "************************************\n"
        sOut += "Import: " + str(sWfConfLong)+" -> "+str(sWfConfTargetLong)+" (CB)\n"
        sOut += "************************************\n\n"
        sOut += "Transpose matrix in column format (CB native)\n"
        return sOut

    # Initialisation
    def init(self):
        res = Result("Antarctic Converter")
        self.setWorking(True)
        return res

    # Reset
    def reset(self):
        pass

    # Converter Antarctic Matrix
    def convert_import_antmat2mat(self, wb, **kwargs):
        """
        Args:
            wb (workbook): Workbook
        Kwargs:
            result_in: Result in
        Return:
            Result
                log (str)
        """

        """
        Args:
            wb (workbook): Workbook
        Kwargs:
            result_in: Result in
        Return:
            Result
                log (str)
        """

        pn = self.__smartprinter.normal
        pv = self.__smartprinter.verbose
        pd = self.__smartprinter.debug

        workflow = self.__smartprinter.workflow
        progress = self.__smartprinter.progress
        status = self.__smartprinter.status

        res = Result(name="Excel Converter Antarctic Matrix")
        self.abort(False)

        # Parameter
        resultIn = kwargs["result_in"] if "result_in" in kwargs else Result("Empty Result Input Dummy")

        # Info
        sWfCnf = "ant_mat"
        sWfCnfLong = self.__model.subsystem_getMapper().workflow_converter_import_short2long(sWfCnf)
        sWfAim = self.__model.subsystem_getMapper().workflow_converter_import_getTarget(sWfCnf)
        sWfAimLong = self.__model.subsystem_getMapper().workflow_short2long(sWfAim)

        workflow(top="Converting", sub="Antarctic Matrix")
        status()
        pn("* Converting Excel Workbook: Antarctic Matrix")
        pn("* Conversion: " + str(sWfCnf)+" ("+str(sWfCnfLong)+") -> "+str(sWfAim)+" ("+str(sWfAimLong)+")", fmt=False)

        # ---

        # Test: Biota Sheet exisitert?
        if wb is None:
            res.addErr("No Excel Workbook found", sys="Converter", kat="Configuration")
            return res
        if "Biota" not in wb.sheetnames:
            res.addErr("Sheet Biota not found", sys="Converter", kat="Configuration")
            return res

        # Definitionen
        lstHeaderBiotaMatrixBaum = ["Cruise", "Station", "Sample", "SubSample", "Subset"] # Indizes der Baum-Header
        lstHeaderBiota = ["Cruise", "Station", "Sample", "SubSample", "Subset", "Biota", "AphiaId", "ScientificName", "Number", "Abundance", "PresenceAbsence","LifeStage"]
        wsBiotaMatrix = wb["Biota"] # Biota Matrix
        wsBiotaMatrix.title = "BiotaMatrix"

        # Test: Header in Biota Sheet
        # Leer
        if len(list(wsBiotaMatrix.rows)) == 0:
            res.addErr("Sheet Biota is empty", sys="Converter", kat="Configuration")
            return res
        # Alle Baum Header vorhanden (Cruise,Station,Sample,Subset)
        lstHeader = [cell.value for cell in list(wsBiotaMatrix.rows)[0] if cell.value is not None]
        lstHeaderMissing = [sCol for sCol in lstHeaderBiotaMatrixBaum if sCol not in lstHeader]
        if len(lstHeaderMissing) > 0:
            res.addErr("Missing Header in Sheet Biota", sheet="Biota", sys="Converter", kat="Configuration", add=PrettyText.lst2strC(lstHeaderMissing))
            return res
        # Keine Taxa in Biota Tabelle
        if len(lstHeader) == 4:
            res.addErr("No Taxa defined", sheet="Biota", sys="Converter", kat="Configuration")
            return res

        # Biota Neu
        # Header
        wsBiota = wb.create_sheet(title="Biota")
        wsBiota.append(lstHeaderBiota)

        # Headers mit Indizes
        dicIdx2col_mat = {n + 1: cell.value for n, cell in enumerate(list(wsBiotaMatrix.rows)[0])}
        dicCol2idx_mat = {cell.value: n + 1 for n, cell in enumerate(list(wsBiotaMatrix.rows)[0])}
        dicIdx2col = {n + 1: cell.value for n, cell in enumerate(list(wsBiota.rows)[0])}
        dicCol2idx = {cell.value: n + 1 for n, cell in enumerate(list(wsBiota.rows)[0])}

        # Extrahiere Aid und Scientifcname aus den Taxa-Headern
        dicIdx2Taxa = {}
        setAidUniq = set()
        for sCol, iIdx in dicCol2idx_mat.items():
            if sCol not in lstHeaderBiotaMatrixBaum and sCol is not None:
                # Praesence/Absence Tag ist in Header -> entfernen -> Flag setzten
                sTaxonType = "count"
                if "[pa]" in sCol:
                    sCol = sCol.replace("[pa]", "")
                    sTaxonType = "pa"
                elif "[a]" in sCol:
                    sCol = sCol.replace("[a]", "")
                    sTaxonType = "abundance"
                elif "[count]" in sCol:
                    sCol = sCol.replace("[count]", "")
                    sTaxonType = "count"
                elif "[" in sCol or "]" in sCol:
                    res.addErr("Unknown taxa number type", sheet="Biota", sys="Converter", kat="Configuration", add=sCol)
                    return res
                else:
                    sTaxonType = "count"
                lstCol = sCol.split()
                sTaxonName = " ".join(lstCol[:-1])
                sTaxonAid = " ".join(lstCol[-1:])
                # Test: AphiaId=Int
                try:
                    iAid = int(sTaxonAid)
                except:
                    res.addErr("Taxon Aid is not valid", sheet="Biota", sys="Converter", kat="Configuration", add="ScientificName: "+str(sTaxonName)+", AphiaId:"+str(sTaxonAid))
                    return res
                # Test: Nichts konnte bestimmt werden
                if sCol == "" or sTaxonName == "" or sTaxonAid == "":
                    res.addErr("Scientificname or AphiaId could not be extracted from taxa header", sheet="Biota", sys="Converter", kat="Configuration",
                               add="ScientificName: "+str(sTaxonName)+", AphiaId:"+str(sTaxonAid))
                    return res
                # Test: Einmaligkeit AphiaId
                if sTaxonAid in setAidUniq:
                    res.addErr("AphiaId is defined multiple times in header", sheet="Biota",
                               sys="Converter", kat="Configuration",
                               add=sTaxonAid)
                    return res
                else:
                    setAidUniq.add(sTaxonAid)
                # Speichern
                dicIdx2Taxa[iIdx] = (sTaxonAid, sTaxonName, sTaxonType)

        # --- Taxa Matrix in Biota-Daten aufbereiten (Entfalte eine Zeile zu n Zeilen)

        # Loop
        iTaxaNumbers = 0
        iIdxCruise_mat = dicCol2idx_mat["Cruise"]-1
        iIdxsStation_mat = dicCol2idx_mat["Station"]-1
        iIdxsSample_mat = dicCol2idx_mat["Sample"]-1
        iIdxsSubSample_mat = dicCol2idx_mat["SubSample"]-1
        iIdxsSubset_mat = dicCol2idx_mat["Subset"]-1
        for i, row in enumerate(wsBiotaMatrix):
            if i > 0:
                sCruise_mat = row[iIdxCruise_mat].value
                sStation_mat = row[iIdxsStation_mat].value
                sSample_mat = row[iIdxsSample_mat].value
                sSubSample_mat = row[iIdxsSubSample_mat].value
                sSubset_mat = row[iIdxsSubset_mat].value
                for iIdCol_mat, (iAid_mat, sScientificName_mat, sNumberType_mat) in dicIdx2Taxa.items():
                    iNumber = row[iIdCol_mat-1].value
                    if iNumber is not None:
                        lstRow = []
                        iTaxaNumbers += 1
                        for iIdx, sCol in dicIdx2col.items():
                            if sCol == "Cruise":
                                lstRow.append(sCruise_mat)
                            elif sCol == "Station":
                                lstRow.append(sStation_mat)
                            elif sCol == "Sample":
                                lstRow.append(sSample_mat)
                            elif sCol == "SubSample":
                                lstRow.append(sSubSample_mat)
                            elif sCol == "Subset":
                                lstRow.append(sSubset_mat)
                            elif sCol == "Biota":
                                lstRow.append(iTaxaNumbers) # Tweak: Vergib eine aufsteigende BiotaID
                            elif sCol == "AphiaId":
                                lstRow.append(str(iAid_mat))
                            elif sCol == "ScientificName":
                                lstRow.append(str(sScientificName_mat))
                            elif sCol == "Number":
                                if sNumberType_mat == "count":
                                    lstRow.append(str(iNumber))
                                else:
                                    lstRow.append(None)
                            elif sCol == "Abundance":
                                if sNumberType_mat == "abundance":
                                    lstRow.append(str(iNumber))
                                else:
                                    lstRow.append(None)
                            elif sCol == "PresenceAbsence":
                                if sNumberType_mat == "pa":
                                    lstRow.append(str(iNumber))
                                else:
                                    lstRow.append(None)
                            elif sCol == "LifeStage":
                                lstRow.append(None)
                        wsBiota.append(lstRow)

        # Entferne Biota Matrix
        wb.remove(wsBiotaMatrix)

        # Debug Save Conversion
        # wb.save(filename="/Users/paul/_temp/huhu.xlsx")

        # Log
        iAnzahlTaxaPerRow = len(dicIdx2Taxa)
        try:
            iMaxRowNewEstimated = (wsBiotaMatrix.max_row - 1) * iAnzahlTaxaPerRow
        except:
            iMaxRowNewEstimated = 0  # Backupwert
        lstTaxa = [[str(sName), str(iAid), sTaxonType] for iIdx, (iAid, sName, sTaxonType) in dicIdx2Taxa.items()]
        sLog = ""
        sLog += "************************************\n"
        sLog += "Converter: Antarctic Matrix\n"
        sLog += str(sWfCnf) + " (" + str(sWfCnfLong) + ") -> " + str(sWfAim) + " (" + str(sWfAimLong) + ")\n"
        sLog += ""
        sLog += "************************************\n\n"
        sLog += "--- Processing Taxa Matrix Biota ---\n"
        sLog += "Number of rows in Biota: " + str(wsBiotaMatrix.max_row - 1) + "\n"
        sLog += "Coded taxa per row: " + str(iAnzahlTaxaPerRow) + "\n"
        sLog += "Taxa total: " + str(iMaxRowNewEstimated) + "\n"
        sLog += "Extracted taxa numbers (count/pa/abundance): " + str(iTaxaNumbers) + "\n"
        sLog += "List of taxa:\n"
        sLog += PrettyText.table([["Scientificname", "AphiaID", "Type"]] + lstTaxa) + "\n"
        sLog += "\n"

        # Debug: Print Biota Conversion
        if self.__model.conf_isDebug():
            sLog += "--- Converted Taxa Matrix Biota ---\n"
            lstTable = []
            for row in wsBiota:
                lstRow = []
                for col in row:
                    lstRow.append(col.value)
                lstTable.append(lstRow)
            sLog += PrettyText.table(lstTable, True)
            sLog += "\n\n"

        # Ergebnis
        res.addData("log", sLog)
        return res

    # Temnplate Antarctic Matrix
    def template_import_antmat2mat(self, sFile):

        res = Result("Write Converter Template")

        # TODO: Tweak: Update Comments
        self.__model.conf_getDbSchema().update()

        # Baue Workbook mit Lookups
        resWb = self.__model.subsystem_getExporter().writeExcel(lookup=True)
        if resWb.isErr():
            res.addResult(resWb)
            return res
        wb = resWb.getData("wb")
        wb.remove(wb["Glossary"])

        # Erweitere um Ingestdaten
        odicIngest = OrderedDict([
            ("Biota", ["Cruise", "Station", "Sample", "Subset", "Scientificname AphiaId [pa/abundance/count]", "Scientificname AphiaId [pa/abundance/count]", "..."]),
            ("Subset", ["Cruise", "Station", "Sample", "Subset", "Project",	"TaxonomicCoverage", "TaxonomicResolution",	"SpatialLevel"]),
            ("Sample", ["Cruise", "Station", "Sample", "Gear", "LatStart", "LonStart", "DateStart",	"DateStartError",	"LatEnd", "LonEnd", "DepthStart", "DepthEnd",	"SampledArea", "ReferenceArea",	"SampledVolume", "ReferenceVolume", "Replicates", "SamplingDistance", "SamplingSpeed"]),
            ("Station", ["Cruise", "Station"]),
            ("Cruise", ["Cruise", "RefSystem"])
        ])
        for sSheet, lstCols in odicIngest.items():
            ws = wb.create_sheet(sSheet, 0)
            lstCells = []
            for sCol in lstCols:
                resComment = self.__model.conf_getDbSchema().getDbSchemaInfo("ingest", sSheet, sCol)
                # Normal
                cell = openpyxl.cell.WriteOnlyCell(ws, value=sCol)
                # Format
                bObl = self.__map.ingest_isObl(sSheet, sCol)
                bPk = self.__map.ingest_isPk(sSheet, sCol)
                bFk = self.__map.ingest_isFk(sSheet, sCol)
                bBold = False
                bItalic = False
                bUnderline = False
                if bPk:
                    bBold = True
                if bFk:
                    bBold = True
                    bItalic = True
                if bObl and not (bPk or bFk):
                    bUnderline = True
                cell.font = openpyxl.styles.Font(italic=bItalic, bold=bBold, underline=("single" if bUnderline else None))
                if resComment.isOk():
                    sComment = resComment.getData("comment-full")
                    if sComment not in (None, ""):
                        cell.comment = openpyxl.comments.Comment(text=sComment, author=None)
                lstCells.append(cell)
            ws.append(lstCells)

        # Save Workbook
        try:
            wb.save(filename=sFile)
        except Exception as ex:
            res.addErr("Writing Converter Template", add=str(ex), kat="I/O", sys="Converter")
            return res

        # Ergebnis
        res.addData("title", "Converter Template: Antarctic (Matrix)")
        res.addData("text", "Template successfully written to file:\n"+str(sFile))
        return res
