# coding=utf-8
import traceback

import openpyxl
from collections import Counter, OrderedDict
import os, sys, csv, getpass, platform, datetime, time, copy, re
import openpyxl.styles.numbers
import warnings
from tools.coding import Result, WorkingObject, DataTools
from tools.excel import ExcelTools, ExcelNavigator
from tools.text import PrettyText
from tools.fileio import FileTools
from tools.ui import Colors
from tools.parallel import Stoppable
from tools.printing import SmartPrinter
from tools.misc import JsonConverter
from tools.geocruncher import GeoCruncher
from pathlib import Path

# -------------------------------------------------------------
# Converter: BSH
# -------------------------------------------------------------

"""
Infos:
CB <-import- T -import-> M
CB <-export- T -export-> M

T 
- Sheet/Col
- Datentyp
- Unit
- Fragen

M/CB Ziele (obligat bei direktem Im-/Export)
- Sheet/Col: CB Import
- DB: CB export, M import (table), M export (view)

CB/M Informationen
- Import/Export-Info: Erklärungen für Import (obligat bei Tweaked Import)
- Info Additional (Intern): Erklärungen für internes Vorgehen (immer fakultativ) nur zum Verständnis

Regeln
- Alle Anfragen an Mapper beziehen sich immer auf T
- Es gibt keine Informationen zu einer Seite (M/CB), wenn kein Bezug zu T besteht (also kein Import/Export möglich)
- Direkter Im-Export: Ok, Explizite Angabe von M/CB-Zielen, fakultative Import/Export-Info
- Indirekter (tweaked) Im-Export: Ok, Tweak, obligate Import/Export-Info, Fakultative Angabe von M/CB-Zielen    
- Import: Ein direktes Ziel darf nicht mehrfach für das Transportformat benutzt werden (BSP: Cruise.Cruise wird nur einmal geschrieben)
- Export: Eine direkte Quelle darf mehrfach für das Transportformat benutzt werden (Cruise.Cruise wird allerdings mehrfach im Export benutzt -> z.B. Station.Cruise, Sample.Crtuise, ...)

"""

class ConverterBsh(Stoppable, WorkingObject):

    # Konstruktor
    def __init__(self, model):
        Stoppable.__init__(self)
        WorkingObject.__init__(self)

        self.setReady(False)

        self.__model = model
        self.__smartprinter = model.ui_getSmartprinter()
        self.__map = model.subsystem_getMapper()
        self.__sWf_conv = "mar_trans"
        self.__sWf_target = self.__map.workflow_converter_import_getTarget(self.__sWf_conv)

        self.__lstSheet_counterParts = [
            ("Cruise", "SURVEY"),
            ("Station", "STATION"),
            ("Sample", "HAUL"),
            ("SubSample", None),
            ("Subset", None),
            ("Biota", None),
            ("Sediment", None)
        ]

        self.__odicDef_trans_opts_allowed = OrderedDict([
            ("obl", "obligate"),
            ("cobl", "conditional obligate"),
            ("lat", "Latitude"),
            ("lon", "Longitude"),
            ("date", "Date"),
            ("time", "Time")
        ])
        self.__odicDef_trans_type_M2CB = {
            "-": None,
            "str": "text",
            "number": "float"
        }
        self.__odicDef_trans_unit_M2CB = {
            "-": None,
            "graddezimal": "dD",
            "qm": "m**2"
        }

    # Initialisation
    def init(self):

        # Helper
        def clean(sTxt):
            if sTxt is None:
                return None
            sTxt = str(sTxt)
            sTxt_clean = sTxt.strip()
            if sTxt != sTxt_clean:
                if res is not None and iRow is not None:
                    res.addInfo("Whitespace at start and end found and removed", add=str(sTxt), kat="Initialisation: Cleaning Definiton Data", sys="Converter: MARLIN",sheet="Def", row=iRow)
                else:
                    print("Converter: MARLIN: Cleaning Definiton Data: Whitespace at start/end removed: "+str(sTxt))
            return sTxt_clean

        def isSet(sA):
            if sA not in [None, ""]:
                return True
            else:
                return False

        def isSet_both(sA, sB):
            if isSet(sA) and isSet(sB):
                return True
            else:
                return False

        def isSetOrNone_both(sA, sB):
            if (sA in [None, ""]) and (sB in [None, ""]):
                return True
            elif (sA not in [None, ""]) and (sB not in [None, ""]):
                return True
            else:
                return False

        # ---

        res = Result("Mapping MARLIN - CB")
        sMap = os.path.dirname(os.path.realpath(__file__)) + "/mapper_convert_bsh.csv"

        pn = self.__smartprinter.normal
        pv = self.__smartprinter.verbose

        # Trans
        self.__odicTrans_sheet2cols = OrderedDict()
        self.__odicTrans_opt = OrderedDict()
        self.__odicTrans_unit = OrderedDict()
        self.__odicTrans_type = OrderedDict()
        self.__odicTrans_notes = OrderedDict()
        self.__odicTrans_question = OrderedDict()
        self.__setTrans_opt = OrderedDict()

        # MARLIN
        self.__setM_import_tweak = set()
        self.__odicM_import_dbTable = OrderedDict()
        self.__lstM_import_singles = []
        self.__odicM_import_notes_general = OrderedDict()
        self.__odicM_import_notes_process = OrderedDict()
        self.__odicM_import_notes_intenal = OrderedDict()

        self.__setM_export_tweak = set()
        self.__odicM_export_dbView = OrderedDict()
        self.__lstM_export_singles = []
        self.__odicM_export_notes_general = OrderedDict()
        self.__odicM_export_notes_process = OrderedDict()
        self.__odicM_export_notes_internal = OrderedDict()

        # CRITTERBASE
        self.__lstCB_import_all = []
        self.__setCB_import_tweak = set()
        self.__odicCB_import_excel = OrderedDict()
        self.__odicCB_import_notes_process = OrderedDict()
        self.__odicCB_import_notes_internal = OrderedDict()
        self.__setCB_import_singles = set()
        self.__odicCB_import_incomp = OrderedDict()

        self.__lstCB_export_all = []
        self.__setCB_export_tweak = set()
        self.__odicCB_export_excel = OrderedDict()
        self.__odicCB_export_notes_process = OrderedDict()
        self.__odicCB_export_notes_internal = OrderedDict()
        self.__setCB_export_singles = set()
        self.__odicCB_export_incomp = OrderedDict()

        # Header der Configdatei
        self.__odicHeaderSectionsMTCB = OrderedDict()
        self.__odicHeaderSectionsKat = OrderedDict()
        self.__odicHeader = OrderedDict()

        # Test
        setTrans_excel_uniq = set()
        dicCB_import_excel_uniq = {}
        dicCB_export_excel_uniq = {}
        dicMarlin_import_db_uniq = {}
        dicMarlin_export_sv_uniq = {}

        try:
            # Einlesen der CSV-Datei
            with open(sMap, newline='', encoding='utf-8') as csvfile:
                csvReader = csv.reader(csvfile, delimiter=',', quotechar='"')
                # Zeilen
                for iRow, lstZeile in enumerate(csvReader):
                    iRow += 1 # Index: 0 -> 1
                    if iRow == 1:
                        # Kommentare zu Zeilen
                        for iNr, sHead in enumerate(lstZeile):
                            if sHead != "":
                                self.__odicHeaderSectionsMTCB[iNr] = sHead
                    elif iRow == 2:
                        # Kommentare zu Zeilen
                        for iNr, sHead in enumerate(lstZeile):
                            if sHead != "":
                                self.__odicHeaderSectionsKat[iNr] = sHead
                    elif iRow == 3:
                        # Kommentare zu Zeilen
                        for iNr, sHead in enumerate(lstZeile):
                            if sHead != "":
                                self.__odicHeader[iNr] = sHead
                    elif iRow > 3:
                        sTrans_ExcelSheet = clean(lstZeile[0])
                        sTrans_ExcelCol = clean(lstZeile[1])
                        sTrans_Opt = clean(lstZeile[2])
                        sTrans_Type = clean(lstZeile[3])
                        sTrans_Unit = clean(lstZeile[4])
                        sTrans_Notes = clean(lstZeile[5])
                        sTrans_Question = clean(lstZeile[6])

                        sCB_Import_Tweak = clean(lstZeile[19])
                        sCB_Import_ExcelSheet = clean(lstZeile[20])
                        sCB_Import_ExcelCol = clean(lstZeile[21])
                        sCB_Import_Info = clean(lstZeile[22])
                        sCB_Import_Info_add = clean(lstZeile[23])

                        sCB_Export_Tweak = clean(lstZeile[24])
                        sCB_Export_ExcelSheet = clean(lstZeile[25])
                        sCB_Export_ExcelCol = clean(lstZeile[26])
                        sCB_Export_Notes_process = clean(lstZeile[27])
                        sCB_Export_Notes_internal = clean(lstZeile[28])

                        sMarlin_Import_Tweak = clean(lstZeile[7])
                        sMarlin_Import_Table = clean(lstZeile[8])
                        sMarlin_Import_Attrib = clean(lstZeile[9])
                        sMarlin_Import_Notes_general = clean(lstZeile[10])
                        sMarlin_Import_Notes_process = clean(lstZeile[11])
                        sMarlin_Import_Notes_internal = clean(lstZeile[12])

                        sMarlin_Export_Tweak = clean(lstZeile[13])
                        sMarlin_Export_SvView = clean(lstZeile[14])
                        sMarlin_Export_SvAttrib = clean(lstZeile[15])
                        sMarlin_Export_Notes_general = clean(lstZeile[16])
                        sMarlin_Export_Notes_process = clean(lstZeile[17])
                        sMarlin_Export_Notes_internal = clean(lstZeile[18])

                        # --------------------------------------------------------------------------------
                        # Transport
                        # --------------------------------------------------------------------------------

                        # --- Konsistenz
                        # T.Sheet.Col: muss einmalig sein
                        if isSet_both(sTrans_ExcelSheet,sTrans_ExcelCol):
                            if (sTrans_ExcelSheet, sTrans_ExcelCol) not in setTrans_excel_uniq:
                                setTrans_excel_uniq.add((sTrans_ExcelSheet, sTrans_ExcelCol))
                            else:
                                res.addErr("Transport", add="Sheet.Col is double: "+str(sTrans_ExcelSheet) + "." + str(sTrans_ExcelCol),
                                           kat="Initialisation",
                                           sys="Converter: MARLIN", sheet="Def", row=iRow)

                        # T.Sheet.Col: muss immer zusammen gesetzt oder eben null sein
                        if not isSetOrNone_both(sTrans_ExcelSheet,sTrans_ExcelCol):
                            res.addErr("Transport", add="Sheet.Col has to be set together or not at all", kat="Initialisation",
                                       sys="Converter: MARLIN", sheet="Def", row=iRow)

                        # T.Sheet -> T.Cols
                        if isSet(sTrans_ExcelSheet):
                            if sTrans_ExcelSheet not in self.__odicTrans_sheet2cols:
                                self.__odicTrans_sheet2cols[sTrans_ExcelSheet] = []
                            if sTrans_ExcelCol not in self.__odicTrans_sheet2cols[sTrans_ExcelSheet]:
                                self.__odicTrans_sheet2cols[sTrans_ExcelSheet].append(sTrans_ExcelCol)
                            else:
                                res.addErr("Transport", add="Double Col in Sheet", kat="Initialisation", sys="Converter: MARLIN", sheet="Def", row=iRow)

                        # Attribute Inhalte
                        # Options
                        if isSet(sTrans_Opt):
                            if isSetOrNone_both(sTrans_ExcelSheet,sTrans_ExcelCol):
                                lstOpts = sTrans_Opt.split()
                                # Check
                                for sOpt in lstOpts:
                                    if sOpt not in self.trans_getOpts_allowed():
                                        res.addErr("Transport",
                                                   add="Option not allowed: " +str(sTrans_ExcelSheet) + "." + str(sTrans_ExcelCol) +" -> "+ str(sOpt),
                                                   kat="Initialisation",
                                                   sys="Converter: MARLIN", sheet="Def", row=iRow)
                                # Save
                                self.__setTrans_opt[(sTrans_ExcelSheet, sTrans_ExcelCol)] = lstOpts
                            else:
                                res.addErr("Transport", add="Options set with out Sheet.Column: "+str(sTrans_Type),
                                           kat="Initialisation",
                                           sys="Converter: MARLIN", sheet="Def", row=iRow)

                        # Datentyp
                        if not isSet(sTrans_Type) and not isSet(sTrans_ExcelSheet) and not isSet(sTrans_ExcelCol):
                            # Alle drei nicht-gesetzt - > OK -> nichts machen
                            pass
                        elif isSet(sTrans_Type) and isSet_both(sTrans_ExcelSheet,sTrans_ExcelCol):
                            # Alle drei gesetzt -> OK -> merken
                            self.__odicTrans_type[(sTrans_ExcelSheet, sTrans_ExcelCol)] = sTrans_Type
                        else:
                            # Rest -> Fehler
                            res.addErr("Transport", add="Sheet.Column and Type have to be set together: " + str(sTrans_ExcelSheet) + "." + str(sTrans_ExcelCol) + " (type:" + str(sTrans_Type)+")",
                                       kat="Initialisation",
                                       sys="Converter: MARLIN", sheet="Def", row=iRow)

                        # Unit
                        if isSet(sTrans_Unit):
                            if isSetOrNone_both(sTrans_ExcelSheet,sTrans_ExcelCol):
                                self.__odicTrans_unit[(sTrans_ExcelSheet, sTrans_ExcelCol)] = sTrans_Unit
                            else:
                                res.addErr("Transport",
                                           add="Unit set without Sheet.Column: " + str(sTrans_Type),
                                           kat="Initialisation",
                                           sys="Converter: MARLIN", sheet="Def", row=iRow)
                        # Notes: Sheet.Col
                        if isSet(sTrans_Notes):
                            if isSetOrNone_both(sTrans_ExcelSheet, sTrans_ExcelCol):
                                self.__odicTrans_notes[
                                    (sTrans_ExcelSheet, sTrans_ExcelCol)] = sTrans_Notes
                            else:
                                res.addErr("Transport",
                                           add="Comment set with out Sheet.Column: " + str(
                                               sTrans_Notes),
                                           kat="Initialisation",
                                           sys="Converter: MARLIN", sheet="Def", row=iRow)

                        # Questions
                        if isSet(sTrans_Question):
                            if isSetOrNone_both(sTrans_ExcelSheet,sTrans_ExcelCol):
                                self.__odicTrans_question[(sTrans_ExcelSheet, sTrans_ExcelCol)] = sTrans_Question
                            else:
                                res.addErr("Transport",
                                           add="Open Question set with out Sheet.Column: " + str(sTrans_Question),
                                           kat="Initialisation",
                                           sys="Converter: MARLIN", sheet="Def", row=iRow)

                        # Vorbereitung:
                        # Einmaligkeit von direkten Import/Export Feldern zu Transport Sheet
                        if isSet(sTrans_ExcelSheet):
                            if sTrans_ExcelSheet not in dicCB_import_excel_uniq:
                                dicCB_import_excel_uniq[sTrans_ExcelSheet] = set()
                            if sTrans_ExcelSheet not in dicCB_export_excel_uniq:
                                dicCB_export_excel_uniq[sTrans_ExcelSheet] = set()
                            if sTrans_ExcelSheet not in dicMarlin_import_db_uniq:
                                dicMarlin_import_db_uniq[sTrans_ExcelSheet] = set()
                            if sTrans_ExcelSheet not in dicMarlin_export_sv_uniq:
                                dicMarlin_export_sv_uniq[sTrans_ExcelSheet] = set()

                        # --------------------------------------------------------------------------------
                        # CRITTERBASE
                        # --------------------------------------------------------------------------------

                        # --- Vorarbeit:

                        # CB Sheet.Col: Wenn * hole die Daten von Gegenseite
                        if (sCB_Import_ExcelSheet == "*" and (sCB_Import_ExcelCol != "*" or sCB_Export_ExcelSheet in ["*", None, ""])) or \
                            (sCB_Import_ExcelCol == "*" and (sCB_Import_ExcelSheet != "*" or sCB_Export_ExcelCol in ["*", None, ""])) or \
                            (sCB_Export_ExcelSheet == "*" and (sCB_Export_ExcelCol != "*" or sCB_Import_ExcelSheet in ["*", None, ""])) or \
                            (sCB_Export_ExcelCol == "*" and (sCB_Export_ExcelSheet != "*" or sCB_Import_ExcelCol in ["*", None, ""])) or \
                            ((sCB_Import_ExcelSheet, sCB_Import_ExcelCol) == ("*","*") and (sCB_Export_ExcelSheet, sCB_Export_ExcelCol) == ("*","*")):
                                res.addErr("CB import/export: Sheet/Col missconfiguration:",
                                           add= "Import {"+str(sCB_Import_ExcelSheet)+"."+str(sCB_Import_ExcelCol)+"} Export {" +str(sCB_Export_ExcelSheet)+"."+str(sCB_Export_ExcelCol)+"}",
                                           kat="Initialisation",
                                           sys="Converter: MARLIN", sheet="Def", row=iRow)
                        if sCB_Import_ExcelSheet == "*":
                            sCB_Import_ExcelSheet = sCB_Export_ExcelSheet
                        if sCB_Import_ExcelCol == "*":
                            sCB_Import_ExcelCol = sCB_Export_ExcelCol
                        if sCB_Export_ExcelSheet == "*":
                            sCB_Export_ExcelSheet = sCB_Import_ExcelSheet
                        if sCB_Export_ExcelCol == "*":
                            sCB_Export_ExcelCol = sCB_Import_ExcelCol

                        # --- CB Import

                        # CB.Sheet.Col: Entweder beide NONE oder beide gesetzt
                        if not isSetOrNone_both(sCB_Import_ExcelSheet, sCB_Import_ExcelCol):
                            res.addErr("CB import: CB Sheet.Col has to be either set or unset both",
                                       add=str(sCB_Import_ExcelSheet) + "." + str(sCB_Import_ExcelCol), kat="Initialisation",
                                       sys="Converter: MARLIN", sheet="Def", row=iRow)

                        # CB.Sheet.Col: Muss einmalig sein sein
                        if isSet_both(sCB_Import_ExcelSheet, sCB_Import_ExcelCol) and isSet(sTrans_ExcelSheet):
                            if (sCB_Import_ExcelSheet, sCB_Import_ExcelCol) not in dicCB_import_excel_uniq[sTrans_ExcelSheet]:
                                dicCB_import_excel_uniq[sTrans_ExcelSheet].add((sCB_Import_ExcelSheet, sCB_Import_ExcelCol))
                            else:
                                res.addErr("CB import: CB Sheet.Col is double",
                                           add= str(sTrans_ExcelSheet) + ": "+str(sCB_Import_ExcelSheet) + "." + str(sCB_Import_ExcelCol), kat="Initialisation",
                                           sys="Converter: MARLIN", sheet="Def", row=iRow)

                        # CB.Sheet.Col: Muss in CB im Ingest/Lookup System exisiteren
                        if isSet_both(sCB_Import_ExcelSheet, sCB_Import_ExcelCol):
                            if not self.__map.all_isSheetColExisting(sCB_Import_ExcelSheet,sCB_Import_ExcelCol):
                                res.addErr("CB import: CB Sheet.Col is not known",
                                           add= str(sCB_Import_ExcelSheet) + "." + str(sCB_Import_ExcelCol),
                                           kat="Initialisation",
                                           sys="Converter: MARLIN", sheet="Def", row=iRow)
                            else:
                                if self.__map.all_getSheetSystem(sCB_Import_ExcelSheet) not in ["ingest","lookup"]:
                                    res.addErr(
                                        "CB import: CB Col.Sheet is not part of Lookup or Ingest system",
                                        add=str(sCB_Import_ExcelSheet) + "." + str(
                                            sCB_Import_ExcelCol) + " -> " + str(
                                            self.__map.all_getSheetSystem(sCB_Import_ExcelSheet)) + " system",
                                        kat="Initialisation",
                                        sys="Converter: MARLIN", sheet="Def", row=iRow)

                        # T.Sheet.Col -> CB.Sheet.Col
                        if self.trans_isSheetColExisting(sTrans_ExcelSheet, sTrans_ExcelCol) and isSet_both(sCB_Import_ExcelSheet, sCB_Import_ExcelCol):
                            # CB.Sheet.Col: einmalige (als Import Ziel)
                            if (sCB_Import_ExcelSheet, sCB_Import_ExcelCol) not in self.__odicCB_import_excel.values():
                                self.__odicCB_import_excel[(sTrans_ExcelSheet, sTrans_ExcelCol)] = (sCB_Import_ExcelSheet, sCB_Import_ExcelCol)
                            else:
                                res.addErr("CB import: CB.Sheet.Col is used more than once as Import Target by Trans.Sheet.Col",
                                           add=str(sCB_Import_ExcelSheet) + "." + str(sCB_Import_ExcelCol),
                                           kat="Initialisation",
                                           sys="Converter: MARLIN", sheet="Def", row=iRow)


                        # Tweak (stumpf merken)
                        if isSet(sCB_Import_Tweak):
                            if self.trans_isSheetColExisting(sTrans_ExcelSheet, sTrans_ExcelCol):
                                self.__setCB_import_tweak.add((sTrans_ExcelSheet, sTrans_ExcelCol))
                            else:
                                res.addErr("CB import: Tweaked: Trans.Sheet.Col not found",
                                           add=str(sTrans_ExcelSheet) + "." + str(sTrans_ExcelCol),
                                           kat="Initialisation",
                                           sys="Converter: MARLIN", sheet="Def", row=iRow)

                        # Notes Process (stumpf merken)
                        if isSet(sCB_Import_Info):
                            if self.trans_isSheetColExisting(sTrans_ExcelSheet, sTrans_ExcelCol):
                                self.__odicCB_import_notes_process[
                                    (sTrans_ExcelSheet, sTrans_ExcelCol)] = sCB_Import_Info
                            else:
                                res.addErr("CB import: Import info: Trans.Sheet.Col not set",
                                           add=str(sTrans_ExcelSheet) + "." + str(sTrans_ExcelCol),
                                           kat="Initialisation", sys="Converter: MARLIN", sheet="Def", row=iRow)

                        # Notes Internal (stumpf merken)
                        if isSet(sCB_Import_Info_add):
                            if self.trans_isSheetColExisting(sTrans_ExcelSheet, sTrans_ExcelCol):
                                self.__odicCB_import_notes_internal[(sTrans_ExcelSheet, sTrans_ExcelCol)] = sCB_Import_Info_add
                            else:
                                res.addErr("CB import: Additional info: Trans.Sheet.Col not set", add=str(sTrans_ExcelSheet) + "." + str(sTrans_ExcelCol),
                                        kat="Initialisation", sys="Converter: MARLIN", sheet="Def", row=iRow)

                        # Check: Import Ok + Tweak: Import Notes Process muss existieren
                        if self.cb_import_isTweaked(sTrans_ExcelSheet, sTrans_ExcelCol) and not self.cb_import_hasNotes_process(sTrans_ExcelSheet, sTrans_ExcelCol):
                            res.addErr("CB import: marked as Tweak: but has no import info",
                                           add=str(sTrans_ExcelSheet) + "." + str(sTrans_ExcelCol),
                                           kat="Initialisation", sys="Converter: MARLIN", sheet="Def", row=iRow)

                        # Leere Transport Felder (es gibt keinen Import für diese CB Felder)
                        if isSet_both(sCB_Import_ExcelSheet, sCB_Import_ExcelCol) and not self.trans_isSheetColExisting(sTrans_ExcelSheet, sTrans_ExcelCol):
                            self.__setCB_import_singles.add((sCB_Import_ExcelSheet, sCB_Import_ExcelCol))

                        # Alle Felder nochmal separat merken
                        self.__lstCB_import_all.append((sCB_Import_ExcelSheet, sCB_Import_ExcelCol))

                        # --- CB Export

                        # CB.Sheet.Col: Entweder beide NONE oder beide gesetzt
                        if not isSetOrNone_both(sCB_Export_ExcelSheet, sCB_Export_ExcelCol):
                            res.addErr("CB export: Sheet.Col has to be either set or unset both",
                                       add=str(sCB_Export_ExcelSheet) + "." + str(sCB_Export_ExcelCol),
                                       kat="Initialisation",
                                       sys="Converter: MARLIN", sheet="Def", row=iRow)

                        # CB.Sheet.Col: Muss einmalig sein sein
                        if isSet_both(sCB_Export_ExcelSheet, sCB_Export_ExcelCol) and isSet(sTrans_ExcelSheet):
                            if (sCB_Export_ExcelSheet, sCB_Export_ExcelCol) not in dicCB_export_excel_uniq[sTrans_ExcelSheet]:
                                dicCB_export_excel_uniq[sTrans_ExcelSheet].add((sCB_Export_ExcelSheet, sCB_Export_ExcelCol))
                            else:
                                res.addErr("CB export: CB Sheet.Col is double",
                                           add=str(sTrans_ExcelSheet) + ":" + str(sCB_Export_ExcelSheet) + "." + str(sCB_Export_ExcelCol),
                                           kat="Initialisation",
                                           sys="Converter: MARLIN", sheet="Def", row=iRow)

                        # CB.Sheet.Col: Muss in CB im Ingest/Lookup System exisiteren
                        if isSet_both(sCB_Export_ExcelSheet, sCB_Export_ExcelCol):
                            if not self.__map.all_isSheetColExisting(sCB_Export_ExcelSheet, sCB_Export_ExcelCol):
                                res.addErr("CB export: CB Sheet.Col is not known",
                                           add=str(sCB_Export_ExcelSheet) + "." + str(sCB_Export_ExcelCol),
                                           kat="Initialisation",
                                           sys="Converter: MARLIN", sheet="Def", row=iRow)
                            else:
                                if self.__map.all_getSheetSystem(sCB_Export_ExcelSheet) not in ["ingest","lookup"]:
                                    res.addErr(
                                        "CB export: CB Sheet.Col is not part of Lookup or Ingest system",
                                        add=str(sCB_Export_ExcelSheet) + "." + str(
                                            sCB_Export_ExcelCol) + " -> " + str(
                                            self.__map.all_getSheetSystem(sCB_Export_ExcelCol)) + " system",
                                        kat="Initialisation",
                                        sys="Converter: MARLIN", sheet="Def", row=iRow)

                        # T.Sheet.Col -> CB.Sheet.Col
                        if self.trans_isSheetColExisting(sTrans_ExcelSheet, sTrans_ExcelCol) and isSet_both(
                                sCB_Export_ExcelSheet, sCB_Export_ExcelCol):
                            self.__odicCB_export_excel[(sTrans_ExcelSheet, sTrans_ExcelCol)] = (sCB_Export_ExcelSheet, sCB_Export_ExcelCol)

                        # Tweak (stumpf merken)
                        if isSet(sCB_Export_Tweak):
                            if self.trans_isSheetColExisting(sTrans_ExcelSheet, sTrans_ExcelCol):
                                self.__setCB_export_tweak.add((sTrans_ExcelSheet, sTrans_ExcelCol))
                            else:
                                res.addErr("CB export: Tweaked: Trans.Sheet.Col not found",
                                           add=str(sTrans_ExcelSheet) + "." + str(sTrans_ExcelCol),
                                           kat="Initialisation",
                                           sys="Converter: MARLIN", sheet="Def", row=iRow)

                        # Notes Process (stumpf merken)
                        if isSet(sCB_Export_Notes_process):
                            if self.trans_isSheetColExisting(sTrans_ExcelSheet, sTrans_ExcelCol):
                                self.__odicCB_export_notes_process[
                                    (sTrans_ExcelSheet, sTrans_ExcelCol)] = sCB_Export_Notes_process
                            else:
                                res.addErr("CB export: Export info: Trans.Sheet.Col not set",
                                           add=str(sTrans_ExcelSheet) + "." + str(sTrans_ExcelCol),
                                           kat="Initialisation", sys="Converter: MARLIN", sheet="Def", row=iRow)

                        # Notes Internal (stumpf merken)
                        if isSet(sCB_Export_Notes_internal):
                            if self.trans_isSheetColExisting(sTrans_ExcelSheet, sTrans_ExcelCol):
                                self.__odicCB_export_notes_internal[
                                    (sTrans_ExcelSheet, sTrans_ExcelCol)] = sCB_Export_Notes_internal
                            else:
                                res.addErr("CB import: Additional info: Trans.Sheet.Col not set",
                                           add=str(sTrans_ExcelSheet) + "." + str(sTrans_ExcelCol),
                                           kat="Initialisation", sys="Converter: MARLIN", sheet="Def", row=iRow)

                        # Check: Export Ok + Tweak: Export Notes Process muss existieren
                        if self.cb_export_isTweaked(sTrans_ExcelSheet,sTrans_ExcelCol) and not self.cb_export_hasNotes_process(sTrans_ExcelSheet, sTrans_ExcelCol):
                            res.addErr("CB export: marked as Tweak: but has no export info",
                                       add=str(sTrans_ExcelSheet) + "." + str(sTrans_ExcelCol),
                                       kat="Initialisation", sys="Converter: MARLIN", sheet="Def", row=iRow)

                        # Leere Transport Felder (es gibt keinen Export für diese CB Felder)
                        if isSet_both(sCB_Export_ExcelSheet,sCB_Export_ExcelCol) and not self.trans_isSheetColExisting(sTrans_ExcelSheet, sTrans_ExcelCol):
                            self.__setCB_export_singles.add((sCB_Export_ExcelSheet, sCB_Export_ExcelCol))

                        # Alle Felder nochmal separat merken
                        self.__lstCB_export_all.append((sCB_Export_ExcelSheet, sCB_Export_ExcelCol))

                        # --- CB Kompatibilität zu T (nur wenn direct/tweak)
                        # CB -> T
                        if self.cb_export_isDirect(sTrans_ExcelSheet,sTrans_ExcelCol) or self.cb_export_isTweaked(sTrans_ExcelSheet,sTrans_ExcelCol):
                            # T
                            bComp_T_obl = self.trans_isObl(sTrans_ExcelSheet, sTrans_ExcelCol)
                            sComp_T_unit = self.trans_getUnit(sTrans_ExcelSheet, sTrans_ExcelCol)
                            sComp_T_type = self.trans_getType(sTrans_ExcelSheet, sTrans_ExcelCol)
                            bComp_T_isGeo = self.trans_isGeo(sTrans_ExcelSheet, sTrans_ExcelCol)
                            sComp_T_geo = self.trans_getGeo(sTrans_ExcelSheet, sTrans_ExcelCol)
                            # CB -> T
                            resSchema = self.__model.conf_getDbSchema().getDbSchemaInfo(self.__map.all_getSheetSystem(sCB_Export_ExcelSheet), sCB_Export_ExcelSheet, sCB_Export_ExcelCol)
                            bComp_CB_obl = self.__map.ingest_isObl(sCB_Export_ExcelSheet,sCB_Export_ExcelCol)
                            sComp_CB_unit = resSchema.getData("comment-part-unit")
                            sComp_CB_type = resSchema.getData("datatype")
                            bComp_CB_geo = self.__map.ingest_isGeo(sCB_Export_ExcelSheet,sCB_Export_ExcelCol)
                            sComp_CB_geo = self.__map.ingest_getGeo(sCB_Export_ExcelSheet,sCB_Export_ExcelCol)
                            # Inkompatabilität
                            sComp_CB_detail = ""
                            sComp_T_detail = ""
                            lstCompFail = []
                            if bComp_CB_obl is not bComp_T_obl:
                                lstCompFail.append("obl")
                                sComp_T_detail += "obl " if bComp_T_obl else ""
                                sComp_CB_detail += "obl " if bComp_CB_obl else ""
                            sComp_T_type_conv = self.typeM2CB(sComp_T_type)
                            bTypeOk, bLoss = self.checkComp(sComp_CB_type, sComp_T_type_conv)
                            if not bTypeOk:
                                lstCompFail.append("type")
                                sComp_T_detail += "type{"+str(sComp_T_type)+"} " if sComp_T_type is not None else ""
                                sComp_CB_detail += "type{"+str(sComp_CB_type)+"} " if sComp_CB_type is not None else ""
                            if bTypeOk and bLoss:
                                lstCompFail.append("loss")
                                sComp_T_detail += "loss{" + str(sComp_T_type) + "} " if sComp_T_type is not None else ""
                                sComp_CB_detail += "loss{" + str(sComp_CB_type) + "} " if sComp_CB_type is not None else ""
                            if sComp_CB_unit != self.unitM2CB(sComp_T_unit):
                                lstCompFail.append("unit")
                                sComp_T_detail += "unit{"+str(sComp_T_unit) + "} " if sComp_T_unit is not None else ""
                                sComp_CB_detail += "unit{"+str(sComp_CB_unit) + "} " if sComp_CB_unit is not None else ""
                            if bComp_CB_geo:
                                if sComp_CB_geo != sComp_T_geo:
                                    lstCompFail.append("geo")
                                    sComp_T_detail += "geo{"+str(sComp_T_geo) + "} " if sComp_T_geo is not None else ""
                                    sComp_CB_detail += "geo{"+str(sComp_CB_geo) + "} " if sComp_CB_geo is not None else ""
                            # ---
                            if len(lstCompFail) > 0:
                                sComp_CB_detail = sComp_CB_detail.strip()
                                sComp_T_detail = sComp_T_detail.strip()
                                self.__odicCB_export_incomp[(sTrans_ExcelSheet, sTrans_ExcelCol)] = (lstCompFail, sComp_CB_detail, sComp_T_detail)

                        # CB <- T
                        if self.cb_import_isDirect(sTrans_ExcelSheet,sTrans_ExcelCol) or self.cb_import_isTweaked(sTrans_ExcelSheet,sTrans_ExcelCol):
                            resSchema = self.__model.conf_getDbSchema().getDbSchemaInfo(
                                self.__map.all_getSheetSystem(sCB_Import_ExcelSheet), sCB_Import_ExcelSheet, sCB_Import_ExcelCol)
                            bComp_CB_obl = self.__map.ingest_isObl(sCB_Import_ExcelSheet, sCB_Import_ExcelCol)
                            sComp_CB_unit = resSchema.getData("comment-part-unit")
                            sComp_CB_type = resSchema.getData("datatype")
                            bComp_CB_geo = self.__map.ingest_isGeo(sCB_Import_ExcelSheet, sCB_Import_ExcelCol)
                            sComp_CB_geo = self.__map.ingest_getGeo(sCB_Import_ExcelSheet, sCB_Import_ExcelCol)
                            # Incompatabilität
                            sComp_CB_detail = ""
                            sComp_T_detail = ""
                            lstCompFail = []
                            if bComp_CB_obl is not bComp_T_obl:
                                lstCompFail.append("obl")
                                sComp_T_detail += "obl " if bComp_T_obl else ""
                                sComp_CB_detail += "obl " if bComp_CB_obl else ""
                            sComp_T_type_conv = self.typeM2CB(sComp_T_type)
                            bTypeOk, bLoss = self.checkComp(sComp_T_type_conv, sComp_CB_type)
                            if not bTypeOk:
                                lstCompFail.append("type")
                                sComp_T_detail += "type{" + str(sComp_T_type) + "} " if sComp_T_type is not None else ""
                                sComp_CB_detail += "type{" + str(sComp_CB_type) + "} " if sComp_CB_type is not None else ""
                            if bTypeOk and bLoss:
                                lstCompFail.append("loss")
                                sComp_T_detail += "loss{" + str(sComp_T_type) + "} " if sComp_T_type is not None else ""
                                sComp_CB_detail += "loss{" + str(sComp_CB_type) + "} " if sComp_CB_type is not None else ""
                            if sComp_CB_unit != self.unitM2CB(sComp_T_unit):
                                lstCompFail.append("unit")
                                sComp_T_detail += "unit{" + str(sComp_T_unit) + "} " if sComp_T_unit is not None else ""
                                sComp_CB_detail += "unit{" + str(
                                    sComp_CB_unit) + "} " if sComp_CB_unit is not None else ""
                            if bComp_CB_geo:
                                if sComp_CB_geo != sComp_T_geo:
                                    lstCompFail.append("geo")
                                    sComp_T_detail += "geo{" + str(
                                        sComp_T_geo) + "} " if sComp_T_geo is not None else ""
                                    sComp_CB_detail += "geo{" + str(
                                        sComp_CB_geo) + "} " if sComp_CB_geo is not None else ""
                            # ---
                            if len(lstCompFail) > 0:
                                self.__odicCB_import_incomp[(sTrans_ExcelSheet, sTrans_ExcelCol)] = (lstCompFail, sComp_T_detail, sComp_CB_detail)

                        # --------------------------------------------------------------------------------
                        # MARLIN
                        # --------------------------------------------------------------------------------

                        # --- Import

                        # Marlin.Table.Attrib: Entweder beide NONE oder beide gesetzt
                        if not isSetOrNone_both(sMarlin_Import_Table, sMarlin_Import_Attrib):
                            res.addErr("MARLIN import: MARLIN Table.Attrib has to be either set or unset both",
                                       add=str(sMarlin_Import_Table) + "." + str(sMarlin_Import_Attrib),
                                       kat="Initialisation",
                                       sys="Converter: MARLIN", sheet="Def", row=iRow)

                        # Marlin.Table.Attrib: Muss einmalig sein sein
                        if isSet_both(sMarlin_Import_Table, sMarlin_Import_Attrib) and isSet(sTrans_ExcelSheet):
                            if (sMarlin_Import_Table, sMarlin_Import_Attrib) not in dicMarlin_import_db_uniq[sTrans_ExcelSheet]:
                                dicMarlin_import_db_uniq[sTrans_ExcelSheet].add((sMarlin_Import_Table, sMarlin_Import_Attrib))
                            else:
                                res.addErr("MARLIN import: MARLIN Table.Attrib is double",
                                           add=str(sTrans_ExcelSheet) + ": "+str(sMarlin_Import_Table) + "." + str(sMarlin_Import_Attrib),
                                           kat="Initialisation",
                                           sys="Converter: MARLIN", sheet="Def", row=iRow)

                        # T.Sheet.Col -> Marlin.Table.Attrib
                        if self.trans_isSheetColExisting(sTrans_ExcelSheet, sTrans_ExcelCol) and isSet_both(
                                sMarlin_Import_Table, sMarlin_Import_Attrib):
                            # Marlin.Table.Attrib: einmalige (als Import Ziel)
                            if (sMarlin_Import_Table, sMarlin_Import_Attrib) not in self.__odicM_import_dbTable.values():
                                self.__odicM_import_dbTable[(sTrans_ExcelSheet, sTrans_ExcelCol)] = (sMarlin_Import_Table, sMarlin_Import_Attrib)
                            else:
                                res.addErr(
                                    "MARLIN import: Marlin.Table.Attrib is used more than once as Import Target by Trans.Sheet.Col",
                                    add=str(sMarlin_Import_Table) + "." + str(sMarlin_Import_Attrib),
                                    kat="Initialisation",
                                    sys="Converter: MARLIN", sheet="Def", row=iRow)

                        # Tweak (stumpf merken)
                        if isSet(sMarlin_Import_Tweak):
                            if self.trans_isSheetColExisting(sTrans_ExcelSheet, sTrans_ExcelCol):
                                self.__setM_import_tweak.add((sTrans_ExcelSheet, sTrans_ExcelCol))
                            else:
                                res.addErr("MARLIN import: Tweaked: Trans.Sheet.Col not found",
                                           add=str(sTrans_ExcelSheet) + "." + str(sTrans_ExcelCol),
                                           kat="Initialisation",
                                           sys="Converter: MARLIN", sheet="Def", row=iRow)

                        # Notes Allgemein (stumpf merken)
                        if isSet(sMarlin_Import_Notes_general):
                            if self.trans_isSheetColExisting(sTrans_ExcelSheet, sTrans_ExcelCol):
                                self.__odicM_import_notes_general[
                                    (sTrans_ExcelSheet, sTrans_ExcelCol)] = sMarlin_Import_Notes_general
                            else:
                                res.addErr("MARLIN import: General Info: Trans.Sheet.Col not set",
                                           add=str(sTrans_ExcelSheet) + "." + str(sTrans_ExcelCol),
                                           kat="Initialisation", sys="Converter: MARLIN", sheet="Def", row=iRow)

                        # Notes Process (stumpf merken)
                        if isSet(sMarlin_Import_Notes_process):
                            if self.trans_isSheetColExisting(sTrans_ExcelSheet, sTrans_ExcelCol):
                                self.__odicM_import_notes_process[(sTrans_ExcelSheet, sTrans_ExcelCol)] = sMarlin_Import_Notes_process
                            else:
                                res.addErr("MARLIN import: Import info: Trans.Sheet.Col not set",
                                           add=str(sTrans_ExcelSheet) + "." + str(sTrans_ExcelCol),
                                           kat="Initialisation", sys="Converter: MARLIN", sheet="Def", row=iRow)

                        # Notes Internal (stumpf merken)
                        if isSet(sMarlin_Import_Notes_internal):
                            if self.trans_isSheetColExisting(sTrans_ExcelSheet, sTrans_ExcelCol):
                                self.__odicM_import_notes_intenal[
                                    (sTrans_ExcelSheet, sTrans_ExcelCol)] = sMarlin_Import_Notes_internal
                            else:
                                res.addErr("MARLIN import: Additional info: Trans.Sheet.Col not set",
                                           add=str(sTrans_ExcelSheet) + "." + str(sTrans_ExcelCol),
                                           kat="Initialisation", sys="Converter: MARLIN", sheet="Def", row=iRow)

                        # Check: Import Ok + Tweak: Import Notes Process muss existieren
                        if self.m_import_isTweaked(sTrans_ExcelSheet, sTrans_ExcelCol) and not self.m_import_hasNotes_process(
                                sTrans_ExcelSheet, sTrans_ExcelCol):
                            res.addErr("MARLIN import: marked as Tweak: but has no import info",
                                       add=str(sTrans_ExcelSheet) + "." + str(sTrans_ExcelCol),
                                       kat="Initialisation", sys="Converter: MARLIN", sheet="Def", row=iRow)

                        # Leere Transport Felder (es gibt keinen Import für diese MARLIN Felder)
                        if isSet_both(sMarlin_Import_Table,sMarlin_Import_Attrib) and not self.trans_isSheetColExisting(sTrans_ExcelSheet,sTrans_ExcelCol):
                            self.__lstM_import_singles.append((sMarlin_Import_Table, sMarlin_Import_Attrib))

                        # --- Export

                        # M.SvView.SvAttrib: Entweder beide NONE oder beide gesetzt
                        if not isSetOrNone_both(sMarlin_Export_SvView, sMarlin_Export_SvAttrib) and isSet(sTrans_ExcelSheet):
                            res.addErr("MARLIN export: Table.Attrib has to be either set or unset both",
                                       add=str(sMarlin_Export_SvView) + "." + str(sMarlin_Export_SvAttrib),
                                       kat="Initialisation",
                                       sys="Converter: MARLIN", sheet="Def", row=iRow)

                        # M.SvView.SvAttrib: Muss einmalig sein sein
                        if isSet_both(sMarlin_Export_SvView, sMarlin_Export_SvAttrib) and isSet(sTrans_ExcelSheet):
                            if (sMarlin_Export_SvView, sMarlin_Export_SvAttrib) not in dicMarlin_export_sv_uniq[sTrans_ExcelSheet]:
                                dicMarlin_export_sv_uniq[sTrans_ExcelSheet].add((sMarlin_Export_SvView, sMarlin_Export_SvAttrib))
                            else:
                                res.addErr("MARLIN export: MARLIN SvView.SvAttrib is double",
                                           add=str(sTrans_ExcelSheet) + ": "+str(sMarlin_Export_SvView) + "." + str(sMarlin_Export_SvAttrib),
                                           kat="Initialisation",
                                           sys="Converter: MARLIN", sheet="Def", row=iRow)

                        # T.Sheet.Col -> M.SvView.SvAttrib
                        if self.trans_isSheetColExisting(sTrans_ExcelSheet, sTrans_ExcelCol) and isSet_both(sMarlin_Export_SvView, sMarlin_Export_SvAttrib):
                            self.__odicM_export_dbView[(sTrans_ExcelSheet, sTrans_ExcelCol)] = (sMarlin_Export_SvView, sMarlin_Export_SvAttrib)

                        # Tweak (stumpf merken)
                        if isSet(sMarlin_Export_Tweak):
                            if self.trans_isSheetColExisting(sTrans_ExcelSheet, sTrans_ExcelCol):
                                self.__setM_export_tweak.add((sTrans_ExcelSheet, sTrans_ExcelCol))
                            else:
                                res.addErr("MARLIN export: Tweaked: Trans.Sheet.Col not found",
                                           add=str(sTrans_ExcelSheet) + "." + str(sTrans_ExcelCol),
                                           kat="Initialisation",
                                           sys="Converter: MARLIN", sheet="Def", row=iRow)

                        # Notes Allgemein (stumpf merken)
                        if isSet(sMarlin_Export_Notes_general):
                            if self.trans_isSheetColExisting(sTrans_ExcelSheet, sTrans_ExcelCol):
                                self.__odicM_export_notes_general[
                                    (sTrans_ExcelSheet, sTrans_ExcelCol)] = sMarlin_Export_Notes_general
                            else:
                                res.addErr("MARLIN export: General Info: Trans.Sheet.Col not set",
                                           add=str(sTrans_ExcelSheet) + "." + str(sTrans_ExcelCol),
                                           kat="Initialisation", sys="Converter: MARLIN", sheet="Def", row=iRow)

                        # Notes Process (stumpf merken)
                        if isSet(sMarlin_Export_Notes_process):
                            if self.trans_isSheetColExisting(sTrans_ExcelSheet, sTrans_ExcelCol):
                                self.__odicM_export_notes_process[
                                    (sTrans_ExcelSheet, sTrans_ExcelCol)] = sMarlin_Export_Notes_process
                            else:
                                res.addErr("MARLIN export: Export info: Trans.Sheet.Col not set",
                                           add=str(sTrans_ExcelSheet) + "." + str(sTrans_ExcelCol),
                                           kat="Initialisation", sys="Converter: MARLIN", sheet="Def", row=iRow)

                        # Notes Internal (stumpf merken)
                        if isSet(sMarlin_Export_Notes_internal):
                            if self.trans_isSheetColExisting(sTrans_ExcelSheet, sTrans_ExcelCol):
                                self.__odicM_export_notes_internal[
                                    (sTrans_ExcelSheet, sTrans_ExcelCol)] = sMarlin_Export_Notes_internal
                            else:
                                res.addErr("CB import: Additional info: Trans.Sheet.Col not set",
                                           add=str(sTrans_ExcelSheet) + "." + str(sTrans_ExcelCol),
                                           kat="Initialisation", sys="Converter: MARLIN", sheet="Def", row=iRow)

                        # Check: Export Ok + Tweak: Export Notes Process muss existieren
                        if self.m_export_isTweaked(sTrans_ExcelSheet, sTrans_ExcelCol) and not self.m_export_hasNotes_process(sTrans_ExcelSheet, sTrans_ExcelCol):
                            res.addErr("MARLIN export: marked as Tweak: but has no export info",
                                       add=str(sTrans_ExcelSheet) + "." + str(sTrans_ExcelCol),
                                       kat="Initialisation", sys="Converter: MARLIN", sheet="Def", row=iRow)

                        # Leere Transport Felder (es gibt keinen Export für diese CB Felder)
                        if isSet_both(sMarlin_Export_SvView,sMarlin_Export_SvAttrib) and not self.trans_isSheetColExisting(sTrans_ExcelSheet,sTrans_ExcelCol):
                            self.__lstM_export_singles.append((sMarlin_Export_SvView, sMarlin_Export_SvAttrib))

        except Exception as ex:
            res.addErr("Unexpected problem: " +str(sTrans_ExcelSheet) + "." + str(sTrans_ExcelCol),
                       add=str(traceback.format_exc()),
                       kat="Initialisation", sys="Converter: MARLIN", sheet="Def", row=iRow)

        # Working
        self.setWorking(True)
        self.setReady(True)

        # Antwort
        if res.isWarn():
            print(res.getWarn())

        if res.isErr():
            print(res.getErr())
            self.setReady(False)

        return res

    # Reset
    def reset(self):
        pass

    # Info
    def info(self, bHtml):

        sWfConfLong = self.__map.workflow_converter_import_short2long(self.__sWf_conv)
        sWfConfTargetShort = self.__map.workflow_converter_import_getTarget(self.__sWf_conv)
        sWfConfTargetLong = self.__map.workflow_short2long(sWfConfTargetShort)

        sColor_bsh_blue_dark = Colors.getColor("bsh_blue_dark", html=True)
        sColor_bsh_blue_mid = Colors.getColor("bsh_blue_mid", html=True)
        sColor_bsh_blue_light = Colors.getColor("bsh_blue_light", html=True)
        sColor_awi_blue_light = Colors.getColor("awi_blue_light", html=True)

        sColor_blue = Colors.getColor("blue", html=True)
        sColor_orange = sColor_bsh_blue_mid
        sColor_red = Colors.getColor("red", html=True)
        sColor_green = Colors.getColor("green", html=True)
        sColor_yellow = Colors.getColor("yellow", html=True)
        sColor_blue = Colors.getColor("blue", html=True)

        iPreview = 5

        # Helper
        def html_h1(sTxt):
            return '<p><hr align="left" width="75%" style="border-top: dotted 1px;"><h2 style="color:'+sColor_bsh_blue_mid+';">' +str(sTxt)+ '</h2><hr align="left" width="75%" style="border-top: dotted 1px;">'

        def html_h2(sTxt):
            return '<p><h2 style="color:'+sColor_bsh_blue_mid+';">' +str(sTxt)+ '</h2>'

        def html_h3(sTxt):
            return '<p><h3 style="color:'+sColor_bsh_blue_light+';">' +str(sTxt)+ '</h3>'

        # ---
        sOut = ""
        if not self.isReady():
            sOut += "Converter MARLIN-CB is not initialized"
            if bHtml:
                sOut = PrettyText.html_txt2html(sOut)
            return sOut

        if not bHtml:
            sOut = "************************************\n"
            sOut += str(sWfConfLong) + " -> " + str(sWfConfTargetLong) + " (CB)\n"
            sOut += "************************************\n\n"

        if bHtml:
            sOut += '<img width="115" height="67" src="https://lindevmarlin61.bsh.de/MARLINDMZ//Images/Marlin_Logo.png" alt="BSH Logo">'
            sOut += '<div style="z-index:1;position:absolute;left:132px;top:7px;width:566px;height:28px;line-height:28px !important;overflow:hidden;text-align:left;background-color:Transparent;color:#425a89;border:medium none;-moz-border-colors:none;border-radius:0;font:normal normal bold 19px Arial;text-decoration:none;">Marine Life Investigator</div>'
            sOut += '<div style="z-index:1;position:absolute;left:132px;top:41px;width:567px;height:19px;line-height:19px;text-align:left;background-color:Transparent;color:#425a89;border:medium none;-moz-border-colors:none;border-radius:0;font:normal normal bold 13px Arial;text-decoration:none;">Bundesamt für Seeschifffahrt und Hydrographie</div>'
            sOut += '<div layout="block" styleclass="DynForm_DfAttributes" style="z-index:0;position:absolute;left:0px;top:68px;right:0px;height:25px;overflow:hidden;text-align:left;background-color:#425a89;border:medium none;-moz-border-colors:none;border-radius:0;"></div>'
            sOut += "\n\n"
            sOut += "&nbsp;"
            sOut += '<h1 style="color:'+sColor_bsh_blue_dark+';">Konsistenzprüfung des Datentransfers zwischen MARLIN und CRITTERBASE</h1>\n\n'

        sOut += "*** Erläuterungen\n\n"
        sOut += "--- Abkürzungen\n\n"
        lstTmp = [["Kürzel", "Erklärung"]]
        lstTmp2 = [
            ("T","Transportformat"),
            ("CB","CRITTERBASE"),
            ("M","MARLIN"),
            ("P","Partner (kann jeweils M oder CB sein)"),
            ("tweak", "Sonderweg"),
            ("type", "Datentyp (z.B. string)"),
            ("opt", "Optionale Infomation - zusätzlich zum Datentyp"),
            ("lon", "Longitude"),
            ("lat", "Latitude"),
            ("date", "Datum"),
            ("time", "Zeit"),
            ("obl", "obligat"),
            ("cobl", "conditional obligat (nur unter bestimmten Umständen) (opt)"),
            ("->lup{table}", "Lookup (Verweis auf einen Lookup-Eintrag in CB - anhand des Namen)"),
            ("lup{table}", "Lookup selber (Ein Lookup Eintrag ist selber gemeint)"),
            ("pk", "Primary Key (Teil eines Primary keys - Identifier für einen Datensatz)"),
            ("fk", "Foreign Key (Teil eines Foreign Keys - Bezug auf den Identifier eines anderen Datensatzes)"),
            ("A -> B", "Transportweg: A exportiert nach B, bzw B importiert von A")
        ]
        lstTmp += lstTmp2
        sOut += PrettyText.table(lstTmp, True, bHtml)
        sOut += "\n"
        sOut += "\n"

        sOut += "--- Sichten auf den Transportprozess\n\n"
        sOut += "* Drei Partner sind in den Transportprozess involviert\n\n"
        sOut += "T\n"
        sOut += "M\n"
        sOut += "CB\n"
        sOut += "\n"
        sOut += "* Zentriert auf einen Partner (nur Aktion mit T)\n\n"
        sOut += "P <-> T \n"
        sOut += "\n"

        sOut += "* Zentriert auf den Transportweg (beide Partner interessant)\n\n"
        sOut += "M -> CB : M -> T -> CB\n"
        sOut += "M <- CB : M <- T <- CB\n"
        sOut += "\n"

        sOut += "--- Aufbau der Konfigurationsdatei\n\n"
        sOut += "* Übersicht\n\n"
        lstTmp = [["Partner", "Richtung", "Spalte"]]
        sSektion = ""
        sKategorie = ""
        for iNr, sHead in self.__odicHeader.items():
            if iNr in self.__odicHeaderSectionsMTCB:
                sSektion = self.__odicHeaderSectionsMTCB[iNr]
            if iNr in self.__odicHeaderSectionsKat:
                sKategorie = self.__odicHeaderSectionsKat[iNr]
            lstTmp.append((sSektion, sKategorie, sHead))
        sOut += PrettyText.table(lstTmp, True, bHtml, preview=iPreview)
        sOut += "\n\n"
        sOut += "* Erklärungen\n\n"
        sOut += "- Datentyp: technischer Datentyp des Feldes (z.B.: string, integer, float)\n"
        sOut += "- Einheit: Einheit des Feldes (z.B.: Kilogramm, Meter, Breitengrad in dD)\n"
        sOut += "- Options: Weitere inhaltliche Zuordnung des Feldes (z.B.: obligat, konditional obligat, Längengrad, Breitengrad - wichtig für genauere Prüfungen)\n"
        sOut += "- OK: Transportweg ist möglich\n"
        sOut += "- Sonderlösung: Transportweg ist nur möglich mit gesondertem Algorithmus\n"
        if bHtml:
            sFile = Path(self.__model.conf_getAppBaseDir(), "doc", "pic", "converter_marlin.jpg")
            sOut += "* Screenshot\n"
            sOut += '<img src="'+str(sFile)+'" alt="Konfigurationsdatei" width="800px"><br>'
        sOut += "\n"

        sOut += "--- Konsistenzprüfungen der beiden Datenmodelle (vor dem Transport)\n\n"
        sOut += "* Transportebene als Referenzpartner\n\n"
        sOut += "- Auf CB-Seite werden alle möglichen Konsistenzprüfungen anhand des vorliegenen Systems (CB Mapper) geprüft.\n"
        sOut += "- Auf MARLIN-Seite ist dies nicht möglich, da wir keinen Zugriff auf das MARLIN Mapping-System (Import-Magic) besitzen.\n"
        sOut += "- CB verläßt sich hierbei auf die Korrektheit der definierten Transport-Ebene\n\n"

        sOut += "* Basisprüfungen\n\n"
        sOut += "- Transportprozesspartner müssen immer komplett (Sheet/Colum oder Table/Attribute) oder gar nicht definiert sein\n"
        sOut += "- Optionen: Alle Optionen müssen valide sein\n"
        sOut += "- Alle Transporteinträge oder -ziele müssen einmalig sein\n"
        sOut += "- Ein Transporteintrag (T Sheet.Column) muss innerhalb eines Transportprozesses immer ein einmaliges Ziel besitzen (oder umgekehrt)\n"
        sOut += "- CB: Alle Ziele werden auf Existenz geprüft\n\n"

        sOut += "* Transport möglich\n\n"
        sOut += "- OK muss gesetzt sein\n"
        sOut += "- Datentyp, Option und Unit beider Partner müssen übereinstimmen\n\n"

        sOut += "* Direkter Transport möglich\n\n"
        sOut += "- Ziel muss gesetzt sein (Sheet/Col, Table/Attribute)\n\n"

        sOut += "* Indirekter Transport möglich\n\n"
        sOut += "- Sonderlösung ist gesetzt\n"
        sOut += "- Notizen zu Im-/Export müssen gegeben sein (Beschreibung der Sonderlösung aus)\n\n"

        # --- CB -> T -> M
        def proz(a,b):
            try:
                return str(a)+"x ~ "+str(int((a/b) * 100))+"%"
            except Exception as ex:
                return ""
        def proz_html(a, b, bPos):
            sColorgrad = "pos" if bPos else "neg"
            return PrettyText.html_bar(a,b, colorgrad=sColorgrad, abs=True)

        iAll = self.trans_getAll(nr=True)
        iAll_cb2m = self.trans_getAll(side="both", dir="export", nr=True)
        iAll_m2cb = self.trans_getAll(side="both", dir="import", nr=True)
        iAll_ok = self.total_working_CB2M() + self.total_working_M2CB()
        iAll_not_ok = (iAll*2) - iAll_ok

        sOut += "*** Transportwege\n\n"
        sOut += "--- CB <-> T <-> M\n\n"
        sOut += "* Alle Transportwege\n\n"
        lstTmp = [["Im-/Export", "Export", "Import"]]
        lstTmp.append(["CB <-> T <-> M", "CB -> T -> M", "CB <- T <- M"])
        lstTmp.append([str(iAll*2), str(iAll_cb2m), str(iAll_m2cb)])
        sOut += PrettyText.table(lstTmp, True, bHtml, preview=iPreview)
        sOut += "\n\n"

        sOut += "* Funktionierende Wege ("+proz(iAll_ok, iAll * 2)+")\n\n"
        lstTmp = [["Gesamtweg", "Teilweg", "Teilweg"]]
        lstTmp.append(["CB -> M", "CB -> T", "T -> M"])
        all = str(iAll) + "x" if iAll > 0 else ""

        a = proz_html(self.total_working_CB2M(), iAll_cb2m, True) if self.total_working_CB2M() > 0 else ""
        b = proz_html(self.total_working_CB2T(), iAll_cb2m, True) if self.total_working_CB2T() > 0 else ""
        c = proz_html(self.total_working_T2M(), iAll_cb2m, True) if self.total_working_T2M() > 0 else ""
        lstTmp.append([a, b, c])
        lstTmp.append(["CB <- M", "CB <- T", "T <- M"])
        all = str(iAll) + "x" if iAll > 0 else ""
        a = proz_html(self.total_working_M2CB(), iAll_m2cb, True) if self.total_working_M2CB() > 0 else ""
        b = proz_html(self.total_working_T2CB(), iAll_m2cb, True) if self.total_working_T2CB() > 0 else ""
        c = proz_html(self.total_working_M2T(), iAll_m2cb, True) if self.total_working_M2T() > 0 else ""
        lstTmp.append([a, b, c])
        sOut += PrettyText.table(lstTmp, True, bHtml, preview=iPreview)
        sOut += "\n\n"

        sOut += "* Nicht funktionierende Wege ("+proz(iAll_not_ok, iAll * 2)+")\n\n"
        lstTmp = [["Gesamtweg", "Teilweg", "Teilweg"]]
        lstTmp.append(["CB -!-> M", "CB -!-> T", "T -!-> M"])
        a = proz_html(self.total_notWorking_CB2M(), iAll_cb2m, False) if self.total_notWorking_CB2M() > 0 else ""
        b = proz_html(self.total_notWorking_CB2T(), iAll_cb2m, False) if self.total_notWorking_CB2T() > 0 else ""
        c = proz_html(self.total_notWorking_T2M(), iAll_cb2m, False) if self.total_notWorking_T2M() > 0 else ""
        lstTmp.append([a,b,c])
        lstTmp.append(["CB <-!- M", "CB <-!- T", "T <-!- M"])
        a = proz_html(self.total_notWorking_M2CB(), iAll_m2cb, False) if self.total_notWorking_M2CB() > 0 else ""
        b = proz_html(self.total_notWorking_T2CB(), iAll_m2cb, False) if self.total_notWorking_T2CB() > 0 else ""
        c = proz_html(self.total_notWorking_M2T(), iAll_m2cb, False) if self.total_notWorking_M2T() > 0 else ""
        lstTmp.append([a,b,c])
        sOut += PrettyText.table(lstTmp, True, bHtml, preview=iPreview)
        sOut += "\n\n"

        lstCBT = self.trans_getAll(side="cb", dir="import", tweak=True)
        lstTM = self.trans_getAll(side="m", dir="export", tweak=True)
        lstTCB = self.trans_getAll(side="cb", dir="export", tweak=True)
        lstMT = self.trans_getAll(side="m", dir="import", tweak=True)
        iTotal = len(lstCBT) + len(lstTM) + len(lstMT) + len(lstTCB)
        sOut += "* Sonderlösungen ("+proz(iTotal, iAll * 2)+")\n\n"
        lstTmp = [["Teilweg", "Teilweg"]]
        lstTmp.append(["CB -> T", "T -> M"])
        a = str(len(lstCBT)) + "x" if len(lstCBT) > 0 else ""
        b = str(len(lstTM)) + "x" if len(lstTM) > 0 else ""
        lstTmp.append([a, b])
        lstTmp.append(["CB <- T", "T <- M"])
        a = str(len(lstTCB)) + "x" if len(lstTCB) > 0 else ""
        b = str(len(lstMT)) + "x" if len(lstMT) > 0 else ""
        lstTmp.append([a, b])
        sOut += PrettyText.table(lstTmp, True, bHtml, preview=iPreview)
        sOut += "\n\n"

        lstCB_export = list(self.cb_export_getSingles())
        lstCB_import = list(self.cb_import_getSingles())
        lstM_export = self.m_export_getSingles()
        lstM_import = self.m_import_getSingles()
        iSingles = len(lstCB_import) + len(lstCB_export) + len(lstM_import) + len(lstM_export)
        sOut += "* Keine Partner in Transportformat ("+str(iSingles)+"x)\n\n"
        lstTmp = [["Export", "Import"]]
        lstTmp.append(["CB -> ?", "CB <- ?"])
        a = str(len(lstCB_export)) + "x" if len(lstCB_export) > 0 else ""
        c = str(len(lstCB_import)) + "x" if len(lstCB_import) > 0 else ""
        lstTmp.append([a, b])
        lstTmp.append(["M -> ?", "M <- ?"])
        a = str(len(lstM_export)) + "x" if len(lstM_export) > 0 else ""
        b = str(len(lstM_import)) + "x" if len(lstM_import) > 0 else ""
        lstTmp.append([a, b])
        sOut += PrettyText.table(lstTmp, True, bHtml, preview=iPreview)
        sOut += "\n\n"


        sOut += "* Gesamtübersicht\n\n"
        lstTmp = [("CB Sheet.Col", "<->", "T Sheet.Col", "<->", "M Table.Attribute")]

        for sTSheet in self.trans_getSheets():
            for sTCol in self.trans_getCols(sTSheet):

                # ------------------------------------------------------------------------------------------------------
                # Transport
                # ------------------------------------------------------------------------------------------------------

                # Hole Daten
                bTrans_Obl = self.trans_isObl(sTSheet, sTCol)
                bTrans_Cobl = self.trans_isCobl(sTSheet, sTCol)
                sTrans_Type = self.trans_getType(sTSheet, sTCol)
                bTrans_Geo = self.trans_isGeo(sTSheet, sTCol)
                sTrans_Geo = self.trans_getGeo(sTSheet, sTCol)
                bTrans_Lat = self.trans_isLat(sTSheet, sTCol)
                bTrans_Lon = self.trans_isLon(sTSheet, sTCol)
                sTrans_Unit = self.trans_getUnit(sTSheet, sTCol)
                sTrans_Comment = self.trans_getComment(sTSheet, sTCol)
                sTrans_Question = self.trans_getQuestion(sTSheet, sTCol)
                sTrans_SheetCol = str(sTSheet) + "." + str(sTCol)
                sInfo_Trans_Opt = ""
                sInfo_Trans_Opt += "[obl]" if self.trans_isObl(sTSheet, sTCol) else ""
                sInfo_Trans_Opt += "[cobl]" if self.trans_isCobl(sTSheet, sTCol) else ""
                sInfo_Trans_Type = sTrans_Type if sTrans_Type not in ["", None] else ""
                sInfo_Trans_Unit = sTrans_Unit if sTrans_Unit not in ["", None] else ""
                lstInfo_Trans_Opt = self.trans_getOpts(sTSheet, sTCol)

                # Info: Sheet/Col
                sInfo_Trans_SheetCol = ""
                sInfo_Trans_SheetCol_popup = ""
                sInfo_Trans_SheetCol_popup_opt = "<sup><small>{datatype: " + str(sTrans_Type) + "}"
                if len(lstInfo_Trans_Opt) > 0:
                    sInfo_Trans_SheetCol_popup_opt += " {options: " + PrettyText.lst2strC(lstInfo_Trans_Opt) + "}"
                sInfo_Trans_SheetCol_popup_opt += "</sup></small>"
                sInfo_Trans_SheetCol_popup += "<strong>" +  sTrans_SheetCol + "</strong><br>" + sInfo_Trans_SheetCol_popup_opt + "<br>" if sTrans_SheetCol not in [None, ""] else ""
                sInfo_Trans_SheetCol_popup += "[" + sTrans_Unit + "]<br>" if sTrans_Unit not in [None, ""] else ""
                sInfo_Trans_SheetCol_popup += "<em>" + sTrans_Comment + "</em><br>" if sTrans_Comment not in [None, ""] else ""
                if sInfo_Trans_SheetCol_popup not in [None, ""] and bHtml:
                    if sTrans_Question not in [None, ""]:
                        sTrans_SheetCol += '<span style="color:'+sColor_orange+';"><sup><strong>?</strong></sup></span>'
                        sInfo_Trans_SheetCol_popup += '<hr style="border-top: dotted 1px;"><div style="color:' + sColor_orange + ';"><b>Offene Frage?</b><br><em>' + sTrans_Question + '</em></div>'
                    bCb2M = self.cb_export_isOk(sTSheet, sTCol) and self.m_import_isOk(sTSheet, sTCol)
                    bM2Cb = self.m_export_isOk(sTSheet, sTCol) and self.cb_import_isOk(sTSheet, sTCol)
                    if bCb2M and bM2Cb:
                        sColor_Trans = sColor_green
                    elif bCb2M != bM2Cb:
                        sColor_Trans = sColor_yellow
                    else:
                        sColor_Trans = sColor_red
                    sInfo_Trans_SheetCol = '<span style="color:'+sColor_Trans+';">&#11044; </span>' + '<div class="tooltip">'+ sTrans_SheetCol + '<span class="right">' + sInfo_Trans_SheetCol_popup + '<i></i></span></div>'
                else:
                    sInfo_Trans_SheetCol = sTrans_SheetCol
                    if sTrans_Question not in [None, ""]:
                        sInfo_Trans_SheetCol += " (Frage?)"

                # ------------------------------------------------------------------------------------------------------
                # CB
                # ------------------------------------------------------------------------------------------------------

                # *** CB Export

                # Hole Daten
                sCB_Export_ExcelSheet, sCB_Export_ExcelCol = self.cb_export_getPartner(sTSheet, sTCol)
                bCB_Export_tweak = self.cb_export_isTweaked(sTSheet, sTCol)
                sCB_Export_notes_process = self.cb_export_getNotes_process(sTSheet, sTCol)
                sCB_Export_notes_internal = self.cb_export_getNotes_internal(sTSheet, sTCol)
                bCB_Export_direct = self.cb_export_isDirect(sTSheet, sTCol)
                sCB_Export_ExcelSheetCol = (str(sCB_Export_ExcelSheet) + "." + str(sCB_Export_ExcelCol)) if not (
                            None in [sCB_Export_ExcelSheet, sCB_Export_ExcelCol]) else ""
                bCB_Export_ExcelSheetCol_obl = self.__map.ingest_isObl(sCB_Export_ExcelSheet, sCB_Export_ExcelCol)
                sCB_Export_ExcelSheetCol_datatype = None
                bCB_Export_ExcelSheetCol_pk = self.__map.ingest_isPk(sCB_Export_ExcelSheet, sCB_Export_ExcelCol)
                bCB_Export_ExcelSheetCol_fk = self.__map.ingest_isFk(sCB_Export_ExcelSheet, sCB_Export_ExcelCol)
                bCB_Export_ExcelSheetCol_lup_ing = self.__map.ingest_isLookup(sCB_Export_ExcelSheet, sCB_Export_ExcelCol)
                sCB_Export_ExcelSheetCol_lup_ing_table = self.__map.ingest_getLookup(sCB_Export_ExcelSheet, sCB_Export_ExcelCol)
                bCB_Export_ExcelSheetCol_lup_itself = (self.__map.all_getSheetSystem(sCB_Export_ExcelSheet) == "lookup")
                sCB_Export_ExcelSheetCol_lup_itself_table = self.__map.lookup_sheet2table(sCB_Export_ExcelSheet)
                bCB_Export_ExcelSheetCol_isGeo = self.__map.ingest_isGeo(sCB_Export_ExcelSheet, sCB_Export_ExcelCol)
                sCB_Export_ExcelSheetCol_getGeo = self.__map.ingest_getGeo(sCB_Export_ExcelSheet, sCB_Export_ExcelCol)
                bCB_Export_ExcelSheetCol_isDate = self.__map.ingest_isDate(sCB_Export_ExcelSheet, sCB_Export_ExcelCol)
                bCB_Export_ExcelSheetCol_isTime = self.__map.ingest_isTime(sCB_Export_ExcelSheet, sCB_Export_ExcelCol)
                resSchema = self.__model.conf_getDbSchema().getDbSchemaInfo(self.__map.all_getSheetSystem(sCB_Export_ExcelSheet), sCB_Export_ExcelSheet, sCB_Export_ExcelCol)
                sCB_Export_ExcelSheetCol_name = resSchema.getData("comment-part-name")
                sCB_Export_ExcelSheetCol_comment = resSchema.getData("comment-part-comment")
                sCB_Export_ExcelSheetCol_unit = resSchema.getData("comment-part-unit")
                sCB_Export_ExcelSheetCol_datatype = resSchema.getData("datatype")
                lstInfo_CbExport_SheetCol_popup = []
                if resSchema.isOk():
                    if bCB_Export_ExcelSheetCol_pk:
                        lstInfo_CbExport_SheetCol_popup.append("pk")
                    if bCB_Export_ExcelSheetCol_fk:
                        lstInfo_CbExport_SheetCol_popup.append("fk")
                    if bCB_Export_ExcelSheetCol_obl:
                        lstInfo_CbExport_SheetCol_popup.append("obl")
                    if bCB_Export_ExcelSheetCol_lup_ing:
                        lstInfo_CbExport_SheetCol_popup.append("->lup("+str(sCB_Export_ExcelSheetCol_lup_ing_table)+")")
                    if bCB_Export_ExcelSheetCol_lup_itself:
                        lstInfo_CbExport_SheetCol_popup.append("lup("+str(sCB_Export_ExcelSheetCol_lup_itself_table)+")")
                    if bCB_Export_ExcelSheetCol_isDate:
                        lstInfo_CbExport_SheetCol_popup.append("date")
                    if bCB_Export_ExcelSheetCol_isTime:
                        lstInfo_CbExport_SheetCol_popup.append("time")
                    if bCB_Export_ExcelSheetCol_isGeo:
                        lstInfo_CbExport_SheetCol_popup.append(str(sCB_Export_ExcelSheetCol_getGeo))

                # Info: Sheet/Col
                sInfo_CB_Export_SheetCol = ""
                sInfo_CbExport_SheetCol_popup = ""
                sCB_Export_ExcelSheetCol_popup_opt = "<sup><small>{datatype: " + str(sCB_Export_ExcelSheetCol_datatype)+"}"
                if len(lstInfo_CbExport_SheetCol_popup) > 0:
                    sCB_Export_ExcelSheetCol_popup_opt += " {options: "+PrettyText.lst2strC(lstInfo_CbExport_SheetCol_popup)+"}"
                sCB_Export_ExcelSheetCol_popup_opt += "</sup></small>"
                sInfo_CbExport_SheetCol_popup += "<strong>" + sCB_Export_ExcelSheetCol_name +"</strong><br>"+sCB_Export_ExcelSheetCol_popup_opt+"<br>" if sCB_Export_ExcelSheetCol_name not in [None, ""] else ""
                sInfo_CbExport_SheetCol_popup += "[" + sCB_Export_ExcelSheetCol_unit +"]<br>" if sCB_Export_ExcelSheetCol_unit not in [None, ""] else ""
                sInfo_CbExport_SheetCol_popup += "<em>" + sCB_Export_ExcelSheetCol_comment +"</em><br>" if sCB_Export_ExcelSheetCol_comment not in [None, ""] else ""
                if None in [sCB_Export_ExcelSheet, sCB_Export_ExcelCol]:
                    sInfo_CB_Export_SheetCol = ""
                else:
                    if sInfo_CbExport_SheetCol_popup not in [None, ""] and bHtml:
                        sInfo_CB_Export_SheetCol = '<div class="tooltip">' + sCB_Export_ExcelSheetCol + '<span class="right">' + sInfo_CbExport_SheetCol_popup + '<i></i></span></div>'
                    else:
                        sInfo_CB_Import_SheetCol = sCB_Export_ExcelSheetCol

                # Info: Export Modus
                sInfo_CB_Export_modus = ""
                sCompFail_info_normal = ""
                sCompFail_info_html_popup = ""
                sCompFail_info_html_arrow = ""

                # Teste Inkompatabilität
                bCompFail = self.cb_export_isCompProblem(sTSheet, sTCol)
                lstCompFail, sCb, sT = self.cb_export_getCompProblem(sTSheet, sTCol)
                if bCompFail:
                    # Inkompatbilität
                    if bCB_Export_direct:
                        # Normal:
                        sColor_problem = sColor_red
                        sCompFail_info_html_popup = '<hr style="border-top: dotted 1px;"><span style="color:' + sColor_problem + ';"><b>Inkompatabilität</b><br><em>' + PrettyText.lst2strC(lstCompFail) + '</em></span>'
                        sCompFail_info_html_arrow = '<span style="color:' + sColor_problem + ';"><sup>' + str(len(lstCompFail)) + '</sup></span>'
                        sCompFail_info_normal = '!' + str(len(lstCompFail))
                    elif bCB_Export_tweak:
                        # Tweak
                        sColor_problem = sColor_orange
                        sCompFail_info_html_popup = '<hr style="border-top: dotted 1px;"><span style="color:'+sColor_problem+';"><b>Überwundende Inkompatabilität</b><br><em>' + PrettyText.lst2strC(lstCompFail) + '</em></span>'
                        sCompFail_info_html_arrow = '<span style="color:'+sColor_problem+';"><sup>'+str(len(lstCompFail))+'</sup></span>'
                        sCompFail_info_normal = '!'+str(len(lstCompFail))

                # Erzeuge Modus-Pfeil
                if bCB_Export_direct:
                    # Direkt
                    sInfo_CB_Export_modus = "-direkt->"
                    if bHtml and (self.cb_export_hasNotes_process(sTSheet, sTCol) or self.cb_export_hasNotes_internal(sTSheet, sTCol) or bCompFail):
                        sPopup = "<strong>direkter Export</strong><br>"
                        if sCB_Export_notes_process is not None:
                            sPopup += "<em>" + str(sCB_Export_notes_process) + "</em><br>"
                        if sCB_Export_notes_internal is not None:
                            sPopup += "<sup><em><small>(" + str(sCB_Export_notes_internal) + ")</sup></em></small>"
                        sPopup += sCompFail_info_html_popup
                        sInfo_CB_Export_modus = '<div style="color:'+sColor_bsh_blue_light+';" class="tooltip">-direkt->'+sCompFail_info_html_arrow+'<span style="color:black;" class="right">' + sPopup + '<i></i></span></div>'
                elif bCB_Export_tweak:
                    # Sonderweg
                    if bHtml and (self.cb_export_hasNotes_process(sTSheet, sTCol) or self.cb_export_hasNotes_internal(sTSheet, sTCol) or bCompFail):
                        sPopup = "<strong>Sonderlösung</strong><br>"
                        if sCB_Export_notes_process is not None:
                            sPopup += "<em>" + str(sCB_Export_notes_process) + "</em><br>"
                        if sCB_Export_notes_internal is not None:
                            sPopup += "<sup><em><small>(" + str(sCB_Export_notes_internal) +")</sup></em></small>"
                        sPopup += sCompFail_info_html_popup
                        sInfo_CB_Export_modus = '<div style="color:'+sColor_bsh_blue_mid+';" class="tooltip">-tweak->'+sCompFail_info_html_arrow+'<span style="color:black;" class="right">' + sPopup + '<i></i></span></div>'
                    else:
                        sInfo_CB_Export_modus = "-tweak->"+sCompFail_info_normal

                # *** CB Import

                # Hole Daten
                sCB_Import_ExcelSheet, sCB_Import_ExcelCol = self.cb_import_getPartner(sTSheet, sTCol)
                bCB_Import_tweak = self.cb_import_isTweaked(sTSheet, sTCol)
                sCB_Import_notes_process = self.cb_import_getNotes_process(sTSheet, sTCol)
                sCB_Import_notes_internal = self.cb_import_getNotes_internal(sTSheet, sTCol)
                bCB_Import_direct = self.cb_import_isDirect(sTSheet, sTCol)
                sCB_Import_ExcelSheetCol = (str(sCB_Import_ExcelSheet) + "." + str(sCB_Import_ExcelCol)) if not (None in [sCB_Import_ExcelSheet, sCB_Import_ExcelCol]) else ""
                bCB_Import_ExcelSheetCol_obl = self.__map.ingest_isObl(sCB_Import_ExcelSheet,sCB_Import_ExcelCol)
                sCB_Import_ExcelSheetCol_datatype = None
                bCB_Import_ExcelSheetCol_pk = self.__map.ingest_isPk(sCB_Import_ExcelSheet,
                                                                     sCB_Import_ExcelCol)
                bCB_Import_ExcelSheetCol_fk = self.__map.ingest_isFk(sCB_Import_ExcelSheet,
                                                                     sCB_Import_ExcelCol)
                bCB_Import_ExcelSheetCol_lup_ing = self.__map.ingest_isLookup(sCB_Import_ExcelSheet,
                                                                              sCB_Import_ExcelCol)
                sCB_Import_ExcelSheetCol_lup_ing_table = self.__map.ingest_getLookup(sCB_Import_ExcelSheet,
                                                                                     sCB_Import_ExcelCol)
                bCB_Import_ExcelSheetCol_lup_itself = (
                            self.__map.all_getSheetSystem(sCB_Import_ExcelSheet) == "lookup")
                sCB_Import_ExcelSheetCol_lup_itself_table = self.__map.lookup_sheet2table(
                    sCB_Import_ExcelSheet)
                bCB_Import_ExcelSheetCol_geo = self.__map.ingest_isGeo(sCB_Import_ExcelSheet,
                                                                       sCB_Import_ExcelCol)
                sCB_Import_ExcelSheetCol_geo = self.__map.ingest_getGeo(sCB_Import_ExcelSheet,
                                                                        sCB_Import_ExcelCol)
                bCB_Import_ExcelSheetCol_date = self.__map.ingest_isDate(sCB_Import_ExcelSheet,
                                                                         sCB_Import_ExcelCol)
                bCB_Import_ExcelSheetCol_time = self.__map.ingest_isTime(sCB_Import_ExcelSheet,
                                                                         sCB_Import_ExcelCol)
                resSchema = self.__model.conf_getDbSchema().getDbSchemaInfo(
                    self.__map.all_getSheetSystem(sCB_Import_ExcelSheet), sCB_Import_ExcelSheet,
                    sCB_Import_ExcelCol)
                sCB_Import_ExcelSheetCol_name = resSchema.getData("comment-part-name")
                sCB_Import_ExcelSheetCol_comment = resSchema.getData("comment-part-comment")
                sCB_Import_ExcelSheetCol_unit = resSchema.getData("comment-part-unit")
                sCB_Import_ExcelSheetCol_datatype = resSchema.getData("datatype")
                lstInfo_CbImport_SheetCol_popup = []
                if resSchema.isOk():
                    if bCB_Import_ExcelSheetCol_pk:
                        lstInfo_CbImport_SheetCol_popup.append("pk")
                    if bCB_Import_ExcelSheetCol_fk:
                        lstInfo_CbImport_SheetCol_popup.append("fk")
                    if bCB_Import_ExcelSheetCol_obl:
                        lstInfo_CbImport_SheetCol_popup.append("obl")
                    if bCB_Import_ExcelSheetCol_lup_ing:
                        lstInfo_CbImport_SheetCol_popup.append(
                            "->lup(" + str(sCB_Import_ExcelSheetCol_lup_ing_table) + ")")
                    if bCB_Import_ExcelSheetCol_lup_itself:
                        lstInfo_CbImport_SheetCol_popup.append(
                            "lup(" + str(sCB_Import_ExcelSheetCol_lup_itself_table) + ")")
                    if bCB_Import_ExcelSheetCol_date:
                        lstInfo_CbImport_SheetCol_popup.append("date")
                    if bCB_Import_ExcelSheetCol_time:
                        lstInfo_CbImport_SheetCol_popup.append("time")
                    if bCB_Import_ExcelSheetCol_geo:
                        lstInfo_CbImport_SheetCol_popup.append(
                            "geo(" + str(sCB_Import_ExcelSheetCol_geo) + ")")

                # Info: Sheet/Col
                sInfo_CB_Import_SheetCol = ""
                sInfo_CbImport_SheetCol_popup = ""
                sCB_Import_ExcelSheetCol_popup_opt = "<sup><small>{datatype: " + str(
                    sCB_Import_ExcelSheetCol_datatype) + "}"
                if len(lstInfo_CbImport_SheetCol_popup) > 0:
                    sCB_Import_ExcelSheetCol_popup_opt += " {options: " + PrettyText.lst2strC(
                        lstInfo_CbImport_SheetCol_popup) + "}"
                sCB_Import_ExcelSheetCol_popup_opt += "</sup></small>"
                sInfo_CbImport_SheetCol_popup += "<strong>" + sCB_Import_ExcelSheetCol_name + "</strong><br>" + sCB_Import_ExcelSheetCol_popup_opt + "<br>" if sCB_Import_ExcelSheetCol_name not in [
                    None, ""] else ""
                sInfo_CbImport_SheetCol_popup += "[" + sCB_Import_ExcelSheetCol_unit + "]<br>" if sCB_Import_ExcelSheetCol_unit not in [
                    None, ""] else ""
                sInfo_CbImport_SheetCol_popup += "<em>" + sCB_Import_ExcelSheetCol_comment + "</em><br>" if sCB_Import_ExcelSheetCol_comment not in [
                    None, ""] else ""
                if None in [sCB_Import_ExcelSheet, sCB_Import_ExcelCol]:
                    sInfo_CB_Import_SheetCol = ""
                else:
                    if sInfo_CbImport_SheetCol_popup not in [None, ""] and bHtml:
                        sInfo_CB_Import_SheetCol = '<div class="tooltip">' + sCB_Import_ExcelSheetCol + '<span class="right">' + sInfo_CbImport_SheetCol_popup + '<i></i></span></div>'
                    else:
                        sInfo_CB_Import_SheetCol = sCB_Import_ExcelSheetCol

                # Info: Import Modus
                sInfo_CB_Import_modus = ""
                sCompFail_info_normal = ""
                sCompFail_info_html_popup = ""
                sCompFail_info_html_arrow = ""

                # Teste Inkompatabilität
                bCompFail = self.cb_import_isCompProblem(sTSheet, sTCol)
                lstCompFail, sT, sCb = self.cb_import_getCompProblem(sTSheet, sTCol)
                sCompFail_info_normal = ""
                if bCompFail:
                    # Inkompatibilität gefunden
                    if bCB_Import_direct:
                        # Normal
                        sColor_problem = sColor_red
                        sCompFail_info_html_popup = '<hr style="border-top: dotted 1px;"><span style="color:' + sColor_problem + ';"><b>Inkompatabilität</b><br><em>' + PrettyText.lst2strC(lstCompFail) + '</em></span>'
                        sCompFail_info_html_arrow = '<span style="color:' + sColor_problem + ';"><sup>' + str(len(lstCompFail)) + '</sup></span>'
                        sCompFail_info_normal = '!' + str(len(lstCompFail))
                    elif bCB_Import_tweak:
                        # Tweak
                        sColor_problem = sColor_orange
                        sCompFail_info_html_popup = '<hr style="border-top: dotted 1px;"><span style="color:' + sColor_problem + ';"><b>Überwundene Inkompatabilität</b><br><em>' + PrettyText.lst2strC(lstCompFail) + '</em></span>'
                        sCompFail_info_html_arrow = '<span style="color:' + sColor_problem + ';"><sup>' + str(len(lstCompFail)) + '</sup></span>'
                        sCompFail_info_normal = '!' + str(len(lstCompFail))

                # Erzeuge Modus-Pfeil
                if bCB_Import_direct:
                    # Direkt
                    sInfo_CB_Import_modus = "<-direkt-"+sCompFail_info_normal
                    if bHtml and (self.cb_import_hasNotes_process(sTSheet,sTCol) or self.cb_import_hasNotes_internal(sTSheet, sTCol) or bCompFail):
                        sPopup = "<strong>direkter Import</strong><br>"
                        if sCB_Import_notes_process is not None:
                            sPopup += "<em>" + str(sCB_Import_notes_process) + "</em><br>"
                        if sCB_Import_notes_internal is not None:
                            sPopup += "<sup><em><small>(" + str(
                                sCB_Import_notes_internal) + ")</sup></em></small>"
                        sPopup += sCompFail_info_html_popup
                        sInfo_CB_Import_modus = '<div style="color:' + sColor_bsh_blue_light + ';" class="tooltip"><-direkt-'+sCompFail_info_html_arrow+'<span style="color:black;" class="right">' + sPopup + '<i></i></span></div>'
                elif bCB_Import_tweak:
                    # Sonderweg
                    if bHtml and (self.cb_import_hasNotes_process(sTSheet,sTCol) or self.cb_import_hasNotes_internal(sTSheet, sTCol) or bCompFail):
                        sPopup = "<strong>Sonderlösung</strong><br>"
                        if sCB_Import_notes_process is not None:
                            sPopup += "<em>" + str(sCB_Import_notes_process) + "</em><br>"
                        if sCB_Import_notes_internal is not None:
                            sPopup += "<sup><em><small>(" + str(
                                sCB_Import_notes_internal) + ")</sup></em></small>"
                        sPopup += sCompFail_info_html_popup
                        sInfo_CB_Import_modus = '<div style="color:' + sColor_bsh_blue_mid + ';" class="tooltip"><-tweak-'+sCompFail_info_html_arrow+'<span style="color:black;" class="right">' + sPopup + '<i></i></span></div>'
                    else:
                        sInfo_CB_Import_modus = "<-tweak-"+sCompFail_info_normal

                # ------------------------------------------------------------------------------------------------------
                # MARLIN
                # ------------------------------------------------------------------------------------------------------

                # *** MARLIN Import 

                bM_Import_tweak = self.m_import_isTweaked(sTSheet, sTCol)
                bM_Import_direct = self.m_import_isDirect(sTSheet, sTCol)
                sM_Import_DbTabl, sM_Import_DbAttr = self.m_import_getPartner(sTSheet, sTCol)
                sM_Import_DbTablAttr = (str(sM_Import_DbTabl) + "." + str(sM_Import_DbAttr)) if not (None in [sM_Import_DbTabl, sM_Import_DbAttr]) else None
                sM_Import_notes_general = self.m_import_getNotes_general(sTSheet, sTCol)
                sM_Import_notes_process = self.m_import_getNotes_process(sTSheet, sTCol)
                sM_Import_notes_internal = self.m_import_getNotes_internal(sTSheet, sTCol)

                # Info: Table/Attribut
                sInfo_M_Import_TableAttrib = ""
                sInfo_M_Import_TableAttrib_popup = "<strong>" + sM_Import_DbTabl +"."+sM_Import_DbAttr+"</strong><br>" if (sM_Import_DbTabl not in [None, ""] and sM_Import_DbAttr not in [None, ""])  else ""
                if sM_Import_notes_general not in [None, ""]:
                    sInfo_M_Import_TableAttrib_popup += "<em>" +sM_Import_notes_general + "</em>"

                if sM_Import_DbTablAttr in [None, ""]:
                    sInfo_M_Import_TableAttrib = ""
                else:
                    if (sInfo_M_Import_TableAttrib_popup not in [None, ""]) and bHtml:
                        sInfo_M_Import_TableAttrib = '<div class="tooltip">' + sM_Import_DbTablAttr + '<span class="right">' + sInfo_M_Import_TableAttrib_popup + '<i></i></span></div>'
                    else:
                        sInfo_M_Import_TableAttrib = sM_Import_DbTablAttr

                # Info: Import Modus
                sInfo_M_Import_Modus = ""
                if bM_Import_direct:
                    # Direkt
                    sInfo_M_Import_Modus = "-direkt->"
                    if bHtml and (self.m_import_hasNotes_process(sTSheet, sTCol) or self.m_import_hasNotes_internal(sTSheet, sTCol)):
                        sPopup = "<strong>direkter Import</strong><br>"
                        if sM_Import_notes_process is not None:
                            sPopup += "<em>" + str(sM_Import_notes_process) + "</em><br>"
                        if sM_Import_notes_internal is not None:
                            sPopup += "<sup><em><small>(" + str(sM_Import_notes_internal) + ")</sup></em></small>"
                        sInfo_M_Import_Modus = '<div style="color:'+sColor_bsh_blue_light+';" class="tooltip">-direkt-><span style="color:black;" class="right">' + sPopup + '<i></i></span></div>'
                elif bM_Import_tweak:
                    # Sonderweg
                    if bHtml and (self.m_import_hasNotes_process(sTSheet, sTCol) or self.m_import_hasNotes_internal(sTSheet, sTCol)):
                        sPopup = "<strong>Sonderlösung</strong><br>"
                        if sM_Import_notes_process is not None:
                            sPopup += "<em>" + str(sM_Import_notes_process) + "</em><br>"
                        if sM_Import_notes_internal is not None:
                            sPopup += "<sup><em><small>(" + str(sM_Import_notes_internal) + ")</sup></em></small>"
                        sInfo_M_Import_Modus = '<div style="color:'+sColor_blue+';" class="tooltip">-tweak-><span style="color:black;" class="right">' + sPopup + '<i></i></span></div>'
                    else:
                        sInfo_M_Import_Modus = "-tweak->"

                # *** MARLIN Export

                bM_Export_tweak = self.m_export_isTweaked(sTSheet, sTCol)
                bM_Export_direct = self.m_export_isDirect(sTSheet, sTCol)
                sM_Export_DbTabl, sM_Export_DbAttr = self.m_export_getPartner(sTSheet, sTCol)
                sM_Export_DbTablAttr = (str(sM_Export_DbTabl) + "." + str(sM_Export_DbAttr)) if not (
                            None in [sM_Export_DbTabl, sM_Export_DbAttr]) else None
                sM_Export_notes_general = self.m_export_getNotes_general(sTSheet, sTCol)
                sM_Export_notes_process = self.m_export_getNotes_process(sTSheet, sTCol)
                sM_Export_notes_internal = self.m_export_getNotes_internal(sTSheet, sTCol)

                # Info: Table/Attribut
                sInfo_M_Export_TableAttrib = ""
                sInfo_M_Export_TableAttrib_popup = "<strong>" + sM_Export_DbTabl + "." + sM_Export_DbAttr + "</strong><br>" if (
                            sM_Export_DbTabl not in [None, ""] and sM_Export_DbAttr not in [None,""]) else ""
                if sM_Export_notes_general not in [None, ""]:
                    sInfo_M_Export_TableAttrib_popup += "<em>" + sM_Export_notes_general + "</em>"
                if sM_Export_DbTablAttr in [None, ""]:
                    sInfo_M_Export_TableAttrib = ""
                else:
                    if (sInfo_M_Export_TableAttrib_popup not in [None, ""]) and bHtml:
                        sInfo_M_Export_TableAttrib = '<div class="tooltip">' + sM_Export_DbTablAttr + '<span class="right">' + sInfo_M_Export_TableAttrib_popup + '<i></i></span></div>'
                    else:
                        sInfo_M_Export_TableAttrib = sM_Export_DbTablAttr

                # Info: Export Modus
                sInfo_M_Export_Modus = ""
                if bM_Export_direct:
                    # Direkt
                    sInfo_M_Export_Modus = "<-direkt-"
                    if bHtml and (self.m_export_hasNotes_process(sTSheet,sTCol) or self.m_export_hasNotes_internal(
                            sTSheet, sTCol)):
                        sPopup = "<strong>direkter Export</strong><br>"
                        if sM_Export_notes_process is not None:
                            sPopup += "<em>" + str(sM_Export_notes_process) + "</em><br>"
                        if sM_Export_notes_internal is not None:
                            sPopup += "<sup><em><small>(" + str(
                                sM_Export_notes_internal) + ")</sup></em></small>"
                        sInfo_M_Export_Modus = '<div style="color:' + sColor_bsh_blue_light + ';" class="tooltip"><-direkt-<span style="color:black;" class="right">' + sPopup + '<i></i></span></div>'
                elif bM_Export_tweak:
                    # Sonderweg
                    if bHtml and (self.m_export_hasNotes_process(sTSheet,sTCol) or self.m_export_hasNotes_internal(
                            sTSheet, sTCol)):
                        sPopup = "<strong>Sonderlösung</strong><br>"
                        if sM_Export_notes_process is not None:
                            sPopup += "<em>" + str(sM_Export_notes_process) + "</em><br>"
                        if sM_Export_notes_internal is not None:
                            sPopup += "<sup><em><small>(" + str(sM_Export_notes_internal) + ")</sup></em></small>"
                        sInfo_M_Export_Modus = '<div style="color:' + sColor_blue + ';" class="tooltip"><-tweak-<span style="color:black;" class="right">' + sPopup + '<i></i></span></div>'
                    else:
                        sInfo_M_Export_Modus = "<-tweak-"

                # ------------------------------------------------------------------------------------------------------
                # Tabellenzeile
                # ------------------------------------------------------------------------------------------------------

                # Format aufbauen
                if bHtml:
                    # HTML (Problem: Zusammenspiel: Doppeleintrag in einer Zeiel in zwei Spalten)
                    def tuplePrint2(a, b):
                        if a in [None, ""]:
                            a = "&nbsp;"
                        if b in [None, ""]:
                            b = "&nbsp;"
                        sOut = str(a) + "<br>" + str(b)
                        return sOut

                    def tuplePrint(a, b):
                        lstTmp = []
                        if a is not None:
                            lstTmp.append(a)
                        if b is not None:
                            lstTmp.append(b)
                        sOut = ""
                        if len(lstTmp) == 1:
                            sOut = str(lstTmp[0])
                        elif len(lstTmp) == 2:
                            sOut = str(a) + "<br>" + str(b)
                        return sOut
                    def tupleCheck(a, b, c, d):
                        if a == "":
                            a = None
                        if b == "":
                            b = None
                        if c == "":
                            c = None
                        if d == "":
                            d = None
                        if a is None and b is not None:
                            a = "&nbsp;"
                        if b is None and a is not None:
                            b = "&nbsp;"
                        if c is None and d is not None:
                            c = "&nbsp;"
                        if d is None and c is not None:
                            d = "&nbsp;"
                        return a, b, c, d
                    # CB
                    if sCB_Export_ExcelSheetCol == sCB_Import_ExcelSheetCol:
                        sInfo_CB_ImportExport_SheetCol =  sInfo_CB_Export_SheetCol
                    else:
                        sInfo_CB_ImportExport_SheetCol = tuplePrint(sInfo_CB_Export_SheetCol, sInfo_CB_Import_SheetCol)
                    sInfo_CB_ImportExport_modus = tuplePrint(sInfo_CB_Export_modus, sInfo_CB_Import_modus)
                    # M
                    sInfo_M_Import_Modus, sInfo_M_Import_TableAttrib, sInfo_M_Export_Modus, sInfo_M_Export_TableAttrib = tupleCheck(sInfo_M_Import_Modus, sInfo_M_Import_TableAttrib, sInfo_M_Export_Modus, sInfo_M_Export_TableAttrib)
                    sInfo_M_ImportExport_TableAttrib = tuplePrint(sInfo_M_Import_TableAttrib, sInfo_M_Export_TableAttrib)
                    sInfo_M_ImportExport_Modus = tuplePrint(sInfo_M_Import_Modus, sInfo_M_Export_Modus)
                    # Eine Zeile
                    lstTmp.append((sInfo_CB_ImportExport_SheetCol, sInfo_CB_ImportExport_modus, sInfo_Trans_SheetCol, sInfo_M_ImportExport_Modus, sInfo_M_ImportExport_TableAttrib))
                else:
                    # Normal
                    # Zwei Zeilen
                    lstTmp.append((sInfo_CB_Export_SheetCol, sInfo_CB_Export_modus, sInfo_Trans_SheetCol, sInfo_M_Import_Modus, sInfo_M_Import_TableAttrib))
                    lstTmp.append((sInfo_CB_Import_SheetCol, sInfo_CB_Import_modus, sInfo_Trans_SheetCol, sInfo_M_Export_Modus, sInfo_M_Export_TableAttrib))

        sOut += PrettyText.table(lstTmp, True, bHtml)
        sOut += "\n\n"

        odicQuestions = self.trans_getQuestion_all()
        sOut += "*** Offene Fragen\n\n"
        sOut += "* Transportformat ("+str(len(odicQuestions.keys()))+"x)\n\n"
        lstTable = [["T Sheet.Col", "Frage"]]
        for (sT_sheet, sT_col), sQuestion in odicQuestions.items():
            sSheetCol = sT_sheet+"."+sT_col
            lstTable.append([sSheetCol, sQuestion])
        sOut += PrettyText.table(lstTable, True, bHtml, preview=iPreview)
        sOut += "\n\n"

        # --- Transportwege nicht möglich

        sOut += "*** Nicht funktionierende Wege\n\n"
        # CB-!->T
        if self.total_notWorking_CB2T() > 0:
            sOut += "--- CB -!-> T\n\n"
            lstSubset = self.trans_getAll(side="cb", dir="import", partner=False, tweak=False)
            if len(lstSubset) > 0:
                sOut += "* Grund: Fehlender Partner in CB (" + str(len(lstSubset)) + "x)\n\n"
                lstTmp = [["T Sheet.Col"]]
                for sTSheet, sCol in lstSubset:
                    lstTmp.append([str(sTSheet) + "." + str(sCol)])
                sOut += PrettyText.table(lstTmp, True, bHtml, preview=iPreview)
                sOut += "\n\n"
            lstSubset = self.trans_getAll(side="cb", dir="import", comp=False, tweak=False)
            if len(lstSubset) > 0:
                sOut += "* Grund: Kompatibilitäsprobleme CB->T (" + str(len(lstSubset)) + "x)\n\n"
                lstTmp = [["T Sheet.Col", "Problem", "CB", "T"]]
                lstTmp_tweak = [["T Sheet.Col", "Problem", "CB", "T"]]
                for sTSheet, sCol in lstSubset:
                    lstProblem, sCb, sT = self.cb_export_getCompProblem(sTSheet, sCol)
                    if not self.cb_export_isTweaked(sTSheet, sCol):
                        lstTmp.append([str(sTSheet) + "." + str(sCol), PrettyText.lst2strC(lstProblem), sCb, sT])
                sOut += PrettyText.table(lstTmp, True, bHtml, preview=iPreview)
                sOut += "\n\n"

        # T-!->M
        if self.total_notWorking_T2M() > 0:
            sOut += "--- T -!-> M\n\n"
            lstSubset = self.trans_getAll(side="m", dir="export", partner=False, tweak=False)
            if len(lstSubset) > 0:
                sOut += "* Grund: Fehlender Partner bei M (" + str(len(lstSubset)) + "x)\n\n"
                lstTmp = [["T Sheet.Column"]]
                for s, c in lstSubset:
                    lstTmp.append([str(s) + "." + str(c)])
                sOut += PrettyText.table(lstTmp, True, bHtml, preview=iPreview)
                sOut += "\n\n"

        # M-!->T
        if self.total_notWorking_M2T() > 0:
            sOut += "--- M -!-> T\n\n"
            lstSubset = self.trans_getAll(side="m", dir="import", partner=False, tweak=False)
            if len(lstSubset) > 0:
                sOut += "* Grund: Fehlender Partner bei M (" + str(len(lstSubset)) + "x)\n\n"
                lstTmp = [["T Sheet.Column"]]
                for s, c in lstSubset:
                    lstTmp.append([str(s) + "." + str(c)])
                sOut += PrettyText.table(lstTmp, True, bHtml, preview=iPreview)
                sOut += "\n\n"

        # T-!>CB
        if self.total_notWorking_T2CB() > 0:
            sOut += "--- T -!-> CB\n\n"
            lstSubset = self.trans_getAll(side="cb", dir="export", partner=False, tweak=False)
            if len(lstSubset) > 0:
                sOut += "* Grund: Fehlender Partner in CB (" + str(len(lstSubset)) + "x)\n\n"
                lstTmp = [["T Sheet.Col"]]
                for s, c in lstSubset:
                    lstTmp.append([str(s) + "." + str(c)])
                sOut += PrettyText.table(lstTmp, True, bHtml, preview=iPreview)
                sOut += "\n\n"
            lstSubset = self.trans_getAll(side="cb", dir="export", comp=False, tweak=False)
            if len(lstSubset) > 0:
                sOut += "* Grund: Kompatibilitäsprobleme T->CB (" + str(len(lstSubset)) + "x)\n\n"
                lstTmp = [["T Sheet.Col", "Problem", "T", "CB"]]
                for s, c in lstSubset:
                    lstProblem, sT, sCB = self.cb_import_getCompProblem(s, c)
                    lstTmp.append([str(s) + "." + str(c), PrettyText.lst2strC(lstProblem), sT, sCB])
                sOut += PrettyText.table(lstTmp, True, bHtml, preview=iPreview)
                sOut += "\n\n"

        # --- Sonderlösungen

        sOut += "*** Sonderlösungen\n"
        sOut += "\n"
        # CB -> T
        lst= self.trans_getAll(side="cb", dir="import", tweak=True)
        lstTmp = []
        sOut += "--- CB -tweak-> T\n\n"
        sOut += "* Sonderlösungen ("+str(len(lst))+"x)\n\n"
        for sTsheet, sTcol in lst:
            sCBsheet, sCBcol = self.cb_export_getPartner(sTsheet, sTcol)
            sInfo = self.cb_export_getNotes_process(sTsheet, sTcol)
            sCBsheetCol = str(sCBsheet) + "." + str(sCBcol)
            sTsheetCol = str(sTsheet) + "." + str(sTcol)
            lstTmp.append([sCBsheetCol, sTsheetCol, sInfo])
        lstTmp.insert(0, ["CB Sheet.Col", "T Sheet.Col", "Sonderlösung"])
        sOut += PrettyText.table(lstTmp, True, bHtml, preview=iPreview)
        sOut += "\n\n"
        # T -> M
        lst= self.trans_getAll(side="m", dir="export", tweak=True)
        lstTmp = []
        sOut += "--- T -tweak-> M\n\n"
        sOut += "* Sonderlösungen (" + str(len(lst)) + "x)\n\n"
        for sTsheet, sTcol in lst:
            sMsheet, sMcol = self.m_import_getPartner(sTsheet, sTcol)
            sInfo = self.m_import_getNotes_process(sTsheet, sTcol)
            if (sMsheet, sMcol) != (None, None):
                sMsheetCol = str(sMsheet) + "." + str(sMcol)
            else:
                sMsheetCol = ""
            sTsheetCol = str(sTsheet) + "." + str(sTcol)
            lstTmp.append([sTsheetCol, sMsheetCol, sInfo])
        lstTmp.insert(0, ["T Sheet.Col", "M View.Attrib", "Sonderlösung"])
        sOut += PrettyText.table(lstTmp, True, bHtml, preview=iPreview)
        sOut += "\n\n"
        # M -> T
        lst = self.trans_getAll(side="m", dir="import", tweak=True, preview=iPreview)
        lstTmp = []
        sOut += "--- T <-tweak- M\n\n"
        sOut += "* Sonderlösungen (" + str(len(lst)) + "x)\n\n"
        for sTsheet, sTcol in lst:
            sMsheet, sMcol = self.m_export_getPartner(sTsheet, sTcol)
            sInfo = self.m_export_getNotes_process(sTsheet, sTcol)
            if (sMsheet, sMcol) != (None, None):
                sMsheetCol = str(sMsheet) + "." + str(sMcol)
            else:
                sMsheetCol = ""
            sTsheetCol = str(sTsheet) + "." + str(sTcol)
            lstTmp.append([sMsheetCol, sTsheetCol, sInfo])
        lstTmp.insert(0, ["M SVView.Attrib", "T Sheet.Col", "Sonderlösung"])
        sOut += PrettyText.table(lstTmp, True, bHtml, preview=iPreview)
        sOut += "\n\n"
        # T -> CB
        lst = self.trans_getAll(side="cb", dir="export", tweak=True)
        lstTmp = []
        sOut += "--- CB <-tweak- T\n\n"
        sOut += "* Sonderlösungen (" + str(len(lst)) + "x)\n\n"
        for sTsheet, sTcol in lst:
            sCBsheet, sCBcol = self.cb_import_getPartner(sTsheet, sTcol)
            sInfo = self.cb_import_getNotes_process(sTsheet, sTcol)
            sCBsheetCol = str(sCBsheet) + "." + str(sCBcol)
            sTsheetCol = str(sTsheet) + "." + str(sTcol)
            lstTmp.append([sTsheetCol, sCBsheetCol, sInfo])
        lstTmp.insert(0, ["T Sheet.Col", "CB Sheet.Col", "Sonderlösung"])
        sOut += PrettyText.table(lstTmp, True, bHtml, preview=iPreview)
        sOut += "\n\n"


        lstCB_export = list(self.cb_export_getSingles())
        lstCB_import = list(self.cb_import_getSingles())
        lstM_export = self.m_export_getSingles()
        lstM_import = self.m_import_getSingles()
        lstCB_export.sort(key=lambda tup: str(tup[0])+str(tup[1]))
        lstCB_import.sort(key=lambda tup: str(tup[0])+str(tup[1]))
        lstM_export.sort(key=lambda tup: str(tup[0])+str(tup[1]))
        lstM_import.sort(key=lambda tup: str(tup[0])+str(tup[1]))
        iSingles = len(lstCB_export) + len(lstCB_import) + len(lstM_export) + len(lstM_import)
        if iSingles > 0:
            sOut += "*** Keine Partner in Transportformat\n"
            sOut += "\n"
            # CB: Nicht angeschlossen
            if len(lstCB_export) > 0:
                lstTmp = []
                sOut += "--- CB\n\n"
                sOut += "* CB -> ? (" + str(len(lstCB_export))+"x)\n\n"
                for sCBsheet, sCBcol in lstCB_export:
                    sCBsheetCol = str(sCBsheet)+"."+str(sCBcol)
                    lstTmp.append([sCBsheetCol])
                lstTmp.insert(0, ["CB Sheet.Col"])
                sOut += PrettyText.table(lstTmp, True, bHtml, preview=iPreview)
                sOut += "\n\n"
            if len(lstCB_import) > 0:
                sOut += "* CB <- ? (" + str(len(lstCB_import)) + "x)\n\n"
                for sCBsheet, sCBcol in lstCB_import:
                    sCBsheetCol = str(sCBsheet) + "." + str(sCBcol)
                    lstTmp.append([sCBsheetCol])
                lstTmp.insert(0, ["CB Sheet.Col"])
                sOut += PrettyText.table(lstTmp, True, bHtml, preview=iPreview)
                sOut += "\n\n"
            # M: Nicht angeschlossen
            if len(lstM_import) > 0:
                lstTmp = []
                sOut += "--- ? -> M (" + str(len(lstM_import)) + "x)\n\n"
                for sMsheet, sMcol in lstM_import:
                    if (sMsheet, sMcol) != (None, None):
                        sMsheetCol = str(sMsheet) + "." + str(sMcol)
                    else:
                        sMsheetCol = ""
                    lstTmp.append([sMsheetCol])
                lstTmp.insert(0, ["M View.Attrib"])
                sOut += PrettyText.table(lstTmp, True, bHtml, preview=iPreview)
                sOut += "\n\n"
            if len(lstM_export) > 0:
                lstTmp = []
                sOut += "--- ? <- M (" + str(len(lstM_export)) + "x)\n\n"
                for sMsheet, sMcol in lstM_export:
                    if (sMsheet, sMcol) != (None, None):
                        sMsheetCol = str(sMsheet) + "." + str(sMcol)
                    else:
                        sMsheetCol = ""
                    lstTmp.append([sMsheetCol])
                lstTmp.insert(0, ["M SvView.Attrib"])
                sOut += PrettyText.table(lstTmp, True, bHtml, preview=iPreview)
                sOut += "\n\n"

        # --- HTML formatieren
        if bHtml:
            # Zeilenumbrüche entfernen
            sOut_tmp = sOut.replace("\n\n\n", "\n\n")
            sOut_tmp = sOut_tmp.replace("\n\n", "\n")
            # -direkt->
            sOut_tmp = sOut_tmp.replace("-direkt->", "&rarr;")
            sOut_tmp = sOut_tmp.replace("<-direkt-", "&larr;")
            # -tweak->
            sOut_tmp = sOut_tmp.replace("-tweak->", "&#10511;")
            sOut_tmp = sOut_tmp.replace("<-tweak-", "&#10510;")
            # !!!
            sOut_tmp = sOut_tmp.replace("!!!", "&#x26A0;")
            # Pfeile einfach
            sOut_tmp = sOut_tmp.replace("-!->", "&#8677;")
            sOut_tmp = sOut_tmp.replace("<-!-", "&#8676;")
            sOut_tmp = sOut_tmp.replace("<->", "&rlarr;")
            sOut_tmp = sOut_tmp.replace("<-!->", "&#8676;") #&#8633;
            sOut_tmp = sOut_tmp.replace("->", "&rarr;")
            sOut_tmp = sOut_tmp.replace("<-", "&larr;")
            # Pfeile doppelt
            sOut_tmp = sOut_tmp.replace("=>", "&#10233;")
            sOut_tmp = sOut_tmp.replace("<=", "&#10232;")
            sOut_tmp = sOut_tmp.replace("=!=>", "&#10499;")
            sOut_tmp = sOut_tmp.replace("<=!=", "&&#10498;")
            sOut_tmp = sOut_tmp.replace("<=>", "&#10234;")
            sOut_tmp = sOut_tmp.replace("<=!=>", "&#10500;")
            # OK
            sOut_tmp = sOut_tmp.replace("[ok]", "&check;")
            # !=
            sOut_tmp = sOut_tmp.replace("!=", "&#8800;")
            # ^2 **2
            sOut_tmp = sOut_tmp.replace("**2", "<sup>2</sup>")
            # ^3 **3
            sOut_tmp = sOut_tmp.replace("**3", "<sup>3</sup>")

            sOut_tmp2 = ""
            for sLine in sOut_tmp.split("\n"):
                # *** Heading MARLIN Spezial (***, ---, *, \n -> h1 h2 h3 p)
                if len(sLine.split("*** ")) == 2:
                    sOut_tmp2 += html_h1(sLine.split("***")[1])
                elif len(sLine.split("--- ")) == 2:
                    sOut_tmp2 += html_h2(sLine.split("---")[1])
                elif len(sLine.split("* ")) == 2:
                    sOut_tmp2 += html_h3(sLine.split("*")[1])
                else:
                    sOut_tmp2 += sLine + "<p>"
            sOut_tmp = sOut_tmp2

            sOut = PrettyText.html_txt2html(sOut_tmp, title="MARLIN-CB Converter", stdfont="Arial, Helvetica, sans-serif")
            sOut += "\n\n"
        if not bHtml:
            sOut += "************************************\n"
            sOut += "Export: CB -> " + str(sWfConfLong) + "\n"
            sOut += "************************************\n\n"
            sOut += "Export all CB native workflows to " + str(sWfConfLong) + "\n"
        # Info
        return sOut

    # ---

    # Helper
    # Mapper für synonyme im M Datenmodell (z.B. str -> text)
    def typeM2CB(self, sType):
        if sType in self.__odicDef_trans_type_M2CB:
            return self.__odicDef_trans_type_M2CB[sType]
        else:
            return sType

    def unitM2CB(self, sUnit):
        if sUnit in self.__odicDef_trans_unit_M2CB:
            return self.__odicDef_trans_unit_M2CB[sUnit]
        else:
            return sUnit

    # Check: Compatapility
    def checkComp(self, sA, sB):

        # Helper: Check Loss durch Groessendifferenz
        def checkSizeLoss(sA, sB, sName):
            sA_val = sA.replace(sName, "")
            sB_val = sB.replace(sName, "")
            sA_val = sA_val.strip()
            sB_val = sB_val.strip()
            try:
                iA_val = int(sA_val)
            except:
                iA_val = None
            try:
                iB_val = int(sB_val)
            except:
                iB_val = None
            if iA_val is None and iB_val is None:
                bLoss = False
            elif iA_val is None and iB_val is not None:
                bLoss = True
            elif iA_val is not None and iB_val is None:
                bLoss = False
            elif iA_val > iB_val:
                bLoss = True
            else:
                bLoss = False
            return bLoss

        # ---

        if None in (sA, sB):
            return False, False
        sA = sA.strip()
        sB = sB.strip()
        if sA == sB:
            # Gleich
            bComp = True
            bLoss = False
        elif "int" in str(sA) and "float" in str(sB):
            # int -> float
            bComp = True
            bLoss = False
        elif "float" in str(sA) and "float" in str(sB):
            # floatX -> floatY
            bComp = True
            bLoss = checkSizeLoss(sA, sB, "float")
        elif "int" in str(sA) and "int" in str(sB):
            # intX -> intY
            bComp = True
            bLoss = checkSizeLoss(sA, sB, "int")
        else:
            # Rest
            bComp = False
            bLoss = False
        return bComp, bLoss

    # --- (Nicht-)Funktionierende Wege

    # OK
    # CB->(T)->M : CB2T + T2M
    def total_working_CB2T(self):
        return (self.trans_getAll(side="cb", dir="import", ok=True, nr=True))
    def total_working_T2M(self):
        return (self.trans_getAll(side="m", dir="export", ok=True, nr=True))
    def total_working_CB2M(self):
        return (self.trans_getAll(side="both", dir="import", ok=True, nr=True))
    # M->(T)->CB : M2T + T2CB
    def total_working_M2T(self):
        return (self.trans_getAll(side="m", dir="import", ok=True, nr=True))
    def total_working_T2CB(self):
        return (self.trans_getAll(side="cb", dir="export", ok=True, nr=True))
    def total_working_M2CB(self):
        return (self.trans_getAll(side="both", dir="export", ok=True, nr=True))

    # Nicht OK
    # CB->(T)->M : CB2T + T2M
    def total_notWorking_CB2T(self):
        return (self.trans_getAll(side="cb", dir="import", ok=False, nr=True))
    def total_notWorking_T2M(self):
        return (self.trans_getAll(side="m", dir="export", ok=False, nr=True))
    def total_notWorking_CB2M(self):
        return (self.trans_getAll(side="both", dir="import", ok=False, nr=True))
    # M->(T)->CB : M2T + T2CB
    def total_notWorking_M2T(self):
        return (self.trans_getAll(side="m", dir="import", ok=False, nr=True))
    def total_notWorking_T2CB(self):
        return (self.trans_getAll(side="cb", dir="export", ok=False, nr=True))
    def total_notWorking_M2CB(self):
        return (self.trans_getAll(side="both", dir="export", ok=False, nr=True))

    # ---

    # Getter

    # Transport
    def trans_getType(self, sSheetTrans, sColTrans):
        if (sSheetTrans,sColTrans) in self.__odicTrans_type:
            return self.__odicTrans_type[(sSheetTrans,sColTrans)]
        else:
            return None

    def trans_isGeo(self, sSheetTrans, sColTrans):
        return (self.__trans_isOpt_set(sSheetTrans, sColTrans,"lat") or self.__trans_isOpt_set(sSheetTrans, sColTrans,"lat"))

    def trans_getGeo(self, sSheetTrans, sColTrans):
        if self.__trans_isOpt_set(sSheetTrans, sColTrans,"lat"):
            return "lat"
        elif self.__trans_isOpt_set(sSheetTrans, sColTrans,"lon"):
            return "lon"
        else:
            return None

    def trans_isLat(self, sSheetTrans, sColTrans):
        return self.__trans_isOpt_set(sSheetTrans, sColTrans,"lat")

    def trans_isLon(self, sSheetTrans, sColTrans):
        return self.__trans_isOpt_set(sSheetTrans, sColTrans,"lon")

    def trans_isObl(self, sSheetTrans, sColTrans):
        return self.__trans_isOpt_set(sSheetTrans, sColTrans,"obl")

    def trans_isCobl(self, sSheetTrans, sColTrans):
        return self.__trans_isOpt_set(sSheetTrans, sColTrans,"cobl")

    def __trans_isOpt_set(self, sSheetTrans, sColTrans, sType):
        if self.trans_isSheetColExisting(sSheetTrans,sColTrans):
            if (sSheetTrans,sColTrans) in self.__setTrans_opt:
                if sType in self.__setTrans_opt[(sSheetTrans,sColTrans)]:
                    return True
            return False
        return None

    def trans_getOpts_allowed(self):
        return self.__odicDef_trans_opts_allowed.keys()

    def trans_getOpts_short2long(self, sShort):
        if sShort in self.__odicDef_trans_opts_allowed:
            return self.__odicDef_trans_opts_allowed[sShort]
        else:
            None

    def trans_getOpts(self, sSheetTrans, sColTrans):
        if self.trans_isSheetColExisting(sSheetTrans, sColTrans):
            if (sSheetTrans,sColTrans) in self.__setTrans_opt:
                return self.__setTrans_opt[(sSheetTrans,sColTrans)]
            return []
        return None

    def trans_getUnit(self, sSheetTrans, sColTrans):
        if (sSheetTrans, sColTrans) in self.__odicTrans_unit:
            return self.__odicTrans_unit[(sSheetTrans, sColTrans)]
        else:
            return None

    def trans_isSheetColExisting(self, sSheetTrans, sColTrans):
        if sSheetTrans in self.__odicTrans_sheet2cols:
            if sColTrans in self.__odicTrans_sheet2cols[sSheetTrans]:
                return True
        return False

    def trans_getSheet_counterpart(self):
        return list(self.__odicTrans_sheet2cols.keys())

    def trans_getSheets(self):
        return list(self.__odicTrans_sheet2cols.keys())

    def trans_getCols(self, sSheetTrans):
        if sSheetTrans in self.__odicTrans_sheet2cols:
            return list(self.__odicTrans_sheet2cols[sSheetTrans])
        else:
            return None

    def trans_getCounterSheet(self, sSheet, sSide="cb"):
        for sS_cb, sS_t in self.__lstSheet_counterParts:
            if sSide == "t":
                if sSheet == sS_cb:
                    return sS_t
            elif sSide == "cb":
                if sSheet == sS_t:
                    return sS_cb
        return  None

    def trans_getAll(self, **kwargs):
        """
        Hole ein Subset aller T Felder

        Kwargs:
            side (str): Welche Partnerseite soll betrachtet werden?: cb/m/both
            direct (bool): Welche Richtung (von T aus gesehen, von full aus gesehen: import=cb->ts) soll betrachtet werden?: import/export
            partner (bool): Partner vorhanden?
            tweak (bool) Sonderweg?
            comp (bool): Kampatibel?
            ok (bool): Weg ist möglich?
            nr (bool): Nur Anzahl
        """

        sSide = kwargs["side"] if "side" in kwargs else None
        sDirection = kwargs["dir"] if "dir" in kwargs else None
        bPartner = kwargs["partner"] if "partner" in kwargs else None
        bTweak = kwargs["tweak"] if "tweak" in kwargs else None
        bComp = kwargs["comp"] if "comp" in kwargs else None
        bOk = kwargs["ok"] if "ok" in kwargs else None
        bNumber = kwargs["nr"] if "nr" in kwargs else False

        # Tests
        if sSide not in [None, "cb", "m", "both"]:
            return None
        if sDirection not in [None, "import", "export"]:
            return None
        if sSide is None and sDirection is not None :
            return None
        if sSide is not None and sDirection is None:
            return None
        if (bPartner is not None or bComp is not None or bTweak) and (sSide is None or sDirection is None):
            return None
        if sSide == "m" and (bComp is not None):
            return None
        if sSide == "both" and sDirection is None:
            return None
        # Hole alle T Einträge
        lstAll = []
        for sSheet in self.trans_getSheets():
            for sCol in self.trans_getCols(sSheet):
                lstAll.append((sSheet,sCol))

        # Antwort
        lstReturn = []
        if sSide is None:
            # *** Sonderfall: Alles zurückgeben
            if bNumber:
                return len(lstAll)
            else:
                return lstAll.copy()
        elif sSide == "both":
            # *** Sonderfall: Voller Durchlauf CB<->T<->M
            if sDirection == "export":
                fkt_cb = self.cb_export_isOk
                fkt_m = self.m_import_isOk
            elif sDirection == "import":
                fkt_cb = self.cb_import_isOk
                fkt_m = self.m_export_isOk
            else:
                return None
            for sSheet, sCol in lstAll:
                if bOk is True:
                    # OK
                    bAdd = True
                    if fkt_cb(sSheet, sCol) is False:
                        bAdd = False
                    if fkt_m(sSheet, sCol) is False:
                        bAdd = False
                elif bOk is False:
                    # Nicht OK
                    bAdd = False
                    if fkt_cb(sSheet, sCol) is False:
                        bAdd = True
                    if fkt_m(sSheet, sCol) is False:
                        bAdd = True
                else:
                    bAdd = True
                # Hinzufügen
                if bAdd:
                    lstReturn.append((sSheet,sCol))
        else:
            # *** Spezieller Teilabschnitt
            partner = None
            tweak = None
            comp = None
            if sSide == "cb":
                if sDirection == "export":
                    partner = self.cb_import_isPartner
                    tweak = self.cb_import_isTweaked
                    comp = self.cb_import_isCompProblem
                elif sDirection == "import":
                    partner = self.cb_export_isPartner
                    tweak = self.cb_export_isTweaked
                    comp = self.cb_export_isCompProblem
            elif sSide == "m":
                if sDirection == "export":
                    partner = self.m_import_isPartner
                    tweak = self.m_import_isTweaked
                elif sDirection == "import":
                    partner = self.m_export_isPartner
                    tweak = self.m_export_isTweaked
            # Loop
            for sSheet, sCol in lstAll:
                if bOk is not None:
                    # Grobe Selektion: OK (Austausch ist möglich oder nicht)
                    # OK = True
                    bAdd = False
                    if not tweak(sSheet, sCol) and partner(sSheet, sCol) and (comp is None or not comp(sSheet, sCol)):
                        bAdd = True
                    if tweak(sSheet, sCol):
                        bAdd = True
                    # OK = False -> OK einfach umdrehen
                    if bOk is False:
                        bAdd = not bAdd
                else:
                    # Feine Selektion: ohne OK
                    bAdd = True
                    if bPartner is True and not partner(sSheet,sCol):
                        bAdd = False
                    if bPartner is False and partner(sSheet,sCol):
                        bAdd = False
                    if bTweak is True and not tweak(sSheet,sCol):
                        bAdd = False
                    if bTweak is False and tweak(sSheet,sCol):
                        bAdd = False
                    if bComp is True and comp(sSheet, sCol):
                        bAdd = False
                    if bComp is False and not comp(sSheet, sCol):
                        bAdd = False
                    if bComp is True and bTweak is True and comp(sSheet, sCol):
                        bAdd = False
                    if bComp is False and bTweak is True and not comp(sSheet, sCol):
                        bAdd = False
                # Hinzufügen
                if bAdd:
                    lstReturn.append((sSheet,sCol))

        # Rückgabe
        if lstReturn is not None and bNumber:
            return len(lstReturn)
        else:
            return lstReturn

    def trans_getComment(self, sSheetTrans, sColTrans):
        if (sSheetTrans, sColTrans) in self.__odicTrans_notes:
            return self.__odicTrans_notes[(sSheetTrans, sColTrans)]
        else:
            return None

    def trans_isQuestion(self, sSheetTrans, sColTrans):
        return (sSheetTrans, sColTrans) in self.__odicTrans_question

    def trans_getQuestion(self, sSheetTrans, sColTrans):
        if (sSheetTrans, sColTrans) in self.__odicTrans_question:
            return self.__odicTrans_question[(sSheetTrans, sColTrans)]
        else:
            return None

    def trans_getQuestion_all(self):
        return self.__odicTrans_question.copy()

    def trans_getInfo(self, sS, sC):
        sOut = ""
        sOut += "*** " +str(sS)+"."+str(sC)+"\n"
        sOut += "--- CB Import\n"
        sS1, sC1 =  self.cb_export_getPartner(sS, sC)
        sOut += "sh/col:  " + ((str(sS1)+"."+str(sC1)) if (sS1,sC1) != (None,None) else "-")+ "\n"
        sOut += "incomp:  " + ("yes" if self.cb_export_isCompProblem(sS, sC) else "no") +"\n"
        if self.cb_export_isCompProblem(sS, sC):
            lstIncomp, sInfo1, sInfo2 = self.cb_export_getCompProblem(sS, sC)
            sOut += "ic lst:  " + PrettyText.lst2strC(lstIncomp)+"\n"
            sOut += "ic inf1: " + str(sInfo1)+"\n"
            sOut += "ic inf2: " + str(sInfo2)+"\n"
        sOut += "partner: " + ("yes" if self.cb_export_isPartner(sS, sC) else "no") +"\n"
        sOut += "direct:  " + ("yes" if self.cb_export_isDirect(sS, sC) else "no") +"\n"
        sOut += "tweak:   " + ("yes" if self.cb_export_isTweaked(sS, sC) else "no") +"\n"
        sOut += "ok:      " + ("yes" if self.cb_export_isOk(sS, sC) else "no") +"\n"
        sOut += "--- CB Export\n"
        sS1, sC1 =  self.cb_import_getPartner(sS, sC)
        sOut += "sh/col:  " + ((str(sS1)+"."+str(sC1)) if (sS1,sC1) != (None,None) else "-")+ "\n"
        sOut += "incomp:  " + ("yes" if self.cb_import_isCompProblem(sS, sC) else "no") +"\n"
        if self.cb_import_isCompProblem(sS, sC):
            lstIncomp, sInfo1, sInfo2 = self.cb_import_getCompProblem(sS, sC)
            sOut += "ic lst:  " + PrettyText.lst2strC(lstIncomp) + "\n"
            sOut += "ic inf1: " + str(sInfo1) + "\n"
            sOut += "ic inf2: " + str(sInfo2) + "\n"
        sOut += "partner: " + ("yes" if self.cb_import_isPartner(sS, sC) else "no") +"\n"
        sOut += "direct:  " + ("yes" if self.cb_import_isDirect(sS, sC) else "no") +"\n"
        sOut += "tweak:   " + ("yes" if self.cb_import_isTweaked(sS, sC) else "no") +"\n"
        sOut += "ok:      " + ("yes" if self.cb_import_isOk(sS, sC) else "no") +"\n"
        sOut += "--- M Import\n"
        sS1, sC1 =  self.m_export_getPartner(sS, sC)
        sOut += "sh/col:  " + ((str(sS1)+"."+str(sC1)) if (sS1,sC1) != (None,None) else "-")+ "\n"
        sOut += "partner: " + ("yes" if self.m_export_isPartner(sS, sC) else "no") +"\n"
        sOut += "direct:  " + ("yes" if self.m_export_isDirect(sS, sC) else "no") +"\n"
        sOut += "tweak:   " + ("yes" if self.m_export_isTweaked(sS, sC) else "no") +"\n"
        sOut += "ok:      " + ("yes" if self.m_export_isOk(sS, sC) else "no") +"\n"
        sOut += "--- M Export\n"
        sS1, sC1 =  self.m_import_getPartner(sS, sC)
        sOut += "sh/col:  " + ((str(sS1)+"."+str(sC1)) if (sS1,sC1) != (None,None) else "-")+ "\n"
        sOut += "partner: " + ("yes" if self.m_import_isPartner(sS, sC) else "no") +"\n"
        sOut += "direct:  " + ("yes" if self.m_import_isDirect(sS, sC) else "no") +"\n"
        sOut += "tweak:   " + ("yes" if self.m_import_isTweaked(sS, sC) else "no") +"\n"
        sOut += "ok:      " + ("yes" if self.m_import_isOk(sS, sC) else "no") +"\n"
        return sOut

    # CB
    # Import
    def cb_import_getAll(self):
        return self.__lstCB_import_all

    def cb_import_isPartner(self, sSheetTrans, sColTrans):
        return (sSheetTrans, sColTrans) in self.__odicCB_import_excel

    def cb_import_getPartner(self, sSheetTrans, sColTrans):
        if (sSheetTrans, sColTrans) in self.__odicCB_import_excel:
            return self.__odicCB_import_excel[(sSheetTrans, sColTrans)]
        else:
            return None, None

    def cb_import_isTweaked(self, sSheetTrans, sColTrans):
        return (sSheetTrans,sColTrans) in self.__setCB_import_tweak

    def cb_import_isDirect(self, sSheetTrans, sColTrans):
        if self.cb_import_getPartner(sSheetTrans, sColTrans) != (None, None) and not self.cb_import_isTweaked(sSheetTrans, sColTrans):
            return True
        else:
            return False

    def cb_import_isOk(self, sSheetTrans, sColTrans):
        if self.cb_import_isDirect(sSheetTrans, sColTrans) and not self.cb_import_isCompProblem(sSheetTrans, sColTrans):
            return True
        elif self.cb_import_isTweaked(sSheetTrans, sColTrans):
            return True
        else:
            return False

    def cb_import_getSingles(self):
        return self.__setCB_import_singles.copy()

    def cb_import_hasNotes_process(self, sSheetTrans, sColTrans):
        return (sSheetTrans, sColTrans) in self.__odicCB_import_notes_process

    def cb_import_getNotes_process(self, sSheetTrans, sColTrans):
        if (sSheetTrans, sColTrans) in self.__odicCB_import_notes_process:
            return self.__odicCB_import_notes_process[(sSheetTrans, sColTrans)]
        else:
            return None

    def cb_import_hasNotes_internal(self, sSheetTrans, sColTrans):
        return (sSheetTrans, sColTrans) in self.__odicCB_import_notes_internal

    def cb_import_getNotes_internal(self, sSheetTrans, sColTrans):
        if (sSheetTrans, sColTrans) in self.__odicCB_import_notes_internal:
            return self.__odicCB_import_notes_internal[(sSheetTrans, sColTrans)]
        else:
            return None

    def cb_import_isCompProblem(self, sSheet, sCol):
        return (sSheet, sCol) in self.__odicCB_import_incomp

    def cb_import_getCompProblem(self, sSheet, sCol):
        if (sSheet, sCol) in self.__odicCB_import_incomp:
            return self.__odicCB_import_incomp[(sSheet, sCol)]
        else:
            return None, None, None

    def cb_import_getCompProblem_all(self):
        # Alle inkompatiblen Verbindungen
        return list(self.__odicCB_import_incomp.keys())

    def cb_import_getCompProblem_all_sum(self):
        # Alle Einzelinkompatibiltäten
        iNr = 0
        for (sSheet, sCol), (lstProblems, sT_details, sCB_details) in self.__odicCB_import_incomp.items():
            iNr += len(lstProblems)
        return iNr

    # Export
    def cb_export_getAll(self):
        return self.__lstCB_export_all

    def cb_export_isPartner(self, sSheetTrans, sColTrans):
        return (sSheetTrans, sColTrans) in self.__odicCB_export_excel

    def cb_export_getPartner(self, sSheetTrans, sColTrans):
        if (sSheetTrans, sColTrans) in self.__odicCB_export_excel:
            return self.__odicCB_export_excel[(sSheetTrans, sColTrans)]
        else:
            return None, None

    def cb_export_isTweaked(self, sSheetTrans, sColTrans):
        return (sSheetTrans, sColTrans) in self.__setCB_export_tweak

    def cb_export_isDirect(self, sSheetTrans, sColTrans):
        if self.cb_export_getPartner(sSheetTrans, sColTrans) != (None, None) and not self.cb_export_isTweaked(sSheetTrans, sColTrans):
            return True
        else:
            return False

    def cb_export_isOk(self, sSheetTrans, sColTrans):
        if self.cb_export_isDirect(sSheetTrans, sColTrans) and not self.cb_export_isCompProblem(sSheetTrans, sColTrans):
            return True
        elif self.cb_export_isTweaked(sSheetTrans, sColTrans):
            return True
        else:
            return False

    def cb_export_hasNotes_process(self, sSheetTrans, sColTrans):
        return (sSheetTrans, sColTrans) in self.__odicCB_export_notes_process

    def cb_export_getNotes_process(self, sSheetTrans, sColTrans):
        if (sSheetTrans, sColTrans) in self.__odicCB_export_notes_process:
            return self.__odicCB_export_notes_process[(sSheetTrans, sColTrans)]
        else:
            return None

    def cb_export_hasNotes_internal(self, sSheetTrans, sColTrans):
        return (sSheetTrans, sColTrans) in self.__odicCB_export_notes_internal

    def cb_export_getNotes_internal(self, sSheetTrans, sColTrans):
        if (sSheetTrans, sColTrans) in self.__odicCB_export_notes_internal:
            return self.__odicCB_export_notes_internal[(sSheetTrans, sColTrans)]
        else:
            return None

    def cb_export_getSingles(self):
        return self.__setCB_export_singles.copy()

    def cb_export_isCompProblem(self, sSheet, sCol):
        return (sSheet, sCol) in self.__odicCB_export_incomp

    def cb_export_getCompProblem(self, sSheet, sCol):
        if (sSheet, sCol) in self.__odicCB_export_incomp:
            return self.__odicCB_export_incomp[(sSheet, sCol)]
        else:
            return None, None, None

    def cb_export_getCompProblem_all(self):
        return list(self.__odicCB_export_incomp.keys())

    def cb_export_getCompProblem_all_sum(self):
        iNr = 0
        for (sSheet, sCol), (lstProblems, sCB_details, sT_details) in self.__odicCB_export_incomp.items():
            iNr += len(lstProblems)
        return iNr

    # MARLIN
    # Import 
    def m_import_isPartner(self, sSheetTrans, sColTrans):
        return (sSheetTrans, sColTrans) in self.__odicM_import_dbTable

    def m_import_getPartner(self, sSheetTrans, sColTrans):
        if (sSheetTrans, sColTrans) in self.__odicM_import_dbTable:
            return self.__odicM_import_dbTable[(sSheetTrans, sColTrans)]
        else:
            return None, None

    def m_import_isTweaked(self, sSheetTrans, sColTrans):
        return (sSheetTrans, sColTrans) in self.__setM_import_tweak

    def m_import_isDirect(self, sSheetTrans, sColTrans):
        if self.m_import_getPartner(sSheetTrans, sColTrans) != (None, None) and not self.m_import_isTweaked(sSheetTrans, sColTrans):
            return True
        else:
            return False

    def m_import_isOk(self, sSheetTrans, sColTrans):
        if self.m_import_isDirect(sSheetTrans, sColTrans) or self.m_import_isTweaked(sSheetTrans, sColTrans):
            return True
        else:
            return False

    def m_import_hasNotes_general(self, sSheetTrans, sColTrans):
        return (sSheetTrans, sColTrans) in self.__odicM_import_notes_general

    def m_import_getNotes_general(self, sSheetTrans, sColTrans):
        if (sSheetTrans, sColTrans) in self.__odicM_import_notes_general:
            return self.__odicM_import_notes_general[(sSheetTrans, sColTrans)]
        else:
            return None

    def m_import_hasNotes_process(self, sSheetTrans, sColTrans):
        return (sSheetTrans, sColTrans) in self.__odicM_import_notes_process

    def m_import_getNotes_process(self, sSheetTrans, sColTrans):
        if (sSheetTrans, sColTrans) in self.__odicM_import_notes_process:
            return self.__odicM_import_notes_process[(sSheetTrans, sColTrans)]
        else:
            return None

    def m_import_hasNotes_internal(self, sSheetTrans, sColTrans):
        return (sSheetTrans, sColTrans) in self.__odicM_import_notes_intenal

    def m_import_getNotes_internal(self, sSheetTrans, sColTrans):
        if (sSheetTrans, sColTrans) in self.__odicM_import_notes_intenal:
            return self.__odicM_import_notes_intenal[(sSheetTrans, sColTrans)]
        else:
            return None

    def m_import_getSingles(self):
        return self.__lstM_import_singles.copy()

    # Export 
    def m_export_isPartner(self, sSheetTrans, sColTrans):
        return (sSheetTrans, sColTrans) in self.__odicM_export_dbView

    def m_export_getPartner(self, sSheetTrans, sColTrans):
        if (sSheetTrans, sColTrans) in self.__odicM_export_dbView:
            return self.__odicM_export_dbView[(sSheetTrans, sColTrans)]
        else:
            return None, None

    def m_export_isTweaked(self, sSheetTrans, sColTrans):
        return (sSheetTrans, sColTrans) in self.__setM_export_tweak

    def m_export_isDirect(self, sSheetTrans, sColTrans):
        if self.m_export_getPartner(sSheetTrans, sColTrans) != (None, None) and not self.m_export_isTweaked(sSheetTrans, sColTrans):
            return True
        else:
            return False

    def m_export_isOk(self, sSheetTrans, sColTrans):
        if self.m_export_isDirect(sSheetTrans, sColTrans) or self.m_export_isTweaked(sSheetTrans, sColTrans):
            return True
        else:
            return False

    def m_export_hasNotes_general(self, sSheetTrans, sColTrans):
        return (sSheetTrans, sColTrans) in self.__odicM_export_notes_general

    def m_export_getNotes_general(self, sSheetTrans, sColTrans):
        if (sSheetTrans, sColTrans) in self.__odicM_export_notes_general:
            return self.__odicM_export_notes_general[(sSheetTrans, sColTrans)]
        else:
            return None

    def m_export_hasNotes_process(self, sSheetTrans, sColTrans):
        return (sSheetTrans, sColTrans) in self.__odicM_export_notes_process

    def m_export_getNotes_process(self, sSheetTrans, sColTrans):
        if (sSheetTrans, sColTrans) in self.__odicM_export_notes_process:
            return self.__odicM_export_notes_process[(sSheetTrans, sColTrans)]
        else:
            return None

    def m_export_hasNotes_internal(self, sSheetTrans, sColTrans):
        return (sSheetTrans, sColTrans) in self.__odicM_export_notes_internal

    def m_export_getNotes_internal(self, sSheetTrans, sColTrans):
        if (sSheetTrans, sColTrans) in self.__odicM_export_notes_internal:
            return self.__odicM_export_notes_internal[(sSheetTrans, sColTrans)]
        else:
            return None

    def m_export_getSingles(self):
        return self.__lstM_export_singles.copy()

    # Converter (Tweaks): Export zu MRALIN Transport
    def convert_export_tweak(self, sT_sheet, sT_col, **kwargs):

        sTweak = "\'<noch kein Sonderweg>\' AS \"" + str(sT_col) + "\", \n"
        return sTweak

    def convert_export_error(self, sT_Sheet, sT_Col, **kwargs):
        lstError = []
        bInComp = self.cb_export_isCompProblem(sT_Sheet, sT_Col)
        bPartner = self.cb_export_isPartner(sT_Sheet, sT_Col)
        if bInComp:
            lstError.append("noch inkompatibel")
        if not bPartner:
            lstError.append("kein Partner")
        sError = PrettyText.lst2strC(lstError)
        sError = "\'<"+sError+">' AS \"" + str(sT_Col) + "\", \n"
        return sError

    # Converter: Export zu MARLIN Transport
    def convert_export_create_sql(self, **kwargs):
        """
        Hier werden nur die SQL Befehle erzeugt
        Args:
            -
        Kwargs:
            -
        Return:
            Result
                data (odic)
                stat (odic)
                log (str)
        """

        # Helper: Erzeuge aus Liste eine String-Liste
        def lst2str(lst):
            if lst is None or len(lst) == 0:
                return ""
            sOut = ""
            for e in lst:
                sOut += "\'" + str(e) + "\',"
            return sOut[:-1]

        # Helper: Erzeuge eine Kurzbezeichung von Schema.tabelle -> kurz
        def short(sT):
            # Tweak: Doppelte Kürzel auflösen: status->sts/station->sta
            if sT == "status":
                return "sts"
            if sT == "subsample":
                return "subsmp"
            return str(sT)[0:3]

        # ---

        res = Result(name="Excel Converter BSH")

        pn = self.__smartprinter.normal
        pv = self.__smartprinter.verbose
        workflow = self.__smartprinter.workflow
        progress = self.__smartprinter.progress
        status = self.__smartprinter.status
        workflow(top="Converting", sub="MARLIN")
        status()
        pn("* Converter: Export to MARLIN (tranport)",fmt=False)

        # Sql
        odicSql = OrderedDict()

        # Create SQL commands.
        pn("* Creating SQL commands")

        # Db Schema
        sDbSchema = self.__model.subsystem_getDb().account_active_getSchema()

        # Log
        sLog = ""
        lstLog_ok_direct = []
        lstLog_ok_tweak = []
        lstLog_err = []

        # Loop MARLIN: Sheets
        for sT_Sheet in self.trans_getSheets():
            # SQL-Befehlsteile
            # Subset
            sSqlSelect = ""
            sSqlFrom = ""
            sSqlJoinFk = ""
            sSqlJoinLookup = ""
            sSqlWhere = "WHERE sub.id IN (SELECT id FROM selector) \n"
            sSqlOrder = ""
            sSqlUnionJoinFk = ""

            sCB_CounterSheet = self.trans_getCounterSheet(sT_Sheet)
            sCB_CounterTable = self.__map.ingest_sheet2table(sCB_CounterSheet)

            # --- Loop MARLIN: Columns
            for sT_Col in self.trans_getCols(sT_Sheet):
                # CB: Sheets/Col bzw. Table/Attrib
                sCB_Sheet, sCB_Col = self.cb_export_getPartner(sT_Sheet, sT_Col)
                sCB_Table, sCB_Attrib = self.__map.ingest_excel2db(sCB_Sheet, sCB_Col)

                # Hole alle Daten zu Feld
                bOk = self.cb_export_isOk(sT_Sheet, sT_Col)
                bDirect = self.cb_export_isDirect(sT_Sheet, sT_Col)
                bTweak = self.cb_export_isTweaked(sT_Sheet, sT_Col)
                bInComp = self.cb_export_isCompProblem(sT_Sheet, sT_Col)
                bPartner = self.cb_export_isPartner(sT_Sheet, sT_Col)
                if bInComp:
                    lstIncomp, sInComp_CB_detail, sInComp_T_detail = self.cb_export_getCompProblem(sT_Sheet, sT_Col)
                else:
                    lstIncomp = []
                    sInComp_CB_detail = sInComp_T_detail = ""
                sInComp_CB_detail = "-" if sInComp_CB_detail == "" else sInComp_CB_detail
                sInComp_T_detail = "-" if sInComp_T_detail == "" else sInComp_T_detail
                bNotes_internal = self.cb_export_hasNotes_internal(sT_Sheet, sT_Col)
                sNotes_internal = ""
                if bNotes_internal:
                    sNotes_internal = self.cb_export_getNotes_internal(sT_Sheet, sT_Col)
                bNotes_process = self.cb_export_hasNotes_process(sT_Sheet, sT_Col)
                sNotes_process = ""
                if bNotes_process:
                    sNotes_process = self.cb_export_getNotes_process(sT_Sheet, sT_Col)
                bLup = self.__map.ingest_isLookup(sCB_Sheet, sCB_Col)
                sLupTable = self.__map.ingest_getLookup(sCB_Sheet, sCB_Col) if bLup else None
                sLupAtr = self.__map.lookup_getKey(sLupTable) if bLup else None
                bFk = self.__map.ingest_isFk(sCB_Sheet, sCB_Col)
                bPk = self.__map.ingest_isPk(sCB_Sheet, sCB_Col)

                # Info
                sInfo_sheet_t = sT_Sheet +"." + sT_Col +" "
                sInfo_sheet_cb = ""
                if None not in (sCB_Sheet, sCB_Col):
                    sInfo_sheet_cb = "(CB:"+str(sCB_Sheet)+"."+str(sCB_Col)+") "
                sInfo_working = ""
                lstInfo_working = []
                lstInfo_working.append("ok" if bOk else "FEHLER")
                if bTweak:
                    lstInfo_working.append("sonderweg")
                if bDirect:
                    lstInfo_working.append("direkt")
                if len(lstInfo_working) > 0:
                    sInfo_working = "[" + PrettyText.lst2str(lstInfo_working, ".") + "] "
                lstInfo_debug_add = []
                sInfo_add = ""
                if not bPartner:
                    lstInfo_debug_add.append("(no partner)")
                if bInComp:
                    sInComp = PrettyText.lst2strC(lstIncomp) + " T:" + sInComp_T_detail + " CB:" + sInComp_CB_detail
                    if bTweak and bOk:
                        lstInfo_debug_add.append("(gelöste Inkompatibilität: "+sInComp+") ")
                    else:
                        lstInfo_debug_add.append("(Inkompatibilität: "+sInComp+") ")
                if len(lstInfo_debug_add) > 0:
                    sInfo_add = PrettyText.lst2strC(lstInfo_debug_add) + " "
                sInfo_tweak_process = ""
                if bOk and bTweak:
                    sInfo_tweak_process = "{Sonderweg: " + sNotes_process + "} "
                sInfo_gesamt = sInfo_sheet_t + sInfo_sheet_cb + sInfo_working + sInfo_add + sInfo_tweak_process
                sInfo_gesamt = PrettyText.clean_wizard(sInfo_gesamt, no_whitespaces_front_back=True, max_one_space_between_words=True, all_in_one_line=True)

                # Log
                if bOk:
                    if bTweak:
                        lstLog_ok_tweak.append(sInfo_gesamt)
                    elif bDirect:
                        lstLog_ok_direct.append(sInfo_gesamt)
                else:
                    lstLog_err.append(sInfo_gesamt)

                # Debug
                if (bPk or bFk):
                    lstDebug = []
                    if bPk:
                        lstDebug.append("PK")
                    if bFk:
                        lstDebug.append("FK")
                    if not bOk:
                        lstDebug.append("ERR")
                    if bTweak:
                        lstDebug.append("TWEAK")
                    if bLup:
                        lstDebug.append("LUP")
                    sDebug_opts = PrettyText.lst2str(lstDebug, " ")
                    sDebug = f"{sT_Sheet}.{sT_Col}(CB.counter:{sCB_CounterSheet}/{sCB_CounterTable})[CB.excel:{sCB_Sheet}.{sCB_Col}][CB.db:{sCB_Table}.{sCB_Attrib}]:{sDebug_opts}"
                    print(sDebug)

                # *** Baue die SQL Befehle zusammen
                # --- SELECT ---
                if bOk and bDirect:
                    print("->", sCB_Sheet, sCB_Col, sCB_Table, sCB_Attrib, bPk)
                    # a) OK & DIREKT
                    if bPk and not bFk:
                        # (nur) PK
                        sDb = self.__map.ingest_sheet2table(sCB_Sheet) # Richtig
                        sSqlSelect += short(sDb) + ".name AS \"" + str(sT_Col) + "\", \n"
                    elif bFk and not bPk:
                        # (nur) FK
                        # sFkDb = self.__map.ingest_sheet2table(sFk)
                        # sSqlSelection = short(sFkDb) + ".name AS \"" + str(sFk) + "\", \n" + sSqlSelection
                        print("!!!", sCB_Sheet, sCB_Col, sCB_Table, sCB_Attrib)
                        sDb = self.__map.ingest_sheet2table(sCB_Sheet) # Falsch
                        sSqlSelect += short(sDb) + ".name AS \"" + str(sT_Col) + "\", \n"
                    elif bPk and bFk:
                        # PK und FK
                        sDb = self.__map.ingest_sheet2table(sCB_Sheet) # Falsch
                        sSqlSelect += short(sDb) + ".name AS \"" + str(sT_Col) + "\", \n"
                    else:
                        #  Rest (kein FK und PK)
                        if not bLup:
                            # kein: Lookup
                            # SELECT
                            sSqlSelect += short(sCB_Table) + "." + str(sCB_Attrib) + " AS \"" + str(sT_Col) + "\", \n"
                        else:
                            # Lookup
                            # SELECT & LEFT JOIN: Lookup Attributes
                            sSqlSelect += short(sLupTable) + "." + str(sLupAtr) + " AS \"" + str(sT_Col) + "\", \n"
                            sSqlJoinLookup += "LEFT JOIN <schema>." + str(sLupTable) + " " + short(
                                sLupTable) + " ON " + short(sCB_Table) + "." + str(sCB_Attrib) + "=" + short(sLupTable) + ".id \n"
                elif bOk and bTweak:
                    # b) OK & TWEAK
                    if not bFk:
                        # --- kein FK
                        sTweak = self.convert_export_tweak(sT_Sheet, sT_Col)
                        sSqlSelect += sTweak
                    else:
                        # --- FK
                        sTweak = self.convert_export_tweak(sT_Sheet, sT_Col)
                        sSqlSelect += sTweak
                else:
                    # c) REST -> FEHLER
                    sError = self.convert_export_error(sT_Sheet, sT_Col)
                    sSqlSelect += sError

                # --- ORDER BY
                if sCB_CounterSheet in ["Cruise", "Station", "Sample", "SubSample", "Subset"] and bOk:
                    # FKs
                    if bFk:
                        sSqlOrder += "length (\"" + sT_Col + "\") ASC, \"" + sT_Col + "\" ASC, "
                    # PK
                    if bPk and not bFk:
                        sSqlOrder += "length (\"" + sT_Col + "\") ASC, \"" + sT_Col + "\" ASC, "

            # --- JOIN: FK
            if sCB_CounterSheet in ["Cruise", "Station", "Sample", "SubSample"]:
                # Tweak: Hole bei Cruise,Station,SubSample,Sample immer das Metadatengerüst Cruise-Station-Sample-SubSample-Subset hinzu
                lstFk = self.__map.ingest_getFk("Subset")
                sFkLast = "Subset"
                sFkDbLast = "subset"
            else:
                lstFk = self.__map.ingest_getFk(sCB_CounterSheet)
                sFkDbLast = sCB_CounterSheet
                sFkLast = sCB_CounterTable
            if lstFk is not None:
                for sFk in reversed(lstFk):
                    sFkDb = self.__map.ingest_sheet2table(sFk)
                    sFkMainCol, sFkMainTable, sFkMainAttrib = self.__map.ingest_getFkMain(sFkLast)
                    sTmp = "JOIN <schema>." + str(sFkDb) + " " + short(sFkDb) + " ON " + short(
                        sFkDb) + ".id = " + short(sFkDbLast) + "." + sFkMainAttrib + "\n"
                    sSqlJoinFk += sTmp
                    sFkDbLast = sFkDb
                    sFkLast = sFk

            # --- FROM: Haupttabelle
            if sCB_CounterSheet in ["Cruise", "Station", "SubSample", "Sample"]:
                sSqlFrom = "FROM <schema>.subset " + short("subset") + "\n"
            else:
                sSqlFrom = "FROM <schema>." + sCB_CounterTable + " " + short(sCB_Table) + "\n"

            # ---

            # Sql-Zusammenbauen
            # IndMeasure -> Kein Distinct -> Es gibt nämlich mehrfache gleiche Einträge pro Biota
            sSqlSelect = "SELECT "+("DISTINCT" if sCB_CounterSheet != "IndividualMeasure" else "") + "\n" + sSqlSelect[:-3] + "\n"
            if len(sSqlOrder) > 0:
                sSqlOrder = "ORDER BY " + sSqlOrder[:-2] + "\n"
            sSql = sSqlSelect + sSqlFrom + sSqlJoinFk + sSqlJoinLookup + sSqlWhere
            sSql = sSql.replace("<schema>", sDbSchema)
            sSql = "SELECT * FROM (\n"+sSql+") t1\n" + sSqlOrder
            odicSql[sT_Sheet] = sSql
            print("\n",sSql)
        # Log
        sLog = ""
        sLog += "* Converter MARLIN\n"
        lstTmp = (
            (lstLog_ok_direct, "Konvertierung funktioniert direkt"),
            (lstLog_ok_tweak, "Konvertierung funktioniert mittels Sonderlösung"),
            (lstLog_err, "Konvertierung funktioniert nicht"),
        )
        sLog += "* Konvertierungen\n"
        for lst, sName in lstTmp:
            if len(lst) > 0:
                sLog += "--- " + sName + " ("+str(len(lst)) + "x)\n"
                sLog += PrettyText.lst2strR(lst) + "\n\n"
        # Ergebnis
        res.addData("log", sLog)
        res.addData("sql", odicSql)
        #quit()
        return res

    # Converter: Import MARLIN Transport -> MARLIN native
    def convert_import(self, wb, **kwargs):
        """
        Args:
            wb (workbook): Workbook
        Kwargs:
            result_in: Result in
        Return:
            Result
                log (str)
        """

        pn = self.__smartprinter.normal
        pv = self.__smartprinter.verbose
        workflow = self.__smartprinter.workflow
        progress = self.__smartprinter.progress
        status = self.__smartprinter.status

        workflow(top="Converting", sub="MARLIN")
        status()
        pn("* Converting Excel Workbook: BSH")

        res = Result(name="Excel Converter BSH")

        # Daten
        odicData = OrderedDict()
        odicStat = OrderedDict()

        # Log
        sLog = ""
        sLog += "* Converter MARLIN\n"
        sLog += "work in progress\n"

        pn("* Converter BSH")

        # Ergebnis
        res.addData("log", sLog)
        return res

    def convert_export_create_excel_wb(self, **kwargs):
        res = Result("MARLIN (Transport)")

        # Parameter
        wb = kwargs.get("wb")
        dicData = kwargs.get("data")

        # Wb
        if wb is None:
            wb = openpyxl.Workbook()
            bFirstSheet = True
        else:
            bFirstSheet = False

        # Loop
        for iIdx, sTSheet in enumerate(self.trans_getSheets()):
            # Sheet
            if bFirstSheet:
                ws = wb.active
                ws.title = sTSheet
                bFirstSheet = False
            else:
                ws = wb.create_sheet(title=sTSheet, index=iIdx+1)

            # Header
            lstHeader = []
            for sTCol in self.trans_getCols(sTSheet):
                # Normal
                cell = openpyxl.cell.WriteOnlyCell(ws, value=sTCol)
                # Format
                bObl = self.trans_isObl(sTSheet, sTCol)
                bBold = False
                bItalic = False
                bUnderline = False
                if bObl:
                    bUnderline = True
                cell.font = openpyxl.styles.Font(italic=bItalic, bold=bBold,
                                                 underline=("single" if bUnderline else None))
                # Hole Daten
                bTrans_Obl = self.trans_isObl(sTSheet, sTCol)
                bTrans_Cobl = self.trans_isCobl(sTSheet, sTCol)
                sTrans_Type = self.trans_getType(sTSheet, sTCol)
                bTrans_Geo = self.trans_isGeo(sTSheet, sTCol)
                sTrans_Geo = self.trans_getGeo(sTSheet, sTCol)
                bTrans_Lat = self.trans_isLat(sTSheet, sTCol)
                bTrans_Lon = self.trans_isLon(sTSheet, sTCol)
                sTrans_Unit = self.trans_getUnit(sTSheet, sTCol)
                sTrans_Comment = self.trans_getComment(sTSheet, sTCol)
                sTrans_Question = self.trans_getQuestion(sTSheet, sTCol)
                lstInfo_Trans_Opt = self.trans_getOpts(sTSheet, sTCol)
                # Info
                sInfo = ""
                sInfo_Trans_SheetCol = str(sTSheet) + "." + str(sTCol) + "\n"
                sInfo_Trans_opt = "options: " + PrettyText.lst2strC(lstInfo_Trans_Opt) + "\n" if len(
                    lstInfo_Trans_Opt) > 0 else ""
                sInfo_Trans_type = "datatype: " + sTrans_Type + "\n" if sTrans_Type not in ["", None] else ""
                sInfo_Trans_unit = "unit: [" + sTrans_Unit + "]\n" if sTrans_Unit not in ["", None] else ""
                sInfo_Trans_comment = "comment: " + sTrans_Comment + "\n" if sTrans_Comment not in [None, ""] else ""
                sInfo += sInfo_Trans_SheetCol
                sInfo += sInfo_Trans_opt
                sInfo += sInfo_Trans_type
                sInfo += sInfo_Trans_unit
                sInfo += sInfo_Trans_comment
                if sInfo not in (None, ""):
                    cell.comment = openpyxl.comments.Comment(text=sInfo, author=None)
                lstHeader.append(cell)
            ws.append(lstHeader)

            # Content
            dicSheet = dicData[sTSheet]
            if dicSheet is not None:
                lstContent = dicSheet.get("content")
                if lstContent is not None:
                    for row in lstContent:
                        ws.append(row)

        res.addData("wb", wb)
        return res

    # Template BSH
    def convert_export_write_excel_template(self, sFile):
        res = Result("MARLIN (Transport)")
        sLog = ""

        # Hole Template
        resExp = self.convert_export_create_excel_wb()
        if resExp.isErr():
            res.addResult(resExp)
        wb = resExp.getData("wb")

        # Save
        try:
            wb.save(filename=sFile)
        except Exception as ex:
            res.addErr("Error during saving Excel file: " + str(ex))

        # Info
        sInfo = PrettyText.html_special_info_fileSaved(sFile, name="Template")
        res.addData("title", "Converter Template")
        res.addData("text", sInfo)
        res.addData("log", sLog)

        return res

