# coding=utf-8


# -------------------------------------------------
# DataPatrol
# -------------------------------------------------

import openpyxl, traceback, pickle, zipfile
from collections import Counter, OrderedDict
import urllib, os, sys, csv, getpass, platform, datetime, time, copy, random, psycopg2, psycopg2.extras
from difflib import SequenceMatcher
import openpyxl.styles.numbers
import warnings
import requests
from psycopg2 import sql
from tools.parallel import Stoppable
from tools.coding import Result, DataTools, ProgressString, WorkingObject
from tools.text import PrettyText
from tools.taxacruncher import TaxaCruncher
from tools.printing import SmartPrinter
from tools.excel import ExcelTools
from tools.fileio import FileTools
from shutil import copyfile
from tools.taxon import Taxon
from tools.ui import Colors

"""
select distinct tax.aid, given_taxon_name, dat.name, ing.file from critterbase.biota bio 
join critterbase.taxon tax on bio.taxon_id=tax.id 
join critterbase.subset sub on bio.subset_id=sub.id
join critterbase.dataset dat on sub.dataset_id=dat.id
join critterbase.ingest ing on bio.ingest_id=ing.id
"""

class DataPatrol(Stoppable, WorkingObject):

    def __init__(self, model):
        Stoppable.__init__(self)
        WorkingObject.__init__(self)

        # -- Member
        self.__model = model
        self.__smartprinter = model.ui_getSmartprinter()
        self.__map = model.subsystem_getMapper()
        self.__myworms = model.subsystem_getMyWorms()

        # Db
        self.__conn = None
        self.__cursor = None

        # Subprozesse die bei einem Abort mit abgebrochen werden sollen
        self.addAbortChild(self.__myworms)

    # ----------------------------------------

    # Info
    def info(self):
        sOut = "*** Status of DataPatrol\n\n"
        return sOut

    def __reset(self):
        self.abort(False)
        self.__conn = None
        self.__cursor = None

    def start_biotaControl(self, **kwargs):
        """
        Biota Check

        Kwargs:
            cursor: Db Cursor
        Returns:
             Result
        """

        # Helper: Aufräumen
        def cleanUi():
            workflow()
            progress("off")
            progress("off")

        def closeDb():
            if self.__model.conf_isDry() and conn is not None:
                conn.rollback()
            elif res.isErr():
                conn.rollback()
            else:
                conn.commit()
            cursor.close()
            conn.close()

        # ---

        pn = self.__smartprinter.normal
        pv = self.__smartprinter.verbose
        pd = self.__smartprinter.debug

        progress = self.__smartprinter.progress
        status = self.__smartprinter.status
        workflow = self.__smartprinter.workflow
        userinput = self.__smartprinter.inputAsk__comboMulti
        getInputAnswer = self.__smartprinter.getInputAnswer
        myworms = self.__model.subsystem_getMyWorms()

        # Abbruch Reset
        self.abort(False)
        res = Result(name="DataPatrol: BiotaControl")
        pn("* BiotaControl")

        # Parameter
        cursor = kwargs["cursor"] if "cursor" in kwargs else None
        if cursor is None:
            bOwnConnection = True
        else:
            bOwnConnection = False

        # ---------

        if bOwnConnection:
            pd("Using own connection")
        else:
            pd("Reusing connection")

        # --- Check Parameter
        progress("on-pulse")
        workflow(top="BiotaCotrol", sub="Getting Biota Data", fmt=False)

        # Daten
        iInfo = 0
        iWarn = 0
        iErr = 0
        dicBiotaNotFound = dict()
        dicAllTaxaWithNameGaps = {}
        dicAllTaxaWithTypos = {}
        dicAllTaxaWithStatusProblems = {}
        dicAllTaxaWithAphiaProblems = {}
        dicAllUnacceptedTaxa = {}
        dicAllUnacceptedTaxa_withAaid_LeadingToNotAcceptedTaxa = {}
        dicAllLimboTaxa = {}

        # Eigene DB Connection
        if bOwnConnection:
            resConn = self.__model.subsystem_getDb().conn_getConnection()
            if resConn.isErr():
                res.addResult(resConn)
                cleanUi()
                return res
            # Connection: An
            conn = resConn.getData("connection")
            cursor = resConn.getData("cursor")

        # Zeit
        iZeitStart = time.time()
        sZeitStart = str(datetime.datetime.now().strftime("%Y-%m-%d %H:%M"))

        resBiotaControl = self.__model.subsystem_getDb().dataPatrol_biotaControl_getAllBiota(cursor=cursor)
        tc = TaxaCruncher(self.__model)

        if resBiotaControl.isOk():
            lstBiota = resBiotaControl.getData("biota")
            iRowTotal = len(lstBiota)
            progress("on-abort")
            workflow(top="BiotaControl", sub="Asking MyWoRMS", fmt=False)
            for iRow, (iAid, sName, sDataset, sFile) in enumerate(lstBiota):
                progress(iRow, iRowTotal, 100, proz=True, abs=True, cli=True, err=iErr, warn=iWarn, info=iInfo)
                if iAid > 0:
                    # Normales Taxon
                    tc.addTaxon(iAid, None, None) # Taxon in TaxonCruncher geben
                    resMyWorms = myworms.getTaxon(iAid)
                    if resMyWorms.isOk() and resMyWorms.getData("found") is True:
                        t = resMyWorms.getData("taxon")
                        # Unakzeptiertes Taxon mit aApid führt zu einem unakzeptierten Taxon
                        if t.isUnaccepted() and t.getAphiaId_accepted() is not None and not t.isRedirected_2_accepted():
                            if sDataset not in dicAllTaxaWithNameGaps:
                                dicAllUnacceptedTaxa_withAaid_LeadingToNotAcceptedTaxa[sDataset] = []
                                dicAllUnacceptedTaxa_withAaid_LeadingToNotAcceptedTaxa[sDataset].append(t)
                                iWarn += 1
                        # Name Problems
                        if t.hasProblems_names():
                            if sDataset not in dicAllTaxaWithNameGaps:
                                dicAllTaxaWithNameGaps[sDataset] = []
                            dicAllTaxaWithNameGaps[sDataset].append(t)
                            iWarn += 1
                        # Status Problems
                        if t.hasProblems_status():
                            if sDataset not in dicAllTaxaWithStatusProblems:
                                dicAllTaxaWithStatusProblems[sDataset] = []
                            dicAllTaxaWithStatusProblems[sDataset].append(t)
                            iWarn += 1
                        # Unaccepted
                        if t.isUnaccepted():
                            if sDataset not in dicAllUnacceptedTaxa:
                                dicAllUnacceptedTaxa[sDataset] = []
                            dicAllUnacceptedTaxa[sDataset].append(t)
                            iInfo += 1
                        # Limbo
                        if t.isLimbo():
                            if sDataset not in dicAllLimboTaxa:
                                dicAllLimboTaxa[sDataset] = []
                            dicAllLimboTaxa[sDataset] .append(t)
                            iInfo += 1
                        # Aphia
                        if t.hasProblems_aphia():
                            if sDataset not in dicAllTaxaWithAphiaProblems:
                                dicAllTaxaWithAphiaProblems[sDataset] = []
                            dicAllTaxaWithAphiaProblems[sDataset].append(t)
                            iWarn += 1
                        if t.getScientificName() is not None and \
                                SequenceMatcher(None, sName, t.getScientificName()).ratio() < self.__model.subsystem_getTaxonWizard().getScientificNameMatchRatio():
                            if sDataset not in dicAllTaxaWithTypos:
                                dicAllTaxaWithTypos[sDataset] = []
                            dicAllTaxaWithTypos[sDataset].append([str(sName),str(t.getScientificName()),t])
                            iWarn += 1
                    else:
                        # AphiaId not found <darf eigentlich nie passieren - dann hätte die ingest QM versagt>
                        if sDataset not in dicBiotaNotFound:
                            dicBiotaNotFound[sDataset] = set()
                        dicBiotaNotFound[sDataset].add(iAid)
                    if self.isAbort():
                        res.addErr("User abort",kat="BiotaControl", sys="Data Patrol")
                else:
                    # Special Taxon
                    pass
        else:
            res.addResult(resBiotaControl, msg_only=True)

        # TaxonCruncher erzeugt Statistic
        res = tc.crunch(html=False)
        taxa = res.getData("taxa")
        sLogTcStat_html = taxa.getStatistic(html=True)
        sLogTcStat_ascii = taxa.getStatistic(html=False)

        # Log
        setDatasets_warnInfo = set()
        setDatasets_warnInfo = setDatasets_warnInfo.union(dicAllLimboTaxa.keys())
        setDatasets_warnInfo = setDatasets_warnInfo.union((dicAllTaxaWithNameGaps.keys()))
        setDatasets_warnInfo = setDatasets_warnInfo.union((dicAllTaxaWithStatusProblems.keys()))
        setDatasets_warnInfo = setDatasets_warnInfo.union((dicAllUnacceptedTaxa.keys()))
        setDatasets_warnInfo = setDatasets_warnInfo.union((dicAllTaxaWithAphiaProblems.keys()))
        setDatasets_warnInfo = setDatasets_warnInfo.union((dicAllTaxaWithTypos.keys()))
        setDatasets_warnInfo = setDatasets_warnInfo.union((dicAllUnacceptedTaxa_withAaid_LeadingToNotAcceptedTaxa.keys()))

        # Zusammenbauen: Info/Warn/Err
        # iWarn = len(dicAllTaxaWithTypos) + len(dicAllTaxaWithAphiaProblems) + len(dicAllTaxaWithNameGaps) + len(dicAllTaxaWithStatusProblems) + len(dicAllUnacceptedTaxa_withAaid_LeadingToNotAcceptedTaxa)
        # iInfo = len(dicAllUnacceptedTaxa) + len(dicAllLimboTaxa)
        # iErr = 0

        # Log: zusammenbauen
        # Log: Datasets: Overview
        sLogDatasets_overview = ""
        lstLogDatasets_overview =[("Dataset","Typo*","Name*","Stat*","Aid*","Lead*","Limbo","Unacc","NotFound*")]
        lstLogDatasets_overview_html =[("Dataset","Typo","Name","Status","AphiaId","Misleading","Limbo","Unaccepted","NotFound")]
        for sDataset in setDatasets_warnInfo:
            sBiotaNotFound_nr = str(len(dicBiotaNotFound[sDataset])) if sDataset in dicBiotaNotFound else ""
            sAllTypos_nr = str(len(dicAllTaxaWithTypos[sDataset])) if sDataset in dicAllTaxaWithTypos else ""
            sAllUnacceptedTaxa_nr = str(len(dicAllUnacceptedTaxa[sDataset])) if sDataset in dicAllUnacceptedTaxa else ""
            sAllLimboTaxa_nr = str(len(dicAllLimboTaxa[sDataset])) if sDataset in dicAllLimboTaxa else ""
            sAllTaxaWithStatusProblems_nr = str(len(dicAllTaxaWithStatusProblems[sDataset])) if sDataset in dicAllTaxaWithStatusProblems else ""
            sAllTaxaWithNameGaps_nr = str(len(dicAllTaxaWithNameGaps[sDataset])) if sDataset in dicAllTaxaWithNameGaps else ""
            sAllTaxaWithAphiaProblems_nr = str(len(dicAllTaxaWithAphiaProblems[sDataset])) if sDataset in dicAllTaxaWithAphiaProblems else ""
            sAllUnacceptedTaxa_nr = str(len(dicAllUnacceptedTaxa[sDataset])) if sDataset in dicAllUnacceptedTaxa else ""
            sAllUnacceptedTaxa_withAaid_LeadingToNotAcceptedTaxa_nr = str(len(dicAllUnacceptedTaxa_withAaid_LeadingToNotAcceptedTaxa[sDataset])) if sDataset in dicAllUnacceptedTaxa_withAaid_LeadingToNotAcceptedTaxa else ""
            lstLogDatasets = [sDataset,
                             sAllTypos_nr,
                             sAllTaxaWithNameGaps_nr,
                             sAllTaxaWithStatusProblems_nr,
                             sAllTaxaWithAphiaProblems_nr,
                             sAllUnacceptedTaxa_withAaid_LeadingToNotAcceptedTaxa_nr,
                             sAllLimboTaxa_nr,
                             sAllUnacceptedTaxa_nr,
                             sBiotaNotFound_nr]
            lstLogDatasets_overview.append(lstLogDatasets)
            lstLogDatasets_html = lstLogDatasets.copy()
            lstLogDatasets_html[0] = '<a style="color:#000; text-decoration: none;" href="#'+sDataset+'">'+sDataset+'</a>'
            lstLogDatasets_overview_html.append(lstLogDatasets_html)

        # Log Prae
        sColor_awi_blue_light = Colors.getColor("awi_blue_light", html=True)
        sColor_awi_blue_mid = Colors.getColor("awi_blue_mid", html=True)
        sColor_awi_blue_dark = Colors.getColor("awi_blue_dark", html=True)
        iSekTotal = time.time() - iZeitStart
        iStd = int(iSekTotal // 3600)
        iMin = int((iSekTotal % 3600) // 60)
        iSek = int(iSekTotal % 60)
        sLogPrae = ""
        sLogPrae_html = ""
        sLogPrae_html += '<img height="67" src="https://recruitingapp-5442.de.umantis.com/pubhtml/logo.gif" alt="AWI Logo">'
        sLogPrae_html += '<div layout="block" styleclass="DynForm_DfAttributes" style="z-index:0;position:absolute;left:0px;top:68px;right:0px;height:25px;overflow:hidden;text-align:left;background-color:' + sColor_awi_blue_light + ';border:medium none;-moz-border-colors:none;border-radius:0; margin-top: 5px; margin-bottom: 5px; ;"></div>'
        sLogPrae_html += '<br>'
        sLogPrae_html += "\n\n"
        sLogPrae_html += "&nbsp;"
        sLogPrae_html += '<h1 style="color:' + sColor_awi_blue_dark + ';">BiotaControl</h1>\n\n'
        sLogPrae += "************************************\n"
        sLogPrae += "BiotaControl\n"
        sLogPrae += "Date: " + str(datetime.datetime.now().strftime("%Y-%m-%d %H:%M:%S")) + "\n"
        sLogPrae_html += "Date: " + str(datetime.datetime.now().strftime("%Y-%m-%d %H:%M:%S")) + "\n"
        sLogPrae += "Duration: " + str(iStd) + "h" + str(iMin) + "m" + str(iSek) + "s\n"
        sLogPrae_html += "Duration: " + str(iStd) + "h" + str(iMin) + "m" + str(iSek) + "s\n"
        sLogPrae += "************************************\n\n"
        if self.__model.conf_isDry():
            sLogPrae+= "Dry mode activated.\n"
            sLogPrae += "\n"

        # Log: Overview
        sLogOverview_asci = ""
        sLogOverview_html = ""
        if len(setDatasets_warnInfo) > 0:
            sLogOverview_asci +=  ">>>>>>>>>>>>>>>>>>><<<<<<<<<<<<<<<<<\n"
            sLogOverview_html +=  "*** Caution\n"
            if iWarn > 0:
                sLogOverview_asci += "Warnings: "+str(iWarn)+"\n"
                #sLogOverview_html += "Warnings: "+str(iWarn)+"\n"
            if iInfo > 0:
                sLogOverview_asci += "Information: "+str(iInfo)+"\n"
                #sLogOverview_html += "Information: "+str(iInfo)+"\n"
            sLogOverview_asci += ">>>>>>>>>>>>>>>>>>><<<<<<<<<<<<<<<<<\n\n"
            if iWarn > 0:
                sLogOverview_asci += "* Warnings: "+str(iWarn)+"\n"
                sLogOverview_html += "* Warnings: "+str(iWarn)+"\n"
                if len(dicBiotaNotFound) >0:
                    iNr = 0
                    for sDataset, lstMessage in dicBiotaNotFound.items():
                        iNr += len(lstMessage)
                    sLogOverview_asci  += "Taxon not found in MyWoRMS (Very strange, Ingest QC failed?): "+str(iNr)+"\n"
                    sLogOverview_html  += "Taxon not found in MyWoRMS (Very strange, Ingest QC failed?): "+str(iNr)+"\n"
                if len(dicAllTaxaWithTypos) >0:
                    iNr = 0
                    for sDataset, lstMessage in dicAllTaxaWithTypos.items():
                        iNr += len(lstMessage)
                    sLogOverview_asci  += "Typos in given ScientifcNames (Typo): "+str(iNr)+"\n"
                    sLogOverview_html  += "Typos in given ScientifcNames (Typo): "+str(iNr)+"\n"
                if len(dicAllTaxaWithNameGaps) >0:
                    iNr = 0
                    for sDataset, lstMessage in dicAllTaxaWithNameGaps.items():
                        iNr += len(lstMessage)
                    sLogOverview_asci  += "Problems with Names (Name): "+str(iNr)+"\n"
                    sLogOverview_html  += "Problems with Names (Name): "+str(iNr)+"\n"
                if len(dicAllTaxaWithAphiaProblems) >0:
                    iNr = 0
                    for sDataset, lstMessage in dicAllTaxaWithAphiaProblems.items():
                        iNr += len(lstMessage)
                    sLogOverview_asci  += "Problems with AphiaIds (Aid): "+str(iNr)+"\n"
                    sLogOverview_html  += "Problems with AphiaIds (Aid): "+str(iNr)+"\n"
                if len(dicAllTaxaWithStatusProblems) >0:
                    iNr = 0
                    for sDataset, lstMessage in dicAllTaxaWithStatusProblems.items():
                        iNr += len(lstMessage)
                    sLogOverview_asci  += "Problems with status (Stat): "+str(iNr)+"\n"
                    sLogOverview_html += "Problems with status (Stat): "+str(iNr)+"\n"
                if len(dicAllUnacceptedTaxa_withAaid_LeadingToNotAcceptedTaxa) > 0:
                    iNr = 0
                    for sDataset, lstMessage in dicAllUnacceptedTaxa_withAaid_LeadingToNotAcceptedTaxa.items():
                        iNr += len(lstMessage)
                    sLogOverview_asci += "Unaccepted Taxa with Accepted AphiaId leading to not accepted Taxa (Lead): " + str(iNr) + "\n"
                    sLogOverview_html += "Unaccepted Taxa with Accepted AphiaId leading to not accepted Taxa (Lead): " + str(iNr) + "\n"
                sLogOverview_asci  += "\n"
            if iInfo > 0:
                sLogOverview_asci += "* Information: " + str(iInfo) + "\n"
                sLogOverview_html += "* Information: " + str(iInfo) + "\n"
                if len(dicAllUnacceptedTaxa) > 0:
                    iNr = 0
                    for sDataset, lstMessage in dicAllUnacceptedTaxa.items():
                        iNr += len(lstMessage)
                    sLogOverview_asci  += "Unaccepted taxa (Unacc): " + str(iNr) + "\n"
                    sLogOverview_html  += "Unaccepted taxa (Unacc): " + str(iNr) + "\n"
                if len(dicAllLimboTaxa) > 0:
                    iNr = 0
                    for sDataset, lstMessage in dicAllLimboTaxa.items():
                        iNr += len(lstMessage)
                    sLogOverview_asci  += "Taxa in Limbo (Limbo): " + str(iNr) + "\n"
                    sLogOverview_html  += "Taxa in Limbo (Limbo): " + str(iNr) + "\n"
                sLogOverview_asci  += "\n"
            sLogOverview_asci += ">>>>>>>>>>>>>>>>>>><<<<<<<<<<<<<<<<<\n\n"

        # Log: Datasets
        sLogDatasets_ascii = ""
        sLogDatasets_html = '<span id="dataset_overview">\n'
        if len(setDatasets_warnInfo) > 0:
            sLogDatasets_ascii += "************************************\n"
            sLogDatasets_ascii += "DATASETS ("+ str(len(setDatasets_warnInfo)) + "x)\n"
            sLogDatasets_html += "*** Datasets (Overview)\n"
            sLogDatasets_ascii += "************************************\n\n"
            sLogDatasets_html += "--- Datasets ("+ str(len(setDatasets_warnInfo)) + "x)\n"
            sLogDatasets_ascii += PrettyText.table(lstLogDatasets_overview) +"\n"
            sLogDatasets_html += PrettyText.table(lstLogDatasets_overview_html, header=True, html=True, preview=5) +"\n"
            sLogDatasets_ascii +="\n"

        # Log Datasets
        iLog_preview = 3
        sLogDatasets_detail_ascii = ""
        sLogDatasets_detail_html = ""
        if len(setDatasets_warnInfo) > 0:
            sLogDatasets_detail_ascii += "************************************\n"
            sLogDatasets_detail_ascii += "DETAILS\n"
            sLogDatasets_detail_html += "*** Datasets (Details)\n"
            sLogDatasets_detail_ascii += "************************************\n\n"
            for sDataset in setDatasets_warnInfo:
                sLogDataset_ascii = ""
                sLogDataset_html = ""
                iLogsDataset = 0
                if sDataset in dicBiotaNotFound:
                    sLogDataset_ascii += "--- Taxon not found in MyWoRMS <INGEST QC FAILED!> (" + str(len(dicBiotaNotFound[sDataset])) + "x)\n"
                    sLogDataset_html += "* Taxon not found in MyWoRMS <INGEST QC FAILED!> (" + str(len(dicBiotaNotFound[sDataset])) + "x)\n"
                    iLogsDataset += len(dicBiotaNotFound[sDataset])
                    for iAid in dicBiotaNotFound[sDataset]:
                        sLogDataset_ascii += str(iAid) + "\n"
                        sLogDataset_html += str(iAid) + "\n"
                if sDataset in dicAllTaxaWithTypos:
                    sLogDataset_ascii += "--- Typos (" + str(len(dicAllTaxaWithTypos[sDataset])) + "x)\n"
                    sLogDataset_html += "* Typos (" + str(len(dicAllTaxaWithTypos[sDataset])) + "x)\n"
                    iLogsDataset += len(dicAllTaxaWithTypos[sDataset])
                    lstHtml = [["Taxon", "ScientificName in Biota", "ScientificName in MyWoRMS"]]
                    for sNameBiota, sNameWorms, t in dicAllTaxaWithTypos[sDataset]:
                        sLogDataset_ascii += "Biota("+sNameBiota + ") != MyWorMS(" + sNameWorms + ") ["+t.signature()+"]\n"
                        lstHtml.append([t.signature(html=True), sNameBiota, sNameWorms])
                    sLogDataset_html += PrettyText.table(lstHtml, header=True, html=True, preview=iLog_preview)
                if sDataset in dicAllUnacceptedTaxa:
                    sLogDataset_ascii += "--- Unaccepted (" + str(len(dicAllUnacceptedTaxa[sDataset])) + "x)\n"
                    sLogDataset_html += "* Unaccepted (" + str(len(dicAllUnacceptedTaxa[sDataset])) + "x)\n"
                    iLogsDataset += len(dicAllUnacceptedTaxa[sDataset])
                    lstHtml = []
                    for t in dicAllUnacceptedTaxa[sDataset]:
                        sLogDataset_ascii += t.signature() + "\n"
                        lstHtml.append([t.signature(html=True)])
                    sLogDataset_html += PrettyText.table(lstHtml, header=False, html=True, preview=iLog_preview)
                if sDataset in dicAllLimboTaxa:
                    sLogDataset_ascii += "--- Limbo (" + str(len(dicAllLimboTaxa[sDataset])) + "x)\n"
                    sLogDataset_html += "* Limbo (" + str(len(dicAllLimboTaxa[sDataset])) + "x)\n"
                    iLogsDataset += len(dicAllLimboTaxa[sDataset])
                    lstHtml = []
                    for t in dicAllLimboTaxa[sDataset]:
                        sLogDataset_ascii += t.signature() + "\n"
                        lstHtml.append([t.signature(html=True)])
                    sLogDataset_html += PrettyText.table(lstHtml, header=False, html=True, preview=iLog_preview)
                if sDataset in dicAllUnacceptedTaxa_withAaid_LeadingToNotAcceptedTaxa:
                    sLogDataset_ascii += "--- Unaccepted Taxa leading with Accepted AphiaId to unaccepted Taxa (" + str(len(dicAllUnacceptedTaxa_withAaid_LeadingToNotAcceptedTaxa[sDataset])) + "x)\n"
                    sLogDataset_html += "* Unaccepted Taxa leading with Accepted AphiaId to unaccepted Taxa (" + str(len(dicAllUnacceptedTaxa_withAaid_LeadingToNotAcceptedTaxa[sDataset])) + "x)\n"
                    iLogsDataset += len(dicAllUnacceptedTaxa_withAaid_LeadingToNotAcceptedTaxa[sDataset])
                    lstHtml = []
                    for t in dicAllUnacceptedTaxa_withAaid_LeadingToNotAcceptedTaxa[sDataset]:
                        sLogDataset_ascii += t.signature() + "\n"
                        lstHtml.append([t.signature(html=True)])
                    sLogDataset_html += PrettyText.table(lstHtml, header=False, html=True, preview=iLog_preview)
                if sDataset in dicAllTaxaWithStatusProblems:
                    sLogDataset_ascii += "--- Status (" + str(len(dicAllTaxaWithStatusProblems[sDataset])) + "x)\n"
                    sLogDataset_html += "* Status (" + str(len(dicAllTaxaWithStatusProblems[sDataset])) + "x)\n"
                    iLogsDataset += len(dicAllUnacceptedTaxa[sDataset])
                    lstHtml = []
                    for t in dicAllTaxaWithStatusProblems[sDataset]:
                        sLogDataset_ascii += t.signature() + "\n"
                        lstHtml.append([t.signature(html=True)])
                    sLogDataset_html += PrettyText.table(lstHtml, header=False, html=True, preview=iLog_preview)
                if sDataset in dicAllTaxaWithNameGaps:
                    sLogDataset_ascii += "--- Name (" + str(len(dicAllTaxaWithNameGaps[sDataset])) + "x)\n"
                    sLogDataset_html += "* Name (" + str(len(dicAllTaxaWithNameGaps[sDataset])) + "x)\n"
                    iLogsDataset += len(dicAllTaxaWithNameGaps[sDataset])
                    lstHtml = []
                    for t in dicAllTaxaWithNameGaps[sDataset]:
                        sLogDataset_ascii += t.signature() + "\n"
                        lstHtml.append([t.signature(html=True)])
                    sLogDataset_html += PrettyText.table(lstHtml, header=False, html=True, preview=iLog_preview)
                if sDataset in dicAllTaxaWithAphiaProblems:
                    lstHtml = []
                    sLogDataset_ascii += "--- AphiaId (" + str(len(dicAllTaxaWithAphiaProblems[sDataset])) + "x)\n"
                    sLogDataset_html += "* AphiaId (" + str(len(dicAllTaxaWithAphiaProblems[sDataset])) + "x)\n"
                    iLogsDataset += len(dicAllTaxaWithAphiaProblems[sDataset])
                    for t in dicAllTaxaWithAphiaProblems[sDataset]:
                        sLogDataset_ascii += t.signature() + "\n"
                        lstHtml.append([t.signature(html=True)])
                    sLogDataset_html += PrettyText.table(lstHtml, header=False, html=True, preview=iLog_preview)
                if sLogDataset_ascii != "":
                    sLogDatasets_detail_ascii += "*** " + sDataset + " (" + str(iLogsDataset) + "x)\n"
                    sLogDatasets_detail_html += '--- ' + '<span id="' + str(sDataset) + '"> ' + sDataset + ' </span> (' + str(iLogsDataset) + 'x)\n'
                    sLogDatasets_detail_html += '<a style="font-size: smaller; color:#000; text-decoration: none;" href="#dataset_overview">&#9651; click here to get back to dataset overview ...</a>\n'
                    sLogDatasets_detail_ascii += sLogDataset_ascii
                    sLogDatasets_detail_html += sLogDataset_html
                    sLogDatasets_detail_ascii += "\n"
                    sLogDatasets_detail_html += "\n"
        # Antwort
        sLog_ascii = sLogPrae + sLogOverview_asci + sLogDatasets_ascii + sLogDatasets_detail_ascii + sLogTcStat_ascii
        sLog_html =  sLogPrae_html + sLogOverview_html + sLogTcStat_html + sLogDatasets_html + sLogDatasets_detail_html
        sLog_html = PrettyText.html_txt2html(sLog_html)

        # Admin Stats schreiben
        if not self.__model.conf_isDry() and not self.isAbort():
            try:
                sSchema = self.__model.subsystem_getDb().account_active_getSchema()
                # Speichere Datum des Updates
                sqlCmd = sql.SQL(
                    "DELETE FROM {schema}.admin WHERE key='last_biota_control_date'").format(
                    schema=sql.Identifier(sSchema)
                )
                cursor.execute(sqlCmd)
                sqlCmd = sql.SQL(
                    "INSERT INTO {schema}.admin (key, value) VALUES ('last_biota_control_date', {val})").format(
                    schema=sql.Identifier(sSchema),
                    val=sql.Literal(str(datetime.datetime.now()))
                )
                cursor.execute(sqlCmd)
                # Speichere Anzahl Fehler
                sqlCmd = sql.SQL(
                    "DELETE FROM {schema}.admin WHERE key='last_biota_control_err'").format(
                    schema=sql.Identifier(sSchema)
                )
                cursor.execute(sqlCmd)
                sqlCmd = sql.SQL(
                    "INSERT INTO {schema}.admin (key, value) VALUES ('last_biota_control_err', {val})").format(
                    schema=sql.Identifier(sSchema),
                    val=sql.Literal(str(iErr))
                )
                cursor.execute(sqlCmd)
                # Speichere Anzahl Warnungen
                sqlCmd = sql.SQL(
                    "DELETE FROM {schema}.admin WHERE key='last_biota_control_warn'").format(
                    schema=sql.Identifier(sSchema)
                )
                cursor.execute(sqlCmd)
                sqlCmd = sql.SQL(
                    "INSERT INTO {schema}.admin (key, value) VALUES ('last_biota_control_warn', {val})").format(
                    schema=sql.Identifier(sSchema),
                    val=sql.Literal(str(iWarn))
                )
                cursor.execute(sqlCmd)
                # Speicher Log
                sqlCmd = sql.SQL(
                    "DELETE FROM {schema}.admin WHERE key='last_biota_control_log'").format(
                    schema=sql.Identifier(sSchema)
                )
                cursor.execute(sqlCmd)
                sqlCmd = sql.SQL(
                    "INSERT INTO {schema}.admin (key, value) VALUES ('last_biota_control_log', {log})").format(
                    schema=sql.Identifier(sSchema),
                    log=sql.Literal(str(sLog_html))
                )
                cursor.execute(sqlCmd)
            except Exception as ex:
                res.addErr("Problems writing to admin tables", add=str(ex), kat="BiotaControl", sys="Data Patrol")

       # *** Action Log
        if not self.__model.conf_isDry() and (iErr == 0) and not self.isAbort() and not res.isErr():
            resActionLog = self.__model.subsystem_getDb().actionlog_new("data-patrol-biota-control", "biota", sLog_ascii, cursor=cursor)
            res.addResult(resActionLog)

        # Antwort
        res.addData("title", "BiotaControl")
        res.addData("text", sLog_html)

        # Save MyWoRMS
        self.__myworms.puffer_save()

        # Aufräumen
        cleanUi()
        if bOwnConnection:
            closeDb()
        return res

