# coding=utf-8

import openpyxl
from collections import Counter, OrderedDict
import os, sys, csv, getpass, platform, datetime, time, copy, re
import openpyxl.styles.numbers
import warnings
from tools.coding import Result, WorkingObject
from tools.text import PrettyText
from tools.parallel import Stoppable
from tools.printing import SmartPrinter
from tools.misc import JsonConverter
from tools.geocruncher import GeoCruncher
import traceback, json


# -------------------------------------
# DbSchema: Alle Informationen zu CB DB Schema
# -------------------------------------
"""
SNIPPET
select c.table_schema,c.table_name,c.column_name,c.ordinal_position,c.column_default,c.data_type,d.description
from information_schema.columns c
inner join pg_class c1
on c.table_name=c1.relname
inner join pg_catalog.pg_namespace n
on c.table_schema=n.nspname
and c1.relnamespace=n.oid
left join pg_catalog.pg_description d
on d.objsubid=c.ordinal_position
and d.objoid=c1.oid
where c.table_name='station'
and c.table_schema='critterbase'
"""
class DbSchema(Stoppable, WorkingObject):

    # Konstruktor
    def __init__(self, model):
        Stoppable.__init__(self)
        WorkingObject.__init__(self)

        self.__model = model
        self.__smartprinter = model.ui_getSmartprinter()
        self.__map = model.subsystem_getMapper()

        self.__bConnectionWorking = False
        self.__sConnection = None
        self.__odicComments = None

        self.setWorking(True)
        self.setReady(False)

    # ------------------------------------------------------------------------------------------------------------

    # ToString
    def __str__(self):
        return self.info()

    # ------------------------------------------------------------------------------------------------------------

    # Info Init
    def info(self):
        sOut = "*** Status of Connected Db Schema\n\n"
        if not self.isReady():
            sOut += "System is not ready"
        else:
            sOut += "Server: " + str(self.__sConnection) + "\n"
            for sSystem, odicSystem in self.__odicComments.items():
                lstTable=[["Sheet","Column","Name","Obl","DataType","Unit","Pangaea","Comment"]]
                for (sSheet, sCol), sCommentJson in odicSystem.items():
                    resComm = self.getDbSchemaInfo(sSystem, sSheet, sCol)
                    if resComm.isErr():
                        sName = "*Err*"
                        sUnit = "*Err*"
                        sComment = "*Err*"
                        sPangaea = "*Err*"
                        bObl = "*Err*"
                        sDataType = "*Err*"
                    else:
                        sName = resComm.getData("comment-part-name")
                        sUnit = resComm.getData("comment-part-unit")
                        sDataType = resComm.getData("datatype")
                        bObl = resComm.getData("obl")
                        sComment = resComm.getData("comment-part-comment")
                        sPangaea = resComm.getData("comment-part-pangaea")
                    sName = "" if sName is None else sName
                    sUnit = "" if sUnit is None else sUnit
                    sComment = "" if sComment is None else sComment
                    sPangaea = "" if sPangaea is None else sPangaea
                    sDataType = "" if sDataType is None else sDataType
                    bObl = False if bObl is None else bObl
                    sObl = "yes" if bObl else ""
                    lstRow = [sSheet, sCol, sName, sObl, sDataType, sUnit, sPangaea, sComment]
                    lstTable.append(lstRow)
                sOut += "--- "+sSystem+"\n"
                sOut += PrettyText.table(lstTable, True)+"\n"

        return sOut

    def reset(self):
        self.__odicComments = dict()
        self.__odicComments["lookup"] = OrderedDict()
        self.__odicComments["ingest"] = OrderedDict()
        self.__odicComments["taxon"] = OrderedDict()
        self.__bConnectionWorking = False
        self.__sConnection = None

    # Erzeuge Mapping
    def update(self):
        status = self.__smartprinter.status

        res = Result("Update DbSchema Infos")
        self.reset()
        self.setReady(False)

        # Connection
        resConn = self.__model.subsystem_getDb().conn_getConnection()
        if resConn.isErr():
            res.addResult(resConn)
            return res
        conn = resConn.getData("connection")
        cursor = resConn.getData("cursor")
        sConnection = self.__model.subsystem_getDb().conn_getConnectionInfo()
        self.__bConnectionWorking = True
        self.__sConnection = sConnection

        # Hole alle Kommentare
        resComments = self.__model.subsystem_getDb().admin_getAllComments(cursor=cursor)
        if resComments.isErr():
            res.addResult(resComments)
            return res
        dicComments = resComments.getData("comments")

        # Lookup/Ingest/Taxon
        lstBatch = [
            ("lookup", self.__map.lookup_getExcelSheets,  self.__map.lookup_getExcelCols, self.__map.lookup_excel2db),
            ("ingest", self.__map.ingest_getExcelSheets, self.__map.ingest_getExcelCols, self.__map.ingest_excel2db),
            ("taxon", self.__map.taxon_getExcelSheets, self.__map.taxon_getExcelCols, self.__map.taxon_excel2db)]
        # Total Comments
        iComment = 0
        iTotalComments = 0
        for sTable in self.__map.admin_getDbTables():
            iTotalComments += len(self.__map.admin_getDbAttributes(sTable))
        for (sSystem, getSheets, getCols, excel2db) in lstBatch:
            for sTable in getSheets():
                iTotalComments += len(getCols(sTable))
        # Loop
        for (sSystem, getSheets, getCols, excel2db) in lstBatch:
            for sSheet in getSheets():
                for sCol in getCols(sSheet):
                    # TWEAK!: Ingest: Alle FK -> Umbiegen auf entsprechendes PK-Sheet (Col->Sheet): (Sample.Cruise -> Cruise.Cruise
                    if sSystem == "ingest" and self.__map.ingest_isFk(sSheet,sCol):
                        sTable, sAttrib = excel2db(sCol, sCol)
                    else:
                        sTable, sAttrib = excel2db(sSheet,sCol)
                    sCommentJson = None
                    sDataType = None
                    # Hole Daten aus geladener Datenmodell-Datenstruktur
                    try:
                        (bOblDb, sDataType, sCommentJson) = dicComments[sTable][sAttrib]
                    except:
                        res.addErr("Data missing", add="System: " + sSystem, sheet=str(sSheet), col=str(sCol), kat="Definition Not Found", sys="Database Schema")

                    # TWEAK!: Ingest: Alle LUP -> sind vom Datentyp in der Excelwelt immer Text und nicht wie im Datenmodell int
                    if sSystem == "ingest" and self.__map.ingest_isLookup(sSheet, sCol):
                        sDataType = "text"

                    # TWEAK!: Ingest: Biota.AphiaId ist zwar ein Lookup aber mit einem Key der vom Datentyp int ist (AphiaId)
                    if (sSystem, sSheet, sCol) == ("ingest","Biota","AphiaId"):
                        sDataType = "int"

                    # Teste auf korrektes JSON Format
                    try:
                        dicJson = json.loads(sCommentJson)
                        sName = str(dicJson["name"]).strip() if "name" in dicJson and dicJson["name"] not in (None, "") else None
                        sUnit = str(dicJson["unit"]).strip() if "unit" in dicJson and dicJson["unit"] not in (None, "") else None
                        sComment = str(dicJson["comment"]).strip() if "comment" in dicJson and dicJson["comment"] not in (None, "") else None
                        sPangaea = str(dicJson["pangaea"]).strip() if "pangaea" in dicJson and dicJson["pangaea"] not in (None, "") else None
                    except Exception as ex:
                        res.addWarn("Corrupt Format", add=str(sCommentJson), sheet=str(sSheet), col=str(sCol), kat="Parsing JSON format", sys="Database Schema")
                    if None in (sSheet, sCol, sTable, sAttrib, sCommentJson):
                        res.addErr("Data missing", add="System: "+sSystem, sheet=str(sSheet), col=str(sCol), kat="Comment Not Found", sys="Database Schema")
                    else:
                        # Alle Daten Speichern
                        bOblMap = self.__map.all_isSheetColObl(sSheet, sCol)

                        # TWEAK!: FK:IndividualMeasure.Biota (Obl) -> PK:Biota.Name (Nicht Obl) -> PK auf obl umbiegen (Eieieieiei!)
                        if (sSheet, sCol) == ("IndividualMeasure", "Biota"):
                            bOblDb = True
                        if bOblMap != bOblDb:
                            res.addWarn("Obligate Definition varies", add="Db:"+("yes" if bOblDb else "no")+" != Map:"+("yes" if bOblMap else "no"), sheet=str(sSheet),
                                       col=str(sCol), kat="Inconsistency Db to Mapper", sys="Database Schema")
                            #print("#",sSheet, sCol, "db:", bOblDb, "map:",bOblMap)
                        sCommentJson = sCommentJson.strip()
                        sCommentJson = sCommentJson.replace("\n", "")
                        self.__odicComments[sSystem][(sSheet, sCol)] = (bOblDb, sDataType, sCommentJson)
                        iComment += 1
                        status("Reading comment ("+str(iComment)+"/"+str(iTotalComments)+"): "+str(sTable)+"."+str(sCol))
        if res.isOk():
            self.setReady(True)
        return res

    # Ist ein Scheet sCol obligat
    def isObl(self, sSheet, sCol):
        """
        Ist ein Sheet.Col obligat
        Args
            sSheet (str): Sheet
            sCol (str): Col
        Return
            Bool
        """
        sSystem = self.__map.all_getSheetSystem(sSheet, sCol)
        resInfo = self.getDbSchemaInfo(sSystem, sSheet, sCol)
        bObl = resInfo.getData("obl")
        if bObl is None:
            return False
        else:
            return bObl

    # Hole Kommentar
    def getDbSchemaInfo(self, sSystem, sSheet, sCol):
        """
        Hole Db Schema Info

        Args
            sSystem (str): taxon, lookup, ingest
            sSheet (str): Sheet
            sCol (str): Col
        Return
            Result
        """
        res = Result("Get commment")
        if not self.isReady():
            res.addErr("DbSchema system is not ready", kat="Internal", sys="Database Schema")
        try:

            # JSON String
            (bObl, sDataType, sCommentJson) = self.__odicComments[sSystem][sSheet, sCol]

            # Parse
            dicJson = json.loads(sCommentJson)

            # Comment: Parts
            sName = str(dicJson["name"]).strip() if "name" in dicJson and dicJson["name"] not in (None, "") else None
            sUnit = str(dicJson["unit"]).strip() if "unit" in dicJson and dicJson["unit"] not in (None, "") else None
            sComment = str(dicJson["comment"]).strip() if "comment" in dicJson and dicJson["comment"] not in (None, "") else None
            sPangaea = str(dicJson["pangaea"]).strip() if "pangaea" in dicJson and dicJson["pangaea"] not in (None, "") else None

            # Comment: Pangaea
            sCommentPangaea = ""
            if sPangaea is not None:
                sCommentPangaea += sPangaea + " "
                if sComment:
                    sCommentPangaea += "@" + sComment
            else:
                if sName:
                    sCommentPangaea += sName + " "
                if sUnit:
                    sCommentPangaea += "[" + sUnit + "] "
                if sComment:
                    sCommentPangaea += "@" + sComment

            # Comment: Full
            sCommentFull = ""
            if sName:
                sCommentFull += sName + " "
            if sUnit:
                sCommentFull += "[" + sUnit + "] "
            if sComment:
                sCommentFull += "(" + sComment + ")"

            res.addData("comment-json", sCommentJson)
            res.addData("comment-full", sCommentFull)
            res.addData("comment-full-pangaea", sCommentPangaea)
            res.addData("comment-part-name", sName)
            res.addData("comment-part-unit", sUnit)
            res.addData("comment-part-comment", sComment)
            res.addData("comment-part-pangaea", sPangaea)
            res.addData("datatype", sDataType)
            res.addData("obl", bObl)
        except Exception as ex:
            res.addErr("Problems getting Info", add=str(sSystem)+": "+str(ex), sheet=str(sSheet), col=str(sCol), kat="Internal", sys="Database Schema")
        return res
