# coding=utf-8

import openpyxl
import openpyxl.styles.numbers
import openpyxl.styles.borders

from collections import OrderedDict
from tools.coding import Result, WorkingObject
from tools.parallel import Stoppable
from tools.text import PrettyText
from psycopg2 import sql


# -------------------------------------
# Exporter: Datenbank -> Export Format
# -------------------------------------

class Exporter(Stoppable, WorkingObject):

    # Konstruktor
    def __init__(self, model):
        Stoppable.__init__(self)
        WorkingObject.__init__(self)
        self.__model = model
        self.__map = model.subsystem_getMapper()
        self.__comments = model.conf_getDbSchema()
        self.__smartprinter = model.ui_getSmartprinter()

        # R-Friendly Tweaks
        self.__odicRfriendlyTweak = {
            "Biota":
                OrderedDict([
                    ("ScientificName", "ScientificName (WoRMS)"),
                    ("ScientificName (user)", "Original scientific name from taxonomic identification in the lab"),
                    ("Status", "Status of Taxon"),
                    ("Accepted AphiaId", "Accepted AphiaId (WoRMS)"),
                    ("Accepted ScientificName", "Accepted ScientificName (WoRMS)")
                ])
        }

    # Info
    def info(self):
        sOut = "*** Status of Exporter System\n\n"
        lstConverters = self.__map.workflow_converter_export_getAll(long=True)
        sOut += "* Registered converters (" + str(len(lstConverters)) + "x)\n"
        for iIdx, sName in enumerate(lstConverters):
            sOut += sName + "\n"
        return sOut

    def getSubset(self, **kwargs):
        """
        Hole ein Subset

        *** 3 Szenarien

        1) Hole alle Daten beginnend bei Subset (Metadatenstümpfe werden nicht mit geholt)
            - Rückgabe Template maximal
        2) Hole alle Daten mit Metadatenstümpfen
            - Selector besteht nur aus Ingest Ids (Liste)
            - Nur so können die Metadaten eingesammelt werden
            a) Wenn alle Ingest Ids vom selben Workflow sind -> Rückgabe Tempate in dem Workflow
            b) Wenn Workflows differieren -> Rückgabe Tempate maximal

        Kwargs
            sheets (lst): Liste der zu Exportierenden Sheets
            showid (bool): Pro Tabelle: Sollen die Ids mit exportiert werden
            rfriendly (bool): Alle Tabellen: Gib einen ID-Block mit aus; Biota: gib erweiterte Taxonomie mit aus
            showingestid (bool): Pro Tabelle: Sollen die Ingest-Ids mit exportiert werden
            selector (dic): Selektor
                sample_date_start (str)
                sample_date_end (str)
                workflow (lst)
                dataset (lst)
                gear (lst)
                taxcover (lst)
                region (lst)
                ingest (lst)
            converter (str): Converter
        Returns:
            Result:
                data (odic): Daten
                wf (str): Tabellen
                log (str): Log
        """

        # Helper: Clean up
        def cleanUp():
            # DB
            try:
                cursor.close()
                conn.close()
            except:
                pass

        # ---

        res = Result(name="Get subset from db")

        pn = self.__smartprinter.normal
        pv = self.__smartprinter.verbose
        pd = self.__smartprinter.debug
        progress = self.__smartprinter.progress
        status = self.__smartprinter.status

        # Parameter
        lstSheetsOnly = kwargs["sheets"] if "sheets" in kwargs else None
        if isinstance(lstSheetsOnly, list) and len(lstSheetsOnly) == 0:
            lstSheetsOnly = None
        dicSelector = kwargs["selector"] if "selector" in kwargs else None
        bShowId = True if "showid" in kwargs and kwargs["showid"] in ["true", "True", True] else False
        bShowIngestId = True if "showingestid" in kwargs and kwargs["showingestid"] in ["true", "True", True] else False
        bRfriendly = True if "rfriendly" in kwargs and kwargs["rfriendly"] in ["true", "True", True] else False
        sConverter = kwargs.get("converter")
        bConverter = False
        bIngest = False
        sIngestWf = None
        lstIngestIds = None
        if (dicSelector is not None) and (len(dicSelector) == 1) and ("ingest" in dicSelector) and len(
                dicSelector["ingest"]) > 0:
            bIngest = True
            lstIngestIds = kwargs["selector"]["ingest"]

        # Parameter Tests
        if dicSelector is None or len(dicSelector) == 0:
            res.addErr("No selection defined")
            return res
        if bIngest:
            for iId in lstIngestIds:
                try:
                    int(iId)
                except:
                    res.addErr("ingest id is not valid")
                    return res
        if lstSheetsOnly is not None:
            lstSheetUnbekannt = list(set(lstSheetsOnly).difference(set(self.__map.ingest_getExcelSheets())))
            if len(lstSheetUnbekannt) > 0:
                cleanUp()
                res.addErr("Unknown sheet: " + PrettyText.lst2strC(lstSheetUnbekannt))
                return res
        if sConverter is not None:
            if self.__map.workflow_converter_export_isExisting(sConverter):
                bConverter = True
            else:
                cleanUp()
                res.addErr("Unknown converter: " + str(sConverter))
                return res

        # ---

        # DB
        resConn = self.__model.subsystem_getDb().conn_getConnection()
        if resConn.isErr():
            cleanUp()
            return resConn
        conn = resConn.getData("connection")
        cursor = resConn.getData("cursor")

        # Check: Mode Export Ingest
        if bIngest and not bConverter:
            # Ermittle Workflow
            setWf = set()
            for iIng in lstIngestIds:
                resWf = self.__model.subsystem_getDb().ingest_getWorkflowOfIngest(iIng, cursor=cursor)
                if resWf.isErr():
                    cleanUp()
                    return resWf
                sWf = resWf.getData("workflow")
                setWf.add(sWf)
            if len(setWf) > 1:
                sIngestWf = None
            else:
                sIngestWf = setWf.pop()

        # Info
        sLog = ""
        sLog = "************************************\n"
        sLog += "Exported Subset System\n"
        sLog += "************************************\n"
        sLog += "\n"
        sLog_converter = ""
        if bConverter:
            # Converter
            sLog += "Mode: Converter: " + sConverter + "\n"
        else:
            # Normal
            if bIngest:
                # Ingest-Subset
                sLog += "Mode: ingest (metadata-stumps will be collected too)\n"
                sLog += "Ingest ids: " + PrettyText.lst2strC(lstIngestIds) + "\n"
                if len(setWf) > 1:
                    sLog += "Multiple workflows found: " + PrettyText.lst2strC(setWf) + "\n"
                    sLog += "Export template: standard workflow\n"
                else:
                    sLog += "Single workflow found: " + sIngestWf + "\n"
                    sLog += "Export template: " + sIngestWf + "\n"
            else:
                # Allgemeines Subset
                sLog += "Mode: normal (no metadata-stumps will be collected)\n"
                sLog += "Export template: standard workflow\n"
            sLog += "Export sheets: " + (
                PrettyText.lst2strC(lstSheetsOnly) if lstSheetsOnly is not None else "all") + "\n"
            sLog += "Export IDs: " + ("yes" if bShowId else "no") + "\n"
            sLog += "Export ingest IDs: " + ("yes" if bShowIngestId else "no") + "\n"
            sLog += "Export R-friendly: " + ("yes" if bRfriendly else "no") + "\n\n"

        # SQL Befehle
        # - Erzeuge SQL: Selector
        resSqlSelector = self.__getSelector_sql(dicSelector)
        if resSqlSelector.isErr():
            cleanUp()
            return resSqlSelector
        sSqlSelector = resSqlSelector.getData("sql")
        #print(sSqlSelector)
        sLog += "--- Selector ---\n" + sSqlSelector + "\n\n"
        # - Erzeuge SQL: Ingest-Daten
        if bConverter:
            # Converter
            c = self.__model.subsystem_getConverter()
            resConverter = c.start_converter_export(sConverter)
            if resConverter.isErr():
                cleanUp()
                return resConverter
            odicSql = resConverter.getData("sql")
            sLog_converter = resConverter.getData("log")
        else:
            # Normal
            resSql = self.__getSubset_sql(showid=bShowId, showingestid=bShowIngestId, ingest=(lstIngestIds, sIngestWf), selector=dicSelector, rfriendly=bRfriendly)
            if resSql.isErr():
                cleanUp()
                return resSql
            odicSql = resSql.getData("sql")

        # - Loop: SQL
        odicAntwort = OrderedDict()
        odicStat = OrderedDict()
        try:
            # Selector: View aufbauen
            cursor.execute(sSqlSelector)
            odicAntwort["Selector"] = {}
            odicAntwort["Selector"]["sql"] = sSqlSelector
            # Loop: Sheets
            iNrMax = len(odicSql)
            iRowsTotal = 0
            for iNr, (sSheet, sSql) in enumerate(odicSql.items()):
                if lstSheetsOnly is None or sSheet in lstSheetsOnly:
                    progress(iNr + 1, iNrMax, 100, proz=False, abs=True, cli=True, status="[" + str(sSheet) + "]")
                    status(str(sSheet) + ", total rows:" + str(iRowsTotal))
                    cursor.execute(sSql)
                    # Antwort
                    lstHeader = [head[0] for head in cursor.description]
                    lstContent = cursor.fetchall()
                    odicAntwort[sSheet] = {}
                    odicAntwort[sSheet]["header"] = lstHeader
                    odicAntwort[sSheet]["content"] = lstContent
                    odicAntwort[sSheet]["rows"] = len(lstContent)
                    odicAntwort[sSheet]["sql"] = sSql
                    iRowsTotal += len(lstContent)
                    odicStat[sSheet] = len(lstContent)
                    sLog += "--- " + str(sSheet) + " ---\n" + str(sSql) + "\n"
        except Exception as ex:
            res.addErr("Creating SQL for Selector", add=str(ex), kat="SQL", sys="Subset Exporter")

        # Log
        if bConverter:
            sLog += sLog_converter

        # Clean
        cleanUp()

        # Ergebnis
        res.addData("log", sLog)
        res.addData("data", odicAntwort)
        res.addData("wf", sIngestWf)
        res.addData("stat", odicStat)
        # quit()
        return res

    def __getSelector_sql(self, dicSelector):
        """
        Hole Subset Ids anhand der Selector Einstellung

        Args:
            selector (dic): Selektor
                sample_date_start (str)
                sample_date_end (str)
                workflow (lst)
                dataset (lst)
                gear (lst)
                taxcover (lst)
                ingest (lst)
        Returns:
            Result:
                subset (lst): Ids der gefundenen Subsets
        """

        # Helper: Erzeuge aus Liste eine String-Liste
        def lst2str(lst):
            if lst is None or len(lst) == 0:
                return ""
            sOut = ""
            for e in lst:
                sOut += "\'" + str(e) + "\',"
            return sOut[:-1]

        res = Result(name="Get SQL for selection view")

        # Parameter
        sSelector_sampleDateStart = None
        sSelector_sampleDateEnd = None
        lstSelector_workflows = None
        lstSelector_dataset = None
        lstSelector_gear = None
        lstSelector_taxcover = None
        lstSelector_ingest = None
        lstSelector_region = None
        if dicSelector is not None:
            sSelector_sampleDateStart = dicSelector[
                "sample_date_start"] if "sample_date_start" in dicSelector and isinstance(
                dicSelector["sample_date_start"], str) else None
            sSelector_sampleDateEnd = dicSelector[
                "sample_date_end"] if "sample_date_end" in dicSelector and isinstance(
                dicSelector["sample_date_end"], str) else None
            lstSelector_workflows = dicSelector["workflow"] if "workflow" in dicSelector and isinstance(
                dicSelector["workflow"], list) else None
            lstSelector_dataset = dicSelector["dataset"] if "dataset" in dicSelector and isinstance(
                dicSelector["dataset"], list) else None
            lstSelector_gear = dicSelector["gear"] if "gear" in dicSelector and isinstance(dicSelector["gear"],
                                                                                           list) else None
            lstSelector_taxcover = dicSelector["taxcover"] if "taxcover" in dicSelector and isinstance(
                dicSelector["taxcover"], list) else None
            lstSelector_region = dicSelector["region"] if "region" in dicSelector and isinstance(
                dicSelector["region"], list) else None
            lstSelector_ingest = dicSelector["ingest"] if "ingest" in dicSelector and isinstance(
                dicSelector["ingest"], list) else None
        # SQL
        # smp.date
        # smp.gear
        # sub.ing_id
        # sub.taxcover
        # sub.dat.name
        # sub.dat.workflow
        sSqlMain = \
            """DROP VIEW IF EXISTS selector CASCADE;
CREATE OR REPLACE TEMP VIEW selector AS 
SELECT DISTINCT sub.id  
FROM <schema>.subset sub 
JOIN <schema>.subsample subsmp ON subsmp.id = sub.subsample_id 
JOIN <schema>.sample smp ON smp.id = subsmp.sample_id 
JOIN <schema>.station sta ON sta.id = smp.station_id 
JOIN <schema>.cruise cru ON cru.id = sta.cruise_id 
JOIN <schema>.ingest ing ON ing.id = sub.ingest_id 
JOIN <schema>.region2sample reg2smp ON reg2smp.sample_id = smp.id 
JOIN <schema>.region reg ON reg.id = reg2smp.region_id
JOIN <schema>.dataset dat ON dat.id = sub.dataset_id 
JOIN <schema>.gear gea ON gea.id = smp.gear_id 
JOIN <schema>.taxcover tax ON tax.id = sub.taxcover_id 
"""
        sSqlWhere = ""
        lstSqlWhere = []
        if None not in (sSelector_sampleDateStart, sSelector_sampleDateEnd):
            lstSqlWhere.append(
                "smp.date_start BETWEEN (\'" + sSelector_sampleDateStart + "\'::date) AND (\'" + sSelector_sampleDateEnd + "\'::date)")
        lstSelector_all = (
            ("ing.workflow", lstSelector_workflows),
            ("dat.name", lstSelector_dataset),
            ("gea.category", lstSelector_gear),
            ("tax.name", lstSelector_taxcover),
            ("reg.name", lstSelector_region),
            ("sub.ingest_id", lstSelector_ingest),
        )
        for (sSelector_name, lstSelector_single) in lstSelector_all:
            if lstSelector_single is not None and len(lstSelector_single) > 0:
                lstSqlWhere.append(sSelector_name + " IN (" + lst2str(lstSelector_single) + ")")
        if len(lstSqlWhere) > 0:
            sSqlWhere = "WHERE\n"
            sSqlWhere += PrettyText.lst2str(lstSqlWhere, " AND\n")

        # Zusammenbauen
        sSql = sSqlMain + sSqlWhere
        sSql = sSql.replace("<schema>", self.__model.subsystem_getDb().account_active_getSchema())

        # Antwort
        res.addData("sql", sSql)
        return res

    """
    Erzeuge die erforderlichen SQL-Befehle für Subset-Download
    Info:
    Datenbank ist Messdatenzentriert 
    Subset dient als Einstiespunkt für Abfrage 
    Spezialmodus: Export von ganzem Ingest
    Sammle zusätzlich noch alle Metadaten ohne Messdaten des Ingests Auf dem Weg zum Subset ein
    Diese wären nämlich sonst verloren, da sie an keinem Subset hängen und Subset als Einstiegspunkt dient  
    """

    def __getSubset_sql(self, **kwargs):
        """
        Kwargs
            showid (bool): Pro Tabelle: Sollen die Ids mit exportiert werden
            showingestid (bool): Pro Tabelle: Sollen die Ingest-Ids mit exportiert werden
            rfriendly (bool): Biota: Füge accepted ScientificName und Aid hinzu
            ingest ((lsrt:int,wf:str)): Special Mode: Hole gesamtes Ingest (mit unvollständigen Metadatenteilen) aus DB
        Returns:
            Result:
                sql (odic): SQL-Befehle
        """

        # Helper: Erzeuge aus Liste eine String-Liste
        def lst2str(lst):
            if lst is None or len(lst) == 0:
                return ""
            sOut = ""
            for e in lst:
                sOut += "\'" + str(e) + "\',"
            return sOut[:-1]

        # Helper: Erzeuge eine Kurzbezeichung von schema.tabelle -> kurz
        def short(sTable):
            # Tweak: Doppelte Kürzel auflösen: status->sts/station->sta
            if sTable == "status":
                return "sts"
            if sTable == "subsample":
                return "subsmp"
            return str(sTable)[0:3]

        # ---

        res = Result(name="Create SQL statements for getting subset from db")

        # Kwargs
        # Id und Ingest Id anzeigen
        bShowId = True if "showid" in kwargs and kwargs["showid"] in ["true", "True", True] else False
        bShowIngestId = True if "showingestid" in kwargs and kwargs["showingestid"] in ["true", "True", True] else False
        bRfriendly = True if "rfriendly" in kwargs and kwargs["rfriendly"] in ["true", "True", True] else False

        # Special Mode: Export Single Ingest
        if "ingest" in kwargs and isinstance(kwargs["ingest"], tuple):
            bIngestMode = bool(kwargs["ingest"][0])
            lstIngestModeIds = kwargs["ingest"][0] if bIngestMode else None
            sIngestModeWf = kwargs["ingest"][1] if bIngestMode else None
        else:
            bIngestMode = False
            lstIngestModeIds = None
            sIngestModeWf = None

        bIngestOnlySubsetTableReached = False
        bIngestOnlyActualTableReached = False

        # Sql Befehle
        odicSql = OrderedDict()

        # Db Schema
        sDbSchema = self.__model.subsystem_getDb().account_active_getSchema()

        # Loop: Sheets
        bIngestOnlySubsetTableReached = False
        lstSheets = self.__map.ingest_getExcelSheets(sIngestModeWf)
        for iCol, sSheet in enumerate(lstSheets):
            lstCols = self.__map.ingest_getExcelCols(sSheet, sIngestModeWf)
            sTable = self.__map.ingest_sheet2table(sSheet)

            # Modus: Ingest-Export: Ab Subset explizit keine Metadaten mehr einsammeln
            if sTable == "subset":
                bIngestOnlySubsetTableReached = True

            # SQL-Befehlsteile
            # Subset
            sSqlSelection = ""
            sSqlFrom = ""
            sSqlJoinFk = ""
            sSqlJoinLookup = ""
            sSqlWhere = "WHERE sub.id IN (SELECT id FROM selector) \n"
            sSqlOrder = ""
            sSqlUnionJoinFk = ""

            # Benutzte Lookup Ids
            # --- Loop: Columns
            for sCol in lstCols:
                # DB-Table/Attribute
                sTableTmp, sAttrib = self.__map.ingest_excel2db(sSheet, sCol)
                # FK
                bFk = self.__map.ingest_isFk(sSheet, sCol)
                # Lookup
                bLup = self.__map.ingest_isLookup(sSheet, sCol)
                sLupTable = self.__map.ingest_getLookup(sSheet, sCol) if bLup else None
                sLupAtr = self.__map.lookup_getKey(sLupTable) if bLup else None

                # --- SELECT: PK / Lookup
                if bRfriendly and sSheet == "Biota" and sCol == "ScientificName":
                    # Mode: R-friendly & Biota: WoRMS AphiaId/ScienticNames
                    sSqlSelection += "tax.scientificname AS \"ScientificName\", \n"
                    sSqlSelection += "bio.given_taxon_name AS \"ScientificName (user)\", \n"
                    sSqlSelection += "tax.status AS \"Status\", \n"
                    sSqlSelection += "tax.aaid AS \"Accepted AphiaId\", \n"
                    sSqlSelection += "tax.aname AS \"Accepted ScientificName\", \n"
                else:
                    # Normal
                    if not bFk:
                        # Kein FK
                        if not bLup:
                            # Kein Lookup
                            # SELECT: Normal
                            sSqlSelection += short(sTable) + "." + str(sAttrib) + " AS \"" + str(sCol) + "\", \n"
                        else:
                            # Lookup
                            # SELECT, LEFT JOIN
                            sSqlSelection += short(sLupTable) + "." + str(sLupAtr) + " AS \"" + str(sCol) + "\", \n"
                            sSqlJoinLookup += "LEFT JOIN <schema>." + str(sLupTable) + " " + short(
                                sLupTable) + " ON " + short(sTable) + "." + str(sAttrib) + "=" + short(
                                sLupTable) + ".id \n"

            # --- JOIN: FK
            if sSheet in ["Cruise", "Station", "Sample", "SubSample"]:
                # Tweak: Hole bei Cruise,Station,SubSample,Sample immer das Metadatengerüst Cruise-Station-Sample-SubSample-Subset hinzu
                lstFk = self.__map.ingest_getFk("Subset")
                sFkLast = "Subset"
                sFkDbLast = "subset"
            else:
                lstFk = self.__map.ingest_getFk(sSheet)
                sFkDbLast = sTable
                sFkLast = sSheet
            if lstFk is not None:
                bIngestOnlyActualTableReached = False
                for sFk in reversed(lstFk):
                    sFkDb = self.__map.ingest_sheet2table(sFk)
                    sFkMainCol, sFkMainTable, sFkMainAttrib = self.__map.ingest_getFkMain(sFkLast)
                    sTmp = "JOIN <schema>." + str(sFkDb) + " " + short(sFkDb) + " ON " + short(
                        sFkDb) + ".id = " + short(sFkDbLast) + "." + sFkMainAttrib + "\n"
                    sSqlJoinFk += sTmp
                    if bIngestMode and bIngestOnlyActualTableReached:
                        sSqlUnionJoinFk += sTmp
                    if sFkDb == sTable:
                        bIngestOnlyActualTableReached = True
                    sFkDbLast = sFkDb
                    sFkLast = sFk

            # --- SELECT: FK
            lstFk = self.__map.ingest_getFk(sSheet)
            if lstFk is not None:
                for sFk in reversed(lstFk):
                    sFkDb = self.__map.ingest_sheet2table(sFk)
                    sSqlSelection = short(sFkDb) + ".name AS \"" + str(sFk) + "\", \n" + sSqlSelection

            # --- ORDER BY
            # Haupttabelle (nur bei Cruise,Station,Sample,Subset)
            if sSheet in ["Cruise", "Station", "Sample", "SubSample", "Subset"]:
                sTable = self.__map.ingest_sheet2table(sSheet)
                sSqlOrder = "length (\"" + sSheet + "\") ASC, \"" + sSheet + "\" ASC, " + sSqlOrder
            # FK (nur bei Cruise,Station,Sample,Subset)
            lstFk = self.__map.ingest_getFk(sSheet)
            if lstFk is not None:
                for sFk in reversed(lstFk):
                    if sFk in ["Cruise", "Station", "Sample", "SubSample", "Subset"]:
                        sFkDb = self.__map.ingest_sheet2table(sFk)
                        # sSqlOrder = "\"" + sFk + "\", " + sSqlOrder
                        sSqlOrder = "length (\"" + sFk + "\") ASC, \"" + sFk + "\" ASC, " + sSqlOrder

            # --- FROM: Haupttabelle
            if sSheet in ["Cruise", "Station", "SubSample", "Sample"]:
                sSqlFrom = "FROM <schema>.subset " + short("subset") + "\n"
            else:
                sSqlFrom = "FROM <schema>." + sTable + " " + short(sTable) + "\n"

            # ---

            # Sql-Zusammenbauen
            # Mode: Show IngestID
            if bShowIngestId:
                sSqlSelection = short(sTable) + ".ingest_id AS ingest_id, \n" + sSqlSelection
            # Mode: R-friendly
            if bRfriendly:
                # ID-Blocks
                sID_block = ""
                # FK
                lstFk_r = self.__map.ingest_getFk(sSheet)
                if lstFk_r is not None:
                    for sSheet_r in lstFk_r:
                        sTable_r = self.__map.ingest_sheet2table(sSheet_r)
                        sID_block += short(sTable_r) + "." + "id AS \"" + str(sSheet_r) + "ID\", \n"
                # PK
                sID_block += short(sTable) + ".id AS \"" + str(sSheet) + "ID\", \n"
                # For das Select bauen
                sSqlSelection = sID_block + sSqlSelection
            # Mode: Show ID
            if bShowId:
                sSqlSelection = short(sTable) + ".id AS id, \n" + sSqlSelection
            sSqlSelection = "SELECT \n" + sSqlSelection[:-3] + "\n"
            # Zusammenbauen
            sSql = sSqlSelection + sSqlFrom + sSqlJoinFk + sSqlJoinLookup + sSqlWhere
            # Mode: Ingest
            if bIngestMode and not bIngestOnlySubsetTableReached:
                sSql += "UNION \n" + sSqlSelection + "FROM <schema>." + sTable + " " + short(
                    sTable) + "\n" + sSqlUnionJoinFk + sSqlJoinLookup + "WHERE " + short(
                    sTable) + ".ingest_id IN (" + lst2str(lstIngestModeIds) + ")\n"
            sSql = sSql.replace("<schema>", sDbSchema)
            # Order
            if len(sSqlOrder) > 0:
                sSqlOrder = "ORDER BY " + sSqlOrder[:-2] + "\n"
            # IndMeasure -> Kein Distinct -> Es gibt nämlich mehrfache gleiche Einträge pro Biota
            if sSheet == "IndividualMeasure":
                sSql = "SELECT * FROM (\n" + sSql + ") t1\n" + sSqlOrder
            else:
                sSql = "SELECT * FROM (\nSELECT DISTINCT * FROM (\n" + sSql + ") t1\n) t2 \n" + sSqlOrder
            # Speichern
            odicSql[sSheet] = sSql
            # if sSheet == "Biota":
            #     print(sSql)

        res.addData("sql", odicSql)
        return res

    def getLookup(self, **kwargs):
        """
        Hole Lookups

        *** Drei Szenarien

        0) Holle alle Lookups (kein Selektor angegeben)
        1) Hole alle Lookups beginnend bei Subset (Metadatenstümpfe werden nicht mit geholt)
        2) Hole alle Lookups mit Metadatenstümpfen
            - Selector besteht nur aus Ingest Ids
            - Nur so können die Metadaten eingesammelt werden

        Kwargs
            selector (dic): Selektor
                    sample_date_start (str)
                    sample_date_end (str)
                    workflow (lst)
                    dataset (lst)
                    gear (lst)
                    taxcover (lst)
                    ingest (lst)
            sheets (lst): Liste der zu Exportierenden Lookup Tabellen
            showid (bool): Sollen die Ids mit exportiert werden
            gold (str): Exportiere nur folgende Gold Lookups (all,not-gold,gold)
        Returns:
            Result:
                header (odic): Header
                content (odic): Tabellen
                rows (dic): Zeilen
                log (str): Log
        """

        # Helper: Clean up
        def cleanUp():
            # DB
            try:
                cursor.close()
                conn.close()
            except:
                pass

        # ---

        pn = self.__smartprinter.normal
        pv = self.__smartprinter.verbose
        pd = self.__smartprinter.debug
        progress = self.__smartprinter.progress
        status = self.__smartprinter.status

        res = Result(name="Get lookups from db")

        # ---

        # Kwargs
        lstSheetsOnly = kwargs["sheets"] if "sheets" in kwargs else None
        bShowId = True if "showid" in kwargs and kwargs["showid"] in ["true", "True", True] else False
        dicSelector = kwargs["selector"] if "selector" in kwargs else None
        sGold = kwargs["gold"] if "gold" in kwargs else "all"

        # Ingest Mode (Metadatastümpfe)
        if (dicSelector is not None) and (len(dicSelector) == 1) and ("ingest" in dicSelector) and len(
                dicSelector["ingest"]) > 0:
            bIngestMode = True
            lstIngest = kwargs["selector"]["ingest"]
        else:
            bIngestMode = False

        # Parameter Tests
        if bIngestMode:
            for iId in lstIngest:
                try:
                    int(iId)
                except:
                    res.addErr("ingest id is not valid")
                    return res

        # Info
        sLog = ""
        sLog = "************************************\n"
        sLog += "Exported Lookups\n"
        sLog += "************************************\n"
        sLog += "\n"
        if dicSelector is None:
            sLog += "No selector given -> getting all lookups\n"
        else:
            if bIngestMode:
                sLog += "Mode: ingest (metadata-stumps are used for collection either)\n"
                sLog += "ingest ids: " + PrettyText.lst2strC(lstIngest) + "\n"
            else:
                sLog += "Mode: normal (no metadata-stumps used)\n"
            sLog += "\n--- Selector ---\n"
            for k, v in dicSelector.items():
                sLog += str(k) + " : " + PrettyText.lst2strC(v) + "\n"
        if sGold in ["not-gold", "gold"]:
            sLog += "Gold: export only " + sGold + "\n"
        sLog += "\n"

        # ---

        if lstSheetsOnly is not None:
            lstSheetUnbekannt = list(set(lstSheetsOnly).difference(set(self.__map.lookup_getExcelSheets())))
            if len(lstSheetUnbekannt) > 0:
                cleanUp()
                res.addErr("Unknown lookup: " + PrettyText.lst2strC(lstSheetUnbekannt))
                return res

        # ---

        # DB
        resConn = self.__model.subsystem_getDb().conn_getConnection()
        if resConn.isErr():
            cleanUp()
            return resConn
        conn = resConn.getData("connection")
        cursor = resConn.getData("cursor")

        # Ids
        dicIds = None
        if dicSelector is not None:
            # Selector
            resIds = self.__getLookup_ids(dicSelector)
            if resIds.isErr():
                cleanUp()
                return resIds
            dicIds = resIds.getData("ids")

        # SQL
        resSql = self.__getLookup_content_sql(ids=dicIds, showid=bShowId, gold=sGold)
        if resSql.isErr():
            cleanUp()
            return resSql
        dicSql = resSql.getData("sql")

        # Loop
        iNrMax = len(dicSql)
        odicAntwort = OrderedDict()
        for iNr, (sSheet, sSql) in enumerate(dicSql.items()):
            progress(iNr + 1, iNrMax, 100, proz=True, abs=True, cli=True, text=sSheet)
            if (lstSheetsOnly is None) or (sSheet in lstSheetsOnly):
                cursor.execute(sSql)
                lstHeader = [head[0] for head in cursor.description]
                lstContent = cursor.fetchall()
                odicAntwort[sSheet] = {
                    "sql": sSql,
                    "header": lstHeader,
                    "content": lstContent,
                    "rows": len(lstContent)
                }
                sLog += "--- " + sSheet + " ---\n" + sSql + "\n"

        # Antwort
        res.addData("log", sLog)
        res.addData("lookup", odicAntwort)

        # End
        cleanUp()
        cursor.close()
        conn.close()

        return res

    """
    Hole erforderliche IDs für lookups
    """

    def __getLookup_ids(self, dicSelector):

        pd = self.__smartprinter.debug
        res = Result(name="Get ids for lookups from db")

        if "ingest" in dicSelector:
            lstIngest = dicSelector["ingest"]
        else:
            lstIngest = None

        # DB
        resConn = self.__model.subsystem_getDb().conn_getConnection()
        if resConn.isErr():
            res.addResult(resConn)
            return res
        conn = resConn.getData("connection")
        cursor = resConn.getData("cursor")
        # SQL-Cmds
        # - Selector
        resSqlSelector = self.__getSelector_sql(dicSelector)
        if resSqlSelector.isErr():
            res.addResult(resSqlSelector)
            return res
        sSqlSelector = resSqlSelector.getData("sql")
        # - Lookups
        resSql = self.__getLookup_ids_sql(lstIngest)
        if resSql.isErr():
            res.addResult(resSql)
            return res
        odicSql = resSql.getData("sql")

        # Sample Lookup IDs
        dicLookupUsed_directly = dict()
        dicLookupUsed_indirectly = dict()
        try:
            pd("* Getting ids for lookups")
            pd("Ingest ids: " + PrettyText.lst2strC(lstIngest))
            pd("* SQL")
            pd("--- Directly")
            # --- 1) Direkt (Ingest)
            # Selector View
            pd("- Selector")
            pd(sSqlSelector)
            cursor.execute(sSqlSelector)
            # Loop: Sheets
            for sSheet, sSql in odicSql.items():
                pd("- " + str(sSheet))
                pd(sSql)
                cursor.execute(sSql)
                # Antwort
                lstHeader = [head[0] for head in cursor.description]
                lstContent = cursor.fetchall()
                # Sammle ids für die Spalten (Lookups) ein
                for row in lstContent:
                    for iCol, col in enumerate(lstHeader):
                        key = col
                        val = row[iCol]
                        if (key is not None) and (val not in [None, ""]):
                            if key in dicLookupUsed_directly:
                                dicLookupUsed_directly[key].add(val)
                            else:
                                dicLookupUsed_directly[key] = {val}

            # --- 2) Indirekt (Sammle alle Lookups ein, die von Lookups gebraucht werden)
            if len(dicLookupUsed_directly) > 0:
                pd("--- Indirectly")
                # Baue Collect Liste zusammen
                lstLookupsNeedingLookups = self.__map.lookup_moreLookupNeeded_getAllNeedingTables()
                lstCollect = []
                for sTable in lstLookupsNeedingLookups:
                    lstNeedingAttribs = self.__map.lookup_moreLookupNeeded_getNeedingAttributes(sTable)
                    for sNeedingAttribute in lstNeedingAttribs:
                        sNeededLookup = self.__map.lookup_moreLookupNeeded_getNeededLookup(sTable, sNeedingAttribute)
                        lstCollect.append((sNeededLookup, sTable, sNeedingAttribute))
                # Einsammeln
                for sLookup, sTable, sAttr in lstCollect:
                    sSheet = self.__map.lookup_table2sheet(sTable)
                    sSheetLookup = self.__map.lookup_table2sheet(sLookup)
                    tpUsedIds = tuple(dicLookupUsed_directly[sSheet])
                    sqlCmd = sql.SQL(
                        "SELECT {attr} FROM {schema}.{table} WHERE id IN %s").format(
                        schema=sql.Identifier(self.__model.subsystem_getDb().account_active_getSchema()),
                        table=sql.Identifier(sTable),
                        attr=sql.Identifier(sAttr)
                    )
                    cursor.execute(sqlCmd, (tpUsedIds,))
                    sSqlQuery = cursor.query.decode("utf-8")
                    pd("- Collecting " + str(sSheetLookup) + " ids from " + sSheet)
                    pd(sSqlQuery)
                    # Antwort
                    lstContent = cursor.fetchall()
                    for row in lstContent:
                        val = row[0]
                        if val not in [None, ""]:
                            if sSheetLookup in dicLookupUsed_indirectly:
                                dicLookupUsed_indirectly[sSheetLookup].add(val)
                            else:
                                dicLookupUsed_indirectly[sSheetLookup] = {val}

            # Kombiniere Ergebnisse
            dicLookupUsed = dicLookupUsed_directly.copy()
            for sSheet, setIds in dicLookupUsed_indirectly.items():
                if sSheet in dicLookupUsed:
                    dicLookupUsed[sSheet].update(setIds)
                else:
                    dicLookupUsed[sSheet] = setIds

            # Info
            pd("* Results")
            pd("--- Directly")
            if len(dicLookupUsed_directly) > 0:
                for sSheet, setIds in dicLookupUsed_directly.items():
                    pd(sSheet + ": " + PrettyText.lst2strC(setIds))
            else:
                pd("nothing")
            pd("--- Indirectly")
            if len(dicLookupUsed_indirectly) > 0:
                for sSheet, setIds in dicLookupUsed_indirectly.items():
                    pd(sSheet + ": " + PrettyText.lst2strC(setIds))
            else:
                pd("nothing")
            pd("--- Combined")
            if len(dicLookupUsed_indirectly) > 0:
                for sSheet, setIds in dicLookupUsed.items():
                    pd(sSheet + ": " + PrettyText.lst2strC(setIds))
            else:
                pd("nothing")
            # Ergebnis
            res.addData("ids", dicLookupUsed)

        except Exception as ex:
            res.addErr("Getting lookup ids from database: " + str(ex))

        return res

    """
    SQL-Kommandos: Einsammeln der Ids für die benötigten Lookups
    """

    def __getLookup_ids_sql(self, lstIngest=None):
        """
        Args:
            ingest (lst): Ingest Liste
        Kwargs:
            gold (str): Nur Ids sammeln für folgenden Gold Status: all,not-gold,gold
        Returns:
            Result:
                sql (odic): SQL-Befehle
        """

        # Helper: Erzeuge eine Kurzbezeichung von schema.tabelle -> kurz
        def short(sTable):
            # Tweak: Doppelte Kürzel auflösen: subset/subsample->sub, status->sts/station->sta
            if sTable == "subsample":
                return "subsmp"
            return str(sTable)[0:3]

        # Helper: Erzeuge aus Liste eine String-Liste
        def lst2str(lst):
            if lst is None or len(lst) == 0:
                return ""
            sOut = ""
            for e in lst:
                sOut += "\'" + str(e) + "\',"
            return sOut[:-1]

        # ---

        res = Result(name="Create SQL statements for lookup download")

        bSubsetTableReached = False
        bActualTableReached = False
        odicSql = OrderedDict()
        sDbSchema = self.__model.subsystem_getDb().account_active_getSchema()
        bIngest = True if (lstIngest is not None and len(lstIngest) > 0) else False

        # Loop: Sheets -> Sammle alle Lookups ein
        bSubsetTableReached = False
        lstSheets = self.__map.ingest_getExcelSheets()
        for iCol, sSheet in enumerate(lstSheets):

            # Hat das Sheet Lookups?
            if self.__map.ingest_hasLookup(sSheet):
                # Ja! -> Einsammeln
                lstCols = self.__map.ingest_getExcelCols(sSheet)
                sTable = self.__map.ingest_sheet2table(sSheet)

                # Modus: Ingest-Export: Ab Subset explizit keine Metadaten mehr einsammeln
                if sTable == "subset":
                    bSubsetTableReached = True

                # SQL-Befehlsteile
                sSqlSelection = ""
                sSqlFrom = ""
                sSqlJoinFk = ""
                sSqlJoinLookup = ""
                sSqlWhere = "WHERE sub.id IN (SELECT id FROM selector) \n"
                sSqlUnionJoinFk = ""

                # --- Loop: Columns
                for sCol in lstCols:
                    # DB-Table/Attribute
                    sTableTmp, sAttrib = self.__map.ingest_excel2db(sSheet, sCol)
                    # FK
                    bFk = self.__map.ingest_isFk(sSheet, sCol)
                    # Lookup
                    bLup = self.__map.ingest_isLookup(sSheet, sCol)
                    sLupTable = self.__map.ingest_getLookup(sSheet, sCol) if bLup else None
                    sLupAtr = self.__map.lookup_getKey(sLupTable) if bLup else None

                    # --- SELECT Lookups: Haupttabelle
                    if not bFk and bLup:
                        # Select eines Lookup Attributes (select plus left join)
                        sLupSheet = self.__map.lookup_table2sheet(sLupTable)
                        sSqlSelection += short(sLupTable) + ".id AS \"" + str(sLupSheet) + "\", \n"
                        sSqlJoinLookup += "LEFT JOIN <schema>." + str(sLupTable) + " " + short(
                            sLupTable) + " ON " + short(sTable) + "." + str(sAttrib) + "=" + short(sLupTable) + ".id \n"

                # --- JOIN: FK
                if sSheet in ["Cruise", "Station", "Sample", "SubSample"]:
                    # Tweak: Hole bei Cruise,Station,Sample immer das Metadatengerüst Cruise-Station-Sample-Subset hinzu
                    lstFk = self.__map.ingest_getFk("Subset")
                    sFkLast = "Subset"
                    sFkDbLast = "subset"
                else:
                    lstFk = self.__map.ingest_getFk(sSheet)
                    sFkDbLast = sTable
                    sFkLast = sSheet
                if lstFk is not None:
                    bActualTableReached = False
                    for sFk in reversed(lstFk):
                        sFkDb = self.__map.ingest_sheet2table(sFk)
                        sFkMainCol, sFkMainTable, sFkMainAttrib = self.__map.ingest_getFkMain(sFkLast)
                        sTmp = "JOIN <schema>." + str(sFkDb) + " " + short(sFkDb) + " ON " + short(
                            sFkDb) + ".id = " + short(sFkDbLast) + "." + sFkMainAttrib + "\n"
                        sSqlJoinFk += sTmp
                        if bActualTableReached:
                            sSqlUnionJoinFk += sTmp
                        if sFkDb == sTable:
                            bActualTableReached = True
                        sFkDbLast = sFkDb
                        sFkLast = sFk

                # --- FROM: Haupttabelle
                if sSheet in ["Cruise", "Station", "Sample", "SubSample"]:
                    sSqlFrom = "FROM <schema>.subset " + short("subset") + "\n"
                else:
                    sSqlFrom = "FROM <schema>." + sTable + " " + short(sTable) + "\n"

                # ---

                # Sql-Zusammenbauen
                sSqlSelection = "SELECT DISTINCT \n" + sSqlSelection[:-3] + "\n"
                sSql = sSqlSelection + sSqlFrom + sSqlJoinFk + sSqlJoinLookup + sSqlWhere
                if bIngest and not bSubsetTableReached:
                    sSql += "UNION \n" + sSqlSelection + "FROM <schema>." + sTable + " " + short(
                        sTable) + "\n" + sSqlUnionJoinFk + sSqlJoinLookup + "WHERE " + short(
                        sTable) + ".ingest_id IN (" + lst2str(lstIngest) + ") \n"
                sSql = sSql.replace("<schema>", sDbSchema)
                odicSql[sSheet] = sSql

        res.addData("sql", odicSql)
        return res

    """
    SQL-Kommandos: Hole die eigentlichen Lookup Daten
    
    Kwargs
        ids (dic): Lookups und Ids zum Exportieren
        showid (bool): Pro Tabelle: Sollen die Ids mit exportiert werden
    Returns:
        Result:
            sql (odic): Befehle
    """

    def __getLookup_content_sql(self, **kwargs):

        # Helper: Erzeuge eine Kurzbezeichung von schema.tabelle -> kurz
        def short(sTable):
            # Tweak: Doppelte Kürzel auflösen: status->sts/station->sta
            if sTable == "status":
                return "sts"
            return str(sTable)[0:3]

        # Helper: Erzeuge aus Liste eine String-Liste
        def lst2str(lst):
            if lst is None or len(lst) == 0:
                return ""
            sOut = ""
            for e in lst:
                sOut += "\'" + str(e) + "\',"
            return sOut[:-1]

        # ---

        pd = self.__smartprinter.debug

        res = Result(name="Create SQL commands for getting lookups content from db")

        # Parameters
        bShowId = True if "showid" in kwargs and kwargs["showid"] in ["true", "True", True] else False
        sGold = kwargs["gold"] if "gold" in kwargs else None

        sDbSchema = self.__model.subsystem_getDb().account_active_getSchema()
        dicIds = kwargs["ids"] if "ids" in kwargs else None

        # Kein Ids -> SQL für alle Lookups erzeugen
        if dicIds is None:
            dicIds = {}
            for sLup in self.__map.lookup_getExcelSheets():
                dicIds[sLup] = set()

        # DB
        resConn = self.__model.subsystem_getDb().conn_getConnection()
        if resConn.isErr():
            res.addResult(resConn)
            return res
        conn = resConn.getData("connection")
        cursor = resConn.getData("cursor")

        # SQL
        dicSql = dict()
        # Loop: Table
        for sSheet, setIds in dicIds.items():
            sTable = self.__map.lookup_sheet2table(sSheet)
            sTableKeyAttrib = self.__map.lookup_getKey(sTable)

            # Andere Lookups
            bTableNeedsMoreLookups = self.__map.lookup_moreLookupNeeded_isTableNeeding(sTable)
            lstNeedingAttributes = self.__map.lookup_moreLookupNeeded_getNeedingAttributes(
                sTable) if bTableNeedsMoreLookups else None

            # SQL
            sSqlSelect = "SELECT \n"
            if bShowId:
                sSqlSelect += short(sTable) + ".id AS \"id\", "
            sSqlFrom = "FROM <schema>." + sTable + " " + short(sTable) + " \n"
            sSqlJoin = ""
            sSqlWhere = ""
            if len(setIds) > 0:
                sSqlWhere = "WHERE " + short(sTable) + ".id in (<ids>) \n"
            # Export: Gold/Not-Gold
            if (sGold in ["not-gold", "gold"]) and (sTable != "taxon"):
                if sSqlWhere == "":
                    sSqlWhere = "WHERE "
                else:
                    sSqlWhere += " AND "
                if sGold == "gold":
                    sSqlWhere += (short(sTable) + ".gold=True \n")
                else:
                    sSqlWhere += (short(sTable) + ".gold=False \n")
            if sTableKeyAttrib is not None:
                sSqlOrder = "ORDER by " + short(sTable) + "." + sTableKeyAttrib + " \n"
            else:
                sSqlOrder = ""

            # Loop: Attribute
            lstAttrib = self.__map.lookup_getDbAttributes(sTable)
            for sAttrib in lstAttrib:
                sSheetTmp, sCol = self.__map.lookup_db2excel(sTable, sAttrib)
                if bTableNeedsMoreLookups and (sAttrib in lstNeedingAttributes):
                    # Attribut brauch anderes Lookup
                    sNeededLookupTable = self.__map.lookup_moreLookupNeeded_getNeededLookup(sTable, sAttrib)
                    sNeededLookupKeyAttrib = self.__map.lookup_getKey(sNeededLookupTable)
                    # SELECT
                    sSqlSelect += short(sNeededLookupTable) + "." + sNeededLookupKeyAttrib + " AS \"" + sCol + "\", \n"
                    # JOIN
                    sSqlJoin += "LEFT JOIN <schema>." + str(sNeededLookupTable) + " " + short(
                        sNeededLookupTable) + " ON " + short(sTable) + "." + str(sAttrib) + "=" + short(
                        sNeededLookupTable) + ".id \n"
                else:
                    # Normales Attribut
                    # SELECT
                    sSqlSelect += short(sTable) + "." + sAttrib + " AS \"" + sCol + "\", \n"
            sSqlSelect = sSqlSelect[:-3] + " \n"
            sSql = sSqlSelect + sSqlFrom + sSqlJoin + sSqlWhere + sSqlOrder
            sSql = sSql.replace("<schema>", sDbSchema)
            if len(setIds) > 0:
                sSql = sSql.replace("<ids>", lst2str(setIds))
            dicSql[sSheet] = sSql

        res.addData("sql", dicSql)

        return res

    def getTaxon(self, **kwargs):
        """
        Hole Taxon System

        Kwargs
            sheets (lst): Liste der zu Exportierenden Lookup Tabellen
            showid (bool): Sollen die Ids mit exportiert werden
        Returns:
            Result:
                header (odic): Header
                content (odic): Tabellen
                rows (dic): Zeilen
                log (str): Log
        """

        # Helper: Clean up
        def cleanUp():
            # DB
            try:
                cursor.close()
                conn.close()
            except:
                pass

        # ---

        pn = self.__smartprinter.normal
        pv = self.__smartprinter.verbose
        pd = self.__smartprinter.debug
        progress = self.__smartprinter.progress
        status = self.__smartprinter.status

        res = Result(name="Get taxon system admin tables from db")

        # ---

        # Kwargs
        lstSheetsOnly = kwargs["sheets"] if "sheets" in kwargs else None
        bShowId = True if "showid" in kwargs and kwargs["showid"] in ["true", "True", True] else False

        # Info
        sLog = ""
        sLog = "************************************\n"
        sLog += "Exported Taxon System\n"
        sLog += "************************************\n"
        sLog += "\n"

        # ---

        if lstSheetsOnly is not None:
            lstSheetUnbekannt = list(set(lstSheetsOnly).difference(set(self.__map.taxon_getExcelSheets())))
            if len(lstSheetUnbekannt) > 0:
                cleanUp()
                res.addErr("Unknown taxon admin table: " + PrettyText.lst2strC(lstSheetUnbekannt))
                return res

        # ---

        # DB
        resConn = self.__model.subsystem_getDb().conn_getConnection()
        if resConn.isErr():
            cleanUp()
            return resConn
        conn = resConn.getData("connection")
        cursor = resConn.getData("cursor")

        # SQL
        resSql = self.__getTaxon_sql(showid=bShowId)
        if resSql.isErr():
            cleanUp()
            return resSql
        dicSql = resSql.getData("sql")
        # Loop
        iNrMax = len(dicSql)
        odicAntwort = OrderedDict()
        for iNr, (sSheet, sSql) in enumerate(dicSql.items()):
            progress(iNr + 1, iNrMax, 100, proz=True, abs=True, cli=True, text=sSheet)
            if (lstSheetsOnly is None) or (sSheet in lstSheetsOnly):
                cursor.execute(sSql)
                lstHeader = [head[0] for head in cursor.description]
                lstContent = cursor.fetchall()
                odicAntwort[sSheet] = {
                    "sql": sSql,
                    "header": lstHeader,
                    "content": lstContent,
                    "rows": len(lstContent)
                }
                sLog += "--- " + sSheet + " ---\n" + sSql + "\n"

        # Antwort
        res.addData("log", sLog)
        res.addData("taxon", odicAntwort)

        # End
        cleanUp()
        cursor.close()
        conn.close()

        return res

    """
    SQL-Kommandos: Hole die Taxon System Tabellen
    
    Kwargs
       showid (bool): Pro Tabelle: Sollen die Ids mit exportiert werden
    Returns:
       Result:
           sql (odic): Befehle
    """

    def __getTaxon_sql(self, **kwargs):

        # Helper: Erzeuge eine Kurzbezeichung von schema.tabelle -> kurz
        def short(sTable):
            # Tweak: Doppelte Kürzel auflösen: status->sts/station->sta
            return str(sTable)[0:3]

        # Helper: Erzeuge aus Liste eine String-Liste
        def lst2str(lst):
            if lst is None or len(lst) == 0:
                return ""
            sOut = ""
            for e in lst:
                sOut += "\'" + str(e) + "\',"
            return sOut[:-1]

        # ---

        pd = self.__smartprinter.debug

        res = Result(name="Create SQL commands for getting taxon admin tables from db")

        # Parameters
        bShowId = True if "showid" in kwargs and kwargs["showid"] in ["true", "True", True] else False
        sDbSchema = self.__model.subsystem_getDb().account_active_getSchema()

        # DB
        resConn = self.__model.subsystem_getDb().conn_getConnection()
        if resConn.isErr():
            res.addResult(resConn)
            return res
        conn = resConn.getData("connection")
        cursor = resConn.getData("cursor")

        # SQL
        dicSql = dict()
        # Loop: Table
        for sSheet in self.__map.taxon_getExcelSheets():
            sTable = self.__map.taxon_sheet2table(sSheet)
            # SQL
            sSqlSelect = "SELECT \n"
            if bShowId:
                sSqlSelect += short(sTable) + ".id AS \"id\", "
            sSqlFrom = "FROM <schema>." + sTable + " \n"

            # Attribute
            lstAttrib = self.__map.taxon_getDbAttributes(sTable)
            for sAttrib in lstAttrib:
                sSheetTmp, sCol = self.__map.taxon_db2excel(sTable, sAttrib)
                sSqlSelect += sTable + "." + sAttrib + " AS \"" + sCol + "\", \n"
            sSqlSelect = sSqlSelect[:-3] + " \n"
            sSql = sSqlSelect + sSqlFrom
            sSql = sSql.replace("<schema>", sDbSchema)
            dicSql[sSheet] = sSql

        res.addData("sql", dicSql)

        return res

    def writeExcel(self, **kwargs):
        """
        Exportiere Template (entweder speichern in Datei oder Rückgabe als Workbook)
        Args:
        Kwargs
            lookup_data (dic/odic): Lookup Daten
            ingest_data (dic/odic): Ingest Daten
            taxon_data (dic/odic): Taxon Daten

            ingest (bool): Ingest exportiern
            ingest_wf (str): Ingest Workflow
            ingest_sheets (lst): Ingest Sheets
            lookup (bool): Lookup exportiern
            lookup_sheets (bool): Lookup Sheets
            taxon (bool): Taxon exportieren
            taxon_sheets (bool): Taxon Sheets

            converter (str): Converter
            pangaea (bool): Alles im PANGAEA format exportieren

            file (str): File
        Returns:
            wb (Workbench): Wenn kein File angegeben wurde, ansonsten wird einfach die Datei geschrieben
        """

        # Helper: Erzeuge eine Kurzbezeichung von tabelle -> kurz
        def short(sTable):
            # Tweak: Doppelte Kürzel auflösen: status->sts/station->sta
            if sTable == "status":
                return "sts"
            if sTable == "subsample":
                return "subsmp"
            return str(sTable)[0:3]

        pn = self.__smartprinter.normal
        pv = self.__smartprinter.verbose
        pd = self.__smartprinter.debug
        progress = self.__smartprinter.progress
        status = self.__smartprinter.status

        res = Result(name="Write data to Excel")

        progress("pulse")

        # --- Parameter
        bPangaea = True if "pangaea" in kwargs and kwargs["pangaea"] in ["true", "True", True] else False
        bRfriendly = True if "rfriendly" in kwargs and kwargs["rfriendly"] in ["true", "True", True] else False
        sConverter = kwargs.get("converter")

        bIngest = True if "ingest" in kwargs and kwargs["ingest"] in ["true", "True", True] else False
        sIngestWf = kwargs["ingest_wf"] if "ingest_wf" in kwargs else None
        lstIngestSheets = kwargs["ingest_sheets"] if "ingest_sheets" in kwargs else None

        bLookup = True if "lookup" in kwargs and kwargs["lookup"] in ["true", "True", True] else False
        lstLookupSheets = kwargs["lookup_sheets"] if "lookup_sheets" in kwargs else None
        bTaxon = True if "taxon" in kwargs and kwargs["taxon"] in ["true", "True", True] else False
        lstTaxonSheets = kwargs["taxon_sheets"] if "taxon_sheets" in kwargs else None

        # Data
        odicLookupData = kwargs["lookup_data"] if "lookup_data" in kwargs else None
        odicIngestData = kwargs["ingest_data"] if "ingest_data" in kwargs else None
        odicTaxonData = kwargs["taxon_data"] if "taxon_data" in kwargs else None

        # Template
        bTemplate = True
        if bIngest and odicIngestData is not None:
            bTemplate = False
        if bLookup and odicLookupData is not None:
            bTemplate = False
        if bTaxon and odicTaxonData is not None:
            bTemplate = False

        # File
        sFile = kwargs["file"] if "file" in kwargs else None

        # Info
        odicInfo = OrderedDict()
        if bIngest:
            odicInfo["ingest"] = OrderedDict()
        if bLookup:
            odicInfo["lookup"] = OrderedDict()
        if bTaxon:
            odicInfo["taxon"] = OrderedDict()

        # --- Tests
        if bIngest and (sIngestWf is not None) and (sIngestWf not in self.__map.workflow_getAll()):
            res.addErr("Workflow not valid: " + sIngestWf)
            return res
        if bIngest and lstIngestSheets is not None:
            lstSheetUnbekannt = list(set(lstIngestSheets).difference(set(self.__map.ingest_getExcelSheets())))
            if len(lstSheetUnbekannt) > 0:
                res.addErr("Unknown ingest sheet: " + PrettyText.lst2strC(lstSheetUnbekannt))
                return res
        if bLookup and lstLookupSheets is not None:
            lstSheetUnbekannt = list(set(lstLookupSheets).difference(set(self.__map.lookup_getExcelSheets())))
            if len(lstSheetUnbekannt) > 0:
                res.addErr("Unknown lookup sheet: " + PrettyText.lst2strC(lstSheetUnbekannt))
                return res
        if bTaxon and lstTaxonSheets is not None:
            lstSheetUnbekannt = list(set(lstTaxonSheets).difference(set(self.__map.taxon_getExcelSheets())))
            if len(lstSheetUnbekannt) > 0:
                res.addErr("Unknown taxon sheet: " + PrettyText.lst2strC(lstSheetUnbekannt))
                return res
        if (lstIngestSheets is not None) and (len(lstIngestSheets) == 0):
            bIngest = False
        if (lstLookupSheets is not None) and (len(lstLookupSheets) == 0):
            bLookup = False
        if (lstTaxonSheets is not None) and (len(lstTaxonSheets) == 0):
            bTaxon = False

        if not bIngest and not bLookup and not bTaxon:
            res.addErr("Nothing to export")
            return res

        # Aufbauen
        wb = openpyxl.Workbook()
        bFirstSheet = True

        # TODO: Tweak: Update Comments
        self.__model.conf_getDbSchema().update()

        # Row/MaxRow
        iRow = 0
        iRowMax = 0
        if bLookup and odicLookupData is not None:
            for sH in odicLookupData:
                if lstLookupSheets is None or sH in lstLookupSheets:
                    iRowMax += odicLookupData[sH]["rows"]
        if bTaxon and odicTaxonData is not None:
            for sH in odicTaxonData:
                if lstTaxonSheets is None or sH in lstTaxonSheets:
                    iRowMax += odicTaxonData[sH]["rows"]
        if bIngest and odicIngestData is not None:
            for sH in odicIngestData:
                if sH != "Selector":
                    if lstIngestSheets is None or sH in lstIngestSheets:
                        iRowMax += odicIngestData[sH]["rows"]

        # --- Ingest
        if bIngest:
            if sConverter is not None:
                c = self.__model.subsystem_getConverter()
                resConverter = c.create_converter_export_excel(sConverter, data=odicIngestData)
                if resConverter.isErr():
                    res.addResult(resConverter)
                    return res
                wb = resConverter.getData("wb")
                bFirstSheet = False
            else:
                # Normal
                for sSheet in self.__map.ingest_getExcelSheets(sIngestWf):
                    if (lstIngestSheets is None) or (sSheet in lstIngestSheets):
                        if bFirstSheet:
                            ws = wb.active
                            ws.title = sSheet
                            bFirstSheet = False
                        else:
                            ws = wb.create_sheet(title=sSheet)
                        # Header
                        lstHeader = []
                        lstCols = self.__map.ingest_getExcelCols(sSheet, sIngestWf)
                        if bRfriendly:
                            # *** TWEAK: R-friendly: Header Veränderung
                            # --- Biota: Taxonomie erweitern
                            # kleiner "Bug": Hier wird fälschlicherweise ScientificName später im Format-Bereich wiedergefunden und als Obligat (weil für ScientificName user given gehalten) definiert
                            if sSheet == "Biota":
                                iIndex_name = lstCols.index("ScientificName")
                                del (lstCols[iIndex_name])
                                for iIdx, (sHeader, sComment) in enumerate(self.__odicRfriendlyTweak["Biota"].items()):
                                    lstCols.insert(iIndex_name + iIdx, sHeader)
                            # --- ID-Block voranstellen
                            # FK
                            sTable = self.__map.ingest_sheet2table(sSheet)
                            # PK
                            lstCols.insert(0, str(sSheet) + "ID")
                            lstFk_r = self.__map.ingest_getFk(sSheet)
                            if lstFk_r is not None:
                                for sSheet_r in reversed(lstFk_r):
                                    sTable_r = self.__map.ingest_sheet2table(sSheet_r)
                                    lstCols.insert(0, str(sSheet_r) + "ID")
                        for sCol in lstCols:
                            resComment = self.__model.conf_getDbSchema().getDbSchemaInfo("ingest", sSheet, sCol)
                            if bPangaea and resComment.isOk():
                                # Pangaea
                                sValPangaea = resComment.getData("comment-full-pangaea")
                                cell = openpyxl.cell.WriteOnlyCell(ws, value=sValPangaea)
                            else:
                                # Normal
                                cell = openpyxl.cell.WriteOnlyCell(ws, value=sCol)
                            # Format
                            bObl = self.__map.ingest_isObl(sSheet, sCol)
                            bPk = self.__map.ingest_isPk(sSheet, sCol)
                            bFk = self.__map.ingest_isFk(sSheet, sCol)
                            # bLup = self.__map.ingest_isLookup(sSheet, sCol)
                            bBold = False
                            bItalic = False
                            bUnderline = False
                            if bPk:
                                bBold = True
                            if bFk:
                                bBold = True
                                bItalic = True
                            if bObl and not (bPk or bFk):
                                bUnderline = True
                            cell.font = openpyxl.styles.Font(italic=bItalic, bold=bBold,
                                                             underline=("single" if bUnderline else None))
                            if resComment.isOk():
                                sComment = resComment.getData("comment-full")
                                if sComment not in (None, ""):
                                    cell.comment = openpyxl.comments.Comment(text=sComment, author=None)
                            # TWEAK: R-friendly: Header-Kommentar Veränderung
                            if bRfriendly and sSheet == "Biota" and sCol in self.__odicRfriendlyTweak["Biota"]:
                                sComment = self.__odicRfriendlyTweak["Biota"][sCol]
                                cell.comment = openpyxl.comments.Comment(text=sComment, author=None)
                            lstHeader.append(cell)
                        ws.append(lstHeader)
                        progress(iRow, iRowMax, 100, proz=True, abs=False, cli=True, text=sSheet)
                        iRow += 1
                        # Daten
                        if odicIngestData is not None and sSheet in odicIngestData:
                            lstDataHeader = odicIngestData[sSheet]["header"]
                            lstDataContent = odicIngestData[sSheet]["content"]
                            if lstDataHeader == lstCols:
                                for row in lstDataContent:
                                    ws.append(row)
                            else:
                                res.addErr("Ingest template/data header not consistent: ", add=PrettyText.lst2strC(
                                    lstCols) + " != " + PrettyText.lst2strC(lstDataHeader), kat="Internal problem",
                                           sys="Export Ingest Template")
                            odicInfo["ingest"][sSheet] = len(lstDataContent)  # Info

        # --- Lookup
        if bLookup:
            for sSheet in self.__map.lookup_getExcelSheets():
                if (lstLookupSheets is None) or (sSheet in lstLookupSheets):
                    if bFirstSheet:
                        ws = wb.active
                        ws.title = sSheet
                        bFirstSheet = False
                    else:
                        ws = wb.create_sheet(title=sSheet)
                    # Header
                    lstHeader = []
                    lstCols = self.__map.lookup_getExcelCols(sSheet)
                    for sCol in lstCols:
                        resComment = self.__model.conf_getDbSchema().getDbSchemaInfo("lookup", sSheet, sCol)
                        if bPangaea and resComment.isOk():
                            # Pangaea
                            sValPangaea = resComment.getData("comment-full-pangaea")
                            cell = openpyxl.cell.WriteOnlyCell(ws, value=sValPangaea)
                        else:
                            # Normal
                            cell = openpyxl.cell.WriteOnlyCell(ws, value=sCol)
                        # Format
                        sTable, sAttrib = self.__map.lookup_excel2db(sSheet, sCol)
                        bObl = self.__map.lookup_isObl(sSheet, sCol)
                        bKey = self.__map.lookup_isKey(sSheet, sCol)
                        bLup2 = self.__map.lookup_moreLookupNeeded_isAttributeNeeding(sTable, sAttrib)
                        bBold = False
                        bItalic = False
                        bUnderline = False
                        if bKey:
                            bBold = True
                        if bObl and not bKey:
                            bUnderline = True
                        if bLup2:
                            bItalic = True
                        cell.font = openpyxl.styles.Font(italic=bItalic, bold=bBold,
                                                         underline=("single" if bUnderline else None))
                        if resComment.isOk():
                            sComment = resComment.getData("comment-full")
                            if sComment not in (None, ""):
                                cell.comment = openpyxl.comments.Comment(text=sComment, author=None)
                        lstHeader.append(cell)
                    ws.append(lstHeader)

                    # Daten
                    if odicLookupData is not None and sSheet in odicLookupData:
                        lstDataHeader = odicLookupData[sSheet]["header"]
                        lstDataContent = odicLookupData[sSheet]["content"]
                        if lstDataHeader == lstCols:
                            for row in lstDataContent:
                                ws.append(row)
                                progress(iRow, iRowMax, 100, proz=True, abs=False, cli=True,
                                         text=str(sSheet))
                                iRow += 1
                        else:
                            res.addErr("Lookup template/data header not consistent",
                                       add=PrettyText.lst2strC(lstCols) + " != " + PrettyText.lst2strC(lstDataHeader),
                                       kat="Internal problem", sys="Export Lookup Template")
                        odicInfo["lookup"][sSheet] = len(lstDataContent)  # Info

        # --- Taxon
        if bTaxon:
            for sSheet in self.__map.taxon_getExcelSheets():
                if (lstTaxonSheets is None) or (sSheet in lstTaxonSheets):
                    if bFirstSheet:
                        ws = wb.active
                        ws.title = sSheet
                        bFirstSheet = False
                    else:
                        ws = wb.create_sheet(title=sSheet)
                    lstHeader = []
                    lstCols = self.__map.taxon_getExcelCols(sSheet)
                    for sCol in lstCols:
                        resComment = self.__model.conf_getDbSchema().getDbSchemaInfo("taxon", sSheet, sCol)
                        if bPangaea and resComment.isOk():
                            # Pangaea
                            sValPangaea = resComment.getData("comment-full-pangaea")
                            cell = openpyxl.cell.WriteOnlyCell(ws, value=sValPangaea)
                        else:
                            # Normal
                            cell = openpyxl.cell.WriteOnlyCell(ws, value=sCol)
                        bObl = self.__map.taxon_isObl(sSheet, sCol)
                        # Format
                        bBold = False
                        bItalic = False
                        bUnderline = False
                        if bObl:
                            bUnderline = True
                        cell.font = openpyxl.styles.Font(italic=bItalic, bold=bBold,
                                                         underline=("single" if bUnderline else None))
                        if resComment.isOk():
                            sComment = resComment.getData("comment-full")
                            if sComment not in (None, ""):
                                cell.comment = openpyxl.comments.Comment(text=sComment, author=None)
                        lstHeader.append(cell)
                    ws.append(lstHeader)

                    # Daten
                    if odicTaxonData is not None and sSheet in odicTaxonData:
                        lstDataHeader = odicTaxonData[sSheet]["header"]
                        lstDataContent = odicTaxonData[sSheet]["content"]
                        if lstDataHeader == lstCols:
                            for row in lstDataContent:
                                ws.append(row)
                                progress(iRow, iRowMax, 100, proz=True, abs=False, cli=True,
                                         text=sSheet)
                                iRow += 1
                        else:
                            res.addErr("TaxSys template/data header not consistent", add=PrettyText.lst2strC(
                                lstCols) + " != " + PrettyText.lst2strC(lstDataHeader), kat="Internal problem",
                                       sys="Export Taxon Template")
                        odicInfo["taxon"][sSheet] = len(lstDataContent)  # Info

        # Glossar
        borderTop = openpyxl.styles.borders.Border(left=None,
                                                   right=None,
                                                   top=None,
                                                   bottom=openpyxl.styles.Side(style='thin'))
        wsGlossary = wb.create_sheet(title="Glossary")
        iRowGlossary = 1
        # Ingest
        if bIngest:
            cellGlossary = wsGlossary.cell(column=1, row=1, value="INGEST")
            cellGlossary.font = openpyxl.styles.Font(bold=True)
            iRowGlossary += 1
            sCommentAttributes = "PK: primary key (bold)\nFK: foreign key (bold, italic)\nOBL: obligate (underscore)\nLUP: lookup"
            for iIdx, sName in enumerate(
                    ("Sheet", "Column", "Attributes", "Name", "Unit", "Comment")):
                cellGlossary = wsGlossary.cell(column=iIdx + 1, row=iRowGlossary, value=sName)
                cellGlossary.font = openpyxl.styles.Font(bold=True)
                # cellGlossary.border = borderTop
                if sName == "Attributes":
                    cellGlossary.comment = openpyxl.comments.Comment(text=sCommentAttributes, author=None)
            iRowGlossary += 1
            for sSheet in self.__map.ingest_getExcelSheets(sIngestWf):
                if (lstIngestSheets is None) or (sSheet in lstIngestSheets):
                    for sCol in self.__map.ingest_getExcelCols(sSheet, sIngestWf):
                        bObl = self.__map.ingest_isObl(sSheet, sCol)
                        bPk = self.__map.ingest_isPk(sSheet, sCol)
                        bFk = self.__map.ingest_isFk(sSheet, sCol)
                        bLup = self.__map.ingest_isLookup(sSheet, sCol)
                        lstAttrib = []
                        if bPk:
                            lstAttrib.append("PK")
                        if bFk:
                            lstAttrib.append("FK")
                        if bObl:
                            lstAttrib.append("OBL")
                        if bLup:
                            lstAttrib.append("LUP")
                        if len(lstAttrib) > 0:
                            sAttrib = PrettyText.lst2strC(lstAttrib)
                        else:
                            sAttrib = ""
                        resComment = self.__model.conf_getDbSchema().getDbSchemaInfo("ingest", sSheet, sCol)
                        if resComment.isOk():
                            sComName = resComment.getData("comment-part-name")
                            sComUnit = resComment.getData("comment-part-unit")
                            sComComment = resComment.getData("comment-part-comment")
                        else:
                            sComName = sComUnit = sComComment = None
                        cellGlossary = wsGlossary.cell(column=1, row=iRowGlossary, value=sSheet)
                        cellGlossary = wsGlossary.cell(column=2, row=iRowGlossary, value=sCol)
                        cellGlossary = wsGlossary.cell(column=3, row=iRowGlossary, value=sAttrib)
                        if sComName is not None:
                            cellGlossary = wsGlossary.cell(column=4, row=iRowGlossary, value=sComName)
                        if sComUnit is not None:
                            cellGlossary = wsGlossary.cell(column=5, row=iRowGlossary, value=sComUnit)
                        if sComComment is not None:
                            cellGlossary = wsGlossary.cell(column=6, row=iRowGlossary, value=sComComment)
                        iRowGlossary += 1
        # Lookup
        if bLookup:
            cellGlossary = wsGlossary.cell(column=1, row=iRowGlossary, value="LOOKUP")
            cellGlossary.font = openpyxl.styles.Font(bold=True)
            iRowGlossary += 1
            sCommentAttributes = "KEY: lookup key name (bold)\n<*>: needs other lookup (italic)\nOBL: obligate (underscore)"
            for iIdx, sName in enumerate(
                    ("Sheet", "Column", "Attributes", "Name", "Unit", "Comment")):
                cellGlossary = wsGlossary.cell(column=iIdx + 1, row=iRowGlossary, value=sName)
                cellGlossary.font = openpyxl.styles.Font(bold=True)
                # cellGlossary.border = borderTop
                if sName == "Attributes":
                    cellGlossary.comment = openpyxl.comments.Comment(text=sCommentAttributes, author=None)
            iRowGlossary += 1
            for sSheet in self.__map.lookup_getExcelSheets():
                if (lstLookupSheets is None) or (sSheet in lstLookupSheets):
                    for sCol in self.__map.lookup_getExcelCols(sSheet):
                        sTable, sAttrib = self.__map.lookup_excel2db(sSheet, sCol)
                        bKey = self.__map.lookup_isKey(sSheet, sCol)
                        bObl = self.__map.lookup_isObl(sSheet, sCol)
                        bLup2 = self.__map.lookup_moreLookupNeeded_isAttributeNeeding(sTable, sAttrib)
                        lstAttrib = []
                        if bKey:
                            lstAttrib.append("KEY")
                        if bLup2:
                            sLup2_table = self.__map.lookup_moreLookupNeeded_getNeededLookup(sTable, sAttrib)
                            sLup2_sheet = self.__map.lookup_table2sheet(sLup2_table)
                            lstAttrib.append("<" + str(sLup2_sheet) + ">")
                        if bObl:
                            lstAttrib.append("OBL")
                        if len(lstAttrib) > 0:
                            sAttrib = PrettyText.lst2strC(lstAttrib)
                        else:
                            sAttrib = ""
                        resComment = self.__model.conf_getDbSchema().getDbSchemaInfo("lookup", sSheet, sCol)
                        if resComment.isOk():
                            sComName = resComment.getData("comment-part-name")
                            sComUnit = resComment.getData("comment-part-unit")
                            sComComment = resComment.getData("comment-part-comment")
                        cellGlossary = wsGlossary.cell(column=1, row=iRowGlossary, value=sSheet)
                        cellGlossary = wsGlossary.cell(column=2, row=iRowGlossary, value=sCol)
                        cellGlossary = wsGlossary.cell(column=3, row=iRowGlossary, value=sAttrib)
                        if sComName is not None:
                            cellGlossary = wsGlossary.cell(column=4, row=iRowGlossary, value=sComName)
                        if sComUnit is not None:
                            cellGlossary = wsGlossary.cell(column=5, row=iRowGlossary, value=sComUnit)
                        if sComComment is not None:
                            cellGlossary = wsGlossary.cell(column=6, row=iRowGlossary, value=sComComment)
                        iRowGlossary += 1
        # Taxon
        if bTaxon:
            cellGlossary = wsGlossary.cell(column=1, row=iRowGlossary, value="TAXON")
            cellGlossary.font = openpyxl.styles.Font(bold=True)
            iRowGlossary += 1
            sCommentAttributes = "OBL: obligate (underscore)"
            for iIdx, sName in enumerate(
                    ("Sheet", "Column", "Attributes", "Name", "Unit", "Comment")):
                cellGlossary = wsGlossary.cell(column=iIdx + 1, row=iRowGlossary, value=sName)
                cellGlossary.font = openpyxl.styles.Font(bold=True)
                if sName == "Attributes":
                    cellGlossary.comment = openpyxl.comments.Comment(text=sCommentAttributes, author=None)
            iRowGlossary += 1
            for sSheet in self.__map.taxon_getExcelSheets():
                if (lstTaxonSheets is None) or (sSheet in lstTaxonSheets):
                    for sCol in self.__map.taxon_getExcelCols(sSheet):
                        bObl = self.__map.taxon_isObl(sSheet, sCol)
                        lstAttrib = []
                        if bObl:
                            lstAttrib.append("OBL")
                        if len(lstAttrib) > 0:
                            sAttrib = PrettyText.lst2strC(lstAttrib)
                        else:
                            sAttrib = ""
                        resComment = self.__model.conf_getDbSchema().getDbSchemaInfo("taxon", sSheet, sCol)
                        if resComment.isOk():
                            sComName = resComment.getData("comment-part-name")
                            sComUnit = resComment.getData("comment-part-unit")
                            sComComment = resComment.getData("comment-part-comment")
                        cellGlossary = wsGlossary.cell(column=1, row=iRowGlossary, value=sSheet)
                        cellGlossary = wsGlossary.cell(column=2, row=iRowGlossary, value=sCol)
                        cellGlossary = wsGlossary.cell(column=3, row=iRowGlossary, value=sAttrib)
                        if sComName is not None:
                            cellGlossary = wsGlossary.cell(column=4, row=iRowGlossary, value=sComName)
                        if sComUnit is not None:
                            cellGlossary = wsGlossary.cell(column=5, row=iRowGlossary, value=sComUnit)
                        if sComComment is not None:
                            cellGlossary = wsGlossary.cell(column=6, row=iRowGlossary, value=sComComment)
                        iRowGlossary += 1

        # Statistik
        res.addData("info", odicInfo)
        res.addData("template", bTemplate)
        res.addData("ingest", bIngest)
        res.addData("lookup", bLookup)
        res.addData("taxon", bTaxon)
        # Save
        progress("pulse")
        if sFile is None:
            res.addData("wb", wb)
        else:
            try:
                wb.save(filename=sFile)
            except Exception as ex:
                res.addErr("Error during saving Excel file: " + str(ex))
        return res
