# coding=utf-8

import openpyxl
from collections import Counter, OrderedDict
import os, sys, csv, getpass, platform, datetime, time, copy, re
import openpyxl.styles.numbers
import warnings
from tools.coding import Result, WorkingObject, DataTools
from tools.fileio import FileTools
from tools.excel import ExcelTools, ExcelNavigator
from tools.text import PrettyText
from tools.parallel import Stoppable
from tools.printing import SmartPrinter
from tools.misc import JsonConverter
from tools.geocruncher import GeoCruncher
from pathlib import Path

# -------------------------------------------------------------
# Importer: Dirigiert Import (Lookup, Ingestdaten, TaxonWizard-Daten, Converter, ExcelTamer)
# -------------------------------------------------------------

class Importer(Stoppable, WorkingObject):

    # Konstruktor
    def __init__(self, model):
        Stoppable.__init__(self)
        WorkingObject.__init__(self)

        self.__model = model
        self.__smartprinter = model.ui_getSmartprinter()
        self.__map = model.subsystem_getMapper()

        # Abort Childs
        self.addAbortChild(self.__model.subsystem_getTamer())
        self.addAbortChild(self.__model.subsystem_getValidator())
        self.addAbortChild(self.__model.subsystem_getLookup())
        self.addAbortChild(self.__model.subsystem_getIngestor())
        self.addAbortChild(self.__model.subsystem_getTaxonWizard())

    # Info
    def info(self):
        sOut = "*** Status of Importer System\n\n"
        lstConverters = self.__map.workflow_converter_import_getAll(long=True)
        sOut += "* Registered converters (" + str(len(lstConverters)) + "x)\n"
        for iIdx, sName in enumerate(lstConverters):
            sOut += sName + "\n"
        return sOut

    def startImport(self, sFile, **kwargs):
        """
        Args
            sFile (str): Excel File
        Kwargs:
            loader_tamer (bool): Excel Tamer währen Loading
            ingest (bool): Ingest
            lookup (bool): Lookup
            taxon (bool): Taxon
            ingest_sheets (lst): Liste aller hinzuzufügenden Sheets
            lookup_sheets (lst): Liste aller hinzuzufügenden Sheets
            taxon_sheets (lst): Liste aller hinzuzufügenden Sheets
            ingest_wf (str)
            ingest_lup_search_single (bool): Teste Lookup -> Befrage einzeln (oder am Anfang alles herunterladen)
            lookup_import_excel (lst): Zu importierende Einträge aus Excel: all/gold/not-gold
            lookup_import_mode (lst): Import modus: new/update/both
            lookup_inport_db (lst): Nur folgende Einträge in DB updaten: all/gold/not-gold
            lookup_import_goldrush (bool): Es werden nur noch gold Lookups beim Import akzeptiert
            lookup_reuse (str): Reaktion wenn bestehende Lookups differieren: erroe/warning/info
            extend (bool): Extension allowed
            extend_react (str): Extension reaction if not matching 100%
            taxon_update (bool): Sollen Tabellen in Taxon System geupdated werden dürfen
            batch (str): Batch Präfix für alle Submodule
        """

        pn = self.__smartprinter.normal
        pv = self.__smartprinter.verbose
        progress = self.__smartprinter.progress
        workflow = self.__smartprinter.workflow

        res = Result(name="Importer")

        # Parameter
        bTamer = True if "tamer" in kwargs and kwargs["tamer"] is True else False
        bIngest = True if "ingest" in kwargs and kwargs["ingest"] is True else False
        bLookup = True if "lookup" in kwargs and kwargs["lookup"] is True else False
        bTaxon = True if "taxon" in kwargs and kwargs["taxon"] is True else False
        lstIngestSheets = kwargs["ingest_sheets"] if "ingest_sheets" in kwargs else None
        lstLookupSheets = kwargs["lookup_sheets"] if "lookup_sheets" in kwargs else None
        lstTaxonSheets = kwargs["taxon_sheets"] if "taxon_sheets" in kwargs else None
        sIngestWf = kwargs["ingest_wf"] if "ingest_wf" in kwargs else None
        bIngestLookupSearchSingle = kwargs["ingest_lup_search_single"] if "ingest_lup_search_single" in kwargs else False
        bExtend = kwargs["extend"] if "extend" in kwargs and kwargs["extend"] in [True,False] else None
        sExtendReact = kwargs["extend_react"] if "extend_react" in kwargs else None
        sLookupImportExcel = kwargs["lookup_import_excel"] if "lookup_import_excel" in kwargs else None
        sLookupImportMode = kwargs["lookup_import_mode"] if "lookup_import_mode" in kwargs else None
        bLookupImportGoldrush = kwargs["lookup_import_goldrush"] if ("lookup_import_goldrush" in kwargs and kwargs["lookup_import_goldrush"] in [True,False]) else None
        sLookupReuse = kwargs["lookup_reuse"] if "lookup_reuse" in kwargs else None
        sLookupImportDb = kwargs["lookup_import_db"] if "lookup_import_db" in kwargs else None
        bTaxonUpdate = True if "taxon_update" in kwargs and kwargs["taxon_update"] is True else False

        # Parameter Test
        if not bIngest and not bLookup and not bTaxon:
            res.addErr("Nothing to do ...", sys="Importer", kat="Configuration")
        if sFile is None:
            res.addErr("No Excel file given", sys="Importer", kat="I/O")

        # None -> Default (Ingest Extend / Lookup Import)
        if sLookupImportExcel is None:
            sLookupImportExcel = self.__model.conf_getLookupImport_excel()
        if bLookupImportGoldrush is None:
            bLookupImportGoldrush = self.__model.conf_getLookupImport_excel_goldrush()
        if sLookupImportMode is None:
            sLookupImportMode = self.__model.conf_getLookupImport_db()
        if sLookupImportDb is None:
            sLookupImportDb = self.__model.conf_getLookupImport_db_update()
        if sLookupReuse is None:
            sLookupReuse = self.__model.conf_getLookupReuseDiffersReactionlevel()
        if sExtendReact is None:
            if bExtend is True:
                res.addErr("No Reaction level given for not 100% matching updates", sys="Importer", kat="Configuration")
            else:
                sExtendReact = self.__model.conf_getRecordExtensionProblemReactionlevel()
        if bExtend is None:
            bExtend = self.__model.conf_isRecordExtensionAllowed()

        # Daten
        iIngestId = None # Ingest: IngestLog Id des erzeugten Ingests#
        iIngestRecsNew = 0

        # Log
        iZeitStart = time.time()
        sLog = ""
        sLogHeader = ""
        sLogDetails_conf = ""
        sLogSummary = ""
        sLogSummary_prae = ""
        sLogSummary_errorWarnInfo = ""
        sLogSummary_subModules = ""
        sLogDetails_errorWarnInfo = ""
        sLogDetails_SubModule = ""
        sLogCaution_prae = ""

        # Progress
        progress("on-pulse")

        # ---

        # Db
        if res.isOk():
            resConn = self.__model.subsystem_getDb().conn_getConnection()
            if resConn.isErr():
                res.addResult(resConn)
            conn = resConn.getData("connection")
            cursor = resConn.getData("cursor")

        # Workbook
        if res.isOk():
            if bTamer or (bIngest and self.__map.workflow_converter_import_isExisting(sIngestWf)):
                bRw = True
            else:
                bRw = False
            workflow(top="Importer",sub="Loading Excel")
            pn("* Loading Excel Workbook")
            resWb = ExcelTools.getWb(sFile, rw=bRw)
            if resWb.isErr():
                res.addResult(resWb, err=True)
            wb = resWb.getData("wb")

        # Converter
        if res.isOk() and bIngest and self.__map.workflow_converter_import_isExisting(sIngestWf):
            resCvt = self.__model.subsystem_getConverter().start_converter_import(wb, result_in=res, ingest_wf=sIngestWf) # wb wird als Referenz direkt veraendert
            res.addResult(resCvt, msg_only=True)
            if resCvt.isOk():
                sLogDetails_SubModule += resCvt.getData("log")
            sIngestWf = self.__model.subsystem_getMapper().workflow_converter_import_getTarget(sIngestWf) # Workflow verändern auf Zielworkflow
        # Tamer
        if res.isOk() and bTamer:
            resTame = self.__model.subsystem_getTamer().startTaming(wb, result_in=res, ingest=bIngest, lookup=bLookup, taxon=bTaxon, ingest_wf=sIngestWf)
            res.addResult(resTame, msg_only=True)
            if resTame.isOk():
                sLogDetails_SubModule += resTame.getData("log")

        # Validator
        if res.isOk():
            resValidate= self.__model.subsystem_getValidator().startValidate(wb,
                result_in=res,
                ingest=bIngest,
                lookup=bLookup,
                taxon=bTaxon,
                ingest_sheets=lstIngestSheets,
                lookup_sheets=lstLookupSheets,
                taxon_sheets=lstTaxonSheets,
                ingest_wf=sIngestWf)
            res.addResult(resValidate, msg_only=True)
            if resValidate.isOk():
                sLogDetails_SubModule += resValidate.getData("log")

        # Lookup
        if res.isOk() and bLookup:
            resLup = self.__model.subsystem_getLookup().addLookupBatch(wb,
                    result_in=res,
                    cursor=cursor,
                    sheets=lstLookupSheets,
                    import_excel=sLookupImportExcel,
                    import_mode=sLookupImportMode,
                    import_db=sLookupImportDb,
                    import_excel_goldrush=bLookupImportGoldrush,
                    reuse=sLookupReuse)
            res.addResult(resLup, msg_only=True)
            if resLup.isOk():
                res.addResultData(resLup,
                    ["lookup-tables-changed",
                     "lookup-excel-recs",
                     "lookup-import-no",
                     "lookup-new",
                     "lookup-new-no",
                     "lookup-update",
                     "lookup-update-no",
                     "lookup-err"])
                sLogDetails_SubModule += resLup.getData("log")
                sLogSummary_subModules += resLup.getData("log_stat")

        # Ingest
        if res.isOk() and bIngest:
            resIngNew = self.__model.subsystem_getIngestor().startIngest(wb, sIngestWf, sFile, result_in=res, cursor=cursor, sheets=lstIngestSheets, lup_search_single=bIngestLookupSearchSingle, extend=bExtend, extend_react=sExtendReact)
            res.addResult(resIngNew, msg_only=True)
            if resIngNew.isOk():
                res.addResultData(resIngNew,
                                  ["ingest-biota-new",
                                   "ingest-counter-new",
                                   "ingest-counter-update",
                                   "ingest-ingest-id",
                                   "ingest-ingest-depend",
                                   "ingest-cruise-extend",
                                   "ingest-station-extend",
                                   "ingest-sample-extend"])
                iIngestId = resIngNew.getData("ingest-ingest-id")
                cntNewRecs = resIngNew.getData("ingest-counter-new")
                iIngestRecsNew = sum(cntNewRecs.values())
                sLogDetails_SubModule += resIngNew.getData("log")
                sLogSummary_subModules += resIngNew.getData("log_stat")

        # TaxonWizard
        if res.isOk() and bTaxon:
            resTax = self.__model.subsystem_getTaxonWizard().addAdmTable_excel(wb, result_in=res, cursor=cursor, sheets=lstTaxonSheets, update=bTaxonUpdate)
            res.addResult(resTax, msg_only=True)
            if resTax.isOk():
                res.addResultData(resTax,
                                  ["taxon-tables-changed",
                                   "taxon-excel-sheets-used",
                                   "taxon-excel-recs",
                                   "taxon-import-no",
                                   "taxon-new",
                                   "taxon-equal",
                                   "taxon-update",
                                   "taxon-update-no",
                                   "taxon-err"])
                sLogDetails_SubModule += resTax.getData("log")

        # *** Log zusammenbauen
        # Log Prae
        iSekTotal = time.time() - iZeitStart
        iStd = int(iSekTotal // 3600)
        iMin = int((iSekTotal % 3600) // 60)
        iSek = int(iSekTotal % 60)
        sLogHeader += "************************************\n"
        sLogHeader += "Import Data\n"
        sLogHeader += "Duration: " + str(iStd) + "h" + str(iMin) + "m" + str(iSek) + "s\n"
        sLogHeader += "************************************\n\n"
        sLogHeader += "File: " + str(FileTools.getFileWithoutPath(sFile)) + "\n\n"

        # Log Conf
        sLogDetails_conf += "************************************\n"
        sLogDetails_conf += "Configuration\n"
        sLogDetails_conf += "************************************\n\n"
        sLogDetails_conf += "File: " + str(FileTools.getFileWithoutPath(sFile)) + "\n"
        if bIngest:
            if self.__map.workflow_isExisting(sIngestWf):
                sLogDetails_conf += "* Import Ingest Data\n"
                sLogDetails_conf += "Workflow: " + str(self.__map.workflow_short2long(sIngestWf)) + "\n"
                sLogDetails_conf += "Sheets: " + ("all" if lstIngestSheets is None else PrettyText.lst2strC(lstIngestSheets)) + "\n"
                sLogDetails_conf += "Extension of records in Db allowed: " + ("yes" if bExtend else "no") + "\n"
                sLogDetails_conf += "Reaction if differences while extension occurr: " + str(sExtendReact) + "\n"
            elif self.__map.workflow_converter_import_isExisting(sIngestWf):
                sLogDetails_conf += "* Import Ingest Data via Converter\n"
                sLogDetails_conf += "Foreign format: " + str(self.__map.workflow_converter_import_short2long(sIngestWf)) + "\n"
                sLogDetails_conf += "Target CB workflow: " + str(
                    self.__map.workflow_short2long(self.__map.workflow_converter_import_getTarget(sIngestWf))) + "\n"
                sLogDetails_conf += "Sheets: " + (
                    "all" if lstIngestSheets is None else PrettyText.lst2strC(lstIngestSheets)) + "\n"
            else:
                sLogDetails_conf += "* Ingest (Uuups...)\n"
        if bLookup:
            sLogDetails_conf += "* Import Lookup Data\n"
            sLogDetails_conf += "Search mode: " + (
                "load every single lookup on-the-fly" if bIngestLookupSearchSingle else "load all lookups at start") + "\n"
            sLogDetails_conf += "Sheets: " + ("all" if lstLookupSheets is None else PrettyText.lst2strC(lstLookupSheets)) + "\n"
            sLogDetails_conf += "GoldRush!\n" if (bLookupImportGoldrush and sLookupImportExcel == "all") else ""
            sLogDetails_conf += "Gold (import from Excel): " + str(sLookupImportExcel) + "\n"
            sLogDetails_conf += "Gold (write mode in Db): " + str(sLookupImportMode) + "\n"
            if sLookupImportDb is not None:
                sLogDetails_conf += "Gold (update mode in Db): " + str(sLookupImportDb) + "\n"
        if bTaxon:
            sLogDetails_conf += "* Import TaxonWizard Data\n"
            sLogDetails_conf += "Update records in admin tables: " + ("yes" if bTaxonUpdate else "no") + "\n"
            sLogDetails_conf += "Sheets: " + ("all" if lstTaxonSheets is None else PrettyText.lst2strC(lstTaxonSheets)) + "\n"
        if bTamer:
            sLogDetails_conf += "* Special Features\n"
            sLogDetails_conf += "ExcelTamer: " + ("active" if bTamer else "deactivated") + "\n"
        sLogDetails_conf +="\n"
        # Log: Err/Warn/Info
        if res.isAnyMessage():
            sLogDetails_errorWarnInfo += res.getErrWarnInfoStringHeader() + res.getErrWarnInfoString() + res.getErrWarnInfoStringFooter()
        # Log: Details
        sLogDetailsHead = ""
        sLogDetailsHead += "************************************\n"
        sLogDetailsHead += "DETAILS ****************************\n"
        sLogDetailsHead += "************************************\n\n"

        # Log: Stat Prae
        sLogSummary_prae += "************************************\n"
        sLogSummary_prae += "SUMMARY ****************************\n"
        sLogSummary_prae += "************************************\n\n"

        # Log: ErrWarnInfo Prae
        sLogCaution = ""
        sLogCaution_prae += "************************************\n"
        sLogCaution_prae += "CAUTION ****************************\n"
        sLogCaution_prae += "************************************\n\n"

        sLogSummary_prae += "Please check\ndetail section below\nfor further information ...\n\n"
        # Log: Stat Error/Warn/Info
        if res.isAnyMessage():
            sLogSummary_errorWarnInfo += ">>>>> CAUTION <<<<<\n"
            if res.isErr():
                sLogSummary_errorWarnInfo += "Errors: "+str(res.getErrNo())+"\n"
            if res.isWarn():
                sLogSummary_errorWarnInfo += "Warnings: "+str(res.getWarnNo())+"\n"
            if res.isInfo():
                sLogSummary_errorWarnInfo += "Information: "+str(res.getInfoNo())+"\n"
            sLogSummary_errorWarnInfo += ">>>>>>>>>><<<<<<<<<\n"
            sLogSummary_errorWarnInfo += "\n"

        # Zusammenbauen
        if sLogSummary_subModules != "" or sLogSummary_errorWarnInfo != "":
            sLogSummary = sLogSummary_prae + sLogSummary_errorWarnInfo + sLogSummary_subModules
        if res.isAnyMessage():
            sLogCaution = sLogCaution_prae + sLogDetails_errorWarnInfo
        sLogDetails = sLogDetailsHead + sLogDetails_conf + sLogDetails_SubModule
        sLog = sLogHeader + sLogSummary + sLogCaution + sLogDetails
        sLog = PrettyText.clean_wizard(sLog, max_one_empty_row=True) # Cleanup
        # Antwort
        res.addData("log", sLog)

        # Zusatz bei Ingest: Update Ingest Log und Action Log
        if res.isOk() and bIngest:
            # Eintrag in Action Log
            sActionLookup = "ing-add"
            resultActionlog = self.__model.subsystem_getDb().actionlog_new(sActionLookup, FileTools.getFileWithoutPath(sFile), sLog, cursor=cursor)
            if resultActionlog.isErr():
                res.addResult(resultActionlog, err=True, sys="Importer", kat="Writing to Action Log")
            # Update Ingest Log
            iWarn = res.getWarnNo()
            iInfo = res.getInfoNo()
            resUpdateIngestLog = self.__model.subsystem_getDb().inglog_update(iIngestId, cursor=cursor, rows=iIngestRecsNew, warn=iWarn, info=iInfo, log=sLog, wf=sIngestWf, file=FileTools.getFileWithoutPath(sFile))
            if resUpdateIngestLog.isAnyMessage():
                res.addResult(resUpdateIngestLog)

        # Db
        try:
            if self.__model.conf_isDry():
                conn.rollback()
            elif res.isErr():
                conn.rollback()
            else:
                conn.commit()
            cursor.close()
            conn.close()
        except:
            pass
        return res