# coding=utf-8

import openpyxl
from collections import Counter, OrderedDict
import os, sys, csv, getpass, platform, datetime, time, copy, re, traceback
import openpyxl.styles.numbers
import warnings
from tools.coding import Result, WorkingObject
from tools.text import PrettyText
from tools.parallel import Stoppable
from tools.printing import SmartPrinter
from tools.misc import JsonConverter
from tools.geocruncher import GeoCruncher

# -------------------------------------
# Mapper: Mapping-Wissen
# -------------------------------------

"""
KURZINFO

*** INGEST
PK/FK
- Excel-Welt: Verbindung von Rows verschiedener Sheets durch Identifier-Kombi: PK/FK
- DB-Welt: Verbindung von Datebnsätzen verschiedenerrt Tabellen: Einzel-Attribute PK(id)/FK(main)
- Kurzinfo: In FK steht immer der PK des zu verbindenen Gegenstücks
- Excel: FK-main ist ein FK-Feld mit Datenbank Ziel (muss es genau einmal für ein Sheet geben)
- Excel: PK fakultativ für wf-Sheet (braucht man nur, wenn etwas per FK drangehängt werden soll)
- Excel: FK obligat für wf-Sheet (weil sonst entkoppelt von datenbaum in db - ausser bei cruise)
- Excel: PK/FK-System gilt für maximalen Workflow

*** LOOKUP
...

*** TAXON
...

Todo: 
Längenermittlung: PK-FK max 
Test: ob Part in Sub-Wf
Test: FK ohne FK-Aim (PK)
Part: bei PK -> PK weg
Part: bei FK->Fehler
Info: Wo sind PK entfernt wegen Part
"""

class Mapper(Stoppable, WorkingObject):

    # Konstruktor
    def __init__(self, model):
        Stoppable.__init__(self)
        WorkingObject.__init__(self)
        self.__model = model
        self.__smartprinter = model.ui_getSmartprinter()

        # Allgemeiner Informationsspeicher für die Info-Seite
        self.__lstInfosEtc = []

        # Konfiguration der ActionLog Tables und ingest Tabellen
        self.__odicAdminTables = OrderedDict([
            ("ingest", ["name", "created_on", "workflow", "log", "rows", "warn", "info", "file"]),
            ("actionlog", ["username", "action", "object", "date", "host", "os", "log", "ip"])
        ])

        # Workflows
        # Nativ (Import, Export)
        self.__odicMapping__Wf = OrderedDict([
            ("arc", "PANABIO"),
            ("ant", "WEECOS"),
            ("nds", "BENOSIS"),
            ("cbmp", "CBMP"),
            ("beast", "BEAST"),
            ("dsr", "DSR"),
            ("loki", "LOKI"),
            ("pn", "PlanktonNet"),
            ("mar", "MARLIN"),
        ])
        # Converter (Import)
        self.__odicMapping__Wf_converter_import = OrderedDict([
            ("ant_mat", ("WEECOS (Matrix)", "ant")),
            ("mar_trans", ("MARLIN (Transport)", "mar")),
        ])
        # Converter (Export)
        self.__odicMapping__Wf_converter_export = OrderedDict([
            ("mar_trans", ("MARLIN (Transport)"))
        ])

        # Ingest Parameter
        self.__lstMapping__Params = ["pk", "fk", "obl", "lat", "lon", "date", "time"]

        # FK/PK Paarungen
        self.__lstMapping__Ingest_FkPk = [("IndividualMeasure", "Biota"),
                  ("Sediment", "Subset"),
                  ("Biota", "Subset"),
                  ("Subset", "SubSample"),
                  ("SubSample", "Sample"),
                  ("Sample", "Station"),
                  ("Station", "Cruise")
                  ]

        # *** All
        # Alias: Sheet Namen
        self.__dicMapping__All_ExcelSheet2AliasSheet = {
            "ingest": {},
            "lookup": {},
            "taxon": {}
        }
        self.__odicMapping__All_ExcelSheetAlias2Sheet = OrderedDict()

        # Alias Column Namen
        self.__dicMapping__All_ExcelSheetCol2AliasCol = {
            "ingest": {},
            "lookup": {},
            "taxon": {}
        }
        self.__odicMapping__All_ExcelSheetCoAlias2col = OrderedDict()

        # *** Ingest
        self.__odicMapping__Ingest_ExcelSheet2col = OrderedDict()
        self.__odicMapping__Ingest_ExcelSheetCol2dbTableAttrib = OrderedDict()
        self.__odicMapping__Ingest_ExcelSheetCol2link = OrderedDict()
        self.__odicMapping__Ingest_ExcelSheetCol2wf = OrderedDict()
        self.__odicMapping__Ingest_ExcelSheetCol2dbLup = OrderedDict()
        self.__odicMapping__Ingest_ExcelSheetCol2geo = OrderedDict()
        self.__odicMapping__Ingest_ExcelSheetCol2obl = OrderedDict()
        self.__odicMapping__Ingest_ExcelSheet2date = OrderedDict()
        self.__odicMapping__Ingest_ExcelSheet2time = OrderedDict()
        self.__odicMapping__Ingest_ExcelSheet2pk = OrderedDict()
        self.__odicMapping__Ingest_ExcelSheet2fk = OrderedDict()
        self.__odicMapping__Ingest_ExcelSheet2fkMain = OrderedDict()

        self.__odicMapping__Ingest_DbTable2Attrib = OrderedDict()
        self.__odicMapping__Ingest_DbTableAttrib2excelSheetCol = OrderedDict()

        self.__lstMapping__Ingest_ExcelSheet2dbTable = []

        # Mapping von (Sheet,Column), deren Kommentare in abweichenden Tabellen gesucht werden müssen
        self.__dicMapping__Ingest_ForgeComments = {
            ("Station", "Cruise"): ("cruise", "name"),
            ("Sample", "Cruise"): ("cruise", "name"),
            ("Sample", "Station"): ("station", "name"),
            ("SubSample", "Cruise"): ("cruise", "name"),
            ("SubSample", "Station"): ("station", "name"),
            ("SubSample", "Sample"): ("sample", "name"),
            ("Subset", "Cruise"): ("cruise", "name"),
            ("Subset", "Station"): ("station", "name"),
            ("Subset", "Sample"): ("sample", "name"),
            ("Subset", "SubSample"): ("subsample", "name"),
            ("Biota", "Cruise"): ("cruise", "name"),
            ("Biota", "Station"): ("cruise", "name"),
            ("Biota", "Sample"): ("station", "name"),
            ("Biota", "SubSample"): ("subsample", "name"),
            ("Biota", "Subset"): ("subset", "name"),
            ("Sediment", "Cruise"): ("cruise", "name"),
            ("Sediment", "Station"): ("station", "name"),
            ("Sediment", "Sample"): ("sample", "name"),
            ("Sediment", "SubSample"): ("subsample", "name"),
            ("Sediment", "Subset"): ("subset", "name"),
            ("IndividualMeasure", "Cruise"): ("cruise", "name"),
            ("IndividualMeasure", "Station"): ("station", "name"),
            ("IndividualMeasure", "Sample"): ("sample", "name"),
            ("IndividualMeasure", "SubSample"): ("subsample", "name"),
            ("IndividualMeasure", "Subset"): ("subset", "name")
        }

        # *** Lookup
        self.__odicMapping__Lookup_excel2db = OrderedDict()
        self.__odicMapping__Lookup_db2excel = OrderedDict()
        self.__odicMapping__Lookup_table2key = OrderedDict()
        self.__odicMapping__Lookup_tableAtr2obl = OrderedDict()

        # Map key/val der Lookups die nicht aus der Mapping Tabelle zu extrahieren sind
        self.__dicMapLupKeyValSeparate = {
            "country": "name"
        }

        # *** Taxon
        self.__odicMapping__Taxon_Table2sheet = OrderedDict()
        self.__odicMapping__Taxon_Sheet2table = OrderedDict()
        self.__odicMapping_Taxon_db2obl = OrderedDict()
        self.____odicMapping__Taxon_table2key = OrderedDict()

        # --------------------------------------------------------------------------------------------------------------

        # Mapping erstellen
        res = self.__createMapping()
        if res.isErr():
            self.__smartprinter.normal(res.getErrWarnInfoString())
            self.setWorking(False)

    # ------------------------------------------------------------------------------------------------------------

    # ToString
    def __str__(self):
        return self.info()

    # Info Init
    def info(self):
        if not self.isReady():
            return None
        sOut = "*** Status of Mapping System\n\n"

        sOut += "************************************\n"
        sOut += "Workflows" + "\n"
        sOut += "************************************\n\n"
        sOut += "--- Native (import/export) ("+str(len(self.workflow_getAll()))+"x)\n"
        for sWf in self.workflow_getAll():
            sWf_long = self.workflow_short2long(sWf)
            sOut += sWf_long + " (" + sWf+")\n"
        sOut += "\n"
        sOut += "--- Converter\n"
        sOut += "* Import (" + str(len(self.workflow_converter_import_getAll())) + "x)\n"
        for sWfconv in self.workflow_converter_import_getAll():
            sWfconv_long = self.workflow_converter_import_short2long(sWfconv)
            sWfconv_target = self.workflow_converter_import_getTarget(sWfconv)
            sWfconv_target_long = self.workflow_short2long(sWfconv_target)
            sOut += sWfconv_long + " -> " + sWfconv_target_long+" (" + sWfconv + "->" + sWfconv_target+")\n"
        sOut += "\n"
        sOut += "* Export (" + str(len(self.workflow_converter_export_getAll())) + "x)\n"
        for sWfconv in self.workflow_converter_export_getAll():
            sWfconv_long = self.workflow_converter_export_short2long(sWfconv)
            sOut += sWfconv_long + " (" + sWfconv + ")\n"
        sOut += "\n"

        sOut += "************************************\n"
        sOut += "Automatic system detection" + "\n"
        sOut += "************************************\n\n"
        lstSheetsToDetect = []
        lstSheetsToDetect += self.ingest_getExcelSheets()
        lstSheetsToDetect += self.lookup_getExcelSheets()
        lstSheetsToDetect += self.taxon_getExcelSheets()
        odicSheetSysDetect = OrderedDict()
        for sSheet in lstSheetsToDetect:
            sSystem = self.all_getSheetSystem(sSheet)
            if sSystem not in odicSheetSysDetect:
                odicSheetSysDetect[sSystem] = [sSheet]
            else:
                odicSheetSysDetect[sSystem].append(sSheet)
        for sSystem, lstSheets in odicSheetSysDetect.items():
            sOut += sSystem+": " + PrettyText.lst2strC(lstSheets)+"\n"
        sOut += "\n"
        sOut += "************************************\n"
        sOut += "Excel Tamer\n"
        sOut += "************************************\n\n"
        sOut += "--- Sheets\n"
        odicTmp = OrderedDict()
        lstSheets = self.all_getExcelSheets()
        for sSheet in lstSheets:
            sSystem = self.all_getSheetSystem(sSheet)
            lstAltSheet = self.all_getAliasesSheet(sSheet)
            if lstAltSheet is not None:
                if sSystem not in odicTmp:
                    odicTmp[sSystem] = OrderedDict()
                if sSheet not in odicTmp[sSystem]:
                    odicTmp[sSystem][sSheet] = []
                odicTmp[sSystem][sSheet] += lstAltSheet
        for sSystem, odicSheetRewrites  in odicTmp.items():
            sOut += "* "+sSystem.capitalize() +"\n"
            for sSheet, lstRewrites in odicSheetRewrites.items():
                sOut += str(sSheet) + ": " + PrettyText.lst2strC(lstRewrites) + "\n"
            sOut += "\n"
        sOut += "--- Columns\n"
        odicTmp = OrderedDict()
        lstSheets = self.all_getExcelSheets()
        for sSheet in lstSheets:
            sSystem = self.all_getSheetSystem(sSheet)
            for sCol in self.all_getExcelCols(sSheet):
                lstAltCol = self.all_getAliasesCol(sSheet, sCol)
                if lstAltCol is not None:
                    for sColAlt in lstAltCol:
                        if sSystem not in odicTmp:
                            odicTmp[sSystem] = OrderedDict()
                        if sSheet not in odicTmp[sSystem]:
                            odicTmp[sSystem][sSheet] = []
                        odicTmp[sSystem][sSheet].append(str(sColAlt) + "->" + str(sCol))
        for sSystem, odicSheetRewrites in odicTmp.items():
            sOut += "* " + sSystem.capitalize() + "\n"
            for sSheet, lstRewrites in odicSheetRewrites.items():
                sOut += str(sSheet) + ": " + PrettyText.lst2strC(lstRewrites) + "\n"
            sOut += "\n"
        sOut += "************************************\n"
        sOut += "Admin System" + "\n"
        sOut += "************************************\n\n"
        sOut += "--- Db tables" + "\n"
        sOut += PrettyText.lst2strC(self.admin_getDbTables()) + "\n"
        sOut += "\n"
        sOut += "************************************\n"
        sOut += "Lookup system" + "\n"
        sOut += "************************************\n\n"

        sOut += "--- Excel sheets (->table)" + "\n"
        sOut += PrettyText.lst2strC([str(sSheet) + " (" + str(self.lookup_sheet2table(sSheet))+")" for sSheet in  self.lookup_getExcelSheets()]) + "\n"
        sOut += "\n"
        sOut += "--- Db tables (->sheet)" + "\n"
        sOut += PrettyText.lst2strC([str(sTable) + " (" + str(self.lookup_table2sheet(sTable)) + ")" for sTable in self.lookup_getDbTables()]) + "\n"
        sOut += "\n"
        sOut += "--- Mapping Excel -> Db" + "\n"
        for sSheet in self.lookup_getExcelSheets():
            lst = []
            sOut += str(sSheet)+": "
            for sCol in self.lookup_getExcelCols(sSheet):
                sTable, sAtr = self.lookup_excel2db(sSheet, sCol)
                lst.append(str(sCol) + "->" + str(sTable) + "."+str(sAtr))
            sOut += PrettyText.lst2strC(lst) + "\n"
        sOut += "\n"
        sOut += "--- Mapping Db -> Excel" + "\n"
        for sTable in self.lookup_getDbTables():
            lst = []
            sOut += str(sTable)+": "
            for sAtr in self.lookup_getDbAttributes(sTable):
                sSheet, sCol = self.lookup_db2excel(sTable, sAtr)
                lst.append(str(sAtr) + "->" + str(sSheet) + "."+str(sCol))
            sOut += PrettyText.lst2strC(lst) + "\n"
        sOut += "\n"
        sOut += "--- Db table keys" + "\n"
        sOut += PrettyText.lst2strC([sSheet+"."+self.lookup_getKey(sSheet) for sSheet in self.lookup_getDbTables()]) + "\n"
        sOut += "\n"
        sOut += "--- Db table lookups" + "\n"
        for sTable in self.lookup_getDbTables():
            if self.lookup_moreLookupNeeded_isTableNeeding(sTable):
                lstTmp = []
                for sAttrib in self.lookup_moreLookupNeeded_getNeedingAttributes(sTable):
                    sLookup = self.lookup_moreLookupNeeded_getNeededLookup(sTable, sAttrib)
                    lstTmp.append(str(sAttrib)+" -> "+str(sLookup))
                sOut += sTable + ": " + PrettyText.lst2strC(lstTmp) + "\n"
        sOut += "\n"
        sOut += "--- Db table obligate attributes" + "\n"
        for sSheet in self.lookup_getExcelSheets():
            lst = []
            for sCol in self.lookup_getExcelCols(sSheet):
                if self.lookup_isObl(sSheet, sCol):
                    lst.append(sCol)
            if len(lst) > 0:
                sOut += str(sTable) + ": " + PrettyText.lst2strC(lst) + "\n"
        sOut += "\n"
        sOut += "************************************\n"
        sOut += "Taxon System" + "\n"
        sOut += "************************************\n\n"
        sOut += "--- Excel sheets (->table)" + "\n"
        sOut += PrettyText.lst2strC([str(sSheet) + " (" + str(self.taxon_sheet2table(sSheet))+")" for sSheet in  self.taxon_getExcelSheets()]) + "\n"
        sOut += "\n"
        sOut += "--- Db tables (->sheet)" + "\n"
        sOut += PrettyText.lst2strC([str(sTable) + " (" + str(self.taxon_table2sheet(sTable)) + ")" for sTable in self.taxon_getDbTables()]) + "\n"
        sOut += "\n"
        sOut += "--- Mapping Excel -> Db" + "\n"
        for sSheet in self.taxon_getExcelSheets():
            lst = []
            sOut += str(sSheet)+": "
            for sCol in self.taxon_getExcelCols(sSheet):
                sTable, sAtr = self.taxon_excel2db(sSheet, sCol)
                lst.append(str(sCol) + "->" + str(sTable) + "."+str(sAtr))
            sOut += PrettyText.lst2strC(lst) + "\n"
        sOut += "\n"
        sOut += "--- Mapping Db -> EXCEL" + "\n"
        for sTable in self.taxon_getDbTables():
            lst = []
            sOut += str(sTable)+": "
            for sAtr in self.taxon_getDbAttributes(sTable):
                sSheet, sCol = self.taxon_db2excel(sTable, sAtr)
                lst.append(str(sAtr) + "->" + str(sSheet) + "."+str(sCol))
            sOut += PrettyText.lst2strC(lst) + "\n"
        sOut += "\n"
        sOut += "--- Obligate cols" + "\n"
        for sSheet in self.taxon_getExcelSheets():
            lst = []
            for sCol in self.taxon_getExcelCols(sSheet):
                if self.taxon_isObl(sSheet, sCol):
                    lst.append(sCol)
            if len(lst) > 0:
                sOut += str(sSheet) + ": " + PrettyText.lst2strC(lst) + "\n"
        sOut += "\n"
        sOut += "--- Keys" + "\n"
        sOut += PrettyText.lst2strC(
            [sSheet + "." + self.taxon_getKey(sSheet) for sSheet in self.taxon_getDbTables()]) + "\n"
        sOut += "\n"
        sOut += "************************************\n"
        sOut += "Ingest System" + "\n"
        sOut += "************************************\n\n"
        sOut += "--- Excel sheets (->table)" + "\n"
        sOut += PrettyText.lst2strC([str(sSheet) + " (" + str(self.ingest_sheet2table(sSheet))+")" for sSheet in  self.ingest_getExcelSheets()]) + "\n"
        sOut += "\n"
        sOut += "--- Db tables (->sheet)" + "\n"
        sOut += PrettyText.lst2strC([str(sTable) + " (" + str(self.ingest_table2sheet(sTable)) + ")" for sTable in self.ingest_getDbTables()]) + "\n"
        sOut += "\n"
        sOut += "--- Workflows" + "\n"
        sOut += PrettyText.lst2strC([sWf +" (" + self.workflow_short2long(sWf) + ")" for sWf in self.workflow_getAll()]) + "\n"
        sOut += "\n"
        sOut += "--- Converter Workflows -> Workflows" + "\n"
        sOut += PrettyText.lst2strC(
            [sWf + "(" + self.workflow_converter_import_short2long(sWf) + ") -> " + self.workflow_converter_import_getTarget(sWf) for sWf in self.workflow_converter_import_getAll()]) + "\n"
        sOut += "\n"
        sOut += "--- Excel sheets (per Workflow)" + "\n"
        for sWf in self.workflow_getAll():
            sOut += "* " + self.workflow_short2long(sWf) + "\n"
            for sSheet in self.ingest_getExcelSheets(sWf):
                lst = []
                sOut += str(sSheet) + ": "
                for sCol in self.ingest_getExcelCols(sSheet, sWf):
                    sTable, sAtr = self.ingest_excel2db(sSheet, sCol)
                    lst.append(str(sCol))
                sOut += PrettyText.lst2strC(lst) + "\n"
            sOut += "\n"
        sOut += "--- Mapping Excel -> Db" + "\n"
        for sSheet in self.ingest_getExcelSheets():
            lst = []
            for sCol in self.ingest_getExcelCols(sSheet):
                sTable, sAtr = self.ingest_excel2db(sSheet, sCol)
                if (sTable, sAtr) == (None, None):
                    sInfo = str(sCol)
                else:
                    sInfo = str(sCol) + "->" + str(sTable) + "." + str(sAtr)
                lst.append(sInfo)
            if len(lst) >0 :
                sOut += str(sSheet) + ": " + PrettyText.lst2strC(lst) + "\n"
        sOut += "\n"
        sOut += "--- Obligate cols" + "\n"
        for sSheet in self.ingest_getExcelSheets():
            lst = []
            for sCol in self.ingest_getExcelCols(sSheet):
                if self.ingest_isObl(sSheet, sCol):
                    lst.append(sCol)
            if len(lst) >0 :
                sOut += str(sSheet) + ": " + PrettyText.lst2strC(lst) + "\n"
        sOut += "\n"
        sOut += "--- Lookup cols" + "\n"
        for sSheet in self.ingest_getExcelSheets():
            lst = []
            for sCol in self.ingest_getExcelCols(sSheet):
                if self.ingest_isLookup(sSheet, sCol):
                    sLupTab = self.ingest_getLookup(sSheet,sCol)
                    lst.append(sCol+" ("+str(sLupTab)+")")
            if len(lst) >0 :
                sOut += str(sSheet) + ": " + PrettyText.lst2strC(lst) + "\n"
        sOut += "\n"
        sOut += "--- Geo columns" + "\n"
        for sSheet in self.ingest_getExcelSheets():
            lst = []
            for sCol in self.ingest_getExcelCols(sSheet):
                if self.ingest_isGeo(sSheet, sCol):
                    lst.append(sCol + " ("+self.ingest_getGeo(sSheet,sCol)+")")
            if len(lst) > 0:
                sOut += str(sSheet) + ": " + PrettyText.lst2strC(lst) + "\n"
        sOut += "\n"
        sOut += "--- Date cols" + "\n"
        for sSheet in self.ingest_getExcelSheets():
            lst = []
            for sCol in self.ingest_getExcelCols(sSheet):
                if self.ingest_isDate(sSheet, sCol):
                    lst.append(sCol)
            if len(lst) > 0:
                sOut += str(sSheet) + ": " + PrettyText.lst2strC(lst) + "\n"
        sOut += "\n"
        sOut += "--- Time cols" + "\n"
        for sSheet in self.ingest_getExcelSheets():
            lst = []
            for sCol in self.ingest_getExcelCols(sSheet):
                if self.ingest_isTime(sSheet, sCol):
                    lst.append(sCol)
            if len(lst) > 0:
                sOut += str(sSheet) + ": " + PrettyText.lst2strC(lst) + "\n"
        sOut += "\n"
        sOut += "--- FK-PK Pairings per Workflow" + "\n"
        for sWf in self.workflow_getAll():
            sOut += "* " + self.workflow_short2long(sWf) + "\n"
            for sSheetFk, sSheetPk in self.ingest_getFkPkPartnerAll(sWf).items():
                sOut += sSheetFk + "(FK)->" +sSheetPk + "(PK)\n"
            sOut += "\n"
        sOut += "--- PK usages per Workflow" + "\n"
        for sWf in self.workflow_getAll():
            sOut += "* " + self.workflow_short2long(sWf) + "\n"
            for sSheetPk, lstSheetFk in self.ingest_getPkUseAll(sWf).items():
                sOut += sSheetPk + "(PK)<-[" +  PrettyText.lst2strC(lstSheetFk)+ "](FK)\n"
            sOut += "\n"
        sOut += "--- PK cols per Workflow" + "\n"
        for sWf in self.workflow_getAll():
            sOut += "* " + self.workflow_short2long(sWf) + "\n"
            for sSheet in self.ingest_getExcelSheets(sWf):
                lst = []
                for sCol in self.ingest_getExcelCols(sSheet,sWf):
                    if self.ingest_isPk(sSheet, sCol, sWf):
                        if self.ingest_isObl(sSheet, sCol):
                            lst.append(sCol)
                        else:
                            lst.append(sCol+"{not obligate}")
                if len(lst) > 0:
                    sOut += str(sSheet) + ": " + PrettyText.lst2strC(lst) + "\n"
            sOut += "\n"
        sOut += "--- FK cols per workflow (<target>)" + "\n"
        for sWf in self.workflow_getAll():
            sOut += "* " + self.workflow_short2long(sWf) + "\n"
            for sSheet in self.ingest_getExcelSheets(sWf):
                lst = []
                for sCol in self.ingest_getExcelCols(sSheet,sWf):
                    if self.ingest_isFk(sSheet, sCol):
                        sInfoMain = ""
                        if self.ingest_isFkTarget(sSheet, sCol):
                            sFkMainCol, sFkMainTable, sFkMainAttrib = self.ingest_getFkMain(sSheet)
                            sInfoMain = "("+sFkMainTable + "."+ sFkMainAttrib+")"
                        lst.append(sCol+sInfoMain)
                if len(lst) > 0:
                    sOut += str(sSheet) + ": " + PrettyText.lst2strC(lst) + "\n"
            sOut += "\n"
        sOut += "--- PK combination per workflow" + "\n"
        for sWf in self.workflow_getAll():
            sOut += "* " + self.workflow_short2long(sWf) + "\n"
            for sSheet in self.ingest_getExcelSheets(sWf):
                lst = self.ingest_getPk(sSheet, sWf)
                if lst is not None and len(lst) > 0:
                    lst2 = []
                    for sCol in lst:
                        if self.ingest_isObl(sSheet, sCol):
                            lst2.append(sCol)
                        else:
                            lst2.append(sCol + "{not obligate}")
                    sOut += str(sSheet) + ": " + PrettyText.lst2str(lst2,"-") + "\n"
            sOut += "\n"
        sOut += "--- FK combination" + "\n"
        for sWf in self.workflow_getAll():
            sOut += "* " + self.workflow_short2long(sWf) + "\n"
            for sSheet in self.ingest_getExcelSheets(sWf):
                lst = self.ingest_getFk(sSheet, sWf)
                if lst is not None and len(lst) > 0:
                    sOut += str(sSheet) + ": " + PrettyText.lst2str(lst,"-") + "\n"
            sOut += "\n"
        if len(self.__lstInfosEtc) > 0:
            sOut += "************************************\n"
            sOut += "Miscellaneous\n"
            sOut += "************************************\n\n"
            sOut += PrettyText.lst2str(self.__lstInfosEtc, "\n")
            sOut += "\n"
        return sOut

    # Erzeuge Mapping
    def __createMapping(self):
        res = Result("Create mapping")
        resWf = self.__checkWorflowNames()
        resLup = self.__createMappingLookup()
        resIng = self.__createMappingIngest()
        resTax = self.__createMappingTaxon()
        resAlias = self.__createMappingAlias()
        resFinal = self.__createMappingFinalize()
        res.addResult(resWf)
        res.addResult(resIng)
        res.addResult(resLup)
        res.addResult(resTax)
        res.addResult(resAlias)
        res.addResult(resFinal)
        return res

    def __checkWorflowNames(self):
        res = Result("Check Worflow Names")
        setTmp = set()
        for sWf_short, sWf_long in self.__odicMapping__Wf.items():
            if sWf_long in setTmp:
                res.addErr("Double long name in workflows: " + str(sWf_long))
                return res
            else:
                setTmp.add(sWf_long)
        for sWf_conv_imp in self.__odicMapping__Wf_converter_import.items():
            if sWf_conv_imp in self.__odicMapping__Wf:
                res.addErr("Double name of Workflow in native and converter import: " + str(sWf_conv_imp))
                return res
        for sWf_conv_exp in self.__odicMapping__Wf_converter_export.items():
            if sWf_conv_exp in self.__odicMapping__Wf:
                res.addErr("Double name of Workflow in native and converter export: " + str(sWf_conv_exp))
                return res
        return res

    # Erzeuge System zur Erkennung von alternativen Colum-Namen für jedes System
    def __createMappingAlias(self):
        res = Result("Create Aliases System")
        # Sheets
        dicSheetAliases = OrderedDict()
        for sSheet in self.all_getExcelSheets():
            # Alle normalen Namen: Lower -> Normal
            if sSheet.lower() in dicSheetAliases:
                res.addErr("Double sheet (in lower case): " + str(sSheet).lower())
            else:
                dicSheetAliases[sSheet.lower()] = sSheet
            # Alle Aliase: Lower -> Normal
            if self.all_getAliasesSheet(sSheet) is not None:
                for sSheetAlias in self.all_getAliasesSheet(sSheet):
                    if sSheetAlias.lower() in dicSheetAliases:
                        res.addErr("Double sheet alias (in lower case): " + str(sSheetAlias).lower())
                    else:
                        dicSheetAliases[sSheetAlias.lower()] = sSheet
        self.__odicMapping__All_ExcelSheetAlias2Sheet = dicSheetAliases

        # Columns
        dicColAliases = OrderedDict()
        for sSystem, f1,f2 in [
            ("ingest", self.ingest_getExcelSheets, self.ingest_getExcelCols),
            ("lookup", self.lookup_getExcelSheets, self.lookup_getExcelCols),
            ("taxon", self.taxon_getExcelSheets, self.taxon_getExcelCols)]:
            # Init
            dicColAliases[sSystem] = OrderedDict()
            for sSheet in f1():
                dicColAliases[sSystem][sSheet] = OrderedDict()
            # Alle normalen Namen: Lower -> Normal
            for sSheet in f1():
                for sCol in f2(sSheet):
                    if sCol.lower() in dicColAliases[sSystem][sSheet]:
                        res.addErr("Double column in sheet "+str(sSheet)+" (in lower case): "+str(sCol).lower())
                    else:
                        dicColAliases[sSystem][sSheet][sCol.lower()] = sCol
            # Alle Aliase: Lower -> Normal
            for sSheet, dicColAltSubSystem in self.__dicMapping__All_ExcelSheetCol2AliasCol[sSystem].items():
                for sCol, lstColAlt in dicColAltSubSystem.items():
                    for sColAlt in lstColAlt:
                        if sColAlt.lower() in dicColAliases[sSystem][sSheet]:
                            res.addErr("Double column in sheet " + str(sSheet) + " (in lower case): " + str(sColAlt).lower())
                        else:
                            dicColAliases[sSystem][sSheet][sColAlt.lower()] = sCol
        self.__odicMapping__All_ExcelSheetColAlias2Col = dicColAliases
        return res

    # Erzeuge Mapping: Lookup
    def __createMappingGeneric(self, sNameMapping, sMap):

        res = Result(sNameMapping)

        pn = self.__smartprinter.normal
        pv = self.__smartprinter.verbose

        dicTest_sheet2table = dict()
        dicTest_table2sheet = dict()
        dicTest_sheetCol2tableAtr = dict()
        dicTest_tableAtr2sheetCol = dict()
        dicTest_tableAtr2sheetCol = dict()

        odicExl2db = OrderedDict()
        odicDb2exl = OrderedDict()
        odicDb2opts = OrderedDict()
        odicDb2extra = OrderedDict() # Variable Anzahl von extra Feldern

        setTestUniqSheetColLower = set()

        dicAliasColNames = {}
        dicAliasSheetNames = {}

        try:
            # Einlesen der CSV-Datei
            with open(sMap, newline='', encoding='utf-8') as csvfile:
                csvReader = csv.reader(csvfile, delimiter=',', quotechar='"')
                odicExtraCol2name = OrderedDict() # Columns und Namen der Extrazeilen
                # Zeilen
                for iZeile, lstZeile in enumerate(csvReader):
                    dicExtraRow = {} # Extra Daten in Zeile
                    if iZeile == 0:
                        # Erste Zeile: Header
                        # Werte zusätzliche Felder aus -> Erzeuge Extra Data Map
                        if len(lstZeile) > 5:
                            for i in range(5, len(lstZeile)):
                                sName = lstZeile[i]
                                if sName in odicDb2extra:
                                    raise NameError("Double definition of extra field in workflow mapping file: file: " + str(sMap) + ", field: " + str(sName))
                                else:
                                    odicDb2extra[sName] = OrderedDict()
                                    odicExtraCol2name[i] = sName
                    elif len(lstZeile) != 0 and not lstZeile[0].startswith("#"):
                        # Nicht leer und kein Kommentar
                        # Hole Inhalte (Spalten wird direkt durch Index angesprochen, Header werden nicht benutzt)
                        # ---------------------------------------------------------------
                        sExcelSheet = lstZeile[0]
                        sExcelCol = lstZeile[1]
                        sExcelSheetAlias = lstZeile[2]
                        sExcelColAlias = lstZeile[3]
                        sDbTable = lstZeile[4]
                        sDbAttrib = lstZeile[5]
                        sOpts = lstZeile[6]
                        # Sammle alle Extra Daten ein
                        for iExtraCol, sExtraName in odicExtraCol2name.items():
                            sExtraVal = lstZeile[iExtraCol]
                            if sExtraVal not in [None, ""]:
                                dicExtraRow[sExtraName] = sExtraVal
                        # Test Eintrag ist leer
                        for sEntry in [sExcelSheet, sExcelCol, sDbTable, sDbAttrib]:
                            if sEntry is None or sEntry.strip() == "":
                                raise NameError("Bad entry in definition: " + str(sExcelSheet) + "." + str(
                                    sExcelCol) + "->" + str(sDbTable) + "." + str(sDbAttrib))
                        # Test: Excel Tuple Einmaligkeit (lowercase)
                        if (sExcelSheet.lower(), sExcelCol.lower()) in setTestUniqSheetColLower:
                            res.addErr("Double sheet.colum (in lowercase): " + str(sExcelSheet).lower() + "." + str(
                                sExcelCol).lower())
                        else:
                            setTestUniqSheetColLower.add((sExcelSheet.lower(), sExcelCol.lower()))
                        # Col Aliase -> Speichern
                        if sExcelColAlias.strip() not in ["", None]:
                            lstColAltName = sExcelColAlias.split(",")
                            if sExcelSheet not in dicAliasColNames:
                                dicAliasColNames[sExcelSheet] = {sExcelCol: lstColAltName}
                            else:
                                dicAliasColNames[sExcelSheet][sExcelCol] = lstColAltName
                        # Sheet Aliase -> Speichern
                        if sExcelSheetAlias.strip() not in ["", None]:
                            lstSheetAltName = sExcelSheetAlias.split(",")
                            if sExcelSheet in dicAliasSheetNames:
                                dicAliasSheetNames[sExcelSheet] += lstSheetAltName
                            else:
                                dicAliasSheetNames[sExcelSheet] = lstSheetAltName
                        # Test Konsistenz: Excel Sheet <-> DB Table
                        if sExcelSheet not in dicTest_sheet2table:
                            dicTest_sheet2table[sExcelSheet] = sDbTable
                        if sDbTable not in dicTest_table2sheet:
                            dicTest_table2sheet[sDbTable] = sExcelSheet
                        if dicTest_sheet2table[sExcelSheet] != sDbTable:
                            raise NameError(
                                "Double definition of Excel sheet -> db table: " + str(sExcelSheet) + ": " + str(
                                    sDbTable) + " -> " + str(dicTest_sheet2table[sExcelSheet]))
                        if dicTest_table2sheet[sDbTable] != sExcelSheet:
                            raise NameError(
                                "Double definition of db table -> Excel sheet: " + str(sDbTable) + ": " + str(
                                    sExcelSheet) + " -> " + str(dicTest_table2sheet[sDbTable]))
                        # Test Konsistenz Excel Sheet/Col <-> DB Table/Attrib
                        if (sExcelSheet, sExcelCol) not in dicTest_sheetCol2tableAtr:
                            dicTest_sheetCol2tableAtr[(sExcelSheet, sExcelCol)] = (sDbTable, sDbAttrib)
                        else:
                            raise NameError(
                                "Double definition of Excel sheet/col -> db table/attribute: " + str(
                                    sExcelSheet) + "." + str(
                                    sExcelCol) + " -> " + str(sDbTable) + "." + str(sDbAttrib) + " and " + str(
                                    dicTest_sheetCol2tableAtr[(sExcelSheet, sExcelCol)][0]) + "." + str(
                                    dicTest_sheetCol2tableAtr[(sExcelSheet, sExcelCol)][1]))

                        if (sDbTable, sDbAttrib) not in dicTest_tableAtr2sheetCol:
                            dicTest_tableAtr2sheetCol[(sDbTable, sDbAttrib)] = (sExcelSheet, sExcelCol)
                        else:
                            raise NameError(
                                "Double definition of db table/attribute -> Excel sheet/col: " + str(
                                    sDbTable) + "." + str(
                                    sDbAttrib) + " -> " + str(sExcelSheet) + "." + str(sExcelCol) + " and " + str(
                                    dicTest_tableAtr2sheetCol[(sDbTable, sDbAttrib)][0]) + "." + str(
                                    dicTest_tableAtr2sheetCol[(sDbTable, sDbAttrib)][1]))

                        # Speichern: Excel: Sheet/Col <-> DB:Table/Attrib
                        if sExcelSheet not in odicExl2db:
                            odicExl2db[sExcelSheet] = OrderedDict()
                        odicExl2db[sExcelSheet][sExcelCol] = (sDbTable, sDbAttrib)
                        if sDbTable not in odicDb2exl:
                            odicDb2exl[sDbTable] = OrderedDict()
                        odicDb2exl[sDbTable][sDbAttrib] = (sExcelSheet, sExcelCol)
                        # Speichern: DB:Table/Attrib -> Opts
                        if sOpts is None:
                            sOpts = ""
                        lstOpts = sOpts.strip().split()
                        if len(lstOpts) > 0:
                            if sDbTable not in odicDb2opts:
                                odicDb2opts[sDbTable] = OrderedDict()
                            odicDb2opts[sDbTable][sDbAttrib] = lstOpts
                        # Speichern: DB:Table/Attrib -> Extra
                        if len(dicExtraRow) != 0:
                            for sExtraName, sExtraVal in dicExtraRow.items():
                                lstExtra = sExtraVal.strip().split()
                                if len(lstExtra) > 0:
                                    if sDbTable not in odicDb2extra[sExtraName]:
                                        odicDb2extra[sExtraName][sDbTable] = OrderedDict()
                                    odicDb2extra[sExtraName][sDbTable][sDbAttrib] = lstExtra

        except Exception as ex:
            res.addErr(str(ex))
        res.addData("exl2db", odicExl2db)
        res.addData("db2exl", odicDb2exl)
        res.addData("db2opts", odicDb2opts)
        res.addData("altcol", dicAliasColNames)
        res.addData("altsheet", dicAliasSheetNames)
        # Extra
        for sNameExtra, odicData in odicDb2extra.items():
            res.addData("db2"+sNameExtra, odicData)
        return res

    # Erzeuge Mapping: Lookup
    def __createMappingLookup(self):
        res = self.__createMappingGeneric(
            "Create mapping: Lookup",
            os.path.dirname(os.path.realpath(__file__)) + "/mapper_lookup.csv")
        self.__odicMapping__Lookup_excel2db = res.getData("exl2db")
        self.__odicMapping__Lookup_db2excel = res.getData("db2exl")
        self.__odicMapping__Lookup_db2lookup = res.getData("db2lookup")
        self.__dicMapping__All_ExcelSheetCol2AliasCol["lookup"] = res.getData("altcol")
        self.__dicMapping__All_ExcelSheet2AliasSheet["lookup"] = res.getData("altsheet")
        odicTable2opts = res.getData("db2opts")
        # Opt: Obligate
        for sLookup, odicAtrrOpts in odicTable2opts.items():
            for sAttrib, lstOpts in odicAtrrOpts.items():
                if "obl" in lstOpts:
                    self.__odicMapping__Lookup_tableAtr2obl[(sLookup, sAttrib)] = True
        # Test Opts: Lookup Key
        # Doppelte Definition eines Key
        odicTable2keyVal = OrderedDict()
        for sLookup, odicAtrrOpts in odicTable2opts.items():
            for sAttrib, lstOpts in odicAtrrOpts.items():
                if "key" in lstOpts:
                    if sLookup not in odicTable2keyVal:
                        odicTable2keyVal[sLookup] = sAttrib
                    else:
                        res.addErr("Double definition of key for lookup table: " + str(sLookup) + " -> " + str(sAttrib))
        # Keine Definition eines Key
        for sTable in self.lookup_getDbTables():
            if sTable not in odicTable2keyVal:
                res.addErr("No key defined for lookup table: " + str(sTable))
        # Speichere key Map
        self.__odicMapping__Lookup_table2key = odicTable2keyVal
        # Test: Lookup (gibt es das Lookup?)
        for sTable, odic in self.__odicMapping__Lookup_db2lookup.items():
            for sAttrib, lstLookup in odic.items():
                if len(lstLookup) > 1:
                    res.addErr("More than one lookup table defined: " + str(sTable) + "." + str(sAttrib) + " -> " + str(PrettyText.lst2strC(lstLookup)))
                else:
                    sLookup = lstLookup[0]
                    if sLookup not in self.lookup_getDbTables():
                        res.addErr("Lookup table not known: " + str(sTable) + "." + str(sAttrib) + " -> " + str(sLookup))
        return res

    # Erzeuge Mapping: Taxon
    def __createMappingTaxon(self):
        res = self.__createMappingGeneric(
            "Create mapping: Taxon",
            os.path.dirname(os.path.realpath(__file__)) + "/mapper_taxon.csv")
        self.__odicMapping__Taxon_excel2db = res.getData("exl2db")
        self.__odicMapping__Taxon_db2excel = res.getData("db2exl")
        self.__dicMapping__All_ExcelSheetCol2AliasCol["taxon"] = res.getData("altcol")
        self.__dicMapping__All_ExcelSheet2AliasSheet["taxon"] = res.getData("altsheet")
        odicTable2opts = res.getData("db2opts")
        # Opt: Obligate
        for sTable, odicAtrrOpts in odicTable2opts.items():
            for sAttrib, lstOpts in odicAtrrOpts.items():
                if "obl" in lstOpts:
                    self.__odicMapping_Taxon_db2obl[(sTable, sAttrib)] = True
        # Opt: Key
        odicTable2keyVal = OrderedDict()
        for sTable, odicAtrrOpts in odicTable2opts.items():
            for sAttrib, lstOpts in odicAtrrOpts.items():
                if "key" in lstOpts:
                    if sTable not in odicTable2keyVal:
                        odicTable2keyVal[sTable] = sAttrib
                    else:
                        # Doppelte Definition eines Key
                        res.addErr("Double definition of key for lookup table: " + str(sTable) + " -> " + str(
                            sAttrib))
        # Speichere key Map
        self.__odicMapping__Taxon_table2key = odicTable2keyVal
        # Keine Definition eines Key
        for sTable in self.taxon_getDbTables():
            if sTable not in odicTable2keyVal:
                res.addErr("No key defined for taxon system table: " + str(sTable))
        return res

    # Erzeuge Mapping: Ingest
    def __createMappingIngest(self):
        pn = self.__smartprinter.normal
        pv = self.__smartprinter.verbose

        # Helper: Werte Parameter-String aus
        def processParams(sParams):

            bErr = False
            sErr = ""

            bPk = False
            bFk = False
            bObl = False
            bDate = False
            bTime = False
            bLat = False
            bLong = False

            lstParams = sParams.split()
            for sParam in lstParams:
                if sParam not in self.__lstMapping__Params:
                    bErr = True
                    sErr += "Unknown parameter found: " + str(sParam) + "\n"
                elif sParam == "pk":
                    bPk = True
                    bObl = True
                elif sParam == "fk":
                    bFk = True
                    bObl = True
                elif sParam == "lat":
                    bLat = True
                elif sParam == "lon":
                    bLong = True
                elif sParam == "date":
                    bDate = True
                elif sParam == "time":
                    bTime = True
                elif sParam == "obl":
                    bObl = True

            # Tests
            if bLat and bLong:
                bErr = True
                sErr += "Parameter lon and lat are set simultaneously" + "\n"
            if bTime and bDate:
                bErr = True
                sErr += "Parameter time and date are set simultaneously" + "\n"

            return bErr, sErr, bPk, bFk, bObl, bDate, bTime, bLat, bLong

        # ---

        res = Result(name="Create mapping: ingest")

        sDateiMap = os.path.dirname(os.path.realpath(__file__)) + "/mapper_ingest.csv"
        odicMap = OrderedDict()

        odicTestUniqSheet2table = {}
        setTestUniqSheetColLower = set()

        sSheetLastProccessed = None

        try:
            # Einlesen der CSV-Datei
            with open(sDateiMap, newline='', encoding='utf-8') as csvfile:
                csvReader = csv.reader(csvfile, delimiter=',', quotechar='"')
                # Zeilen
                for iZeile, lstZeile in enumerate(csvReader):
                    # Nicht leer und kein Kommentar und nicht erste Zeile
                    if len(lstZeile) != 0 and not lstZeile[0].startswith("#") and iZeile != 0:

                        # Hole Inhalte (Spalten wird direkt durch Index angesprochen, Header werden nicht benutzt)
                        # ---------------------------------------------------------------
                        sExcelSheet = lstZeile[0]
                        sExcelCol = lstZeile[1]
                        sExcelSheetAlias = lstZeile[2]
                        sExcelColAlias = lstZeile[3]
                        sDbTable = lstZeile[4]
                        sDbAttrib = lstZeile[5]
                        sDbLupTable = lstZeile[6]
                        sParam = lstZeile[7]
                        sWf = lstZeile[8]

                        # Werte Parameterzeile aus
                        bErrParam, sErrParam, bPk, bFk, bObl, bDate, bTime, bLat, bLong = processParams(sParam)
                        if bErrParam:
                            res.addErr(sErrParam)

                        # Wf: * = alle
                        if sWf == "*":
                            sWf = PrettyText.lst2str(self.workflow_getAll(), " ")

                        # ---------------------------------------------------------------

                        # Konsistenzchecks des Eintrages in der Mapping-Definitionsdatei
                        if sExcelSheet != "":
                            # Excel Tuple OK?
                            if (sExcelCol == ""):
                                res.addErr(" Missing column for sheet: " + sExcelSheet, row=iZeile)
                            # Test: Excel Tuple Einmaligkeit (lowercase)
                            if (sExcelSheet.lower(), sExcelCol.lower()) in setTestUniqSheetColLower:
                                res.addErr("Double sheet.colum (in lowercase): " + str(sExcelSheet).lower() + "." + str(sExcelCol).lower())
                            else:
                                setTestUniqSheetColLower.add((sExcelSheet.lower(), sExcelCol.lower()))
                            # DB-Tuple OK
                            if (bool(sDbTable) is not bool(sDbAttrib)):
                                res.addErr("DB table.attribute not set both:: " +str(sDbTable)+"."+str(sDbAttrib))
                            # ODic Excel Sheet/Col -> Db table attribute
                            if (sExcelSheet, sExcelCol) not in self.__odicMapping__Ingest_ExcelSheetCol2dbTableAttrib \
                                    and (sDbTable != "" and sDbAttrib != ""):
                                self.__odicMapping__Ingest_ExcelSheetCol2dbTableAttrib[(sExcelSheet, sExcelCol)] = (sDbTable, sDbAttrib)
                            # Workflow definiert
                            if sWf in [None, ""]:
                                res.addErr("Workflow missing: " + str(sExcelSheet) + "." + sExcelCol)
                            # ODic Excel Sheet/Col -> List Workflows
                            if (sExcelSheet, sExcelCol) not in self.__odicMapping__Ingest_ExcelSheetCol2wf:
                                lstWf = sWf.split()
                                for sWfSingle in lstWf:
                                    if sWfSingle not in self.__odicMapping__Wf:
                                        res.addErr("Workflow not allowed: " + sWfSingle)
                                self.__odicMapping__Ingest_ExcelSheetCol2wf[(sExcelSheet, sExcelCol)] = lstWf
                            # ODic Excel Sheet/Col -> Db Lookup
                            if ((sExcelSheet, sExcelCol) not in self.__odicMapping__Ingest_ExcelSheetCol2dbLup) and \
                                    sDbLupTable not in ["", None]:
                                self.__odicMapping__Ingest_ExcelSheetCol2dbLup[(sExcelSheet, sExcelCol)] = sDbLupTable
                            # ODic Excel Sheet -> List Cols
                            if sExcelSheet not in self.__odicMapping__Ingest_ExcelSheet2col:
                                self.__odicMapping__Ingest_ExcelSheet2col[sExcelSheet] = [sExcelCol]
                            else:
                                if sExcelCol not in self.__odicMapping__Ingest_ExcelSheet2col[sExcelSheet]:
                                    self.__odicMapping__Ingest_ExcelSheet2col[sExcelSheet].append(sExcelCol)
                            # ODic Excel Sheet -> Alias Sheet Name
                            if sExcelSheetAlias.strip() not in ["", None]:
                                lstSheetAliasName = sExcelSheetAlias.split(",")
                                if sExcelSheet not in self.__dicMapping__All_ExcelSheet2AliasSheet["ingest"]:
                                    self.__dicMapping__All_ExcelSheet2AliasSheet["ingest"][sExcelSheet] = lstSheetAliasName
                                else:
                                    self.__dicMapping__All_ExcelSheet2AliasSheet["ingest"][sExcelSheet] += lstSheetAliasName
                            # ODic Excel Sheet/Col -> Alias Col Name
                            if sExcelColAlias.strip() not in ["", None]:
                                lstColAliasName = sExcelColAlias.split(",")
                                if sExcelSheet not in self.__dicMapping__All_ExcelSheetCol2AliasCol["ingest"]:
                                    self.__dicMapping__All_ExcelSheetCol2AliasCol["ingest"][sExcelSheet] = {
                                        sExcelCol: lstColAliasName}
                                else:
                                    self.__dicMapping__All_ExcelSheetCol2AliasCol["ingest"][sExcelSheet][
                                        sExcelCol] = lstColAliasName
                            # ODic Excel Sheet/Col -> Obl
                            if (sExcelSheet, sExcelCol) not in self.__odicMapping__Ingest_ExcelSheetCol2obl:
                                self.__odicMapping__Ingest_ExcelSheetCol2obl[(sExcelSheet, sExcelCol)] = bObl
                            # Odict Excel Sheet Col -> lon/lat
                            if bLat or bLong:
                                sEntry = "lat" if bLat else "lon"
                                self.__odicMapping__Ingest_ExcelSheetCol2geo[(sExcelSheet, sExcelCol)] = sEntry
                            # Odict Excel Sheet -> List Cols time
                            if bTime:
                                if sExcelSheet not in self.__odicMapping__Ingest_ExcelSheet2time:
                                    self.__odicMapping__Ingest_ExcelSheet2time[sExcelSheet] = [sExcelCol]
                                else:
                                    if sExcelCol not in self.__odicMapping__Ingest_ExcelSheet2time[sExcelSheet]:
                                        self.__odicMapping__Ingest_ExcelSheet2time[sExcelSheet].append(sExcelCol)
                            # Odict Excel Sheet -> List cols date
                            if bDate:
                                if sExcelSheet not in self.__odicMapping__Ingest_ExcelSheet2date:
                                    self.__odicMapping__Ingest_ExcelSheet2date[sExcelSheet] = [sExcelCol]
                                else:
                                    if sExcelCol not in self.__odicMapping__Ingest_ExcelSheet2date[sExcelSheet]:
                                        self.__odicMapping__Ingest_ExcelSheet2date[sExcelSheet].append(sExcelCol)
                            # Excel Sheet: List Cols Primary Keys
                            if bPk:
                                if sExcelSheet != "" and sExcelCol != "":
                                    if sExcelSheet not in self.__odicMapping__Ingest_ExcelSheet2pk:
                                        self.__odicMapping__Ingest_ExcelSheet2pk[sExcelSheet] = [sExcelCol]
                                    else:
                                        self.__odicMapping__Ingest_ExcelSheet2pk[sExcelSheet].append(sExcelCol)
                                else:
                                    res.addErr(" Missing reference for primary key: " + (
                                    sExcelSheet if sExcelSheet != "" else "?") + "." + (
                                               sExcelCol if sExcelCol != "" else "?"), row=iZeile)
                            # Excel Sheet: List Cols Foreign Keys
                            if bFk:
                                if sExcelSheet != "" and sExcelCol != "":
                                    if sExcelSheet not in self.__odicMapping__Ingest_ExcelSheet2fk:
                                        self.__odicMapping__Ingest_ExcelSheet2fk[sExcelSheet] = [sExcelCol]
                                    else:
                                        self.__odicMapping__Ingest_ExcelSheet2fk[sExcelSheet].append(sExcelCol)
                                else:
                                    res.addErr("Missing reference for foreign key: " + (
                                    sExcelSheet if sExcelSheet != "" else "?") + "." + (
                                               sExcelCol if sExcelCol != "" else "?"), row=iZeile)
                            # Main FK
                            if bFk and sDbTable != "" and sDbAttrib != "":
                                if sExcelSheet not in self.__odicMapping__Ingest_ExcelSheet2fkMain:
                                    self.__odicMapping__Ingest_ExcelSheet2fkMain[sExcelSheet] = (sExcelCol, sDbTable, sDbAttrib)
                                else:
                                    res.addErr("More than one main (db) foreign key defined in sheet: " + str(sExcelSheet) + " -> "
                                               + str(sExcelCol)+"/"+str(sDbTable) +"/"+str(sDbAttrib) +" and "
                                               + str(self.__odicMapping__Ingest_ExcelSheet2fkMain[sExcelSheet][0]) + "/"
                                               + str(self.__odicMapping__Ingest_ExcelSheet2fkMain[sExcelSheet][1]) + "/"
                                               + str(self.__odicMapping__Ingest_ExcelSheet2fkMain[sExcelSheet][2]))

                            # Einmaligkeit von: Excel Sheet <-> Table
                            if not bFk:
                                if sExcelSheet not in odicTestUniqSheet2table:
                                    odicTestUniqSheet2table[sExcelSheet] = sDbTable
                                    self.__lstMapping__Ingest_ExcelSheet2dbTable.append((sExcelSheet,sDbTable))
                                elif odicTestUniqSheet2table[sExcelSheet] != sDbTable:
                                    res.addErr("Excel sheet is mapped to more than one db table: " + str(sExcelSheet) + " -> " + str(sDbTable) + " and " + str(odicTestUniqSheet2table[sExcelSheet]))
                            # Wird ein Sheet nochmal wieder bearbeitet?
                            if (sSheetLastProccessed is not None) and (sExcelSheet != sSheetLastProccessed) and (sExcelSheet in odicMap):
                                res.addErr("Definittion of Excel sheet already finishd: " + str(sExcelSheet))

                            # FK: Gibt es den FK überhaupt (FK:Col -> PK:Sheet)
                            if bFk:
                                pass
                            # Eintrag in Mapping-Karte
                            if sExcelSheet not in odicMap:
                                odicMap[sExcelSheet] = [(sExcelSheet, sExcelCol)]
                            else:
                                odicMap[sExcelSheet].append((sExcelSheet, sExcelCol))
                            # Letztes bearbeitetes Sheet merken
                            sSheetLastProccessed = sExcelSheet

                # Eintrag in DB Table/Attribute Liste
                odicDbAttrib = OrderedDict()
                for (sSheet, sCol), (sDbTable, sDbAttrib) in self.__odicMapping__Ingest_ExcelSheetCol2dbTableAttrib.items():
                    if (sDbTable != "" and sDbAttrib != ""):
                        if sDbTable in self. __odicMapping__Ingest_DbTable2Attrib:
                            if sDbAttrib not in self.__odicMapping__Ingest_DbTable2Attrib[sDbTable]:
                                self.__odicMapping__Ingest_DbTable2Attrib[sDbTable].append(sDbAttrib)
                            if bool(sDbTable) != bool(sDbAttrib):
                                res.addErr("Db table.attribute not set both: "+str(sDbTable)+"."+str(sDbAttrib))
                        else:
                            self.__odicMapping__Ingest_DbTable2Attrib[sDbTable] = [sDbAttrib]
                # Eintrag: DB:table,attrib -> Excel:sheet,col
                for (sSheet, sCol), (sDbTable, sDbAttrib) in self.__odicMapping__Ingest_ExcelSheetCol2dbTableAttrib.items():
                    if not (sDbTable, sDbAttrib) in self.__odicMapping__Ingest_DbTableAttrib2excelSheetCol:
                        self.__odicMapping__Ingest_DbTableAttrib2excelSheetCol[(sDbTable, sDbAttrib)] = (sSheet, sCol)
                    else:
                        res.addErr("Double usage of table.attribute: " +  str(sDbTable) +"."+ str(sDbAttrib))

            # Teste Konsistenz: FK -> PK
            for sWf in self.workflow_getAll():
                for sSheet in self.ingest_getExcelSheets(sWf):
                    lstFk = self.ingest_getFk(sSheet,sWf)
                    if lstFk is not None:
                        for sFk in lstFk:
                            if self.ingest_getPk(sFk, sWf) is None or sFk not in self.ingest_getPk(sFk, sWf):
                                res.addErr("Foreign key is not a primary key on counterpart: Workflow("+str(sWf)+") Fk(" + str(sSheet) + "." + str(sFk)+") -> Pk(" + str(sFk) + "." + str(sFk)+")")

            # Teste ob jedes Sheet genau einen Main-FK besitzt (ausser Cruise)
            for sSheet in self.ingest_getExcelSheets():
                if (sSheet != "Cruise") and (sSheet not in self.__odicMapping__Ingest_ExcelSheet2fkMain):
                    res.addErr("Sheet has no main (db) FK: "+sSheet)

            # Teste ob PKs in Subworkflow zerissen wurden -> Fehler
            odicPkFull = OrderedDict()
            for sWf in self.workflow_getAll():
                for sSheet in self.ingest_getExcelSheets():
                    lstPk = self.ingest_getPk(sSheet)
                    if (lstPk is not None) and (len(lstPk)>0):
                        odicPkFull[sSheet] = lstPk
            for sWf in self.workflow_getAll():
                for sSheet in self.ingest_getExcelSheets(sWf):
                    lstPk = self.ingest_getPk(sSheet, sWf)
                    sPk = ("|").join(lstPk) if lstPk is not None else ""
                    sPkFull = ("|").join(odicPkFull[sSheet]) if sSheet in odicPkFull else ""
                    setPk = set(lstPk) if (lstPk is not None) and (len(lstPk) > 0) else None
                    if ((setPk is None) and (sSheet in odicPkFull)):
                        sErr = "PK removed due to cropping full PK (no data will be attachable to table): Workflow:" + str(self.workflow_short2long(sWf)) + ", Sheet:" + str(sSheet) + ", PK:" + str(sPkFull)
                        res.addErr(sErr)
                    elif ((setPk is not None) and (sSheet not in odicPkFull)):
                        sErr = "Strange Error - PK popping up in sheet: Workflow:" + str(self.workflow_short2long(sWf)) + ", Sheet:" + str(sSheet) + ", PK:" + str(sPk)
                        res.addErr(sErr)
                    elif ((setPk is not None) and (sSheet in odicPkFull) and (setPk != set(odicPkFull[sSheet]))):
                        sErr = "Strange Error - PK differing in sheet: Workflow:" + str(
                            self.workflow_short2long(sWf)) + ", Sheet:" + str(sSheet) + ", PK:" + str(sPk) + ", Full-PK:" + str(sPkFull)
                        res.addErr(sErr)

            # Teste ob FKs in Subworkflow zerissen wurden -> Fehler
            odicFkFull = OrderedDict()
            for sWf in self.workflow_getAll():
                for sSheet in self.ingest_getExcelSheets():
                    lstFk = self.ingest_getFk(sSheet)
                    if (lstFk is not None) and (len(lstFk) > 0):
                        odicFkFull[sSheet] = lstFk
            for sWf in self.workflow_getAll():
                for sSheet in self.ingest_getExcelSheets(sWf):
                    lstFk = self.ingest_getFk(sSheet, sWf)
                    sFk = ("|").join(lstFk) if lstFk is not None else ""
                    sFkFull = ("|").join(odicFkFull[sSheet]) if sSheet in odicFkFull else ""
                    setFk = set(lstFk) if (lstFk is not None) and (len(lstFk) > 0) else None
                    if ((setFk is not None) and (sSheet in odicFkFull)) and (setFk != set(odicFkFull[sSheet])):
                        sErr = "FK cropped: Workflow:" + str(
                            self.workflow_short2long(sWf)) + ", Sheet:" + str(sSheet) + ", Fk:" + str(sFkFull)
                        res.addErr(sErr)
                    elif ((setFk is not None) and (sSheet not in odicFkFull)):
                        sErr = "Strange Error - FK popping up in sheet: Workflow:" + str(
                            self.workflow_short2long(sWf)) + ", Sheet:" + str(sSheet) + ", FK:" + str(sFk)
                        res.addErr(sErr)

        except Exception as ex:
            res.addErr(str(ex) + str(traceback.print_exc()))

        if res.isAnything():
            pn(res.getErrWarnInfoString())

        # Map Speichern
        self.__odicDefMap = odicMap
        return res

    # Abschliessende Tests
    def __createMappingFinalize(self):
        res = Result("Finalization of Mapping System")

        # Converter Ziele müssen valide sein
        for sWfCnv in self.workflow_converter_import_getAll():
            sWfCnvTarget = self.workflow_converter_import_getTarget(sWfCnv)
            if sWfCnvTarget not in self.workflow_getAll():
                res.addErr("Converter: target workflow is unknonw: " + str(sWfCnv)+"->"+str(sWfCnvTarget))

    # Sheet names müssen (lower) uniq sein -> Lower Alias merken
        setSheets = set()
        lstFkt = [
            self.ingest_getExcelSheets,
            self.lookup_getExcelSheets,
            self.taxon_getExcelSheets
        ]
        for f in lstFkt:
            for sSheet in f():
                if sSheet.lower() in setSheets:
                    res.addErr("Sheet name is not uniq (lower case): "+str(sSheet))
                else:
                    setSheets.add(sSheet.lower())
                    self.__odicMapping__All_ExcelSheetAlias2Sheet[sSheet.lower()] = sSheet
        return res


    # -----------------------------------------------------------------------------------------------------------

    # --- All ---

    # Table System
    def all_isTableIngest(self, sTable):
        return sTable in self.ingest_getDbTables()
    def all_isTableLookup(self, sTable):
        return sTable in self.lookup_getDbTables()
    def all_isTableTaxon(self, sTable):
        return sTable in self.taxon_getDbTables()
    def all_isTableAdmin(self, sTable):
        return sTable in self.admin_getDbTables()
    def all_getTableSystem(self, sTable):
        if sTable in self.admin_getDbTables():
            return "admin"
        elif sTable in self.lookup_getDbTables():
            return "lookup"
        elif sTable in self.taxon_getDbTables():
            return "taxon"
        elif sTable in self.ingest_getDbTables():
            return "ingest"
        else:
            return None

    # Sheet System
    def all_isSheetIngest(self, sSheet):
        return sSheet in self.ingest_getExcelSheets()
    def all_isSheetLookup(self, sSheet):
        return sSheet in self.lookup_getExcelSheets()
    def all_isSheetTaxon(self, sSheet):
        return sSheet in self.taxon_getExcelSheets()
    def all_getSheetSystem(self, sSheet):
        if sSheet in self.lookup_getExcelSheets():
            return "lookup"
        elif sSheet in self.taxon_getExcelSheets():
            return "taxon"
        elif sSheet in self.ingest_getExcelSheets():
            return "ingest"
        else:
            return None

    # Hole Columns
    def all_getExcelSheets(self, sSystem=None):
        if sSystem == "ingest":
            return self.ingest_getExcelSheets()
        elif sSystem == "lookup":
            return self.lookup_getExcelSheets()
        elif sSystem == "taxon":
            return self.taxon_getExcelSheets()
        elif sSystem is None:
            return self.ingest_getExcelSheets() + self.lookup_getExcelSheets() + self.taxon_getExcelSheets()
        else:
            return None

    # Hole Columns
    def all_getExcelCols(self, sSheet, sWf=None):
        sSystem = self.all_getSheetSystem(sSheet)
        if sSystem == "ingest":
            return self.ingest_getExcelCols(sSheet, sWf)
        elif sSystem == "lookup":
            return self.lookup_getExcelCols(sSheet)
        elif sSystem == "taxon":
            return self.taxon_getExcelCols(sSheet)
        return None

    # Existiert eine Excel Sheet
    def all_isSheetExisting(self, sSheet):
        if (sSheet in self.ingest_getExcelSheets()) or (sSheet in self.lookup_getExcelSheets()) or (sSheet in self.taxon_getExcelSheets()):
            return True
        else:
            return False

    # Hole alternative Colum Namen
    def all_getAliasesSheet(self, sSheet):
        sSystem = self.all_getSheetSystem(sSheet)
        try:
            return self.__dicMapping__All_ExcelSheet2AliasSheet[sSystem][sSheet]
        except:
            return None

    # Hole Original Sheetnamen für ein Alias (nur case insensitive)
    def all_getProperSheetName(self, sSheet):
        # Sheet existiert korrekt
        if self.all_isSheetExisting(sSheet):
            return sSheet
        # Case Insensitive
        sSheet = sSheet.strip()
        try:
            return self.__odicMapping__All_ExcelSheetAlias2Sheet[sSheet.lower()]
        except:
            return None

    # Existiert eine Excel Sheet/Colum Kombination
    def all_isSheetColExisting(self, sSheet, sCol):
        sSystem = self.all_getSheetSystem(sSheet)
        if sSystem == "ingest" and sCol in self.ingest_getExcelCols(sSheet):
                return True
        elif sSystem == "lookup" and sCol in self.lookup_getExcelCols(sSheet):
            return True
        elif sSystem == "taxon" and sCol in self.taxon_getExcelCols(sSheet):
            return True
        else:
            return False

    # Ist Excel Sheet/Colum obligat
    def all_isSheetColObl(self, sSheet, sCol):
        sSystem = self.all_getSheetSystem(sSheet)
        if sSystem == "ingest":
            return self.ingest_isObl(sSheet, sCol)
        elif sSystem == "lookup":
            return self.lookup_isObl(sSheet, sCol)
        elif sSystem == "taxon":
            return self.taxon_isObl(sSheet, sCol)
        else:
            return False

    # Hole alternative Colum Namen
    def all_getAliasesCol(self, sSheet, sCol):
        sSystem = self.all_getSheetSystem(sSheet)
        try:
            return self.__dicMapping__All_ExcelSheetCol2AliasCol[sSystem][sSheet][sCol]
        except:
            return None

    # Hole Original Colname für ein Alias
    def all_getProperColName(self, sSheet, sCol):
        sSystem = self.all_getSheetSystem(sSheet)
        # Sheet/Col existiert korrekt
        if self.all_isSheetColExisting(sSheet, sCol):
            return sCol
        # Alias und CaseInsensitive
        sCol = sCol.strip()
        try:
            return self.__odicMapping__All_ExcelSheetColAlias2Col[sSystem][sSheet][sCol.lower()]
        except:
            return None

    # --- Admin ---

    def admin_getDbTables(self):
        return list(self.__odicAdminTables.keys())

        # Hole alle Attribute einer Tabelle

    def admin_getDbAttributes(self, sTable):
        try:
            return self.__odicAdminTables[sTable]
        except:
            return None

    # --- Lookup ---

    def lookup_getExcelSheets(self):
        return list(self.__odicMapping__Lookup_excel2db.keys())

    # Hole alle Colums eines Excel Sheets
    def lookup_getExcelCols(self, sSheet):
        try:
            return list(self.__odicMapping__Lookup_excel2db[sSheet].keys())
        except:
            return None

    # Hole alle DB Tabellen
    def lookup_getDbTables(self):
        return list(self.__odicMapping__Lookup_db2excel.keys())

    # Hole alle Attribute einer Tabelle
    def lookup_getDbAttributes(self, sTable):
        try:
            return list(self.__odicMapping__Lookup_db2excel[sTable].keys())
        except:
            return None

    # Excel sheet -> db table
    def lookup_sheet2table(self, sSheet):
        if sSheet not in self.__odicMapping__Lookup_excel2db:
            return None
        setSheet = {t[0] for t in self.__odicMapping__Lookup_excel2db[sSheet].values()}
        if len(setSheet) == 1:
            return setSheet.pop()
        else:
            return None # // Fehler: Es waren mehr als eine Datenbank Tabelle in der Lookup Definition

    # DB table -> Excel Sheet
    def lookup_table2sheet(self, sTable):
        if sTable not in self.__odicMapping__Lookup_db2excel:
            return None
        setTable = {t[0] for t in self.__odicMapping__Lookup_db2excel[sTable].values()}
        if len(setTable) == 1:
            return setTable.pop()
        else:
            return None  # // Fehler: Es waren mehr als eine Datenbank Tabelle in der Lookup Definition

    # Excel Sheet,Col -> Db Table,Attrib
    def lookup_excel2db(self, sSheet, sCol):
        try:
            return self.__odicMapping__Lookup_excel2db[sSheet][sCol]
        except:
            return (None, None)

    # Db Table,Attrib -> Excel Sheet,Col
    def lookup_db2excel(self, sTable, sAttrib):
        try:
            return self.__odicMapping__Lookup_db2excel[sTable][sAttrib]
        except:
            return (None, None)

    # Alle Lookups, die weiter brauchen
    def lookup_moreLookupNeeded_getAllNeedingTables(self):
        return list(self.__odicMapping__Lookup_db2lookup.keys())

    # DB table -> Braucht die Tabelle weiter Lookups
    def lookup_moreLookupNeeded_isTableNeeding(self, sTable):
        if self.lookup_moreLookupNeeded_getNeededLookups(sTable) is None:
            return False
        else:
            return True

    # Db Table,Attrib -> Braucht Atrtribute weiters Lookup
    def lookup_moreLookupNeeded_isAttributeNeeding(self, sTable, sAttrib):
        try:
            if self.__odicMapping__Lookup_db2lookup[sTable][sAttrib][0] != None:
                return True
            else:
                return False
        except:
            return False

    # DB table -> Hole die brauchenden Attribute
    def lookup_moreLookupNeeded_getNeedingAttributes(self, sTable):
        if sTable in self.__odicMapping__Lookup_db2lookup:
            return [sAttrib for sAttrib, lstLookup in self.__odicMapping__Lookup_db2lookup[sTable].items()]
        else:
            return None

    # DB table -> Hole die gebrauchten weiteren Lookups
    def lookup_moreLookupNeeded_getNeededLookups(self, sTableSearch):
        lstResult = []
        for sTable, odic in self.__odicMapping__Lookup_db2lookup.items():
            if sTable == sTableSearch:
                for sAttrib, lstLookup in odic.items():
                    lstResult.append(lstLookup[0])
        if len(lstResult) > 0:
            return lstResult
        else:
            return None

    # Db Table,Attrib -> Hole Lookup Tabelle (Es wird dieses andere Lookup)
    def lookup_moreLookupNeeded_getNeededLookup(self, sTable, sAttrib):
        try:
            return self.__odicMapping__Lookup_db2lookup[sTable][sAttrib][0]
        except:
            return None

    # Get Key Value für Lookup
    def lookup_getKey(self, sTable):
        try:
            return  self.__odicMapping__Lookup_table2key[sTable]
        except:
            return None

    # Is Lookup Attribute Key?
    def lookup_isKey(self, sSheet, sCol):
        sTable, sAttrib = self.lookup_excel2db(sSheet, sCol)
        try:
            if sAttrib == self.__odicMapping__Lookup_table2key[sTable]:
                return True
            else:
                return False
        except:
            return None

    # Is Lookup Attribute Obligate?
    def lookup_isObl(self, sSheet, sCol):
        sTable, sAttrib = self.lookup_excel2db(sSheet, sCol)
        try:
            return self.__odicMapping__Lookup_tableAtr2obl[(sTable,sAttrib)]
        except:
            return False

    # Hole alle Ingest Tabellen und Attribute in denen ein/alle Lookup(s) benutzt werden
    def lookup_getAllUsages_ingest(self, sLupSearch=None):
        dicLup = {}
        for sSheet in self.ingest_getExcelSheets():
            for sCol in self.ingest_getExcelCols(sSheet):
                if self.ingest_isLookup(sSheet, sCol):
                    sLupTab = self.ingest_getLookup(sSheet, sCol)
                    if sLupSearch is None or sLupSearch == sLupTab:
                        if sLupTab not in dicLup:
                            dicLup[sLupTab] = []
                        sTable, sAttr = self.ingest_excel2db(sSheet,sCol)
                        dicLup[sLupTab].append((sTable, sAttr))
        if sLupSearch is None:
            return dicLup
        elif sLupSearch in dicLup:
            return dicLup[sLupSearch]
        else:
            return None

    # Hole alle anderen Lookups in denen ein/alle Lookup(s) benutzt werden
    def lookup_getAllUsages_lookup(self, sLupSearch=None):
        dicLup = {}
        for sLup in self.lookup_moreLookupNeeded_getAllNeedingTables():
            lstAtr = self.lookup_moreLookupNeeded_getNeedingAttributes(sLup)
            for sAtr in lstAtr:
                sLupNeeded = self.lookup_moreLookupNeeded_getNeededLookup(sLup,sAtr)
                if sLupSearch is None or sLupSearch == sLupNeeded:
                    if sLupNeeded not in dicLup:
                        dicLup[sLupNeeded] = []
                    dicLup[sLupNeeded].append((sLup, sAtr))
        if sLupSearch is None:
            return dicLup
        elif sLupSearch in dicLup:
            return dicLup[sLupSearch]
        else:
            return None

    # --- Workflow
    # Normal
    def workflow_getAll(self, **kwargs):
        bLong = kwargs.get("long")
        lst = [(k, v) for k, v in self.__odicMapping__Wf.items()]
        lst.sort(key=lambda x: x[1])
        if bLong is True:
            lst = [v for k, v in lst]
        else:
            lst = [k for k, v in lst]
        return lst

    def workflow_isExisting(self, sWf):
        return sWf in self.workflow_getAll()

    def workflow_long2short(self, sWf_search):
        for sWf_short, sWf_long in self.__odicMapping__Wf.items():
            if sWf_search == sWf_long:
                return sWf_short
        return None

    def workflow_short2long(self, sWf):
        if sWf in self.__odicMapping__Wf:
            return self.__odicMapping__Wf[sWf]
        else:
            return None

    # Converter: Import
    def workflow_converter_import_isExisting(self, sWf):
        return sWf in self.workflow_converter_import_getAll()

    def workflow_converter_import_getAll(self, **kwargs):
        bLong = kwargs.get("long")
        bLong = True if bLong is True else False
        lst = [(k, v[0], v[1]) for k, v in self.__odicMapping__Wf_converter_import.items()]
        lst.sort(key=lambda x: x[1])
        if bLong:
            lst = [self.workflow_converter_import_short2long(k) for k, v, v2 in lst]
        else:
            lst = [k for k, v, v2 in lst]
        return lst

    def workflow_converter_import_short2long(self, sWf):
        if sWf in self.__odicMapping__Wf_converter_import:
            return self.__odicMapping__Wf_converter_import[sWf][0]
        else:
            return None

    def workflow_converter_import_getTarget(self, sWf):
        if sWf in self.__odicMapping__Wf_converter_import:
            return self.__odicMapping__Wf_converter_import[sWf][1]
        else:
            return None

    # Converter: Export
    def workflow_converter_export_isExisting(self, sWf):
        return sWf in self.__odicMapping__Wf_converter_export

    def workflow_converter_export_getAll(self, **kwargs):
        bLong = kwargs.get("long")
        bLong = True if bLong is True else False
        lst = [(k, v) for k, v in self.__odicMapping__Wf_converter_export.items()]
        lst.sort(key=lambda x: x[1])
        if bLong:
            lst = [self.workflow_converter_export_short2long(k) for k, v in lst]
        else:
            lst = [k for k, v in lst]
        return lst

    def workflow_converter_export_short2long(self, sWfconf):
        return self.__odicMapping__Wf_converter_export.get(sWfconf)


    # --- Taxon ---

    # Hole alle Excel Sheets
    def taxon_getExcelSheets(self):
        return list(self.__odicMapping__Taxon_excel2db.keys())

    # Hole alle DB Tabellen
    def taxon_getDbTables(self):
        return list(self.__odicMapping__Taxon_db2excel.keys())

    # Excel sheet -> db table
    def taxon_sheet2table(self, sSheet):
        setSheet = {t[0] for t in self.__odicMapping__Taxon_excel2db[sSheet].values()}
        if len(setSheet) == 1:
            return setSheet.pop()
        else:
            return None  # // Fehler: Es waren mehr als eine Datenbank Tabelle in der Lookup Definition

    # DB table -> Excel Sheet
    def taxon_table2sheet(self, sTable):
        setTable = {t[0] for t in self.__odicMapping__Taxon_db2excel[sTable].values()}
        if len(setTable) == 1:
            return setTable.pop()
        else:
            return None  # // Fehler: Es waren mehr als eine Datenbank Tabelle in der Lookup Definition

    # Hole alle Colums eines Excel Sheets
    def taxon_getExcelCols(self, sSheet):
        try:
            return list(self.__odicMapping__Taxon_excel2db[sSheet].keys())
        except:
            return None

    # Hole alle Colums eines Excel Sheets
    def taxon_getDbAttributes(self, sTable):
        try:
            return list(self.__odicMapping__Taxon_db2excel[sTable].keys())
        except:
            return None

    # Excel Sheet,Col -> Db Table,Attrib
    def taxon_excel2db(self, sSheet, sCol):
        try:
            return self.__odicMapping__Taxon_excel2db[sSheet][sCol]
        except:
            return (None, None)

    # Db Table,Attrib -> Excel Sheet,Col
    def taxon_db2excel(self, sTable, sAttrib):
        try:
            return self.__odicMapping__Taxon_db2excel[sTable][sAttrib]
        except:
            return (None, None)

    # Hole Obligatinfo zu Sheet,Column
    def taxon_isObl(self, sSheet, sCol):
        sTable, sAttrib = self.taxon_excel2db(sSheet, sCol)
        if (sTable, sAttrib) in self.__odicMapping_Taxon_db2obl:
            return self.__odicMapping_Taxon_db2obl[(sTable, sAttrib)]
        else:
            return False

    # Hole Key Value für Taxon Table
    def taxon_getKey(self, sTable):
        try:
            return self.__odicMapping__Taxon_table2key[sTable]
        except:
            return None

    # Ist Taxon Table Attribute ein Key?
    def taxon_isKey(self, sSheet, sCol):
        sTable, sAttrib = self.taxon_excel2db(sSheet, sCol)
        try:
            if sAttrib == self.__odicMapping__Taxon_table2key[sTable]:
                return True
            else:
                return False
        except:
            return None

    # --- Ingest ---

    # Excel sheet -> db table
    def ingest_sheet2table(self, sSheet):
        for (sS, sT) in self.__lstMapping__Ingest_ExcelSheet2dbTable:
            if sS == sSheet:
                return sT
        return None

    # DB table -> Excel sheet
    def ingest_table2sheet(self, sTable):
        for (sS, sT) in self.__lstMapping__Ingest_ExcelSheet2dbTable:
            if sT == sTable:
                return sS
        return None

    # Hole alle Sheets (pro Workflow)
    def ingest_getExcelSheets(self, sWf=None):
        try:
            if sWf is None:
                lstSheets = []
                for (sSheet, sCol), v in self.__odicMapping__Ingest_ExcelSheetCol2wf.items():
                    if sSheet not in lstSheets:
                        lstSheets.append(sSheet)
                return lstSheets
            else:
                lstSheets = []
                for (sSheet, sCol), v in self.__odicMapping__Ingest_ExcelSheetCol2wf.items():
                    if sSheet not in lstSheets and sWf in v:
                        lstSheets.append(sSheet)
                return lstSheets
        except:
            return None

    # Hole alle Columns eines Sheet (pro Workflow)
    def ingest_getExcelCols(self, sSheet, sWf=None):
        try:
            if sWf is None:
                return list(sC for (sS,sC), lstWf in self.__odicMapping__Ingest_ExcelSheetCol2wf.items() if sS == sSheet)
            else:
                lstCol = []
                for (sS,sC), lstWf in self.__odicMapping__Ingest_ExcelSheetCol2wf.items():
                    if sS == sSheet:
                        # if (self.ingest_isFk(sS,sC) or self.ingest_isFk(sS,sC)): # Immer PK/FK
                        #     lstCol.append(sC)
                        if sWf in lstWf: # Check Workflow
                            lstCol.append(sC)
                return lstCol
        except Exception as ex:
            return None

    # Hole alle DB-Tabellen
    def ingest_getDbTables(self):
        return list(self.__odicMapping__Ingest_DbTable2Attrib.keys())

    # Hole alle Attribute einer Tabelle
    def ingest_getDbAttributes(self, sTable):
        try:
            return list(self.__odicMapping__Ingest_DbTable2Attrib[sTable])
        except:
            return None

    # Hole table,attribute
    def ingest_excel2db(self, sSheet, sCol):
        try:
            return self.__odicMapping__Ingest_ExcelSheetCol2dbTableAttrib[sSheet, sCol]
        except:
            return None, None

    # Db Table,Attrib -> Excel Sheet,Col
    def ingest_db2excel(self, sTable, sAttrib):
        try:
            return self.__odicMapping__Ingest_DbTableAttrib2excelSheetCol[(sTable,sAttrib)]
        except:
            return (None, None)

    # Ist Lookup-Tabelle?
    def ingest_hasLookup(self, sSheetTest):
        for sSheet, sCol in self.__odicMapping__Ingest_ExcelSheetCol2dbLup:
            if sSheetTest == sSheet:
                return True
        return False

    # Hole alle Lookup Cols
    def ingest_getLookupCols(self, sSheetTest):
        lstCol = []
        for sSheet, sCol in self.__odicMapping__Ingest_ExcelSheetCol2dbLup:
            if sSheetTest == sSheet:
                lstCol.append(sCol)
        return lstCol

    # Ist Lookup-Tabelle?
    def ingest_isLookup(self, sSheet, sCol):
        if (sSheet,sCol) in self.__odicMapping__Ingest_ExcelSheetCol2dbLup:
            return True
        else:
            return False

    # Hole Lookup-Tabelle
    def ingest_getLookup(self, sSheet, sCol):
        try:
            return self.__odicMapping__Ingest_ExcelSheetCol2dbLup[sSheet, sCol]
        except:
            return None

    # Ist Link?
    def ingest_isLink(self, sSheet, sCol):
        if (sSheet, sCol) in self.__odicMapping__Ingest_ExcelSheetCol2link:
            return True
        else:
            return False

    # Hole alle FK/PK-Paarungen
    def ingest_getFkPkPartnerAll(self, sWf=None):
        lstSheets = self.ingest_getExcelSheets(sWf)
        odicFk = OrderedDict()
        for sSheetFk, sSheetPk in self.__lstMapping__Ingest_FkPk:
            if (sSheetFk in lstSheets) and (sSheetPk in lstSheets):
                odicFk[sSheetFk] = sSheetPk
        return odicFk

    # Hole PK zu FK
    def ingest_getFkPkPartner(self, sFkSheet, sWf=None):
        try:
            return self.ingest_getFkPkPartnerAll(sWf)[sFkSheet]
        except:
            return None

    # Hole alle FK-Benutzungen zu den PK
    def ingest_getPkUseAll(self, sWf=None):
        lstSheets = self.ingest_getExcelSheets(sWf)
        odicPk = OrderedDict()
        for sSheetFk, sSheetPk in self.__lstMapping__Ingest_FkPk:
            if (sSheetPk in lstSheets) and (sSheetFk in lstSheets):
                if sSheetPk not in odicPk:
                    odicPk[sSheetPk] = [sSheetFk]
                else:
                    odicPk[sSheetPk].append(sSheetFk)
        return odicPk

    # Hole alle benutzenden FK zu einem PK
    def ingest_getPkUse(self, sPkSheet, sWf=None):
        try:
            return self.ingest_getPkUseAll(sWf)[sPkSheet]
        except:
            return None

    # Hole alle Primary Keys eines Sheets
    def ingest_getPk(self, sSheet, sWf=None):
        try:
            # PK für vollen Workflow
            # Info:
            # PK-System funktioniert nur wenn keine PK-Teile im Subworkflow glöscht werden
            # Wird ein PK-Teil entfernt ist er nicht mehr einmalig -> Kein PK
            if sSheet not in self.__odicMapping__Ingest_ExcelSheet2pk:
                return []
            lstPkMaxWf = list(self.__odicMapping__Ingest_ExcelSheet2pk[sSheet])
            if sWf is None:
                # Alle PKs
                return lstPkMaxWf
            else:
                # Hole PKs
                lstCols = []
                for sCol in self.ingest_getExcelCols(sSheet,sWf):
                    if sCol in self.__odicMapping__Ingest_ExcelSheet2pk[sSheet]:
                        lstCols.append(sCol)
                # Teste ob ein Teil fehlt -> Nichts
                for sPk in lstPkMaxWf:
                    if sPk not in lstCols:
                        return None
                return lstCols
        except Exception as ex:
            print("Internal error getting PK: ",ex)
            return None

    # Hole alle Foreign Keys eines Sheets
    def ingest_getFk(self, sSheet, sWf=None):
        try:
            if sWf is None:
                return list(self.__odicMapping__Ingest_ExcelSheet2fk[sSheet])
            else:
                lstCols = []
                for sCol in self.ingest_getExcelCols(sSheet, sWf):
                    if self.ingest_isFk(sSheet, sCol):
                        lstCols.append(sCol)
                return lstCols
        except:
            return None

    # Hole FK Main Info
    def ingest_getFkMain(self, sSheet):
        if sSheet in self.__odicMapping__Ingest_ExcelSheet2fkMain:
            return self.__odicMapping__Ingest_ExcelSheet2fkMain[sSheet]
        else:
            return (None, None, None )

    # Hole PK Info
    def ingest_isPk(self, sSheet, sCol, sWf=None):
        lstPks = self.ingest_getPk(sSheet, sWf)
        if (lstPks is not None) and (sCol in lstPks):
            return True
        else:
            return False

    # Hole FK Info
    def ingest_isFk(self, sSheet, sCol):
        if sSheet in self.__odicMapping__Ingest_ExcelSheet2fk and sCol in self.__odicMapping__Ingest_ExcelSheet2fk[sSheet]:
            return True
        else:
            return False

    # Hole FK Main Info
    def ingest_isFkTarget(self, sSheet, sCol):
        if (sSheet in self.__odicMapping__Ingest_ExcelSheet2fkMain) and (self.__odicMapping__Ingest_ExcelSheet2fkMain[sSheet][0] == sCol):
            return True
        else:
            return False

    # Hole Date Info
    def ingest_isDate(self, sSheet, sCol):
        if sSheet in self.__odicMapping__Ingest_ExcelSheet2date and sCol in self.__odicMapping__Ingest_ExcelSheet2date[sSheet]:
            return True
        else:
            return False

    # Hole Time Info
    def ingest_isTime(self, sSheet, sCol):
        if sSheet in self.__odicMapping__Ingest_ExcelSheet2time and sCol in \
                self.__odicMapping__Ingest_ExcelSheet2time[sSheet]:
            return True
        else:
            return False

    # Hole Workflows
    def ingest_getWf(self, sSheet, sCol):
        try:
            return self.__odicMapping__Ingest_ExcelSheetCol2wf[sSheet, sCol]
        except:
            return None

    # Hole Obligatinfo zu Sheet,Column
    def ingest_isObl(self, sSheet, sCol):
        if (sSheet,sCol) in self.__odicMapping__Ingest_ExcelSheetCol2obl:
            return self.__odicMapping__Ingest_ExcelSheetCol2obl[(sSheet,sCol)]
        else:
            return False

    # Ist Sheet,Colum Geoinfo
    def ingest_isGeo(self, sSheet, sCol):
        if (sSheet, sCol) in self.__odicMapping__Ingest_ExcelSheetCol2geo:
            return True
        else:
            return False

    # Hole Geoinfo zu Sheet,Column
    def ingest_getGeo(self, sSheet, sCol):
        if (sSheet, sCol) in self.__odicMapping__Ingest_ExcelSheetCol2geo:
            return self.__odicMapping__Ingest_ExcelSheetCol2geo[(sSheet, sCol)]
        else:
            return None

    # --- Comments

    # Hole Kommentar zu Sheet/Column
    # TODO: Hierzu muss der das System Online sein und die Datenbank frqagen
    def ingest_getComment(self, sSheet, sCol):
        if (sSheet, sCol) in self.__dicMapping__Ingest_ForgeComments:
            tpForge = self.__dicMapping__Ingest_ForgeComments[(sSheet, sCol)]
        sComment = "Kommentar"
        return sComment

        # Hole alle Sheets pro Converter


