# coding=utf-8

import openpyxl
from collections import Counter, OrderedDict, namedtuple
import csv, time, psycopg2, warnings, traceback
from tools.coding import Result, WorkingObject, DataTools
from tools.excel import ExcelTools
from tools.misc import JsonConverter
from tools.text import PrettyText
from tools.excel import ExcelTools
from tools.parallel import Stoppable

# -------------------------------------
# Rwgiona (Hält die Regions Daten)
# -------------------------------------

class Regions(Stoppable, WorkingObject):

    # Konstruktor
    def __init__(self, model):
        Stoppable.__init__(self)
        WorkingObject.__init__(self)

        self.__model = model
        self.__smartprinter = model.ui_getSmartprinter()
        self.__db = self.__model.subsystem_getDb()

        # Daten
        self.__sLeftover_name = "non-FAO"
        self.__iSamples_total = 0

        # Init
        self.reset()
        self.setWorking(True)

    # To String
    def __str__(self):
        return self.info()

    # ----------------------------------------------------------------------------------------------------

    # Info
    def info(self):
        sOut = "*** Status of Region System\n\n"
        sConn = self.__model.subsystem_getDb().conn_getConnectionInfo()
        bConnected = self.__model.subsystem_getConnector().db_isConnected()
        sOut+= "Server: "+ str(sConn)+"\n"
        sOut+= "Connected: "+ ("yes" if bConnected else "no")+"\n\n"
        if bConnected:
            sOut += "* Geometries (regional polygons)\n"
            sOut += "Buffered: " + str(self.regions_getPufferSize_shape()) + " shapes with " + str(self.regions_getPufferSize_shape_points()) + " points\n\n"
            sOut += "* Leftover\n"
            sOut += "Category name: " + str(self.conf_getLeftover_name()) + "\n\n"
            if self.regions_getAll_nr(leftover=False) > 0:
                sOut += "* Geometries (regional polygons)\n"
                sOut += self.info_regions()
                sOut += "\n"
            sOut += self.info_map(short=True)
        return sOut

    # -----------------------------------------------------------------------------

    # Lösche Region System
    def reset(self, lookup=None):
        self.__iSamples_unmapped = 0
        self.__iSamples_leftover = 0
        self.__puffer_shape = {}
        self.__odicRegions = OrderedDict()
        self.__lstSamples_unmapped_datasets = []
        self.__dicMap_region_name2id = {}
        self.__odicMap_region = OrderedDict()

    # Reload
    def reload(self):
        res = Result(name="Reload region system")
        self.setReady(False)
        # Regions All
        resRegions = self.__db.fao_regions_getRegions()
        if resRegions.isErr():
            res.addResult(resRegions)
            return res
        self.__odicRegions = resRegions.getData("regions")
        # Map
        resMap = self.__db.fao_map_getData()
        if resMap.isErr():
            res.addResult(resMap)
            return res
        self.__iSamples_total = resMap.getData("sample_nr")
        self.__iSamples_unmapped = resMap.getData("unmapped_nr")
        self.__iSamples_leftover = resMap.getData("leftover_nr")
        self.__lstSamples_unmapped_datasets = resMap.getData("unmapped")
        self.__dicMap_region_name2id = resMap.getData("region_dic_name2id")
        self.__odicMap_region = resMap.getData("region_dic")
        if res.isOk():
            self.setReady(True)
        return res

    # Leftover Name
    def conf_getLeftover_code(self):
        return self.__sLeftover_name.lower()
    def conf_getLeftover_name(self):
        return self.__sLeftover_name

    # Levels (all)
    def regions_getLevels(self):
        setLevels = set()
        for sRegion, (id, code, level, points) in self.__odicRegionsitems():
            setLevels.add(level)
        return sorted(list(setLevels))
    def regions_getLevels_nr(self):
        return len(self.regions_getLevels())

    # Regions (all)
    def regions_getAll(self, **kwargs):
        bLeftover = kwargs["leftover"] if "leftover" in kwargs else True
        lstRegions_sorted = sorted([it for it in self.__odicRegions.keys() if it != self.conf_getLeftover_name()])
        if bLeftover:
            lstRegions_sorted.insert(0, self.conf_getLeftover_name())
        return lstRegions_sorted
    def regions_getAll_nr(self, **kwargs):
        bLeftover = kwargs["leftover"] if "leftover" in kwargs else True
        return len(self.regions_getAll(leftover=bLeftover))
    def regions_getInfo(self, sRegion):
        if sRegion in self.__odicRegions:
            return self.__odicRegions[sRegion]
        else:
            return None, None, None, None

    # Map Data
    def map_getSamples_all_nr(self):
        return self.__iSamples_total
    def map_getSamples_unmapped_nr(self):
        return self.__iSamples_unmapped
    def map_getSamples_leftover_nr(self):
        return self.__iSamples_leftover

    def map_getSamples_unmapped_datasets(self):
        return self.__lstSamples_unmapped_datasets.copy()

    def map_getSamples_leftover_proz(self):
        fProz_leftover = 0
        try:
            fProz_leftover = self.map_getSamples_leftover_nr() / self.map_getSamples_all_nr() * 100
            if 0 < fProz_leftover < 1:
                fProz_leftover = round(fProz_leftover, 2)
            else:
                fProz_leftover = int(fProz_leftover)
        except:
            pass
        return fProz_leftover

    def map_getSamples_unmapped_proz(self):
        fProz_new = 0
        try:
            fProz_new = self.map_getSamples_unmapped_nr() / self.map_getSamples_all_nr() * 100
            if 0 < fProz_new < 1:
                fProz_new = round(fProz_new, 2)
            else:
                fProz_new = int(fProz_new)
        except:
            pass
        return fProz_new

    def map_hasLeftover(self):
        return self.__sLeftover_name in self.__dicMap_region_name2id.keys()

    def map_getRegions(self, **kwargs):
        bLeftover = kwargs["leftover"] if "leftover" in kwargs else True
        lstRegs = list(it for it in self.__dicMap_region_name2id.keys() if it != self.conf_getLeftover_name())
        if bLeftover and self.map_hasLeftover():
            lstRegs.insert(0, self.__sLeftover_name)
        return lstRegs

    def map_getRegions_nr(self, **kwargs):
        bLeftover = kwargs["leftover"] if "leftover" in kwargs else True
        return len(self.map_getRegions(**kwargs))

    # Shapes
    def regions_getPufferSize_shape(self):
        return len(self.__puffer_shape)
    def regions_getPufferSize_shape_points(self):
        iPoints = 0
        for sReg, (sShape, iSize) in self.__puffer_shape.items():
            iPoints += iSize
        return iPoints
    def regions_getShape(self, sRegion):
        if sRegion in self.__puffer_shape:
            # Puffer
            return self.__puffer_shape[sRegion]
        else:
            # Db
            resShape = self.__db.fao_regions_getShape(sRegion)
            if resShape.isErr():
                return None, None
            shape = resShape.getData("shape")
            iSize = resShape.getData("shape-size")
            # Save in Puffer
            self.__puffer_shape[sRegion] = (shape, iSize)
            return shape, iSize

    # ---

    # (Re)Create FAO regions
    def regions_create(self, **kwargs):
        res = Result()
        resFaoClean = self.__model.subsystem_getDb().fao_map_clean()
        if resFaoClean.isErr():
            res.addResult(resFaoClean)
        lstLevel = kwargs.get("level_new")
        bMirror = kwargs.get("mirror")
        resFaoNew = self.__model.subsystem_getDb().fao_regions_create(level=lstLevel, leftover=self.conf_getLeftover_name(), mirror=bMirror)
        res.addData("title", "Report on FAO region")
        if resFaoNew.isErr():
            res.addResult(resFaoNew)
        else:
            sLog = "*** Report on: FAO region creation\n\n"
            sLog += str(resFaoNew.getData("log"))
            res.addData("text", sLog)
            res.addData("server_reconnect", True)
        return res

    # ---

    # Import FAO regions from file
    def regions_import(self, sFile):
        res = Result()
        resFaoClean = self.__model.subsystem_getDb().fao_map_clean()
        if resFaoClean.isErr():
            res.addResult(resFaoClean)
        resFaoImport = self.__model.subsystem_getDb().fao_regions_imExport(sFile, "import")
        if resFaoImport.isErr():
            res.addResult(resFaoImport)
        res.addData("import", True)
        res.addData("file", sFile)
        if res.isOk():
            sLog = "*** Report on: FAO region import\n\n"
            sLog += str(resFaoImport.getData("log"))
            res.addData("log", sLog)
            res.addData("server_reconnect", True)
        return res

    # Export FAO regions to file
    def regions_export(self, sFile):
        res = Result()
        resFaoExport = self.__model.subsystem_getDb().fao_regions_imExport(sFile, "export")
        if resFaoExport.isErr():
            res.addResult(resFaoExport)
        res.addData("export", True)
        res.addData("file", sFile)
        if res.isOk():
            sLog = "*** Report on: FAO region export\n\n"
            sLog += str(resFaoExport.getData("log"))
            res.addData("log", sLog)
        return res

    # Map: Update
    def map_update(self):
        res = Result()
        sLog = ""
        bFaoUpdate_idle = False
        bServer_reconnect = False
        resFao = self.__model.subsystem_getDb().fao_map_update()
        if resFao.isErr():
            res.addResult(resFao)
        else:
            sLogFao = resFao.getData("log")
            bFaoUpdate_idle = resFao.getData("idle")
            if sLogFao is not None:
                sLog += sLogFao + "\n\n"
        if bFaoUpdate_idle is not True:
            bServer_reconnect = True
            # Statistik wenn Samples neu gemapped wurden
        res.addData("title", "FAO region update")
        res.addData("text", sLog)
        res.addData("server_reconnect", bServer_reconnect)
        return res

    # Map: Clean
    def map_clean(self):
        res = Result()
        resFao = self.__model.subsystem_getDb().fao_map_clean()
        if resFao.isErr():
            res.addResult(resFao)
        iDel = resFao.getData("del")
        sLog = "*** Clean FAO region mapping\n\n"
        sLog += "Removed " + str(iDel) + " samples from mapping"
        res.addData("title", "Clean FAO region mapping")
        res.addData("text", sLog)
        res.addData("server_reconnect", True)
        return res

    # ---

    # Info
    def info_regions(self):
        sOut = ""
        "* Regions (" + str(self.regions_getAll_nr(leftover=False)) + ")\n"
        lstTable = [["Region", "Code", "Level", "Size (points)"]]
        for sRegion in self.regions_getAll(leftover=False):
            iId, sCode, sLevel, iSize = self.regions_getInfo(sRegion)
            lstTable.append((sRegion, sCode, sLevel, iSize))
        sOut += PrettyText.table(lstTable) + "\n"
        return sOut

    def info_map(self, **kwargs):
        bShort = kwargs["short"] if "short" in kwargs else False

        iRegions = self.map_getRegions_nr(leftover=False)
        iLeftover = self.map_getSamples_leftover_nr()
        fProz_leftover = self.map_getSamples_leftover_proz()
        iNew = self.map_getSamples_unmapped_nr()
        fProz_new = self.map_getSamples_unmapped_proz()

        sLog = ""
        sLog_region_overview = ""
        sLog_region = ""
        sLog_new = ""
        sLog_leftover = ""
        sLog_region_datasets = ""

        # Mapped Samples
        lstTable_regions_overview = [("Region", "Size (pts)", "Samples")]
        for iId, (sReg, sCode, sLevel, iPoints, lstDat, iCnt) in self.__odicMap_region.items():
            if sCode != self.conf_getLeftover_code().lower():
                # Overview
                lstTable_regions_overview.append((sReg, iPoints, iCnt))
                if not bShort:
                    # Datasets
                    sLog_region_datasets += "* " + str(sReg) + "\n"
                    lstTable_regions = [("Dataset", "Samples")]
                    lstTable_regions += lstDat
                    sLog_region_datasets += PrettyText.table(lstTable_regions) + "\n\n"
            else:
                # Leftover
                lstTable_leftover = [("Dataset", "Samples")]
                lstTable_leftover += lstDat
                sLog_leftover += PrettyText.table(lstTable_leftover) + "\n\n"

        # New Samples
        if iNew > 0:
            lstTable_new = [("Dataset", "Samples")]
            lstTable_new += self.map_getSamples_unmapped_datasets()
            sLog_new += PrettyText.table(lstTable_new) + "\n\n"

        # Overview
        sLog_region_overview += PrettyText.table(lstTable_regions_overview) + "\n\n"

        # *** Log zusammenbauen
        if bShort:
            sLog = ""
            sLog += "------------------------------------\n"
            sLog += "Status of region mapping\n"
            sLog += "------------------------------------\n\n"
            if iNew > 0:
                sLog += "* Unmapped samples: " + str(iNew) + "x " + str(fProz_new) + "%\n"
            if iLeftover > 0:
                sLog += "* Leftover: " + str(iLeftover) + "x " + str(fProz_leftover) + "%\n"
            if iRegions > 0:
                sLog += "* Used Regions: " + str(iRegions) + "x\n"
                sLog += sLog_region_overview
        else:
            if iNew > 0:
                sLog += "------------------------------------\n"
                sLog += "New/Unmapped samples (" + str(iNew) + "x " + str(fProz_new) + "%)\n"
                sLog += "------------------------------------\n\n"
                sLog += sLog_new
            if iLeftover > 0:
                sLog += "------------------------------------\n"
                sLog += "Leftover (" + str(iLeftover) + "x " + str(fProz_leftover) + "%)\n"
                sLog += "------------------------------------\n\n"
                sLog += sLog_leftover
            if iRegions > 0:
                sLog_region += sLog_region_overview
                sLog_region += sLog_region_datasets
                sLog += "------------------------------------\n"
                sLog += "Used regions (" + str(iRegions) + "x)\n"
                sLog += "------------------------------------\n\n"
                sLog += sLog_region

        return sLog
