# coding=utf-8

import openpyxl
from collections import Counter, OrderedDict
import os, sys, csv, getpass, platform, datetime, time, copy, re
import openpyxl.styles.numbers
import warnings
from tools.coding import Result, WorkingObject, DataTools
from tools.excel import ExcelTools, ExcelNavigator
from tools.text import PrettyText
from tools.parallel import Stoppable
from tools.printing import SmartPrinter
from tools.misc import JsonConverter
from tools.geocruncher import GeoCruncher
from pathlib import Path

# -------------------------------------------------------------
# Import: Tamer: Excel File -> Korrigiere Formatprobleme
# -------------------------------------------------------------

class Tamer(Stoppable, WorkingObject):

    # Konstruktor
    def __init__(self, model):
        Stoppable.__init__(self) # Nichts besonderes: Hier wird nur die Unterbrechbarkeit im Thread vererbt
        WorkingObject.__init__(self)

        self.__model = model
        self.__smartprinter = model.ui_getSmartprinter() # Intelligente Printing Modul (CLI/GUI
        self.__map = model.subsystem_getMapper()

    # Info: Hier wird nur Informaiton zum Systemstatus ausgegeben
    def info(self):
        sOut = "*** Status of Excel Tamer System\n\n"
        # Zu untersuchende Sheets

        lst = (
            ("Ingest", self.__map.ingest_getExcelSheets, self.__map.ingest_getExcelCols),
            ("Lookup", self.__map.lookup_getExcelSheets, self.__map.lookup_getExcelCols),
            ("Ingest", self.__map.taxon_getExcelSheets, self.__map.taxon_getExcelCols)
        )

        for sSys, fktSheets, fktCols in lst:
            lstTableAliasesSheets = [["Sheet", "Aliases"]]
            lstTableAliasesCols = [["Sheet","Col","Aliases"]]
            for sSheet in fktSheets():
                # Aliases Sheet
                lstAliasSheet = self.__map.all_getAliasesSheet(sSheet)
                if lstAliasSheet is not None:
                    lstTableAliasesSheets.append((sSheet, PrettyText.lst2strC(lstAliasSheet)))
                # Aliases Col
                for sCol in fktCols(sSheet):
                    lstAliasCol = self.__map.all_getAliasesCol(sSheet, sCol)
                    if lstAliasCol is not None:
                        lstTableAliasesCols.append((sSheet,sCol,PrettyText.lst2strC(lstAliasCol)))

            sOut += "* "+sSys +"\n\n"
            sOut += "--- Aliases for Sheet Names\n"
            sOut += PrettyText.table(lstTableAliasesSheets) + "\n\n"
            sOut += "--- Aliases for Column Names\n"
            sOut += PrettyText.table(lstTableAliasesCols) +"\n\n"

        return sOut

    # Import Eingang
    def startTaming(self, wb, **kwargs):

        """
        Args:
            wb (workbook): Excel Workbook (Modul: OpenPyxl)
        Kwargs:
            result_in: Result in (Objekt mit Warnungen und Informationen gesammelt aus verherigen Modulen)
            ingest (bool): Ingest (Ingest Sheets zähmen)
            lookup (bool): Lookup (Lookup Sheets zähmen)
            taxon (bool): Taxon (TaxonWizard Sheets zähmen)
            batch (str): Batch Info die weitergegeben wird
        Return
            Result
                log (str)
                tamed (int) Bezähmte Einheiten (Sheets/Columns)
                unknown (int) Unzähmbare Einheiten (Sheets/Columns)
        """

        # Alles GUI Methoden
        pn = self.__smartprinter.normal
        pv = self.__smartprinter.verbose
        workflow = self.__smartprinter.workflow
        progress = self.__smartprinter.progress
        status = self.__smartprinter.status
        progress("on-pulse")

        res = Result(name="Excel Tamer") # Objekt in dem Antwortdaten gespeichert werden

        # Parameter
        resultIn = kwargs["result_in"] if "result_in" in kwargs else Result("Empty Result Input Dummy")
        sIngestWf = kwargs["ingest_wf"] if "ingest_wf" in kwargs else None # Arbeitsablauf (Nordsee/Arktis/Antarktis/BSH/...)
        bIngest = True if "ingest" in kwargs and kwargs["ingest"] is True else False
        bLookup = True if "lookup" in kwargs and kwargs["lookup"] is True else False
        bTaxon = True if "taxon" in kwargs and kwargs["taxon"] is True else False

        # Parameter Test
        if not bIngest and not bLookup and not bTaxon:
            res.addErr("Nothing to do ...", sys="Excel Tamer", kat="Idle")
        if wb is None:
            res.addErr("No Excel Workbook given", sys="Excel Tamer", kat="I/O")
        # ---

        pn("* Taming Excel Workbook")
        workflow(top="ExcelTamer", sub="Initialisation", fmt=False)
        status()

        # Log
        bLog = False
        sLog = "************************************\n"
        sLog += "Taming Excel Workbook\n"
        sLog += "************************************\n\n"
        if sIngestWf is not None:
            sLog += "Ingest Workflow: " + str(sIngestWf) +"\n\n"

        # Init
        lstExcludeSheets = ["Glossary","glossary"] # Excel-Sheets die ausgenommen sind von zähmen
        lstSystems = ["ingest","lookup","taxon"] # Alle System werden benutzt

        # Gezähmte Sheets/Columns
        odicTamedSheets = OrderedDict()
        odicTamedCols = OrderedDict()
        iTamed = 0

        # Unzähmbare Sheets/Columns
        iUntamed = 0
        lstUntamedSheets = []
        odicUntamedCols = OrderedDict()
        for sSystem in lstSystems:
            odicTamedSheets[sSystem] = []
            odicTamedCols[sSystem] = OrderedDict()
            odicUntamedCols[sSystem] = OrderedDict()

        # Durch den EXCEL-Tamer hinzugefügte/ehemals fehlende Sheets/Columns
        iAddedSheets = 0
        iAddedCols = 0
        iAddedSheetsBlocked = 0
        iAddedColsBlocked = 0

        # *** Phase 1: Aliase auflösen

        # Liste der zu untersuchende Sheets zusammenstellen
        lstSheetsToWorkWith = []
        if bIngest:
            lstSheetsToWorkWith += self.__map.ingest_getExcelSheets(sIngestWf) # Hole aus Mapper-Modul alle Namen der Ingest-Sheets
        if bLookup:
            lstSheetsToWorkWith += self.__map.lookup_getExcelSheets() # Hole aus Mapper-Modul alle Namen der Lookup-Sheets
        if bTaxon:
            lstSheetsToWorkWith += self.__map.taxon_getExcelSheets() # Hole aus Mapper-Modul alle Namen der TaxonWizard-Sheets

        # 1) Sheets
        for sheet in wb: # Alle Sheets der EXCEL-Tabelle
            sSheet = sheet.title # Name des aktuellen Sheets
            if not self.__map.all_isSheetExisting(sSheet) and sSheet not in lstExcludeSheets: # Gibt es das Sheet nicht und ist es nicht ausgeklammert
                sSheetProper = self.__map.all_getProperSheetName(sSheet) # Hier wird geprüft, ob ein Alias erkannt wurde im Mapper
                if sSheetProper is not None:
                    # Tamed (Sheet wurde gezähmt: Es kann auch der ursprüngliche Name sein wenn kein Alias angewendet wurde)
                    sSystem = self.__map.all_getSheetSystem(sSheetProper) # Ermittle aus dem Sheet das entsprechende System (ingest/lookup/taxon(
                    sTxt = sSheet + " -> " + sSheetProper # Merke dir das gezähmte Sheet für das Log
                    odicTamedSheets[sSystem].append(sTxt)
                    sheet.title = sSheetProper # <- Verändere den Sheetnamen
                    res.addInfo("Tamed sheets names", add="->" + sSheetProper, sheet=sSheet, sys="Excel Tamer", kat="Tamed") # Log: Info in Result-Objekt
                    iTamed += 1
                else:
                    # Untamed (Sheet konnte nicht gezähmt werden)
                    lstUntamedSheets.append(sSheet)
                    res.addInfo("Untamed sheet names", sheet=sSheet, sys="Excel Tamer", kat="Untameable")

        # 2) Columns
        for sSheet in lstSheetsToWorkWith: # Alle Sheet-Name, die geprüft werden sollen durchlaufen
            sSystem = self.__map.all_getSheetSystem(sSheet) # Ermittle System des aktuellen Sheet
            if sSheet in wb.sheetnames: # Wenn es Sheet in EXCEL Workbook gibt? -> weiter
                header = wb[sSheet][1] # Erste Zeile des betrefenden Sheets
                for cell in header: # Duchlaufe alle Felder der ersten Zeile
                    if cell.value is not None:
                        sCol = str(cell.value)
                        bMiss = False
                        sColProper = None
                        if not self.__map.all_isSheetColExisting(sSheet, sCol): # Kennt der Mapper die Kombination Sheet.Column nicht?
                            bMiss = True
                            sColProper = self.__map.all_getProperColName(sSheet, sCol) # Suche nach Aliasen
                        if bMiss:
                            if sColProper is not None: # Alias gefunden
                                # Tamed (Gezähmt)
                                sTxt = sCol + " -> " + sColProper # Info merken
                                if sSheet not in odicTamedCols[sSystem]:
                                    odicTamedCols[sSystem][sSheet] = []
                                odicTamedCols[sSystem][sSheet].append(sTxt)
                                cell.value = sColProper # <- Verändere die Column
                                res.addInfo("Tamed column names", add="->"+sColProper, sheet=sSheet, col=sCol, sys="Excel Tamer", kat="Tamed")
                                iTamed += 1
                            else:
                                # Untamed (Ungezähmt)
                                if sSheet not in odicUntamedCols[sSystem]:
                                    odicUntamedCols[sSystem][sSheet] = []
                                odicUntamedCols[sSystem][sSheet].append(sCol)
                                res.addInfo("Untamed column names",sheet=sSheet, col=sCol, sys="Excel Tamer", kat="Untameable")


        # Log: Hier werden alle Veränderungen protokolliert
        sLog += "------------------------------------\n"
        sLog += "Phase 1)\nRenaming stubborn sheets names\nand columns names using\naliases, typos and case insensivity\n"
        sLog += "------------------------------------\n\n"
        bLog = False
        # Loop
        for sSystem in lstSystems:
            if (len(odicTamedCols[sSystem]) + len(odicUntamedCols[sSystem]) +len(odicTamedCols[sSystem])) > 0:
                bLog = True
                sLog += sSystem.upper() + "\n"
                if len(odicTamedSheets[sSystem]) > 0:
                    sLog += "--- Tamed sheets ---\n"
                    for sTmp in odicTamedSheets[sSystem]:
                        sLog += sTmp + "\n"
                        iTamed += 1
                if len(odicTamedCols[sSystem]) > 0:
                    sLog += "--- Tamed column names ---\n"
                    for sSheet, lstCols in odicTamedCols[sSystem].items():
                        for sCol in lstCols:
                            sLog += sSheet+": "+sCol + "\n"
                        iTamed += 1
                if len(odicUntamedCols[sSystem]) > 0:
                    sLog += "--- Untameable column names ---\n"
                    for sSheet, lstCols in odicUntamedCols[sSystem].items():
                        for sCol in lstCols:
                            sLog += sSheet+": " +sCol + "\n"
                        iUntamed += 1
        if len(lstUntamedSheets) > 0:
            bLog = True
            sLog += "\n"
            sLog += "UNKNOWN SYSTEM\n"
            sLog += "--- Untameable sheets ---\n"
            if len(lstUntamedSheets) > 0:
                for sTmp in lstUntamedSheets:
                    sLog += sTmp + "\n"
                    iUntamed += 1
        if not bLog:
            sLog += "... nothing to do ...\n"
        sLog +="\n"

        # *** Phase 2: Füge fehlende Header hinzu
        # Log
        bLog = False
        sLog += "------------------------------------\n"
        sLog += "Phase 2)\nAdding missing sheets names\nand column names (if possible)\n"
        sLog += "------------------------------------\n\n"
        lstAddMissingSheetsBlocked = []
        lstAddMissingSheets = []
        lstAddMissingColsBlocked = []
        lstAddMissingCols = []

        # *** Fehlende Sheets
        for sSheet in lstSheetsToWorkWith: # Durchlaufe alle Sheets, die gefprüft werden sollen
            if sSheet not in wb.sheetnames: # Findet man ein Sheet nicht im EXCEL Worbook?
                bLog = True
                # Hier kommt jetzt mal ein kleiner Trick:

                if len(lstUntamedSheets) > 0:
                    # Untameable Sheetnames ausser Glossary
                    # Achtung: Gab es unzähmbare Sheets (Ausgenommen der Ausnahmen wie z.B. "Glossary"
                    # Wenn ja, kann man das fehlende Sheet nicht ergänzen
                    # Denn ein ungezähmtes Sheet könnte ein falschgeschriebene Variante des Sheets sein.
                    # Wenn man das Sheet nun leer ergänzen würde, könnte Datenverlust entstehen.
                    lstAddMissingSheetsBlocked.append(str(sSheet) + " -> blocked by: "+PrettyText.lst2strC(lstUntamedSheets))
                    iAddedSheetsBlocked += 1
                    res.addInfo("Blocked adding missing sheet", add="blocked by: "+PrettyText.lst2strC(lstUntamedSheets), sheet=sSheet, sys="Excel Tamer",
                                kat="Blocked Adding Missing")
                else:
                    # Adding Sheet mit cols
                    # Alles OK: Ergänze das fehlende Sheet mit allen dazugehörigen Columns
                    res.addInfo("Added missing sheet with columns", sheet=sSheet, sys="Excel Tamer", kat="Adding Missing")
                    sSystem = self.__map.all_getSheetSystem(sSheet)
                    lstColsWanted = self.__map.all_getExcelCols(sSheet, sIngestWf).copy()
                    lstAddMissingSheets.append(str(sSheet))
                    wb.create_sheet(title=sSheet)
                    wb[sSheet].append(lstColsWanted)
                    lstHeader = [cell.value for cell in wb[sSheet][1]]
                    iAddedSheets += 1
                    try:
                        iAddedCols += len(lstHeader)
                    except:
                        pass

        # *** Fehlende Columns
        for sheet in wb: # Durchlaufe alle Sheets in EXCEL Workbook
            sSheet = sheet.title
            if self.__map.all_isSheetExisting(sSheet): # Wird Sheet.Column erkannt?
                sSystem = self.__map.all_getSheetSystem(sSheet) # Ermittle das System
                lstColsWanted = self.__map.all_getExcelCols(sSheet, sIngestWf).copy() # Hole alle validen Columns zu Sheet
                lstColsFound = [cell.value for cell in sheet[1] if cell.value is not None]  # Alle erkannten Columns in EXCEL-Workbook
                try:
                    lstUntamedCols = odicUntamedCols[sSystem][sSheet]
                except:
                    lstUntamedCols = []
                sLogSheetAdd = ""
                resCmp = DataTools.cmpLists(lstColsFound, lstColsWanted) # In dieser statischen Klassenmethode wird einfach ermittelt welche Colums fehlen
                lstColsMissing = resCmp.getData("istFehlt")
                if lstColsMissing is not None:
                    # Es fehlen Columns
                    bLog = True
                    if len(lstUntamedCols) > 0:
                        # Es exisiteren unbezähmbare Columns im Sheet? -> Ergänzung nicht möglich ohne potentiellen Datenverlust (siehe analoge Problem oben: Sheet)
                        iAddedColsBlocked += len(lstColsMissing)
                        lstAddMissingColsBlocked.append(str(sSheet)+": "+PrettyText.lst2strC(lstColsMissing)+" -> blocked by: "+PrettyText.lst2strC(lstUntamedCols))
                        res.addInfo("Blocked adding missing columns in sheet", add=PrettyText.lst2strC(lstColsMissing)+" -> blocked by: "+PrettyText.lst2strC(lstUntamedCols), sheet=sSheet, sys="Excel Tamer", kat="Blocked Adding Missing")
                    else:
                        # Alle fehlenden Columns können ergänzt werden
                        iColNew = len(sheet[1]) + 1 # Letzer Column Index in Header
                        lstAddMissingCols.append(str(sSheet)+": "+str(PrettyText.lst2strC(lstColsMissing)))
                        # Add New Col
                        res.addInfo("Added missing columns in sheet", add=PrettyText.lst2strC(lstColsMissing), sheet=sSheet, sys="Excel Tamer", kat="Adding Missing")
                        for sCol in lstColsMissing:
                            sheet.cell(column=iColNew, row=1, value=sCol) # <- Neue Column an das Ende der Headerzeile anhängen
                            iColNew += 1
                        iAddedCols += len(lstColsMissing)

        # Log
        if len(lstAddMissingSheets) > 0:
            sLog += "--- Added missing sheets ---\n"
            for sMsg in lstAddMissingSheets:
                sLog += sMsg +"\n"
            sLog += "\n"
        if len(lstAddMissingSheetsBlocked) > 0:
            sLog += "--- Warning: Blocked adding missing sheets (because untameable sheets are found in workbook) ---\n"
            for sMsg in lstAddMissingSheetsBlocked:
                sLog += sMsg + "\n"
            sLog += "\n"
        if len(lstAddMissingCols) > 0:
            sLog += "--- Added missing columns ---\n"
            for sMsg in lstAddMissingCols:
                sLog += sMsg + "\n"
            sLog += "\n"
        if len(lstAddMissingColsBlocked) > 0:
            sLog += "--- Warning: Blocked adding missing columns (because non tameable columns are found in sheet) ---\n"
            for sMsg in lstAddMissingColsBlocked:
                sLog += sMsg + "\n"
            sLog += "\n"
        if not bLog:
            sLog += "... nothing to do ...\n"
        sLog += "\n"

        # *** Final Check
        # Hier wird nocheinmal das virtuelle Workbook auf Vollständigkeit geprüft
        bLog = False
        sLog += "------------------------------------\n"
        sLog += "Final Checkup\n"
        sLog += "------------------------------------\n\n"

        # Log
        if iTamed + iUntamed + iAddedSheets + iAddedCols + iAddedSheetsBlocked + iAddedColsBlocked >0:
            sLog += "--- Actions ---\n"
            if iTamed >0:
                sLog += "Tamed sheet and column names: " +str(iTamed) +"\n"
            if iUntamed>0:
                sLog += "Untameable sheet and column names: " +str(iUntamed) +"\n"
            if iAddedSheets > 0:
                sLog += "Added sheets: " + str(iAddedSheets)+"\n"
            if iAddedCols > 0:
                sLog += "Added columns: " + str(iAddedCols)+"\n"
            if iAddedSheetsBlocked > 0:
                sLog += "Blocked adding sheets: " + str(iAddedSheetsBlocked)+"\n"
            if iAddedColsBlocked > 0:
                sLog += "Blocked adding columns: " + str(iAddedColsBlocked)+"\n"
            sLog += "\n"

        sLog += "--- Result ---\n"

        # 1) Überprüfe alle Sheets
        bMissingSheets = False
        resCmp = DataTools.cmpLists(wb.sheetnames, lstSheetsToWorkWith)
        lstSheetsMissing = resCmp.getData("istFehlt")
        if lstSheetsMissing is not None:
            sLog += "Missing Sheets in workbook: " + str(PrettyText.lst2strC(lstSheetsMissing)) +"\n"
            bMissingSheets = True
        else:
            sLog += "Sheets: all found\n"

        # 2) Überprüfe alle Columns
        bMissingCols = False
        for sSheet in lstSheetsToWorkWith:
            try:
                sheet = wb[sSheet]
                sSystem = self.__map.all_getSheetSystem(sSheet)
                lstColsWanted = self.__map.all_getExcelCols(sSheet, sIngestWf).copy()
                lstColsFound = [cell.value for cell in sheet[1] if cell.value is not None]
                resCmp = DataTools.cmpLists(lstColsFound, lstColsWanted)
                lstColsMissing = resCmp.getData("istFehlt")
                if lstColsMissing is not None:
                    sLog += "Missing columns in sheet " + str(sSheet)+ ": "+ PrettyText.lst2strC(lstColsMissing) +"\n"
                    bMissingCols = True
            except:
                pass

        # Finales LOG
        if not bMissingCols:
            sLog += "Columns: all found\n"
        if bMissingCols or bMissingSheets:
            sLog += "Workbook: Fail, not 100% tameable -> problems may occurr\n"
            res.addWarn("Workbook is not 100% tameable -> problems may occurr", sys="Excel Tamer", kat="Final Checkup")
        else:
            sLog += "Workbook: 100% tamed and Ok\n"
            #res.addInfo("Excel Workbook is 100% tamed and Ok", sys="Excel Tamer", kat="Final Checkup")

        sLog += "\n"

        # Ergebnis Objekt erzeugen
        res.addData("log", sLog) # Log speichern
        res.addData("ok", not bMissingCols and not bMissingCols) # Bool OK: Ja/Nein speichern
        return res



