# coding=utf-8


# -------------------------------------------------
# TaxonWizard
# -------------------------------------------------

import openpyxl, traceback, pickle, zipfile
from collections import Counter, OrderedDict
import urllib, os, sys, csv, getpass, platform, datetime, time, copy, random, psycopg2, psycopg2.extras
import openpyxl.styles.numbers
import warnings
import requests
from psycopg2 import sql
from tools.parallel import Stoppable
from tools.coding import Result, DataTools, ProgressString, WorkingObject
from tools.text import PrettyText
from tools.printing import SmartPrinter
from tools.excel import ExcelTools
from tools.fileio import FileTools
from shutil import copyfile
from tools.taxon import Taxon
from tools.ui import Colors
from tools.taxacruncher import TaxaCruncher

class TaxonWizard(Stoppable, WorkingObject):

    def __init__(self, model):
        Stoppable.__init__(self)
        WorkingObject.__init__(self)

        # -- Member
        self.__model = model
        self.__smartprinter = model.ui_getSmartprinter()
        self.__map = model.subsystem_getMapper()
        self.__myworms = model.subsystem_getMyWorms()

        # Special Taxa
        self.__dicTaxaSpecial = {-1: "Artifact" }

        # Biota: ÄhnlichkeitsWert Scientificnames
        self.__fScientificNameMatchRatio = 0.9

        # DB-Namen der Admin Tables
        self.__odicAdminTables = OrderedDict([
            ("Base", "wormsbase"),
            ("Aaid","wormsaaid"),
             ("Private","wormsprivate"),
             ("Colony","wormscolony"),
             ("Change","wormschange"),
             ("Out", "noworms")
        ])

        # Admin Tables
        self.__dicWormsBase = {}
        self.__dicWormsAaid = {}
        self.__dicWormsPrivate = {}
        self.__dicWormsColony = {}
        self.__dicWormsChange = {}
        self.__dicNoWorms = {}

        # Taxon List Aid
        self.__lstTaxonListAid = []
        self.__lstTaxonListAaid = []

        # Taxon List Headers
        self.__dicTaxonListHeader = {}

    # ----------------------------------------

    # Info
    def info(self):
        sOut = "*** Status of TaxonWizard\n\n"
        return sOut


    def __reset(self):
        self.abort(False)
        self.__dicWormsBase = {}
        self.__dicWormsAaid = {}
        self.__dicWormsPrivate = {}
        self.__dicWormsColony = {}
        self.__dicNoWorms = {}
        self.__lstTaxonListAid = []
        self.__lstTaxonListAaid = []

    # Ermittle alle benötigten Aphia Ids
    def __gatheringNeededAphiaIds(self):
        res = Result()

        # Sammel benötigte Aid aus anderen Tabellen ein
        try:
            setTaxonAid = set(self.__lstTaxonListAid)
            setWormsChange = set(self.__dicWormsChange)
            setWormsColony = set(self.__dicWormsColony)
            setWormsPrivate = set([v["aid"] for sScientificname,v in self.__dicWormsPrivate.items()])
            setWormsBase = set(self.__dicWormsBase)
            setWormsAaid = set()
            for iAid, v in self.__dicWormsAaid.items():
                setWormsAaid.add(iAid) # Aid
                setWormsAaid.add(v["aaid"]) # Aaid
            setAll = set().union(setWormsBase,setWormsAaid,setWormsPrivate,setWormsColony,setWormsChange)
        except Exception as ex:
            res.addErr("Unexpected Error", add=str(ex), kat="Gathering Aphia Ids from other Admin Tables", sys="TaxonWizard")
            return res

        # Bestimme fehlende Aid
        setTaxonMissing = set(setAll).difference(setTaxonAid)

        # Antwort
        res.addData("base", setWormsBase)
        res.addData("aaid", setWormsAaid)
        res.addData("private", setWormsPrivate)
        res.addData("colony", setWormsPrivate)
        res.addData("change", setWormsChange)
        res.addData("taxon-list-missing", setTaxonMissing)
        res.addData("taxon-list", setTaxonAid)
        res.addData("all", setAll)
        return res

    # Checke Ingest-Daten
    def __checkImportData(self, excelFile, lstTabs=None):
        pv = self.__smartprinter.verbose
        res = Result(name="Check Taxon Admin Import Bulk Data")

        # --- Check: Sheets
        pv("- Checking sheets")
        lstTaxonSheets = self.__model.subsystem_getMapper().taxon_getExcelSheets()
        lstExcelNames = excelFile.sheetnames
        lstSheetFound = []
        lstSheetMissing = []
        lstSheetsImport = []
        for sSheet in lstTaxonSheets:
            if sSheet in lstExcelNames:
                lstSheetFound.append(sSheet)
                if (len(lstTabs) == 0) or (sSheet in lstTabs):
                    lstSheetsImport.append(sSheet)
            else:
                lstSheetMissing.append(sSheet)
        pv("Allowed sheets: " + PrettyText.lst2strC(lstTaxonSheets))
        pv("Found allowed sheets in file: " + PrettyText.lst2strC(lstSheetFound))
        if len(lstTabs) > 0:
            pv("Only use these sheets: " + PrettyText.lst2strC(lstTabs))
        pv("These Sheets will be imported: " + PrettyText.lst2strC(lstSheetsImport))
        res.addData("sheetsImport", lstSheetsImport)

        # Vergleiche alle Attribute mit Headern aus entsprechendem Sheet
        dicHeader = {}
        dicHeaderIndex = {}
        for sSheet in lstSheetsImport:
            resExcelHeader = ExcelTools.getHeadIndex(excelFile[sSheet])
            lstExcelHeader = resExcelHeader.getData("header")
            dicExcelHeaderIndex = resExcelHeader.getData("index")
            if resExcelHeader.isErr() or lstExcelHeader is None:
                res.addResult(resExcelHeader)
                return res
            dicHeader[sSheet] = lstExcelHeader
            dicHeaderIndex[sSheet] = dicExcelHeaderIndex
            resCheck = DataTools.cmpLists(lstExcelHeader, self.__model.subsystem_getMapper().taxon_getExcelCols(sSheet))
            if resCheck.getData("istFehlt") is not None:
                res.addErr("Header not found in sheet", sheet=str(sSheet), add=PrettyText.lst2strC(resCheck.getData("istFehlt")), kat="Check Import", sys="TaxonWizard")
            if resCheck.getData("istDoppelt") is not None:
                res.addErr("Found double headers in sheet", sheet=str(sSheet), add=PrettyText.lst2strC(resCheck.getData("istDoppelt")), kat="Check Import", sys="TaxonWizard")
            if resCheck.getData("istPlus") is not None:
                pv("Found surplus headers in sheet", sheet=str(sSheet), add=PrettyText.lst2strC(resCheck.getData("istPlus")), kat="Check Import", sys="TaxonWizard")
        res.addData("header", dicHeader)
        res.addData("headerIndex", dicHeaderIndex)
        # Bestimme Anzahl der gesamt einzulesenden Zeilenanzahl
        pv("- Getting sum of new entries")
        iRowsTotal = 0
        for sSheet in lstSheetsImport:
            pv(sSheet + ": " + str(excelFile[sSheet].max_row - 1))
            iRowsTotal += excelFile[sSheet].max_row - 1
        pv("total: " + str(iRowsTotal))
        res.addData("entriesTotal", iRowsTotal)
        return res

    # Minimaler Ähnlichkeitswert zweier ScientificNames
    def getScientificNameMatchRatio(self):
        return self.__fScientificNameMatchRatio

    # Special Taxa
    def getAllSpecialTaxa(self):
        return self.__dicTaxaSpecial.copy()
    def isSpecialTaxon(self, iAidSpecial):
        if iAidSpecial in self.__dicTaxaSpecial:
            return True
        else:
            return False
    def getSpecialTaxonName(self, iAidSpecial):
        if iAidSpecial in self.__dicTaxaSpecial:
            return self.__dicTaxaSpecial[iAidSpecial]
        else:
            return None


    def __logBaseTaxa_changes(self, **kwargs):
        """
        Protokolliere Veränderungen bei Base Taxa
        Kwargs:
            cursor: cursor
        Args:
            new (bool): New Base Taxon
            delete (bool): Delete Base Taxon
            aid (lst): Liste mit AphiaIds
            taxa (lst): Liste mit Taxa
        """
        res = Result("Log in new taxa to TaxonWizard base list")

        pn = self.__smartprinter.normal
        pv = self.__smartprinter.verbose
        pd = self.__smartprinter.debug
        progress = self.__smartprinter.progress
        workflow = self.__smartprinter.workflow
        status = self.__smartprinter.status

        # Parameter
        cursor = kwargs["cursor"] if "cursor" in kwargs else None
        bNew = kwargs["new"] if "new" in kwargs and (kwargs["new"] in (True,False)) else False
        bDel = kwargs["delete"] if "delete" in kwargs and (kwargs["delete"] in (True,False)) else False
        lstAid = kwargs["aid"] if "aid" in kwargs else None
        lstTaxa = kwargs["taxa"] if "taxa" in kwargs else None

        # Test: Mode entweder: New oder Del
        if bNew is bDel:
            res.addErr("Mode has to either: new or del", kat="Log in new base taxa", sys="TaxonWizard")
            return res

        # Test: Entweder Liste mit AphiaIds oder mit Taxa übergeben
        if (lstTaxa is None) is (lstAid is None):
            res.addErr("Either Taxa or List of AphiaIds should be given", kat="Log in new base taxa", sys="TaxonWizard")
            return res

        sLog = ""
        if lstAid is not None:
            # AphiaIds
            lstAid = list(set(lstAid))
            lstTaxa = []
            if len(lstAid) > 0:
                if bNew:
                    workflow(top="TaxonWizard", sub="New Base Taxa - Asking MyWoRMS", fmt=False)
                    progress("on-pulse")
                    pn("* TaxonWizard: New Base Taxa: Asking MyWoRMS", fmt=False)
                else:
                    workflow(top="TaxonWizard", sub="Del Base Taxa - Asking MyWoRMS", fmt=False)
                    progress("on-pulse")
                    pn("* TaxonWizard: Del Base Taxa: Asking MyWoRMS", fmt=False)
                iErr = iWarn = iInfo = 0
                for iNr, iAid in enumerate(lstAid):
                    resMyWorms = self.__model.subsystem_getMyWorms().getTaxon(iAid)
                    if resMyWorms.isAnyMessage():
                        res.addResult(resMyWorms, msg_only=True)
                    iErr += resMyWorms.getErrNo()
                    iWarn += resMyWorms.getWarnNo()
                    iInfo += resMyWorms.getInfoNo()
                    progress(iNr, len(lstAid), 100, proz=True, abs=True, cli=True, err=iErr, warn=iWarn, info=iInfo)
                    if resMyWorms.isOk() and resMyWorms.getData("found"):
                        t = resMyWorms.getData("taxon")
                        lstTaxa.append(t)
                        status("Asking MyWoRMS: " + str(t.signature()), fmt=False)
                    else:
                        res.addResult(resMyWorms, msg_only=True)
                sLog = self.__model.subsystem_getMyWorms().getStatisticProblemTaxa(lstTaxa)
                self.__model.subsystem_getMyWorms().puffer_save()
        else:
            # Taxa
            sLog = self.__model.subsystem_getMyWorms().getStatisticProblemTaxa(lstTaxa)

        # Action Log
        sAction = "tax-base-new" if bNew else "tax-base-del"
        result_actionlog = self.__model.subsystem_getDb().actionlog_new(sAction, "wormsbase", sLog, cursor=cursor)
        if result_actionlog.isErr():
            res.addResult(result_actionlog, msg_only=True)

        # Result
        res.addResult(result_actionlog)
        res.addData("log", sLog)
        res.addData("taxa", lstTaxa)
        return res

    def delBaseTaxa(self, **kwargs):
        """
        Löschen von Base Taxa aus Base List
        Kwargs:
            taxa (lst): Liste mit Taxon Objekten
            aphia (lst): Liste mit AphiaIds
            ids (lst): Liste mit Ids (der Einträge in der tabelle wormsbase)
            cursor: Db Cursor
        Returns:
             Result
        """

        # Helper: Aufräumen
        def cleanUi():
            workflow()
            progress("off")

        def closeDb():
            if self.__model.conf_isDry() and conn is not None:
                conn.rollback()
            elif res.isErr():
                conn.rollback()
            else:
                conn.commit()
            cursor.close()
            conn.close()

        def id2aphia():
            resTmp = Result()
            lstAphiaTmp = []
            try:
                sqlCmd = sql.SQL(
                    "SELECT aid FROM {schema}.{table} WHERE id IN ({ids})").format(
                    schema=sql.Identifier(sDbSchema),
                    table=sql.Identifier("wormsbase"),
                    ids=sql.SQL(', ').join(map(sql.Literal, lstIds_final))
                )
                cursor.execute(sqlCmd)
                lstHeader = [head[0] for head in cursor.description]
                lstContent = cursor.fetchall()
                lstAphiaTmp = [row[0] for row in lstContent]
            except Exception as ex:
                resTmp.addErr("Problems getting AphiaIds for Basa Taxa to delete", add=str(ex), kat="Delete Base Taxa", sys="TaxonWizard")
            resTmp.addData("aphia", lstAphiaTmp)
            return resTmp

        def aphia2tax():
            resTmp = Result()
            iErr = iWarn = iInfo = 0
            lstTaxaTmp = []
            for iNr, iAid in enumerate(lstAphia_final):
                resMyWorms = self.__model.subsystem_getMyWorms().getTaxon(iAid)
                if resMyWorms.isAnyMessage():
                    res.addResult(resMyWorms, msg_only=True)
                iErr += resMyWorms.getErrNo()
                iWarn += resMyWorms.getWarnNo()
                iInfo += resMyWorms.getInfoNo()
                progress(iNr, len(lstAphia_final), 100, proz=True, abs=True, cli=True, err=iErr, warn=iWarn, info=iInfo)
                if resMyWorms.isOk() and resMyWorms.getData("found"):
                    t = resMyWorms.getData("taxon")
                    lstTaxaTmp.append(t)
                    status("Asking MyWoRMS: " + str(t.signature()), fmt=False)
                else:
                    resTmp.addResult(resMyWorms, msg_only=True, kat="Delete Base Taxa", sys="TaxonWizard")
            resTmp.addData("taxa", lstTaxaTmp)
            return resTmp

        def delTax():
            resTmp = Result()
            try:
                sqlCmd = sql.SQL(
                    "DELETE FROM {schema}.{table} WHERE id IN ({ids})").format(
                    schema=sql.Identifier(sDbSchema),
                    table=sql.Identifier("wormsbase"),
                    ids=sql.SQL(', ').join(map(sql.Literal, lstIds_final))
                )
                cursor.execute(sqlCmd)
            except Exception as ex:
                resTmp.addErr("Problems deleting Base Taxa", add=str(ex), kat="Delete Base Taxa", sys="TaxonWizard")
            return resTmp

        # ------------------------------

        pn = self.__smartprinter.normal
        pv = self.__smartprinter.verbose
        pd = self.__smartprinter.debug

        progress = self.__smartprinter.progress
        status = self.__smartprinter.status
        workflow = self.__smartprinter.workflow

        self.abort(False)
        res = Result(name="Del Base Taxa")

        # Parameter
        cursor = kwargs["cursor"] if "cursor" in kwargs else None
        if cursor is None:
            bOwnConnection = True
        else:
            bOwnConnection = False
        lstTaxa = kwargs["taxa"] if "taxa" in kwargs else None
        lstAid = kwargs["aphia"] if "aphia" in kwargs else None
        lstId = kwargs["ids"] if "ids" in kwargs else None


        # Test: Mutual Exclusion (Taxa, Aid, Id)
        bOk = False
        for lst in (lstTaxa,lstAid,lstId):
            if (lst is not None):
                if bOk is True:
                    bOk = False
                    break
                else:
                    bOk = True
        if not bOk:
            res.addErr("Taxa source has to be either a list of: Taxa, AphiaIds or Ids", sys="TaxonWizard", kat="Del Base Taxa")
            return res

        # Einzeldaten zu Listen machen
        if lstTaxa is not None and type(lstTaxa) == Taxon:
            lstTaxa = [lstTaxa]
        if lstAid is not None and type(lstAid) == int:
            lstAid = [lstAid]
        if lstId is not None and type(lstId) == int:
            lstTaxa = [lstTaxa]

        # Eigene DB Connection
        if bOwnConnection:
            resConn = self.__model.subsystem_getDb().conn_getConnection()
            if resConn.isErr():
                res.addResult(resConn)
                cleanUi()
                return res
            # Connection: An
            conn = resConn.getData("connection")
            cursor = resConn.getData("cursor")

        # Info
        pn("* TaxonWizard : Delete Base Taxa")
        workflow(top="TaxonWizard", sub="Delete Base Taxa", fmt=False)
        # Daten
        sLog = ""
        sDbSchema = self.__model.subsystem_getDb().account_active_getSchema()
        iRecs = 0
        iDel = 0
        iDelFail = 0
        lstIds_final = []
        lstAphia_final = []
        lstTaxa_final = []
        if lstId is not None:
            # *** List Ids
            if res.isOk():
                # --- Hole Aphia Ids
                workflow(top="TaxonWizard", sub="Del Base Taxa - Getting AphiaIds", fmt=False)
                progress("on-pulse")
                pn("* TaxonWizard: Del Base Taxa: Getting AphiaIds", fmt=False)
                lstIds_final = lstId.copy()
                resAphia = id2aphia()
                if resAphia.isErr():
                    res.addResult(resAphia, msg_only=True)
                lstAphia_final = resAphia.getData("aphia")
            if res.isOk():
                # --- Erzeuge Taxa
                workflow(top="TaxonWizard", sub="Del Base Taxa - Ask MyWoRMS", fmt=False)
                progress("on-pulse")
                pn("* TaxonWizard: Del Base Taxa: Asking MyWoRMS", fmt=False)
                resTax = aphia2tax()
                if resTax.isErr():
                    resTax.addResult(resAphia, msg_only=True)
                lstAphia_final = resTax.getData("aphia")
            if res.isOk():
                # --- Delete Taxa
                progress("on-pulse")
                pn("* TaxonWizard: Del Base Taxa", fmt=False)
                resDel = delTax()
                if resDel.isErr():
                    resDel.addResult(resAphia, msg_only=True)
                lstTaxa_final = resTax.getData("taxa")

        elif lstAid is not None:
            # *** List Aid
            for iId in lstId:
                pass

        elif lstTaxa is not None:
            # *** List Taxa
            pass

        # Log: New Taxa in Base List
        if res.isOk():
            resLogDelBase = self.__logBaseTaxa_changes(delete=True, taxa=lstTaxa_final, cursor=cursor)
            if resLogDelBase.isOk():
                sLogWormsbaseTaxa = resLogDelBase.getData("log")
            else:
                res.addResult(resLogDelBase, msg_only=True)

        # Antwort
        res.addData("log", sLog)

        # Aufräumen
        cleanUi()
        if bOwnConnection:
            closeDb()

        # Ende
        return res

    def addBaseTaxa(self, lstTaxa, **kwargs):
        """
        Hinzufügen: Base Taxa zu Base List
        Args:
            lstTaxa (lst): Base Taxa (Liste mit Taxon Objekten)
            quick_update_taxon_list (bool): (Über)Schreibe alle Aphia Ids in die Taxonliste
        Kwargs:
            result_in: Result In
            cursor: Db Cursor
        Returns:
             Result
        """

        # Helper: Aufräumen
        def cleanUi():
            workflow()
            progress("off")

        def closeDb():
            if self.__model.conf_isDry() and conn is not None:
                conn.rollback()
            elif res.isErr():
                conn.rollback()
            else:
                conn.commit()
            cursor.close()
            conn.close()

        pn = self.__smartprinter.normal
        pv = self.__smartprinter.verbose
        pd = self.__smartprinter.debug

        progress = self.__smartprinter.progress
        status = self.__smartprinter.status
        workflow = self.__smartprinter.workflow

        self.abort(False)
        res = Result(name="Add Base Taxa")

        # Parameter
        resultIn = kwargs["result_in"] if "result_in" in kwargs else Result("Empty Result Input Dummy")
        bUpdateTaxonList = kwargs["quick_update_taxon_list"] if ("quick_update_taxon_list" in kwargs and kwargs["quick_update_taxon_list"] in (True, False)) else False
        cursor = kwargs["cursor"] if "cursor" in kwargs else None
        if cursor is None:
            bOwnConnection = True
        else:
            bOwnConnection = False

        # Einzeltaxon zu liste machen
        if type(lstTaxa) == Taxon:
            lstTaxa = [lstTaxa]

        # Test
        if lstTaxa is None:
            res.addErr("No Taxa given", sys="TaxonWizard", kat="Add Base Taxa")
            return res

        # Eigene DB Connection
        if bOwnConnection:
            resConn = self.__model.subsystem_getDb().conn_getConnection()
            if resConn.isErr():
                res.addResult(resConn)
                cleanUi()
                return res
            # Connection: An
            conn = resConn.getData("connection")
            cursor = resConn.getData("cursor")

        # Info
        if bUpdateTaxonList:
            pn("* Add new taxa to Base List and final Taxon List")
        else:
            pn("* Add new taxa to Base List")
        if bOwnConnection:
            pv("Using own connection")
        else:
            pv("Reusing connection")
        workflow(top="TaxonWizard", sub="Add Base Taxa", fmt=False)

        # Daten
        sLog = ""
        sDbSchema = self.__model.subsystem_getDb().account_active_getSchema()
        iQuickAddToBaseList = 0
        iQuickAddToTaxonList = 0

        # *** Add to Base List / Quick Add to Taxon List
        for taxon in lstTaxa:
            res.addInfo("New Taxa", add=str(taxon.getAphiaId()) + ": " + str(taxon.getScientificName()), kat="Add Taxon to Base List", sys="TaxonWizard")
            dicRec = {"aid": taxon.getAphiaId(), "scientificname": taxon.getScientificName()}
            # Quick Add/Update Taxon to in Base List
            try:
                sqlCmd = sql.SQL(
                    "INSERT INTO {schema}.{table} ({attr}) VALUES ({val}) ON CONFLICT (aid) DO UPDATE SET ({attr}) = ({val})").format(
                    schema=sql.Identifier(sDbSchema),
                    table=sql.Identifier("wormsbase"),
                    attr=sql.SQL(', ').join(map(sql.Identifier, dicRec.keys())),
                    val=sql.SQL(', ').join(map(sql.Placeholder, dicRec.keys()))
                )
                cursor.execute(sqlCmd, dicRec)
                sSqlQuery = cursor.query.decode("utf-8")
                iQuickAddToBaseList += 1
            except Exception as ex:
                res.addErr("Creation failed", add=str(ex), kat="Create", sys="Access to Table in Db")
            if bUpdateTaxonList:
                # Quick Add/Update Taxon to Taxon List
                try:
                    sqlCmd = sql.SQL(
                        "INSERT INTO {schema}.{table} ({attr}) VALUES ({val}) ON CONFLICT (aid) DO UPDATE SET ({attr}) = ({val})").format(
                        schema=sql.Identifier(sDbSchema),
                        table=sql.Identifier("taxon"),
                        attr=sql.SQL(', ').join(map(sql.Identifier, dicRec.keys())),
                        val=sql.SQL(', ').join(map(sql.Placeholder, dicRec.keys()))
                    )
                    cursor.execute(sqlCmd, dicRec)
                    sSqlQuery = cursor.query.decode("utf-8")
                    iQuickAddToTaxonList += 1
                except Exception as ex:
                    res.addErr("Creation failed", add=str(ex), kat="Create", sys="Access to Table in Db")

        # Log: New Taxa in Base List
        if len(lstTaxa) > 0:
            resLogNewBase = self.__logBaseTaxa_changes(new=True, taxa=lstTaxa, cursor=cursor)
            if resLogNewBase.isOk():
                sLogWormsbaseTaxa = resLogNewBase.getData("log")
            else:
                res.addResult(resLogNewBase, msg_only=True)

        # *** Log
        if iQuickAddToBaseList > 0:
            sLog += "Added New Taxa to Base List and Taxon List: "+str(iQuickAddToBaseList)+"\n"
            for taxon in lstTaxa:
                sLog+= str(taxon.getAphiaId()) + ": " + str(taxon.getScientificName()) + "\n"

        # Antwort
        res.addData("log", sLog)

        # Aufräumen
        cleanUi()
        if bOwnConnection:
            closeDb()

        # Ende
        return res


    # Lade neue TaxonAdmin Daten ein
    def addAdmTable_excel(self, wb, **kwargs):
        """
        Ingest von Lookup Admin-Tables

        Args:
            wb (workbook): Dateiname der zu importierenden Datei
        Kwargs:
            result_in: Result In
            cursor: Db Cursor
            sheets (lst): Liste aller hinzuzufügenden Sheets
            update (bool): Update Sheets in DB
            batch (str): Batch Info die weitergegeben wird
        Returns:
             Result
        """

        # Helper: Aufräumen
        def cleanUi():
            workflow()
            progress("off")

        def closeDb():
            if self.__model.conf_isDry() and conn is not None:
                conn.rollback()
            elif res.isErr():
                conn.rollback()
            else:
                conn.commit()
            cursor.close()
            conn.close()

        pn = self.__smartprinter.normal
        pv = self.__smartprinter.verbose
        pd = self.__smartprinter.debug

        progress = self.__smartprinter.progress
        status = self.__smartprinter.status
        workflow = self.__smartprinter.workflow

        self.abort(False)
        res = Result(name="Add TaxonWizard Data Batch")
        pn("* Importing TaxonWizard Data")

        # Parameter
        resultIn = kwargs["result_in"] if "result_in" in kwargs else Result("Empty Result Input Dummy")
        lstImportTables = kwargs["sheets"] if "sheets" in kwargs else None
        bUpdate = kwargs["update"] if ("update" in kwargs and kwargs["update"] in [True, "true", "t"]) else False
        cursor = kwargs["cursor"] if "cursor" in kwargs else None
        if cursor is None:
            bOwnConnection = True
        else:
            bOwnConnection = False

        # ---------

        if bOwnConnection:
            pv("Using own connection")
        else:
            pv("Reusing connection")
        pn("Update existing records in Db: " + ("yes" if bUpdate else "no"))
        workflow(top="TaxonWizard", sub="Import Admin - Checking data", fmt=False)

        # --- Check Parameter

        if lstImportTables is None:
            lstImportTables = []
        if wb is None:
            res.addErr("No Excel workbook given", sys="TaxonWizard", kat="Bulk import")
        for sSheet in lstImportTables:
            if sSheet not in self.__model.subsystem_getMapper().taxon_getExcelSheets():
                res.addErr("Sheet not valid", sheet=str(sSheet), sys="TaxonWizard", kat="Bulk import")
        if res.isErr():
            cleanUi()
            return res


        # --- Teste Eingangsdaten
        resCheck = self.__checkImportData(wb, lstImportTables)
        res.addResult(resCheck)
        if res.isErr():
            cleanUi()
            return res

        # --- Starte Ingest
        lstSheets = resCheck.getData("sheetsImport")
        dicHeader = resCheck.getData("header")
        dicHeaderIndex = resCheck.getData("headerIndex")
        iRowsTotal = resCheck.getData("entriesTotal")

        progress("on-pulse")

        # Eigene DB Connection
        if bOwnConnection:
            resConn = self.__model.subsystem_getDb().conn_getConnection()
            if resConn.isErr():
                res.addResult(resConn)
                cleanUi()
                return res
            # Connection: An
            conn = resConn.getData("connection")
            cursor = resConn.getData("cursor")

        # Log Total
        sLogTotal = "------------------------------------\n" \
                    "Detailed logs per Taxon System Table\n" \
                    "------------------------------------\n\n"

        if self.__model.conf_isDry():
            sLogTotal += "Dry mode activated.\n\n"

        lstSheetsUsed = []
        lstTablesChanged = []
        iExcelRowsTotal = 0
        iRowAktuell = 0
        iRecNewTotal = 0
        iRecErrTotal = 0
        iRecUpdateTotal = 0
        iRecUpdateNoTotal = 0
        iRecUpdateNoEqualTotal = 0
        odicRecErrTotal = OrderedDict()

        # Loop Sheets
        for sSheet in lstSheets:
            workflow("TaxonWizard | add "+str(sSheet))
            pn("- Add TaxonWizard Data: " + sSheet)
            status("Add TaxonWizard Admin Table: " + sSheet)
            sTable = self.__map.taxon_sheet2table(sSheet)

            # Erlaubte Header für Taxon Admin Tables
            lstHeadersAllowed = self.__model.subsystem_getMapper().taxon_getExcelCols(sSheet)

            # Log: Single
            lstRecNew = []
            lstRecUpdateDenied = []
            lstRecUpdateDeniedGold = []
            lstRecUpdate = []
            lstRecUpdateNotNecessaryEqual = []
            lstRecErr = []

            # Wormsbase: Sammle alle benutzeten Aphia Ids ein
            setWormsBase_aphiaIds = set()

            # Log
            sLog = ""
            sLogWormsbaseTaxa = ""

            ws = wb[sSheet]
            bSheetEmpty = True
            # Rows
            for iRow, row in enumerate(ws.iter_rows()):
                if iRow > 0:
                    odicRowExcel = ExcelTools.row2dic(row, dicHeaderIndex[sSheet], conv_val_2_str=True)

                    # Entferne: Unerlaubte Header aus Adin Table
                    for sCol in list(odicRowExcel.keys()):
                        if sCol not in lstHeadersAllowed:
                            del odicRowExcel[sCol]

                    # Teste: Leere Zeile?
                    bEmpty = True
                    for col, val in odicRowExcel.items():
                        if val not in [None, ""]:
                            bEmpty = False

                    # Translate Namen: Excel -> DB
                    odicRowDb = OrderedDict()
                    for sCol, sVal in odicRowExcel.items():
                        sTable, sAttribute = self.__model.subsystem_getMapper().taxon_excel2db(sSheet, sCol)
                        odicRowDb[sAttribute] = sVal

                    # # Wenn ein Gold-Attribut existiert -> None -> False
                    if "gold" in odicRowDb:
                        if str(odicRowDb["gold"]).strip().lower() in ["true", "t", "1"]:
                            odicRowDb["gold"] = True
                        else:
                            odicRowDb["gold"] = False

                    if not bEmpty:
                        # Zeile is voll
                        bSheetEmpty = False
                        iExcelRowsTotal += 1

                        # Test Key
                        sTestKey = self.__model.subsystem_getMapper().taxon_getKey(sTable)
                        sTestVal = odicRowDb[sTestKey]

                        # Tweak: Gold -> Immer auf Bool verändern True/False
                        if "gold" in odicRowDb:
                            bGoldIn = True if str(odicRowDb["gold"]).strip().lower() in ["true", "1", "t"] else False
                        else:
                            bGoldIn = None

                        # Test: Eintrag schon in DB?
                        resExist = self.__model.subsystem_getDb().check_isRecordExistingInTable(sTable, sTestKey, dict(odicRowDb), cursor=cursor)
                        if resExist.getData("found"):
                            iId = resExist.getData("id")
                            # --- Record bereits in DB gefunden
                            if resExist.getData("equal"):
                                # --- Record gleich
                                lstRecUpdateNotNecessaryEqual.append(
                                    "#" + str(iRow) + ": " + sTestKey + "=" + str(sTestVal))
                                res.addInfo("No Update - exactly same entry in Db", sheet=sSheet, col=sTestKey,
                                            add=str(sTestVal), row=iRow, sys="TaxonWizard Import", kat="Update Db")
                                iRecUpdateNoEqualTotal += 1
                            else:
                                # --- Record differiert
                                dicDiff = resExist.getData("diff")
                                if "gold" in dicDiff:
                                    bGoldDb = True if str(dicDiff["gold"]).strip().lower() in ["true", "1","t"] else False
                                else:
                                    bGoldDb = None
                                sDiff = ""
                                for kDb, vDb in dicDiff.items():
                                    if kDb in odicRowDb:
                                        vIn = odicRowDb[kDb]
                                    else:
                                        vIn = "-"
                                    sDiff += str(kDb) + "(db=\'" + str(vDb) + "\' in=\'" + str(vIn) + "\') "
                                if bUpdate:
                                    # Update: Erlaubt
                                    # Gold-Test: Not-Gold kann Gold nicht überschreiben
                                    if (bGoldDb is not None and bGoldIn is not None) and (bGoldIn is False and bGoldDb is True):
                                        # Kein Update
                                        sNoUpdate = "#" + str(iRow) + ": " + sTestKey + "=" + str(sTestVal) + ", diff=" + sDiff
                                        lstRecUpdateDeniedGold.append(sNoUpdate)
                                        iRecUpdateNoTotal += 1
                                        res.addWarn("Update denied", sheet=sSheet, col=sTestKey,
                                                add=str(sTestVal) + ", diff:" + sDiff, row=iRow,
                                                sys="TaxonWizard Import",
                                                kat="Update entry in Db denied (not-gold cannot override gold)")
                                    else:
                                        # Update
                                        resUpdate = self.__model.subsystem_getDb().table_updateInTable(sTable, iId, odicRowDb, cursor=cursor)
                                        lstRecUpdate.append( "#" + str(iRow) + ": " + sTestKey + "=" + str(sTestVal) + ", diff=" + sDiff)
                                        res.addInfo("Update entry in Db", sheet=sSheet, col=sTestKey,
                                                    add=str(sTestVal) + ", diff:" + sDiff, row=iRow, sys="TaxonWizard Import",
                                                    kat="Update entry in Db")
                                        iRecUpdateTotal += 1
                                        if resUpdate.isErr():
                                            lstRecErr.append(("#" + str(iRow) + ": " + resUpdate.getResult()))
                                            iRecErrTotal += 1
                                            res.addResult(resUpdate, err=True, sys="TaxonWizard Import", kat="Update entry in Db")
                                else:
                                    # Update: Nicht erlaubt
                                    sNoUpdate = "#" + str(iRow) + ": " + sTestKey + "=" + str(sTestVal) +", diff=" + sDiff
                                    lstRecUpdateDenied.append(sNoUpdate)
                                    res.addInfo("Update denied", sheet=sSheet, col=sTestKey,
                                                add=str(sTestVal) + ", diff:" + sDiff, row=iRow, sys="TaxonWizard Import",
                                                kat="Update entry in Db denied (user deactivated option)")
                                    #iRecUpdateDenied += 1
                                    iRecUpdateNoTotal += 1
                        else:
                            # --- Create new
                            resCreate = self.__model.subsystem_getDb().table_insertIntoTable(sTable, odicRowDb, cursor=cursor, none=False)
                            if resCreate.isErr():
                                lstRecErr.append(("#" + str(iRow) + ": " + resCreate.getResult()))
                                res.addResult(resCreate, err=True, sys="TaxonWizard Import", kat="Create new entry in Db")
                                iRecErrTotal += 1
                            else:
                                sSqlQuery = resCreate.getData("sql")
                                iId = resCreate.getData("id")
                                lstRecNew.append(
                                    "#" + str(iRow) + ": " + sTestKey + "=" + str(sTestVal))
                                iRecNewTotal += 1
                            # *** Neuer Eintrag in WormsBase: Extrahiere AphiaId -> Log New Taxon
                            if sTable == "wormsbase":
                                try:
                                    iAid = int(odicRowDb["aid"])
                                    if iAid >= 0:
                                        setWormsBase_aphiaIds.add(iAid)
                                except:
                                    res.addInfo("AphiaId is no integer", sheet="Base", col="AphiaId", sys="TaxonWizard Import", kat="Base AphiaId check")

                    # Info
                    iRowAktuell += 1
                    # Status
                    lstStatusEWI = list()
                    sStatus = ""
                    if len(lstRecNew) > 0:
                        lstStatusEWI.append("New:" + str(len(lstRecNew)))
                    if len(lstRecUpdate) > 0:
                        lstStatusEWI.append("Upd:" + str(len(lstRecUpdate)))
                    if len(lstRecUpdateNotNecessaryEqual) > 0:
                        lstStatusEWI.append("Eq:" + str(len(lstRecUpdateNotNecessaryEqual)))
                    if len(lstRecUpdateDenied) > 0:
                        lstStatusEWI.append("UpdNo:" + str(len(lstRecUpdateDenied)))
                    if len(lstRecUpdateDeniedGold) > 0:
                        lstStatusEWI.append("UpdNoGold:" + str(len(lstRecUpdateDeniedGold)))
                    if len(lstStatusEWI) != 0:
                        sStatus = "[" + (" ".join(lstStatusEWI)) + "]"
                    status("Reading Sheet: "+sSheet+" #"+str(iRowAktuell)+"/"+str(iRowsTotal)+" "+sStatus)
                    iErr = iRecErrTotal + res.getErrNo()
                    iWarn = res.getWarnNo()
                    iInfo = res.getInfoNo()
                    # Progress
                    progress(iRowAktuell, iRowsTotal, 100, proz=True, abs=True, cli=True, err=iErr, warn=iWarn, info=iInfo, status=sStatus)

                    # Abbruch durch Benutzer
                    if self.isAbort():
                        res.addErr("Aborted by user ...", sys="TaxonWizard", kat="Abort")
                        cleanUi()
                        closeDb()
                        return res

            # Wormsbase: Überprüfe alle Taxa
            if len(setWormsBase_aphiaIds) > 0:
                resLogNewBase = self.__logBaseTaxa_changes(new=True, aid=list(setWormsBase_aphiaIds), cursor=cursor)
                if resLogNewBase.isOk():
                    sLogWormsbaseTaxa = resLogNewBase.getData("log")
                else:
                    res.addResult(resLogNewBase, msg_only=True)

            # Statistik für eine Lookup Table
            sLog += "--- " + str(sTable) + " ---\n"
            sLog += "Update existing records in Db: " + ("yes" if bUpdate else "no") +"\n"
            if len(lstRecNew) > 0 or \
                len(lstRecUpdateNotNecessaryEqual) > 0 or \
                len(lstRecUpdateDenied) > 0 or \
                len(lstRecErr) > 0 or \
                len(lstRecUpdate) > 0:
                # Bearbeitete Sheets
                lstSheetsUsed.append(sSheet)
                if len(lstRecNew) > 0 or len(lstRecUpdate) > 0:
                    # Veränderte Tabellen
                    lstTablesChanged.append(sSheet)
                if len(lstRecNew) > 0:
                    # Neu
                    sLog += "- Created records: " + str(len(lstRecNew)) + "\n"
                    for sTxt in lstRecNew:
                        sLog += sTxt + "\n"
                    sLog += "\n"
                if len(lstRecUpdate) != 0:
                    sLog += "- Updated records in db: " + str(len(lstRecUpdate)) + "\n"
                    for sTxt in lstRecUpdate:
                        sLog += sTxt + "\n"
                    sLog += "\n"
                if len(lstRecUpdateNotNecessaryEqual) != 0:
                    # Update nicht nötig (in==db)
                    sLog += "- Update not necessary (record existing exactly in db: " + str(len(lstRecUpdateNotNecessaryEqual)) + "\n"
                    for sTxt in lstRecUpdateNotNecessaryEqual:
                        sLog += sTxt + "\n"
                    sLog += "\n"
                if len(lstRecUpdateDenied) != 0:
                    sLog += "- Update denied (due to deactivated option): " + str(len(lstRecUpdateDenied)) + "\n"
                    for sTxt in lstRecUpdateDenied:
                        sLog += sTxt + "\n"
                    sLog += "\n"
                if len(lstRecUpdateDeniedGold) != 0:
                    sLog += "- Update denied (not-gold cannot override gold): " + str(len(lstRecUpdateDeniedGold)) + "\n"
                    for sTxt in lstRecUpdateDeniedGold:
                        sLog += sTxt + "\n"
                    sLog += "\n"
                if len(lstRecErr) != 0:
                    # Fehler
                    sLog += "- Failed creations: " + str(len(lstRecErr)) + "\n"
                    for sTxt in lstRecErr:
                        sLog += sTxt + "\n"
                    sLog += "\n"
                # Aktuelles Taxon Log im Action Log speichern
                sActionTaxon = "tax-adm-add"
                result_actionlog = self.__model.subsystem_getDb().actionlog_new(sActionTaxon, sTable, sLog, cursor=cursor)
                res.addResult(result_actionlog)
            else:
                sLog += "... nothing to do\n\n"

            # Aktuelles Taxon Log an gesamtes Log anhängen
            sLogTotal += sLog

            # Fehlerliste an Gesamtfehlerliste herandhängen
            if len(lstRecErr) > 0:
                odicRecErrTotal[sSheet] = lstRecErr

        # Sind Fehler aufgetreten -> Setze diese an Gesamt Log an den Anfang
        if iRecErrTotal > 0:
            sLogTmp = "------------------------------------\n" \
                    " Overview of occurred errors (" + str(iRecErrTotal) + ")\n" \
                    "------------------------------------\n\n"
            for sSheet, lstErr in odicRecErrTotal.items():
                sLogTmp += "- " + str(sSheet) + "\n"
                for sErr in lstErr:
                    sLogTmp += str(sErr) + "\n"
            sLogTotal = sLogTmp + "\n" + sLogTotal

        # Log Präefix
        sLogPrae = "************************************\n"
        sLogPrae += "Import TaxonWizard Data\n"
        sLogPrae += "************************************\n"
        sLogPrae += "\n"
        sLogTotal = sLogPrae + sLogTotal
        sLogPrae += "------------------------------------\n" \
                    "Overview\n" \
                    "------------------------------------\n\n"
        sLogPrae += "Processed Excel sheets: " + PrettyText.lst2strC(lstSheetsUsed) + "\n"
        sLogPrae += "Proccessed EXCEL records: " + str(iExcelRowsTotal) + "\n"
        sLogPrae += "Tables changed: " + PrettyText.lst2strC(lstTablesChanged) + "\n"
        sLogPrae += "Created records in db: " + str(iRecNewTotal) + "\n"
        sLogPrae += "Updated records in db: " + str(iRecUpdateTotal) + "\n"
        sLogPrae += "Denied Updates: " + str(iRecUpdateNoTotal) + "\n\n"
        sLogTotal = sLogPrae + sLogTotal + "\n"

        # Log Total
        res.addData("taxon-excel-sheets-used", lstSheetsUsed)
        res.addData("taxon-tables-changed", lstTablesChanged)
        res.addData("taxon-new", iRecNewTotal)
        res.addData("taxon-update", iRecUpdateTotal)
        res.addData("taxon-equal", iRecUpdateNoEqualTotal)
        res.addData("taxon-update-no", iRecUpdateNoTotal)
        res.addData("taxon-err", iRecErrTotal)
        res.addData("log", sLogTotal)

        # Aufräumen
        cleanUi()
        if bOwnConnection:
            closeDb()
        # Ende
        return res

    def update(self, **kwargs):
        """
        Volles Update der Taxonliste
        Kwargs:
            cursor (cursor): Cursor
        """

        # Helper: Aufräumen
        def cleanUi():
            workflow()
            progress("off")

        def closeDb():
            if self.__model.conf_isDry() and conn is not None:
                conn.rollback()
            elif res.isErr():
                conn.rollback()
            else:
                conn.commit()
            cursor.close()
            conn.close()

        # ---

        pn = self.__smartprinter.normal
        pv = self.__smartprinter.verbose
        pd = self.__smartprinter.debug
        progress = self.__smartprinter.progress
        workflow = self.__smartprinter.workflow
        status = self.__smartprinter.status

        res = Result(name="Update taxon list")
        iZeitStart = time.time()
        sZeitStart = str(datetime.datetime.now().strftime("%Y-%m-%d %H:%M"))
        self.__reset()

        bErr = False
        bWarn = False

        # Parameter
        cursor = kwargs["cursor"] if "cursor" in kwargs else None
        if cursor is None:
            bOwnConnection = True
        else:
            bOwnConnection = False

        # Eigene DB Connection
        if bOwnConnection:
            resConn = self.__model.subsystem_getDb().conn_getConnection()
            if resConn.isErr():
                res.addResult(resConn)
                return res
            conn = resConn.getData("connection")
            cursor = resConn.getData("cursor")

        # *** Zähler

        # --- Check Admin Tables
        iAdminUpdateNames = 0
        lstAdminUpdateNames = [["Admin list", "Aphia Id", "Updated differing names"]]
        iAdminErrors = 0
        lstAdminErrors = [["Admin list", "Aphia Id", "Error"]]

        # --- Log Daten
        # Neues Taxon wegen anderer Aaid
        iTaxonNewAidWegenAaid = 0
        lstTaxonNewAidWegenAaid = [["Aphia Id", "ScientificName"]]
        # Rekursiv ermittelte  Aaid
        lstWormsAaid = [("Aphia Id", "Scientificname", "Accepted Aphia Id", "Accepted ScientificName", "Status", "Trace route)")]
        iTaxonAaid = 0
        # Expert Knowledge: Aaid
        lstExpertAaid = [("Aphia Id", "ScientificName", "Accepted Aphia Id", "Accepted ScientificName")]
        iExpertAaid = 0
        # Expert Knowledge: Is Colony
        lstExpertColony = [["Aphia Id", "ScientificName"]]
        iExpertColony = 0
        # Neues Taxon in Taxonliste erzeugt
        iTaxonNew = 0
        # Update eines vorhanden Taxon in Taxonliste
        iTaxonUpdate = 0
        # Anzahl der geupdateten Felder
        iTaxonUpdateFields = 0
        # Fehler beim Update oder Erzeugen eines Taxon in Taxonliste
        iTaxonError = 0
        lstTaxonWormsError = [("Aphia Id", "Message")]
        # Alle geupdateten Felder
        dicTaxonUpdateFields = {}
        # Alle Updates gelogt
        lstLogUpdates = [("Aphia Id", "ScientificName", "Updates", "Fields", "Differences")]

        # --- Problematische Taxa
        # Probeme Taxon: Namen
        setTaxaProblem_no_name= set()
        setTaxaProblem_no_aname_and_accepted = set()
        # Probleme Taxon: AphiaId
        setTaxaProblem_aphia = set()
        # Probleme Taxon: Status
        setTaxaProblem_status = set()

        # --- Taxa in Limbo
        setTaxaLimbo = set()

        # --- Unakzeptierte Taxa die auf akzeptierte Taxa zeigen, die keine sind
        setTaxaUnaccept_withAaid_leadingToNotacceptetTaxa = set()

        # --- Teste Einmaligkeit einiger Suchbegriffe
        lstUniqPrivateTaxon = [["Private name"]]
        lstUniqNowormsTaxon = [["Scientificname"]]
        lstUniqPrivateNoworms = [["Private name"]]

        # Teste auf mehrfache Scientificnames
        lstAmbigousScientificnames = [["ScientificName", "Aphia Id", "Valid Aphia Id", "Status"]]
        iAmbigousScientificnames = 0
        lstAmbigousScientificnamesAccepted = [["ScientificName", "Aphia Id", "Status"]]
        iAmbigousScientificnamesAccepted = 0

        progress("on-pulse")

        # --------------------------------------------------------------------------------------------------------------
        # Vorarbeiten
        # --------------------------------------------------------------------------------------------------------------

        pn("* TaxonWizard")
        # Check Compatibility to WoRMS
        progstr = ProgressString(prae="Checking compatability to WoRMS", post=None, now=1, max=6)
        workflow(top="TaxonWizard", sub="Checking compatability to WoRMS", fmt=False)
        pn(progstr.cli())
        resCompatibility = Taxon.checkCompatibility2Worms()
        if resCompatibility.isErr():
            res.addResult(resCompatibility, err=True)
            return res

        # Hole alle Admin Tabellen
        progstr.up(prae="Collecting Data", post=None)
        pn(progstr.cli())

        try:
            # WormsBase
            resTable = self.__model.subsystem_getDb().table_getWholeTable("wormsbase", cursor=cursor)
            lstHeader = resTable.getData("header")
            lstContent = resTable.getData("content")
            if resTable.isErr():
                res.addResult(resTable)
                raise NameError("Problems getting Table from database")
            self.__dicWormsBase = DataTools.tab2dic(lstHeader, lstContent, "aid", ["id"]).getData("dic")
            pv("Base List")

            # WormsAaid
            resTable = self.__model.subsystem_getDb().table_getWholeTable("wormsaaid", cursor=cursor)
            lstHeader = resTable.getData("header")
            lstContent = resTable.getData("content")
            if resTable.isErr():
                res.addResult(resTable)
                raise NameError("Problems getting Table from database")
            self.__dicWormsAaid = DataTools.tab2dic(lstHeader, lstContent, "aid", ["id"]).getData("dic")
            pv("Expert Accepted AphiaIds")

            # WormsPrivate
            pv("Expert Private Names")
            resTable = self.__model.subsystem_getDb().table_getWholeTable("wormsprivate", cursor=cursor)
            lstHeader = resTable.getData("header")
            lstContent = resTable.getData("content")
            if resTable.isErr():
                res.addResult(resTable)
                raise NameError("Problems getting Table from database")
            self.__dicWormsPrivate = DataTools.tab2dic(lstHeader, lstContent, "private_name", ["id"]).getData("dic")

            # WormsColony
            resTable = self.__model.subsystem_getDb().table_getWholeTable("wormscolony", cursor=cursor)
            lstHeader = resTable.getData("header")
            lstContent = resTable.getData("content")
            if resTable.isErr():
                res.addResult(resTable)
                raise NameError("Problems getting Table from database")
            self.__dicWormsColony = DataTools.tab2dic(lstHeader, lstContent, "aid", ["id"]).getData("dic")
            pv("Expert Colony Taxa")

            # WormsChange
            resTable = self.__model.subsystem_getDb().table_getWholeTable("wormschange", cursor=cursor)
            lstHeader = resTable.getData("header")
            lstContent = resTable.getData("content")
            if resTable.isErr():
                res.addResult(resTable)
                raise NameError("Problems getting Table from database")

            self.__dicWormsChange = DataTools.tab2dic(lstHeader, lstContent, "aid", ["id"]).getData("dic")
            pv("Expert Changes")

            # NoWorms
            resTable = self.__model.subsystem_getDb().table_getWholeTable("noworms", cursor=cursor)
            lstHeader = resTable.getData("header")
            lstContent = resTable.getData("content")
            if resTable.isErr():
                res.addResult(resTable)
                raise NameError("Problems getting Table from database")
            self.__dicNoWorms = DataTools.tab2dic(lstHeader, lstContent, "scientificname", ["id"]).getData("dic")
            pv("Expert NoWoRMS")
            # Hole Aid from taxon list
            self.__lstTaxonListAid = self.__model.subsystem_getDb().taxon_getAllAidFromTaxonList().getData("aid")
            pv("TaxonList Aphia Ids")
            # Hole Aaid from taxon list
            self.__lstTaxonListAaid = self.__model.subsystem_getDb().taxon_getAllAaidFromTaxonList().getData("aaid")
            pv("TaxonList Accepted Aphia Ids")

        except Exception as ex:
            res.addErr("Error while getting tables", add=str(ex), kat="Update", sys="TaxonWizard")
            progress("off")
            return res

        # --------------------------------------------------------------------------------------------------------------
        # Bearbeite Admin tables
        # --------------------------------------------------------------------------------------------------------------

        progress("on-pulse-abort")
        lstAdmin = [("Expert Accepted AphiaIds", self.__dicWormsAaid, "wormsaaid"),
                    ("Base List", self.__dicWormsBase, "wormsbase"),
                    ("Expert Colony Taxa ", self.__dicWormsColony, "wormscolony"),
                    ("Expert Private Names", self.__dicWormsPrivate, "wormsprivate"),
                    ("Expert Change", self.__dicWormsChange, "wormschange")]

        iNr = 0
        iNrMax = 0
        progstr.up(prae="Processing admin tables", post=None)
        workflow(top="TaxonWizard", sub="Processing admin tables", fmt=False)
        pn(progstr.cli())
        # Summe aller Einträge
        for sAdminTableName, dicAdmin, sDbAdminTableName in lstAdmin:
            iNrMax += len(dicAdmin)
        # Loop: alle Tabellen
        for sAdminTableName, dicAdmin, sDbAdminTableName in lstAdmin:
            pv(sAdminTableName)
            if self.isAbort():
                break
            for key, val in dicAdmin.items():
                taxonWorms = None
                if self.isAbort():
                    break
                # Extrahiere Aphia Id aus Eintrag
                if sAdminTableName == "Expert Private":
                    iAid = val["aid"]
                else:
                    iAid = key
                # Frage Worms: Exisitert der Eintrag bei MyWoRMS
                if iAid > 0:
                    resMyWorms = self.__myworms.getTaxon(iAid)
                    taxonWorms = resMyWorms.getData("taxon")
                    res.addResult(resMyWorms)
                    iNr += 1
                    if resMyWorms.isErr():
                        # Fehler bei Verbindung
                        bErr = True
                        res.addResult(taxonWorms.getResult(), err=True)
                        lstAdminErrors.append([sAdminTableName, iAid, PrettyText.lst2strC(taxonWorms.getResult().getErrLst(short=True))])
                        pv("- Aphia Id: " + str(iAid) + PrettyText.lst2strC(taxonWorms.getResult().getErrLst(short=True)))
                    elif taxonWorms.isNotFound():
                        # Aphia Id Nicht gefunden
                        lstAdminErrors.append([sAdminTableName,iAid,"Aphia Id not found in WoRMS"])
                        sLogErr_ascii = "Aphia Id not found in WoRMS: Table: " + sAdminTableName + ", Aphia Id: " + str(iAid)
                        pv("- "+sLogErr_ascii)
                        bErr = True
                        res.addErr("Aphia Id not found in WoRMS", add= "Table: " + sAdminTableName + ", Aphia Id: " + str(iAid), kat="Update", sys="TaxonWizard")
                    else:
                        # Aphia Id gefunden
                        if sAdminTableName == "Expert Aaid":
                            # Tabelle: Expert Aaid
                            iMappedAaid = val["aaid"]
                            # Frage WoRMS: Existiert die Accepted Aphia Id?
                            resMyWorms = self.__myworms.getTaxon(iAid)
                            taxonWorms = resMyWorms.getData("taxon")
                            res.addResult(resMyWorms)
                            if resMyWorms.isErr():
                                # Fehler bei Bestimmung
                                sLogErr_ascii = "Table: "+sAdminTableName+", Aphia Id: " + str(iAid) + " -> Mapped Accepted Aphia Id "+str(iMappedAaid)+" has errors"
                                lstAdminErrors.append([sAdminTableName, iAid, "Mapped Accepted Aphia Id " + str(iMappedAaid) + " has errors"])
                                pv("- "+sLogErr_ascii)
                                res.addErr("Error with Mapped Accepted Aphia Id", add=sLogErr_ascii, kat="Update", sys="TaxonWizard")
                                bErr = True
                            elif taxonWorms.isNotFound() is None:
                                # Excepted Aphia Id existiert nicht
                                sLogErr_ascii = "Table: " + sAdminTableName + ", Aphia Id: " + str(iAid) + " -> Mapped Accepted Aphia Id " + str(iMappedAaid) + " not found in WoRMS"
                                lstAdminErrors.append([sAdminTableName, iAid, "Mapped Accepted Aphia Id " + str(iMappedAaid) + " not found in WoRMS"])
                                pv("- "+sLogErr_ascii)
                                res.addErr("Mapped Accepted Aphia Id not found in WoRMS", add="Table: " + sAdminTableName + ", Aphia Id: " + str(iAid) + " -> Mapped Accepted Aphia Id " + str(iMappedAaid), kat="Update", sys="TaxonWizard")
                                bErr = True
                            else:
                                # Accepted Aphia Id existiert bei WoRMS
                                iVaid = taxonWorms.getAphiaId_valid()
                                if iMappedAaid != iVaid:
                                    # Ist die Excepted Aphia Id wirklich accepted
                                    sLogErr_ascii = "Table: " + sAdminTableName + ", Aphia Id: " + str(iAid) + " -> Mapped Accepted Aphia Id "+str(iMappedAaid)+" is not accepted"
                                    lstAdminErrors.append([sAdminTableName, iAid, "Mapped Accepted Aphia Id "+str(iMappedAaid)+" is not Accepted"])
                                    pv("- "+sLogErr_ascii)
                                    res.addErr("Mapped Accepted Aphia Id is not Accepted", add= "Table: " + sAdminTableName + ", Aphia Id: " + str(iAid) + " -> Mapped Accepted Aphia Id "+str(iMappedAaid), kat="Update", sys="TaxonWizard")
                                    bErr = True

                        # *** Update Namen in Admin Tabellen
                        # Scientficname
                        bDiffScientificname = False
                        if sAdminTableName == "Expert Private":
                            sAdmTabScientificname = key
                        else:
                            sAdmTabScientificname = dicAdmin[iAid]["scientificname"] if "scientificname" in dicAdmin[iAid] else None
                        sWormsScientificname = taxonWorms.getScientificName()
                        # Accepted Scientificname
                        bDiffAcceptedScientificname = False
                        if sAdminTableName == "Expert Private":
                            sAdmTabAcceptedScientificname = dicAdmin[key]["accepted_scientificname"] if "accepted_scientificname" in dicAdmin[key] else None
                        else:
                            sAdmTabAcceptedScientificname = dicAdmin[iAid]["accepted_scientificname"] if "accepted_scientificname" in dicAdmin[iAid] else None
                        sWormsAcceptedScientificname = taxonWorms.getScientificName_accepted()
                        # Aaid Scientificname
                        if sAdminTableName == "Expert Aaid":
                            bDiffAaidAcceptedScientificname = False
                            sAdmTabAaidScientificname = dicAdmin[iAid]["aaid_scientificname"] if "aaid_scientificname" in dicAdmin[iAid] else None
                            sWormsAaidScientificname = taxonWorms.getScientificName_accepted()

                        # Welche Daten liegen anders vor?
                        sSqlCmdTeil = ""
                        lstSqlVal = []

                        if sAdminTableName != "Expert Private" and \
                                        sAdmTabScientificname != sWormsScientificname:
                            sSqlCmdTeil += "\"scientificname\"=%s,"
                            lstSqlVal.append(sWormsScientificname)
                            bDiffScientificname = True
                        if sAdmTabAcceptedScientificname != sWormsAcceptedScientificname:
                            sSqlCmdTeil += "\"accepted_scientificname\"=%s,"
                            lstSqlVal.append(sWormsAcceptedScientificname)
                            bDiffAcceptedScientificname = True
                        if sAdminTableName == "Expert Aaid":
                            if sAdmTabAaidScientificname != sWormsAaidScientificname:
                                sSqlCmdTeil += "\"aaid_scientificname\"=%s,"
                                lstSqlVal.append(sWormsAaidScientificname)
                                bDiffAaidAcceptedScientificname = True

                        # Sql-Cmd
                        if len(lstSqlVal) > 0:
                            iAdminUpdateNames += 1
                            sInfo = "Aphia Id: "+str(iAid)+" - Updating differing names: "
                            sInfo2 = ""
                            if bDiffScientificname:
                                sInfo2 += "Scientificname("+str(sAdmTabScientificname) + "->" + str(sWormsScientificname) + ") "
                            if bDiffAcceptedScientificname:
                                sInfo2 += "Accepted Scientificname("+str(sAdmTabAcceptedScientificname) + "->" + str(sWormsAcceptedScientificname) + ") "
                            if sAdminTableName == "Expert Aaid":
                                if bDiffAaidAcceptedScientificname:
                                    sInfo2 += "Aaid Scientificname(" + str(sAdmTabAaidScientificname) + "->" + str(sWormsAaidScientificname) + ") "
                            sInfo = sInfo+sInfo2
                            sInfo = sInfo[:-1]
                            pv(sInfo)
                            lstAdminUpdateNames.append([sAdminTableName, iAid, sInfo2])
                            sqlCmd = "update " + self.__model.subsystem_getDb().account_active_getSchema() + "." + sDbAdminTableName + " set " + sSqlCmdTeil[:-1] + " where aid=" + str(iAid) + ";"
                            cursor.execute(sqlCmd, lstSqlVal)
                            #iId = cursor.fetchone()[0]  # Antwort
                            sSqlQuery = cursor.query.decode("utf-8")
                            #pd("SQL: "+str(sSqlQuery) + " -> " + str(iId))

                # ---------------------

                # Progress / Status
                lstStatusEWI = list()
                sStatusEWI = ""
                iAdminErrors = len(lstAdminErrors)-1
                if iAdminErrors > 0:
                    lstStatusEWI.append("Adm-Err:" + str(iAdminErrors))
                if iAdminUpdateNames > 0:
                    lstStatusEWI.append("Adm-Upd:" + str(iAdminUpdateNames))
                if len(lstStatusEWI) != 0:
                    sStatusEWI = "[" + (" ".join(lstStatusEWI)) + "]"
                progress(iNr, iNrMax, 100, proz=True, abs=False, cli=True, status=sStatusEWI)
                sTaxon = taxonWorms.signature() if taxonWorms is not None else str(iAid)
                status(sAdminTableName+": " + sTaxon + " (" + str(iNr) + "/" + str(iNrMax) + ") " + sStatusEWI)

        if len(lstAdminErrors) > 1:
            pv("Errors in admin tables")

        # --------------------------------------------------------------------------------------------------------------
        # Update Artenlistengerüst
        # --------------------------------------------------------------------------------------------------------------
        sLogFrame_ascii = ""
        sLogTcStat_ascii = ""
        sLogTcStat_html = ""
        setBase = set()
        setPrivate = set()
        setAaid = set()
        setColony = set()
        setChange = set()
        setTaxonList = set()
        setTaxonListMissing = set()

        if not self.isAbort():
            # Bestimme fehlende Aphia Ids der Taxon Liste
            progstr.up(prae="Create taxon list framework", post=None)
            workflow(top="TaxonWizard", sub="Create taxon list framework", fmt=False)
            pn(progstr.cli())
            resGatherAid = self.__gatheringNeededAphiaIds()
            if resGatherAid.isErr():
                progress("off")
                return resGatherAid
            setTaxonAdminAll = resGatherAid.getData("all")
            setBase = resGatherAid.getData("base")
            setPrivate = resGatherAid.getData("private")
            setAaid = resGatherAid.getData("aaid")
            setColony = resGatherAid.getData("colony")
            setChange = resGatherAid.getData("change")
            setTaxonList = resGatherAid.getData("taxon-list")
            setTaxonListMissing = resGatherAid.getData("taxon-list-missing")

            # --------------------------------------------------------------------------------------------------------------
            # Update Artenlisteninhalt
            # --------------------------------------------------------------------------------------------------------------

            # Update Taxon List
            progress("on-pulse-abort")
            progstr.up(prae="Update Taxon List", post=None)
            workflow(top="TaxonWizard", sub="Update Taxon List", fmt=False)
            pn(progstr.cli())

            tc = TaxaCruncher(self.__model)

            try:
                # Loop ***
                lstTaxonTotal = list(set().union(setTaxonAdminAll, setTaxonList))
                lstTaxonTotalStart = lstTaxonTotal.copy()

                iNr = 0
                iNrMax = len(lstTaxonTotal)
                while lstTaxonTotal:
                    taxonWorms = None
                    iAid = lstTaxonTotal.pop(0)
                    iAid = int(iAid)
                    iNr += 1
                    tc.addTaxon(iAid, None, None)  # Taxon in TaxonCruncher geben
                    if self.isAbort():
                        break
                    if iAid > 0:
                        # Hole Daten aus WoRMS
                        resMyWorms = self.__myworms.getTaxon(iAid)
                        taxonWorms = resMyWorms.getData("taxon")
                        res.addResult(resMyWorms)
                        if resMyWorms.isErr():
                            # Fehler bei WoRMS-Abfrage
                            pv("Aphia Id: " + (str(iAid) + " - Error asking MyWoRMS: " + PrettyText.lst2strC(taxonWorms.getResult().getErrLst(short=True))))
                            lstTaxonWormsError.append((iAid, PrettyText.lst2strC(taxonWorms.getResult().getErrLst(short=True))))
                            iTaxonError += 1
                            bErr = True
                        elif taxonWorms.isNotFound():
                            # Aphia Id nicht gefunden
                            sLogErr_ascii = "Aphia Id: " + str(iAid) + " not found"
                            lstTaxonWormsError.append((str(iAid), "Aphia Id not found"))
                            iTaxonError += 1
                            bErr = True
                        else:
                            # WoRMS hat eine Taxon geliefert
                            # Hole Taxon Daten aus interner Artenliste
                            # TODO: Baue SQL (sicher)
                            sqlCmd = sql.SQL(
                                "select * from {schema}.{table} where aid=%s").format(
                                schema=sql.Identifier(self.__model.subsystem_getDb().account_active_getSchema()),
                                table=sql.Identifier("taxon")
                            )
                            cursor.execute(sqlCmd, [str(iAid)])
                            lstHeader = [head[0] for head in cursor.description]
                            lstContent = cursor.fetchone()

                            # Logging bei Veränderungen
                            sOutUpdatedFieldsNumber = ""
                            sOutUpdatedFields = ""
                            sOutUpdatedContent = ""
                            
                            bTaxonNew = False
                            bTaxonAaidChanged = False
                            bTaxonUpdate = False

                            if lstContent is not None:

                                # Taxon existiert in Taxon-Liste
                                dicTaxonIntern = dict(zip(lstHeader, lstContent))
                                dicTaxonIntern["aaid_way"] = [] # Setze Weg
                                del dicTaxonIntern["id"] # Id braucht man nicht
                                taxonIntern = Taxon(blueprint=dicTaxonIntern) # Erzeuge Taxon

                                # Bestimme alle Felder die für den Vergleich nicht benutzt werden sollen
                                lstExcludeAttribute = ["aaid_way", "aauth"]
                                if iAid in self.__dicWormsAaid: # Wenn Aid in Expertenliste Aaid befinden -> kein Vergleich!
                                    lstExcludeAttribute.append("aaid")

                                # Vergleiche Taxon intern mit Worms Taxon
                                resDiff = taxonIntern.cmp(taxonWorms, exclude=lstExcludeAttribute)
                                bEqual = resDiff.getData("equal")
                                dicDiff = resDiff.getData("diff")

                                if not bEqual:
                                    # Es existieren Veränderungen
                                    # Statistik über Veränderungen erstellen
                                    bTaxonNew = all(v is None for k, v in dicTaxonIntern.items() if k != "aid")
                                    if bTaxonNew:
                                        sOutUpdatedFieldsNumber = "new"
                                        # Zähler: Neues Taxon
                                        iTaxonNew += 1
                                    else:
                                        # Zähler: Veränderungen (kategorisiert)
                                        bTaxonUpdate = True
                                        bAaidChanged = False
                                        iTaxonUpdate += 1
                                        for key, val in dicDiff.items():
                                            if key in dicTaxonUpdateFields:
                                                dicTaxonUpdateFields[key] += 1
                                            else:
                                                dicTaxonUpdateFields[key] = 1
                                        dicTmp = {}
                                        for key, val in dicDiff.items():
                                            dicTmp[key] = "{" + str(val[0]) + "}->{" + str(val[1])+"}"
                                            if key == "aaid":
                                                bTaxonAaidChanged = True # Hat aidChanged = True
                                        sOutUpdatedFieldsNumber = str(len(dicDiff.keys()))
                                        sOutUpdatedFields = PrettyText.lst2strC(dicDiff.keys())
                                        sOutUpdatedContent = PrettyText.lst2strC(dicTmp)
                                    pv("UPD: " +taxonWorms.signature()+" ["+sOutUpdatedFieldsNumber + " fields: " + sOutUpdatedFields+"]")
                            else:
                                # Taxon existiert nicht in Liste (wurde während dieses Loops eingefügt)
                                # Zähler: Neues Taxon
                                sOutUpdatedFieldsNumber = "new"
                                iTaxonNew += 1
                                pv("NEW: "+taxonWorms.signature())

                            # Logging: New / Update
                            if sOutUpdatedFieldsNumber != "" and sOutUpdatedFields != "" and sOutUpdatedContent != "":
                                lstLogUpdates.append((str(iAid), taxonWorms.getData("scientificname"), sOutUpdatedFieldsNumber, sOutUpdatedFields, sOutUpdatedContent))

                            # *** Expert: Aaid?
                            if iAid in self.__dicWormsAaid:
                                # Check: Aaid in Expertenliste
                                taxonWorms.overide("aaid", self.__dicWormsAaid[iAid]["aaid"])
                                try:
                                    tmpAaid = self.__dicWormsAaid[iAid]["aaid"]
                                    tmpAaidScientificname = self.__dicWormsAaid[iAid]["aaid_scientificname"]
                                except:
                                    tmpAai = None
                                    tmpAaidScientificname = None
                                lstExpertAaid.append((iAid, taxonWorms.getData("scientificname"), tmpAaid, tmpAaidScientificname))
                                iExpertAaid += 1

                            # *** Problematische Taxa
                            lstTaxonProblem  = []
                            if taxonWorms.isUnaccepted() and taxonWorms.getAphiaId_accepted() is not None and not taxonWorms.isRedirected_2_accepted():
                                # Unaktzeptiertes Taxon mi aAid -> aTaxon ist aber nicht akzeptiert
                                setTaxaUnaccept_withAaid_leadingToNotacceptetTaxa.add(taxonWorms)
                                res.addWarn("Unaccepted taxon with Accepted AphiaId leading to unaccepted taxon",add=str(taxonWorms.signature()), kat="Misfits", sys="TaxonWizard")
                                bWarn = True
                                lstTaxonProblem.append("Unaccepted taxon with Accepted AphiaId leading unaccepted taxon")
                            if taxonWorms.getScientificName() is None:
                                # Kein SceintificName
                                setTaxaProblem_no_name.add(taxonWorms)
                                res.addWarn("Missing ScientificName", add=str(taxonWorms.signature()), kat="Misfits", sys="TaxonWizard")
                                bWarn = True
                                lstTaxonProblem.append("Missing ScientificName")
                            if taxonWorms.getScientificName_accepted() is None and taxonWorms.isAccepted():
                                # Kein Accepted ScientificName und Accepted
                                setTaxaProblem_no_aname_and_accepted.add(taxonWorms)
                                res.addWarn("Missing AcceptedScientificName and accepted", add=str(taxonWorms.signature()), kat="Misfits",sys="TaxonWizard")
                                bWarn = True
                                lstTaxonProblem.append("Missing AcceptedScientificName and accepted")
                            if taxonWorms.hasProblems_aphia():
                                # AphiaIds
                                setTaxaProblem_aphia.add(taxonWorms)
                                res.addWarn("AphiaId Problems", add=str(taxonWorms.signature()), kat="Misfits",sys="TaxonWizard")
                                lstTaxonProblem.append("AphiaId")
                                bWarn = True
                            if taxonWorms.hasProblems_status():
                                # Status
                                setTaxaProblem_status.add(taxonWorms)
                                res.addWarn("Status Problems", add=str(taxonWorms.signature()), kat="Mistifs",sys="TaxonWizard")
                                lstTaxonProblem.append("Status")
                                bWarn = True
                            sTaxonProblem = "["+PrettyText.lst2strC(lstTaxonProblem)+"]" if len(lstTaxonProblem) > 0 else ""
                            if sTaxonProblem != "":
                                pv("MFT: "+taxonWorms.signature()+" "+sTaxonProblem)

                            # Limbo Taxa
                            if taxonWorms.isLimbo():
                                setTaxaLimbo.add(taxonWorms)
                                res.addWarn("Limbo", add=str(taxonWorms.signature()), kat="Taxon in Limbo",sys="TaxonWizard")
                                setTaxaLimbo.add(taxonWorms)
                                pv("LBO: " + taxonWorms.signature())
                                bWarn = True

                            # *** Log: War Taxon nicht accepted? und Aaid in Taxon List updated -> Log Aaid und den Weg dorthin
                            if taxonWorms.isUnaccepted() and (bTaxonAaidChanged or bTaxonNew):
                                lstWormsAaid.append((
                                    taxonWorms.getData("aid"), taxonWorms.getData("scientificname"),
                                    taxonWorms.getData("aaid"), taxonWorms.getData("aname"),
                                    taxonWorms.getData("status"),
                                    PrettyText.lst2strC(taxonWorms.getData("aaid_way"))))
                                iTaxonAaid += 1

                            # *** Aaid noch nicht in Aid-Liste -> Eintragen
                            if (taxonWorms.getAphiaId_accepted() is not None) and (taxonWorms.getAphiaId_accepted() not in lstTaxonTotalStart):
                                lstTaxonTotalStart.append(taxonWorms.getAphiaId_accepted())
                                lstTaxonTotal.insert(0, taxonWorms.getAphiaId_accepted())
                                iTaxonNewAidWegenAaid += 1
                                lstTaxonNewAidWegenAaid.append([taxonWorms.getAphiaId_accepted(), taxonWorms.getScientificName()])

                            # *** Is Colony?
                            if iAid in self.__dicWormsColony:
                                taxonWorms.getData("is_colony") is True
                                iExpertColony += 1
                                lstExpertColony.append([iAid, taxonWorms.getScientificName()])

                            # *** Update Taxon Daten
                            # TODO: Baue SQL (sicher)
                            lstSqlAtr = []
                            dicSqlVal = {}
                            # Durchlaufe alle Attribute und Werte
                            lstExcludeAttribute = ["aaid_way", "aauth"]
                            for sAttribut, sWert in taxonWorms.getAllData().items():
                                if sAttribut not in lstExcludeAttribute:
                                    lstSqlAtr.append(sAttribut)
                                    dicSqlVal[sAttribut] = sWert
                            sqlCmd = sql.SQL("insert into {schema}.{table} ({attr}) values ({val}) on conflict (aid) do update set ({attr}) = ({val})").format(
                                schema=sql.Identifier(self.__model.subsystem_getDb().account_active_getSchema()),
                                table=sql.Identifier("taxon"),
                                attr=sql.SQL(', ').join(map(sql.Identifier, lstSqlAtr)),
                                val=sql.SQL(', ').join(map(sql.Placeholder, lstSqlAtr))
                            )
                            cursor.execute(sqlCmd, dicSqlVal)
                            sSqlQuery = cursor.query.decode("utf-8")
                            pd("SQL: " + str(sSqlQuery))

                    # Progress / Status
                    iErrGesamt = iAdminErrors+iTaxonError
                    iWarn = res.getWarnNo()
                    iInfo = res.getInfoNo()
                    sTaxon = taxonWorms.signature() if taxonWorms is not None else str(iAid)
                    progress(iNr, iNrMax, 100, proz=True, abs=False, cli=True, err=iErrGesamt, warn=iWarn, info=iInfo)
                    status("Update: "+sTaxon)

            except Exception as ex:
                # Fehler
                exc_type, exc_obj, exc_tb = sys.exc_info()
                fname = os.path.split(exc_tb.tb_frame.f_code.co_filename)[1]
                sLogErr_ascii = "DB error: "+str(ex)+" err_type=" + str(exc_type) + " methode=" + str(fname) + " line=" + str(exc_tb.tb_lineno)
                pd(sLogErr_ascii)
                res.addErr("Uexpected Error", add=sLogErr_ascii, kat="Update", sys="TaxonWizard")
                return res

            # TaxonCruncher erzeugt Statistic
            resTc = tc.crunch(html=False)
            taxa = resTc.getData("taxa")
            if resTc.isErr():
                res.addResult(resTc, msg_only=True)
                sLogTcStat_html = ""
                sLogTcStat_ascii = ""
            else:
                sLogTcStat_html = taxa.getStatistic(html=True)
                sLogTcStat_ascii = taxa.getStatistic(html=False)

        # --------------------------------------------------------------------------------------------------------------
        # Nacharbeiten
        # --------------------------------------------------------------------------------------------------------------

        progstr.up(prae="Check uniqueness of search names in expert knowledge", post=None)
        workflow(top="TaxonWizard", sub="Check expert knowledge", fmt=False)

        pn(progstr.cli())

        try:
            # Uniq Scientificname: Private <-> Taxonliste
            # TODO: Baue SQL (sicher)
            sqlCmd = sql.SQL(
                "select distinct t2.private_name from {schema}.{table1} t1 join {schema}.{table2} t2 on t1.scientificname=t2.private_name").format(
                schema=sql.Identifier(self.__model.subsystem_getDb().account_active_getSchema()),
                table1=sql.Identifier("taxon"),
                table2=sql.Identifier("wormsprivate"),
            )
            cursor.execute(sqlCmd)
            sSqlQuery = cursor.query.decode("utf-8")
            lstContent = [[tp[0]] for tp in cursor.fetchall()]
            lstUniqPrivateTaxon += lstContent
            pd("SQL: " + str(sSqlQuery))
            if len(lstUniqPrivateTaxon) > 1:
                sLogErr_ascii = "Some search names in WoRMS privat list found in taxon list"
                res.addErr("Some search names in WoRMS privat list found in taxon list", add=PrettyText.lst2strC(lstUniqPrivateTaxon), kat="Update", sys="TaxonWizard")

            # Uniq Scientificname: NoWORMS <-> Taxonliste
                # TODO: Baue SQL (sicher)
                sqlCmd = sql.SQL(
                    "select distinct t2.scientificname from {schema}.{table1} t1 join {schema}.{table2} t2 on t1.scientificname=t2.scientificname").format(
                    schema=sql.Identifier(self.__model.subsystem_getDb().account_active_getSchema()),
                    table1=sql.Identifier("taxon"),
                    table2=sql.Identifier("noworms"),
                )
                cursor.execute(sqlCmd)
                sSqlQuery = cursor.query.decode("utf-8")
                lstContent = [[tp[0]] for tp in cursor.fetchall()]
                lstUniqNowormsTaxon += lstContent
                pd("SQL: " + str(sSqlQuery))
                if len(lstUniqNowormsTaxon) > 1:
                    sLogErr_ascii = "Some search names in No WoRMS list found in taxon list"
                    res.addErr("Some search names in No WoRMS list found in taxon list", add=PrettyText.lst2strC(lstUniqNowormsTaxon), kat="Update", sys="TaxonWizard")

            # Uniq Scientificname: Private <-> NoWORMS
            # TODO: Baue SQL (sicher)
            sqlCmd = sql.SQL(
                "select distinct t2.private_name from {schema}.{table1} t1 join {schema}.{table2} t2 on t1.scientificname=t2.private_name").format(
                schema=sql.Identifier(self.__model.subsystem_getDb().account_active_getSchema()),
                table1=sql.Identifier("noworms"),
                table2=sql.Identifier("wormsprivate"),
            )
            cursor.execute(sqlCmd)
            sSqlQuery = cursor.query.decode("utf-8")
            lstContent = [[tp[0]] for tp in cursor.fetchall()]
            lstUniqPrivateNoworms += lstContent
            pd("SQL: " + str(sSqlQuery))
            if len(lstUniqPrivateNoworms) > 1:
                sLogErr_ascii = "Some search names in WoRMS private list found in No WoRMS list"
                res.addErr("Some search names in WoRMS private list found No WoRMS list",
                           add=PrettyText.lst2strC(lstUniqPrivateNoworms), kat="Update", sys="TaxonWizard")

            # Taxa with same Scientificname
            # TODO: Baue SQL (sicher)
            sqlCmd = sql.SQL(
                "select t1.scientificname as sn, t1.aid, t1.vaid, t1.status from {schema}.taxon t1, {schema}.taxon t2 where t1.scientificname=t2.scientificname and t1.id<>t2.id order by sn").format(
                schema=sql.Identifier(self.__model.subsystem_getDb().account_active_getSchema())
            )
            cursor.execute(sqlCmd)
            sSqlQuery = cursor.query.decode("utf-8")
            lstContent = cursor.fetchall()
            lstAmbigousScientificnames += lstContent
            pd("SQL: " + str(sSqlQuery))
            if len(lstAmbigousScientificnames) > 1:
                iAmbigousScientificnames = len(lstAmbigousScientificnames)-1
                sInfo = "Ambigous scientificname found in taxon list"
                res.addInfo(sInfo)

            # Accepted Taxa with same Scientificname
            # TODO: Baue SQL (sicher)
            sqlCmd = sql.SQL(
                "select scientificname, aid, status from {schema}.taxon where scientificname in (select scientificname from {schema}.taxon where aid=vaid and status='accepted' group by scientificname having count(scientificname)>1) order by scientificname").format(
                schema=sql.Identifier(self.__model.subsystem_getDb().account_active_getSchema())
            )
            cursor.execute(sqlCmd)
            sSqlQuery = cursor.query.decode("utf-8")
            lstContent = cursor.fetchall()
            lstAmbigousScientificnamesAccepted += lstContent
            pd("SQL: " + str(sSqlQuery))
            if len(lstAmbigousScientificnamesAccepted) > 1:
                bWarn = True
                iAmbigousScientificnamesAccepted = len(lstAmbigousScientificnamesAccepted)-1
                sInfo = "Accepted taxa found with ambigous Scientificname"
                res.addInfo(sInfo)

        except Exception as ex:
            # Fehler
            exc_type, exc_obj, exc_tb = sys.exc_info()
            fname = os.path.split(exc_tb.tb_frame.f_code.co_filename)[1]
            sLogErr_ascii = "DB error: "+str(ex)+"err_type=" + str(exc_type) + " methode=" + str(fname) + " line=" + str(exc_tb.tb_lineno)
            pd(sLogErr_ascii)
            res.addErr("Unexcepted Error", add=sLogErr_ascii, kat="Update", sys="TaxonWizard")

        progress("off")

        # ---------------------------——---------------------------——---------------------------——

        # *** Log
        iPreview = 3
        # Frame
        # ascii
        sLogFrame_ascii += "TaxonWizard gathered AphiaIds to build new taxon List from ...\n"
        if len(setBase) + len(setPrivate) + len(setAaid) + len(setColony) + len(setChange) + len(setTaxonList) > 0:
            if len(setBase) > 0:
                sLogFrame_ascii += "Base List: " + str(len(setBase)) + "\n"
            if len(setPrivate) > 0:
                sLogFrame_ascii += "Expert Knowledge Private Names: " + str(len(setPrivate)) + "\n"
            if len(setAaid) > 0:
                sLogFrame_ascii += "Expert Knowledge Accepted AphiaIds: " + str(len(setAaid)) + "\n"
            if len(setColony) > 0:
                sLogFrame_ascii += "Expert Knowledge Colony: " + str(len(setColony)) + "\n"
            if len(setChange) > 0:
                sLogFrame_ascii += "Expert Knowledge Change: " + str(len(setChange)) + "\n"
            if len(setTaxonList) > 0:
                sLogFrame_ascii += "Old Taxon List: " + str(len(setTaxonList)) + "\n"
            sLogFrame_ascii += "\n"
        else:
            sLogFrame_ascii += "no AphiaIds available\n\n"
        # html
        sLogFrame_html = sLogFrame_ascii

        # Errors
        sLogErr_ascii = ""
        sLogErr_html = ""
        if bErr:
            # ascii
            sLogErr_ascii += ">>>>>>>>>>>>>>>>>><<<<<<<<<<<<<<<<\n"
            sLogErr_ascii += "Errors: "+str(res.getErrNo())+"x\n"
            sLogErr_ascii += ">>>>>>>>>>>>>>>>>><<<<<<<<<<<<<<<<\n\n"
            if len(lstAdminErrors) > 1:
                sLogErr_ascii += "--- Errors in admin tables: "+str(iAdminErrors)+"x\n"
                sLogErr_ascii += PrettyText.table(lstAdminErrors) +"\n"
            if len(lstTaxonWormsError) > 1:
                sLogErr_ascii += "--- Errors asking WoRMS: "+str(iTaxonError)+"x\n"
                sLogErr_ascii += PrettyText.table(lstTaxonWormsError) +"\n"
            if len(lstUniqPrivateTaxon) > 1:
                sLogErr_ascii += "--- Private Names are found in Taxon List: "+str(len(lstUniqPrivateTaxon)-1)+"x\n"
                sLogErr_ascii += PrettyText.table(lstUniqPrivateTaxon) +"\n"
            if len(lstUniqNowormsTaxon) > 1:
                sLogErr_ascii += "--- ScientificNames of NoWoRMS-List are found in Taxon List: "+str(len(lstUniqNowormsTaxon)-1)+"x\n"
                sLogErr_ascii += PrettyText.table(lstUniqNowormsTaxon) +"\n"
            if len(lstUniqPrivateNoworms) > 1:
                sLogErr_ascii += "--- Private Names are found in NoWoRMS-List: " + str(len(lstUniqPrivateNoworms) - 1) + "x\n"
                sLogErr_ascii += PrettyText.table(lstUniqPrivateNoworms) + "\n"
            sLogErr_ascii +="\n"
            # html
            sLogErr_html += "*** Errors\n"
            if len(lstAdminErrors) > 1:
                sLogErr_html += "* Errors in admin tables: "+str(iAdminErrors)+"x\n"
                sLogErr_html += PrettyText.table(lstAdminErrors, html=True, preview=iPreview) +"\n"
            if len(lstTaxonWormsError) > 1:
                sLogErr_html += "* Errors asking WoRMS: "+str(iTaxonError)+"x\n"
                sLogErr_html += PrettyText.table(lstTaxonWormsError, html=True, preview=iPreview) +"\n"
            if len(lstUniqPrivateTaxon) > 1:
                sLogErr_html += "* Private Names are found in Taxon List: "+str(len(lstUniqPrivateTaxon)-1)+"x\n"
                sLogErr_html += PrettyText.table(lstUniqPrivateTaxon, html=True, preview=iPreview) +"\n"
            if len(lstUniqNowormsTaxon) > 1:
                sLogErr_html += "* ScientificNames of NoWoRMS-List are found in Taxon List: "+str(len(lstUniqNowormsTaxon)-1)+"x\n"
                sLogErr_html += PrettyText.table(lstUniqNowormsTaxon, html=True, preview=iPreview) +"\n"
            if len(lstUniqPrivateNoworms) > 1:
                sLogErr_html += "* Private Names are found in NoWoRMS-List: " + str(len(lstUniqPrivateNoworms) - 1) + "x\n"
                sLogErr_html += PrettyText.table(lstUniqPrivateNoworms, html=True, preview=iPreview) + "\n"
            sLogErr_html +="\n"

        # Warnings
        sLogWarn_ascii = ""
        sLogWarn_html = ""
        if bWarn:
            # ascii
            sLogWarn_ascii = ">>>>>>>>>>>>>>>>>><<<<<<<<<<<<<<<<\n"
            sLogWarn_ascii += "Warnings: "+str(res.getWarnNo())+"x\n"
            sLogWarn_ascii += ">>>>>>>>>>>>>>>>>><<<<<<<<<<<<<<<<\n\n"
            if len(setTaxaProblem_no_name) > 0:
                sLogWarn_ascii += "--- Taxa with missing ScientificName - update expert knowledge to fix this (" + str(len(setTaxaProblem_no_name)) + "x)\n"
                for t in setTaxaProblem_no_name:
                    sLogWarn_ascii += t.signature() + "\n"
                sLogWarn_ascii += "\n"
            if len(setTaxaProblem_no_aname_and_accepted) > 0:
                sLogWarn_ascii += "--- Accepted taxon with missing AcceptedScientificName - update Expert Knowledge to fix this (" + str(len(setTaxaProblem_no_aname_and_accepted)) + "x)\n"
                for t in setTaxaProblem_no_aname_and_accepted:
                    sLogWarn_ascii += t.signature() + "\n"
                sLogWarn_ascii += "\n"
            if len(setTaxaProblem_aphia) > 0:
                sLogWarn_ascii += "--- Taxa with AphiaIds problems (" + str(len(setTaxaProblem_aphia)) + "x)\n"
                for t in setTaxaProblem_aphia:
                    sLogWarn_ascii += t.signature(debug=True) + "\n"
                sLogWarn_ascii += "\n"
            if len(lstAmbigousScientificnamesAccepted) > 1:
                sLogWarn_ascii += "--- Accepted taxa with ambigous ScientificNames - update Expert Knowledge to fix this (" + str(len(lstAmbigousScientificnamesAccepted)-1) + "x)\n"
                sLogWarn_ascii += PrettyText.table(lstAmbigousScientificnamesAccepted, header=True) + "\n"
                sLogWarn_ascii += "\n"
            if len(setTaxaUnaccept_withAaid_leadingToNotacceptetTaxa) > 0:
                sLogWarn_ascii += "--- Unaccepted Taxa with Accepted AphiaId leading to not accepted Taxa (" + str(len(setTaxaUnaccept_withAaid_leadingToNotacceptetTaxa)) + "x)\n"
                for t in setTaxaUnaccept_withAaid_leadingToNotacceptetTaxa:
                    sLogWarn_ascii += t.signature() + "\n"
                sLogWarn_ascii += "\n"
            if len(setTaxaLimbo) > 0:
                sLogWarn_ascii += "--- Taxa in Limbo (" + str(len(setTaxaLimbo)) + "x)\n"
                for t in setTaxaLimbo:
                    sLogWarn_ascii += t.signature() + "\n"
                sLogWarn_ascii += "\n"
            if len(setTaxaProblem_status) > 0:
                sLogWarn_ascii += "--- Taxa with status problems (" + str(len(setTaxaProblem_status)) + "x)\n"
                for t in setTaxaProblem_status:
                    sLogWarn_ascii += t.signature() + "\n"
                sLogWarn_ascii += "\n"
            sLogWarn_ascii += "\n"
            # html
            sLogWarn_html += "*** Warnings\n"
            if len(setTaxaProblem_no_name) > 0:
                sLogWarn_html += "* Taxa with missing ScientificName (" + str(len(setTaxaProblem_no_name)) + "x)\n"
                sLogWarn_html += "<i>update expert knowledge to fix this</i>\n"
                sLogWarn_html += PrettyText.table([[t.signature(html=True)] for t in setTaxaProblem_no_name], header=False, html=True, preview=iPreview) + "\n"
            if len(setTaxaProblem_no_aname_and_accepted) > 0:
                sLogWarn_html += "* Accepted taxon with missing AcceptedScientificName (" + str(
                    len(setTaxaProblem_no_aname_and_accepted)) + "x)\n"
                sLogWarn_html += "<i>update expert knowledge to fix this</i>\n"
                sLogWarn_html += PrettyText.table([[t.signature(html=True)] for t in setTaxaProblem_no_aname_and_accepted], header=False, html=True, preview=iPreview) + "\n"
            if len(setTaxaProblem_aphia) > 0:
                sLogWarn_html += "* Taxa with AphiaIds problems (" + str(len(setTaxaProblem_aphia)) + "x)\n"
                sLogWarn_html += PrettyText.table([[t.signature(html=True)] for t in setTaxaProblem_aphia], header=False, html=True, preview=iPreview) + "\n"
            if len(lstAmbigousScientificnamesAccepted) > 1:
                sLogWarn_html += "* Accepted taxa with ambigous ScientificNames (" + str(len(lstAmbigousScientificnamesAccepted) - 1) + "x)\n"
                sLogWarn_html += "<i>update expert knowledge to fix this</i>\n"
                sLogWarn_html += PrettyText.table(lstAmbigousScientificnamesAccepted, header=True, html=True, preview=iPreview) + "\n"
            if len(setTaxaUnaccept_withAaid_leadingToNotacceptetTaxa) > 0:
                sLogWarn_html += "* Unaccepted Taxa with Accepted AphiaId leading to not accepted Taxa (" + str(
                    len(setTaxaUnaccept_withAaid_leadingToNotacceptetTaxa)) + "x)\n"
                sLogWarn_html += PrettyText.table([[t.signature(html=True)] for t in setTaxaUnaccept_withAaid_leadingToNotacceptetTaxa], header=False, html=True, preview=iPreview) + "\n"
            if len(setTaxaLimbo) > 0:
                sLogWarn_html += "* Taxa in Limbo (" + str(len(setTaxaLimbo)) + "x)\n"
                sLogWarn_html += PrettyText.table([[t.signature(html=True)] for t in setTaxaLimbo], header=False, html=True, preview=iPreview) + "\n"
            if len(setTaxaProblem_status) > 0:
                sLogWarn_html += "* Taxa with status problems (" + str(len(setTaxaProblem_status)) + "x)\n"
                sLogWarn_html += PrettyText.table([[t.signature(html=True)] for t in setTaxaProblem_status], header=False, html=True, preview=iPreview) + "\n"

        # Rest
        # ascii
        sLogRest_ascii = ""
        bResultsRest = False
        if len(lstWormsAaid) > 1:
            sLogRest_ascii += "--- Added/Updated Traced Accepted AphiaIds: " + str(iTaxonAaid) + "x\n"
            sLogRest_ascii += PrettyText.table(lstWormsAaid) + "\n\n"
            bResultsRest = True
        if len(lstTaxonNewAidWegenAaid) > 1:
            sLogRest_ascii += "--- Added Taxon because of Traced AcceptedAphia Ids: " + str(iTaxonNewAidWegenAaid) + "x\n"
            sLogRest_ascii += PrettyText.table(lstTaxonNewAidWegenAaid) + "\n\n"
            bResultsRest = True
        if len(lstExpertAaid) > 1:
            sLogRest_ascii += "--- Used Expert Knowledge Accepted AphiaIds: " + str(iExpertAaid) + "x\n"
            sLogRest_ascii += PrettyText.table(lstExpertAaid) + "\n\n"
            bResultsRest = True
        if len(lstExpertColony) > 1:
            sLogRest_ascii += "--- Used Expert Knowledge Colony: " + str(iExpertColony) + "x\n"
            sLogRest_ascii += PrettyText.table(lstExpertColony) + "\n\n"
            bResultsRest = True
        if len(lstAdminUpdateNames) > 1:
            sLogRest_ascii += "--- Updated Records in Admin Tables: " + str(iAdminUpdateNames) + "x\n"
            sLogRest_ascii += PrettyText.table(lstAdminUpdateNames) + "\n\n"
            bResultsRest = True
        if len(lstAmbigousScientificnames) > 1:
            sLogRest_ascii += "--- Ambigous ScientificNames Found: " + str(iAmbigousScientificnames) + "x\n"
            sLogRest_ascii += PrettyText.table(lstAmbigousScientificnames) + "\n\n"
            bResultsRest = True
        if len(lstLogUpdates) > 1:
            sLogRest_ascii += "--- Updated Fields: "+str(len(lstLogUpdates)-1)+"\n"
            sLogRest_ascii += PrettyText.table(lstLogUpdates) + "\n\n"
            bResultsRest = True
        if not bResultsRest :
            sLogRest_ascii += "... nothing special happend, everything was fine ...\n\n"
        # html
        sLogRest_html = ""
        if len(lstWormsAaid) > 1:
            sLogRest_html += "* Added/Updated Traced Accepted AphiaIds (" + str(iTaxonAaid) + "x)\n"
            sLogRest_html += PrettyText.table(lstWormsAaid, html=True, preview=iPreview) + "\n\n"
        if len(lstTaxonNewAidWegenAaid) > 1:
            sLogRest_html += "* Added Taxon because of Traced AcceptedAphia Ids (" + str(iTaxonNewAidWegenAaid) + "x)\n"
            sLogRest_html += PrettyText.table(lstTaxonNewAidWegenAaid, html=True, preview=iPreview) + "\n\n"
        if len(lstExpertAaid) > 1:
            sLogRest_html += "* Used Expert Knowledge Accepted AphiaIds (" + str(iExpertAaid) + "x)\n"
            sLogRest_html += PrettyText.table(lstExpertAaid, html=True, preview=iPreview) + "\n\n"
        if len(lstExpertColony) > 1:
            sLogRest_html += "* Used Expert Knowledge Colony (" + str(iExpertColony) + "x)\n"
            sLogRest_html += PrettyText.table(lstExpertColony, html=True, preview=iPreview) + "\n\n"
        if len(lstAdminUpdateNames) > 1:
            sLogRest_html += "* Updated Records in Admin Tables (" + str(iAdminUpdateNames) + "x)\n"
            sLogRest_html += PrettyText.table(lstAdminUpdateNames, html=True, preview=iPreview) + "\n\n"
        if len(lstAmbigousScientificnames) > 1:
            sLogRest_html += "* Ambigous ScientificNames Found (" + str(iAmbigousScientificnames) + "x)\n"
            sLogRest_html += PrettyText.table(lstAmbigousScientificnames, html=True, preview=iPreview) + "\n\n"
        if len(lstLogUpdates) > 1:
            sLogRest_html += "* Updated Fields ("+str(len(lstLogUpdates)-1)+")\n"
            sLogRest_html += PrettyText.table(lstLogUpdates, html=True, preview=iPreview) + "\n\n"
        if not bResultsRest :
            sLogRest_html += "... nothing special happend, everything was fine ...\n\n"
        else:
            sLogRest_html = "*** Other messages\n" + sLogRest_html + "\n"

            
        # Log Prae
        sColor_awi_blue_light = Colors.getColor("awi_blue_light", html=True)
        sColor_awi_blue_mid = Colors.getColor("awi_blue_mid", html=True)
        sColor_awi_blue_dark = Colors.getColor("awi_blue_dark", html=True)
        iSekTotal = time.time() - iZeitStart
        iStd = int(iSekTotal // 3600)
        iMin = int((iSekTotal % 3600) // 60)
        iSek = int(iSekTotal % 60)
        # ascii
        sLogPrae_ascii = "************************************\n"
        sLogPrae_ascii += "TaxonWizard\n"
        sLogPrae_ascii += "Date: " + str(datetime.datetime.now().strftime("%Y-%m-%d %H:%M:%S")) + "\n"
        sLogPrae_ascii += "Duration: " + str(iStd) + "h" + str(iMin) + "m" + str(iSek) + "s\n"
        sLogPrae_ascii += "************************************\n\n"
        # html
        sLogPrae_html = ""
        sLogPrae_html += '<img height="67" src="https://recruitingapp-5442.de.umantis.com/pubhtml/logo.gif" alt="AWI Logo">'
        sLogPrae_html += '<div layout="block" styleclass="DynForm_DfAttributes" style="z-index:0;position:absolute;left:0px;top:68px;right:0px;height:25px;overflow:hidden;text-align:left;background-color:' + sColor_awi_blue_light + ';border:medium none;-moz-border-colors:none;border-radius:0; margin-top: 5px; margin-bottom: 5px; ;"></div>'
        sLogPrae_html += '<br>'
        sLogPrae_html += "\n\n"
        sLogPrae_html += "&nbsp;"
        sLogPrae_html += '<h1 style="color:' + sColor_awi_blue_dark + ';">TaxonWizard</h1>\n\n'
        sLogPrae_html += "Date: " + str(datetime.datetime.now().strftime("%Y-%m-%d %H:%M:%S")) + "\n"
        sLogPrae_html += "Duration: " + str(iStd) + "h" + str(iMin) + "m" + str(iSek) + "s\n"

        # Overview
        # ascii
        sLogWarnErrOverview_ascii = ""
        sLogWarnErrOverviewEndRule_ascii = ""
        if self.isAbort() or bErr or bWarn:
            sLogWarnErrOverview_ascii +=  ">>>>>>>>>>>>>>>>>>><<<<<<<<<<<<<<<<<\n"
            if self.isAbort():
                sLogWarnErrOverview_ascii += "User aborted update\n"
            if bErr:
                sLogWarnErrOverview_ascii += "Errors: "+str(res.getErrNo())+"x\n"
            if bWarn:
                sLogWarnErrOverview_ascii += "Warnings: "+str(res.getWarnNo())+"x\n"
            sLogWarnErrOverview_ascii += ">>>>>>>>>>>>>>>>>>><<<<<<<<<<<<<<<<<\n\n"
        sLogWarnErrOverviewEndRule_ascii = ">>>>>>>>>>>>>>>>>>><<<<<<<<<<<<<<<<<\n\n"
        # html
        sLogWarnErrOverview_html = ""
        if self.isAbort() or bErr or bWarn:
            sLogWarnErrOverview_html += "*** Caution\n"
            if self.isAbort():
                sLogWarnErrOverview_html += "User aborted update\n"
            if bErr:
                sLogWarnErrOverview_html += "* Errors: " + str(res.getErrNo()) + "x\n"
            if bWarn:
                sLogWarnErrOverview_html += "* Warnings: " + str(res.getWarnNo()) + "x\n"

        # Log zusammenbauen
        sLog = sLogPrae_ascii + sLogFrame_ascii +  sLogWarnErrOverview_ascii + sLogTcStat_ascii + sLogErr_ascii + sLogWarn_ascii + sLogWarnErrOverviewEndRule_ascii + sLogRest_ascii
        sLog = sLog.replace("\n\n\n", "\n\n")
        sLog_html = sLogPrae_html + sLogFrame_html + sLogWarnErrOverview_html + sLogTcStat_html + sLogErr_html + sLogWarn_html + sLogRest_html
        sLog_html = PrettyText.html_txt2html(sLog_html)

        # *** Antwort
        # Titel
        if bErr:
            sTitle = "Taxon List Updated"
        elif self.__model.conf_isDry():
            sTitle = "Taxon List Updated (in dry mode)"
        else:
            sTitle = "Taxon List Updated"
        res.addData("title", sTitle)
        res.addData("text", sLog_html)
        res.addData("abort", self.isAbort())

        # *** Action Log
        if not self.__model.conf_isDry() and not bErr and not self.isAbort() and not res.isErr():
            resActionLog = self.__model.subsystem_getDb().actionlog_new("tax-list-update", "taxon", sLog, cursor=cursor)
            res.addResult(resActionLog)

        # *** Admin
        if not self.__model.conf_isDry() and not self.isAbort():

            try:
                sSchema = self.__model.subsystem_getDb().account_active_getSchema()
                # Speichere Datum des Updates
                dateTaxWizLastUpdateDate = datetime.datetime.now()
                sqlCmd = sql.SQL(
                    "DELETE FROM {schema}.admin WHERE key='taxon_wizard_last_update_date'").format(
                    schema=sql.Identifier(sSchema)
                )
                cursor.execute(sqlCmd)
                sqlCmd = sql.SQL(
                    "INSERT INTO {schema}.admin (key, value) VALUES ('taxon_wizard_last_update_date', {val})").format(
                    schema=sql.Identifier(sSchema),
                    val=sql.Literal(str(dateTaxWizLastUpdateDate))
                )
                cursor.execute(sqlCmd)
                # Speichere Anzahl Fehler
                sqlCmd = sql.SQL(
                    "DELETE FROM {schema}.admin WHERE key='taxon_wizard_last_update_err'").format(
                    schema=sql.Identifier(sSchema)
                )
                cursor.execute(sqlCmd)
                sqlCmd = sql.SQL(
                    "INSERT INTO {schema}.admin (key, value) VALUES ('taxon_wizard_last_update_err', {val})").format(
                    schema=sql.Identifier(sSchema),
                    val=sql.Literal(str(res.getErrNo()))
                )
                cursor.execute(sqlCmd)
                # Speichere Anzahl Warnungen
                sqlCmd = sql.SQL(
                    "DELETE FROM {schema}.admin WHERE key='taxon_wizard_last_update_warn'").format(
                    schema=sql.Identifier(sSchema)
                )
                cursor.execute(sqlCmd)
                sqlCmd = sql.SQL(
                    "INSERT INTO {schema}.admin (key, value) VALUES ('taxon_wizard_last_update_warn', {val})").format(
                    schema=sql.Identifier(sSchema),
                    val=sql.Literal(str(res.getWarnNo()))
                )
                cursor.execute(sqlCmd)
                # Speicher Log
                sqlCmd = sql.SQL(
                    "DELETE FROM {schema}.admin WHERE key='taxon_wizard_last_update_log'").format(
                    schema=sql.Identifier(sSchema)
                )
                cursor.execute(sqlCmd)
                sqlCmd = sql.SQL(
                    "INSERT INTO {schema}.admin (key, value) VALUES ('taxon_wizard_last_update_log', {log})").format(
                    schema=sql.Identifier(sSchema),
                    log=sql.Literal(str(sLog_html))
                )
                cursor.execute(sqlCmd)
            except Exception as ex:
                res.addErr("Problems writing to admin tables", add=str(ex), kat="Update", sys="TaxonWizard")

        # Aufräumen
        cleanUi()
        closeDb()

        # MyWoRMS Puffer saven
        self.__myworms.puffer_save()

        return res
