# coding=utf-8

import openpyxl
from collections import Counter, OrderedDict
import os, sys, csv, getpass, platform, datetime, time, copy, re
import openpyxl.styles.numbers
import warnings
from tools.coding import Result, WorkingObject, DataTools
from tools.excel import ExcelTools, ExcelNavigator
from tools.text import PrettyText
from tools.parallel import Stoppable
from tools.printing import SmartPrinter
from tools.misc import JsonConverter
from tools.geocruncher import GeoCruncher
from pathlib import Path

# -------------------------------------------------------------
# Import: Validiere Excel Workbook (Sheets, Headers)
# -------------------------------------------------------------

class ExcelValidator(Stoppable, WorkingObject):

    # Konstruktor
    def __init__(self, model):
        Stoppable.__init__(self)
        WorkingObject.__init__(self)

        self.__model = model
        self.__smartprinter = model.ui_getSmartprinter()
        self.__map = model.subsystem_getMapper()

    # Info
    def info(self):
        sOut = "*** Status of ExcelValidator System\n\n"
        return sOut

    # Validate Excel Format
    def startValidate(self, wb, **kwargs):
        """
        Args:
            wb (workbook): Workbook
        Kwargs:
            result_in: Result in
            ingest_wf (str): Ingest Workflow
            ingest (bool): Ingest
            lookup (bool): Lookup
            taxon (bool): Taxon
            ingest_sheets (lst): Liste aller hinzuzufügenden Sheets
            lookup_sheets (lst): Liste aller hinzuzufügenden Sheets
            taxon_sheets (lst): Liste aller hinzuzufügenden Sheets
            batch (str): Batch Info die weitergegeben wird
        Return
            Result
        """

        pn = self.__smartprinter.normal
        pv = self.__smartprinter.verbose
        workflow = self.__smartprinter.workflow
        progress = self.__smartprinter.progress
        status = self.__smartprinter.status

        res = Result(name="Validate Excel Workbook")

        # Parameter
        resultIn = kwargs["result_in"] if "result_in" in kwargs else Result("Empty Result Input Dummy")
        bIngest = True if "ingest" in kwargs and kwargs["ingest"] is True else False
        bLookup = True if "lookup" in kwargs and kwargs["lookup"] is True else False
        bTaxon = True if "taxon" in kwargs and kwargs["taxon"] is True else False
        sIngestWf = kwargs["ingest_wf"] if "ingest_wf" in kwargs else None
        lstSheetsIngest = kwargs["ingest_sheets"] if "ingest_sheets" in kwargs else None
        lstSheetsLookup = kwargs["lookup_sheets"] if "lookup_sheets" in kwargs else None
        lstSheetsTaxon = kwargs["taxon_sheets"] if "taxon_sheets" in kwargs else None

        progress("on-pulse")
        pn("* Validate Excel Workbook")
        workflow(top="Validator", sub="Checking Exel")
        status()

        # Parameter Test
        if not bIngest and not bLookup and not bTaxon:
            res.addErr("Nothing to do ...", kat="Idle", sys="Excel Checker")
        if wb is None:
            res.addErr("No Excel Workbook given", sys="Excel Checker", kat="I/O")

        # Zu untersuchende Sheets
        lstSystems = [
            ("ingest", bIngest, lstSheetsIngest, self.__map.ingest_getExcelSheets(sIngestWf)),
            ("lookup", bLookup, lstSheetsLookup, self.__map.lookup_getExcelSheets()),
            ("taxon", bTaxon, lstSheetsTaxon, self.__map.taxon_getExcelSheets())
        ]

        # Loop Systeme
        sLog = "************************************\n"
        sLog += "Validate Excel Workbook\n"
        sLog += "************************************\n"

        lstSheetsNotAllowedAtAll = wb.sheetnames
        try:
            lstSheetsNotAllowedAtAll.remove("Glossary") # Special: Entferne GLossary (das ist OK)
        except:
            pass

        for tpSystem in lstSystems:
            sSystem = tpSystem[0]
            bUse = tpSystem[1]
            lstSheetsWanted = tpSystem[2]
            lstSheetsMap = tpSystem[3]
            lstSheetAllowedFound = []
            lstSheetImport = []
            if bUse:
                sLog += "\n"
                sLog += "------------------------------------\n"
                sLog += sSystem.capitalize() + "\n"
                sLog += "------------------------------------\n\n"
                if sSystem == "ingest":
                    if sIngestWf is not None:
                        sLog += "Workflow: " + str(self.__map.workflow_short2long(sIngestWf)) + "\n"
                # Sheets
                sLog += "--- Sheets ---\n"
                for sSheetMap in lstSheetsMap:
                    if sSheetMap in wb.sheetnames:
                        lstSheetAllowedFound.append(sSheetMap)
                        lstSheetsNotAllowedAtAll.remove(sSheetMap)
                        if (lstSheetsWanted is None) or (len(lstSheetsWanted)==0) or (sSheetMap in lstSheetsWanted):
                            lstSheetImport.append(sSheetMap)
                sLog += "- allowed: " + PrettyText.lst2strC(lstSheetsMap) +"\n"
                sLog += "- found allowed: " + PrettyText.lst2strC(lstSheetAllowedFound) +"\n"
                if (lstSheetsWanted is not None) and len(lstSheetsWanted) > 0:
                    sLog += "- use only these: " + PrettyText.lst2strC(lstSheetsWanted) +"\n"
                sLog += "- will be imported: " + PrettyText.lst2strC(lstSheetImport) +"\n"
                #res.addInfo("Sheets allowed by Workflow", add=PrettyText.lst2strC(lstSheetsMap), kat="Excel Sheets: "+sSystem, sys="Excel Checker")
                #res.addInfo("Found allowed sheets in File", add=PrettyText.lst2strC(lstSheetAllowedFound), kat="Excel Sheets: "+sSystem,sys="Excel Checker")
                #res.addInfo("Sheets choosen by user", add=PrettyText.lst2strC(lstSheetsWanted), kat="Excel Sheets: "+sSystem,sys="Excel Checker")
                #res.addInfo("Imported Sheets", add=PrettyText.lst2strC(lstSheetImport), kat="Excel Sheets: "+sSystem,sys="Excel Checker")
                # Columns
                sLogCol = ""
                for sSheet in lstSheetImport:
                    resExcelHeader = ExcelTools.getHeadIndex(wb[sSheet])
                    lstExcelHeader = resExcelHeader.getData("header")
                    if resExcelHeader.isErr() or lstExcelHeader is None:
                        res.addResult(resExcelHeader)
                        return res
                    resCheck = DataTools.cmpLists(lstExcelHeader, self.__model.subsystem_getMapper().all_getExcelCols(sSheet, sIngestWf))
                    if resCheck.getData("istFehlt") is not None:
                        sErr = sSheet+": header not found: " + PrettyText.lst2strC(resCheck.getData("istFehlt"))
                        sLogCol += sErr + "\n"
                        res.addErr("Header not found", sheet=sSheet, add=PrettyText.lst2strC(resCheck.getData("istFehlt")), kat="Excel Header", sys="ExcelValidator")
                    if resCheck.getData("istDoppelt") is not None:
                        sErr = sSheet + ": found double headers: " + PrettyText.lst2strC(resCheck.getData("istDoppelt"))
                        sLogCol += sErr + "\n"
                        res.addErr(sErr, sheet=sSheet, kat="Excel Header", sys="ExcelValidator")
                    if resCheck.getData("istPlus") is not None:
                        sInfo = sSheet + ": surplus headers found: " + PrettyText.lst2strC(resCheck.getData("istPlus"))
                        sLogCol += sInfo + "\n"
                        res.addInfo("Surplus headers found", sheet=sSheet, add=PrettyText.lst2strC(resCheck.getData("istPlus")), kat="Excel Header", sys="ExcelValidator")
                if sLogCol != "":
                    sLog += "--- Columns ---\n" + sLogCol
        if len(lstSheetsNotAllowedAtAll) > 0:
            res.addInfo("Surplus sheets found", add=PrettyText.lst2strC(lstSheetsNotAllowedAtAll), kat="Excel Sheets", sys="ExcelValidator")

        sLog += "\n"
        res.addData("log", sLog)
        return res


