# coding=utf-8


# -------------------------------------------------
# TaxaPolishing
# -------------------------------------------------

import openpyxl, traceback, pickle, zipfile
from collections import Counter, OrderedDict
import urllib, os, sys, csv, getpass, platform, datetime, time, copy, random, psycopg2, psycopg2.extras
from difflib import SequenceMatcher
import openpyxl.styles.numbers
import warnings
import requests
from psycopg2 import sql
from tools.parallel import Stoppable
from tools.coding import Result, DataTools, ProgressString, WorkingObject
from tools.text import PrettyText
from tools.printing import SmartPrinter
from tools.excel import ExcelTools
from tools.fileio import FileTools
from tools.taxacruncher import TaxaCruncher
from shutil import copyfile
from tools.taxon import Taxon

class TaxaPolish(Stoppable, WorkingObject):

    def __init__(self, model):
        Stoppable.__init__(self)
        WorkingObject.__init__(self)

        # -- Member
        self.__model = model
        self.__smartprinter = model.ui_getSmartprinter()
        self.__map = model.subsystem_getMapper()
        self.__myworms = model.subsystem_getMyWorms()
        self.__wizard = model.subsystem_getTaxonWizard()

        # Db
        self.__conn = None
        self.__cursor = None

        # Subprozesse die bei einem Abort mit abgebrochen werden sollen
        self.addAbortChild(self.__myworms)

    # ----------------------------------------

    # Info
    def info(self):
        sOut = "*** Status of VirtualAssistant\n\n"
        return sOut

    def __reset(self):
        self.abort(False)
        self.__conn = None
        self.__cursor = None

    def start_polishing(self, sFileOrig, sFilePolished=None, **kwargs):
        """
        Ingest von Lookup Admin-Tables

        Args:
            file (Excel): Dateiname der zu polierenden Datei
        Kwargs:
            cursor: Db Cursor
            file_polished (str): Polished Output File
            sheet_biota (str): Biota Sheet
            col_aid (str): COlumn mit Aphia Ids
            col_sn (str): Column mit ScientificNames
            puffer (bool): Use Buffer for Name Queries
            fuzzy (bool): Use Fuzzy Name Queries
            marine_only (bool): Suche nur nach marinen Lebewesen
        Returns:
             Result
        """

        # Helper: Aufräumen
        def cleanUi():
            workflow()
            progress("off")

        def closeDb():
            if self.__model.conf_isDry() and conn is not None:
                conn.rollback()
            elif res.isErr():
                conn.rollback()
            else:
                conn.commit()
            cursor.close()
            conn.close()

        # Test Aid: Integer
        def checkAid(sAid):
            iAid = None
            try:
                iAid = int(sAid)
            except:
                pass
            return iAid

        def chooseTaxonName(t):
            """
            Hole denn Namen aus einem Taxon (Name -> Aname -> Vname -> PLATZHALTER)
            return:
                (str) Name, (str) Comment, (bool ok): Name, Kommentar, Ok - wenn FALSE dann wurde etwas anderes als der Scientifiname gewählt
            """
            if t.getScientificName() is not None:
                return t.getScientificName(), "used ScientificName of found taxon", True
            elif t.getScientificName_accepted() is not None:
                return t.getScientificName_accepted(), "found taxon: missing ScientificName -> used AcceptedScientificName", False
            elif t.getScientificName_valid() is not None:
                return t.getScientificName_valid(), "found taxon: missing ScientificName and AcceptedScientificName -> used ValidScientificName", False
            else:
                return "PLACEHOLDER", "found taxon has no names -> used placeholder", False


        # ---

        pn = self.__smartprinter.normal
        pv = self.__smartprinter.verbose
        pd = self.__smartprinter.debug

        progress = self.__smartprinter.progress
        status = self.__smartprinter.status
        workflow = self.__smartprinter.workflow
        userinput = self.__smartprinter.inputAsk__comboMulti
        getInputAnswer = self.__smartprinter.getInputAnswer

        # Abbruch Reset
        self.abort(False)
        res = Result(name="VirtualAssistant: Polish Taxa")
        pn("* Polishing Taxa")

        # Parameter
        bQueryUsePuffer = kwargs["puffer"] if ("puffer" in kwargs and kwargs["puffer"] in [True,False]) else True
        bQueryFuzzy = kwargs["fuzzy"] if ("fuzzy" in kwargs and kwargs["fuzzy"] in [True,False]) else False
        bQueryLike = kwargs["like"] if ("like" in kwargs and kwargs["like"] in [True,False]) else False
        bQueryMarineOnly = kwargs["marine_only"] if ("marine_only" in kwargs and kwargs["marine_only"] in (True, False)) else False
        sSheetBiota = kwargs["sheet_biota"] if ("sheet_biota" in kwargs and kwargs["sheet_biota"] is not None)else "Biota"
        sColAid = kwargs["col_aid"] if ("col_aid" in kwargs and kwargs["col_aid"] is not None) else "AphiaId"
        sColSn = kwargs["col_scientificname"] if ("col_scientificname" in kwargs and kwargs["col_scientificname"] is not None) else "ScientificName"
        cursor = kwargs["cursor"] if "cursor" in kwargs else None
        if cursor is None:
            bOwnConnection = True
        else:
            bOwnConnection = False

        # ---------

        if bOwnConnection:
            pd("Using own connection")
        else:
            pd("Reusing connection")

        # --- Check Parameter

        if sFileOrig is None:
            res.addErr("No Excel file given", sys="TaxaPolishing", kat="I/O")
        if res.isErr():
            cleanUi()
            return res

        workflow(top="TaxaPolishing", sub="Working", fmt=False)
        progress("on-pulse")

        # Eigene DB Connection
        if bOwnConnection:
            resConn = self.__model.subsystem_getDb().conn_getConnection()
            if resConn.isErr():
                res.addResult(resConn)
                cleanUi()
                return res
            # Connection: An
            conn = resConn.getData("connection")
            cursor = resConn.getData("cursor")

        # Log Total
        sLogTotal = "************************************\n" \
                    "TaxaPolishing\n" \
                    "************************************\n\n"
        sLogTotal += "File: " + str(sFileOrig) + "\n"
        sLogTotal += "Sheet Biota: " + str(sSheetBiota) + "\n"
        sLogTotal += "Column AphiaId: " + str(sColAid) + "\n"
        sLogTotal += "Column ScientificName: " + str(sColSn) + "\n"
        sLogTotal += "Use Buffer: " + ("yes" if bQueryUsePuffer else "no") + "\n"
        sLogTotal += "Fuzzy Special Name Search: " + ("yes" if bQueryFuzzy else "no") + "\n"
        sLogTotal += "Fuzzy Generic Name Search: " + ("yes" if bQueryLike else "no") + "\n"
        if self.__model.conf_isDry():
            sLogTotal += "Dry mode activated.\n"
        sLogTotal += "\n"

        # Hole Wb
        resWb = ExcelTools.getWb(sFileOrig, rw=True)
        if resWb.isErr():
            res.addResult(resWb, err=True)
            return res
        wb = resWb.getData("wb")

        # Hole Indizes der Spalte: AphiaId und Scientificname
        odicWb = OrderedDict()
        iIdxColAid = None
        iIdxColScientificname = None
        for sheet in wb:
            # Sheets
            sSheet = sheet.title
            if sSheet not in odicWb:
                odicWb[sSheet] = []
                # Header
                try:
                    header = wb[sSheet][1]
                    for cell in header:
                        if cell.value is not None:
                            sCol = str(cell.value)
                            odicWb[sSheet].append(sCol)
                            if sSheet == sSheetBiota:
                                if sCol == sColAid:
                                    iIdxColAid = cell.column -1 # AphiaId
                                elif sCol == sColSn:
                                    iIdxColScientificname = cell.column -1 # Scientificname
                except:
                    # Sheet ist leer
                    pass

        # Tests
        if sSheetBiota not in odicWb:
            res.addErr("Sheet with Biota data not not found", add=str(sSheetBiota) +" -> sheets found: "+PrettyText.lst2strC(odicWb.keys()), sys="TaxaPolishing", kat="I/O")
        else:
            if sColAid not in odicWb[sSheetBiota]:
                res.addErr("Column with Aphia Ids not found in Biota Sheet", add=str(sSheetBiota)+"."+str(sColAid)+" -> columns found: " + PrettyText.lst2strC(odicWb[sSheetBiota]), sys="I/O",
                           kat="TaxaPolishing")
            if sColSn not in odicWb[sSheetBiota]:
                res.addErr("Column with Scientificnames not found in Biota Sheet",
                           add=str(sSheetBiota)+"."+str(sColSn)+" -> columns found: " + PrettyText.lst2strC(odicWb[sSheetBiota]), sys="TaxaPolishing",
                           kat="TaxaPolishing")
        if None in [iIdxColAid, iIdxColScientificname]:
            res.addErr("Internal problem: Column Index for Aphia Ids/Acientificnames not available",
                       add=str(sSheetBiota) + "." + str(sColAid) + ": "+ str(iIdxColAid)+", "+str(sSheetBiota) + "." + str(sColSn) + ": "+ str(iIdxColScientificname), sys="TaxaPolishing",kat="I/O")

        # *** Excel: Biota holen
        # Hole alle Taxa Dataen aus EXCEL Tabelle
        # Teile in 3 Szenarien auf (AphiaId/Name)
        workflow(top="TaxaPolishing", sub="Extracting Taxa")
        pn("* Extracting Taxa Data")
        lstWb = []
        taxaCruncher = TaxaCruncher(self.__model)
        if res.isOk():
            progress("on-abort")
            iRowAll = len(list(wb[sSheetBiota].rows))
            for iRow, row in enumerate(wb[sSheetBiota]):
                iRow += 1
                if self.isAbort(): # Abort
                    res.addErr("Aborted by user...", sys="TaxaPolishing", kat="Abort")
                    break
                if iRow > 1: # Header
                    progress(iRow, iRowAll, 100, proz=True, abs=True, cli=True)
                    if ExcelTools.isRowEmpty(row):
                        # Zeile leer
                        res.addInfo("Empty row", sheet=sSheetBiota, row=iRow, sys="TaxaPolishing", kat="I/O")
                    else:
                        # Zeile nicht leer
                        sAid = row[iIdxColAid].value
                        sSn = row[iIdxColScientificname].value
                        taxaCruncher.addTaxon(sAid, sSn, iRow)
                        lstWb.append((sAid, sSn))

        # *** TeaxCruncher
        dicName2taxa = {}
        if res.isOk():
            resTaxaList = taxaCruncher.crunch(fuzzy=bQueryFuzzy, like=bQueryLike, puffer=bQueryUsePuffer, marine_only=bQueryMarineOnly)
            res.addResult(resTaxaList, msg_only=True)
            sLogTaxaList = resTaxaList.getData("log")
            taxa = resTaxaList.getData("taxa")
            dicName2taxa = resTaxaList.getData("name2taxa")
            if sLogTaxaList is not None:
                sLogTotal += sLogTaxaList

        # Fülle Lücken in Excel
        progress("on-pulse")
        workflow(top="TaxaPolishing", sub="Fill in gaps")
        for iRow, row in enumerate(wb[sSheetBiota]):
            iRow += 1
            if iRow > 1: # Header
                progress(iRow, iRowAll, 100, proz=True, abs=True, cli=True)
                if ExcelTools.isRowEmpty(row):
                    pass
                else:
                    # Zeile nicht leer
                    sAid = row[iIdxColAid].value
                    sSn = row[iIdxColScientificname].value
                    sAid_clean = None
                    sSn_clean = None
                    try:
                        iAid = int(sAid)
                    except:
                        iAid = None
                    if iAid is not None and sSn is None:
                        # *** Aid
                        if self.__wizard.isSpecialTaxon(iAid):
                            # Special
                            sSn_clean = self.__wizard.getSpecialTaxonName(iAid)
                        else:
                            # Normal
                            taxon = taxa.getTaxon(iAid)
                            if taxon is not None:
                                sSn_clean = str(taxon.getScientificName())
                    elif (iAid is None) and (sSn is not None) and (sSn in dicName2taxa):
                        # *** Sn
                        taxon = dicName2taxa[sSn]
                        sAid_clean = str(taxon.getAphiaId())
                    # Reinsurance Check
                    pass
                    # Eintragen
                    if sAid_clean is not None: row[iIdxColAid].value = sAid_clean
                    if sSn_clean is not None: row[iIdxColScientificname].value = sSn_clean

        # Alle Warnungen vor das Log stellen
        if res.isAnyMessage():
            sLogTotal = res.getErrWarnInfoStringHeader() + res.getErrWarnInfoString() + res.getErrWarnInfoStringFooter() + sLogTotal
        sLogTotal = PrettyText.clean_wizard(sLogTotal, max_one_empty_row=True)

        # Write Excel
        pn("* Writing Excel File")
        if sFilePolished is None:
            # Update Original File (no polished file given)
            resBackup = FileTools.prepareSaveFile(sFileOrig)
            if resBackup.isErr():
                res.addResult(resBackup, sys="TaxaPolishing", kat="I/O", msg_only=True)
            else:
                try:
                    wb.save(filename=sFileOrig)
                except Exception as ex:
                    res.addErr("Error during saving Excel file", add=str(ex), sys="TaxaPolishing", kat="I/O")
        else:
            # Own Polished File Given
            resBackup = FileTools.prepareSaveFile(sFilePolished)
            if resBackup.isErr():
                res.addResult(resBackup, sys="TaxaPolishing", kat="I/O", msg_only=True)
            else:
                try:
                    wb.save(filename=sFilePolished)
                except Exception as ex:
                    res.addErr("Error during saving Excel file", add=str(ex), sys="TaxaPolishing", kat="I/O")

        # Write Log
        pn("* Writing Log File")
        if sFilePolished is None:
            sFile_log = FileTools.getFileWithoutType(sFileOrig)+"_log.txt"
        else:
            sFile_log = FileTools.getFileWithoutType(sFilePolished) + "_log.txt"
        FileTools.prepareSaveFile(sFile_log)
        try:
            with open(sFile_log, "w") as f:
                f.write(sLogTotal)
        except Exception as ex:
            pass

        # Antwort
        res.addData("title", "TaxaPolishing")
        res.addData("text", sLogTotal)

        # Save MyWoRMS
        self.__myworms.puffer_save()

        # Aufräumen
        cleanUi()
        if bOwnConnection:
            closeDb()

        return res

