"""
Modul mit verschiedensten Hilfsklassen zum Umgang mit Excel-Daten
"""
# coding=utf-8

import openpyxl, warnings
from collections import OrderedDict
from tools.coding import Result
from tools.text import PrettyText

class ExcelNavigator():
    """
    Wrapper für Excel Sheets
    """

    # Konstruktor
    def __init__(self, wb):

        self.__wb = wb
        self.__lstSheets = []
        self.__odicHeaders = OrderedDict()
        self.__dicRowMax = {}
        self.__dicIdx = {}
        self.__dicIdxRev = {}
        self.__resInit = self.__initialize()

    def __str__(self):
        # print(self.__dicIdx)
        # print(self.__dicIdxRev)
        # print(self.__lstSheets)
        # print(self.__odicHeaders)
        # print(self.__dicRowMax)

        sOut = "* Excel workbench\n"
        if self.isOk():
            sOut += "Sheets: " + PrettyText.lst2strC(self.__lstSheets) +"\n"
            for sSheet in self.getSheets():
                sOut += sSheet +" (rows: "+ str(self.getMaxRow(sSheet)) +"): " + PrettyText.lst2strC(self.getHeader(sSheet)) +"\n"
            sOut += "Total rows: " +str(self.getMaxRow()) + "\n"
        else:
            sOut += "Not working: " +self.getErr()
            return sOut
        return sOut

    def __initialize(self):
        res = Result("Excel navigator")

        for sheet in self.__wb:
            sSheet = sheet.title
            if sSheet not in self.__lstSheets:
                # Sheets
                self.__lstSheets.append(sSheet)
            else:
                res.addErr("Multiple sheet names found in workbench", kat="Sheet Structure", sys="Excel Tools")

            # Max Row
            self.__dicRowMax[sSheet] = sheet.max_row - 1
            # Index
            head = sheet[1]
            lstHeader = []
            dicIdx = {}
            dicIdxRev = {}
            lstMulti = []
            for idx, col in enumerate(head):
                if col.value is not None:
                    if col.value in lstHeader:
                        if col.value not in lstMulti:
                            lstMulti.append(col.value)
                    lstHeader.append(col.value)
                    dicIdx[idx] = col.value
                    dicIdxRev[col.value] = idx


            if len(lstMulti) > 0:
                res.addErr("Multiple header found", add= PrettyText.lst2strC(lstMulti), sheet=str(sheet.title), kat="Sheet Structure", sys="Excel Tools")

            self.__dicIdx[sSheet] = dicIdx
            self.__dicIdxRev[sSheet] = dicIdxRev
            # Header
            self.__odicHeaders[sSheet] = lstHeader
        return res

    def isErr(self):
        return not self.isOk()

    def isOk(self):
        return self.__resInit.isOk()

    def getErr(self):
        return self.__resInit.getErr()

    def isEmpty (self):
        if self.getMaxRow() > 0:
            return False
        else:
            return True

    def getSheets(self):
        try:
            return self.__lstSheets
        except:
            return None

    def getHeader(self, sSheet):
        try:
            return list(self.__odicHeaders.key())
        except:
            return None

    def getHeaderIdx(self, sSheet, sHead):
        try:
            return None
        except:
            return None

    def getMaxRow(self,sSheet=None):
        if sSheet is None:
            # Total
            iRowTotal = 0
            for iRow in self.__dicRowMax.items():
                iRowTotal += 1
            return iRowTotal
        else:
            # Ein Sheet
            try:
                return self.__dicRowMax[sSheet]
            except:
                return None

    def getHeadIdx(self, sHead):
        pass

    def getHead(self, idx):
        pass


class ExcelTools:
    """
    Hilfsklasse zum Umgang mit Excel-Daten
    """

    # Get Workbook
    def getWb(sFile, **kwargs):
        res = Result("Get Excel Workbook for file: "+str(sFile))
        bRw = kwargs["rw"] if "rw" in kwargs else False
        if sFile is None:
            res.addErr("No file given", sys="Excel Tools", kat="I/O")
        else:
            try:
                warnings.simplefilter("ignore")  # Ignoriere Warnungen
                wb = openpyxl.load_workbook(sFile, data_only=True, read_only=not bRw)  # Keine Formeln sondern letzter gespeicherter Wert
                warnings.simplefilter("default")  # Warnungen wieder aktivieren
                res.addData("wb",wb)
            except Exception as ex:
                res.addErr("Structural problems found in EXCEL file" , add=str(ex), sys="Excel Tools", kat="I/O")
        return res


    # Ist ein Sheet leer
    def isSheetEmpty (sheet):
        """
        Ist ein Excel-Sheet leer?
            Args:
                sheet (openpyxl.worksheet.worksheet): Excel Sheet
            Return:
                result (tools.coding.Result): header (odic:head->idx)
        """
        head = sheet[1]
        if len(head) == 1 and head[0].value is None:
            return True
        else:
            return False

    def isRowEmpty(row):
        if row is None:
            return None
        bEmptyRow = True
        for cell in row:
            if cell.value is not None:
                bEmptyRow = False
                break
        return bEmptyRow

    # Get Rows max
    def getRowsTotal(wb):
        res =Result("Get number of rows")
        dicRows = {}
        iRowTotal = 0
        if wb is None:
            res.addErr("No workbench given", kat="Configuration", sys="Excel Tools")
            return res
        for sheet in wb:
            iRow = sheet.max_row
            if iRow != 0:
                dicRows[sheet.title] = sheet.max_row - 1
            else:
                iRow = 0
            dicRows[sheet.title] = iRow
            iRowTotal += iRow
        res.addData("rows", dicRows)
        res.addData("total", iRowTotal)
        return res

    # Get Header
    def getHeadIndex(sheet):
        res = Result(name="Get Excel header")
        if ExcelTools.isSheetEmpty(sheet):
            res.addInfo("sheet is empty", sys="Excel Tools", kat="I/O")
            return res
        head = sheet[1]
        lstHeader = []
        dicIndex = {}
        dicIndexRev = {}
        lstMulti = []
        for idx, col in enumerate(head):
            if col.value is not None:
                if col.value in lstHeader:
                    if col.value not in lstMulti:
                        lstMulti.append(col.value)
                lstHeader.append(col.value)
                dicIndex[idx] = col.value
                dicIndexRev[col.value] = idx
        if len(lstMulti) > 0:
            res.addErr("Multiple header found", sheet=sheet.title, add=PrettyText.lst2strC(lstMulti), kat="Workbook Structure", sys="Excel Tools")
        res.addData("header", lstHeader)
        res.addData("index", dicIndex)
        res.addData("indexRev", dicIndexRev)
        return res

    # Transform row to dictionary
    def row2dic(row, dicIndex2name, **kwargs):
        """
        Args:
            row: Row
            dicIndex2name: Index
        kwargs:
             conv_2_str: Konvertiere alle Werte in einen String
        """
        bConvert2String = True if "conv_val_2_str" in kwargs and kwargs["conv_val_2_str"] is True else False
        odicRow = OrderedDict()
        try:
            for idx, cell in enumerate(row):
                if idx in dicIndex2name:
                    sColName = dicIndex2name[idx]
                    if bConvert2String:
                        sVal = str(cell.value) if cell.value is not None else cell.value
                        odicRow[sColName] = sVal
                        #print(odicRow)
                    else:
                        odicRow[sColName] = cell.value
            return odicRow
        except Exception as ex:
            return None






