'''
# Create virtual env
python3 -m venv env

# Activate virtual env
source env/bin/activate

# Install new dependencies
python3 -m pip install requests

# (Optional) Write requirements
python3 -m pip freeze > requirements.txt

# Load requirements
python3 -m pip install < requirements.txt
'''

import getpass
import json
import os.path
import psycopg2
import requests

overwrite=False


host = "localhost"
database = "critterbase"
schema = "critterbase"
print('Database credentials for '+schema+'.'+database+' @ '+host+')')
username = input('Username: ')
password = getpass.getpass('Password: ')

# download FAO

if not overwrite or not os.path.isfile('fao.json'):
    r = requests.get('http://www.fao.org/fishery/geoserver/fifao/ows?service=WFS&request=GetFeature&version=1.0.0&typeName=fifao:FAO_AREAS_CWP&outputFormat=json')
    with open('fao.json', 'w') as f:
        f.write(r.text)


# open database connection and recreate regions table

conn = psycopg2.connect(host=host, database=database, user=username, password=password)
cur = conn.cursor()

sql = \
'''
DROP TABLE IF EXISTS <schema>.region;
CREATE TABLE <schema>.region (id SERIAL PRIMARY KEY, code VARCHAR, name VARCHAR, level VARCHAR, geometry GEOMETRY);
'''
sql = sql.replace("<schema>", schema)
cur.execute(sql)

# parse and select geometries

with open('fao.json') as f:
    fao = json.load(f)

    for feature in fao['features']:
        properties = feature['properties']
        code = properties['F_CODE']
        name = properties['NAME_EN']
        level = properties['F_LEVEL']
        geom = feature['geometry']
        geometry = json.dumps(geom)

        # only harvest major names.
        if not level in ('MAJOR', 'SUBAREA'): continue

        # skip areas without a name
        if not name: continue

        print('{}\t{}\t{}'.format(level, code, name))

        cur.execute('INSERT INTO '+schema+'.region (code, name, level, geometry) VALUES (%s, %s, %s, ST_AsText(ST_GeomFromGeoJSON(%s)));', (code, name, level, geometry))

conn.commit()
conn.close()