"""
Modul mit Hilfsklassen, die bei der Arbeit mit Gereferenzen untersützt
"""

# coding=utf-8

import traceback, re, math
from tools.printing import SmartPrinter
from tools.coding import Result

class GeoCruncher:
    """
    KLASSE ZUM QUALTITÄSTESTEN UND KONVERTIEREN VON GEOREFERENZEN
    """

    # Konstruktor
    def __init__(self, sp=None):
        if sp is None:
            sp = SmartPrinter()

        # Member
        self.__smartprinter = sp
        self.__lstHemi = ["s", "n", "e", "o", "w"]

    # ---

    def getInfo(self):
        """
        Informationen zu Funktionalitäten
        """

        return """
Allowed input formats:

[NSWEO] D° M' S" [NSWEO]
[NSWEO] D° M' dS" [NSWEO]
[NSWEO] D° dM' [NSWEO]
[NSWEO] dD° [NSWEO]
[+-] dD

The hemisphere is mandatory (it may stand front or back).
When using decimal degrees, minutes or seconds, a decimal dot is mandatory.
(e.g.: N3 ° is not allowed, it should be N3.0 ° or just N3 ° 0'0 ")
General information: 1.2 is the same as 1.2°N, but in two different spellings.
"""

    def __removeWhitespaces(self, sGeo):
        return None if sGeo is None else re.sub(r'[\ \t]', "", str(sGeo))

    def __processChars(self, sGeo):
        """
        GeoRef String säubern (nur noch erlaubte Zeichen, transformiere °'" -> gms)

         Args:
            sGeo (str): Geo String

        Returns:
            (coding.Result) result:
                geo (str): aufbereiteter Geo String
        """

        res = Result(name="process chars")

        # Test: Unerlaubte Zeichen?
        sNotAllowedChars = re.sub(r'[0-9°\'\\."nswoe]', "", sGeo)
        if sNotAllowedChars != "":
            res.addErr("Found not allowed characters in geo string", add=str(sNotAllowedChars), kat="Characters", sys="Geo Cruncher")
            return res

        # Entferne unerlaubte Zeichen
        sGeo = re.sub(r'[^0-9°\'\\."nswoe]', "", sGeo)

        # Übersetze °"' in ungefährlichere GMS
        sGeo = re.sub(r'[\°]', "g", sGeo)
        sGeo = re.sub(r'[\']', "m", sGeo)
        sGeo = re.sub(r'[\"]', "k", sGeo)

        # Ergebnis
        res.addData("geo", sGeo)
        return res

    def __consumeHemisphere(self, sGeo):
        """
        Verarbeite Himmelsrichtung

        Args:
            sGeo (str): Geo String

        Returns:
            (coding.Result) result:
                geo (str): Geo String ohne Hemisphere
                hemi (str): ermittelte Hemisphere
                coor (str): ermittelte Coordinate
                dir (str): ermitteltes vorzeichen für das dD
        """

        res = Result(name="consume hemisphere")

        sHemisphere = re.sub(r'[^nswoe]', "", sGeo)

        # Test: Keine Himmelsrichtung angegeben
        if len(sHemisphere) == 0:
            res.addErr("No hemisphere code found", kat="Hemisphere", sys="Geo Cruncher")
            return res

        # Test: Mehr als eine Himmelsrichtung?
        if len(sHemisphere) > 1:
            res.addErr("Too many hemisphere codes found", add=str(sHemisphere), kat="Hemisphere", sys="Geo Cruncher")
            return res

        # Hemisphere ist nicht an Anfang oder Ende
        if sGeo[0] not in self.__lstHemi and sGeo[-1] not in self.__lstHemi:
            res.addErr("Hemisphere code found in wrong place", add=str(sHemisphere), kat="Hemisphere", sys="Geo Cruncher")
            return res

        # Entferne Hemisphere aus String
        sGeo = re.sub(r'[nswoe]', "", sGeo)

        # Ergebnis
        res.addData("hemi", sHemisphere)
        res.addData("geo", sGeo)
        if sHemisphere in ["s", "w"]:
            res.addData("dir", -1)
            res.addData("coor", "lat" if sHemisphere == "s" else "lon")
        else:
            res.addData("dir", 1)
            res.addData("coor", "lat" if sHemisphere == "n" else "lon")
        return res

    # Helper: Teste den korrekten Wertebereich
    def __isRangeOk(self, fWert, sCoordinate=None):
        """
        GeoRef Teste Wertebereich

         Args:
            fWert (float): Wert
            sCoordinate (str): Koordinate (lat/lon)

        Returns:
            (coding.Result) result
        """

        res = Result(name="Testing Range")
        if fWert is None:
            res.addErr("No data given", kat="Ranges", sys="Geo Cruncher")
        if (sCoordinate is None or sCoordinate == "lon") and math.fabs(fWert) <= 180:
            # Lon
            return res
        elif sCoordinate == "lat" and math.fabs(fWert) <= 90:
            # Lat
            return res
        else:
            if sCoordinate is None:
                res.addErr("Value is out of range (assuming to be lon)", add=str(fWert), kat="Range", sys="Geo Cruncher")
            else:
                res.addErr("Value is out of range", add=sCoordinate+": " + str(fWert), kat="Range", sys="Geo Cruncher")
            return res

    def crunch(self, sGeo, sCoordinate=None):
        """

        Konvertierung (mit Verifikation) eines GeoRef-Strings in das dD-System

        --- Hintergrund
        Lat (Breite) +-[0-90] (+N -S) Horizontale Schnitte (oben->unten)
        Lon (Länge) +-[0-180] (+E -W) Schnitte durch die Eedachse (um die Erde herum)

        --- Umrechungen
        D° M' S" (Grad/Min/Sek)
        1° = 60' = 3600"
        1' = 60"
        D = D + M/60 + S/3600

        --- Wertebereiche
        - Einzelne Maximalwerte
        Lat: D90°, M60', S60"
        Lon: D180°, M60', S60"
        - Gesamt Maximalwerte
        Lat: 90
        Lon: 180
        Je nachdem wie hoch die "vorherigen" werte sind, können nachfolgenden nicht maximal werden
        Bsp: Lat: 89° 60' 1" ist zuviel da Summe > 90

        --- Erlaubte Eingabeformate
        Die Hemispherenangabe is zwingend notwendig (sie darf vorn oder hinten stehen)
        Ein Plus ist nicht als Zeichen erlaubt
        Bei der Verwendung von Dezimal-Stunden/Minuten/Sekunden ist ein Dezimalpunkt zwingend
        (z.B. ist N3° nicht erlaubt, es müsste N3.0° lauten oder eben N3°0'0")
        Info: 1.2 ist zwar inhaltlich dasselbe wie 1.2°N, aber in zwei unterschiedlichen schreibweisen

        [NSWEO] D° M' S" [NSWEO]  (Klassisch Sexagesimal)
        [NSWEO] D° M' dS" [NSWEO]
        [NSWEO] D° dM' [NSWEO]
        [NSWEO] dD° [NSWEO]
        [-] dD

        --- Ausgabeformat:
        [-] dD (Float)

        Args:
            bLatLong (bool=None): Ist die Richtung bekannt?
            sGeo (str) : Georeferenz

        Returns:
            (tools.coding.Result) res:
                val (float): Wert in -dD
                format (str): Erkanntes Eingabeformat
        """

        """
        RegEx-Snippets
        
        regFloat = r'[-+]?([0-9]+(\.[0-9]+)?|\.[0-9]+)'

        """

        # ---

        res = Result(name="convert geo data")

        # *** Ergebnisse
        sGeoIn = None
        sGeoInHemisphere = None
        sGeoInCoordinateActual = None
        sGeoInCoordinateSetpoint = None
        iDirection = None
        fValue = None
        sFormat = None
        sInfo = None

        # Eingangstests

        if sGeo in ["", None]:
            res.addInfo("Nothing to convert", kat="Configuration", sys="Geo Cruncher")
            return res
        sGeoIn = self.__removeWhitespaces(sGeo)

        if sCoordinate is not None:
            if str(sCoordinate).lower() not in ["lat", "lon"]:
                res.addErr("Unknown type of coordinate", add=str(sCoordinate), kat="Coordinates", sys="Geo Cruncher")
                return res
            sGeoInCoordinateSetpoint = str(sCoordinate).lower()

        # *** Es liegt bereits dD-Format vor

        try:
            # Funktioniert eine einfache Float-Konversion?...
            fValue = float(sGeoIn)

            # ... funktioniert offensichtlich.

            # Teste Wertebereich
            resRange = self.__isRangeOk(fValue, sGeoInCoordinateSetpoint)
            if resRange.isErr():
                res.addResult(resRange)
                return res

            # Werte speichern
            res.addData("value", fValue)
            res.addData("format", "dd")

            # Info
            #sInfo = "in=" + sGeoIn + " -> detected format=dD -> out=" + str(fValue)
            #res.addInfo(sInfo, kat="Conversion", sys="Geo Cruncher")
            return res
        except:
            # ... funktioniert offensichtlich nicht.
            pass


        # *** Es liegt kein dd-Format vor

        # *** Eingabe aufbereiten
        sGeoWork = sGeoIn.lower()

        # Entferne nicht erlaubte Zeichen
        resProcess = self.__processChars(sGeoWork)
        if resProcess.isErr():
            res.addResult(resProcess)
            return res
        sGeoWork = resProcess.getData("geo")

        # Verarbeite Hemisphere
        resHemi = self.__consumeHemisphere(sGeoWork)
        if resHemi.isErr():
            res.addResult(resHemi)
            return res

        iDirection = resHemi.getData("dir")
        sGeoInHemisphere = resHemi.getData("hemi")
        sGeoInCoordinateActual = resHemi.getData("coor")
        sGeoWork = resHemi.getData("geo")

        # Teste ob ermittelte Hemisphere mit Vorgabe übereinstimmt?
        if sGeoInCoordinateSetpoint is not None and sGeoInCoordinateActual != sGeoInCoordinateSetpoint:
            res.addErr("Found hemisphere code does not match the expected coordinate", add=str(sGeoInHemisphere)+" != "+str(sGeoInCoordinateSetpoint), kat="Hemisphere", sys="Geo Cruncher")
            return res

        # --- Teste Formate
        sFormat = None

        # DMS
        erg = re.fullmatch(r'^[0-9]+g[0-9]+m[0-9]+k$', sGeoWork)
        if erg is not None:
            sFormat = "DMS"

        # DMdS
        if sFormat is None:
            erg = re.fullmatch(r'^[0-9]+g[0-9]+m([0-9]+(\.[0-9]+)|\.[0-9]+)k$', sGeoWork)
            if erg is not None:
                sFormat = "DMdS"

        # DdM
        if sFormat is None:
            erg = re.fullmatch(r'^[0-9]+g([0-9]+(\.[0-9]+)|\.[0-9]+)m$', sGeoWork)
            if erg is not None:
                sFormat = "DdM"

        # dD
        if sFormat is None:
            erg = re.fullmatch(r'^([0-9]+(\.[0-9]+)|\.[0-9]+)g$', sGeoWork)
            if erg is not None:
                sFormat = "dD"

        # Umrechnung
        if sFormat == "DMS":
            # *** DMS
            try:
                lstTmp1 = sGeoWork.split("g")
                lstTmp2 = lstTmp1[1].split("m")
                lstTmp3 = lstTmp2[1].split("k")
                iD = int(lstTmp1[0])
                iM = int(lstTmp2[0])
                iS = int(lstTmp3[0])
                fValue = iDirection * (iD + (iM /60) + (iS / 3600))
                #fValue = iDirection * (iD + ((iM + (iS / 60))/60))
            except:
                res.addErr("Conversion of detected format DMS failed", kat="Conversion", sys="Geo Cruncher")
            # Teste auf Wertebereich min, sek
            if iS > 60:
                res.addErr("seconds > 60", kat="Conversion", sys="Geo Cruncher")
            if iM > 60:
                res.addErr("minutes > 60", kat="Conversion", sys="Geo Cruncher")
            # Info
            sInfo = "in=" + str(sGeo) + " -> detected format=DMS -> hemi=" + str(sGeoInHemisphere) + ", deg=" + str(iD) + ", min=" + str(iM) + ", sec=" + str(iS) + " -> out=" + str(fValue)
            res.addInfo(sInfo, kat="Conversion", sys="Geo Cruncher")

        elif sFormat == "DMdS":
            # *** DMdS
            try:
                lstTmp1 = sGeoWork.split("g")
                lstTmp2 = lstTmp1[1].split("m")
                lstTmp3 = lstTmp2[1].split("k")
                iD = int(lstTmp1[0])
                iM = int(lstTmp2[0])
                fS = float(lstTmp3[0])
                fValue = iDirection * (iD + (iM / 60) + (fS / 3600))
                #fValue = iDirection * (iD + ((iM + (fS / 60)) / 60))
            except:
                res.addErr("Conversion of detected format DMdS failed", kat="Conversion", sys="Geo Cruncher")
            # Teste auf Wertebereich min, sek
            if fS > 60:
                res.addErr("seconds > 60", kat="Conversion", sys="Geo Cruncher")
            if iM > 60:
                res.addErr("minutes > 60", kat="Conversion", sys="Geo Cruncher")
            # Info
            sInfo = "in=" + str(sGeo) + " -> detected format=DMdS -> hemi=" + str(sGeoInHemisphere) + ", deg=" + str(
                iD) + ", min=" + str(iM) + ", dsec=" + str(fS) + " -> out=" + str(fValue)
            #res.addInfo(sInfo)

        elif sFormat == "DdM":
            # *** DdM
            try:
                lstTmp1 = sGeoWork.split("g")
                lstTmp2 = lstTmp1[1].split("m")
                iD = int(lstTmp1[0])
                fM = float(lstTmp2[0])
                fValue = iDirection * (iD + ((fM / 60)))
            except:
                res.addErr("Conversion of detected format DdM failed", kat="Conversion", sys="Geo Cruncher")
            # Teste auf Wertebereich min
            if fM > 60:
                res.addErr("minutes > 60", kat="Conversion", sys="Geo Cruncher")
            # Info
            sInfo = "in=" + str(sGeo) + " -> detected format=DdM -> hemi=" + str(sGeoInHemisphere) + ", deg=" + str(
                iD) + ", dmin=" + str(fM) + " -> out=" + str(fValue)
            #res.addInfo(sInfo)
        elif sFormat == "dD":
            # *** dD
            try:
                lstTmp1 = sGeoWork.split("g")
                fD = float(lstTmp1[0])
                fValue = iDirection * fD
            except:
                res.addErr("Conversion of detected format dD failed", kat="Conversion", sys="Geo Cruncher")
            # Info
            sInfo = "in=" + str(sGeo) + " -> detected format=dD -> hemi=" + str(sGeoInHemisphere) + ", ddeg=" + str(
                fD) + " -> out=" + str(fValue)
            #res.addInfo(sInfo)
        else:
            res.addErr("Unknown format", kat="Conversion", sys="Geo Cruncher")
            return res

        # Teste Wertebereich
        resRange = self.__isRangeOk(fValue, sGeoInCoordinateSetpoint)
        if resRange.isErr():
            res.addResult(resRange)
            return res

        # Antwort
        res.addData("value", fValue)
        res.addData("format", sFormat)
        return res