"""
Modul mit Hilfsklassen, zur Arbeit mit TaxaListen
"""

# coding=utf-8

import openpyxl, traceback, pickle, zipfile
from difflib import SequenceMatcher
from collections import Counter, OrderedDict
import urllib, os, sys, csv, getpass, platform, datetime, time, copy, random, psycopg2, psycopg2.extras
import openpyxl.styles.numbers
import warnings
import requests
from psycopg2 import sql
from tools.parallel import Stoppable
from tools.coding import Result, DataTools, ProgressString, WorkingObject
from tools.text import PrettyText
from tools.taxa import Taxa
from tools.printing import SmartPrinter
from tools.excel import ExcelTools
from tools.fileio import FileTools
from shutil import copyfile

class TaxaCruncher(Stoppable):
    """
    TaxaList
    """

    def __init__(self, model):
        super().__init__()
        self.__model = model
        self.__smartprinter = model.ui_getSmartprinter()
        self.__myworms = model.subsystem_getMyWorms()
        self.__map = model.subsystem_getMapper()
        self.__wizard = model.subsystem_getTaxonWizard()
        self.reset()

    def reset(self):
        # Parameter
        self.__bFuzzy = None
        self.__bLike = None
        self.__bPuffer = None
        self.__bMarineOnly = None
        self.__bHTML = False

        # Log
        self.__sLog = None

        # Eingesammelte Daten
        self.__setAid = set()
        self.__odicAidSn = OrderedDict()
        self.__setSn = set()

        # In MyWoRMS
        self.__setAid_myworms_not = set()
        self.__setAid_myorms_ok = set()
        self.__setAid_myworms_old = set()
        self.__setAid_myworms_special = set()
        self.__setAid_myworms_special_notFound = set()

        # WoRMS Anfragen
        # Search by AphiaId
        self.__lstWorms_searchByAid_notFound = []
        # Search by Name
        self.__setWorms_searchByName = set()
        self.__setWorms_searchByName_foundInPuffer = set()
        self.__setWorms_searchByName_aphiaIds = set()
        self.__dicWorms_searchByName_all = {}
        self.__dicWorms_searchByName_answer_single = {}
        self.__dicWorms_SearchByName_answer_multi = {}
        self.__dicWorms_SearchByName_answerChoices_multi_userSelection = {}
        self.__setWorms_searchByName_notFound = set()

        # Ein: Taxa Liste
        self.__lstTaxa_user = []

        # Aus: Taxon Menge
        self.__taxa_final = Taxa(self.__model)

    def addTaxon(self, sAid, sSn, iRow=None):
        if (sAid, sSn) == (None,None):
            return
        else:
            self.__lstTaxa_user.append((sAid, sSn, iRow))

    def crunch(self, **kwargs):
        """
        TaxaCruncher: Verarbeite User Taxon Liste -> Protokoll -> Liste von Taxa Objekten
        Kwargs:
            puffer (bool): Use Buffer for Name Query
            fuzzy (bool): Use Special Fuzzy Name Query
            fuzzy (bool): Use normal Like Fuzzy Name Query
            marine_only (bool): Suche nur nach marinen Lebewesen
            html (bool): HTML Log
        Return:
             Result:
                taxa (lst): Taxa
                search (dic): ScientificNamen -> Taxa
        """

        res = Result(name="TaxaCruncher")

        # Parameter
        self.__bFuzzy = kwargs["fuzzy"] if ("fuzzy" in kwargs and kwargs["fuzzy"] in [True, False]) else False
        self.__bLike = kwargs["like"] if ("like" in kwargs and kwargs["like"] in [True, False]) else True
        self.__bPuffer = kwargs["puffer"] if ("puffer" in kwargs and kwargs["puffer"] in [True, False]) else True
        self.__bMarineOnly = kwargs["marine_only"] if ("marine_only" in kwargs and kwargs["marine_only"] in [True, False]) else True
        self.__bHTML = kwargs["html"] if ("html" in kwargs and kwargs["html"] in [True, False]) else False

        # ---

        self.__sLog = ""
        resAidSn = self.__readTaxaList()
        res.addResult(resAidSn)

        resNameQueries = self.__updateMyWorms_nameQueries()
        res.addResult(resNameQueries)

        resUpdate = self.__updateMyWorms_taxa()
        res.addResult(resUpdate)

        resVerify = self.__verifyTaxaList()
        res.addResult(resVerify)

        res.addData("log", self.__sLog)
        res.addData("taxa", self.__taxa_final)
        res.addData("name2taxa", self.__dicWorms_searchByName_all)
        return res

    def __verifyTaxaList(self):
        pn = self.__smartprinter.normal
        pv = self.__smartprinter.verbose
        pd = self.__smartprinter.debug
        progress = self.__smartprinter.progress
        status = self.__smartprinter.status
        workflow = self.__smartprinter.workflow
        userinput = self.__smartprinter.inputAsk__comboMulti
        getInputAnswer = self.__smartprinter.getInputAnswer

        # Test Aid: Integer
        def checkAid(sAid):
            iAid = None
            try:
                iAid = int(sAid)
            except:
                pass
            return iAid

        # ---

        res = Result(name="TaxaCruncher: Verify")
        lstVerify_add_aid = []
        lstVerify_add_sn = []
        lstVerify_add_specialname = []
        lstVerify_aidSn_normalTaxon_namediff = []
        lstVerify_aidSn_specialTaxon_namediff = []
        lstVerify_aidSn_notFound = []
        bAction = False

        progress("on-abort")
        workflow(top="TaxaCruncher", sub="Verify Taxa List", fmt=False)
        pn("* Verify Taxon List", fmt=False)
        lstTaxa_fresh = []
        iNr_all = len(self.__lstTaxa_user)
        for iNr, (sAid, sSn, iRow_info) in enumerate(self.__lstTaxa_user):
            iNr += 2
            progress(iNr, iNr_all, 100, proz=True, abs=True, cli=True)
            if self.isAbort():  # Abort
                res.addErr("Aborted by user...", sys="TaxaCruncher", kat="Abort")
                break
            # Szenarien
            if (sAid not in ("", None)) and (sSn in ("", None)):
                # --- Aid
                iAid = checkAid(sAid)
                if iAid is None:
                    # Aid ist kein Integer -> Das hääte schon vorher entdeckt werden müssem
                    res.addErr("AphiaId is not integer", row=iRow_info, add=str(iAid), sheet="sheet", col="aphiaid", sys="TaxaCruncher", kat="Verify Input List")
                elif iAid < 0:
                    # Special
                    sSpecial_name = self.__model.subsystem_getTaxonWizard().getSpecialTaxonName(iAid)
                    if sSpecial_name is None:
                        res.addErr("Special Taxa Name not found", row=iRow_info, add=str(iAid), sheet="sheet", col="aphiaid", sys="TaxaCruncher", kat="Verify Input List")
                    else:
                        lstVerify_add_specialname.append((str(iRow_info), str(iAid), str(sSpecial_name)))
                        bAction = True
                else:
                    # Normal
                    taxon = self.__taxa_final.getTaxon(iAid)
                    if taxon is None:
                        res.addErr("Taxon not found by AphiaId", row=iRow_info, add=str(iAid), sheet="sheet", col="aphiaid", sys="TaxaCruncher", kat="Verify Input List")
                    else:
                        sTaxon_sn = taxon.getScientificName()
                        lstVerify_add_sn.append((str(iRow_info), str(iAid), str(sTaxon_sn)))
                        bAction = True
            elif (sAid in ("", None)) and (sSn not in ("", None)):
                # --- Sn
                if sSn not in self.__dicWorms_searchByName_all:
                    res.addErr("Taxon not found by Name", row=iRow_info, add=str(sSn), sheet="sheet", col="aphiaid", sys="TaxaCruncher", kat="Verify Input List")
                else:
                    taxon = self.__dicWorms_searchByName_all[sSn]
                    if taxon is None:
                        res.addErr("Taxon not found by Name", row=iRow_info, add=str(sSn), sheet="sheet", col="aphiaid", sys="TaxaCruncher", kat="Verify Input List")
                    else:
                        iTaxon_aid = taxon.getAphiaId()
                        sTaxon_sn = taxon.getScientificName()
                        lstVerify_add_aid.append((str(iRow_info), str(sSn), str(iTaxon_aid)))
                        bAction = True
                        if SequenceMatcher(None, sSn,sTaxon_sn).ratio() < self.__wizard.getScientificNameMatchRatio():
                            res.addWarn(
                                "Reinsurance Name Check: ScientificName (given) differs Too Much From Norm",
                                row=iRow_info,
                                sheet="Biota",
                                col="ScientificName",
                                add=str(iTaxon_aid) + ": " + sSn + " != " + sTaxon_sn,
                                sys="TaxonSpotter", kat="Checking Taxa")
            elif (sAid not in ("", None)) and (sSn not in ("", None)):
                # --- Aid/Sn
                iAid = checkAid(sAid)
                if iAid < 0:
                    # Special Taxon
                    if self.__model.subsystem_getTaxonWizard().isSpecialTaxon(iAid):
                        # Sonderfall: Spezial-Taxa
                        sScientificNameSpecial = self.__model.subsystem_getTaxonWizard().getSpecialTaxonName(iAid)
                        if SequenceMatcher(None, sSn,sScientificNameSpecial).ratio() < self.__wizard.getScientificNameMatchRatio():
                            res.addWarn(
                                "Reinsurance Name Check: Special Taxa Name (given) differs Too Much From Norm",
                                row=iRow_info,
                                sheet="Biota",
                                col="ScientificName",
                                add=str(iAid) + ": " + sSn + " != " + sScientificNameSpecial,
                                sys="TaxonSpotter", kat="Checking Taxa")
                            lstVerify_aidSn_specialTaxon_namediff.append((str(iRow_info), str(sSn), str(sScientificNameSpecial)))
                            bAction = True
                    else:
                        res.addErr("Reinsurance Name Check: Special Taxa Aid not known", row=iRow_info, sheet="Biota",col="AphiaId given", add=str(iAid), sys="TaxaCruncher", kat="Verify Input List")
                else:
                    # Normales Taxon
                    taxon = self.__taxa_final.getTaxon(iAid)
                    if taxon is None:
                        res.addErr("Taxon not found by AphiaId", row=iRow_info, add=str(iAid), sheet="sheet", col="aphiaid",sys="TaxaCruncher", kat="Verify Input List")
                    else:
                        # Teste alle Namen auf relative Ähnlichkeit mit MyWoRMS-Eintrag
                        if taxon.getScientificName() is None:
                            lstVerify_aidSn_normalTaxon_namediff.append((str(iRow_info) , str(sSn), "Taxon in MyWoRMS has no ScientificName"))
                            bAction = True
                        elif SequenceMatcher(None, sSn, taxon.getScientificName()).ratio() < self.__model.subsystem_getTaxonWizard().getScientificNameMatchRatio():
                            res.addWarn(
                                "Reinsurance Name Check: ScientificName (given) differs Too Much From Norm",
                                row=iRow_info,
                                sheet="Biota",
                                col="ScientificName",
                                add=str(iAid) + ": " + sSn + " != " + taxon.getScientificName(),
                                sys="TaxonSpotter", kat="Checking Taxa")
                            lstVerify_aidSn_normalTaxon_namediff.append((str(iRow_info),str(sSn),str(taxon.getScientificName())))
                            bAction = True

        # *** Log
        if bAction:
            self.__sLog += "*** Verify Taxon List\n\n"

        # ScientificName added
        if len(lstVerify_add_sn) > 0:
            self.__sLog += "* ScientificName added (" + str(len(lstVerify_add_sn)) + "x)\n"
            lstTab = [["Row","AphiaId","ScientificName"]]
            for sRow, sAid, sSn in lstVerify_add_sn:
                lstTab.append((str(sRow), str(sAid), str(sSn)))
            self.__sLog += PrettyText.table(lstTab, True, self.__bHTML) + "\n"
            self.__sLog += "\n"

        # SpecialName added
        if len(lstVerify_add_specialname) > 0:
            self.__sLog += "* Special Taxa Name added (" + str(len(lstVerify_add_specialname)) + "x)\n"
            lstTab = [["Row","Special Taxon Id","Special Taxon Name"]]
            for sRow, sAid, sSn in lstVerify_add_specialname:
                lstTab.append((str(sRow), str(sAid), str(sSn)))
            self.__sLog += PrettyText.table(lstTab, True, self.__bHTML) + "\n"
            self.__sLog += "\n"

        # AphiaId added
        if len(lstVerify_add_aid) > 0:
            self.__sLog += "* AphiaId added (" + str(len(lstVerify_add_aid)) + "x)\n"
            lstTab = [["Row","ScientificName","AphiaId"]]
            for sRow, sAid, sSn in lstVerify_add_aid:
                lstTab.append((str(sRow), str(sAid), str(sSn)))
            self.__sLog += PrettyText.table(lstTab, True, self.__bHTML) + "\n"
            self.__sLog += "\n"

        # Reinsurance Name Check: ScientificName differs too much from MyWoRMS
        if len(lstVerify_aidSn_normalTaxon_namediff) > 0:
            self.__sLog += "* ScientificName (given) differs too much from MyWoRMS (" + str(len(lstVerify_aidSn_normalTaxon_namediff)) + "x)\n"
            lstTab = [["Row","ScientificName (given)","ScientificName (MyWoRMS)"]]
            for sRow, sSn_given, sSn in lstVerify_aidSn_normalTaxon_namediff:
                lstTab.append((str(sRow), str(sSn_given), str(sSn)))
            self.__sLog += PrettyText.table(lstTab, True, self.__bHTML) + "\n"
            self.__sLog += "\n"

        # Reinsurance Name Check: SpecialName differs too much from MyWoRMS
        if len(lstVerify_aidSn_specialTaxon_namediff) > 0:
            self.__sLog += "* Special Taxa Name (given) differs Too Much From Norm ( " + str(len(lstVerify_aidSn_specialTaxon_namediff)) + "x)\n"
            lstTab = [["Row","Special Taxon Name (given)","Special Taxon Name (TaxonWizard)"]]
            for sRow, sSn_given, sSn  in lstVerify_aidSn_specialTaxon_namediff:
                lstTab.append((str(sRow), str(sSn_given), str(sSn)))
            self.__sLog += PrettyText.table(lstTab, True, self.__bHTML) + "\n"
            self.__sLog += "\n"

        return res

    def __readTaxaList(self):
        """
        Sammle die AphiaIds und ScientificNames aus Eingangsdaten
        """

        # --- Helper
        # Test Aid: Integer
        def checkAid(sAid):
            iAid = None
            try:
                iAid = int(sAid)
            except:
                pass
            return iAid

        # ---

        self.abort(False)
        res = Result(name="TaxonCruncher: Gather AphiaIds and ScientificNames")

        pn = self.__smartprinter.normal
        pv = self.__smartprinter.verbose
        pd = self.__smartprinter.debug
        progress = self.__smartprinter.progress
        status = self.__smartprinter.status
        workflow = self.__smartprinter.workflow
        userinput = self.__smartprinter.inputAsk__comboMulti
        getInputAnswer = self.__smartprinter.getInputAnswer

        # *** Teile in 3 Szenarien auf (AphiaId/Name)
        # Taxa Szenarien
        iRecTotal = 0

        # ---

        if self.__lstTaxa_user is None:
            res.addErr("No Taxa List given", sys="TaxaCruncher", kat="Configuration")
        workflow(top="TaxaCruncher", sub="Read Taxa List", fmt=False)
        if res.isOk():
            pn("* Collecting AphiaIds and ScientificNames", fmt=False)
            progress("on-abort")
            iNr_all = len(self.__lstTaxa_user)
            for iNr, tp in enumerate(self.__lstTaxa_user):
                if self.isAbort():  # Abort
                    res.addErr("Aborted by user...", sys="TaxaCruncher", kat="Abort")
                    break
                progress(iNr, iNr_all, 100, proz=True, abs=True, cli=True)
                iRecTotal += 1
                sAid = tp[0]
                sSn = tp[1]
                sRow_info = tp[2]
                if sAid not in ["", None] and sSn in ["", None]:
                    # *** Nur Aphia Id
                    iAid = checkAid(sAid)
                    if iAid is None:
                        res.addErr("AphiaId is no integer", sheet="sheet", col="col", row=sRow_info, add=str(sAid),sys="TaxaCruncher", kat="Collecting taxa data")
                    else:
                        # AphiaId ist OK
                        self.__setAid.add(iAid)
                elif sAid in ["", None] and sSn not in ["", None]:
                    # *** Nur Scientificname
                    self.__setSn.add(sSn)
                elif sAid not in ["", None] and sSn not in ["", None]:
                    # *** Beides gegeben
                    iAid = checkAid(sAid)
                    if iAid is None:
                        res.addErr("AphiaId is no integer", sheet="sheet", col="col",  row=sRow_info, add=str(sAid), sys="TaxaCruncher", kat="Collecting taxa data")
                    else:
                        # AphiAid_given ist OK
                        if iAid not in self.__odicAidSn:
                            self.__odicAidSn[iAid] = []
                        if sSn not in self.__odicAidSn[iAid]:
                            self.__odicAidSn[iAid].append(sSn)
                else:
                    # *** Beides nicht (kann eigentlich nicht sein)
                    res.addErr("Either AphiaId nor ScientificName is set", sheet="sheet", col="col", row=sRow_info, sys="TaxaCruncher", kat="Collecting taxa data")

            # Log
            self.__sLog += "*** Overview\n"
            self.__sLog += "* Found Taxa Scenarios\n"
            lstTab = [("Scenario", "Counts")]
            if len(self.__setAid) > 0:
                lstTab.append(("Unique AphiaIds only", str(len(self.__setAid)) + "x"))
            if len(self.__setSn) > 0:
                lstTab.append(("Unique ScientificNames only", str(len(self.__setSn)) + "x"))
            if len(self.__odicAidSn) > 0:
                lstTab.append(("Unique AphiaIds in combination with ScientificName" , str(len(self.__odicAidSn)) + "x"))
                iMultiSn = 0
                for iAid, lstSn in self.__odicAidSn.items():
                    if len(lstSn) > 1:
                        iMultiSn += 1
                if iMultiSn > 0:
                    lstTab.append(("Unique AphiaIds in combination with multiple ScientificNames" , str(iMultiSn) + "x"))
            self.__sLog += PrettyText.table(lstTab, True, self.__bHTML) + "\n"
        return res

    def __updateMyWorms_nameQueries(self):
        """
        Bearbeite Anfragen über Namen
        """

        # --- Helper
        # Test Aid: Integer
        def checkAid(sAid):
            iAid = None
            try:
                iAid = int(sAid)
            except:
                pass
            return iAid

        # ---

        self.abort(False)
        res = Result(name="TaxaCruncher: Process MyWoRMS - NameQueries")

        pn = self.__smartprinter.normal
        pv = self.__smartprinter.verbose
        pd = self.__smartprinter.debug
        progress = self.__smartprinter.progress
        status = self.__smartprinter.status
        workflow = self.__smartprinter.workflow
        userinput = self.__smartprinter.inputAsk__comboMulti
        getInputAnswer = self.__smartprinter.getInputAnswer


        progress("on-pulse")
        # --- Ermittle alle Queries, die erneuert werden müssen
        if len(self.__setSn) > 0:
            self.__sLog += "*** Process Name Queries\n\n"
            workflow(top="TaxaCruncher", sub="Update Name Queries", fmt=False)
            for sSn in self.__setSn:
                if self.__bPuffer:
                    # Benutze Puffer für Name Queries
                    bFound = self.__myworms.isQueryInPuffer(sSn)
                    if bFound:
                        # Name in Puffer gefunden
                        self.__setWorms_searchByName_foundInPuffer.add(sSn)
                    else:
                        # Name Nicht Gefunden
                        self.__setWorms_searchByName.add(sSn)
                else:
                    # Benutze keinen Puffer für Name Queries
                    self.__setWorms_searchByName.add(sSn)  # Alle Namen Suchen

            # Log
            if len(self.__setSn) > 0:
                pn("* Queries by ScientificName", fmt=False)
                self.__sLog += "* Queries by ScientificName\n"
                lstTab = [("Message", "Counts")]
                lstTab.append(("Missing Aphia Ids", str(len(self.__setSn)) + "x"))
                if len(self.__setWorms_searchByName_foundInPuffer) > 0:
                    lstTab.append(("Queries found in buffer", str(len(self.__setWorms_searchByName_foundInPuffer)) + "x"))
                self.__sLog += PrettyText.table(lstTab, True, self.__bHTML) + "\n"
                if len(self.__setWorms_searchByName) > 0:
                    self.__sLog += "* MyWoRMS has to contact WoRMS: " + str(len(self.__setWorms_searchByName)) + "x" + "\n"
                else:
                    self.__sLog += "* All Name Queries are in MyWoRMS\n"
                    self.__sLog += "* No contact to WoRMS needed\n"
            self.__sLog += "\n"

            # --- Update aller Queries, die erneuert werden müssen
            if len(self.__setWorms_searchByName) > 0:
                pn("* Update Name Queries in MyWoRMS", fmt=False)
                self.__sLog += "* Update Name Queries in MyWoRMS\n"
                lstTab = [("Name Query", "Taxon")]
                iNrAll = len(self.__setWorms_searchByName)
                progress("on-abort")
                for iNrNow, sSn in enumerate(self.__setWorms_searchByName):
                    if self.isAbort():  # Abort
                        res.addErr("Aborted by user...", sys="TaxaCruncher", kat="Abort")
                        break
                    progress(iNrNow, iNrAll, 100, proz=True, abs=True, cli=True, status=str(sSn))
                    sInfo = ""
                    resMyWorms = self.__myworms.queryByName(sSn, fuzzy=self.__bFuzzy, like=self.__bLike, puffer=self.__bPuffer,marine_only=self.__bMarineOnly)
                    if resMyWorms.isErr():
                        # Fehler
                        res.addResult(resMyWorms, msg_only=True)
                        sInfo += "error"
                    else:
                        # Kein Fehler
                        if resMyWorms.isWarn():  # Warnungen sichern
                            res.addResult(resMyWorms, warn=True)
                        lstTaxa = resMyWorms.getData("taxa")
                        bFound = resMyWorms.getData("found")
                        bFallbackUsed = resMyWorms.getData("fallback_used")
                        bFoundInPuffer = resMyWorms.getData("found_in_puffer")
                        lstTaxa_discardedUnaccepted = resMyWorms.getData("taxa_discarded_unaccepted")
                        lstTaxa_addedAcceptedTraced = resMyWorms.getData("taxa_added_accepted_traced")
                        lstTaxa_wormsOriginal = resMyWorms.getData("taxa_worms_original")
                        if bFound:
                            # Taxon/Taxa wurden gefunden
                            # Info
                            if len(lstTaxa) == 1:
                                # Ein Ergebnis
                                try:
                                    iAid = lstTaxa[0].getAphiaId()
                                    sAid = str(iAid) + " "
                                except:
                                    res.addErr("WoRMS: Problems getting AphiaId",
                                               add="ScientificName: " + str(sSn),
                                               sys="TaxaCruncher", kat="WoRMS")
                                    sAid = "<uuups> "
                            elif len(lstTaxa) > 1:
                                # Mehrfache Ergebnisse
                                sAid = str(len(lstTaxa)) + "x choices: " + PrettyText.lst2strC([t.getAphiaId() for t in lstTaxa]) + " "
                            else:
                                # Kein Ergebnis
                                sAid = "nothing found/left "
                            # Info
                            # Add Discarded/Add Info
                            sWormsOriginal = ""
                            if bFallbackUsed:
                                sInfo += "[fallback used to prevent empty result: used original unaccepted taxa by WoRMS] "
                            if len(lstTaxa_discardedUnaccepted) > 0 or len(lstTaxa_addedAcceptedTraced) > 0:
                                sWormsOriginal = "[" + str(
                                    len(lstTaxa_wormsOriginal)) + " taxa found by WoRMS: " + PrettyText.lst2strC(
                                    [t.getAphiaId() for t in lstTaxa_wormsOriginal]) + "] "
                            sDiscardedTaxa = ""
                            if len(lstTaxa_discardedUnaccepted) > 0:
                                sDiscardedTaxa = "[" + str(
                                    len(lstTaxa_discardedUnaccepted)) + " discarded unaccepted taxa: " + PrettyText.lst2strC(
                                    [t.getAphiaId() for t in lstTaxa_discardedUnaccepted]) + "] "
                            sAddedTaxa = ""
                            if len(lstTaxa_addedAcceptedTraced) > 0:
                                sAddedTaxa = "[" + str(
                                    len(lstTaxa_addedAcceptedTraced)) + " added traced accepted taxa: " + PrettyText.lst2strC(
                                    [t.getAphiaId() for t in lstTaxa_addedAcceptedTraced]) + "] "
                            sInfo += sAid + sWormsOriginal + sDiscardedTaxa + sAddedTaxa
                            if bFoundInPuffer:
                                sInfo += "[buffer]"
                        else:
                            # Taxon/Taxa wurde nicht gefunden
                            sInfo += "not found"
                            self.__setWorms_searchByName_notFound.add(sSn)
                    # Log
                    pv(sSn+": "+sInfo, fmt=False)
                    if sSn not in self.__setWorms_searchByName_notFound:
                        lstTab.append((sSn, sInfo))
                if len(lstTab) > 1:
                    self.__sLog += PrettyText.table(lstTab, True, self.__bHTML) + "\n"
                else:
                    self.__sLog += "Nothing to do or nothing  found ...\n"
                self.__sLog += "\n"

                # Puffer speichern
                self.__myworms.puffer_save()

        # --- Trenne Mehrfachauswahlen (Suchname -> mehrere Taxa) von eindeutig gefundenen Taxa
        progress("on-pulse")
        workflow(top="TaxaCruncher", sub="Process Name Queries", fmt=False)
        lstWormsSearchByScientificName_notFound = []
        if len(self.__setSn) > 0:
            pn("* Process Name Queries", fmt=False)
            for iNrNow, sSn in enumerate(self.__setSn):
                if sSn not in self.__setWorms_searchByName_notFound:
                    resMyWorms = self.__myworms.queryByName(sSn, fuzzy=self.__bFuzzy, like=self.__bLike, marine_only=self.__bMarineOnly, puffer=True)
                    if resMyWorms.isAnyMessage():
                        res.addResult(resMyWorms, msg_only=True)
                    else:
                        lstTaxa = resMyWorms.getData("taxa")
                        bFound = resMyWorms.getData("found")
                        if not bFound:
                            # Kein Taxon gefunden
                            res.addErr("Taxon not found by name", add="ScientificName: " + str(sSn),
                                        sys="TaxaCruncher",
                                        kat="Name Query")  #
                            lstWormsSearchByScientificName_notFound.append(sSn)
                        elif len(lstTaxa) == 1:
                            # Ein Taxon
                            try:
                                t = lstTaxa[0]
                                self.__dicWorms_searchByName_answer_single[sSn] = t
                            except Exception as ex:
                                res.addErr("WoRMS: Problems getting AphiaId",
                                           add="ScientificName: " + str(sSn) + " - " + str(ex),
                                           sys="TaxaCruncher",
                                           kat="Name Query")
                        else:
                            # Mehrere Taxa
                            self.__dicWorms_SearchByName_answer_multi[sSn] = lstTaxa  # Taxa und User Auswahl

        # --- Mehrfachauswahlen -> GUI -> User Auswahl
        progress("on-pulse")
        workflow(top="TaxaCruncher", sub="User Selection", fmt=False)
        if len(self.__dicWorms_SearchByName_answer_multi) > 0:
            pn("* Asking user", fmt=False)
            # Auswahl durch Benutzer (GUI)
            odicUserChoices = OrderedDict()
            dicUSerChoicesAnswers = {}
            for sSn, lstTaxa in self.__dicWorms_SearchByName_answer_multi.items():
                lstChoice = [t.signature() for t in lstTaxa]
                odicUserChoices[sSn] = lstChoice
            userinput("Multiple Answers for Name Query", odicUserChoices)
            bUserInputDone = False
            # Warte auf Antwort von Benutzer ...
            while not bUserInputDone:
                userInput = getInputAnswer()
                if userInput is not None:
                    dicUserChoice = userInput
                    bUserInputDone = True
                time.sleep(1)
                # Abort
                if self.isAbort():
                    res.addErr("Action aborted by user...", kat="Abort", sys="TaxaCruncher")

            # User hat Eingabe abgebrochen
            if len(dicUserChoice) == 0:
                res.addErr("Selection aborted by user...", kat="Abort", sys="TaxaCruncher")

            # Choice auswählen
            if len(self.__dicWorms_SearchByName_answer_multi) > 0:
                for sSn, lstTaxa in self.__dicWorms_SearchByName_answer_multi.items():
                    try:
                        iIdxChoose_user = dicUserChoice[sSn]
                        taxon = lstTaxa[iIdxChoose_user]
                        sTaxon = taxon.signature()
                        iAid = taxon.getAphiaId()
                        self.__dicWorms_SearchByName_answerChoices_multi_userSelection[sSn] = taxon
                    except:
                        res.addErr("Problems with user selection", kat="Abort", sys="TaxaCruncher")

        # Füge alle Namenssuchen zusammen
        self.__dicWorms_searchByName_all = {**self.__dicWorms_searchByName_answer_single, **self.__dicWorms_SearchByName_answerChoices_multi_userSelection}
        if len(self.__dicWorms_searchByName_all) != (len(self.__dicWorms_searchByName_answer_single) + len(self.__dicWorms_SearchByName_answerChoices_multi_userSelection)):
            res.addErr("Internal Problem: Search by Name: Single and User Selection has same Search Name",
                       sys="TaxaCruncher",
                       kat="Name Query")

        # --- Nicht gefundene Taxa
        if len(self.__setWorms_searchByName_notFound) > 0:
            self.__sLog += "* Taxa not found by name (" + str(
                len(self.__setWorms_searchByName_notFound)) + "x)\n"
            lstTab = [["Name Query"]]
            for sSn in self.__setWorms_searchByName_notFound:
                lstTab.append([str(sSn)])
            self.__sLog += PrettyText.table(lstTab, True, self.__bHTML) + "\n"
            self.__sLog += "\n"

        # --- Eindeutige Taxa Auswahl
        if len(self.__dicWorms_searchByName_answer_single) > 0:
            lstTab = [["Name Query", "Taxon"]]
            self.__sLog += "* Taxa found directly by name (" + str(
                len(self.__dicWorms_searchByName_answer_single)) + "x)\n"
            for sSn, t in self.__dicWorms_searchByName_answer_single.items():
                lstTab.append([str(sSn), t.signature(html=self.__bHTML)])
            self.__sLog += PrettyText.table(lstTab, True, self.__bHTML) + "\n"
            self.__sLog += "\n"

        # --- Durch User ausgewählte Taxa
        if len(self.__dicWorms_SearchByName_answerChoices_multi_userSelection) > 0:
            self.__sLog += "* Taxa selected by user (" + str(
                len(self.__dicWorms_SearchByName_answerChoices_multi_userSelection)) + "x)\n"
            lstTab = [["Name Query", "Taxon"]]
            for sSn, t in self.__dicWorms_SearchByName_answerChoices_multi_userSelection.items():
                lstTab.append([str(sSn), t.signature(html=self.__bHTML)])
            self.__sLog += PrettyText.table(lstTab, True, self.__bHTML) + "\n"
            self.__sLog += "\n"

        return res

    def __updateMyWorms_taxa(self):
        """
        Update MyWoRMS
        """

        self.abort(False)
        res = Result(name="TaxaCruncher: Update MyWoRMS")

        pn = self.__smartprinter.normal
        pv = self.__smartprinter.verbose
        pd = self.__smartprinter.debug
        progress = self.__smartprinter.progress
        status = self.__smartprinter.status
        workflow = self.__smartprinter.workflow
        userinput = self.__smartprinter.inputAsk__comboMulti
        getInputAnswer = self.__smartprinter.getInputAnswer

        # ---
        progress("on-pulse")
        workflow(top="TaxaCruncher", sub="Update MyWoRMS", fmt=False)
        self.__sLog += "*** Update MyWoRMS\n\n"
        pn("* Update Taxa", fmt=False)

        # Sammle alle AphiaIds ein -> für WoRMS Update
        # SearchByName: Single
        setAid_final = set()
        for sSn, t in self.__dicWorms_searchByName_answer_single.items():
            iAid = t.getAphiaId()
            setAid_final.add(iAid)
        # SearchByName: (Multi -> User selected)
        for sSn, t in self.__dicWorms_SearchByName_answerChoices_multi_userSelection.items():
            iAid = t.getAphiaId()
            setAid_final.add(iAid)
        # SearchByAid: (Aid Only)
        for iAid in self.__setAid:
            setAid_final.add(iAid)
        # SearchByAid: (Aid/Sn)
        for iAid in self.__odicAidSn.keys():
            if iAid < 0:
                self.__setAid_myworms_special.add(iAid) # Special
            else:
                setAid_final.add(iAid)
        # Durchlaufe alle Aid
        for iAid in setAid_final:
            if iAid < 0:
                # Special
                self.__setAid_myworms_special.add(iAid)
                if not self.__model.subsystem_getTaxonWizard().isSpecialTaxon(iAid):
                    # Nicht gefunden
                    self.__setAid_myworms_special_notFound.add(iAid)
                    res.addErr("Special Taxon not found", add="AphiaId: " + str(iAid), sys="TaxaCruncher",kat="MyWoRMS")
            else:
                if self.__bPuffer:
                    # Puffer soll befragt werden
                    resMyWorms = self.__myworms.getTaxon(iAid, puffer=True, worms=False, deage=False)
                    if resMyWorms.getData("found"):
                        t = resMyWorms.getData("taxon")
                        self.__taxa_final.addTaxon(t)
                        if resMyWorms.getData("puffer_age_too_old"):
                            self.__setAid_myworms_old.add(iAid)  # Gefunden und zu alt -> WoRMS fragen
                        else:
                            self.__setAid_myorms_ok.add(iAid)  # Gefunden und Ok
                            self.__taxa_final.addTaxon(t) # Merken
                    else:
                        # Nicht gefunden -> WoRMS fragen
                        self.__setAid_myworms_not.add(iAid)
                else:
                    # Puffer soll nicht befragt werden
                    self.__setAid_myworms_not.add(iAid)
        setWormsSearchByAphiaId = self.__setAid_myworms_not.union(self.__setAid_myworms_old)
        # Log
        self.__sLog += "* Update Taxa in MyWoRMS\n"
        lstTab = [["Message", "Counts"]]
        if len(setAid_final) > 0:
            lstTab.append(("All Taxa (including Special Taxa)", str(len(setAid_final)) + "x"))
        if len(self.__setAid_myworms_special_notFound) > 0:
            lstTab.append(("Special Taxa not found in MyWoRMS", str(len(self.__setAid_myworms_special_notFound)) + "x"))
        if len(self.__setWorms_searchByName_aphiaIds) > 0:
            lstTab.append(("Taxa (normal) determined by name query", str(len(self.__setWorms_searchByName_aphiaIds)) + "x"))
        if len(self.__setAid_myorms_ok) > 0:
            lstTab.append(("Taxa (normal) found in MyWoRMS (young enough)" , str(len(self.__setAid_myorms_ok)) + "x"))
        if len(self.__setAid_myworms_old) > 0:
            lstTab.append(("Taxa (normal) found in MyWoRMS (too old)", str(len(self.__setAid_myworms_old)) + "x"))
        if len(self.__setAid_myworms_not) > 0:
            lstTab.append(("Taxa (normal) not found in MyWoRMS", str(len(self.__setAid_myworms_not)) + "x"))
        self.__sLog += PrettyText.table(lstTab, True, self.__bHTML) + "\n"
        if len(setWormsSearchByAphiaId) > 0:
            self.__sLog += "* MyWoRMS has to contact WoRMS: " + str(len(setWormsSearchByAphiaId)) + "x\n"
            pn("MyWoRMS has to contact WoRMS: " + str(len(setWormsSearchByAphiaId)) + "x", fmt=False)
        else:
            self.__sLog += "* No need to contact WoRMS\n"
            self.__sLog += "* MyWoRMS is Up-To-Date\n"
            pv("MyWoRMS is Up-To-Date and ready to use!", fmt=False)
        self.__sLog += "\n"

        # Update
        progress("on-abort")
        workflow(top="TaxaCruncher", sub="Update MyWoRMS", fmt=False)
        lstWormsSearchByAphiaId_notFound = []
        if len(setWormsSearchByAphiaId) > 0:
            self.__sLog += "* Contact WoRMS\n"
            pn("* Contact WoRMS", fmt=False)
            lstTab = [["AphiaId", "Answer"]]
            iNrAll = len(setWormsSearchByAphiaId)
            for iNrNow, iAid in enumerate(setWormsSearchByAphiaId):
                if self.isAbort():  # Abort
                    res.addErr("Aborted by user...", sys="TaxaCruncher", kat="Abort")
                    break
                progress(iNrNow, iNrAll, 100, proz=True, abs=True, cli=True, status=str(iAid))
                resMyWorms = self.__myworms.getTaxon(iAid, puffer=False, worms=True, deage=True)
                sInfo = ""
                sInfo_ascii = ""
                sStatus_actualTaxon = str(iAid)
                if resMyWorms.isErr():
                    res.addErr("Taxon not found by AphiaId", add="AphiaId: " + str(iAid), sys="TaxaCruncher", kat="WoRMS")
                    sInfo += "error"
                    sInfo_ascii += "error"
                else:
                    t = resMyWorms.getData("taxon")
                    bFound = resMyWorms.getData("found")
                    if bFound:
                        self.__taxa_final.addTaxon(t) # Merken
                        sStatus_actualTaxon = str(t.signature())
                        sInfo += str(t.signature(html=self.__bHTML))
                        sInfo_ascii += str(t.signature())
                    else:
                        sInfo += "not found"
                        res.addErr("Taxon not found by AphiaId", add="AphiaId: " + str(iAid), sys="TaxaCruncher",kat="WoRMS")
                        lstWormsSearchByAphiaId_notFound.append(iAid)
                # Log
                sStatus_sumNotFound = " [" + str(len(lstWormsSearchByAphiaId_notFound)) + "]" if len(
                    lstWormsSearchByAphiaId_notFound) > 0 else ""
                status("Update Taxon: " + str(sStatus_actualTaxon) + sStatus_sumNotFound, fmt=False)
                if iAid not in lstWormsSearchByAphiaId_notFound:
                    lstTab.append((str(iAid),sInfo))
                pv(str(iAid)+": "+sInfo_ascii, fmt=False)
            if len(lstTab) > 1:
                self.__sLog += PrettyText.table(lstTab, True, self.__bHTML) + "\n"
            else:
                self.__sLog += "Nothing to do or nothing found ...\n"
            self.__sLog += "\n"

            # Log
            if len(lstWormsSearchByAphiaId_notFound) > 0:
                self.__sLog += "* Taxa not found in WoRMS\n"
                lstTab = [["AphiaId"]]
                for sTxt in lstWormsSearchByAphiaId_notFound:
                    lstTab.append([str(iAid)])
                self.__sLog += PrettyText.table(lstTab, True, self.__bHTML) + "\n"
            self.__sLog += "\n"

            # Puffer speichern
            self.__myworms.puffer_save()

        return res
