"""
Modul mit Hilfsklassen, die bei der Arbeit mit Taxa
"""

# coding=utf-8

import openpyxl, traceback, pickle, zipfile
from collections import Counter, OrderedDict
import urllib, os, sys, csv, getpass, platform, datetime, time, copy, random, psycopg2, psycopg2.extras, re
import openpyxl.styles.numbers
import warnings
import requests
from psycopg2 import sql
from tools.parallel import Stoppable
from tools.coding import Result, DataTools, ProgressString, WorkingObject
from tools.text import PrettyText
from tools.ui import Colors
from tools.printing import SmartPrinter
from tools.excel import ExcelTools
from tools.fileio import FileTools
from shutil import copyfile

class Taxon:
    """
    Taxon
    """

    # Klassen-Variablen/Methoden
    __dicWormsHeaderMapping = {
        'aid': 'AphiaID',
        'vaid': 'valid_AphiaID',
        'scientificname': 'scientificname',
        'valid_name': 'valid_name',
        'rank': 'rank',
        'kingdom': 'kingdom',
        'phylum': 'phylum',
        'class': 'class',
        'order': 'order',
        'family': 'family',
        'genus': 'genus',
        'lsid': 'lsid',
        'unacceptreason': 'unacceptreason',
        'status': 'status',
        'match_type': 'match_type',
        'isfreshwater': 'isFreshwater',
        'isbrackish': 'isBrackish',
        'isextinct': 'isExtinct',
        'ismarine': 'isMarine',
        'isterrestrial': 'isTerrestrial',
        'modified': 'modified',
        'citation': 'citation',
        'authority': 'authority',
        'valid_authority': 'valid_authority',
        'url': 'url'
    }
    __lstAdditionalHeader = ["aaid", "aname", "aauth", "aaid_way", "is_colony"]

    @staticmethod
    def getMapHeaderIntern2Worms():
        return Taxon.__dicWormsHeaderMapping.copy()

    def getMapHeaderWorms2Intern():
        dicWorms2intern = {v: k for k, v in Taxon.getMapHeaderIntern2Worms().items()}
        return dicWorms2intern

    @staticmethod
    def isHeaderWorms(sHeader):
        if sHeader in Taxon.getMapHeaderWorms2Intern():
            return True
        else:
            return False

    @staticmethod
    def mapHeaderWorms2Intern(sHead):
        dic = Taxon.getMapHeaderWorms2Intern()
        if sHead in dic:
            return dic[sHead]
        else:
            return None

    @staticmethod
    def getHeaderIntern():
        lstHeader = [k for k, v in Taxon.__dicWormsHeaderMapping.items()]
        lstHeader += Taxon.__lstAdditionalHeader
        return lstHeader

    @staticmethod
    def getHeaderWorms():
        lstHeader = [v for k, v in Taxon.__dicWormsHeaderMapping.items()]
        return lstHeader

    @staticmethod
    def checkCompatibility2Worms():
        res = Result(name="Check compatability to WoRMS")
        # Tweak: No Connection
        # return res
        lstUnusedHeaders = ["taxonRankID", "parentNameUsageID"]
        try:
            # Anfrage
            sRequest = 'http://www.marinespecies.org/rest/AphiaRecordByAphiaID/141433'
            resp = requests.get(sRequest)
            # Antwort
            try:
                dicWorms = resp.json()
            except:
                dicWorms = None
            # Alles OK?
            if resp.status_code != 200 or dicWorms is None:
                raise NameError("Connection Problems")
            # Entferne nicht benötigte Einträge von WoRMS
            for sName in lstUnusedHeaders:
                dicWorms.pop(sName)
            # Vergleichen
            set1 = set((dicWorms))
            set2 = set(Taxon.getMapHeaderIntern2Worms().values())
            setMissing = set(set1).difference(set(set2))
            setSurplus = set(set2).difference(set(set1))
            bMissing = len(setMissing) != 0
            bSurplus = len(setSurplus) != 0
            if bSurplus:
                res.addInfo("Surplus headers in internal taxon data structure", add=PrettyText.lst2strC(setSurplus),
                            kat="Check Compatibility with WoRMS", sys="Taxon")
            if bMissing:
                res.addErr("Missing required headers in internal taxon data structure",
                           add=PrettyText.lst2strC(setMissing), kat="Check Compatibility with WoRMS", sys="Taxon")
        except Exception as ex:
            # Fehler
            res.addErr("Problems connecting WoRMS", add=str(ex), kat="Check Compatibility with WoRMS", sys="Taxon")
        return res

    # ------------------------------------

    # Konstruktor
    def __init__(self, iAid=None, **kwargs):
        """

        Args:
            (int) AphiaId: Automatischer aufbau durch WoRMS-Anfrage
        Kwargs:
            blueprint (dict): Manuelles Aufbauen des Taxon
        Returns:
            (Result) result
        """

        # Parameter
        dicBlueprint = kwargs["blueprint"] if "blueprint" in kwargs else None

        # Member
        self.__bHandmade = False
        self.__bSpecial = False
        self.__result = Result()
        self.__dicData = None
        self.__dateCreate = datetime.datetime.now()
        self.__bLeadingToRealAcceptedTaxon = False # Achtung: Sehr speziell und bei Abstieg ermittelt: Bei Abstieg wird nicht nur auf Aid=vAid=aAid geprüft sondern auch auf sStatus=accepted
        if iAid is not None and type(iAid) is int and dicBlueprint is None:
            # 1) Erzeuge Taxon mit WoRMS
            self.__dicData = self.__askWorms(iAid)
            # 2) Basis-Tests
            # Aphia Id gefunden
            if self.getAphiaId() is None:
                if iAid is None:
                    self.__result.addErr("AphiaId missing",
                                         add="ScientificName: " + str(self.getScientificName()),
                                         sys="Taxon", kat="Automatic Configuration")
                elif type(iAid) is not int:
                    self.__result.addErr("AphiaId not integer",
                                         add="AphiaId: " + str(iAid) + ", ScientificName:" + str(self.getScientificName()),
                                         sys="Taxon", kat="Automatic Configuration")
            # 3) Hole die Aaid
            if self.isFound() and self.isOk():
                if self.getAphiaId() == self.getAphiaId_valid():
                    # Taxon ist bereits akzeptiert
                    self.__dicData["aaid"] = self.getAphiaId()
                    self.__dicData["aname"] = self.getScientificName()
                    self.__dicData["aauth"] = self.getAuthority()
                    self.__dicData["aaid_way"] = None
                else:
                    # Suche Akzeptiertes Taxon mit WoRMS
                    iAaid, sSn, sAuth, lstVisited = self.__getAaid(iAid)
                    self.__dicData["aaid"] = iAaid
                    self.__dicData["aname"] = sSn
                    self.__dicData["aauth"] = sAuth
                    self.__dicData["aaid_way"] = lstVisited

        elif dicBlueprint is not None and type(dicBlueprint) is dict and iAid is None:
            # 2) Taxon manuell erzeugen
            self.__bHandmade = True
            sSn = dicBlueprint["scientificname"] if "scientificname" in dicBlueprint else None
            iAid = dicBlueprint["aid"] if "aid" in dicBlueprint else None
            if iAid is None:
                self.__result.addErr("AphiaId missing",
                                     add="ScientificName: "+str(sSn),
                                     sys="Taxon", kat="Handmade Configuration")
            elif type(iAid) is not int:
                self.__result.addErr("AphiaId not integer",
                                     add="AphiaId: "+str(iAid)+", ScientificName:"+str(sSn),
                                     sys="Taxon", kat="Handmade Configuration")
            else:
                # Alles OK
                # Warnung bei fehlendem Scientificname
                if sSn in [None, ""]:
                    self.__result.addWarn("ScientificName missing",
                                      add="AphiaId: " + str(iAid),
                                      sys="Taxon", kat="Handmade Configuration")
                # Special Taxon
                self.__bSpecial = True if iAid < 0 else False # Special Taxon
                # Alle nicht erlaubten Header entfernen
                lstSuprplusHeader = []
                for sHead in list(dicBlueprint.keys()):
                    if sHead not in self.getHeaderIntern():
                        del dicBlueprint[sHead]
                        lstSuprplusHeader.append(sHead)
                if len(lstSuprplusHeader) > 0:
                    self.__result.addInfo("Discarded surplus headers", add=str(iAid)+": "+PrettyText.lst2strC(lstSuprplusHeader), sys="Taxon", kat="Handmade configuration")

                # Speichern
                self.__dicData = dicBlueprint
        else:
            # Beides geht nicht
            self.__result.addErr("Taxon has to be configured either by hand or by WoRMS automatically", sys="Taxon", kat="Configuration")

        # Check auf andere oOnfigurationsfehler
        if self.isRedirected_2_accepted() and self.isRedirected_2_unaccepted():
            self.__result.addErr("redirected and directed at same time", sys="Taxon", kat="Configuration")
        if self.isRedirected() and self.isAccepted():
            self.__result.addErr("redirected and acccepted at same time", sys="Taxon", kat="Configuration")
        if self.isLimbo() and (self.isAccepted() or self.isUnaccepted()):
            self.__result.addErr("Limbo and acccepted/unaccepted at the same time", sys="Taxon", kat="Configuration")

    def __str__(self):
        return self.signature()

    def info(self, **kwargs):
        """
        Farbcodes:
            direkt akzeptiert: grün
            indirekt akzeptiert (unacc->acc): hellgruen
            direkt unakzeptiert:  gelb
            indirekt unakzeptiert (unacc->uncacc): gelb1
            limbo: lila
            problem: rot
        """
        bDebug = kwargs["debug"] if ("debug" in kwargs and kwargs["debug"] in (True,False)) else False

        sOut = ""
        # Normal
        if self.isErr():
            sOut += "Error: yes\n"
        else:
            sOut += self.signature() + "\n"
            sOut += "\n"
            if bDebug:
                sOut += "Found:                     " + ("yes" if self.isFound() else "no") + "\n"
                sOut += "\n"
            sOut += "AphiaId:                   " + (str(self.getAphiaId()) if self.getAphiaId() is not None else "-") + "\n"
            sOut += "ScientificName:            " + (self.getScientificName() if self.getScientificName() is not None else "-") + "\n"
            sOut += "Authority:                 " + (self.getAuthority() if self.getAuthority() is not None else "-")   + "\n"
            sOut += "AphiaId (valid):           " + (str(self.getAphiaId_valid()) if self.getAphiaId_valid() is not None else "-")  + "\n"
            sOut += "ScientificName (valid):    " + (self.getScientificName_valid() if self.getScientificName_valid() is not None else "-") + "\n"
            sOut += "Authority (valid):         " + (self.getAuthority_valid() if self.getAuthority_valid() is not None else "-")+ "\n"
            sOut += "AphiaId (accepted):        " + (str(self.getAphiaId_accepted()) if self.getAphiaId_accepted()is not None else "-")  + "\n"
            sOut += "ScientificName (accepted): " + (self.getScientificName_accepted() if self.getScientificName_accepted() is not None else "-")+ "\n"
            sOut += "Authority (accepted):      " + (self.getAuthority_accepted() if self.getAuthority_accepted() is not None else "-")  + "\n"
            sOut += "Way of Acceptance:         " + PrettyText.lst2strC(self.getAcceptedWay()) + "\n"
            sOut += "\n"
            sOut += "Created on:                " + (str(self.getDateCreate()) if self.getDateCreate() is not None else "-")  + "\n"
            sOut += "Age (days):                " + (str(self.getAge()) if self.getAge() is not None else "-") + "\n"
            sOut += "\n"
            sOut += "Status:                    "
            if self.isAccepted():
                sOut += "<accepted>\n"
            elif self.isUnaccepted():
                sOut += "<unaccepted>\n"
            elif self.isLimbo():
                sOut += "<limbo>\n"
            else:
                sOut += "<?>\n"
            sOut += "Leading to accepted taxon: " + ("yes" if self.isRedirected_2_accepted() else "no") + "\n"
            sOut += "Status field:              " + (self.getStatus() if self.getStatus() is not None else "-")  + "\n"
            sOut += "UnacceptReason field:      " +(self.getUnacceptReason() if self.getUnacceptReason() is not None else "-")  + "\n"
            sOut += "\n"
            if self.hasProblems():
                lstProblems = []
                if self.hasProblems_aphia():
                    lstProblems.append("AphiaIds")
                if self.hasProblems_names():
                    lstProblems.append("Names")
                if self.hasProblems_status():
                    lstProblems.append("Status")
                sOut += "Problems:                  yes ("+PrettyText.lst2strC(lstProblems)+")\n"
            else:
                sOut += "Problems:                  no\n"
            sOut += "Name gaps:                 " + ("yes" if self.hasProblems_names() else "no") + "\n"
            sOut += "All Names are equal:       " + ("yes" if self.allNamesEqual() else "no") + "\n"
            sOut += "All AphiaIds are equal:    " + ("yes" if self.allAphiaIdsEqual() else "no") + "\n"
            sOut += "\n"
            sOut += "Rank:                      " + (self.getRank() if self.getRank() is not None else "-") +"\n"
            sOut += "Kingdom:                   " + (self.getKingdom() if self.getKingdom() is not None else "-") +"\n"
            sOut += "Phylum:                    " + (self.getPhylum() if self.getPhylum() is not None else "-") +"\n"
            sOut += "Class:                     " + (self.getClass() if self.getClass() is not None else "-") + "\n"
            sOut += "Order:                     " + (self.getOrder() if self.getOrder() is not None else "-") + "\n"
            sOut += "Family:                    " + (self.getFamily() if self.getFamily() is not None else "-") + "\n"
            sOut += "Genus:                     " + (self.getGenus() if self.getGenus() is not None else "-") + "\n"
            sOut += "\n"
            sOut += "Freshwater:                " + ("yes" if self.isFreshwater() else "no") + "\n"
            sOut += "Brackish:                  " + ("yes" if self.isBrackish() else "no") + "\n"
            sOut += "Extinct:                   " + ("yes" if self.isExtinct() else "no") + "\n"
            sOut += "Marine:                    " + ("yes" if self.isMarine() else "no") + "\n"
            sOut += "Terrestrial:               " + ("yes" if self.isTerrestrial() else "no") + "\n"
            sOut += "\n"
            sOut += "Modified:                  " + (self.getModified() if self.getModified() is not None else "-") +"\n"
            sOut += "Citation:                  " + (self.getCitation() if self.getCitation() is not None else "-") +"\n"
            sOut += "URL:                       " + (self.getUrl() if self.getUrl() is not None else "-") +"\n"
            sOut += "LSID:                      " + (self.getLsid() if self.getLsid() is not None else "-") +"\n"
            if bDebug:
                sOut += "\n"
                sOut += "All Data:                  " + PrettyText.lst2strC(self.__dicData) + "\n"
            if self.__result.isAnyMessage():
                sOut += "\n"
                sOut += "Intenal Build Protocol:\n"
                sOut += self.__result.getErrWarnInfoString(short=True) + "\n"

        return sOut

    # Signature zu Taxon erstellen
    def signature(self, **kwargs):
        bHtml = kwargs["html"] if ("html" in kwargs and kwargs["html"] in (True,False)) else False
        if not bHtml:
            return self.__signature_ascii(**kwargs)
        else:
            return self.__signature_html(**kwargs)

    def __signature_html(self, **kwargs):
        def clean(sTxt):
            if sTxt in ("",None):
                sTxt = "&lt;empty&gt;"
            sTxt = str(sTxt)
            rexHTML = re.compile(r'<[^>]+>')
            sTxt = rexHTML.sub('', sTxt)
            sTxt = sTxt.replace("(", "")
            sTxt = sTxt.replace(")", "")
            sTxt = sTxt.replace("[", "")
            sTxt = sTxt.replace("]", "")
            # sTxt = sTxt.replace("{", "")
            # sTxt = sTxt.replace("}", "")
            sTxt = sTxt.replace("<", "&lt;")
            sTxt = sTxt.replace(">", "&gt;")
            sTxt = sTxt.strip()
            return sTxt

        # Parameter
        bDebug = kwargs["debug"] if ("debug" in kwargs and kwargs["debug"] in (True,False)) else False

        # Bei Problemen Debug anmachen
        if self.hasProblems():
            bDebug = True

        # Colors
        sColor_awi_blue_dark = Colors.getColor("awi_blue_dark", html=True)
        sColor_awi_blue_mid = Colors.getColor("awi_blue_mit", html=True)
        sColor_awi_blue_light = Colors.getColor("awi_blue_light", html=True)
        sColor_blue = Colors.getColor("blue", html=True)
        sColor_orange = Colors.getColor("orange", html=True)
        sColor_red = Colors.getColor("red", html=True)
        sColor_green = Colors.getColor("green", html=True)
        sColor_green_light = Colors.getColor("green_light", html=True)
        sColor_yellow = Colors.getColor("yellow", html=True)
        sColor_purple = Colors.getColor("purple", html=True)
        sColor_rose = Colors.getColor("rose", html=True)
        sColor_white = Colors.getColor("white", html=True)
        sColor_accepted = sColor_green
        sColor_unaccepted_2_accepted = sColor_green_light
        sColor_unaccepted = sColor_rose
        sColor_unaccepted_2_unaccepted = sColor_yellow
        sColor_problems = sColor_red
        sColor_limbo = sColor_purple

        # Info
        sInfo_main = ""
        sInfo_accept = ""
        sInfo_problems = ""
        sInfo_problems_extra = ""
        sInfo_aphia = ""
        sInfo_way = ""
        sInfo_rank = ""
        sInfo_name = ""
        sInfo_name_valid = ""
        sInfo_name_redirected = ""
        sInfo_auth = ""
        sInfo_add = ""
        sInfo_status = ""
        sInfo_handmade = ""
        sInfo_err = ""

        # *** Main Name
        if self.hasProblems() or self.isErr():
            sColor = sColor_problems
        elif self.isAccepted():
            sColor = sColor_accepted
        elif self.isLimbo():
            sColor = sColor_limbo
        elif not self.isAccepted() and not self.isRedirected():
            sColor = sColor_unaccepted
        elif not self.isAccepted() and self.isRedirected_2_accepted():
            sColor = sColor_unaccepted_2_accepted
        elif not self.isAccepted() and not self.isRedirected_2_accepted():
            sColor = sColor_unaccepted_2_unaccepted
        else:
            sColor = sColor_white
        sInfo_notmarine_main = ""
        sUrl = self.getUrl()
        bUrl = True if sUrl not in (None, "") else False
        sName_any = clean(self.getAnyName()[0])
        if not self.isMarine():
            sInfo_notmarine_main = ' <sup><span style="color:' + sColor_blue + ';">not marine</span></sup>'
        if bUrl:
            # Mit URL
            sInfo_main += '<span style="white-space: nowrap;"><span style="color:' + sColor + '">&#11044;</span><a style="text-decoration:none; color: black;" href="' + str(sUrl) + '" onclick=\'window.open(this.href, "WoRMS", "toolbar=0,menubar=0,scrollbars=1,height=400,width=600"); return false;\'> ' + sName_any + '</a>'+sInfo_notmarine_main+'</span>'
        else:
            # Keine URL
            sInfo_main += '<span style="white-space: nowrap;"><span style="color:' + sColor + '">&#11044;</span> ' + sName_any + sInfo_notmarine_main+'</span>'

        if bDebug:
            # *** Debug/Problem Taxon
            # Problems
            lstProblems = []
            if not self.isFound():
                lstProblems.append("not found")
            if self.hasProblems_status():
                lstProblems.append("status")
            if self.hasProblems_names():
                lstProblems.append("name")
            if self.hasProblems_aphia():
                lstProblems.append("aphia")
                # *** Error
            if self.__result.isErr():
                lstProblems.append("internal")
            if len(lstProblems) > 0:
                sInfo_problems += '<span style="white-space: nowrap; color:' + sColor_red + ';">'
                sInfo_problems += "SEVERE PROBLEMS: "
                sInfo_problems += PrettyText.lst2strC(lstProblems).upper()
                sInfo_problems += '</span><br><hr>'
                sInfo_problems_extra += '<hr><code>'
                sInfo_problems_extra += '<span style="white-space: nowrap; font-weight: bold;">Debug Info</span><br>'
                sInfo_problems_extra += '<span style="white-space: nowrap; font-weight: 100;">AphiaId: '+str(self.getAphiaId())+'</span><br>'
                sInfo_problems_extra += '<span style="white-space: nowrap; font-weight: 100;">AphiaId (valid): '+str(self.getAphiaId_valid())+'</span><br>'
                sInfo_problems_extra += '<span style="white-space: nowrap; font-weight: 100;">AphiaId (accepted): '+str(self.getAphiaId_accepted())+'</span><br>'
                if self.hasProblems_names():
                    sInfo_problems_extra += '<span style="white-space: nowrap; font-weight: 100;">ScientificName: '+str(self.getScientificName()) +'</span><br>'
                    sInfo_problems_extra += '<span style="white-space: nowrap; font-weight: 100;">ScientificName (valid): '+str(self.getScientificName_valid())+'</span><br>'
                    sInfo_problems_extra += '<span style="white-space: nowrap; font-weight: 100;">ScientificName (accepted): '+str(self.getScientificName_accepted())+'</span><br>'
                if self.hasProblems_status():
                    sInfo_problems_extra += '<span style="white-space: nowrap; font-weight: 100;">Status: '+str(self.getStatus()) +'</span><br>'
                    sInfo_problems_extra += '<span style="white-space: nowrap; font-weight: 100;">Unaccept Reason: '+str(self.getUnacceptReason())+'</span><br>'
                if self.isErr():
                    lstErr = self.getResult().getErrLst(short=True)
                    sErr = PrettyText.lst2str(lstErr, "<br>")
                    sInfo_problems_extra += '<span style="white-space: nowrap; font-weight: 100;">Errors: '+ sErr +'</span><br>'
                sInfo_problems_extra += '</code>'

        # *** Normales Taxon
        # * Accepted/Unaccepted/Limbo
        if self.isAccepted() :
            # akzeptiert (direkt)
            sInfo_accept += '<span style="white-space: nowrap; color:' + sColor_accepted + ';">accepted</span><br>'
        elif self.isLimbo():
            # Limbo (direkt)
            sInfo_accept += '<span style="white-space: nowrap; color:' + sColor_limbo + ';">limbo</span><br>'
        elif self.isUnaccepted():
            # unakzeptiert
            if not self.isRedirected():
                # unakzeptiert (direkt)
                sInfo_accept += '<span style="white-space: nowrap; color:' + sColor_unaccepted + ';">unaccepted</span><br>'
            elif self.isRedirected_2_accepted():
                # unakzeptiert redirekted ...
                sInfo_accept += '<span style="white-space: nowrap;">unaccepted</span><br>'
            elif self.isRedirected_2_unaccepted():
                # unakzeptiert redirekted ...
                sInfo_accept += '<span style="white-space: nowrap;">unaccepted</span><br>'
        # * AphiaId/ScientificName/Authority
        # Rank
        if self.getRank() is not None:
            sInfo_rank += "["+clean(self.getRank())+"]"
        if self.getScientificName() not in ("", None):
            # Direkt
            sInfo_name += '<span style="white-space: nowrap; font-weight: bold;"><span style="font-weight: 100;">'+clean(self.getAphiaId())+'</span> '+clean(self.getScientificName())+' <span style="font-weight: 100;">'+sInfo_rank+'</span></span><br>'
            if self.getAuthority() not in ("", None):
                sInfo_name += '<span style="white-space: nowrap; font-weight: 100;">('+clean(self.getAuthority())+')</span><br>'
        if not self.isAccepted() and self.isRedirected():
            # Redirected
            # Way of Acceptance
            if self.getAcceptedWay() is not None and len(self.getAcceptedWay()) > 2:
                lstTmp = self.getAcceptedWay()
                try:
                    lstTmp.pop()
                    del lstTmp[0]
                    sInfo_way += "&horbar; " + PrettyText.lst2str(lstTmp, " &horbar; ") + " "
                except:
                    pass
            if self.getScientificName_accepted() not in ("", None):
                if self.isRedirected_2_accepted():
                    # -> akzeptiert (Accepted ScientificName)
                    sInfo_name_redirected += '<span style="white-space: nowrap; font-style: italic; color:' + sColor_unaccepted_2_accepted + ';">'+sInfo_way+'&horbar; accepted &rarr;</span><br>'
                else:
                    # -> unakzeptiert
                    sInfo_name_redirected += '<span style="white-space: nowrap; font-style: italic; color:' + sColor_unaccepted_2_unaccepted+ ';">'+sInfo_way+'&horbar; not accepted &rarr;</span><br>'
                sInfo_name_redirected += '<span style="white-space: nowrap; font-weight: bold;"><span style="font-weight: 100;">'+clean(self.getAphiaId_accepted())+'</span> '+clean(self.getScientificName_accepted())+'</span><br>'
                if self.getAuthority_accepted() not in ("", None):
                    sInfo_name_redirected += '<span style="white-space: nowrap; font-weight: 100;">(' + clean(self.getAuthority_accepted()) + ')</span><br>'
        # Nur Valid Name vorhanden
        if (self.getAuthority_valid is not None) and (self.getScientificName(), self.getScientificName_accepted()) == (None, None):
            sInfo_name_valid += '<span style="white-space: nowrap; font-weight: bold;"><span style="font-weight: 100;">' + clean(
                self.getAphiaId_valid()) + '</span> ' + clean(
                self.getScientificName_valid()) + '<span style="font-weight: 100;"> ' + sInfo_rank + '</span></span><br>'
            if self.getAuthority_accepted() not in ("", None):
                sInfo_name_valid += '<span style="white-space: nowrap; font-weight: 100;">(' + clean(
                    self.getAuthority_valid()) + ')</span><br>'

        # * Status
        if self.isUnaccepted() or self.isLimbo():
            if self.getStatus() not in ["unaccepted", None]:
                sInfo_status += '<span style="white-space: nowrap; font-style: italic;">status: <span style="font-weight: 100;">' + clean(self.getStatus()) + '</span></span><br>'
            if self.getUnacceptReason() is not None:
                sInfo_status += '<span style="font-style: italic;">unaccept reason:<span style="font-weight: 100;"><br>' + clean(self.getUnacceptReason()) + '</span></span><br>'

        # Antwort
        sInfo_start = '<div class="tooltip">'
        sInfo_popup_start = '<span class="right">'
        sInfo_popup_end = '<i></i></span>'
        sInfo_end = '</div>'
        sInfo = ""
        sInfo += sInfo_start
        sInfo += sInfo_main
        sInfo += sInfo_popup_start
        sInfo += sInfo_problems
        sInfo += sInfo_accept
        sInfo += sInfo_name
        sInfo += sInfo_status
        sInfo += sInfo_name_valid
        sInfo += sInfo_name_redirected
        sInfo += sInfo_problems_extra
        sInfo += sInfo_popup_end
        sInfo += sInfo_end
        return sInfo

    def __signature_ascii(self, **kwargs):
        def clean(sTxt):
            if sTxt is not None:
                sTxt = sTxt.replace("(", "")
                sTxt = sTxt.replace(")", "")
                sTxt = sTxt.strip()
            return sTxt

        # Parameter
        bDebug = kwargs["debug"] if ("debug" in kwargs and kwargs["debug"] in (True,False)) else False
        # Bei Problemen Debug anmachen
        if self.hasProblems_status() or self.hasProblems_names() or self.hasProblems_aphia():
            bDebug = True

        # Daten
        iAid = self.getAphiaId()
        iVaid = self.getAphiaId_valid()
        iAaid = self.getAphiaId_accepted()
        sName = self.getScientificName()
        sVname = self.getScientificName_valid()
        sAname = self.getScientificName_accepted()
        sAuth =  clean(self.getAuthority())
        sVauth =  clean(self.getAuthority_valid())
        sAauth = clean(self.getAuthority_accepted())
        lstAway = self.getAcceptedWay()
        sRank = self.getRank()
        sStatus = self.getStatus()
        bAccepted = self.isAccepted()
        sUnacceptReason = self.getUnacceptReason()

        # Info
        sInfo = ""
        sInfoAphia = ""
        sInfoWay = ""
        sInfoName = ""
        sInfoAuth = ""
        sInfoAccept = ""
        sInfoAdd = ""
        sInfoHandmade = ""
        sInfoErr = ""

        # Error
        if self.__result.isErr():
            sInfoErr = "<ERROR> "

        # AphiaIds
        if bDebug:
            # Debug
            sInfoAphia = ""
            sInfoAphia += ""+str(iAid)+" " if iAid is not None else ""
            sInfoAphia += "("+str(iVaid)+") " if iVaid is not None else ""
            sInfoAphia += "{"+str(iAaid)+"} " if iAaid is not None else ""
            if sInfoAphia == "":
                sInfoAphia = "<No AphiaIds at all>"
        else:
            if self.isAccepted() or self.isLimbo():
                # Akzeptiert und Limbo
                if iAid is not None:
                    # Aid
                    sInfoAphia += str(iAid)+" "
                elif iAaid is not None: # Hier darf man gar nicht hinkommen ...
                    # Aaid
                    sInfoAphia += "{" + str(iAaid) + "} "
                else: # Hier darf man gar nicht hinkommen ...
                    sInfoAphia += "<no AphiaIds at all> "
            else:
                # Unakzeptiert
                if iAid is not None:
                    # Aid
                    sInfoAphia += str(iAid) + " "
                if iAaid is not None:
                    sInfoAphia += "{"+str(iAaid)+"} "
                if iAid is None and iAaid is None:
                    # Gar kein AphiaIds
                    sInfoAphia += "<no AphiaIds at all> "

        # AphiaId Way
        if lstAway is not None and len(lstAway) > 2:
            # - Normal
            sInfoWay += "["+PrettyText.lst2strC(lstAway)+"] "

        # Namen
        if bDebug:
            # Debug
            sInfoName += "" + str(sName) + " " if sName is not None else ""
            sInfoName += "(" + str(sVname) + ") " if sVname is not None else ""
            sInfoName += "{" + str(sAname) + "} " if sAname is not None else ""
            if sInfoName == "":
                sInfoName = "<no names at all> "
            else:
                sInfoName = "Names: " + sInfoName
        else:
            # Akzeptiert und Limbo
            if self.isAccepted() or self.isLimbo():
                # Akzeptiert
                if sName is not None:
                    # Name
                    sInfoName += sName + " "
                elif sAname is not None: # Hier darf man gar nicht hinkommen ...
                    # Aname
                    sInfoName += "{" + str(sAname) + "} "
                else: # Hier darf man gar nicht hinkommen ...
                    # Kein Name
                    sInfoName += "<no names at all> "
            else:
                # Unakzeptiert
                if sName is not None and sAname is None:
                    # Name
                    sInfoName += sName + " "
                elif sName is not None and sAname is not None:
                    # Name und Aname
                    sInfoName += sName + " {" + str(sAname) + "} "
                elif sName is None and sAname is not None:
                    # Aname
                    sInfoName += "<no ScientificName> {" + str(sAname) + "} "
                if sName is None and sAname is None:
                    # Gar keine Namen
                    sInfoName += "<no names at all> "

        # Auth
        if bDebug:
            # Debug
            sInfoAuth = ""
            sInfoAuth += "" + str(sAuth) + " " if sAuth is not None else ""
            sInfoAuth += "(" + str(sVauth) + ") " if sVauth is not None else ""
            sInfoAuth += "{" + str(sAauth) + "} " if sAauth is not None else ""
            if sInfoAuth != "":
                sInfoAuth = "Authority: "+sInfoAuth
        elif self.isAccepted() or self.isLimbo():
            # Akzeptiert und Limbo
            if sAuth is not None:
                # Auth
                sInfoAuth += "("+sAuth + ") "
            elif sAauth is not None:  # Hier darf man gar nicht hinkommen ...
                sInfoAuth += "{" + str(sAauth) + "} "

        # Add
        if self.isAccepted() or self.isLimbo() or bDebug:
            if sRank is not None:
                # Rank
                sInfoAdd += "["+sRank+ "] "

        # Accept
        if bDebug:
            # Debug
            if self.isAccepted():
                sAccepted = "<accepted>"
            elif self.isUnaccepted():
                sAccepted = "<unaccepted>"
            elif self.isLimbo():
                sAccepted = "<limbo>"
            else:
                sAccepted = "<UUUPS!>" # Das soltle nicht passieren dürfen
            sInfoAccept = sAccepted+" [status:" + str(sStatus) + ", unaccept_reason:" + str(sUnacceptReason)+"]"
        else:
            if self.isUnaccepted():
                # Accepted
                sInfoAccept += "<unaccepted"
                lstStatus = []
                if sStatus is not None and sStatus != "unaccepted":
                    lstStatus.append(sStatus)
                if sUnacceptReason is not None:
                    lstStatus.append(sUnacceptReason)
                if len(lstStatus) > 0:
                    sInfoAccept+= ": "+PrettyText.lst2strC(lstStatus)
                sInfoAccept +="> "
                if self.getAphiaId_accepted() is not None and not self.isRedirected_2_accepted():
                    sInfoAccept += "<caution: not leading to accepted taxon> "
            elif self.isLimbo():
                # Limbo
                sInfoAccept += "<limbo"
                lstStatus = [str(sStatus)]
                if sUnacceptReason is not None:
                    lstStatus.append(sUnacceptReason)
                if len(lstStatus) > 0:
                    sInfoAccept += ": " + PrettyText.lst2strC(lstStatus)
                sInfoAccept += "> "
            elif self.isAccepted():
                pass
            else:
                sInfoAccept += "<UUUPS!>" # Das soltle nicht passieren dürfen

        if self.isHandmade():
            if self.isSpecial():
                # Special
                sInfoWay = sInfoAuth = sInfoAdd = sInfoAccept = ""
                sInfoHandmade = "<SPECIAL TAXON> "
            else:
                # Handmade
                sInfoHandmade = "<CAUTION:HANDMADE TAXON> "

        # Zusammenbauen
        sInfo += sInfoErr + sInfoAphia + sInfoWay + sInfoName + sInfoAuth + sInfoAdd + sInfoAccept + sInfoHandmade
        if sInfo is not None:
            sInfo = sInfo.strip()
            sInfo = PrettyText.clean_2ascii_long(sInfo)
        return sInfo

    # ---

    def hasProblems(self):
        if not self.isFound() or self.isErr():
            return True
        if self.hasProblems_status() or self.hasProblems_names() or self.hasProblems_aphia():
            return True
        return False

    def hasProblems_status(self):
        # Nur ein Status ist erlaubt
        if (self.isAccepted() and (self.isLimbo() or self.isUnaccepted())) or \
            (self.isLimbo() and (self.isAccepted() or self.isUnaccepted())) or \
            (self.isUnaccepted() and (self.isAccepted() or self.isLimbo())):
            return True
        # Akzeptiert, nicht Limbo und unakzeptierter Grund ist vorhanden
        if self.isAccepted() and not self.isLimbo() and self.getUnacceptReason() is not None:
            return True
        else:
            return False

    def hasProblems_names(self):
        if self.getScientificName() is None:
            # Kein ScientificName
            return True
        if self.isAccepted() and self.getScientificName_accepted() is None:
            # Akzeptiert und kein AcceptedScientificName
            return True
        if (self.isAccepted() or self.isLimbo()) and not self.allNamesEqual():
            # Akzeptiert und die Namen differieren
            return True
        return False

    def hasProblems_aphia(self):
        if self.getAphiaId() is None:
            # Keine Aphia Id (das darf gar nicht passieren)
            return True
        if self.getAphiaId() is not None and self.getAphiaId_valid() is None and self.getAphiaId_accepted() is not None:
            # Aphia Id, keine ValidAphiaId, AcceptedAphiaId
            return True
        if self.getAphiaId() is not None and self.getAphiaId_valid() is not None and self.getAphiaId_accepted() is not None:
            # Alle Aphia Ids sind gesetzt und ...
            if (self.getAphiaId() == self.getAphiaId_accepted()) and (self.getAphiaId_valid() != self.getAphiaId()):
                # Aphia Id == AcceptedAphiaId != ValidAphiaId (Zirkelschluss)
                return True
        return False

    # ---

    def cmp(self, taxon, **kwargs):
        lstExcludeAttributes = []
        if "exclude" in kwargs:
            lstExcludeAttributes = kwargs["exclude"]
        res = Result()
        diff = {}
        taxonData = taxon.getAllData()
        for k, v in taxonData.items():
            valSelf = self.getData(k)
            valOther = taxonData[k]
            # "" -> None
            valSelf = None if valSelf == "" else valSelf
            valOther = None if valOther == "" else valOther
            if (k not in lstExcludeAttributes) and (valSelf != valOther):
                diff[k] = (valSelf, valOther)
        bEqual = True if len(diff) == 0 else False
        res.addData("equal", bEqual)
        res.addData("diff", diff)
        return res

    def getDateCreate(self):
        return self.__dateCreate

    def debug__setDateCreate(self, datum):
        self.__dateCreate = datum

    def getAge(self):
        return (datetime.datetime.now() - self.getDateCreate()).days

    def isHandmade(self):
        return self.__bHandmade

    def isSpecial(self):
        return self.__bSpecial

    def isFound(self):
        return True if self.__dicData is not None else False

    def isNotFound(self):
        return not self.isFound()

    def isOk(self):
        return self.__result.isOk()

    def isErr(self):
        return not self.isOk()

    def allAphiaIdsEqual(self):
        if self.getAphiaId() == self.getAphiaId_accepted() == self.getAphiaId_valid():
            return True
        else:
            return False

    def allNamesEqual(self):
        if self.getScientificName() == self.getScientificName_valid() == self.getScientificName_accepted():
            return True
        else:
            return False

    def isAccepted(self):
        # aid==vaid==aaid
        # Status == accepted
        if self.allAphiaIdsEqual() and str(self.getStatus()).lower() == "accepted":
            return True
        else:
            return False

    def isUnaccepted(self):
        # !(aid==vaid==aAid)
        # Status == unaccepted
        if not self.allAphiaIdsEqual() or str(self.getStatus()).lower() == "unaccepted":
            return True
        else:
            return False

    def isRedirected(self):
        if self.getAphiaId_accepted() is not None and (self.getAphiaId_accepted() != self.getAphiaId()):
            return True
        else:
            return False

    def isRedirected_2_accepted(self):
        # Taxon: (aid==vaid==aaid) und status==unaccepted
        # Redirected Taxon kann immer noch Probleme haben
        if self.__bLeadingToRealAcceptedTaxon:
            return True
        else:
            return False

    def isRedirected_2_unaccepted(self):
        if (self.isUnaccepted() or self.isLimbo()) and self.isRedirected() and not self.isRedirected_2_accepted():
            return True
        else:
            return False

    def isLimbo(self):
        # aid==vaid==aaid
        # Status != accepted, unaccepted
        if self.allAphiaIdsEqual() and str(self.getStatus()).lower() not in ("accepted", "unaccepted"):
            return True
        else:
            return False

    def hasNoNameAtAll(self):
        if self.getScientificName() is None and self.getScientificName_valid() is None and self.getScientificName_accepted() is None:
            return True
        else:
            return False

    def hasNameGaps(self):
        if (self.getScientificName() in ["", None]) or (self.getScientificName_accepted() in ["", None]):
            # Kein ScientificName oder AcceptedScientificName
            return True
        else:
            return False

    def hasAnyName(self):
        return not self.hasNoNameAtAll()

    def getAnyName(self):
        if self.getScientificName() is not None:
            return self.getScientificName(), "name"
        if self.getScientificName_accepted() is not None:
            return self.getScientificName_accepted()+" {accepted name}", "aname"
        if self.getScientificName_valid() is not None:
            return self.getScientificName_valid()+" {valid name}", "vname"
        return "{no name at all}", None

    def hasNoAaid(self):
        return not self.hasAaid()

    def hasAaid(self):
        if self.getData("aaid") is None:
            return False
        else:
            return True

    def getResult(self):
        return self.__result

    def getAphiaId(self):
        try:
            return self.__dicData["aid"]
        except:
            return None

    def getScientificName(self):
        try:
            return self.__dicData["scientificname"]
        except:
            return None

    def getAuthority(self):
        try:
            return self.__dicData["authority"]
        except:
            return None

    def getAphiaId_valid(self):
        try:
            return self.__dicData["vaid"]
        except:
            return None

    def getScientificName_valid(self):
        try:
            return self.__dicData["valid_name"]
        except:
            return None

    def getAuthority_valid(self):
        try:
            return self.__dicData["valid_authority"]
        except:
            return None

    def getAphiaId_accepted(self):
        try:
            return self.__dicData["aaid"]
        except:
            return None

    def getScientificName_accepted(self):
        try:
            return self.__dicData["aname"]
        except:
            return None

    def getAuthority_accepted(self):
        try:
            return self.__dicData["aauth"]
        except:
            return None

    def getAcceptedWay(self):
        lstWay = self.__dicData.get("aaid_way")
        if lstWay is not None:
            lstWay = list(lstWay)
        return lstWay

    def getRank(self):
        try:
            return self.__dicData["rank"]
        except:
            return None

    def getKingdom(self):
        try:
            return self.__dicData["kingdom"]
        except:
            return None

    def getPhylum(self):
        try:
            return self.__dicData["phylum"]
        except:
            return None

    def getClass(self):
        try:
            return self.__dicData["class"]
        except:
            return None

    def getOrder(self):
        try:
            return self.__dicData["order"]
        except:
            return None

    def getFamily(self):
        try:
            return self.__dicData["family"]
        except:
            return None

    def getGenus(self):
        try:
            return self.__dicData["genus"]
        except:
            return None

    def getLsid(self):
        try:
            return self.__dicData["lsid"]
        except:
            return None

    def getMatchType(self):
        try:
            return self.__dicData["match_type"]
        except:
            return None

    def isFreshwater(self):
        try:
            return bool(self.__dicData["isfreshwater"])
        except:
            return None

    def isBrackish(self):
        try:
            return bool(self.__dicData["isbrackish"])
        except:
            return None

    def isTerrestrial(self):
        try:
            return bool(self.__dicData["isterrestrial"])
        except:
            return None

    def isExtinct(self):
        try:
            return bool(self.__dicData["isexctinct"])
        except:
            return None

    def isMarine(self):
        try:
            return bool(self.__dicData["ismarine"])
        except:
            return None

    def getCitation(self):
        try:
            return self.__dicData["citation"]
        except:
            return None

    def getUrl(self):
        try:
            return self.__dicData["url"]
        except:
            return None

    def getModified(self):
        try:
            return self.__dicData["modified"]
        except:
            return None

    def getStatus(self):
        try:
            return self.__dicData["status"]
        except:
            return None

    def getUnacceptReason(self):
        try:
            return self.__dicData["unacceptreason"]
        except:
            return None

    def getAllData(self):
        if not self.isFound() or self.isErr():
            return None
        else:
            return self.__dicData.copy()

    def getData(self, sName):
        if sName is None or sName == "" or not self.isFound() or self.isErr():
            return None
        elif sName in self.__dicData:
            return self.__dicData[sName]
        else:
            return None

    def overide(self, key, val):
        if self.isFound() and self.isOk:
            self.__dicData[key] = val
            return True
        else:
            return False

    def __cmpHeader(self, lst1, lst2):
        res = Result()
        setMissing = set(lst1).difference(set(lst2))
        setSurplus = set(lst2).difference(set(lst1))
        bMissing = len(setMissing) != 0
        bSurplus = len(setSurplus) != 0
        if bSurplus:
            res.addInfo("Surplus headers in given data", add=PrettyText.lst2strC(setSurplus), kat="Configuration",
                        sys="Taxon")
        if bMissing:
            res.addErr("Missing required headers in given data", add=PrettyText.lst2strC(setMissing),
                       kat="Configuration", sys="Taxon")
        return res

    def __askWorms(self, iAid):
        """
        Initialisiere das Taxon mit Hilfe von WoRMS
        """

        # Sonderfall: AphiaId < 0 (Artefakt oder Eigene Nomenklatur... WoRMS gar nicht erst fragen)
        if iAid < 0:
            self.__result.addErr("AphiaId not found", add=str(iAid), kat="Asking WoRMS", sys="Taxon")
            return None

        iVersuch = 1
        iVersuchMax = 5
        jsonAnswer = None
        dicJsonAnswer = None

        # Versuche eine Verbindung herzustellen
        while True:
            try:
                if iVersuch > 1:
                    # Retry: Cooldown
                    time.sleep(1)

                # Anfrage
                sRequest = 'http://www.marinespecies.org/rest/AphiaRecordByAphiaID/' + str(iAid)
                resp = requests.get(sRequest)

                # Werte antwort aus
                try:
                    jsonAnswer = resp.json()
                except:
                    pass

                # Info Response
                iStatus = resp.status_code

                if iStatus == 200:
                    # Aphia Id gefunden
                    break
                elif iStatus == 204:
                    # Aphia Id nicht gefunden
                    sOut = "WoRMS Response: aid " + str(iAid) + "  not found (return code " + str(iStatus) + ")"
                    jsonAnswer = None
                    self.__result.addErr("AphiaId not found", add=str(iAid), kat="Asking WoRMS", sys="Taxon")
                    break
                elif iStatus == 400:
                    # Fehler
                    raise NameError("WoRMS Response: aid " + str(
                        iAid) + " not found because of bad or missing properites (return code " + str(iStatus) + ")")
                else:
                    # Fehler
                    raise NameError(
                        "WoRMS Response: aid " + str(iAid) + " not found because of other problems (return code " + str(
                            iStatus) + ")")

            except Exception as ex:
                # Fehler
                iVersuch += 1
                # Nicht sofort ein Fehler -> Retry ausprobieren
                sOut = "Problems with aid=" + str(iAid) + ", msg=" + str(ex)
                self.__result.addInfo("AphiaId not found", add=sOut, kat="Asking WoRMS", sys="Taxon")
                if iVersuch > iVersuchMax:
                    # Fehler
                    sOut = "Connection failed " + str(iVersuch - 1) + " times. I give up ..."
                    self.__result.addErr("AphiaId not found", add=sOut, kat="Asking WoRMS", sys="Taxon")
                    break

        # Retries loggen
        if iVersuch > 1:
            sOut = "Connection to WoRMS: worked after " + str(iVersuch) + " tries. "
            self.__result.addInfo("Reconnect worked", add=sOut, kat="Asking WoRMS", sys="Taxon")

        # Kovertiere Header
        if jsonAnswer is not None:
            dicTmp = {}
            for k, v in Taxon.getMapHeaderIntern2Worms().items():
                dicTmp[k] = jsonAnswer[v]
            dicJsonAnswer = dicTmp

        # Speichern
        return dicJsonAnswer

    def __getAaid(self, iAid, lstVisited=None):
        """
        Hole Aaid zu Aid
            Args:
                Aid (int): Aphia Id
                lstVisited (lst): Bereits besuchte AphiaIds
            Returns:
                (Aaid, AScientificName, AAuthority, lstVisited)
        """

        # Eingangstest
        if lstVisited is None:
            lstVisited = []
        if iAid is None or iAid in lstVisited:  # Loop oder None -> Fehler:rauf
            return (None, None, None, lstVisited)

        lstVisited.append(iAid)  # Eintrag in Liste der besuchten Aids

        # Auf-/abstiegstest
        try:
            rec = self.__askWorms(iAid)
            iVaid = rec["vaid"]  # Valid Aid
            sStatus = rec["status"]
            sScientificname = rec["scientificname"]  # Aktueller Scientificname
            sAuthority = rec["authority"] # Aktuelle Authority
        except:
            # Uuups... aus irgendeinem Grund wird zur aid die vaid nicht gefunden
            return (None, None, None, lstVisited)
        if (iVaid is not None) and (iAid == iVaid):
            # ACHTUNG:
            # Spezieller schneller Test bei Abstieg
            # Wenn gefundenes Taxon wirklich akzeptiert ist -> Setzte Flag: Aaid=Vaid=Aid und aTaxon.status == "accepted")
            # Diese information ist nämlich nur hier vorhanden
            if str(sStatus).lower() == "accepted":
                self.__bLeadingToRealAcceptedTaxon = True
            else:
                self.__bLeadingToRealAcceptedTaxon = False
            return (iVaid, sScientificname, sAuthority, lstVisited)  # Fund: wieder hoch
        else:
            return self.__getAaid(iVaid, lstVisited)  # (Noch) nicht akzeptiert oder Fehler -> weiter runter


