"""
Modul mit verschiedensten Gui Widgets
"""

try:
    from PyQt5.QtCore import *
    from PyQt5.QtWidgets import *
    from PyQt5.QtGui import *
    from collections import OrderedDict
except ImportError as ex:
    print("Module missing ("+str(__file__)+"): " + ex.name + "\nPlease install.")

    quit()

class CssPresets:
    progressBar_grey = \
"""
QProgressBar{
    text-align: center
}
QProgressBar::chunk {
    background-color: rgba(0,0,0,35);
}
"""

# coding=utf-8
class QMultiCombo(QToolButton):
    """
    MultiCombobox
    """
    def __init__(self, methodClick=None):
        super().__init__()
        self.setSizePolicy(QSizePolicy.Preferred, QSizePolicy.Preferred)
        font = self.font()
        font.setPointSize(10)
        font.setBold(False)
        self.setFont(font)
        self.setPopupMode(QToolButton.InstantPopup)
        # ---
        self.menu = QMenu()
        self.menu.setFont(font)
        self.setStyleSheet("QMenu { padding: 3px; menu-scrollable: 1;  font-size: 7}")
        self.click = methodClick

    def setName(self, sName):
        self.setText(sName)

    def clear(self):
        self.clearName()
        self.clearEntries()

    def clearName(self):
        self.setText(" ")

    def clearEntries(self):
        self.menu = QMenu(self)
        self.setMenu(self.menu)

    def entryClicked(self, action):
        self.click()

    def setEntries(self, odicEntries, **kwargs):
        # Alle Einträge nach Key aufsteigend sortieren
        # TODO: Wenn ein Key = None ist knall es hier
        #odicEntries = OrderedDict(sorted(dicEntries.items(), key=lambda t: t[0]))

        # Loop durch alle Einträge
        for sEntry, (bEffective, bActive) in odicEntries.items():

            # Effektive Einträge herausfiltern
            if bEffective:
                # Action
                sEntry = sEntry if sEntry is not None else "<not defined>"
                action = self.menu.addAction(sEntry)
                action.setCheckable(True)
                action.setEnabled(True)
                action.setChecked(bActive)
                # Font
                font = self.font()
                font.setPointSize(10)
                action.setFont(font)

                # Was soll bei einem User-Click gemacht werden
                if self.click is not None:
                    action.triggered.connect(self.click)

        self.setMenu(self.menu)

    def getEntries(self, **kwargs):
        bInverted = True if "invert" in kwargs else False
        bToggle = True if "toggle_on_off" in kwargs else False
        dicEntries = {}
        if bToggle:
            # Toggle all: an/aus
            bAllOff = False  # Soll alles an oder ausgeschaltet werden
            for action in self.menu.actions():
                bAllOff |= action.isChecked()
            for action in self.menu.actions():
                dicEntries[action.iconText()] = not bAllOff
        else:
            # Invert all
            for action in self.menu.actions():
                dicEntries[action.iconText()] = action.isChecked() if not bInverted else not action.isChecked()
        return dicEntries

class ColorGradient:
    '''
    Color Gradient
    '''

    def __init__(self, start_hex, finish_hex, n):
        # start_hex = start_hex[1:] if start_hex[0] == "#" else start_hex
        # finish_hex = finish_hex[1:] if finish_hex[0] == "#" else finish_hex
        self.__start_hex = start_hex
        self.__finish_hex= finish_hex
        self.__n = n
        self.__cols = {}
        self.__create()

    def getHtml(self, i):
        r,g,b = self.getRGB(i)
        return '#%02x%02x%02x' % (r, g, b)

    def getRGB(self, i):
        if i < 0:
            i = 0
        if i > 100:
            i = 100
        try:
            r = self.__cols['r'][i-1]
            g = self.__cols['g'][i-1]
            b = self.__cols['b'][i-1]
            return r,g,b
        except:
            return 255,255,255

    def __create(self):
        # Erzeuge Gradienten
        s = self.__hex_to_RGB(self.__start_hex)
        f = self.__hex_to_RGB(self.__finish_hex)
        RGB_list = [s]
        for t in range(1, self.__n):
            curr_vector = [
                int(s[j] + (float(t) / (self.__n - 1)) * (f[j] - s[j]))
                for j in range(3)
            ]
            RGB_list.append(curr_vector)
        # Gradient
        self.__cols = self.__color_dict(RGB_list)

    def __hex_to_RGB(self, hex):
        return [int(hex[i:i + 2], 16) for i in range(1, 6, 2)]

    def __RGB_to_hex(self, RGB):
        # Components need to be integers for hex to make sense
        RGB = [int(x) for x in RGB]
        return "#" + "".join(["0{0:x}".format(v) if v < 16 else
                              "{0:x}".format(v) for v in RGB])
    def __color_dict(self, gradient):
        return {"hex": [self.__RGB_to_hex(RGB) for RGB in gradient],
                "r": [RGB[0] for RGB in gradient],
                "g": [RGB[1] for RGB in gradient],
                "b": [RGB[2] for RGB in gradient]}

class Colors:
    """
    Colors
    """
    __colors = {
        "black": "000000",
        "white": "FFFFFF",
        "awi_blue_dark": "173E6A",
        "awi_blue_mid": "425a89",
        "awi_blue_light": "4FAAE2",
        "bsh_blue_dark": "163B66",
        "bsh_blue_mid": "425a89",
        "bsh_blue_light": "4c709c",
        "blue": "488AC3",
        "red": "D37753",
        "red_01": "E93531",
        "green": "6B8E23",
        "green_light": "62BE48",
        "yellow": "E6D881",
        "purple": "D43281",
        "orange": "EDA766",
        "rose": "E9885C"
    }

    @staticmethod
    def getNames():
        return Colors.__colors.keys()

    @staticmethod
    def getColor(sName, **kwargs):
        bHex = kwargs["hex"] if "hex" in kwargs else False
        sColor_code = Colors.__colors.get(sName)
        if sColor_code is None:
            return None
        if bHex:
            sColor_ret = sColor_code
        else:
            sColor_ret = "#" + sColor_code
        return sColor_ret







