"""
Modul mit View
"""

try:
    import sys, os, time, psycopg2, csv, re, folium, webbrowser, ctypes, random, traceback, subprocess, warnings, psutil, webbrowser, datetime
    import openpyxl
    from pathlib import Path
    from distutils.dir_util import copy_tree
    from folium.plugins import MarkerCluster
    from collections import OrderedDict
    from functools import partial
    from shutil import copyfile
    from PyQt5 import uic
    from PyQt5.QtCore import *
    from PyQt5.QtGui import *
    from PyQt5.QtWidgets import *
    from PyQt5.QtWebEngineCore import QWebEngineUrlRequestInterceptor
    from PyQt5.QtWebEngineWidgets import QWebEngineView
    from helper.thread_container import *
    from tools.parallel import Worker, Stoppable
    from tools.printing import SmartPrinter
    from tools.taxon import Taxon
    from tools.text import TextEncoding, PrettyText
    from tools.csvfiles import CsvTools
    from tools.excel import ExcelTools
    from tools.coding import Result, Package
    from tools.ui import QMultiCombo, ColorGradient, CssPresets
    from tools.misc import JsonConverter

except ImportError as ex:
    print("Module missing ("+str(__file__)+"): " + ex.name + "\nPlease install.")
    quit()

# UI-Definitionen
fileHauptfenster = os.path.dirname(os.path.realpath(__file__)) + "/ressource/mainwindow.ui"
fileOptionen = os.path.dirname(os.path.realpath(__file__)) + "/ressource/dialog_options.ui"
fileProgress = os.path.dirname(os.path.realpath(__file__)) + "/ressource/progress.ui"
fileWaiting = os.path.dirname(os.path.realpath(__file__)) + "/ressource/waiting.ui"
fileCsvExtractNewData = os.path.dirname(os.path.realpath(__file__)) + "/ressource/dialog_csv_extract_new_data.ui"
fileDiff = os.path.dirname(os.path.realpath(__file__)) + "/ressource/dialog_diff.ui"
fileFaoCreate = os.path.dirname(os.path.realpath(__file__)) + "/ressource/dialog_fao_create.ui"
fileBatch = os.path.dirname(os.path.realpath(__file__)) + "/ressource/dialog_batch.ui"
fileImport = os.path.dirname(os.path.realpath(__file__)) + "/ressource/dialog_import.ui"
fileExSub = os.path.dirname(os.path.realpath(__file__)) + "/ressource/dialog_exportsubset.ui"
fileUpdate = os.path.dirname(os.path.realpath(__file__)) + "/ressource/dialog_update.ui"
fileIngest = os.path.dirname(os.path.realpath(__file__)) + "/ressource/dialog_ingest.ui"
fileVaBiotaPolish = os.path.dirname(os.path.realpath(__file__)) + "/ressource/dialog_taxapolish.ui"
fileVaTaxonSpotter = os.path.dirname(os.path.realpath(__file__)) + "/ressource/dialog_taxonspotter.ui"
fileGenericComboChoice = os.path.dirname(os.path.realpath(__file__)) + "/ressource/dialog_generic_combo_choice.ui"
fileExport = os.path.dirname(os.path.realpath(__file__)) + "/ressource/dialog_export.ui"
fileExportSubset = os.path.dirname(os.path.realpath(__file__)) + "/ressource/dialog_export_subset.ui"
fileSysInfo = os.path.dirname(os.path.realpath(__file__)) + "/ressource/dialog_systeminfo.ui"

class View(QMainWindow, uic.loadUiType(fileHauptfenster)[0]):
    """
    VIEW

    - Diese Klasse definiert das Hauptfenster
    - Die View ist "intelligent", denn sie besitzt Verwaltungslogik
    """

    def __init__(self, model):
        """
        Args:
            model (gui.model.Model): Model
        """

        # Super
        QMainWindow.__init__(self)

        # Instanzvariablen
        self.__model = model
        self.__map = model.subsystem_getMapper()
        self.__smartprinter = SmartPrinter(model=model)
        self.__statusbar = QStatusBar()
        font = QFont()
        font.setPixelSize(10)
        self.__statusbar.setFont(font)
        self.__browser = None

        self.__bMapForceDrawFirstTimeQuestion = True
        self.__bMapForceDraw = False
        self.__lstSelectedIngests = []
        self.__lstSelectedActionLogs = []

        # -----------

        # Spezielle Gui Elemente als Objektvariablen merken

        # --- Initialisierug

        self.setupUi(self)
        self.setWindowFlags(Qt.Dialog | Qt.CustomizeWindowHint | Qt.WindowTitleHint | Qt.WindowCloseButtonHint)

        # Taxonsystem Daten
        self.__odicTaxonSystemData = OrderedDict([
            ("wormsbase", ("TaxBase", self.taxSysBaseInfo)),
            ("wormsaaid", ("TaxAaid", self.taxSysAaidInfo)),
            ("wormscolony", ("TaxColony", self.taxSysColonyInfo)),
            ("wormsprivate", ("TaxPrivate", self.taxSysPrivateInfo)),
            ("wormschange", ("TaxChange", self.taxSysChangeInfo)),
            ("noworms", ("TaxOut", self.taxSysNowormsInfo))
        ])

        # Window Icon
        self.setWindowIcon(QIcon(os.path.dirname(os.path.realpath(__file__)) + "/ressource/critterbase.png"))
        self.setWindowFlags(Qt.Dialog | Qt.CustomizeWindowHint | Qt.WindowTitleHint | Qt.WindowCloseButtonHint)

        # Spezial Windows: Setzte Taskbar Icon für Windows
        try:
            if os.name == "nt":
                myappid = u'awi.functional-ecology.tools.0'
                ctypes.windll.shell32.SetCurrentProcessExplicitAppUserModelID(myappid)
        except:
            pass
            # Titel Dryrun
        self.status_refresh_dryMode()
            # Statusbar einfügen
        self.setStatusBar(self.__statusbar)
            # Progressbar ausblenden
        self.ui_hideProgress()
        self.ui_hideBatch()

        # *** Menu
        # BatchBot
        action = QAction("BatchBot ", self)
        action.triggered.connect(self.main_batchBot)
        self.menuBatchBot.addAction(action)
        # Ingest/Template
        for sWf in self.__model.subsystem_getMapper().workflow_getAll():
            sWfLong = self.__model.subsystem_getMapper().workflow_short2long(sWf)
            action = QAction("Ingest "+sWfLong, self)
            action.triggered.connect(partial(self.import_ingest, sWf))
            self.menuIngest.addAction(action)
            action = QAction("Template "+sWfLong, self)
            action.triggered.connect(partial(self.import_ingest_template, wf=sWf, ingest=True, lookup=True))
            self.menuTemplate.addAction(action)
        # Convert
        menuConvertIngest = QMenu("Convert and Ingest", self)
        menuConvertTemplate = QMenu("Template", self)
        menuConvertDoc = QMenu("Consistency check", self)
        self.menuConvert.addMenu(menuConvertIngest)
        self.menuConvert.addMenu(menuConvertTemplate)
        self.menuConvert.addMenu(menuConvertDoc)
        for sWf_conf in self.__model.subsystem_getMapper().workflow_converter_import_getAll():
            sWf_conf_long = self.__model.subsystem_getMapper().workflow_converter_import_short2long(sWf_conf)
            sWf_conf_target = self.__model.subsystem_getMapper().workflow_converter_import_getTarget(sWf_conf)
            sWf_conf_target_long = self.__model.subsystem_getMapper().workflow_short2long(sWf_conf_target)
            actionConvert = QAction(str(sWf_conf_long)+" -> " +str(sWf_conf_target_long), self)
            actionConvert.triggered.connect(partial(self.import_ingest, sWf_conf))
            menuConvertIngest.addAction(actionConvert)
            actionTemplate = QAction(sWf_conf_long, self)
            actionTemplate.triggered.connect(partial(self.converter_templateConverter, sWf_conf))
            menuConvertTemplate.addAction(actionTemplate)
        actionMarlinDok = QAction("MARLIN (Transport)", self)
        actionMarlinDok.triggered.connect(self.converter_marlin_doc)
        menuConvertDoc.addAction(actionMarlinDok)
        # Lookup
        lstTables = self.__map.lookup_getDbTables()
        menu = QMenu("New Entry", self)
        self.menuLup.addMenu(menu)
        for sTabName in lstTables:
            if sTabName != "taxon":
                sTitel = self.__map.lookup_table2sheet(sTabName)
                action = QAction(sTitel, self)
                action.triggered.connect(partial(self.admin_createLookupOrTaxon, sTabName, "New: " + sTitel))
                menu.addAction(action)
                #self.toolButtonLookupMenu.addAction(action)
        action = QAction("Import", self)
        action.triggered.connect(self.import_lookups)
        self.menuLup.addAction(action)
        action = QAction("Export", self)
        action.triggered.connect(self.admin_lookup_export)
        self.menuLup.addAction(action)
        action = QAction("Template", self)
        action.triggered.connect(partial(self.import_ingest_template, lookup=True))
        self.menuLup.addAction(action)
        # Regions
        action = QAction("Create FAO regions", self)
        action.triggered.connect(partial(self.fao_cmd, "reg_create"))
        self.menuRegions.addAction(action)
        action = QAction("Import regions from file", self)
        action.triggered.connect(partial(self.fao_cmd, "import"))
        self.menuRegions.addAction(action)
        action = QAction("Export regions to file", self)
        action.triggered.connect(partial(self.fao_cmd, "export"))
        self.menuRegions.addAction(action)
        action = QAction("Show FAO regions", self)
        action.triggered.connect(partial(self.fao_cmd, "reg_show"))
        self.menuRegions.addAction(action)
        action = QAction("Clean mapping", self)
        action.triggered.connect(partial(self.fao_cmd, "reg_clean"))
        self.menuRegions.addAction(action)
        action = QAction("Show mapping", self)
        action.triggered.connect(partial(self.fao_cmd, "reg_stat"))
        self.menuRegions.addAction(action)
        action = QAction("Update mapping", self)
        action.triggered.connect(partial(self.fao_cmd, "reg_update"))
        self.menuRegions.addAction(action)
        # TaxonWizard
        menu = QMenu("New entry", self)
        self.menuTaxWiz.addMenu(menu)
        for sTabName, (sViewName, widgetInfo) in self.__odicTaxonSystemData.items():
            action = QAction(sViewName, self)
            action.triggered.connect(partial(self.admin_createLookupOrTaxon, sTabName, "New: " + self.__map.taxon_table2sheet(sTabName), taxHelper=True))
            menu.addAction(action)
        action = QAction("Start Update", self)
        action.triggered.connect(partial(self.taxon_cmd, "update"))
        self.menuTaxWiz.addAction(action)
        action = QAction("Log of Last Update", self)
        action.triggered.connect(partial(self.ui_table_stat, "taxlistupdate"))
        self.menuTaxWiz.addAction(action)
        action = QAction("Import", self)
        action.triggered.connect(self.import_taxon)
        self.menuTaxWiz.addAction(action)
        action = QAction("Export", self)
        action.triggered.connect(self.admin_export_taxonList)
        self.menuTaxWiz.addAction(action)
        action = QAction("Template", self)
        action.triggered.connect(partial(self.import_ingest_template, taxon=True))
        self.menuTaxWiz.addAction(action)
        # MyWorms
        action = QAction("Start DeAging", self)
        action.triggered.connect(partial(self.taxon_cmd, "puffer-update"))
        self.menuMyworms.addAction(action)
        action = QAction("Take Snapshot", self)
        action.triggered.connect(partial(self.taxon_cmd, "puffer-snapshot"))
        self.menuMyworms.addAction(action)
        action = QAction("Show Status", self)
        action.triggered.connect(partial(self.taxon_cmd, "puffer-status"))
        self.menuMyworms.addAction(action)
        action = QAction("Clean all", self)
        action.triggered.connect(partial(self.taxon_cmd, "puffer-clean"))
        self.menuMyworms.addAction(action)
        action = QAction("Reload all", self)
        action.triggered.connect(partial(self.taxon_cmd, "puffer-reload"))
        self.menuMyworms.addAction(action)
        action = QAction("Save", self)
        action.triggered.connect(partial(self.taxon_cmd, "puffer-save"))
        self.menuMyworms.addAction(action)
        action = QAction("Import", self)
        action.triggered.connect(partial(self.taxon_cmd, "puffer-import"))
        self.menuMyworms.addAction(action)
        action = QAction("Export", self)
        action.triggered.connect(partial(self.taxon_cmd, "puffer-export"))
        self.menuMyworms.addAction(action)
        # DataPatrol
        action = QAction("BiotaControl", self)
        action.triggered.connect(partial(self.dataPatrol_cmd, "biota_control"))
        self.menuDataPatrol.addAction(action)
        # Assistant
        action = QAction("TaxaPolishing", self)
        action.triggered.connect(partial(self.virtualAssistent_taxaPolishing))
        self.menuAssistant.addAction(action)
        action = QAction("TaxonSpotter", self)
        action.triggered.connect(partial(self.virtualAssistant_taxonSpotter))
        self.menuAssistant.addAction(action)
        # Toolbox
        action = QAction("Extract new Entries from CSV File", self)
        action.triggered.connect(partial(self.toolbox_extractNewDataFromCsv))
        self.menuTools.addAction(action)
        action = QAction("Show Content of CSV File", self)
        action.triggered.connect(partial(self.toolbox_showCsvFile))
        self.menuTools.addAction(action)
        action = QAction("Version Control", self)
        action.triggered.connect(partial(self.toolbox_versionControl))
        self.menuTools.addAction(action)
        # Extra
        action = QAction("Optionen", self)
        action.triggered.connect(self.main_optionen)
        self.menuExtra.addAction(action)
        # UI
        action = QAction("Toggle console orientation", self)
        action.triggered.connect(partial(self.ui_consoleViewChangeOrientation))
        self.menuUi.addAction(action)
        # Console View: Change visibility
        action = QAction("Toggle console visibility", self)
        action.triggered.connect(partial(self.ui_consoleViewChangeVisibility))
        self.menuUi.addAction(action)
        # Info
        action = QAction("Update App", self)
        action.triggered.connect(self.help_update)
        self.menuHelp.addAction(action)
        action = QAction("Changelog App", self)
        action.triggered.connect(self.help_appChangelog)
        self.menuHelp.addAction(action)
        action = QAction("Changelog Data Model", self)
        action.triggered.connect(self.help_datenmodellChangelog)
        self.menuHelp.addAction(action)
        action = QAction("System", self)
        action.triggered.connect(self.help_system)
        self.menuHelp.addAction(action)
        action = QAction("Documentation", self)
        action.triggered.connect(self.help_help)
        self.menuHelp.addAction(action)
        action = QAction("Info", self)
        action.triggered.connect(self.help_info)
        self.menuHelp.addAction(action)

        # *** ToolBar
        self.toolBar.setContextMenuPolicy(Qt.PreventContextMenu)
        buttonSize = QSize(48, 48)
        # Ingest
        toolButtonIngest = QToolButton()
        toolButtonIngest.setIcon(QIcon(os.path.dirname(os.path.realpath(__file__)) + "/ressource/icon_import.png"))
        toolButtonIngest.setIconSize(buttonSize)
        toolButtonIngest.setToolTip("Ingest data")
        toolButtonIngestMenu = QMenu()
        for sWf in self.__model.subsystem_getMapper().workflow_getAll():
            sWfLong = self.__model.subsystem_getMapper().workflow_short2long(sWf)
            action = QAction("Ingest "+sWfLong, self)
            action.triggered.connect(partial(self.import_ingest, sWf))
            toolButtonIngestMenu.addAction(action)
        toolButtonIngest.setMenu(toolButtonIngestMenu)
        toolButtonIngest.setPopupMode(QToolButton.InstantPopup)
        self.toolBar.addWidget(toolButtonIngest)
        # Template
        toolButtonTempl = QToolButton()
        toolButtonTempl.setIcon(QIcon(os.path.dirname(os.path.realpath(__file__)) + "/ressource/icon_excel.png"))
        toolButtonTempl.setIconSize(buttonSize)
        toolButtonTempl.setToolTip("Export Excel template")
        toolButtonTemplMenu = QMenu()
        for sWf in self.__model.subsystem_getMapper().workflow_getAll():
            sWfLong = self.__model.subsystem_getMapper().workflow_short2long(sWf)
            action = QAction("Template "+sWfLong, self)
            action.triggered.connect(partial(self.import_ingest_template, wf=sWf, ingest=True, lookup=True))
            toolButtonTemplMenu.addAction(action)
        toolButtonTempl.setMenu(toolButtonTemplMenu)
        toolButtonTempl.setPopupMode(QToolButton.InstantPopup)
        self.toolBar.addWidget(toolButtonTempl)
        # Converter
        toolButtonConvert = QToolButton()
        toolButtonConvert.setIcon(QIcon(os.path.dirname(os.path.realpath(__file__)) + "/ressource/icon_convert.png"))
        toolButtonConvert.setIconSize(buttonSize)
        toolButtonConvert.setToolTip("Convert data")
        toolButtonConvertMenu = QMenu()
        toolButtonConvertMenuIngest = toolButtonConvertMenu.addMenu("Convert and Ingest")
        toolButtonConvertMenuTemplate = toolButtonConvertMenu.addMenu("Template")
        toolButtonConvertMenuDoc = toolButtonConvertMenu.addMenu("Consistency check")
        for sWf_conf in self.__model.subsystem_getMapper().workflow_converter_import_getAll():
            sWf_conf_long = self.__model.subsystem_getMapper().workflow_converter_import_short2long(sWf_conf)
            sWf_conf_target = self.__model.subsystem_getMapper().workflow_converter_import_getTarget(sWf_conf)
            sWf_conf_target_long = self.__model.subsystem_getMapper().workflow_short2long(sWf_conf_target)
            actionConvert = QAction(str(sWf_conf_long) + " -> " + str(sWf_conf_target_long), self)
            actionConvert.triggered.connect(partial(self.import_ingest, sWf_conf))
            actionConvert = QAction(str(sWf_conf_long)+" -> "+str(sWf_conf_target_long), self)
            actionConvert.triggered.connect(partial(self.import_ingest, sWf_conf))
            toolButtonConvertMenuIngest.addAction(actionConvert)
            actionTemplate = QAction(sWf_conf_long, self)
            actionTemplate.triggered.connect(partial(self.converter_templateConverter, sWf_conf))
            toolButtonConvertMenuTemplate.addAction(actionTemplate)
        toolButtonConvert.setMenu(toolButtonConvertMenu)
        toolButtonConvert.setPopupMode(QToolButton.InstantPopup)
        actionMarlinDok = QAction("MARLIN (Transport)", self)
        actionMarlinDok.triggered.connect(self.converter_marlin_doc)
        toolButtonConvertMenuDoc.addAction(actionMarlinDok)
        self.toolBar.addWidget(toolButtonConvert)
        # Lookuptable
        toolButtonLookup = QToolButton()
        toolButtonLookup.setIcon(QIcon(os.path.dirname(os.path.realpath(__file__)) + "/ressource/icon_lookup2.png"))
        toolButtonLookup.setIconSize(buttonSize)
        toolButtonLookup.setToolTip("Lookup")
        toolButtonLookupMenu = QMenu()
        toolButtonLookup.setMenu(toolButtonLookupMenu)
        toolButtonLookup.setPopupMode(QToolButton.InstantPopup)
        toolButtonLookup.setMenu(self.menuLup)
        self.toolBar.addWidget(toolButtonLookup)
        # FAO Regions
        toolButtonFaoRegions = QToolButton()
        toolButtonFaoRegions.setIcon(
            QIcon(os.path.dirname(os.path.realpath(__file__)) + "/ressource/icon_world2.png"))
        toolButtonFaoRegions.setIconSize(buttonSize)
        toolButtonFaoRegions.setToolTip("FAO Regions")
        toolButtonFaoRegionsMenu = QMenu()
        toolButtonFaoRegions.setMenu(toolButtonFaoRegionsMenu)
        toolButtonFaoRegions.setPopupMode(QToolButton.InstantPopup)
        toolButtonFaoRegions.setMenu(self.menuRegions)
        self.toolBar.addWidget(toolButtonFaoRegions)

        # TaxonWizard
        toolButtonTaxonsystem = QToolButton()
        toolButtonTaxonsystem.setIcon(QIcon(os.path.dirname(os.path.realpath(__file__)) + "/ressource/icon_gandalf1.png"))
        toolButtonTaxonsystem.setIconSize(buttonSize)
        toolButtonTaxonsystem.setToolTip("TaxonWizard")
        toolButtonTaxonsystemMenu = QMenu()
        toolButtonTaxonsystem.setMenu(toolButtonTaxonsystemMenu)
        toolButtonTaxonsystem.setPopupMode(QToolButton.InstantPopup)
        toolButtonTaxonsystem.setMenu(self.menuTaxWiz)
        self.toolBar.addWidget(toolButtonTaxonsystem)
        # DataPatrol
        toolButtonDataPatrol = QToolButton()
        toolButtonDataPatrol.setIcon(
            QIcon(os.path.dirname(os.path.realpath(__file__)) + "/ressource/icon_police1.png"))
        toolButtonDataPatrol.setIconSize(buttonSize)
        toolButtonDataPatrol.setToolTip("DataPatrol")
        toolButtonDataPatrolMenu = QMenu()
        toolButtonDataPatrol.setMenu(toolButtonTaxonsystemMenu)
        toolButtonDataPatrol.setPopupMode(QToolButton.InstantPopup)
        toolButtonDataPatrol.setMenu(self.menuDataPatrol)
        self.toolBar.addWidget(toolButtonDataPatrol)
        # MyWoRMS
        toolButtonWymorms = QToolButton()
        toolButtonWymorms.setIcon(QIcon(os.path.dirname(os.path.realpath(__file__)) + "/ressource/icon_myworms.png"))
        toolButtonWymorms.setIconSize(buttonSize)
        toolButtonWymorms.setToolTip("MyWoRMS")
        toolButtonWymormsMenu = QMenu()
        toolButtonWymorms.setMenu(toolButtonWymormsMenu)
        toolButtonWymorms.setPopupMode(QToolButton.InstantPopup)
        toolButtonWymorms.setMenu(self.menuMyworms)
        self.toolBar.addWidget(toolButtonWymorms)
        # Virtual Assistent
        toolButtonAssistent = QToolButton()
        toolButtonAssistent.setIcon(QIcon(os.path.dirname(os.path.realpath(__file__)) + "/ressource/icon_assistant2.png"))
        toolButtonAssistent.setIconSize(buttonSize)
        toolButtonAssistent.setToolTip("VirtualAssistant")
        toolButtonAssistentMenu = QMenu()
        toolButtonAssistent.setMenu(self.menuAssistant)
        toolButtonAssistent.setPopupMode(QToolButton.InstantPopup)
        self.toolBar.addWidget(toolButtonAssistent)
        # BatchBot
        toolButtonBatchBot = QToolButton()
        toolButtonBatchBot.setIcon(QIcon(os.path.dirname(os.path.realpath(__file__)) + "/ressource/icon_batchbot.png"))
        toolButtonBatchBot.setToolTip("BatchBot")
        toolButtonBatchBot.setIconSize(buttonSize)
        toolButtonBatchBot.clicked.connect(self.main_batchBot)
        self.toolBar.addWidget(toolButtonBatchBot)
        # Toolbox
        # toolButtonTools = QToolButton()
        # toolButtonTools.setIcon(QIcon(os.path.dirname(os.path.realpath(__file__)) + "/ressource/icon_toolbox.png"))
        # toolButtonTools.setIconSize(buttonSize)
        # toolButtonTools.setToolTip("Toolbox")
        # toolButtonToolsMenu = QMenu()
        # toolButtonTools.setMenu(self.menuTools)
        # toolButtonTools.setPopupMode(QToolButton.InstantPopup)
        # self.toolBar.addWidget(toolButtonTools)
        # Optionen
        toolButtonOpt = QToolButton()
        toolButtonOpt.setIcon(QIcon(os.path.dirname(os.path.realpath(__file__)) + "/ressource/icon_settings.png"))
        toolButtonOpt.setIconSize(buttonSize)
        toolButtonOpt.setToolTip("Settings")
        toolButtonOpt.clicked.connect(self.main_optionen)
        self.toolBar.addWidget(toolButtonOpt)
        # Beenden
        toolButtonExit = QToolButton()
        toolButtonExit.setIcon(QIcon(os.path.dirname(os.path.realpath(__file__)) + "/ressource/icon_exit.png"))
        toolButtonExit.setIconSize(buttonSize)
        toolButtonExit.setToolTip("Quit program")
        toolButtonExit.clicked.connect(self.main_beenden)
        self.toolBar.addWidget(toolButtonExit)
        # Combobox
        # Preset veraendern
        self.infoComboDbPreset.currentIndexChanged.connect(self.sys_presetChanged)

        # Andere Buttons
        # Snapshot
        self.buttonSnapshot.setIcon(QIcon(os.path.dirname(os.path.realpath(__file__)) + "/ressource/icon_snapshot.png"))
        self.buttonSnapshot.clicked.connect(self.core_snapshot)
        # Console View: ButtonMenu
        self.toolConsoleView.setIcon(QIcon(os.path.dirname(os.path.realpath(__file__)) + "/ressource/icon_console.png"))
        self.toolConsoleView.setIconSize(buttonSize)
        self.toolConsoleView.setToolTip("Change view of console output ")
        toolConsoleViewMenu = QMenu()
        self.toolConsoleView.setMenu(toolConsoleViewMenu)
        self.toolConsoleView.setPopupMode(QToolButton.InstantPopup)
            # Console View: Change orientation
        action = QAction("Toggle console orientation", self)
        action.triggered.connect(partial(self.ui_consoleViewChangeOrientation))
        toolConsoleViewMenu.addAction(action)
            # Console View: Change visibility
        action = QAction("Toggle console visibility", self)
        action.triggered.connect(partial(self.ui_consoleViewChangeVisibility))
        toolConsoleViewMenu.addAction(action)
            # Visual Explorer
        self.buttonDrawMap.clicked.connect(self.explorer_drawMap)
        self.buttonVizInAllRst.clicked.connect(self.explorer_reset)
        self.buttonVizInAllRst.setIcon(QIcon(os.path.dirname(os.path.realpath(__file__)) + "/ressource/icon_reset.png"))
        self.buttonVizInAllRst.setAutoRaise(True)
        self.buttonVizInDateStartMin.clicked.connect(self.explorer_setDateStartToMin)
        self.buttonVizInDateStartMin.setIcon(QIcon(os.path.dirname(os.path.realpath(__file__)) + "/ressource/icon_down.png"))
        self.buttonVizInDateStartMin.setAutoRaise(True)
        self.buttonVizInDateEndMax.clicked.connect(self.explorer_setDateEndToMax)
        self.buttonVizInDateEndMax.setIcon(QIcon(os.path.dirname(os.path.realpath(__file__)) + "/ressource/icon_up.png"))
        self.buttonVizInDateEndMax.setAutoRaise(True)
        self.buttonVizInDatasetAll.clicked.connect(self.explorer_toggleActivateDataset)
        self.buttonVizInDatasetAll.setIcon(QIcon(os.path.dirname(os.path.realpath(__file__)) + "/ressource/icon_0_1.png"))
        self.buttonVizInDatasetAll.setAutoRaise(True)
        self.buttonVizInRegionAll.clicked.connect(self.explorer_toggleActivateRegion)
        self.buttonVizInRegionAll.setIcon(QIcon(os.path.dirname(os.path.realpath(__file__)) + "/ressource/icon_0_1.png"))
        self.buttonVizInRegionAll.setAutoRaise(True)
        self.buttonVizInTypeAll.clicked.connect(self.explorer_toggleActivateType)
        self.buttonVizInTypeAll.setIcon(QIcon(os.path.dirname(os.path.realpath(__file__)) + "/ressource/icon_0_1.png"))
        self.buttonVizInTypeAll.setAutoRaise(True)
        self.buttonVizInGearAll.clicked.connect(self.explorer_toggleActivateGear)
        self.buttonVizInGearAll.setIcon( QIcon(os.path.dirname(os.path.realpath(__file__)) + "/ressource/icon_0_1.png"))
        self.buttonVizInGearAll.setAutoRaise(True)
        self.buttonVizInTaxcoverAll.clicked.connect(self.explorer_toggleActivateTaxcover)
        self.buttonVizInTaxcoverAll.setIcon(QIcon(os.path.dirname(os.path.realpath(__file__)) + "/ressource/icon_0_1.png"))
        self.buttonVizInTaxcoverAll.setAutoRaise(True)
        self.buttonVizInIngestAll.clicked.connect(self.explorer_toggleActivateIngest)
        self.buttonVizInIngestAll.setIcon(QIcon(os.path.dirname(os.path.realpath(__file__)) + "/ressource/icon_0_1.png"))
        self.buttonVizInIngestAll.setAutoRaise(True)
        self.buttonVizInDatasetInv.clicked.connect(self.explorer_invActivateDataset)
        self.buttonVizInDatasetInv.setIcon(QIcon(os.path.dirname(os.path.realpath(__file__)) + "/ressource/icon_inverse.png"))
        self.buttonVizInDatasetInv.setAutoRaise(True)
        self.buttonVizInRegionInv.clicked.connect(self.explorer_invActivateRegion)
        self.buttonVizInRegionInv.setIcon(QIcon(os.path.dirname(os.path.realpath(__file__)) + "/ressource/icon_inverse.png"))
        self.buttonVizInRegionInv.setAutoRaise(True)
        self.buttonVizInTypeInv.clicked.connect(self.explorer_invActivateType)
        self.buttonVizInTypeInv.setIcon(
            QIcon(os.path.dirname(os.path.realpath(__file__)) + "/ressource/icon_inverse.png"))
        self.buttonVizInTypeInv.setAutoRaise(True)
        self.buttonVizInGearInv.clicked.connect(self.explorer_invActivateGear)
        self.buttonVizInGearInv.setIcon(QIcon(os.path.dirname(os.path.realpath(__file__)) + "/ressource/icon_inverse.png"))
        self.buttonVizInGearInv.setAutoRaise(True)
        self.buttonVizInTaxcoverInv.clicked.connect(self.explorer_invActivateTaxcover)
        self.buttonVizInTaxcoverInv.setIcon( QIcon(os.path.dirname(os.path.realpath(__file__)) + "/ressource/icon_inverse.png"))
        self.buttonVizInTaxcoverInv.setAutoRaise(True)
        self.buttonVizInIngestInv.clicked.connect(self.explorer_invActivateIngest)
        self.buttonVizInIngestInv.setIcon(QIcon(os.path.dirname(os.path.realpath(__file__)) + "/ressource/icon_inverse.png"))
        self.buttonVizInIngestInv.setAutoRaise(True)
        self.vizInDateStart.dateChanged.connect(self.explorer_update)
        self.vizInDateEnd.dateChanged.connect(self.explorer_update)
        self.vizProjectionType.currentIndexChanged.connect(self.explorer_update)
        self.vizMarkerType.currentIndexChanged.connect(self.explorer_update)
        self.vizWorldWrap.clicked.connect(self.explorer_update)
        self.buttonSaveMap.clicked.connect(self.explorer_saveMap)
        self.buttonSaveMap2.clicked.connect(self.explorer_saveMap)
        self.buttonBrowserMap.clicked.connect(self.explorer_openMapInExternalBrowser)
        self.buttonBrowserMap2.clicked.connect(self.explorer_openMapInExternalBrowser)
        # Show Subset
        toolButtonShowSubset = QMenu()
        self.buttonShowSubset.setMenu(toolButtonShowSubset)
        self.buttonShowSubset.setPopupMode(QToolButton.InstantPopup)
        actionShowAWI = QAction("CRITTERBASE", self)
        actionShowAWI.triggered.connect(partial(self.explorer_showSubset))
        actionShowAWI_rfriendly = QAction("CRITTERBASE {R-friendly}", self)
        actionShowAWI_rfriendly.triggered.connect(partial(self.explorer_showSubset, rfriendly=True))
        actionShowBSH = QAction("MARLIN (Transport)", self)
        actionShowBSH.triggered.connect(partial(self.explorer_showSubset, converter="mar_trans"))
        toolButtonShowSubset.addAction(actionShowAWI)
        toolButtonShowSubset.addAction(actionShowAWI_rfriendly)
        toolButtonShowSubset.addAction(actionShowBSH)
        # Export Subset
        toolButtonExportSubset = QMenu()
        self.buttonExportSubset.setMenu(toolButtonExportSubset)
        self.buttonExportSubset.setPopupMode(QToolButton.InstantPopup)
        actionExportAWI = QAction("CRITTERBASE", self)
        actionExportAWI.triggered.connect(partial(self.explorer_exportSubset, rfriendly=False))
        actionExportBSH = QAction("MARLIN (Transport)", self)
        actionExportBSH.triggered.connect(partial(self.explorer_exportSubset, converter="mar_trans"))
        toolButtonExportSubset.addAction(actionExportAWI)
        toolButtonExportSubset.addAction(actionExportBSH)
            # Ingest-Kommandos
        self.btnIngestCsv.clicked.connect(partial(self.import_ingest_cmd, "excel"))
        self.btnIngestCsv.setIcon(QIcon(os.path.dirname(os.path.realpath(__file__)) + "/ressource/icon_excel.png"))
        self.btnIngestLookup.clicked.connect(partial(self.import_ingest_cmd, "lookup"))
        self.btnIngestLookup.setIcon(QIcon(os.path.dirname(os.path.realpath(__file__)) + "/ressource/icon_lookup2.png"))
        self.btnIngestTab.clicked.connect(partial(self.import_ingest_cmd, "tab"))
        self.btnIngestTab.setIcon(QIcon(os.path.dirname(os.path.realpath(__file__)) + "/ressource/icon_sheet.png"))
        self.btnIngestMap.clicked.connect(partial(self.import_ingest_cmd, "map"))
        self.btnIngestMap.setIcon(QIcon(os.path.dirname(os.path.realpath(__file__)) + "/ressource/icon_world.png"))
        self.btnIngestLog.clicked.connect(partial(self.import_ingest_cmd, "log"))
        self.btnIngestLog.setIcon(QIcon(os.path.dirname(os.path.realpath(__file__)) + "/ressource/icon_log.png"))
        self.btnIngestDel.clicked.connect(partial(self.import_ingest_cmd, "del"))
        self.btnIngestDel.setIcon(QIcon(os.path.dirname(os.path.realpath(__file__)) + "/ressource/icon_trash.png"))
        self.btnIngestDepShow.clicked.connect(partial(self.import_ingest_cmd, "dep-show"))
        self.btnIngestDepShow.setIcon(QIcon(os.path.dirname(os.path.realpath(__file__)) + "/ressource/icon_zoom_in.png"))
        self.btnIngestDepHide.clicked.connect(partial(self.import_ingest_cmd, "dep-hide"))
        self.btnIngestDepHide.setIcon(QIcon(os.path.dirname(os.path.realpath(__file__)) + "/ressource/icon_zoom_out.png"))
            # Action Log Kommandos
        self.btnActionLogDel.clicked.connect(partial(self.actionlog_cmd, "del"))
        self.btnActionLogDelAll.clicked.connect(partial(self.actionlog_cmd, "del-all"))
        self.btnActionLogShow.clicked.connect(partial(self.actionlog_cmd, "win"))
        self.comboActionLogFilter.currentIndexChanged.connect(partial(self.actionlog_cmd, "filter"))
            # Taxon-Wizard
        self.taxSysAgeStatus.setIcon(QIcon(os.path.dirname(os.path.realpath(__file__)) + "/ressource/icon_frage.png"))
        self.taxSysAgeProgress.setMaximum(self.__model.subsystem_getMyWorms().puffer_getMaxAgeAllowed())
        self.taxSysAgeProgress.setValue(0)
        self.taxSysAgeProgress.setFormat("")
        self.taxSysBaseAdd.setIcon(QIcon(os.path.dirname(os.path.realpath(__file__)) + "/ressource/icon_add.png"))
        self.taxSysAaidAdd.setIcon(QIcon(os.path.dirname(os.path.realpath(__file__)) + "/ressource/icon_add.png"))
        self.taxSysPrivateAdd.setIcon(QIcon(os.path.dirname(os.path.realpath(__file__)) + "/ressource/icon_add.png"))
        self.taxSysColonyAdd.setIcon(QIcon(os.path.dirname(os.path.realpath(__file__)) + "/ressource/icon_add.png"))
        self.taxSysChangeAdd.setIcon(QIcon(os.path.dirname(os.path.realpath(__file__)) + "/ressource/icon_add.png"))
        self.taxSysNowormsAdd.setIcon(QIcon(os.path.dirname(os.path.realpath(__file__)) + "/ressource/icon_add.png"))
        self.taxSysBaseStatus.setIcon(QIcon(os.path.dirname(os.path.realpath(__file__)) + "/ressource/icon_info1.png"))
        self.taxSysAaidStatus.setIcon(QIcon(os.path.dirname(os.path.realpath(__file__)) + "/ressource/icon_info1.png"))
        self.taxSysPrivateStatus.setIcon(QIcon(os.path.dirname(os.path.realpath(__file__)) + "/ressource/icon_info1.png"))
        self.taxSysColonyStatus.setIcon(QIcon(os.path.dirname(os.path.realpath(__file__)) + "/ressource/icon_info1.png"))
        self.taxSysChangeStatus.setIcon(QIcon(os.path.dirname(os.path.realpath(__file__)) + "/ressource/icon_info1.png"))
        self.taxSysNowormsStatus.setIcon(QIcon(os.path.dirname(os.path.realpath(__file__)) + "/ressource/icon_info1.png"))
        self.mywormsSearchAid_go.setIcon(QIcon(os.path.dirname(os.path.realpath(__file__)) + "/ressource/icon_lupe.png"))
        self.mywormsSearchName_go.setIcon(QIcon(os.path.dirname(os.path.realpath(__file__)) + "/ressource/icon_lupe.png"))
        self.taxSysBaseAdd.setToolTip("Add WoRMS Base")
        self.taxSysAaidAdd.setToolTip("Add WoRMS Aaid")
        self.taxSysPrivateAdd.setToolTip("Add WoRMS Private")
        self.taxSysColonyAdd.setToolTip("Add WoRMS Colony")
        self.taxSysChangeAdd.setToolTip("Add WoRMS Change")
        self.taxSysNowormsAdd.setToolTip("Add No WoRMS")
        self.taxSysBaseProgress.setTextVisible(True)
        self.taxSysBaseProgress.setValue(0)
        self.taxSysBaseProgress.setMaximum(100)
        self.taxSysBaseProgress.setFormat("")
        font = self.taxSysBaseProgress.font()
        font.setPointSize(10)
        font.setBold(False)
        self.taxSysBaseProgress.setFont(font)
        self.taxSysBaseProgress.setAlignment(Qt.AlignRight)
        self.taxSysBaseProgress.setToolTip("Checked taxa")
        self.taxSysBaseAdd.clicked.connect(partial(self.admin_createLookupOrTaxon, "wormsbase", "New: " + self.__map.taxon_table2sheet("wormsbase"), taxHelper=True))
        self.taxSysAaidAdd.clicked.connect(partial(self.admin_createLookupOrTaxon, "wormsaaid", "New: " + self.__map.taxon_table2sheet("wormsaaid"), taxHelper=True))
        self.taxSysPrivateAdd.clicked.connect(partial(self.admin_createLookupOrTaxon, "wormsprivate", "New: " + self.__map.taxon_table2sheet("wormsprivate"), taxHelper=True))
        self.taxSysColonyAdd.clicked.connect(partial(self.admin_createLookupOrTaxon, "wormscolony", "New: " + self.__map.taxon_table2sheet("wormscolony"), taxHelper=True))
        self.taxSysChangeAdd.clicked.connect(partial(self.admin_createLookupOrTaxon, "wormschange", "New: " + self.__map.taxon_table2sheet("wormschange"), taxHelper=True))
        self.taxSysNowormsAdd.clicked.connect(partial(self.admin_createLookupOrTaxon, "noworms", "New: " + self.__map.taxon_table2sheet("noworms"), taxHelper=True))
        self.taxSysBaseEdit.clicked.connect(partial(self.taxon_cmd, "worms-base-edit"))
        btnMenu = QMenu()
        self.taxSysBaseEdit.setMenu(btnMenu)
        self.taxSysBaseEdit.setPopupMode(QToolButton.InstantPopup)
        action = QAction("all", self)
        actionGold = QAction("only gold", self)
        actionNotGold = QAction("only not-gold", self)
        action.triggered.connect(partial(self.taxon_cmd, "worms-base-edit"))
        actionGold.triggered.connect(partial(self.taxon_cmd, "worms-base-edit", gold=True))
        actionNotGold.triggered.connect(partial(self.taxon_cmd, "worms-base-edit", gold=False))
        btnMenu.addAction(action)
        btnMenu.addAction(actionGold)
        btnMenu.addAction(actionNotGold)
        self.taxSysTaxList.clicked.connect(partial(self.table_show, "taxon"))
        self.taxSysAaidEdit.clicked.connect(partial(self.taxon_cmd, "worms-aaid-edit"))
        self.taxSysPrivateEdit.clicked.connect(partial(self.taxon_cmd, "worms-private-edit"))
        self.taxSysColonyEdit.clicked.connect(partial(self.taxon_cmd, "worms-colony-edit"))
        self.taxSysChangeEdit.clicked.connect(partial(self.taxon_cmd, "change-edit"))
        self.taxSysNowormsEdit.clicked.connect(partial(self.taxon_cmd, "noworms-edit"))
        self.taxSysUpdate.clicked.connect(partial(self.taxon_cmd, "update"))
        self.taxSysUpdate.setIcon(QIcon(os.path.dirname(os.path.realpath(__file__)) + "/ressource/icon_go.png"))
        self.taxSysBaseStatus.clicked.connect(partial(self.ui_table_stat, "wormsbase"))
        self.taxSysAaidStatus.clicked.connect(partial(self.ui_table_stat, "wormsaaid"))
        self.taxSysPrivateStatus.clicked.connect(partial(self.ui_table_stat, "wormsprivate"))
        self.taxSysColonyStatus.clicked.connect(partial(self.ui_table_stat, "wormscolony"))
        self.taxSysChangeStatus.clicked.connect(partial(self.ui_table_stat, "wormschange"))
        self.taxSysNowormsStatus.clicked.connect(partial(self.ui_table_stat, "noworms"))
        self.taxSysAgeStatus.clicked.connect(partial(self.ui_table_stat, "taxlistupdate"))
            # MyWoRMS
        taxSysPufferResetMenu = QMenu()
        action = QAction("Clean whole MyWoRMS", self)
        action.triggered.connect(partial(self.taxon_cmd, "puffer-clean"))
        taxSysPufferResetMenu.addAction(action)
        action = QAction("Clean Taxa", self)
        action.triggered.connect(partial(self.taxon_cmd, "puffer-clean-taxa"))
        taxSysPufferResetMenu.addAction(action)
        action = QAction("Clean Name Queries", self)
        action.triggered.connect(partial(self.taxon_cmd, "puffer-clean-query"))
        taxSysPufferResetMenu.addAction(action)
        self.taxSysPufferReset.setMenu(taxSysPufferResetMenu)
        self.taxSysPufferReset.setPopupMode(QToolButton.InstantPopup)
        self.taxSysPufferReset.setIcon(QIcon(os.path.dirname(os.path.realpath(__file__)) + "/ressource/icon_trash.png"))
        self.taxSysPufferLoad.clicked.connect(partial(self.taxon_cmd, "puffer-reload"))
        self.taxSysPufferLoad.setIcon(QIcon(os.path.dirname(os.path.realpath(__file__)) + "/ressource/icon_reload.png"))
        self.taxSysPufferSave.clicked.connect(partial(self.taxon_cmd, "puffer-save"))
        self.taxSysPufferSave.setIcon(QIcon(os.path.dirname(os.path.realpath(__file__)) + "/ressource/icon_save.png"))
        self.taxSysPufferUpdate.clicked.connect(partial(self.taxon_cmd, "puffer-update"))
        self.taxSysPufferUpdate.setIcon(QIcon(os.path.dirname(os.path.realpath(__file__)) + "/ressource/icon_new.png"))
        self.taxSysPufferStatus.clicked.connect(partial(self.taxon_cmd, "puffer-status"))
        self.taxSysPufferStatus.setIcon(QIcon(os.path.dirname(os.path.realpath(__file__)) + "/ressource/icon_info1.png"))
        self.taxSysMisfitStatus.clicked.connect(partial(self.taxon_cmd, "misfits-status"))
        self.taxSysMisfitStatus.setIcon(QIcon(os.path.dirname(os.path.realpath(__file__)) + "/ressource/icon_info1.png"))
        self.taxSysQueryStatus.clicked.connect(partial(self.taxon_cmd, "query-status"))
        self.taxSysQueryStatus.setIcon(QIcon(os.path.dirname(os.path.realpath(__file__)) + "/ressource/icon_info1.png"))
        self.taxSysPufferSnapshot.clicked.connect(partial(self.taxon_cmd, "puffer-snapshot"))
        self.taxSysPufferSnapshot.setIcon(QIcon(os.path.dirname(os.path.realpath(__file__)) + "/ressource/icon_snapshot.png"))
        self.taxSysPufferImport.clicked.connect(partial(self.taxon_cmd, "puffer-import"))
        self.taxSysPufferImport.setIcon(QIcon(os.path.dirname(os.path.realpath(__file__)) + "/ressource/icon_import.png"))
        self.taxSysPufferExport.clicked.connect(partial(self.taxon_cmd, "puffer-export"))
        self.taxSysPufferExport.setIcon(QIcon(os.path.dirname(os.path.realpath(__file__)) + "/ressource/icon_export.png"))
        self.taxSysImport.setIcon(QIcon(os.path.dirname(os.path.realpath(__file__)) + "/ressource/icon_import.png"))
        self.taxSysImport.clicked.connect(self.import_taxon)
        self.taxSysExport.setIcon(QIcon(os.path.dirname(os.path.realpath(__file__)) + "/ressource/icon_export.png"))
        self.taxSysExport.clicked.connect(self.admin_export_taxonList)
        self.fao_mapping_import.setIcon(QIcon(os.path.dirname(os.path.realpath(__file__)) + "/ressource/icon_import.png"))
        self.fao_mapping_import.clicked.connect(partial(self.fao_cmd, "import"))
        self.fao_mapping_export.setIcon(QIcon(os.path.dirname(os.path.realpath(__file__)) + "/ressource/icon_export.png"))
        self.fao_mapping_export.clicked.connect(partial(self.fao_cmd, "export"))
        self.labelTaxSysGreaterMaxAge.setText("Age >" + str(self.__model.subsystem_getMyWorms().puffer_getMaxAgeAllowed()))
        self.mywormsSearchAid_go.clicked.connect(partial(self.taxon_cmd, "search-aid"))
        self.mywormsSearchName_go.clicked.connect(partial(self.taxon_cmd, "search-name"))
        def mywormsSearchName_fuzzy_changed():
            iState = self.mywormsSearchName_fuzzy.checkState()
            if iState == 0:
                sTxt = "Normal search active"
                self.mywormsSearchName_fuzzy.setToolTip(sTxt)
                self.ui_setStatusbarText(sTxt)
            elif iState == 1:
                sTxt = "Generic fuzzy search active"
                self.mywormsSearchName_fuzzy.setToolTip(sTxt)
                self.ui_setStatusbarText(sTxt)
            else:
                sTxt = "Special fuzzy search active"
                self.mywormsSearchName_fuzzy.setToolTip(sTxt)
                self.ui_setStatusbarText(sTxt)
        self.mywormsSearchName_fuzzy.stateChanged.connect(mywormsSearchName_fuzzy_changed)
        self.mywormsSearchName_fuzzy.setCheckState(1)
        # DataPatrol
        self.dataPatrol_biotaControl.setIcon(QIcon(os.path.dirname(os.path.realpath(__file__)) + "/ressource/icon_go.png"))
        self.dataPatrol_biotaControl.clicked.connect(partial(self.dataPatrol_cmd, "biota_control"))
        self.dataPatrol_biotaControl_lastCheckLog.setIcon(QIcon(os.path.dirname(os.path.realpath(__file__)) + "/ressource/icon_info1.png"))
        self.dataPatrol_biotaControl_lastCheckLog.clicked.connect(partial(self.dataPatrol_cmd, "biota_control_log"))
        # Regionen
        self.fao_regions.clicked.connect(partial(self.fao_cmd, "reg_show"))
        self.fao_regions_recreate.setIcon(QIcon(os.path.dirname(os.path.realpath(__file__)) + "/ressource/icon_world2.png"))
        self.fao_regions_recreate.clicked.connect(partial(self.fao_cmd, "reg_create"))
        self.fao_mapping_clean.setIcon(QIcon(os.path.dirname(os.path.realpath(__file__)) + "/ressource/icon_trash.png"))
        self.fao_mapping_clean.clicked.connect(partial(self.fao_cmd, "reg_clean"))
        self.fao_mapping_stat.setIcon(QIcon(os.path.dirname(os.path.realpath(__file__)) + "/ressource/icon_info1.png"))
        self.fao_mapping_stat.clicked.connect(partial(self.fao_cmd, "reg_stat"))
        self.fao_mapping_update.setIcon(QIcon(os.path.dirname(os.path.realpath(__file__)) + "/ressource/icon_go.png"))
        self.fao_mapping_update.clicked.connect(partial(self.fao_cmd, "reg_update"))
        # Rest
        self.buttonConsoleClear.clicked.connect(self.console_clear)
        self.buttonConnectionToggle.clicked.connect(self.ui_serverConnectionToggle)
        self.buttonSqlCmd.clicked.connect(self.core_sqlCmd_single)
        self.buttonConnectionDry.clicked.connect(self.guiConnectionDry)
        self.sysDryRunSwitch.clicked.connect(self.guiConnectionDry)
        self.comboVerbosity.currentIndexChanged.connect(self.console_changeVerbosity)
        self.buttonConsoleSave.clicked.connect(self.console_save)
        self.progressAbort.clicked.connect(self.worker_abort)
            # Admin Lock
        self.buttonAdminLock.clicked.connect(self.core_lock)
            # Tab
        self.tabMain.setCurrentIndex(2)
        self.tabMain.currentChanged.connect(self.ui_tabWasSwitched)

        # Admin Access Felder
        self.admin_db_host.textEdited.connect(self.guiAdminAccessDataEdited)
        self.admin_db_port.textEdited.connect(self.guiAdminAccessDataEdited)
        self.admin_db_user.textEdited.connect(self.guiAdminAccessDataEdited)
        self.admin_db_pw.textEdited.connect(self.guiAdminAccessDataEdited)
        self.admin_db_name.textEdited.connect(self.guiAdminAccessDataEdited)
        self.admin_db_schema.textEdited.connect(self.guiAdminAccessDataEdited)
            # Admin Access: File Auswähler
        self.button_admin_cmd_restore_schema_open_file.clicked.connect(self.ui_setRestoreSchemaFile)
        self.button_admin_cmd_backup_schema_open_file.clicked.connect(self.ui_setBackupSchemaFile)
            # Admin: Commandos
        self.button_admin_cmd_fetch.clicked.connect(self.core_fetchCredentials)
        self.button_admin_cmd_list_all_users.clicked.connect(self.core_listAllUsers)
        self.button_admin_cmd_list_all_users_marlin.clicked.connect(self.core_listAllUsersMarlin)
        self.button_admin_cmd_list_all_databases.clicked.connect(self.core_listAllDatabases)
        self.button_admin_cmd_sql_script.clicked.connect(self.core_sqlCmd_script)
        self.button_admin_cmd_list_schemas.clicked.connect(self.core_listAllSchemas)
        self.button_admin_cmd_rename_schema.clicked.connect(self.core_renameSchema)
        self.button_admin_cmd_backup_schema.clicked.connect(self.core_backupSchema)
        self.button_admin_cmd_restore_schema.clicked.connect(self.core_restoreSchema)
        self.button_admin_cmd_delete_schema.clicked.connect(self.core_deleteSchema)
        self.button_admin_cmd_new_schema.clicked.connect(self.core_CreateNewSchema)
        self.button_admin_cmd_start_psql.clicked.connect(self.core_startPsqlConsole)
        self.button_admin_version_db.clicked.connect(self.core_showPgServerVersion)
        self.button_admin_version_schema.clicked.connect(self.core_showSchemaVersion)
            # MyWoRMS
        # Baue DB-Tabellenstatus-Box
            # Zusatztabellen
        self.__lstTableStatusZusatz = ["ingest"]
            # Zu entfernende Tabellen
        self.__lstTableStatusBlacklist = []
            # Zeiger auf DB-Status-Box-Elemente
        self.__dicStatus_widgets_ingest = {}
        self.__dicStatus_widgets_lookup = {}
        self.__dicStatus_widgets_lookup_special = {}
        self.__dicStatus_widgets_lookup_status = {}
        self.__widgetStatus_widgets_lookup_dataset_status = None
        self.__dicStatus_widgets_lookup_gold = {}
        self.__dicStatus_widgets_lookup_count = {}
        self.__dicStatus_widgets_lookup_usedNotGoldLookup = {}
        self.__dicStatus_widgets_lookup_notUsedNotGoldLookup = {}
        self.__widgetStatus_widgets_lookup_dataset_allFreeRequest = None
        self.__widgetStatus_widgets_lookup_dataset_allUsedFreeRequest = None
        self.__widgetStatus_widgets_lookup_dataset_allUsedNotGoldFreeRequest = None
        self.__widgetStatus_widgets_lookup_dataset_DOI_OkNotOk = None
        self.__widgetStatus_widgets_lookup_dataset_DOI_UsedOkNotOk = None
        self.__widgetStatus_widgets_lookup_dataset_UsedNotOkReachable = None
        self.__widgetStatus_widgets_lookup_dataset_checkDois = None

            # Aufbauen
        self.build_admin_status_tabellen()

        # Erweitere Viz-Input um MultiCombo Felder
        self.vizInDataset = QMultiCombo(self.explorer_update)
        self.vizInput.layout().replaceWidget(self.vizPlaceholderDataset, self.vizInDataset)
        self.vizInRegion = QMultiCombo(self.explorer_update)
        self.vizInput.layout().replaceWidget(self.vizPlaceholderRegion, self.vizInRegion)
        self.vizInType = QMultiCombo(self.explorer_update)
        self.vizInput.layout().replaceWidget(self.vizPlaceholderType, self.vizInType)
        self.vizInGear = QMultiCombo(self.explorer_update)
        self.vizInput.layout().replaceWidget(self.vizPlaceholderGear, self.vizInGear)
        self.vizInTaxcover = QMultiCombo(self.explorer_update)
        self.vizInput.layout().replaceWidget(self.vizPlaceholderTaxcover, self.vizInTaxcover)
        self.vizInIngest = QMultiCombo(self.explorer_update)
        self.vizInput.layout().replaceWidget(self.vizPlaceholderIngest, self.vizInIngest)
        self.explorer_selectorEnable(False)

        # Millometer
        self.milliometer.setStyleSheet(CssPresets.progressBar_grey)
        # Splitter (Console bekommt bei Fenster Resize allen Platz)
        self.splitterConsole.setStretchFactor(1, 1)
        self.splitterConsole.setStretchFactor(2, 0)

        # Browser
        self.build_browser()
        self.build_browser_emptyMap()
        self.explorer_browser_reloadMap()
        self.browserBox.addWidget(self.__browser)
        self.buttonBrowserBack.clicked.connect(self.__browser.back)
        self.buttonBrowserForward.clicked.connect(self.__browser.forward)
        self.buttonBrowserReload.clicked.connect(self.__browser.reload)

        # Groesse auf Minimum
        self.resize(self.minimumSizeHint().width(), self.sizeHint().height())

        # Etc
            # Refresh View-Widget
        self.status_refresh_all()
        self.tableIngest.horizontalHeader().show()
        self.tableLog.horizontalHeader().show()

    # ---- Private Methoden

    def popup_generic(self, **kwargs):
        """
        Generische Erzeugung eines Dialogfensters

        Kwargs:
            titel (str): Titel
            head (str): Haupttext
            info (str): Zusatztext
            detail (str): Detailtext
            button_special (str): Text
            button_yes (str): Text
            button_no (str): Text
            icon="error" (str): "info","warn" und "error" erzeugen das entsprechende Icon
            button="ok" (str): "ok","close" erzeugen den entsprechenen Abschlussbutton
            timer (int): Zeit in Sekunden nachdem sich das Fenster schliesst
        """

        class QMessageBoxTweak(QMessageBox):
            def __init__(self, **kwargs):
                super().__init__()

                # Member
                sShowSpecial = kwargs["button_special"] if "button_special" in kwargs else None
                sYesButton = kwargs.get('button_yes')
                sNoButton = kwargs.get('button_no')
                if sYesButton is None:
                    sYesButton = "OK"

                self.__dicReturnCode = {}
                iReturnCode_buttonNr = 0

                # Button: No
                if sNoButton is not None:
                    self.__dicReturnCode[iReturnCode_buttonNr] = 0
                    iReturnCode_buttonNr += 1
                    self.__btnNo = QPushButton()
                    self.__btnNo.setText(sNoButton)
                    self.__btnNo.clicked.connect(lambda: QMessageBoxTweak.done(self, 0))
                    self.addButton(self.__btnNo, QMessageBoxTweak.NoRole)

                # Button: Special
                if sShowSpecial is not None:
                    self.__dicReturnCode[iReturnCode_buttonNr] = 2
                    iReturnCode_buttonNr += 1
                    self.__btnShowSpecial = QPushButton()
                    self.__btnShowSpecial.setText(str(sShowSpecial))
                    self.__btnShowSpecial.clicked.connect(lambda: QMessageBoxTweak.done(self, 2))
                    self.addButton(self.__btnShowSpecial, QMessageBoxTweak.NoRole)

                # Button: Yes
                self.__dicReturnCode[iReturnCode_buttonNr] = 1
                iReturnCode_buttonNr += 1
                self.__btnYes = QPushButton()
                self.__btnYes.setText(sYesButton)
                self.__btnYes.clicked.connect(lambda: QMessageBoxTweak.done(self, 1))
                self.addButton(self.__btnYes, QMessageBoxTweak.NoRole)

            def getRetCodeMapped(self, iCode):
                return self.__dicReturnCode.get(iCode)

            def setNoButtonText(self, sText):
                self.__btnNo.setText(sText)

            def setYesButtonText(self, sText):
                self.__btnYes.setText(sText)

        class QTimerMessageBox(QMessageBoxTweak):

            def __init__(self, iTimeout=1):
                super().__init__()

                # Member
                self.timer = QTimer(self)
                self.iTimeout = iTimeout
                # Init
                # Icon
                #self.setIcon(QTimerMessageBox.Warning)

                self.timer.setInterval(1000)
                self.setYesButtonText("Close (in " + str(self.iTimeout) + " seconds)")
                self.timer.timeout.connect(self.timerUpdate)
                self.timer.start()

            def timerUpdate(self):
                self.iTimeout -= 1
                self.setYesButtonText("Close (in " + str(self.iTimeout) + " seconds)")
                if self.iTimeout <= 0:
                    self.timer.stop()
                    self.close()

            def closeEvent(self, event):
                self.timer.stop()
                event.accept()

        sTitel = kwargs.get('title')
        sHead = kwargs.get('head')
        sInfo = kwargs.get('info')
        sDetail = kwargs.get('detail')
        sIcon = kwargs.get('icon')
        iTimer = kwargs.get('timer')
        sYesButton = kwargs.get('button_yes')
        sNoButton = kwargs.get('button_no')
        sSpecialButton = kwargs.get("button_special")

        # Dialog zusammenbauen
        if iTimer is not None:
            msg = QTimerMessageBox(iTimer)
        else:
            msg = QMessageBoxTweak(button_special=sSpecialButton, button_yes=sYesButton, button_no=sNoButton)

        # WindowsFlags
        msg.setWindowFlags(Qt.Dialog | Qt.CustomizeWindowHint | Qt.WindowTitleHint | Qt.WindowCloseButtonHint)

        # Icon
        if sIcon == "info":
            icon = QMessageBox.Information
        elif sIcon == "warn":
            icon = QMessageBox.Warning
        elif sIcon == "error":
            icon = QMessageBox.Critical
        else:
            icon = QMessageBox.NoIcon
        msg.setIcon(icon)

        # Text: Font
        font = msg.font()
        font.setPointSize(10)
        font.setFamily("Arial, Helvetica, sans-serif")
        msg.setFont(font)

        # Text: Format
        # msg.setTextFormat(Qt.PlainText)

        # Style
        msg.setStyleSheet("QPushButton{ font-size: 10pt; }");

        # Titel
        if sTitel is not None:
            msg.setWindowTitle(sTitel)
        # Head
        if sHead is not None:
            sHead = '<span style="font-size:12pt;">' + sHead + '</span>'
            msg.setText(sHead)
        # Info
        if sInfo is not None:
            sInfo = '<span style="font-size:10pt;">' + sInfo + '</span>'
            msg.setInformativeText(sInfo)
        # Details
        if sDetail is not None:
            msg.setDetailedText(sDetail)

        self.ui_centerWidget(msg, self)
        iCode = msg.exec_()
        iCode_mapped = msg.getRetCodeMapped(iCode)

        return iCode_mapped

    def table_showGeneric(self, multisheet, **kwargs):
        """
        Erzeuge ein generisches Dialogfenster mit dem Inhalt einer Tabelle

        Args:
            multisheet (multisheet): Multisheet Datenobjekt
        Kwargs:
            diaTitle (str): Dialogfenster-Titel
            ingId (int): IngestionID - Tabelle ist Teil einer Ingestion
            select (list): Nur diese Attribute sollen benutzt werden (muss weitergericht werden, falls die Tabelle erneut geöffnet werden soll)
            del_entry (bool); Zeilen könenn gelöscht werden
            del_entry (bool); Zeilen könenn gelöscht werden
            edit (bool): Zellen können editiert werden
            save (bool): Tabelle kann als csv-Datei gespeichert werden
            editId (int): Id-Feldname für Edit-Modus
            range (int): Anzeigefenster (Zeilen pro Tabelle)
            rangeIdx (): Anzeigefenster (Zeilen pro Tabelle)
            id (id): Datensatz Id auf die fokussiert werden soll
            gold (bool): Es wurden nur die gold/not-gold Einträge abgefragt (Lookup-Special)
            used (bool): Es wurden nur die used/not-used Einträge abgefragt (Lookup-Special)
            gold_edit (bool): Ermögliche den Gold Status massenweise zu verändern
            used (bool): Lookup benutzt
            doi (bool): DOI ok
            doi_pangaea (bool): DOI führt zu PANGAEA
            access (bool): Zugang gewährleistet (irgendwie)
            gear_double (bool): Gear Double Config
        """

        pn = self.ui_printConsoleNormal
        pv = self.ui_printConsoleVerbose
        pd = self.ui_printConsoleDebug
        workflow = self.ui_setWorkflow

        self.ui_showProgress()
        workflow(top="Building table", sub="assembling rows")

        # Hilfsklassen
        # Windowtitel setzen (normal/selection mode)
        def setDialogTitle():
            sSelectAppendix = ""
            if len(setSelectedIds) > 0:
                sSelectAppendix = " [selected rows: " + str(len(setSelectedIds)) + "/" + str(len(lstContentFull)) + "]"
            sSpecial = ""
            dia.setWindowTitle(str(sDiaTitle) + str(sSpecial) + str(sSelectAppendix))
            if (bUsed, bGold, bAccess, bDoi, bDoiPangaea, bGear_double) == (None, True, None, None, None, None):
                sSpecial = " [only gold]"
            elif (bUsed, bGold, bAccess, bDoi, bDoiPangaea, bGear_double) == (None, False, None, None, None, None):
                sSpecial = " [only not-gold]"
            elif (bUsed, bGold, bAccess, bDoi, bDoiPangaea, bGear_double) == (True, None, None, None, None, None):
                sSpecial = " [only used]"
            elif (bUsed, bGold, bAccess, bDoi, bDoiPangaea, bGear_double) == (False, None, None, None, None, None):
                sSpecial = " [only not-used -> purge hard?]"
            elif (bUsed, bGold, bAccess, bDoi, bDoiPangaea, bGear_double) == (False, False, None, None, None, None):
                sSpecial = " [only not used and not gold -> purge soft?]"
            elif (bUsed, bGold, bAccess, bDoi, bDoiPangaea, bGear_double) == (True, False, None, None, None, None):
                sSpecial = " [only used and not gold -> check for gold!]"
            elif (bUsed, bGold, bAccess, bDoi, bDoiPangaea, bGear_double) == (None, None, True, None, None, None):
                sSpecial = " [only access somehow]"
            elif (bUsed, bGold, bAccess, bDoi, bDoiPangaea, bGear_double) == (None, None, False, None, None, None):
                sSpecial = " [only no access at all]"
            elif (bUsed, bGold, bAccess, bDoi, bDoiPangaea, bGear_double) == (True, None, True, None, None, None):
                sSpecial = " [only used and access somehow]"
            elif (bUsed, bGold, bAccess, bDoi, bDoiPangaea, bGear_double) == (True, None, False, None, None, None):
                sSpecial = " [only used and no access at all]"
            elif (bUsed, bGold, bAccess, bDoi, bDoiPangaea, bGear_double) == (True, False, True, None, None, None):
                sSpecial = " [only used and access somehow and not-gold (check for gold!)]"
            elif (bUsed, bGold, bAccess, bDoi, bDoiPangaea, bGear_double) == (None, None, None, None, None, True):
                sSpecial = " [only double configurations]"
            elif (bUsed, bGold, bAccess, bDoi, bDoiPangaea, bGear_double) == (None, None, None, True, None, None):
                sSpecial = " [only DOI OK]"
            elif (bUsed, bGold, bAccess, bDoi, bDoiPangaea, bGear_double) == (None, None, None, False, None, None):
                sSpecial = " [only DOI not OK]"
            elif (bUsed, bGold, bAccess, bDoi, bDoiPangaea, bGear_double) == (True, None, None, True, None, None):
                sSpecial = " [only used and DOI OK]"
            elif (bUsed, bGold, bAccess, bDoi, bDoiPangaea, bGear_double) == (True, None, None, False, None, None):
                sSpecial = " [only used an DOI not OK]"
            elif (bUsed, bGold, bAccess, bDoi, bDoiPangaea, bGear_double) == (True, None, True, False, None, None):
                sSpecial = " [only used and not-ok and access somehow (bad DOI online)]"
            elif (bUsed, bGold, bAccess, bDoi, bDoiPangaea, bGear_double) == (None, None, None, True, False, None):
                sSpecial = " [only DOI not PANGAEA]"

        # Selection Buttons aktivieren/deaktivieren
        def setSelectionButtons():
            try:
                if len(setSelectedIds) == 0:
                    if bSave:
                        btnSaveSel.setEnabled(False)
                    if bEdit:
                        btnDelSel.setEnabled(False)
                        btnDelAll.setEnabled(False)
                    if bGoldEdit:
                        btnGoldSet.setEnabled(False)
                        btnGoldUnset.setEnabled(False)
                else:
                    if bSave:
                        btnSaveSel.setEnabled(True)
                    if bDel:
                        btnDelSel.setEnabled(True)
                        btnDelAll.setEnabled(True)
                    if bGoldEdit:
                        btnGoldSet.setEnabled(True)
                        btnGoldUnset.setEnabled(True)
            except Exception as ex:
                pass

        # Zelle angeklickt
        def cellClicked(iRowSel, iColSel):
            # Ermittle Zeile in gesamten Datensatz
            iRow = iRangeIdx * iRange + iRowSel # Row in sichtbarer Tabelle
            if sEditId in lstHeader:
                iIdxId = lstHeader.index(sEditId)
            else:
                pn("Internal error while selection of entires in Table: Column " + sEditId + " not found in Table " + sTabName)
                return
            try:
                #iId = lstContent[iRow][iIdxId]  # Id des ausgewählten Datensatzes aus dem Datnfeld holen (dann funktioniert die Sortierung nicht mehr)
                iId = int(tableWidget.item(iRowSel, iIdxId).text())  # Id des ausgewählten Datensatzes aus dem Fenster holen
            except:
                pn("Internal error while selection of entires in Table: Column " + sEditId + " contains no interger")
                return
            # Ermittle Name des Fokusfeldes
            sFokusName = None
            try:
                sFokusName = lstHeader[iColSel]
            except:
                pass
            pd("row-table:"+str(iRowSel)+"/col-table:"+str(iColSel)+ " -> row-total:"+str(iRow)+" -> id-entry:"+str(iId)+", fokus-name:"+str(sFokusName))
            self.table_entryEdit(sTabName, iId, closeWin=dia, fokusCol=sFokusName, gold=bGold, gold_edit=bGoldEdit)

        # Helper Range wurde verändert
        # Bestimme ausgewählte Zeilen
        def rangeChanged():
            lstRanges = tableWidget.selectedRanges()
            setSelectedIds.clear()
            for r in lstRanges:
                iRowStart = r.topRow()
                iRowEnde = r.bottomRow()
                for i in range(iRowStart, iRowEnde + 1):
                    try:
                        iId = int(tableWidget.item(i, 0).text())  # Hole die Ingest Id aus der View-Tabelle
                        setSelectedIds.add(iId)
                    except Exception as ex:
                        pass
            setDialogTitle()
            setSelectionButtons()

        # Kwargs
        bSelect = False
        setSelectedIds = set() # Liste der ausgewählten Zeilen (Ids)
        bSave = False
        bDel = False
        bEdit = False
        iIngId = None
        sTabName = None
        sEditId = "id"
        iRange = 500
        iRangeIdx = 0
        iId = None
        dicIngStat = None
        bGold = None
        bUsed = None
        bGoldEdit = False
        bAccess = None
        bDoi = None
        bDoiPangaea = None
        bGear_double = None
        lstSelect = None

        sDiaTitle = "Table"
        for arg in kwargs:
            if "ingId" in kwargs:  # Ingestion Id
                iIngId = kwargs['ingId']
            if "ingStat" in kwargs:  # Ingestion Statistik über Tabellengrößen
                dicIngStat = kwargs['ingStat']
            if "edit" in kwargs:  # Edit-Modus
                bEdit = kwargs['edit']
            if "del_entry" in kwargs:  # Edit-Modus
                bDel = kwargs['del_entry']
            if "save" in kwargs:  # Tabelle kann als csv-Datei gespeichert werden
                bSave = kwargs['save']
            if "editId" in kwargs:  # Id-Feldname für Edit-Modus
                sEditId = kwargs['editId']
            if "range" in kwargs:  # Anzeigefenster (Zeilen pro Tabelle)
                iRange = kwargs['range']
            if "rangeIdx" in kwargs:  # Index des anzuzeigenden Fensters
                iRangeIdx = kwargs['rangeIdx']
            if "id" in kwargs:  # Datensatz Id des Anzuzeigenden Eintrages
                iId = kwargs['id']
            if "gold" in kwargs:  # (Not-)Gold Modus
                bGold = kwargs['gold']
            if "used" in kwargs:  # (Not-)Used Modus
                bUsed = kwargs['used']
            if "gold_edit" in kwargs:  # Gold Bulk Edit
                bGoldEdit = kwargs['gold_edit']
            if "diaTitle" in kwargs:  # Datensatz Id des Anzuzeigenden Eintrages
                sDiaTitle = kwargs['diaTitle']
            if "gold" in kwargs:  # (Not-)Gold Modus
                bGold = kwargs['gold']
            if "doi" in kwargs:
                bDoi = kwargs["doi"]
            if "doi_pangaea" in kwargs:
                bDoiPangaea = kwargs["doi_pangaea"]
            if "access" in kwargs:
                bAccess = kwargs["access"]
            if "gear_double" in kwargs:
                bGear_double = kwargs["gear_double"]

        # Hole aktuelle Tabelle aus Multisheet
        lstHeader, lstContentFull = multisheet.getSheet()

        # Header: ingest_id -> ING-ID (Achtung: Wird im Objekt verändert)
        try:
            lstHeader[lstHeader.index("ingest_id")] = "<IngestId>"
        except:
            pass

        # Hole Multi Sheet Statistik
        dicIngStat = multisheet.getStat()

        # Hole aktuellen Tabellennamen
        sTabName = multisheet.getOpenSheetName()

        # Bestimme den Tabellen-Index der Id für den Editierungsmodus
        iIdxEditId = None
        if sEditId in lstHeader:
            iIdxEditId = lstHeader.index(sEditId)
        else:
            bEdit = False

        # Falls eine explizite Datensatz ID uebergeben wurde -> Stelle die entsprechende Range-Seite ein
        if iId is not None:
            for row, elem in enumerate(lstContentFull):  # Suche Zeilennummer der ID in Daten
                if elem[iIdxEditId] == iId:
                    iRangeIdx = int(row / iRange)
                    break

        # Hole alle gespeicherten Ids
        lstIdsAll = []
        if iIdxEditId is not None:
            for elem in lstContentFull:
                lstIdsAll.append(elem[iIdxEditId])

        # Fenster Index überprüfen
        if iRangeIdx < 0:
            iRangeIdx = 0
        if (iRangeIdx * iRange) > len(lstContentFull):
            iRangeIdx = int(len(lstContentFull) / iRange)

        # Elemente in Inhaltsfenster auswählen
        iRangeElementStartIdx = iRangeIdx * iRange
        iRangeElementEndIdx = ((iRangeIdx + 1) * iRange) - 1
        lstContent = lstContentFull[
                     iRangeElementStartIdx:iRangeElementEndIdx + 1]  # Achtung das letzte Element wird von Python nie genommen, also +1
        iRangeIdxMax = int(len(lstContentFull) / iRange)

        pd("* Show content of a table")
        pd("Title: " + str(sDiaTitle))
        pd("All Ids: "+ str(len(lstIdsAll)))
        pd("IngId: " + str(iIngId))
        pd("IngStat:" + str(dicIngStat))
        pd("TabName: " + str(sTabName))
        pd("Edit: " + str(bEdit))
        pd("Delete: " + str(bDel))
        pd("Save: " + str(bSave))
        pd("Table edit id: " + sEditId)
        pd("Window width: " + str(iRange))
        pd("Window idx: " + str(iRangeIdx))
        pd("Window idx max: " + str(iRangeIdxMax))
        pd("Elements Idx:" + str(iRangeElementStartIdx) + "-" + str(iRangeElementEndIdx))
        pd("Id of entry: " + str(iId))
        pd("Multi Sheet (sheets): " + str(multisheet.getSheetNames()))
        pd("Multi Sheet (open sheet name): " + str(multisheet.getOpenSheetName()))
        pd("Multi Sheet (open sheet idx): " + str(multisheet.getOpenSheetIdx()))
        pd("Gold edit mode: " +str(bGoldEdit))
        pd("Filter: use: " +str(bUsed))
        pd("Filter: gold: " +str(bGold))
        pd("Filter: access: " +str(bAccess))
        pd("Filter: DOI: " +str(bDoi))
        pd("Filter: DOI PANGAEA: " +str(bDoiPangaea))
        pd("Filter: Gear double config: " +str(bGear_double))
        pd("Select: " +PrettyText.lst2strC(lstSelect))

        # Zeilen/Spalten
        iRows = len(lstContent)
        iCols = len(lstHeader)

        # Dialog erzeugen
        dia = QDialog()
        setDialogTitle()
        dia.setWindowModality(Qt.ApplicationModal)
        dia.setWindowFlags(Qt.Dialog | Qt.CustomizeWindowHint | Qt.WindowTitleHint | Qt.WindowCloseButtonHint)

        vbox = QVBoxLayout(dia)

        # Tabelle
        tableWidget = QTableWidget(0, iCols)
        tableWidget.setSortingEnabled(False)

        # Table anklickbar machen
        # Einfacher Klick -> Markiert ganze Row
        tableWidget.setSelectionBehavior(QTableView.SelectRows)
        try:
            tableWidget.itemSelectionChanged.disconnect()
        except:
            pass
        # Auswahl veraendert
        tableWidget.itemSelectionChanged.connect(rangeChanged)

        if "id" in lstHeader:
            tableWidget.setColumnHidden(lstHeader.index("id"),True) # Id Feld ausblenden
        if bEdit:
            tableWidget.cellDoubleClicked.connect(cellClicked)

        # Header
        tableWidget.setHorizontalHeaderLabels(lstHeader)
        tableWidget.horizontalHeader().setStretchLastSection(True)

        # Fonts
        # Tabelle
        font = tableWidget.font()
        font.setPointSize(10)
        tableWidget.setFont(font)
        # Header
        font = tableWidget.horizontalHeader().font()
        font.setBold(True)
        tableWidget.horizontalHeader().setFont(font)
        tableWidget.horizontalHeader().setDefaultAlignment(Qt.AlignLeft)

        # Progress
        self.ui_setProgressbarTextAnzeigen(True)
        self.ui_setProgressbarMinMax(1, 100)
        self.ui_giveQtTimeToUpdate()

        # Zeilen
        for iZeilennummer, row in enumerate(lstContent):
            # Progress
            iProzent = self.tools_getProgress(iZeilennummer, iRows, 20)
            if iProzent is not None:
                self.ui_setProgressbarWert(iProzent)
                self.ui_setProgressbarText(str(iProzent) + "%")
                self.ui_giveQtTimeToUpdate()
            rowPosition = tableWidget.rowCount()
            tableWidget.insertRow(rowPosition)

            # Spalten (eine Ingestion)
            for iSpaltennummer, col in enumerate(row):
                item = QTableWidgetItem()
                item.setFlags(Qt.ItemIsEnabled | Qt.ItemIsSelectable)
                sCol = str(col) if col is not None else ""
                # Inhalt differenzieren: a) Zahl:Float/Int b) String -> Wegen Sortierung
                bNumber = False
                try:
                    iColInhalt = int(col)
                    bNumber = True
                except:
                    pass
                try:
                    fColInhalt = float(col)
                    bNumber = True
                except:
                    pass
                if bNumber:
                    if col not in [None]:
                        item.setData(Qt.DisplayRole, sCol)  # Integer
                else:
                    if col not in [None]:
                        item.setText(str(col))  # String
                # Tabelle ist Ingest
                # Ingest ID weicht von Vorgabe ab -> IngestId -> Rot
                if lstHeader[iSpaltennummer] == "<IngestId>" and int(col) != iIngId:
                    item.setForeground(QColor(255, 0, 0))
                tableWidget.setItem(rowPosition, iSpaltennummer, item)

        # Zeilen Nummerieren
        lst = list(str(x) for x in range(iRangeElementStartIdx + 1, iRangeElementEndIdx + 2))
        tableWidget.setVerticalHeaderLabels(lst)

        # Setze den Tabellenfokus auf definierte Zeile (zuletzt bearbeitet, definiert anhand der id)
        if iId is not None:
            try:
                lstW = tableWidget.findItems(str(iId),Qt.MatchExactly)  # Durchsuche gesamte Tabelle nach gegebener ID
                w = None
                for w in lstW:
                    if w.column() == iIdxEditId:  # Wurde Treffer in ID-Spalte gefunden -> Super
                        break
                # print("Fokus:", "Id=",iId,"Page=",iRangeIdx,"Col=", w.column(), "Row=", w.row(), "Inhalt=", w.text())
                if w is not None:
                    tableWidget.scrollToItem(w, QAbstractItemView.EnsureVisible)
                    tableWidget.selectRow(w.row())
            except Exception as ex:
                print("Unexpected Error in Table Show: " + str(ex))

        # Zeilen und Spaltenbreite an Textinhalt/-größe anpassen
        tableWidget.resizeColumnsToContents()
        tableWidget.resizeRowsToContents()

        # Auswahl Range
        # Widget
        rangeWidget = QWidget()
        hbox = QHBoxLayout()
        hbox.setSizeConstraint(QLayout.SetFixedSize)
        hbox.setContentsMargins(0, 0, 0, 0)
        hbox.setSpacing(1)
        # Button Zurueck
        btnZurueck = QToolButton()
        btnZurueck = QToolButton()
        btnZurueck.setIcon(QIcon(os.path.dirname(os.path.realpath(__file__)) + "/ressource/icon_zurueck.png"))
        btnZurueck.setAutoRaise(True)
        btnZurueck.setToolButtonStyle(Qt.ToolButtonIconOnly)
        font = btnZurueck.font()
        font.setPointSize(10)
        btnZurueck.setFont(font)
        btnZurueck.setToolTip("Window back")
        # Button Anfang
        btnAnfang = QToolButton()
        btnAnfang = QToolButton()
        btnAnfang.setIcon(QIcon(os.path.dirname(os.path.realpath(__file__)) + "/ressource/icon_anfang.png"))
        btnAnfang.setAutoRaise(True)
        btnAnfang.setToolButtonStyle(Qt.ToolButtonIconOnly)
        font = btnAnfang.font()
        font.setPointSize(10)
        btnAnfang.setFont(font)
        btnAnfang.setToolTip("Jump to first window")
        # Button Vor
        btnVor = QToolButton()
        btnVor.setIcon(QIcon(os.path.dirname(os.path.realpath(__file__)) + "/ressource/icon_vor.png"))
        btnVor.setAutoRaise(True)
        btnVor.setToolButtonStyle(Qt.ToolButtonIconOnly)
        ont = btnVor.font()
        font.setPointSize(10)
        btnVor.setFont(font)
        btnVor.setToolTip("Window forward")
        # Button Ende
        btnEnde = QToolButton()
        btnEnde.setIcon(QIcon(os.path.dirname(os.path.realpath(__file__)) + "/ressource/icon_ende.png"))
        btnEnde.setAutoRaise(True)
        btnEnde.setToolButtonStyle(Qt.ToolButtonIconOnly)
        font = btnEnde.font()
        font.setPointSize(10)
        btnEnde.setFont(font)
        btnEnde.setToolTip("Jump to last window")
        # Button Go
        btnGo = QToolButton()
        #btnGo.setText("go")
        btnGo.setIcon(QIcon(os.path.dirname(os.path.realpath(__file__)) + "/ressource/icon_jump.png"))
        btnGo.setAutoRaise(True)
        font = btnGo.font()
        font.setPointSize(10)
        btnGo.setFont(font)
        btnGo.setToolTip("Jump to window")
        btnGo.setToolButtonStyle(Qt.ToolButtonIconOnly)
        # Range
        iBreiteS = 50
        iBreiteM = 100
        # Fenster Idx
        ausgabeFeldRangeFensterIdx = QLineEdit(str(iRangeIdx + 1))
        ausgabeFeldRangeFensterIdx.setAlignment(Qt.AlignRight)
        ausgabeFeldRangeFensterIdx.setFixedWidth(iBreiteS)
        # Fenster Idx Max
        ausgabeFeldRangeFensterIdxMax = QLineEdit(str(iRangeIdxMax + 1))
        ausgabeFeldRangeFensterIdxMax.setAlignment(Qt.AlignRight)
        ausgabeFeldRangeFensterIdxMax.setFixedWidth(iBreiteS)
        ausgabeFeldRangeFensterIdxMax.setEnabled(True)
        # ausgabeFeldRangeFensterIdxMax.setFrame(False)
        # Start
        ausgabeFeldRangeElementStart = QLineEdit(str(iRangeElementStartIdx + 1))
        ausgabeFeldRangeElementStart.setAlignment(Qt.AlignRight)
        ausgabeFeldRangeElementStart.setFixedWidth(iBreiteM)
        ausgabeFeldRangeElementStart.setEnabled(True)
        # ausgabeFeldRangeElementStart.setFrame(False)
        # Ende
        iLetztesElement = iRangeElementEndIdx + 1 if iRangeElementEndIdx + 1 <= len(lstContentFull) else len(
            lstContentFull)
        ausgabeFeldRangeElementEnde = QLineEdit(str(iLetztesElement))
        ausgabeFeldRangeElementEnde.setAlignment(Qt.AlignRight)
        ausgabeFeldRangeElementEnde.setFixedWidth(iBreiteM)
        ausgabeFeldRangeElementEnde.setEnabled(True)
        # ausgabeFeldRangeElementEnde.setFrame(False)
        # Gesamt
        ausgabeFeldRangeElementGesamt = QLineEdit(str(len(lstContentFull)))
        ausgabeFeldRangeElementGesamt.setAlignment(Qt.AlignRight)
        ausgabeFeldRangeElementGesamt.setFixedWidth(iBreiteM)
        ausgabeFeldRangeElementGesamt.setEnabled(True)
        # ausgabeFeldRangeElementGesamt.setFrame(False)
        # Zusammenbauen
        hbox.addWidget(QLabel("Window:"))
        hbox.addWidget(btnAnfang)
        hbox.addWidget(btnZurueck)
        hbox.addWidget(btnGo)
        hbox.addWidget(ausgabeFeldRangeFensterIdx)
        hbox.addWidget(QLabel("|"))
        hbox.addWidget(ausgabeFeldRangeFensterIdxMax)
        hbox.addWidget(btnVor)
        hbox.addWidget(btnEnde)
        rangeWidget.setLayout(hbox)

        # Tabelle gehört zu Ingest
        # Widget
        ingestWidget = QWidget()
        # Grids (LookUp/Ingest)
        gridIng = QGridLayout()
        gridIng.setSizeConstraint(QLayout.SetFixedSize)
        gridIng.setSpacing(4)
        gridIng.setContentsMargins(4, 4, 4, 4)
        if multisheet.getNoOfSheets() > 1:
            for iNr, sTabNameTmp in enumerate(multisheet.getSheetNames()):
                sTextLabel = sTabNameTmp.replace("_", " ")
                sTextCount = "("+str(dicIngStat[sTabNameTmp])+"x)"
                if sTabNameTmp == sTabName:
                    # Label
                    label = QLabel(sTextLabel)
                    font = QFont()
                    font.setBold(True)
                    font.setPointSize(10)
                    label.setFont(font)
                    gridIng.addWidget(label, 0, iNr, Qt.AlignCenter)
                    # Count
                    label = QLabel(sTextCount)
                    font = QFont()
                    font.setBold(False)
                    font.setPointSize(10)
                    label.setFont(font)
                    gridIng.addWidget(label, 1, iNr, Qt.AlignCenter)
                else:
                    # Button
                    btnTab = QToolButton()
                    btnTab.setText(sTextLabel)
                    font = btnTab.font()
                    font.setPointSize(10)
                    font.setBold(False)
                    btnTab.setFont(font)
                    btnTab.setToolTip("Show table " + sTabNameTmp)
                    btnTab.setAutoRaise(True)
                    btnTab.setToolButtonStyle(Qt.ToolButtonTextOnly)
                    btnTab.setFixedWidth(btnTab.minimumSizeHint().width())
                    btnTab.setFixedHeight(btnTab.minimumSizeHint().height())
                    #btnTab.clicked.connect((lambda t: lambda: QDialog.done(dia, t))(iTabCode)) # Lambda braucht fest gebundene Variable, sonst zeigt es auf immer die selbe Variable im Geltungsbereich
                    btnTab.clicked.connect(partial(QDialog.done, dia, 20+iNr)) # Partial ist hier besser
                    if dicIngStat[sTabNameTmp] == 0:
                        btnTab.setEnabled(False)
                    gridIng.addWidget(btnTab, 0, iNr, Qt.AlignCenter)
                    # Count
                    label = QLabel(sTextCount)
                    font = QFont()
                    font.setBold(False)
                    font.setPointSize(10)
                    label.setFont(font)
                    if dicIngStat[sTabNameTmp] == 0:
                        label.setEnabled(False)
                    gridIng.addWidget(label, 1, iNr, Qt.AlignCenter)
            ingestWidget.setLayout(gridIng)

        # Datensaetze
        # Widget
        datensaetzeWidget = QWidget()
        hbox = QHBoxLayout()
        hbox.setSizeConstraint(QLayout.SetFixedSize)
        hbox.setContentsMargins(0, 0, 0, 0)
        hbox.setSpacing(1)
        # Zusammenbauen
        hbox.addWidget(QLabel("Rows:"))
        hbox.addWidget(ausgabeFeldRangeElementStart)
        hbox.addWidget(QLabel("-"))
        hbox.addWidget(ausgabeFeldRangeElementEnde)
        hbox.addWidget(QLabel("/"))
        hbox.addWidget(ausgabeFeldRangeElementGesamt)
        datensaetzeWidget.setLayout(hbox)

        # Dialog Buttonbox
        buttonBox = QDialogButtonBox(Qt.Horizontal, dia)
        buttonBox.setCenterButtons(True)
        if bDel:
            # Delete All
            btnDelAll = QPushButton()
            btnDelAll.setText("Delete All")
            btnDelAll.setToolTip("Select and delete all entries")
            btnDelAll.clicked.connect(lambda: QDialog.done(dia, 14))
            btnDelAll.setEnabled(False)
            buttonBox.addButton(btnDelAll, QDialogButtonBox.NoRole)
            # Delete selection
            btnDelSel = QPushButton()
            btnDelSel.setText("Delete")
            btnDelSel.setToolTip("Delete selected entries")
            btnDelSel.clicked.connect(lambda: QDialog.done(dia, 13))
            btnDelSel.setEnabled(False)
            buttonBox.addButton(btnDelSel, QDialogButtonBox.NoRole)
        if bGoldEdit:
            # Gold Bulk Edit: Unset gold
            btnGoldUnset = QPushButton()
            btnGoldUnset.setText("Not-Gold")
            btnGoldUnset.setToolTip("Set not-gold status for selected entries")
            btnGoldUnset.clicked.connect(lambda: QDialog.done(dia, 15))
            btnGoldUnset.setEnabled(False)
            buttonBox.addButton(btnGoldUnset, QDialogButtonBox.NoRole)
            # Gold Bulk Edit: Set gold
            btnGoldSet = QPushButton()
            btnGoldSet.setText("Gold")
            btnGoldSet.setToolTip("Set gold status for selected entries")
            btnGoldSet.clicked.connect(lambda: QDialog.done(dia, 16))
            btnGoldSet.setEnabled(False)
            buttonBox.addButton(btnGoldSet, QDialogButtonBox.NoRole)
        if bSave:
            # Save selection (CSV)
            btnSaveSel = QPushButton()
            btnSaveSel.setText("Save")
            btnSaveSel.setToolTip("Saving selected rows to csv file")
            # QObject.connect(btnSave, SIGNAL("clicked()"), lambda: QDialog.done(dia, 7))
            btnSaveSel.clicked.connect(lambda: QDialog.done(dia, 8))
            buttonBox.addButton(btnSaveSel, QDialogButtonBox.NoRole)
            btnSaveSel.setEnabled(False)

        # Ok
        btnOk = QPushButton()
        btnOk.setText("Ok")
        btnOk.clicked.connect(lambda: QDialog.done(dia, 1))
        buttonBox.addButton(btnOk, QDialogButtonBox.NoRole)

        # Button Verbinden
        btnVor.clicked.connect(lambda: QDialog.done(dia, 2))
        btnEnde.clicked.connect(lambda: QDialog.done(dia, 3))
        btnZurueck.clicked.connect(lambda: QDialog.done(dia, 4))
        btnAnfang.clicked.connect(lambda: QDialog.done(dia, 5))
        btnGo.clicked.connect(lambda: QDialog.done(dia, 6))

        # Dialog zusammenbauen
        if multisheet.getNoOfSheets() > 1:
            vbox.addWidget(ingestWidget)
            vbox.setAlignment(ingestWidget, Qt.AlignHCenter)
        vbox.addWidget(tableWidget)
        vbox.addWidget(datensaetzeWidget)
        vbox.addWidget(rangeWidget)
        vbox.addWidget(buttonBox)
        vbox.setAlignment(rangeWidget, Qt.AlignHCenter)
        vbox.setAlignment(datensaetzeWidget, Qt.AlignHCenter)
        dia.resize(800, 600)

        # Warteanzeige aus
        self.ui_hideProgress()

        # Zentrieren zum Hauptfenster
        self.ui_centerWidget(dia)

        # Am Ende sortierbar machen
        tableWidget.setSortingEnabled(True)

        # Font
        font = QFont()
        font.setFamily("Arial")
        font.setPointSize(10)
        dia.setFont(font)

        # Dialog anzeigen
        iRet = dia.exec()

        # Antwort auswerten
        if iRet == 1:
            # Ende
            return
        elif iRet == 2:
            # Vor
            kwargs['select'] = bSelect
            if len(setSelectedIds) != 0:
                kwargs['setSel'] = setSelectedIds
            kwargs['id'] = None
            kwargs["rangeIdx"] = iRangeIdx + 1
            self.table_showGeneric(multisheet, **kwargs)
        elif iRet == 3:
            # Ende
            kwargs['select'] = bSelect
            if len(setSelectedIds) != 0:
                kwargs['setSel'] = setSelectedIds
            kwargs['id'] = None
            kwargs["rangeIdx"] = iRangeIdxMax
            self.table_showGeneric(multisheet, **kwargs)
        elif iRet == 4:
            # Zurück
            kwargs['select'] = bSelect
            if len(setSelectedIds) != 0:
                kwargs['setSel'] = setSelectedIds
            kwargs['id'] = None
            kwargs["rangeIdx"] = iRangeIdx - 1
            self.table_showGeneric(multisheet, **kwargs)
        elif iRet == 5:
            # Anfang
            kwargs['select'] = bSelect
            if len(setSelectedIds) != 0:
                kwargs['setSel'] = setSelectedIds
            kwargs['id'] = None
            kwargs["rangeIdx"] = 0
            self.table_showGeneric(multisheet, **kwargs)
        elif iRet == 6:
            # Go
            kwargs['select'] = bSelect
            if len(setSelectedIds) != 0:
                kwargs['setSel'] = setSelectedIds
            kwargs['id'] = None
            kwargs["rangeIdx"] = int(ausgabeFeldRangeFensterIdx.text()) - 1
            self.table_showGeneric(multisheet, **kwargs)
        elif iRet == 8:
            # Save Csv (selection)
            lstContentSelected = [list(lst) for lst in lstContentFull if lst[iIdxEditId] in setSelectedIds]
            self.ui_setStatusbarText("Extracted rows: " + str(len(lstContentSelected)))
            # Humanisieren
            # Entferne "ids"
            if "id" in lstHeader:
                self.ui_showProgress()
                self.ui_setWorkflow(top="Removing ids", sub="Working")
                self.ui_setProgressbarTextAnzeigen(True)
                self.ui_setProgressbarMinMax(1, 100)
                iIdIdx = lstHeader.index("id")
                iRows = len(lstContentSelected)
                # Entferne gesamte Spalte "id" (Header und Inhalt)
                for i, row in enumerate(lstContentSelected):
                    # Progress
                    iProzent = self.tools_getProgress(i, iRows, 100)
                    if iProzent is not None:
                        self.ui_setProgressbarWert(iProzent)
                        self.ui_setProgressbarText(str(iProzent) + "%")
                        self.ui_giveQtTimeToUpdate()
                    del row[iIdIdx]
                lstHeader.remove("id")
                self.ui_hideProgress()
            self.tools_saveDataCsv([lstHeader] + lstContentSelected, sDiaTitle + "_selection")
        elif iRet == 13:
            # Del selection
            if sEditId in lstHeader:
                iIdxId = lstHeader.index(sEditId)
            else:
                pn("Internal error while selection of entires in Table: Column "+sEditId+ "not found in Table "+sTabName)
                return
            lstIdsSelected = list(setSelectedIds)
            self.ui_setStatusbarText("Selected rows: " + str(len(lstIdsSelected)))
            self.table_entryDel(sTabName, lstIdsSelected, select=lstSelect)
        elif iRet == 14:
            # Del All in Windoww
            self.ui_setStatusbarText("Selected rows: All")
            self.table_entryDel(sTabName, lstIdsAll)
        elif iRet == 15:
            # Set Gold
            if sEditId in lstHeader:
                iIdxId = lstHeader.index(sEditId)
            else:
                pn("Internal error while selection of entires in Table: Column "+sEditId+ "not found in Table "+sTabName)
                return
            lstIdsSelected = list(setSelectedIds)
            self.ui_setStatusbarText("Selected rows: " + str(len(lstIdsSelected)))
            self.table_entryEditGold(sTabName, lstIdsSelected, False)
        elif iRet == 16:
            # Unset Gold
            if sEditId in lstHeader:
                iIdxId = lstHeader.index(sEditId)
            else:
                pn("Internal error while selection of entires in Table: Column "+sEditId+ "not found in Table "+sTabName)
                return
            lstIdsSelected = list(setSelectedIds)
            self.ui_setStatusbarText("Selected rows: " + str(len(lstIdsSelected)))
            self.table_entryEditGold(sTabName, lstIdsSelected, True)
        elif iRet >= 20:
            # Zeige neue Ingest Table an
            multisheet.setOpenSheetIdx(iRet-20) # Kodierung des Index der anzuzeigenden Tabelle: 20+Index (speicherung in Multisheet)
            self.table_showGeneric(multisheet,
                                   ingId=iIngId, edit=False, diaTitle=sDiaTitle,
                                   rangeIdx=0, save=False, gold=bGold, used=bUsed, gear_double=bGear_double, doi=bDoi, doi_pangaea=bDoiPangaea, access=bAccess)

    def dialog_input_createGeneric(self, sTitel, lstAttr, sTabelle, dicAttrObligat={}, dicAttrComments={}, **kwargs):

        """
        Erzeuge den Dialog für eine generischen Datensatzeingabe

        Args:
            sTitel (str): Titel des Dialogfensters
            lstAttr (list): Liste aller Attribute
            sTabelle (str): Tabellenname
            dicAttrObligat (dict): Dictionary mit Obligatinfos zu den entsprechenden Attributen
            dicAttrComments (dict): Dictionary mit Kommentaren zu den entsprechenden Attributen

        Kwargs:
            showCsv (bool): Soll ein CSV-Import zum massenweisen Ingesten von neuen Datensätzen angeboten werden?
            defaultData (dict): Ein Dictionary mit entsprechenen Defaultwerten zu Attributen
            fokusCol (str): Name des Eingsabefeldes, auf das der Fokus liegen soll
        Returns:
            (PyQt5.QtGui.QDialog, dict): Das erzeugte Dialogfenster, eine Liste mit den Eingabefelder für die entsprechenden Attribute

        """
        pn = self.ui_printConsoleNormal

        # Kwargs
        dicDefaultData = None
        bShowCsv = False
        sFokusCol = None
        if "defaultData" in kwargs:  # Default-Werte
            dicDefaultData = kwargs['defaultData']
        if "showCsv" in kwargs:  # Zeige CSV-Import-Funktion
            bShowCsv = kwargs['showCsv']
        if "fokusCol" in kwargs:
            sFokusCol = kwargs['fokusCol']

        # Dialog erzeugen
        eingabeFeldFokus = None
        dia = QDialog()
        dia.setWindowTitle(sTitel)
        dia.setWindowModality(Qt.ApplicationModal)
        dia.setWindowFlags(Qt.Dialog | Qt.CustomizeWindowHint | Qt.WindowTitleHint | Qt.WindowCloseButtonHint)

        vbox = QVBoxLayout(dia)

        # Tabelle
        tabelle = QWidget()
        grid = QGridLayout(tabelle)
        iNummer = 1  # Aktuelle Attributnummer
        iRow = 1  # dazugehörige Zeile in Tabelle
        iCol = 1  # dazugehörige Spalte in Tabelle
        dicWidgets = dict()

        # Umbruch berechnen
        iMaxAttr = len(lstAttr)
        iAttrPerCol = 5  # Attribute pro Spalte
        if iMaxAttr > 5:
            iAttrPerCol = int(iMaxAttr / 2)

        # Durchlaufe alle Attribute
        for sName in lstAttr:
            # Ganz Platt: wenn ein Feld gold heist ist es nicht zu editieren (gold sind in Tabellen: Lookup ausser Taxon, TaxonWizard Base)
            if sName != "gold":

                # Label
                # Lookup/Taxon -> Table.Attribute/Sheet.Column
                if self.__map.all_isTableLookup(sTabelle):
                    sSheet, sCol  = self.__map.lookup_db2excel(sTabelle, sName)
                    if None in (sSheet, sCol):
                        pn("Internal error building lookup entry: "+str(sTabelle)+"."+str(sName))
                        sCol = sName
                    sLabel = sCol
                elif self.__map.all_isTableTaxon(sTabelle):
                    sSheet, sCol = self.__map.taxon_db2excel(sTabelle, sName)
                    if None in (sSheet, sCol):
                        pn("Internal error building taxon entry: " + str(sTabelle) + "." + str(sName))
                        sCol = sName
                    sLabel = sCol
                else:
                    sLabel = sName.replace("_", " ").title()
                label = QLabel(sLabel)

                if sName in dicAttrObligat and dicAttrObligat[sName]:  # Obligat
                    font = QFont()
                    font.setBold(True)
                    font.setPointSize(10)
                    label.setFont(font)
                else:
                    font = QFont()
                    font.setPointSize(10)
                    #font.setItalic(True)
                    label.setFont(font)
                grid.addWidget(label, iRow, iCol)

                # Label
                if sName != "gold": # Gold niemals zum Editieren freigeben
                    eingabeFeld = QLineEdit()
                    font = QFont()
                    font.setPointSize(10)
                    eingabeFeld.setFont(font)

                # Defaultwert
                if dicDefaultData is not None and sName in dicDefaultData and dicDefaultData[sName] is not None:
                    eingabeFeld.setText(str(dicDefaultData[sName]))
                    eingabeFeld.setCursorPosition(0)

                # ---------------------------------------------------------------
                # Tweak/Spezialisierung
                # ---------------------------------------------------------------

                # TODO: Das läuft hier im GUI Thread -> Ist nicht so gut!
                # Update Lookup System für Person und Country
                if sTabelle == "dataset":
                    # Dataset: Update Lookups: Person, Country
                    self.__model.subsystem_getLookup().update(["person", "country"])

                # Gear
                # category -> Combobox
                if sTabelle == "gear" and sName == "category":
                    combo = QComboBox()
                    combo.addItems(["grab", "trawl", "trap", "video", "sensor", "scrape", "rocks", "census", "remote sensing" ])
                    # Defaultwert
                    if dicDefaultData != None and sName in dicDefaultData:
                        sWert = str(dicDefaultData[sName])
                    else:
                        sWert = "grab"
                    # Einstellen
                    if sWert == "grab":
                        combo.setCurrentIndex(0)
                    elif sWert == "trawl":
                        combo.setCurrentIndex(1)
                    elif sWert == "trap":
                        combo.setCurrentIndex(2)
                    elif sWert == "video":
                        combo.setCurrentIndex(3)
                    elif sWert == "sensor":
                        combo.setCurrentIndex(4)
                    elif sWert == "scrape":
                        combo.setCurrentIndex(5)
                    elif sWert == "rocks":
                        combo.setCurrentIndex(6)
                    eingabeFeld = combo
                # StandardUnit -> Combobox
                if sTabelle == "gear" and sName == "smpunit":
                    combo = QComboBox()
                    combo.addItems(["both", "area", "vol"])
                    # Defaultwert
                    if dicDefaultData != None and sName in dicDefaultData:
                        sWert = str(dicDefaultData[sName])
                    else:
                        sWert = "both"
                    # Einstellen
                    if sWert == "both":
                        combo.setCurrentIndex(0)
                    elif sWert == "area":
                        combo.setCurrentIndex(1)
                    elif sWert == "vol":
                        combo.setCurrentIndex(2)
                    eingabeFeld = combo
                # Dataset
                    # Collection type Monitoring / Research -> Bool
                if sTabelle == "dataset" and sName == "collection":
                    combo = QComboBox()
                    combo.addItems(["M", "R"])
                    # Defaultwert
                    if dicDefaultData != None and sName in dicDefaultData:
                        sWert = str(dicDefaultData[sName])
                    else:
                        sWert = "R"
                    # Einstellen
                    if sWert == "M":
                        combo.setCurrentIndex(0)
                    elif sWert == "R":
                        combo.setCurrentIndex(1)
                    eingabeFeld = combo

                    # Access level data -> Combobox
                if sTabelle == "dataset" and sName == "access_level_data":
                    combo = QComboBox()
                    combo.addItems(["confidential",  "request", "free"])
                    # Defaultwert
                    if dicDefaultData != None and sName in dicDefaultData:
                        sWert = str(dicDefaultData[sName])
                    else:
                        sWert = "confidential"
                    # Einstellen
                    if sWert == "confidential":
                        combo.setCurrentIndex(0)
                    elif sWert == "request":
                        combo.setCurrentIndex(1)
                    elif sWert == "free":
                        combo.setCurrentIndex(2)
                    eingabeFeld = combo

                    # Access level data product -> Combobox
                if sTabelle == "dataset" and sName == "access_level_data_product":
                    combo = QComboBox()
                    combo.addItems(["confidential", "request", "free"])
                    # Defaultwert
                    if dicDefaultData != None and sName in dicDefaultData:
                        sWert = str(dicDefaultData[sName])
                    else:
                        sWert = "confidential"
                    # Einstellen
                    if sWert == "confidential":
                        combo.setCurrentIndex(0)
                    elif sWert == "request":
                        combo.setCurrentIndex(1)
                    elif sWert == "free":
                        combo.setCurrentIndex(2)
                    eingabeFeld = combo

                    # Person -> Lookuptable -> Combobox
                if sTabelle == "dataset" and sName == "contact_person_id":
                    combo = QComboBox()
                    dicPerson2id = self.__model.subsystem_getLookup().getLookupCacheAll("person").getData("lookup")
                    dicId2person = self.__model.subsystem_getLookup().getLookupCacheAll("person", reverse=True).getData("lookup")
                    lstPerson = sorted(list(dicPerson2id))
                    combo.addItems(lstPerson)
                    # Defaultwert
                    if dicDefaultData != None and sName in dicDefaultData:
                        try:
                            iId = dicDefaultData[sName]
                            sPerson = dicId2person[str(iId)]
                            combo.setCurrentIndex(combo.findData(sPerson, Qt.DisplayRole))
                        except:
                            pn("Error while building dialog: Default value for field contact_person_id not found")
                    eingabeFeld = combo

                    # Country -> Lookuptable -> Combobox
                if sTabelle == "dataset" and sName == "owner_country_id":
                    combo = QComboBox()
                    dicCountry2id = self.__model.subsystem_getLookup().getLookupCacheAll("country").getData("lookup")
                    dicId2country = self.__model.subsystem_getLookup().getLookupCacheAll("country", reverse=True).getData("lookup")
                    lstCountry = list(dicCountry2id)
                    combo.addItems(lstCountry)
                    # Defaultwert
                    if dicDefaultData != None and sName in dicDefaultData:
                        try:
                            dicId2country = dict((v, k) for k, v in dicCountry2id.items())
                            iId = dicDefaultData[sName]
                            sCountry = dicId2country[str(iId)]
                            combo.setCurrentIndex(combo.findData(sCountry, Qt.DisplayRole))
                        except:
                            pn("Error while building dialog: Default value for field owner_country_id not found")
                    eingabeFeld = combo

                # ---------------------------------------------------------------

                # Kommentar dranhaengen
                if sName in dicAttrComments:
                    eingabeFeld.setToolTip(dicAttrComments[sName])

                # Fokus auf Feld legen
                if sName == sFokusCol:
                    eingabeFeldFokus = eingabeFeld

                # Widgets in Dictionary merken
                dicWidgets[sName] = eingabeFeld

                # Widget in Layout einfuegen
                grid.addWidget(eingabeFeld, iRow, iCol + 1)
                if (iNummer % iAttrPerCol) == 0:
                    iCol += 2
                    iRow = 0
                iNummer += 1
                iRow += 1

        # Buttonbox
        buttonBox = QDialogButtonBox(QDialogButtonBox.Ok | QDialogButtonBox.Cancel, Qt.Horizontal, dia)
        buttonBox.accepted.connect(dia.accept)
        buttonBox.rejected.connect(dia.reject)
        if bShowCsv:
            buttonCsv = buttonBox.addButton("Import file", QDialogButtonBox.NoRole)
            buttonCsv.clicked.connect(lambda: QDialog.done(dia, 2))

        # Dialog zusammenbauen
        vbox.addWidget(tabelle)
        vbox.addWidget(buttonBox)
        dia.setMinimumSize(dia.sizeHint())  # Groeße fixieren

        # Font
        font = QFont()
        font.setFamily("Arial")
        font.setPointSize(10)
        dia.setFont(font)

        # Setze Fokus auf ein Eingabefeld (f
        if eingabeFeldFokus is not None:
            try:
                eingabeFeldFokus.selectAll()
                eingabeFeldFokus.setFocus()
            except:
                pass

        return (dia, dicWidgets)

    def input_dialog_generic_getAttr(self, cursor, sTabelle, lstAttributAusschluss):
        """
        Hole Attribute, Obligatinfo, Kommentare zur generischer Eingabe eines neuen Datenstatzes in eine Tabelle

        Args:
            cursor (psycopg2.cursor.Cursor): Psql-Cursor
            sTabelle (str): Tabellenname
            lstAttributAusschluss (list): auszuschliessende Attribute

        Returns:
            (list,dict,dict): Liste der Attributnamen, Dictionary mit Obligatinfos zu Attributnamen, Dictionary mit Kommentare zu Attributnamen,

        """

        # Entferne nicht gewuenscht Attribute
        resAttr = self.__model.subsystem_getDb().admin_getAttributNames(sTabelle, cursor=cursor)
        if resAttr.isErr():
            raise NameError(resAttr.getResult())
        lstTmp = resAttr.getData("header")
        lstAttr = [x for x in lstTmp if (x not in lstAttributAusschluss)]

        # Hole Kommentare der Attribute
        dicAttrComments = dict()
        for sAttr in lstAttr:
            sSystem = None
            sSheet = None
            sCol = None
            sComment = None
            sSystem = self.__map.all_getTableSystem(sTabelle)
            if self.__map.all_isTableLookup(sTabelle):
                sSheet, sCol  = self.__map.lookup_db2excel(sTabelle, sAttr)
                resComment = self.__model.conf_getDbSchema().getDbSchemaInfo(sSystem, sSheet, sCol)
                sComment = resComment.getData("comment-full")
            if self.__map.all_isTableTaxon(sTabelle):
                sSheet, sCol = self.__map.taxon_db2excel(sTabelle, sAttr)
                resComment = self.__model.conf_getDbSchema().getDbSchemaInfo(sSystem, sSheet, sCol)
                sComment = resComment.getData("comment-full")
            if sComment is not None:
                dicAttrComments[sAttr] = sComment

        # Hole Obligatinfos der Attribute
        dicAttrObligat = dict()
        for sAttr in lstAttr:
            resObl = self.__model.subsystem_getDb().admin_getObligat(sTabelle, sAttr, cursor=cursor)
            bObl = resObl.getData("obl") if resObl.isOk() else False
            dicAttrObligat[sAttr] = bObl

        return (lstAttr, dicAttrObligat, dicAttrComments)

    def dialog_input_generic_evalDialog(self, dicWidgets, sTabelle=None):
        """
        Werte die einzelnen Qt-Widgets eines Dialogs einer generische Datensatzeingabe aus

        Args:
            dicWidgets (PyQt5.QtWidgets.QWidgets): Die abgeleiteten Widgets können hierbei Textfelder oder Comboboxen sein
            sTabelle (str): Tabellenname

        Returns:
            (dict): Dictionary mit den entsprechenden Key/Value-Paaren für die eingegebenen Daten
        """

        # Durchlaufe alle Eingabe-Widgets
        dicVal = dict()
        for sName, feld in dicWidgets.items():
            # Combobox
            if isinstance(feld, QComboBox):

                # ---------------------------------------------------------------
                # TWEAK/SPEZIALISIERUNG
                # ---------------------------------------------------------------

                # Humane Sicht auf die Daten
                # Dataset
                # Name -Auflösen via Lookup-Table-> contact_person_id
                if sTabelle == "dataset" and sName == "contact_person_id":
                    dicPerson = self.__model.subsystem_getLookup().getLookupCacheAll("person").getData("lookup")
                    person = feld.currentText()
                    if person in dicPerson:
                        dicVal[sName] = dicPerson[person]
                # Name -Auflösen via Lookup-Table-> owner_country_id
                elif sTabelle == "dataset" and sName == "owner_country_id":
                    dicCountry = self.__model.subsystem_getLookup().getLookupCacheAll("country").getData("lookup")
                    country = feld.currentText()
                    if country in dicCountry:
                        dicVal[sName] = dicCountry[country]
                else:
                    dicVal[sName] = feld.currentText()

                # ---------------------------------------------------------------

            # Textfeld
            else:
                if feld.text() == "":
                    sVal = None
                else:
                    sVal = feld.text()
                dicVal[sName] = sVal

        return dicVal

    def tools_saveDataCsv(self, lstData, sFilenameDefault="data"):
        """
        Speichere Daten in CSV-Datei

        Args:
            lstData (list): Daten in Form einer Listen-Liste (Rows/Columns)
            sFilenameDefault="data" (str): Dateiname

        Returns:
            (bool): Vorgang hat funktioniert?
        """

        pn = self.ui_printConsoleNormal
        pv = self.ui_printConsoleVerbose
        pd = self.ui_printConsoleDebug

        sModulname = "Save table in csv file"

        # Nichts zu speichern
        if lstData is None or lstData == []:
            return True

        # "Säubere" den Dateinamen
        sFilenameDefault = re.sub('[^A-Za-z0-9_,. ]+', '', sFilenameDefault)
        sFilenameDefault = sFilenameDefault.replace(",", " ")
        sFilenameDefault = sFilenameDefault.replace(".", " ")
        sFilenameDefault = " ".join(sFilenameDefault.split())
        sFilenameDefault = sFilenameDefault.replace(" ", "_").lower() + ".csv"

        # Hole CSV-Dateinamen
        csvFile = QFileDialog.getSaveFileName(self, "csv file", sFilenameDefault)[0]
        if csvFile == "" or csvFile is None:
            return

        # Worker starten
        self.worker_start(Worker_SaveDataToFile(self.__model.ui_getSmartprinter(), lstData, csvFile), self.worker_finished_saveDataToFile)

    def ui_giveQtTimeToUpdate(self, max=1):
        """
        Gib dem QT-Framework Zeit ausstehende Veränderungen an der GUI vorzunehmen

        Args:
            max (int): Anzahl der Aufrufe an QT, Tweak: Häufig reicht ein Aufruf nicht zum GUI-Refresh, sondern erst 10 ...
        """

        for i in range(max):
            # pass
            QApplication.processEvents()

    def ui_blockSignals(self, lstWidgets, bBlock=True):
        """
        Block Signals of widgets

        Args:
            lstWidgets (list): Widgets
            bBlock (bool): Block/Deblock
        """
        for widget in lstWidgets:
            widget.blockSignals(bBlock)

    def ui_deblockSignals(self, lstWidgets):
        self.ui_blockSignals(lstWidgets, False)

    def tools_getProgress(self, iWertAktuell, iWertMax, iSchritte):
        """
        Ermittle Fortschritt in Prozent (Schrittweise)

        Args:
            iWertAktuell (int): Aktueller Wert
            iWertMax (int): Maximalwert der Fortschrittskala
            iSchritte (int): Schrittanzahl in die die gesamte Skala durchlaufen werden soll

        Returns:
            int: Fortschritt in Prozent (0-100%)
        """
        try:
            # iSchritte teilt Gesamtfortschritt in Einzelteile auf (1 ~ einem 100%-Schritt, 100 ~ 100 1%-Schritten)
            if iSchritte > 100:
                # Mehr als 100 Teile gibt es nicht (1%-Schritte)
                iSchritte = 100
            fStepsZeilen = iWertMax / iSchritte
            if iWertAktuell % fStepsZeilen < 1:
                # Nächster Schritt erreicht -> String zurückgeben
                iProzent = int(iWertAktuell / fStepsZeilen / iSchritte * 100)
                return iProzent
        except:
            pass

        # Nächster Schritt noch nicht erreicht (oder Fehler)
        return None

    # --- Thread Tools

    def worker_start(self, worker, finish, **kwargs):

        """
        Starte einen Worker
        """

        if worker is None or finish is None:
            return

        worker.connectFinish(finish) # Abschlussmethode

        if len(kwargs) != 0:
            worker.init(**kwargs) # Initialisierung, wenn vorhanden

        if not self.sys_isWorkerRunning(status=True, popup=True):
            self.__model.thread_setWorker(worker)
            worker.start()

    def worker_abort(self):
        """
        Weise einen Worker an sich zu beenden
        """

        if self.sys_isWorkerRunning():
            self.ui_setStatusbarText("An abort command has been send to the running process. Please wait for the reaction ...")
            # mb = QMessageBox.warning(self, " ", "Do you really want to abort the running process?", QMessageBox.Yes | QMessageBox.No)
            # if mb == QMessageBox.Yes:
            self.__model.thread_getWorker().abort()

    def sys_isWorkerRunning(self, **kwargs):
        """
        Teste ob schon ein Worker läuft

        Kwargs
            popup (bool): Popup falls ein Prozess bereits läuft
            status (bool): Statusmeldung falls ein Prozess bereits läuft
        Returns: bool
        """

        bPopup = False
        bStatus = False
        if "popup" in kwargs:
            bPopup = True
        if "status" in kwargs:
            bStatus = True

        if self.__model.thread_getWorker() is None:
            return False
        else:
            if self.__model.thread_getWorker().isRunning():
                sMsg1 = "Another process is runnning ..."
                sMsg2 = "Please abort it or wait until it has finished."
                if bPopup:
                    self.popup_infoTimer(sMsg1, sMsg2)
                elif bStatus:
                    self.ui_setStatusbarText(sMsg1 + " " + sMsg2)
                return True
            else:
                return False

    def worker_finished(self, result):
        """
        Worker ist fertig und gibt hier ein Ergebnis zurück
        """
        self.ui_hideProgress()
        if result is not None:
            if result.isOk():
                sText = (result.getName() if result.getName() is not None else "Work") + " completed"
                sTextExtra = "For more information, please refer to the console"
                sTextDetails = ""
                if result.isWarn():
                    sTextDetails += "*** Warnings\n" + result.getWarn()
                if result.isInfo():
                    sTextDetails += "*** Infos\n" + result.conf_getInfo()
                self.popup_generic(title="Info", head=sText, info=sTextExtra, detail=sTextDetails)
            else:
                sText = (result.getName() if result.getName() is not None else "Work") + " failed."
                sTextExtra = "For more information, please refer to the console"
                sTextDetails = ""
                sTextDetails += "*** Errors\n" + result.getResult()
                if result.isWarn():
                    sTextDetails += "*** Warnings\n" + result.getWarn()
                if result.isInfo():
                    sTextDetails += "*** Infos\n" + result.getData()
                self.popup_error(sText, sTextExtra, sTextDetails)

    def worker_finished_serverConnection(self, result):
        """
        Worker Server Connection ist beended
        """
        self.ui_hideProgress()
        pn = self.ui_printConsoleNormal
        pv = self.ui_printConsoleVerbose
        status = self.ui_setStatusbarText
        if result is not None:
            if result.isErr():
                pn("* Connetion to server failed")
                pn(result.getErrWarnInfoString(short=True))
            elif result.isWarn():
                pn("* Connection to server established")
                status("Connection to server established")
                pn(result.getErrWarnInfoString(short=True))
            else:
                status("Connection to server established")
        # Vis Selector einstellen
        self.explorer_reset_visual()
        self.explorer_selectorEnable(True)
        self.status_refresh_all()

    def worker_finished_adminCmd(self, result):
        """
        Das Ergebnis des Admin Kommandos wird angezeigt
        """

        self.ui_hideProgress()
        if result is not None:
            if result.isOk():
                if result.getData("cmd") == "snapshot":
                    # Snapshot
                    sSnapshotFile = result.getData("snapshot_file")
                    sSnapshotFolder = self.__model.conf_getAppSnapshotDirSchema()
                    sInfo = '<html><head/><body><p>Snapshot<br><b>'+str(sSnapshotFile)+'</b><br>was saved in <a href="file:///'+str(sSnapshotFolder)+'"><span style=" text-decoration: underline; color:#0000ff;">snapshot folder</span></a></p></body></html>'
                    self.popup_generic(title="Snapshot", head="Snapshot was created sucessfully", info=sInfo)
                else:
                    # Rest
                    self.admin_out.setPlainText(result.getData("out"))
                    if result.getData("popup"):
                        self.popup_generic(title="Core", head="Everything went well", info=result.getData("title"))
            else:
                # Fehler
                sDetail = "*** Error\n" + result.getErr() + "\n"
                self.popup_error(result.getName(), "An error occurred.", sDetail)

    def worker_finished_exportExcel(self, result):
        """
        Worker Export Data to Excel ist beeendet
        """

        self.ui_hideProgress()
        pn = self.ui_printConsoleNormal
        pv = self.ui_printConsoleVerbose

        # Info Popup
        bIngest = result.getData("ingest")
        bLookup = result.getData("lookup")
        bTaxSys = result.getData("taxon")
        bTemplate = result.getData("template")
        odicInfo = result.getData("info")
        lstUsedIng = result.getData("used-ingests")
        sLog = result.getData("log")

        if bTemplate:
            sTitle = "Template creation completed"
            sInfo = "Created template parts"
            sInfo += "<ul>"
            if bIngest:
                sInfo += "<li>Ingest</li>"
            if bLookup:
                sInfo += "<li>Lookup</li>"
            if bTaxSys:
                sInfo += "<li>TaxonWizard</li>"
            sInfo += "</ul>"
        else:
            sTitle = "Export to Excel completed"
            sInfo = "<b>Exported data</b>"
            sInfo += "<ul>"
            if bIngest:
                sInfo += "<li>Ingest</li>"
            if bLookup:
                sInfo += "<li>Lookup</li>"
            if bTaxSys:
                sInfo += "<li>TaxonWizard</li>"
            sInfo += "</ul>"

        # Fehler
        sExtra = None
        if result.isErr():
            sInfo += "<b>Errors occurred</b>"
            sExtra = result.getErrWarnInfoString(short=True)
        elif not bTemplate:
            # Kein Template
            # Info
            # Konsole
            sStat = "************************************\n"
            sStat += "Export to Excel\n"
            sStat += "************************************\n"
            sStat += "\n"
            for sTable, odicSheets in odicInfo.items():
                lstConsole = [("Table","Entries")]
                for sSheet, nr in odicSheets.items():
                    lstConsole.append((sSheet, str(nr)))
                sStat+= sTable.upper() +" \n"
                sStat += PrettyText.table(lstConsole, True)
                sStat += "\n\n"
                #self.printConsoleNormal(sStat)
            # Extra (Export Report)
            if sLog is not None:
                sExtra = sStat
                sExtra += sLog
        if bTemplate:
            # Template
            ant = self.popup_generic(title="Export Template", head=sTitle, info=sInfo)
        else:
            # Export von Datem
            ant = self.popup_generic(title="Export data", head=sTitle, info=sInfo, button_special="Show Protocol")
            # Show Extra
            if ant == 2:
                self.popup_text(title="EXPORT PROTOCOL", text=sExtra, save=True)

    def worker_finished_ingestionLookup(self, result):
        """
        Worker Ingestion Lookup ist beeendet
        """

        iNewRecs = result.getData("recs")
        iNoRecsExistEqual = result.getData("exist-equal")
        iNoRecsExistDiff = result.getData("exist-diff")
        iNoRecsErr = result.getData("exist-err")

        self.ui_hideProgress()
        pn = self.ui_printConsoleNormal

        lstSheetsUsed = result.getData("lookup-excel-sheets-used")
        lstTablesChanged = result.getData("lookup-tables-changed")
        iImportNo = result.getData("lookup-import-no")
        iRecProc = result.getData("lookup-excel-recs")
        iRecNew = result.getData("lookup-new")
        iRecNewNo = result.getData("lookup-new-no")
        iRecUpd = result.getData("lookup-update")
        iRecUpdNo = result.getData("lookup-update-no")
        iErr = result.getData("lookup-err")
        sLog = result.getData("log")

        # Info Popup
        sTitle = "Lookup ingestion complete"
        sInfo = ""
        if self.__model.conf_isDry():
            sInfo += "<i>dry mode was activated</i><p>"
        if result.isErr():
            sInfo += "<b>Errors occurred</b>"
            sInfo += "<ul><li>" + str(iErr) + "</li></ul>"
        else:
            sInfo += "<b>Processed EXCEL records</b>"
            sInfo += "<ul><li>" + str(iRecProc) + "</li></ul>"
            if lstTablesChanged is not None and len(lstTablesChanged) >0:
                sInfo += "<b>Changed Db tables</b>"
                sInfo += "<ul>"
                for sTable in lstTablesChanged:
                    sInfo += "<li>" + str(sTable) + "</li>"
                sInfo += "</ul>"
            if iImportNo is not None and (iImportNo > 0):
                sInfo += "<b>Discarded Excel records</b>"
                sInfo += "<ul><li>" + str(iImportNo) + "</li></ul>"
            if (iRecNew != 0 and iRecNew is not None) or (iRecNewNo != 0 and iRecNewNo is not None):
                lstInfo = []
                lstVal = []
                if (iRecNew != 0 and iRecNew is not None):
                    lstInfo.append("New Db records")
                    lstVal.append(str(iRecNew))
                if (iRecNewNo != 0 and iRecNewNo is not None):
                    lstInfo.append("Denied Db creations")
                    lstVal.append(str(iRecNewNo))
                sInfo += "<b>"+PrettyText.lst2str(lstInfo, " / ")+"</b>"
                sInfo += "<ul><li>" +PrettyText.lst2str(lstVal, " / ")+ "</li></ul>"
            if (iRecUpd!= 0 and iRecUpd is not None) or (iRecUpdNo != 0 and iRecUpdNo is not None):
                lstInfo = []
                lstVal = []
                if (iRecUpd != 0 and iRecUpd is not None):
                    lstInfo.append("Updated Db records")
                    lstVal.append(str(iRecUpd))
                if (iRecUpdNo != 0 and iRecUpdNo is not None):
                    lstInfo.append("Denied Db updates")
                    lstVal.append(str(iRecUpdNo))
                sInfo += "<b>"+PrettyText.lst2str(lstInfo, " / ")+"</b>"
                sInfo += "<ul><li>" +PrettyText.lst2str(lstVal, " / ")+ "</li></ul>"

        ant = self.popup_generic(title="Lookup", head=sTitle + ("in dry mode" if self.__model.conf_isDry() else ""), info=sInfo, button_special="Show Protocol")
        # Reconnect to Server?

        if result.isOk() and not self.__model.conf_isDry() and (iRecNew > 0 or iRecUpd > 0):
            pn("Workflows will be reinitialized ...")
            self.sys_server_disconnect()
            self.sys_server_connect()
        if ant == 1:
            self.popup_text(title="IMPORT PROTOCOL", text=sLog, save=True)

    def worker_finished_ingestionTaxon(self, result):
        """
        Worker Ingestion Taxon Admin Table ist beeendet
        """
        self.ui_hideProgress()
        pn = self.ui_printConsoleNormal

        lstTablesChanged = result.getData("taxon-tables-changed")
        iRecNew = result.getData("taxon-new")
        iRecUpd = result.getData("taxon-update")
        iRecUpdNo = result.getData("taxon-update-no")
        iRecUpdEqual = result.getData("taxon-equal")
        iErr = result.getData("taxon-err")
        sLog = result.getData("log")

        # Info Popup
        sTitle = "Taxon System Ingestion Complete"
        sInfo = ""
        if self.__model.conf_isDry():
            sInfo += "<i>dry mode was activated</i><p>"
        if result.isErr():
            sInfo += "<b>Errors occurred</b>"
            sInfo += "<ul><li>" + str(result.getErrNo()) + "</li></ul>"
        if lstTablesChanged is not None and len(lstTablesChanged) > 0:
            sInfo += "<b>Changed Db tables</b>"
            sInfo += "<ul>"
            for sTable in lstTablesChanged:
                sInfo += "<li>" + str(sTable) + "</li>"
            sInfo += "</ul>"
        if (iRecNew != 0 and iRecNew is not None):
            sInfo += "<b>New records</b>"
            sInfo += "<ul><li>" + str(iRecNew) + "</li></ul>"
        if (iRecUpd != 0 and iRecUpd is not None):
            sInfo += "<b>Updated records</b>"
            sInfo += "<ul><li>" + str(iRecUpd) + "</li></ul>"
        if (iRecUpdNo != 0 and iRecUpdNo is not None):
            sInfo += "<b>Update denied (opt out or gold block)</b>"
            sInfo += "<ul><li>" + str(iRecUpdNo) + "</li></ul>"
        if (iRecUpdEqual != 0 and iRecUpdEqual is not None):
            sInfo += "<b>Update not necessary (in=Db)</b>"
            sInfo += "<ul><li>" + str(iRecUpdEqual) + "</li></ul>"
        # Reconnect to Server?
        if not self.__model.conf_isDry():
            pn("Workflows will be reinitialized ...")
            self.sys_server_disconnect()
            self.sys_server_connect()

        ant = self.popup_generic(title="Taxon", head=sTitle + ("in dry mode" if self.__model.conf_isDry() else ""), info=sInfo, button_special="Show Protocol")

        if not self.__model.conf_isDry():
            pn("Workflows will be reinitialized ...")
            self.sys_server_disconnect()
            self.sys_server_connect()
        if ant == 1:
            self.popup_text(title="IMPORT PROTOCOL", text=sLog, save=True)

    def worker_finished_goldEdit(self, result):
        """
        Worker Edit Gold
        """

        pn = self.ui_printConsoleNormal

        self.ui_hideProgress()
        self.ui_setWorkflow()

        # Parameter
        iEdit = result.getData("edit")
        bBool = result.getData("bool")
        iIdStart = result.getData("idstart")
        sTab = result.getData("tab")

        # Reconnect to Server?
        if not self.__model.conf_isDry():
            if iEdit != 0:
                pn("Workflows will be reinitialized ...")
                self.sys_server_disconnect()
                self.sys_server_connect()

        # Info Popup
        sInfo = "<b>Table</b>"
        sInfo += "<ul><li><i>" + sTab + "</i></li></ul>"
        if self.__model.conf_isDry():
            sInfo += "<b>Records gold status would have been edited</b>"
        else:
            sInfo += "<b>Records gold status edited</b>"
        sInfo += "<ul><li>" + str(iEdit) + "</li></ul>"
        if result.getErrMsgNo() != 0:
            sInfo += "<b>Failed operations</b>"
            sInfo += "<ul><li>" + str(result.getErrMsgNo()) + "</li></ul>"
        sTitle = "Records gold status edited"
        if self.__model.conf_isDry():
            sTitle += " in dry mode"
        sDetails = ""
        if result.isErr():
            sInfo += "<b>Errors occurred</b>: " + str(result.getErrNo()) + "<br>"
            sDetails = result.getResult()

        # Popup
        self.popup_generic(title="Gold", head=sTitle, info=sInfo, detail=sDetails)

        # Tabelle wieder anzeigen mit der Id des veraenderten Eintrages
        self.table_show(sTab, gold_edit=True)  # Tweak: Zeige den Eintrag mit der ID "davor" an, in der Hoffnung, dass es ihn gibt

    def worker_finished_delIngest(self, result):
        """
        Worker Delete Ingestion
        """

        pn = self.ui_printConsoleNormal

        self.ui_hideProgress()
        self.ui_setWorkflow()

        # Parameter
        lstDelTotal = result.getData("del-total")
        lstDelErr = result.getData("del-err")
        lstDelOk = result.getData("del-ok")
        if lstDelTotal is None:
            lstDelTotal = []
        if lstDelOk is None:
            lstDelOk = []
        if lstDelErr is None:
            lstDelErr= []

        # Details
        sDetails = ""
        if len(lstDelTotal) > 0:
            sDetails += "Selected Ingests: "+PrettyText.lst2strC(lstDelTotal) + "\n"
        if len(lstDelOk) > 0:
            sDetails += "Deleted Ingests: " + PrettyText.lst2strC(lstDelOk) + "\n"
        if len(lstDelErr) > 0:
            sDetails += "Not deleted Ingests: " + PrettyText.lst2strC(lstDelErr) + "\n"
        if result.isErr():
            sDetails += result.getErr()

        # Dialog
        if result is not None:
            sInfo = "<b>Deleted ingests</b>: " + str(len(lstDelOk)) + "/" + str(len(lstDelTotal)) + "<br>"
            if result.isErr():
                sInfo += "<b>Errors occurred</b>: " + str(len(lstDelErr)) + "<br>"
            if self.__model.conf_isDry():
                sTitle = "Simulated deletion in dry mode"
            else:
                sTitle = "Delete ingestions"
                pn("Workflows will be reinitialized ...")
                self.sys_server_disconnect()
                self.sys_server_connect()
            self.popup_generic(title="Delete Ingest", head=sTitle, info=sInfo, detail=sDetails)

    def worker_finished_purgeLookups(self, result):
        """
        Purge Lookups
        """

        pn = self.ui_printConsoleNormal

        self.ui_hideProgress()
        self.ui_setWorkflow()

        # Parameter
        dicPurged = result.getData("purged")
        if dicPurged is None:
            dicPurged = {}

        # Details/Info
        sDetails = ""
        sInfo = ""
        for sSheet, iPurged in dicPurged.items():
            sDetails += sSheet + ": "+str(iPurged) + "\n"
            sInfo += sSheet + ": " + str(iPurged) + "x<br>"
        if sDetails == "":
            sInfo = "Nothing to purge ..."
            sDetails = "Nothing to purge ..."
        else:
            sDetails = "* Purged records\n" + sDetails
        if result.isErr():
            sDetails += result.getErr()

        # Dialog
        if result is not None:
            if result.isErr():
                sInfo += "<b>Errors occurred</b>: " + str(result.getErrNo()) + "<br>"
            if self.__model.conf_isDry():
                sTitle = "Simulated purge in dry mode"
            else:
                sTitle = "Purged entries"
                pn("Workflows will be reinitialized ...")
                self.sys_server_disconnect()
                self.sys_server_connect()
            self.popup_generic(title="Purge", head=sTitle, info=sInfo, detail=sDetails)

    def worker_finished_delEntry(self, result):
        """
        Worker Delete Entries in Table
        """

        pn = self.ui_printConsoleNormal

        self.ui_hideProgress()
        self.ui_setWorkflow()

        # Parameter

        sTab = result.getData("tab")
        iIdStart = result.getData("id-start")
        iDelTotal = result.getData("del-total")
        iDelErr = result.getData("del-err")
        iDelOk = result.getData("del-ok")
        if iDelTotal is None:
            iDelTotal = 0
        if iDelOk is None:
            iDelOk = 0
        if iDelErr is None:
            iDelErr= 0

        # Details
        sDetails = ""
        if result.isErr():
            sDetails += result.getErr(short=True)

        # Dialog0
        sTitle = "Delete entries"
        sInfo = "<b>Deleted entries</b>: " + str(iDelOk) + "/" + str(iDelTotal) + "<br>"
        if result.isErr():
            sInfo += "<b>Errors occurred</b>: " + str(iDelErr) + "<br>"
        self.popup_generic(title="Delete", head=sTitle, info=sInfo, detail=sDetails)

        # Reconnect to Server?
        if iDelOk > 0:
            pn("Workflows will be reinitialized ...")
            # Erneuere Status fuer die GUI
            if self.__map.all_isTableTaxon(sTab):
                sSystem = "wizard"
            else:
                sSystem = "lookup"
            # Starte Initialisierung im Thread Container
            self.worker_start(Worker_ServerReconnect(self.__model, start=sSystem, status=sSystem), self.worker_finished_serverConnection)

        # # Tabelle wieder anzeigen mit der Id des veraenderten Eintrages
        # self.tableShow(sTab, id=iIdStart)


    def worker_finished_ingestion(self, result):
        """
        Worker Ingestion ist beended
        """

        self.ui_hideProgress()
        if result is not None:

            # Title/File/Ingest Id
            iIngestId = result.getData("ingest-ingest-id")

            # Info: Gesamt
            sInfoGesamt = ""
            sLog = result.getData("log")

            # Info: Error, Warnung, Info
            sInfoErrWarnInfo = ""
            if result.isErr():
                sInfoErrWarnInfo += "Errors: " + str(result.getErrNo()) + "x\n"
            if result.isWarn():
                sInfoErrWarnInfo += "Warnings: " + str(result.getWarnNo()) + "x\n"
            if result.isInfo():
                sInfoErrWarnInfo += "Information: " + str(result.getInfoNo()) + "x\n"

            if result.isOk():

                # Info: Neue Einträge
                if self.__model.conf_isDry():
                    sInfoNewRecords = "Simulated entries in database"
                else:
                    sInfoNewRecords = "Generated entries in database"
                sInfoNewRecords += "<p>"
                for k, v in result.getData("ingest-counter-new").items():
                    if v > 0:
                        sInfoNewRecords += k + ": " + str(v) + "x<br>"
                sInfoGesamt += sInfoNewRecords

                # Info: Updated Records
                bInfoVorhanden = False
                if self.__model.conf_isDry():
                    sInfoExtendedRecords = "Simulated extensions of database entries"
                else:
                    sInfoExtendedRecords = "Extensions of database entries"
                sInfoExtendedRecords += "<p>"
                for k, v in result.getData("ingest-counter-update").items():
                    if v > 0:
                        bInfoVorhanden = True
                        sInfoExtendedRecords += "" + k + ": " + str(v) + "x<br>"
                if bInfoVorhanden:
                    sInfoGesamt += sInfoExtendedRecords

                # Info: Dependant ingests
                bInfoVorhanden = False
                if self.__model.conf_isDry():
                    sInfoDependentIngests = "Ingest would depend on other ingests:"
                else:
                    sInfoDependentIngests = "Ingest depends on other ingests:"
                sInfoDependentIngests += "<ul>"
                lstTmpIngests = result.getData("ingest-ingest-depend")
                if len(lstTmpIngests) != 0:
                    bInfoVorhanden = True
                    for it in result.getData("ingest-ingest-depend"):
                        sInfoDependentIngests += "<li><i>" + str(it) + "</i></li>"
                sInfoDependentIngests += "</ul>"
                if bInfoVorhanden:
                    sInfoGesamt += sInfoDependentIngests

                # Info: Ingest Id
                bInfoVorhanden = False
                sInfoIngestId = ""
                if not self.__model.conf_isDry():
                    sInfoIngestId += "Ingest Id: " + str(iIngestId) + "<br>"
                    bInfoVorhanden = True
                if bInfoVorhanden:
                    sInfoGesamt += sInfoIngestId

                # Workflow neu initialisieren
                if not self.__model.conf_isDry():
                    self.ui_printConsoleNormal("Workflows will be reinitialized ...")
                    self.sys_server_disconnect()
                    self.sys_server_connect()

                # OK
                ant = self.popup_generic(title="Ingestion", head="Ingestion completed " + ("in dry mode" if self.__model.conf_isDry() else ""), info=sInfoGesamt, button_special="Show Protocol")
                if ant == 2:
                    self.popup_text(title="IMPORT PROTOCOL", text=sLog, save=True)
            else:
                # Fehler
                ant = self.popup_generic(title="Ingestion", head="Ingestion failed " + ("in dry mode" if self.__model.conf_isDry() else ""), info=sInfoErrWarnInfo, button_special="Show Protocol")
                if ant == 2:
                    self.popup_text(title="IMPORT PROTOCOL", text=sLog, save=True)

    def worker_finished_showTable(self, result):
        """
        Eine Tabelle soll in einem Fenster angezeigt werden
        """
        pn = self.ui_printConsoleNormal

        self.ui_hideProgress()
        if result is not None:
            if result.isOk():

                # Hole Einstellungen

                """
                result kwargs:
                      ingId (int): IngestionId - Tabelle is eine ist eine Ingesttabelle
                      edit (bool): Soll die Tabelle editierbar sein 
                      noids (bool): Die Ids wurden entfernt
                      human (bool): Die DAten wurden humanisiert
                      save (bool): Soll die Tabelle speicherbar sein 
                      gold (bool): Wurden nur die gold/not-gold Einträge angezeigt
                      used (bool): Wurden nur die used/not-used Einträge angezeigt
                      gold_edit (bool): Edit Gold Status Bulk
                      doi (bool): DOI ok
                      doi_pangaea (bool): DOI führt zu PANGAEA
                      access (bool): Zugang gewährleistet (irgendwie)
                      gear_double (bool): Gear Double Config
                      id (int): Auf die Zeile mit dieser id soll fokussiert werden 
                """
                bEdit = result.getData("edit")
                bEdit = True if bEdit is True else False  # Mache die Variable definitiv boolsch
                bDel = result.getData("del_entry")
                bDel = True if bDel is True else False
                bNoId = result.getData("noids")
                bNoId = True if bNoId is True else False
                bSave = result.getData("save")
                bSave = True if bSave is True else False
                iIngId = result.getData("ingId")
                sTabName = result.getData("tabName")
                iFokusId = result.getData("id")
                sDiaTitle = result.getData("diaTitle")
                bGold = result.getData("gold")
                bUsed = result.getData("used")
                bAccess = result.getData("access")
                bGear_double = result.getData("gear_double")
                bDoi = result.getData("doi")
                bDoiPangaea = result.getData("doi_pangaea")
                bGoldEdit = result.getData("gold_edit")
                multisheet = result.getData("multisheet")

                # Checke Fehler in Einstellungen
                if bNoId and iFokusId is not None:
                    pn("Uupps: Internal error while showing a Table. Ids are removed and I shall fokus to an id.")
                # Öffne Tabelle
                if multisheet.isOk():
                    # Es kam eine Tabelle an (kann auch leer sein)
                    self.ui_setStatusbarText("Database has answered: " + str(multisheet.getRows()) + " records")
                    self.table_showGeneric(multisheet,
                                           ingId=iIngId, edit=bEdit, del_entry=bDel, diaTitle=sDiaTitle,
                                           rangeIdx=0, save=bSave, id=iFokusId, gold=bGold, used=bUsed, gold_edit=bGoldEdit, access=bAccess, doi=bDoi, doi_pangaea=bDoiPangaea, gear_double=bGear_double)
                else:
                    # Es kam keine Tabelle an
                    self.popup_generic(title="Table", head=result.getName(), info="The server answerd with nothing", detail=result.getErrWarnInfoString())
            else:
                # Fehler
                sDetail = "*** Error\n" + result.getErrWarnInfoString(short=True) + "\n"
                self.popup_error(result.getName(), "An error occurred.", sDetail)

    def worker_finished_showUpdateCheck(self, res):
        """
        Das Ergebnis der Updatesuche wird angezeigt
        """
        self.ui_hideProgress()

        if res.isErr():
            self.popup_error("Updating is not possible", "Are you connected to the AWI intranet?...", res.getErrWarnInfoString(short=True))
        else:
            bUpdate = res.getData("update")
            bIncompatabile = res.getData("update-incomp")
            sIncompatabileVersion = str(res.getData("update-incomp-version"))
            if bUpdate:
                if bIncompatabile :
                    ant = self.popup_generic(title="CAUTION", head="Incompatability Warning",
                                           info="The Version on server is breaking compatability with your running databases.<br>If you continue downloading, all your databases are not accessable any more.<br>You have to recreate them and reingest all your data.<p><b>Do you really want to continue?</b>", button_yes="Yes", button_no="No")
                    if ant == 0:
                        return
                # Es gibt ein Update
                # Dialog Update
                dia = DialogUpdate()
                dia.version_installed.setText(res.getData("version-installed"))
                dia.version_server.setText(res.getData("version-server"))
                dia.update_level.setText(res.getData("update-level"))
                dia.changelog.insertPlainText(res.getData("changelog-server"))
                if bIncompatabile:
                    dia.warnings.insertPlainText("Compatibilty broken since version " + sIncompatabileVersion+"!")
                else:
                    dia.warnings.insertPlainText("all right")
                dia.changelog.moveCursor(QTextCursor.Start)
                self.ui_centerWidget(dia)
                ant = dia.exec_()
                if ant == 0:
                    return
                # Update durchführen
                # Worker starten
                self.worker_start(Worker_DownloadUpdate(self.__model), self.worker_finished_downloadUpdate)
            else:
                # Es gibt kein Update
                self.popup_generic(title="Update", head="CRITTERBASE is up to date", info=res.getData("update-info"))

    def worker_finished_downloadUpdate(self, res):
        """
        Download ist fertig -> Installation
        """
        pn = self.ui_printConsole
        self.ui_hideProgress()

        if res.isErr():
            self.popup_error("Update is not possible", res.getErrWarnInfoString(short=True))
        else:
            pn("* Installation")
            sSys = res.getData("sys")
            sSrc = res.getData("src")
            sTarget = res.getData("target")
            sRestart = res.getData("restart")
            pn("System:  " + sSys)
            pn("Source:  " + sSrc)
            pn("Target:  " + sTarget)
            pn("Restart: " + sRestart)
            ant = self.popup_generic(title="Installation", head="The update has been downloaded", info="Everything is prepared for installation.\nDo you want to start the installation now?", button_yes="Yes", button_no="No")
            if ant == 0:
                pn("Installation aborted ...")
                return

            # Installation ---
            pn("Copying files:")
            files = copy_tree(sSrc, sTarget)
            pn(PrettyText.lst2strR(files))
            pn("Installation is finished")
            pn("Restart: " + sRestart)
            # Timer Info
            self.popup_infoTimer("Update is installed", "App will be restarted automatically.", 15)

            # App beenden
            self.main_beenden(True)

    def worker_finished_showPopupInfo(self, result):
        """
        Ein Popup wird im Fenster angezeigt
        """
        self.ui_hideProgress()
        if result is not None:
            sTitle = str(result.getData("title"))
            sText = str(result.getData("text"))
            sDetails = result.getData("details")
            if result.isOk():
                # Ok
                self.popup_generic(title="Info", head=sTitle, info=sText, detail=sDetails)
            else:
                # Fehler
                sDetail = "*** Error\n" + result.getErrWarnInfoString(short=True) + "\n"
                self.popup_error(sTitle, "An error occurred.", sDetail)

    def worker_finished_checkDoi(self, result):
        """
        DOIS wurden gecheckt
        """
        self.ui_hideProgress()
        if result is not None:
            if result.isOk():
                # Hole Einstellungen
                sLog = result.getData("log")
                dicErr = result.getData("err")
                dicOk = result.getData("ok")
                dicNone = result.getData("no")

                dicDOIs_clustered_byPublisher = result.getData("cluster-journal")
                dicDatasets_clustered_byDoi = result.getData("cluster-doi")
                iPangaea_nr = result.getData("cluster-pangaea")
                setPangaea_ids = result.getData("cluster-pangaea-ids")
                iPangaea_proz = result.getData("cluster-pangaea-proz")
                setNotPangaea_ids = result.getData("cluster-not-pangaea-ids")
                iNotPangaea_nr = result.getData("cluster-not-pangaea")
                iNotPangaea_proz = result.getData("cluster-not-pangaea-proz")

                packDoi = Package()
                packDoi.add("log", sLog)
                packDoi.add("err", dicErr)
                packDoi.add("ok", dicOk)
                packDoi.add("no", dicNone)
                packDoi.add("cluster-journal", dicDOIs_clustered_byPublisher)
                packDoi.add("cluster-doi", dicDatasets_clustered_byDoi)
                packDoi.add("cluster-pangaea", iPangaea_nr)
                packDoi.add("cluster-pangaea-ids", setPangaea_ids)
                packDoi.add("cluster-pangaea-proz", iPangaea_proz)
                packDoi.add("cluster-not-pangaea-ids", setNotPangaea_ids)
                packDoi.add("cluster-not-pangaea", iNotPangaea_nr)
                packDoi.add("cluster-not-pangaea-proz", iNotPangaea_proz)

                self.__model.memory_save("doi", packDoi)
                self.status_refresh_lookup()
                # Öffne Tabelle
                if sLog is not None:
                    self.popup_text(title="Report on: DOI status", text=sLog, browser=True, html=True, save=True)
                else:
                    # Es kam keine Tabelle an
                    self.popup_generic(title="DOI", head=result.getName(), info="The server answerd with nothing", detail=result.getErrWarnInfoString())
            else:
                # Fehler
                sDetail = "*** Error\n" + result.getErrWarnInfoString(short=True) + "\n"
                self.popup_error(result.getName(), "An error occurred.", sDetail)

    def worker_finished_batchBot(self, result):
        self.ui_hideProgress()
        self.ui_hideBatch()
        if result is not None:
            if result.isOk():
                # Hole Einstellungen
                lstFiles = result.getData("files")
                sLogDir = result.getData("dirlog")
                sLog = result.getData("log")
                sTimeStamp = result.getData("timestamp")
                if sLog is not None:
                    # Dialog
                    dia = QDialog()
                    dia.setWindowTitle("BatchLog Results")
                    dia.setWindowModality(Qt.ApplicationModal)
                    dia.setWindowFlags(Qt.Dialog | Qt.CustomizeWindowHint | Qt.WindowTitleHint | Qt.WindowCloseButtonHint)
                    vbox = QVBoxLayout(dia)
                    vbox.setContentsMargins(4, 4, 4, 4)
                    vbox.setSpacing(6)
                    text = QTextEdit()
                    font = QFont()
                    font.setFamily("Courier")
                    font.setPointSize(10)
                    text.setFont(font)
                    text.setLineWrapMode(QTextEdit.NoWrap)
                    text.setReadOnly(True)
                    text.setText(sLog)
                    vbox.addWidget(text)
                    buttonBox = QDialogButtonBox()
                    buttonBox = QDialogButtonBox()
                    # OK
                    btnOk = QPushButton()
                    btnOk.setText("OK")
                    btnOk.clicked.connect(lambda: QDialog.done(dia, 0))
                    buttonBox.accepted.connect(dia.accept)
                    buttonBox.addButton(btnOk, QDialogButtonBox.NoRole)
                    vbox.addWidget(buttonBox)
                    # Logs
                    btnLog = QPushButton()
                    btnLog.setText("All logs")
                    btnLog.setToolTip("Open log directory")
                    btnLog.clicked.connect(lambda: QDialog.done(dia, 1))
                    buttonBox.accepted.connect(dia.accept)
                    buttonBox.addButton(btnLog, QDialogButtonBox.NoRole)
                    vbox.addWidget(buttonBox)
                    dia.resize(800, 600)
                    # Start
                    ant = dia.exec_()
                    if ant == 1:
                        sPath = 'file://' + str(sLogDir)
                        webbrowser.open('file://' + str(sLogDir))
                else:
                    # Es kam keine Tabelle an
                    self.popup_generic(title="BatchBot", head=result.getName(), info="The server answerd with nothing", detail=result.getErrWarnInfoString())
            else:
                # Fehler
                sDetail = "*** Error\n" + result.getErrWarnInfoString(short=True) + "\n"
                self.popup_error(result.getName(), "An error occurred.", sDetail)

            self.ui_printConsoleNormal("Workflows will be reinitialized ...")
            self.sys_server_disconnect()
            self.sys_server_connect()

    def worker_finished_fao(self, result):
        """
        FAO Operation beendet
        """
        self.ui_hideProgress()
        if result is not None:
            if result.isOk():
                # Hole Einstellungen
                bServerReconnect = result.getData("server_reconnect")
                sTitle = str(result.getData("title"))
                sText = str(result.getData("text"))
                bHtml = result.getData("html")
                bBrowser = result.getData("browser")
                bExport = result.getData("export")
                bImport = result.getData("import")
                sFile = result.getData("file")
                sFile_path = FileTools.getFilePath(sFile)
                sFile_name = FileTools.getFileWithoutPath(sFile)
                sLog = result.getData("log")
                bDontShow = result.getData("dont_show") if result.getData("dont_show") is not None else False
                if not bDontShow:
                    bTitleCapital = result.getData("title_capitalize")
                    if bHtml is None:
                        bHtml = False
                    if bImport is True:
                        # Import from file
                        sInfo = '<html><head/><body><p>FAO regions are imported:<br><a href="file:///' + str(
                            sFile_path) + '"><span style=" text-decoration: underline; color:#0000ff;">' + str(
                            sFile_name) + '</span></a></p></body></html>'
                        self.popup_generic(title="FAO regions", head="Import FAO regions from file", info=sInfo)
                    elif bExport is True:
                        # Export in file
                        sInfo = '<html><head/><body><p>FAO regions are saved:<br><a href="file:///' + str(sFile_path) + '"><span style=" text-decoration: underline; color:#0000ff;">'+str(sFile_name)+'</span></a></p></body></html>'
                        self.popup_generic(title="FAO regions", head="Export FAO regions to file", info=sInfo)
                    else:
                        # Rest
                        # Öffne Text
                        if sText is not None:
                            self.popup_text(title=sTitle, text=sText, save=True, browser=bBrowser, html=bHtml, titel_capitalize=bTitleCapital)
                        else:
                            # Es kam nichts an
                            self.popup_generic(title="FAO", head=result.getName(), info="The server answerd with nothing", detail=result.getErrWarnInfoString())

                    if bServerReconnect is True:
                        self.ui_printConsoleNormal("Workflows will be reinitialized ...")
                        self.sys_server_disconnect()
                        self.sys_server_connect()

            else:
                # Fehler
                sDetail = "*** Error\n" + result.getErrWarnInfoString(short=True) + "\n"
                self.popup_error("FAO Regions", "An error occurred.", sDetail)


    def worker_finished_showText(self, result):
        """
        Ein Text wird in einem Fenster angezeigt
        """
        self.ui_hideProgress()
        if result is not None:
            if result.isOk():
                # Hole Einstellungen
                sTitle = str(result.getData("title"))
                sText = str(result.getData("text"))
                bHtml = result.getData("html")
                bBrowser = result.getData("browser")
                bDontShow = result.getData("dont_show") if result.getData("dont_show") is not None else False
                if not bDontShow:
                    bTitleCapital = result.getData("title_capitalize")
                    if bHtml is None:
                        bHtml = False
                    # Öffne Tabelle
                    if sText is not None:
                        self.popup_text(title=sTitle, text=sText, save=True, browser=bBrowser, html=bHtml, titel_capitalize=bTitleCapital)
                    else:
                        # Es kam keine Tabelle an
                        self.popup_generic(title="Info", head=result.getName(), info="The server answerd with nothing", detail=result.getErrWarnInfoString())
            else:
                # Fehler
                sDetail = "*** Error\n" + result.getErrWarnInfoString(short=True) + "\n"
                self.popup_error(result.getName(), "An error occurred.", sDetail)

    def worker_finished_virtualAssistant_taxaReport(self, result):
        """
        Taxon Spotter
        """
        self.ui_hideProgress()
        self.status_refresh_taxonWizard()
        if result is not None:
            if result.getData("abort") is True:
                # User Abort
                self.popup_generic(title="TaxonSpotter", head="TaxonSpotter Aborted", info="TaxonSpotter was aborted by user.")
            elif result.isErr():
                # Fehler
                self.popup_generic(title="TaxonSpotter", head="Errors occured while TaxaSpotter", info="Please check the following log text.", detail=result.getErr() if result.isErr() else None)

            # Öffne Tabelle
            sTitle = result.getData("title")
            sText = result.getData("text")
            bHtml = result.getData("html")

            if sText is not None and sTitle is not None:
                self.popup_text(title=sTitle, text=sText, save=True, browser=bHtml)
            else:
                # Es kam keine Tabelle an
                self.popup_generic(title="Info", head=result.getName(), info="The server answerd with nothing", detail=result.getErrWarnInfoString())

    def worker_finished_virtualAssistant_taxaPolish(self, result):
        """
        Taxa Polish
        """
        self.ui_hideProgress()
        self.status_refresh_taxonWizard()
        if result is not None:
            if result.getData("abort") is True:
                # User Abort
                self.popup_generic(title="TaxaPolishing", head="TaxaPolishing Aborted", info="TaxaPolishing was aborted by user.")
            elif result.isErr():
                # Fehler
                self.popup_generic(title="TaxaPolishing", head="Errors occured while TaxaPolishing", info="Please check the following log text.", detail=result.getErr() if result.isErr() else None)

            # Öffne Tabelle
            sTitle = result.getData("title")
            sText = result.getData("text")
            if sText is not None and sTitle is not None:
                self.popup_text(title=sTitle, text=sText, save=True)
            else:
                # Es kam keine Tabelle an
                self.popup_generic(title="Info", head=result.getName(), info="The server answerd with nothing", detail=result.getErrWarnInfoString())

    def worker_finished_biotaControl(self, result):
        """
        Biota Kontrolle durchgeführt
        """
        self.ui_hideProgress()
        self.status_refresh_dataPatrol()
        if result is not None:
            if result.getData("abort") is True:
                # User Abort
                self.popup_generic(title="BiotaControl", head="BiotaControl aborted", info="Update was aborted by user. Nothing has been written to database. You might want to check the following log text.")
            elif result.isErr():
                # Fehler
                self.popup_generic(title="BiotaControl", head="Errors occured", info="Please check the following log text", detail=result.getErrWarnInfoString(short=True) if result.isErr() else None)

            # Öffne Tabelle
            sTitle = result.getData("title")
            sText = result.getData("text")
            if sText is not None and sTitle is not None:
                self.popup_text(title=sTitle, text=sText, browser=True, html=True, save=True)
            else:
                # Es kam keine Tabelle an
                self.popup_generic(title="Info", head="BiotaCotrol", info="The server answerd with nothing", detail=result.getErrWarnInfoString())

    def worker_finished_updateTaxonList(self, result):
        """
        Taxon System Update durchgeführt
        """
        self.ui_hideProgress()
        self.status_refresh_taxonWizard()
        if result is not None:
            if result.getData("abort") is True:
                # User Abort
                self.popup_generic(title="TaxonWizard", head="TaxonWizard aborted", info="Update was aborted by user. Nothing has been written to database. You might want to check the following log text.")
            elif result.isErr():
                # Fehler
                self.popup_generic(title="TaxonWizard", head="Errors occured - nothing was written to database", info="Please check the following log text, fix the errors and rerun the taxon system update. Next run will be way faster, because all WoRMS requests are locally puffered.", detail=result.getErrWarnInfoString(short=True) if result.isErr() else None)

            # Öffne Tabelle
            sTitle = result.getData("title")
            sText = result.getData("text")
            if sText is not None and sTitle is not None:
                self.popup_text(title=sTitle, text=sText, html=True, browser=True, save=True)
            else:
                # Es kam keine Tabelle an
                self.popup_generic(title="TaxonWizard", head=result.getName(), info="The server answerd with nothing", detail=result.getErrWarnInfoString())

            # Workflows neu starten
            if not self.__model.conf_isDry():
                self.ui_printConsoleNormal("Workflow has to be restarted ...")
                if self.sys_isWorkerRunning():
                    self.popup_infoTimer("There is a process running", "Please reconnect to the server manually when the process has finished.")
                    return
                self.sys_server_disconnect()
                self.sys_server_connect()

    def worker_finished_antiageMyWoRMS(self, result):
        """
        Taxon System Puffer Update durchgeführt
        """
        self.ui_hideProgress()
        self.status_refresh_taxonWizard()
        if result is not None:
            if result.isErr():
                # Fehler
                self.popup_generic(title="MyWoRMS", head="Errors occured", info="Please check the following log text", detail=result.getResult() if result.isErr() else None)
            self.status_refresh_myWorms()
            self.ui_setStatusbarText("WoRMS puffer loaded from disk")
            # Öffne Tabelle
            sTitle = result.getData("title")
            sText = result.getData("text")
            if sText is not None and sTitle is not None:
                self.popup_text(title=sTitle, text=sText)
            else:
                # Es kam keine Tabelle an
                self.popup_generic(title="MyWoRMS", head=result.getName(), info="Buffer Update answered nothing", detail=result.getErrWarnInfoString())

    def worker_finished_saveDataToFile(self, result):
        """
        Daten wurden in Datei gespeichert
        """
        self.ui_hideProgress()
        if result is not None:
            if result.isOk():
                self.popup_generic(title="Save file", head=result.getName(), info="Data was successfully saved to file", detail=result.getErrWarnInfoString())
            else:
                # Fehler
                self.popup_error("Error: " + result.getName(), "Data was not saved to file", result.getErrWarnInfoString())

    def worker_finished_saveMapToFile(self, result):
        """
        Map wurde in Datei gespeichert
        """
        self.ui_hideProgress()
        if result is not None:
            if result.isOk():
                self.popup_generic(title="Map", head="Map was saved to file", info="Map was successfully saved to file", detail=result.getErrWarnInfoString())
            else:
                # Fehler
                self.popup_error("Error: " + result.getName(), "Map was not saved to file", result.getErrWarnInfoString())

    def worker_finished_updateVis(self, result):
        """
        Der Visual Explorer wurde geupdated
        """

        self.ui_hideProgress()
        self.explorer_selectorEnable(True)

        # --- Prüfe auf Force Mode
        bSamplesWarn = result.getData("samples_warn")
        bSamplesWarn = bSamplesWarn if bSamplesWarn is not None else False

        # Evaluiere Force Draw Modi
        if bSamplesWarn:
            # Es waren zu viele Samples
            if self.__bMapForceDraw:
                # Force
                self.buttonDrawMap.setStyleSheet('QPushButton {color: red;}')
                self.buttonDrawMap.setText("redraw")
                self.buttonDrawMap.setToolTip("Force redraw")
            else:
                # Kein Force
                self.buttonDrawMap.setStyleSheet('QPushButton {color: red;}')
                self.buttonDrawMap.setText("draw")
                self.buttonDrawMap.setToolTip("Force draw")
                self.__bMapForceDraw = True  # Ab jetzt Force
            # Einmalige Meldung
            if self.__bMapForceDrawFirstTimeQuestion:
                self.popup_generic(title="One Time Information", head="There are far too many markers to draw.", info="Automatic drawing disabled, force to draw or adjust the selection.")
                self.__bMapForceDrawFirstTimeQuestion = False
        else:
            # Es waren nicht zu viele Samples
            self.buttonDrawMap.setStyleSheet(None)
            self.buttonDrawMap.setText("redraw")
            self.buttonDrawMap.setToolTip(None)
            self.__bMapForceDraw = False

        if result.isErr():
            self.ui_printConsoleNormal(result.getErrWarnInfoString(short=True))
            self.explorer_reset_visual()
        else:
            # Selector einstellen
            self.explorer_setSelector()
            # Statistik einstellen
            self.status_update_selector_statistics()
            # Browser reload
            self.explorer_browser_reloadMap()
            self.ui_setStatusbarText("Refreshing map... just a few seconds ...")

    def worker_finished_showActionLogContent(self, result):
        """
        Zeige Action Log Inhalt an
        """
        self.ui_hideProgress()
        if result is not None:
            if result.isOk():
                sText = str(result.getData("content"))
                self.tableLogContent.setText(sText)
            else:
                # Fehler
                sDetail = "*** Error\n" + result.getResult() + "\n"
                self.popup_error(result.getName(), "An error occurred.", sDetail)


    # --- Observer

    def sys_connectToObserver(self, observer):
        """
        Verbinde Observer mit View

        Args:
            observer (gui.observer.Observer): Observer
        """

        # Print to console
        observer.signalPrint.connect(self.ui_printConsole)
        # Progress
        observer.signalProgressHide.connect(self.ui_hideProgress)
        observer.signalProgressShow.connect(self.ui_showProgress)
        observer.signalProgressSetMinMax.connect(self.ui_setProgressbarMinMax)
        observer.signalProgressSetWert.connect(self.ui_setProgressbarWert)
        observer.signalProgressSetText.connect(self.ui_setProgressbarText)
        observer.signalProgressShowText.connect(self.ui_setProgressbarTextAnzeigen)
        observer.signalBatch.connect(self.ui_setBatch)
        observer.signalProgressErrWarnInfo.connect(self.ui_setProgressbarErrWarnInfo)
        # Eingabe
        observer.signalInputComboMulti.connect(self.ui_userInputComboMulti)
        # Workflowstatus
        observer.signalWorkflowSet.connect(self.ui_setWorkflow__wrapper)
        # Status
        observer.signalStatusbarSetText.connect(self.ui_setStatusbarText)
        # MyWorms
        observer.signalUpdateMyworms.connect(self.status_refresh_myWorms)
        # SysMon
        observer.signalUpdateSysMon.connect(self.status_refresh_sysMon)
        # IngestTable (BatchBot)
        observer.signalUpdateIngestTable.connect(self.status_update_batchBot)

    # --- GUI-Funktionen
    def ui_userInputComboMulti(self, sTitel, odicChoices):
        answerInput = self.__model.ui_getSmartprinter().setInputAnswer
        dia = DialogGenericComboChoice(self.__model, sTitel, odicChoices)
        ant = dia.exec()
        if ant == 0:
            dicUserChoice = {}
        else:
            dicUserChoice = dia.getSelection()
        answerInput(dicUserChoice)

    def ui_printConsole(self, text, appendix="\n"):
        """
        Print in Konsole
        Args:
            text (str): Text
            appendix (str): Appendig
        """

        self.textBox.insertPlainText(text + appendix)
        self.textBox.moveCursor(QTextCursor.End)

    def ui_printConsoleNormal(self, text, appendix="\n"):
        self.ui_printConsole(text, appendix)

    def ui_printConsoleVerbose(self, text, appendix="\n"):
        if self.__model.conf_isVerbose() or self.__model.conf_isDebug():
            self.ui_printConsole(text, appendix)

    def ui_printConsoleDebug(self, text, appendix="\n"):
        if self.__model.conf_isDebug():
            self.ui_printConsole(text, appendix)

    def ui_showProgress(self, bAbort=False, bPulse=True):
        """
        Zeige Progressbar

        Args:
            bAbort (bool): Abort-Button anzeigen
            bPulse (bool): Pulsieren
        """
        if bAbort:
            self.progressAbort.show()
        else:
            self.progressAbort.hide()
        # self.progressErr.show()
        # self.progressWarn.show()
        # self.progressInfo.show()
        #self.setWorkflow(top=sWorkflowText)
        self.progressBar.reset()
        if bPulse:
            self.progressBar.setMinimum(0)
            self.progressBar.setMaximum(0)
        self.progressBar.show()
        self.ui_giveQtTimeToUpdate()

    def ui_showBatch(self):
        #QProgressBar.setMaximumHeight()
        self.progressBar.setMaximumHeight(13)
        self.batch.show()

    def ui_hideBatch(self):
        self.batch.hide()
        self.progressBar.setMaximumHeight(26)

    def ui_hideProgress(self):
        self.progressBar.hide()
        self.labelExpanderWorkflowStep.setText("")
        self.labelExpanderWorkflowInfo.setText("")
        self.progressBar.reset()
        self.progressAbort.hide()
        self.progressErr.hide()
        self.progressWarn.hide()
        self.progressInfo.hide()

    def ui_setProgressbarMinMax(self, min=0, max=0):
        """
        Min Max Progressbar

        Args:
            min (int): Min
            max (max): Max
        """

        self.progressBar.setMinimum(min)
        self.progressBar.setMaximum(max)


    def ui_setProgressbarWert(self, wert, **kwargs):
        """
        Setze Progresbar

        Args:
            wert (int): Wert
        """
        # Wert <0 -> Pulse
        if wert < 0:
            self.progressBar.setMinimum(0)
            self.progressBar.setMaximum(0)
        else:
            self.progressBar.setValue(wert)
        #self.guiRefreshSysMon()

    def ui_setProgressbarTextAnzeigen(self, bBool=True):
        """
        Text bei Progressbar anzeigen?

        Args:
            bBool (bool): Text anzeigen
        """

        self.progressBar.setTextVisible(bBool)

    def ui_setProgressbarText(self, text):
        """
        Progressbar Text

        Args:
            text (str): Text
        """

        if text != "":
            self.progressBar.setFormat(text)

    def ui_setProgressbarErrWarnInfo(self, iErr, iWarn, iInfo):
        """
        Progressbar Error Warn Info

        Args:
            iErr
            iWarn
            iInfo
        """
        lst = [(iErr, self.progressErr, QColor("#ff9494")),(iWarn, self.progressWarn, QColor(4, 100, 69)),(iInfo, self.progressInfo, QColor(4, 159, 74))]
        for iVal, widget, color in lst:
            widget.show()
            if iVal in (None, 0):
                sVal = "-"
                widget.setPalette(QPalette())
            else:
                palette = QPalette(widget.palette())
                palette.setColor(QPalette.Text, color)
                widget.setPalette(palette)
                sVal = ">999" if iVal > 999 else str(iVal)
            widget.setText(sVal)

    def ui_setWorkflow__wrapper(self, sTop, sSub, iStepAkt, iStepMax):
        self.ui_setWorkflow(top=sTop, sub=sSub, steps=(iStepAkt, iStepMax))

    def ui_setWorkflow(self, **kwargs):
        """
        WorkflowMultisheetTesting Status Text setzen

        Args:
            sTop (str): Text
            sSub (str): Text
            iStepAkt (int): Aktueller Schritt
            iStepMax (int): Maximaler Schritt
        """
        sTop = kwargs["top"] if "top" in kwargs else None
        sSub = kwargs["sub"] if "sub" in kwargs else None
        tpSteps = kwargs["steps"] if "steps" in kwargs else None
        try:
            iStepAkt = tpSteps[0]
            iStepMax = tpSteps[1]
        except:
            iStepAkt = 0
            iStepMax = 0
        # Steps
        if iStepAkt not in [0, None] and iStepMax not in [0, None]:
            sSteps = " [" + str(iStepAkt) + "/" + str(iStepMax) + "]"
        else:
            sSteps= ""
        # Top Text
        if sTop in [None, ""]:
            self.labelExpanderWorkflowStep.setText("Working")
        else:
            self.labelExpanderWorkflowStep.setText(str(sTop))
        # Sub Text
        if sSub in [None, ""]:
            self.labelExpanderWorkflowInfo.setText("...")
        else:
            self.labelExpanderWorkflowInfo.setText(str(sSub)+sSteps)

    def ui_setStatusbarText(self, sText="", **kwargs):
        """
        Text im Statusbar setzen

        Args:
            sText (str): Text
        KWargs:
            time (int): Anzeigezeit in Ms
        """

        iTime = kwargs["time"] if "time" in kwargs else 2000
        try:
            iTime = int(iTime)
        except:
            iTIme = 2000
        if sText == "":
            self.__statusbar.clearMessage()
        else:
            self.__statusbar.showMessage("  " + sText, iTime)

    def tab_switchToIngest(self):
        """
        Switch Tab to Ingest
        """

        self.tabIngViz.setCurrentIndex(0)

    def tab_switchToViz(self):
        """
        Switch Tab to Visualisation
        """

        self.tabMain.setCurrentIndex(1)

    # --- Server-Kontrolle

    def sys_server_connect(self):
        """
        Verbinde mit dem Server
        """
        if self.sys_isWorkerRunning(popup=True, status=True):
            return False
        pn = self.ui_printConsoleNormal
        pv = self.ui_printConsoleVerbose
        status = self.ui_setStatusbarText

        if self.__model.subsystem_getDb().account_single_getPart("host") == "" or \
                        self.__model.subsystem_getDb().account_single_getPart("port") == "" or \
                        self.__model.subsystem_getDb().account_single_getPart("user") == "" or \
                        self.__model.subsystem_getDb().account_single_getPart("db") == "" or \
                        self.__model.subsystem_getDb().account_single_getPart("schema") == "":
            self.popup_error("Problems occured while conneting to database", "Credentials to connect to database server are incomplete.")
            self.explorer_selectorEnable(False)
            return False

        if self.__model.subsystem_getDb().account_single_getPart("pw") == "":
            pv("Db Password missing -> getting password manually from user")
            self.getPassword()  # Passwort manuell holen

        # Starte Initialisierung im Thread Container
        self.worker_start(Worker_ServerReconnect(self.__model), self.worker_finished_serverConnection, server=True, comment=False, ingdep=False)
        return True

    def sys_server_disconnect(self):
        """
        Trenne Verbindung mit dem Server
        """

        if self.sys_isWorkerRunning(popup=True, status=True):
            return False

        pn = self.ui_printConsoleNormal
        pv = self.ui_printConsoleVerbose
        status = self.ui_setStatusbarText

        pv("* Disconneting from server")
        self.__model.subsystem_getConnector().disconnect()
        #self.__model.newWorkflow()
        pv("Connection is closed")
        status("Connection is closed")

        self.__model.memory_del("doi")

        self.status_refresh_all()
        self.exporer_selectorClear()

        #self.build_browser_emptyMap()
        self.explorer_browser_reloadMap()
        self.status_clean_ingestDepenencies()
        #self.guiActionLogInfoClear()
        self.status_clean_selector_statistics()
        self.status_clean_taxonWizard()
        self.explorer_selectorEnable(False)
        return True

    def ui_serverConnectionToggle(self):
        """
        Verbindung zum Server herstellen/unterbrechen
        """

        if self.sys_isWorkerRunning(popup=True):
            return

        if self.__model.subsystem_getConnector().db_isConnected():
            self.sys_server_disconnect()
        else:
            self.sys_server_connect()

    # --- GUI-Administration

    def sys_isPsqlCommandLineToolsWorking(self):
        """
        Teste ob PostgreSQL Command line tools alle laufen
        """
        if not self.__model.core_isAdminPsqlWorking() or not self.__model.core_isAdminPgdumpWorking():
            sTitle = "Problems with PostgreSQL tools"
            sText = "Some PostgreSQL command line tools are not running."
            sDetail = ""
            sDetail += "psql: " + ("ok" if self.__model.core_isAdminPsqlWorking() else "fail") + "\n"
            sDetail += "pg_dump: " + ("ok" if self.__model.core_isAdminPgdumpWorking() else "fail") + "\n"
            self.popup_error(sTitle, sText, sDetail)
            return False
        return True

    def ui_centerWidget(self, childWidget, parentWidget = None):
        """
        Zentriere Kindfenster zu View (oder anderem Fenster)

        Args:
            childWidget (PyQt5.QtWidgets.QWidgets): Das zu zentrierende Objekt
            parentWidget PyQt5.QtWidgets.QWidgets: Elternobjekt zu dem es zentriert werden soll
        """

        if childWidget is None:
            return
        if parentWidget is None:
            parentWidget = self
        childWidget.move(parentWidget.frameGeometry().center().x() - int(childWidget.frameGeometry().width() / 2),
                 parentWidget.frameGeometry().center().y() - int(childWidget.frameGeometry().height() / 2))

    def build_browser_emptyMap(self):
        """
        Neue leere Map
        """
        map = folium.Map(no_wrap=True, zoom_start=2, min_zoom=2)
        map.save(self.__model.conf_getMap())

    def build_admin_status_tabellen(self):
        """
        Baue GUI auf: Status-Box aller Tabellen
        - Die Tabellennamen werden aus der WorkflowMultisheetTesting-Defintionsdatei extrahiert
        """

        # Grids (LookUp/Ingest)
        gridLookup = QGridLayout()
        gridLookup.setSizeConstraint(QLayout.SetFixedSize)
        gridLookup.setSpacing(4)
        gridLookup.setContentsMargins(4,4,4,4)
        gridIngest = QGridLayout()
        gridIngest.setSizeConstraint(QLayout.SetFixedSize)
        gridIngest.setSpacing(4)
        gridIngest.setContentsMargins(4, 4, 4, 4)

        # Zähler für Widgets
        # Lookup
        iWidgets_lookup = 0 # Anzahl der aktuell aufgebauten Widgets
        iWidgetLookup_witdh = 10 # Breite eines Widgets
        iWidgetLookup_maxPerCol = 6  # Anzahl der Widgets in einer Spalte
        iColLookup = iRowLookup = 1
        # Ingest
        iWidgets_ingest = 0
        iWidgetIngest_witdh = 2
        iWidgetIngest_perCol = 2
        iColIngest = iRowIngest = 1

        # Ingest
        for sTabName in self.__model.subsystem_getMapper().ingest_getDbTables():
            # Titel
            sTitle = self.__map.ingest_table2sheet(sTabName)
            # Label/Button
            btnOpen = QToolButton()
            btnOpen.setText(sTitle)
            font = btnOpen.font()
            font.setPointSize(10)
            font.setBold(False)
            btnOpen.setFont(font)
            btnOpen.setToolTip("Show content ... ")
            btnOpen.setAutoRaise(True)
            btnOpen.setToolButtonStyle(Qt.ToolButtonTextOnly)
            btnOpen.setFixedWidth(btnOpen.minimumSizeHint().width())
            btnOpen.setFixedHeight(btnOpen.minimumSizeHint().height())

            # Connect
            btnOpen.clicked.connect(partial(self.table_show, sTabName))
            # Ausgabefeld
            ausgabeFeld = QLineEdit()
            ausgabeFeld.setFixedWidth(65)
            ausgabeFeld.setEnabled(True)
            ausgabeFeld.setFrame(True)
            font = ausgabeFeld.font()
            font.setPointSize(10)
            font.setBold(False)
            ausgabeFeld.setFont(font)
            ausgabeFeld.setAlignment(Qt.AlignRight)
            ausgabeFeld.setReadOnly(True)
            self.__dicStatus_widgets_ingest[sTabName] = ausgabeFeld  # Widgets in Dictionary merken
            # Zusammenbauen
            gridIngest.addWidget(btnOpen, iRowIngest, iColIngest + 1)
            gridIngest.addWidget(ausgabeFeld, iRowIngest, iColIngest + 2)
            iColIngest = iColIngest + iWidgetIngest_perCol
            if iColIngest >= (iWidgetIngest_perCol * (iWidgetIngest_witdh)):
                iRowIngest += 1
                iColIngest = 1

        # Lookup
        for sTabName in self.__model.subsystem_getMapper().lookup_getDbTables():
            sTabName_sheet = self.__model.subsystem_getMapper().lookup_table2sheet(sTabName)
            # Label/Button
            sTitle = self.__model.subsystem_getMapper().lookup_table2sheet(sTabName)
            btnOpen = QToolButton()
            bBold = False
            if sTabName == "taxon":
                sTitle = "Taxon List"
                bBold = False
            btnOpen.setText(sTitle)
            font = btnOpen.font()
            font.setPointSize(10)
            font.setBold(bBold)
            btnOpen.setFont(font)
            btnOpen.setToolTip("Show taxon list")
            btnOpen.setAutoRaise(True)
            btnOpen.setToolButtonStyle(Qt.ToolButtonTextOnly)
            btnOpen.setFixedWidth(btnOpen.minimumSizeHint().width())
            btnOpen.setFixedHeight(btnOpen.minimumSizeHint().height())
            # Add
            btnAdd = QToolButton()
            btnAdd.setIcon(QIcon(os.path.dirname(os.path.realpath(__file__)) + "/ressource/icon_add.png"))
            btnAdd.setAutoRaise(True)
            btnAdd.setToolTip("Add " + sTabName)
            # Status
            btnStat = QToolButton()
            if sTabName == "taxon":
                btnStat.setIcon(QIcon(os.path.dirname(os.path.realpath(__file__)) + "/ressource/icon_info1.png"))
            else:
                btnStat.setIcon(QIcon(os.path.dirname(os.path.realpath(__file__)) + "/ressource/icon_frage.png"))
            self.__dicStatus_widgets_lookup_status[sTabName] = btnStat # Merken
            # btnStat.setText("?")
            btnStat.setAutoRaise(True)
            font = btnStat.font()
            font.setPointSize(10)
            font.setBold(False)
            btnStat.setFont(font)
            btnStat.setToolTip("Show status report on: " + sTabName)
            # btnStat.setToolButtonStyle(Qt.ToolButtonTextOnly)
            btnStat.setFixedWidth(btnStat.minimumSizeHint().width())
            btnStat.setFixedHeight(btnStat.minimumSizeHint().height())

            # Purge
            btnPurge = QToolButton()
            menuPurge = QMenu()
            btnPurge.setMenu(menuPurge)
            btnPurge.setPopupMode(QToolButton.InstantPopup)
            btnPurge.setToolTip("Purge: "+sTabName_sheet)
            btnPurge.setIcon(QIcon(os.path.dirname(os.path.realpath(__file__)) + "/ressource/icon_trash.png"))
            btnPurge.setAutoRaise(True)
            actionPurge_notUsed = QAction(str(sTabName_sheet)+": purge hard (all not-used)", self)
            actionPurge_notUsed_notGold = QAction(str(sTabName_sheet)+": purge soft (all not-used and not-gold)", self)
            actionPurge_notUsed.triggered.connect(partial(self.admin_lookup_purge, lookup=sTabName, full=True))
            actionPurge_notUsed_notGold.triggered.connect(partial(self.admin_lookup_purge, lookup=sTabName, full=False))
            menuPurge.addAction(actionPurge_notUsed)
            menuPurge.addAction(actionPurge_notUsed_notGold)

            # Connect
            if sTabName == "taxon":
                btnStat.clicked.connect(partial(self.ui_table_stat, sTabName))
                btnOpen.clicked.connect(partial(self.table_show, sTabName))
            else:
                # Open
                btnMenu = QMenu()
                btnMenu_gold = QMenu("gold", self)
                btnMenu_use = QMenu("use", self)
                btnMenu_special = QMenu("special", self)
                btnOpen.setMenu(btnMenu)
                btnMenu.addMenu(btnMenu_use)
                btnMenu.addMenu(btnMenu_gold)
                btnMenu.addMenu(btnMenu_special)
                btnOpen.setPopupMode(QToolButton.InstantPopup)

                action = QAction("all", self)
                action.triggered.connect(partial(self.table_show, sTabName, gold_edit=True))
                actionGold = QAction("gold", self)
                actionGold.triggered.connect(partial(self.table_show, sTabName, gold=True, gold_edit=True))
                actionNotGold = QAction("not-gold", self)
                actionNotGold.triggered.connect(partial(self.table_show, sTabName, gold=False, gold_edit=True))
                actionUsed = QAction("used", self)
                actionUsed.triggered.connect(partial(self.table_show, sTabName, used=True, gold_edit=True))
                actionNotUsed = QAction("not-used (purge hard?)", self)
                actionNotUsed.triggered.connect(partial(self.table_show, sTabName, used=False, gold_edit=True))
                actionUsedNotGold = QAction("not-gold and used (check for gold?)", self)
                actionUsedNotGold.triggered.connect(partial(self.table_show, sTabName, used=True, gold=False, gold_edit=True))
                actionNotUsedNotGold = QAction("not-gold and not-used (purge soft?)", self)
                actionNotUsedNotGold.triggered.connect(partial(self.table_show, sTabName, used=False, gold=False, gold_edit=True))

                # Special: Dataset
                if sTabName == "dataset":
                    # Access
                    btnMenu_access = QMenu("access", self)
                    btnMenu.addMenu(btnMenu_access)
                    actionAccessOk = QAction("access somehow", self)
                    actionAccessOk.triggered.connect(partial(self.table_show, sTabName, access=True, gold_edit=True))
                    actionAccessRestricted = QAction("no access at all", self)
                    actionAccessRestricted.triggered.connect(partial(self.table_show, sTabName, access=False, gold_edit=True))
                    actionAccessOk_used = QAction("used and access somehow", self)
                    actionAccessOk_used.triggered.connect(partial(self.table_show, sTabName, access=True, used=True, gold_edit=True))
                    actionAccessRestricted_used = QAction("used and no access at all", self)
                    actionAccessRestricted_used.triggered.connect(partial(self.table_show, sTabName, access=False, used=True, gold_edit=True))
                    actionAccessOk_used_notGold = QAction("used and access somehow and not-gold (check?)", self)
                    actionAccessOk_used_notGold.triggered.connect(partial(self.table_show, sTabName, access=True, used=True, gold=False, gold_edit=True))
                    btnMenu_access.addAction(actionAccessOk)
                    btnMenu_access.addAction(actionAccessRestricted)
                    btnMenu_access.addAction(actionAccessOk_used)
                    btnMenu_access.addAction(actionAccessRestricted_used)
                    btnMenu_access.addAction(actionAccessOk_used_notGold)
                    # DOI
                    btnMenu_doi = QMenu("doi", self)
                    btnMenu.addMenu(btnMenu_doi)
                    accessDoiOk = QAction("ok", self)
                    accessDoiOk.triggered.connect(partial(self.table_show, sTabName, doi=True, gold_edit=True))
                    accessDoiNotOk = QAction("not-ok", self)
                    accessDoiNotOk.triggered.connect(partial(self.table_show, sTabName, doi=False, gold_edit=True))
                    accessDoiOk_used = QAction("used ok", self)
                    accessDoiOk_used.triggered.connect(partial(self.table_show, sTabName, doi=True, used=True, gold_edit=True))
                    accessDoiNotOk_used = QAction("used and not-ok", self)
                    accessDoiNotOk_used.triggered.connect(partial(self.table_show, sTabName, doi=False, used=True, gold_edit=True))
                    accessDoiNotOk_used_accessOk = QAction("used and not-ok and access somehow (bad DOI online!)", self)
                    accessDoiNotOk_used_accessOk.triggered.connect(partial(self.table_show, sTabName, access=True, used=True, doi=False, gold_edit=True))
                    accessDoiNotPangaea = QAction("not PANGAEA (raw data?)", self)
                    accessDoiNotPangaea.triggered.connect(partial(self.table_show, sTabName, doi=True, doi_pangaea=False, gold_edit=True))
                    btnMenu_doi.addAction(accessDoiOk)
                    btnMenu_doi.addAction(accessDoiNotOk)
                    btnMenu_doi.addAction(accessDoiOk_used)
                    btnMenu_doi.addAction(accessDoiNotOk_used)
                    btnMenu_doi.addAction(accessDoiNotOk_used_accessOk)
                    btnMenu_doi.addAction(accessDoiNotPangaea)
                if sTabName == "gear":
                    # Double Configuration
                    btnMenu_config = QMenu("config", self)
                    btnMenu.addMenu(btnMenu_config)
                    actionAccessDouble = QAction("double (merge?)", self)
                    actionAccessDouble.triggered.connect(partial(self.table_show, sTabName, gear_double=True, gold_edit=True))
                    btnMenu_config.addAction(actionAccessDouble)

                btnMenu_special.addAction(actionNotUsedNotGold)
                btnMenu_special.addAction(actionUsedNotGold)
                btnMenu_use.addAction(actionUsed)
                btnMenu_use.addAction(actionNotUsed)
                btnMenu_gold.addAction(actionGold)
                btnMenu_gold.addAction(actionNotGold)
                btnMenu.addAction(action)
                btnStat.clicked.connect(partial(self.ui_table_stat, sTabName))
                btnAdd.clicked.connect(partial(self.admin_createLookupOrTaxon, sTabName, "New: " + self.__map.lookup_table2sheet(sTabName)))
            # Ausgabefeld
            ausgabeFeld = QLineEdit()
            ausgabeFeld.setFixedWidth(65)
            ausgabeFeld.setEnabled(True)
            ausgabeFeld.setFrame(True)

            font = ausgabeFeld.font()
            font.setPointSize(10)
            font.setBold(False)
            ausgabeFeld.setFont(font)
            ausgabeFeld.setAlignment(Qt.AlignRight)
            ausgabeFeld.setReadOnly(True)
            # Lookup Gold
            goldFeld = QProgressBar()
            #goldFeld.setStyleSheet(CssPresets.progressBar_grey)
            goldFeld.setFixedWidth(40)
            goldFeld.setEnabled(True)
            palette = QPalette()
            palette.setColor(QPalette.ColorRole.Text, QColor(Colors.getColor("black")))
            palette.setColor(QPalette.ColorRole.HighlightedText, QColor(Colors.getColor("black")))
            goldFeld.setPalette(palette)
            goldFeld.setTextVisible(True)
            goldFeld.setValue(0)
            goldFeld.setMaximum(100)
            goldFeld.setFormat("")
            font = goldFeld.font()
            font.setPointSize(10)
            font.setBold(False)
            goldFeld.setFont(font)
            goldFeld.setAlignment(Qt.AlignRight)
            # Lookup Used
            usedFeld = QProgressBar()
            usedFeld.setStyleSheet(CssPresets.progressBar_grey)
            usedFeld.setFixedWidth(40)
            usedFeld.setEnabled(True)
            palette = QPalette()
            palette.setColor(QPalette.ColorRole.Text, QColor(Colors.getColor("black")))
            palette.setColor(QPalette.ColorRole.HighlightedText, QColor(Colors.getColor("black")))
            usedFeld.setPalette(palette)
            usedFeld.setTextVisible(True)
            usedFeld.setMaximum(100)
            usedFeld.setFormat("")
            font = usedFeld.font()
            font.setPointSize(10)
            font.setBold(False)
            usedFeld.setFont(font)
            usedFeld.setAlignment(Qt.AlignRight)
            # Lookup Used and not-gold
            usedFeld_notgold = QLineEdit()
            usedFeld_notgold.setFixedWidth(25)
            usedFeld_notgold.setEnabled(True)
            usedFeld_notgold.setFrame(True)
            font = usedFeld_notgold.font()
            font.setPointSize(10)
            font.setBold(False)
            usedFeld_notgold.setFont(font)
            usedFeld_notgold.setAlignment(Qt.AlignCenter)
            usedFeld_notgold.setReadOnly(True)
            # Lookup Not-Used and not-gold
            notusedFeld_notgold = QLineEdit()
            notusedFeld_notgold.setFixedWidth(25)
            notusedFeld_notgold.setEnabled(True)
            notusedFeld_notgold.setFrame(True)
            font = notusedFeld_notgold.font()
            font.setPointSize(10)
            font.setBold(False)
            notusedFeld_notgold.setFont(font)
            notusedFeld_notgold.setAlignment(Qt.AlignCenter)
            notusedFeld_notgold.setReadOnly(True)
            # Lookup: Special Info
            specialInfoFeld = QLineEdit()
            specialInfoFeld.setFixedWidth(25)
            specialInfoFeld.setEnabled(True)
            specialInfoFeld.setFrame(True)
            font = specialInfoFeld.font()
            font.setPointSize(10)
            font.setBold(False)
            specialInfoFeld.setFont(font)
            specialInfoFeld.setAlignment(Qt.AlignCenter)
            specialInfoFeld.setReadOnly(True)
            if sTabName == "dataset":
                # Lookup (Dataset): Reachable
                dataset_AllFreeRequest = QProgressBar()
                dataset_AllFreeRequest.setStyleSheet(CssPresets.progressBar_grey)
                dataset_AllFreeRequest.setFixedWidth(40)
                dataset_AllFreeRequest.setEnabled(True)
                dataset_AllFreeRequest.setTextVisible(True)
                dataset_AllFreeRequest.setValue(0)
                dataset_AllFreeRequest.setMaximum(100)
                dataset_AllFreeRequest.setFormat("")
                palette = QPalette()
                palette.setColor(QPalette.Highlight, QColor(Colors.getColor("black")))
                dataset_AllFreeRequest.setPalette(palette)
                font = dataset_AllFreeRequest.font()
                font.setPointSize(10)
                font.setBold(False)
                dataset_AllFreeRequest.setFont(font)
                dataset_AllFreeRequest.setAlignment(Qt.AlignRight)
                # Lookup (Dataset): Used and reachable
                dataset_AllUsedFreeRequest = QProgressBar()
                dataset_AllUsedFreeRequest.setStyleSheet(CssPresets.progressBar_grey)
                dataset_AllUsedFreeRequest.setFixedWidth(40)
                dataset_AllUsedFreeRequest.setEnabled(True)
                dataset_AllUsedFreeRequest.setTextVisible(True)
                dataset_AllUsedFreeRequest.setValue(0)
                dataset_AllUsedFreeRequest.setMaximum(100)
                dataset_AllUsedFreeRequest.setFormat("")
                palette = QPalette()
                palette.setColor(QPalette.Highlight, QColor(Colors.getColor("black")))
                dataset_AllUsedFreeRequest.setPalette(palette)
                font = dataset_AllUsedFreeRequest.font()
                font.setPointSize(10)
                font.setBold(False)
                dataset_AllUsedFreeRequest.setFont(font)
                dataset_AllUsedFreeRequest.setAlignment(Qt.AlignRight)
                # Lookup (Dataset): Used, not-gold and reachable
                dataset_AllUsedNotGoldFreeRequest = QLineEdit()
                dataset_AllUsedNotGoldFreeRequest.setFixedWidth(25)
                dataset_AllUsedNotGoldFreeRequest.setEnabled(True)
                dataset_AllUsedNotGoldFreeRequest.setFrame(True)
                font = dataset_AllUsedNotGoldFreeRequest.font()
                font.setPointSize(10)
                font.setBold(False)
                dataset_AllUsedNotGoldFreeRequest.setFont(font)
                dataset_AllUsedNotGoldFreeRequest.setAlignment(Qt.AlignCenter)
                dataset_AllUsedNotGoldFreeRequest.setReadOnly(True)
                # Lookup (Dataset): DOI Ok/NotOK/empty
                dataset_DOI_OkNotOk = QProgressBar()
                dataset_DOI_OkNotOk.setStyleSheet(CssPresets.progressBar_grey)
                dataset_DOI_OkNotOk.setFixedWidth(40)
                dataset_DOI_OkNotOk.setEnabled(True)
                dataset_DOI_OkNotOk.setTextVisible(True)
                dataset_DOI_OkNotOk.setValue(0)
                dataset_DOI_OkNotOk.setMaximum(100)
                dataset_DOI_OkNotOk.setFormat("")
                font = dataset_DOI_OkNotOk.font()
                font.setPointSize(10)
                font.setBold(False)
                dataset_DOI_OkNotOk.setFont(font)
                dataset_DOI_OkNotOk.setAlignment(Qt.AlignRight)
                palette = QPalette()
                palette.setColor(QPalette.Highlight, QColor(Colors.getColor("black")))
                dataset_DOI_OkNotOk.setPalette(palette)
                # Lookup (Dataset): DOI Used Ok/NotOK/empty
                dataset_DOI_UsedOkNotOk = QProgressBar()
                dataset_DOI_UsedOkNotOk.setStyleSheet(CssPresets.progressBar_grey)
                dataset_DOI_UsedOkNotOk.setFixedWidth(40)
                dataset_DOI_UsedOkNotOk.setEnabled(True)
                dataset_DOI_UsedOkNotOk.setTextVisible(True)
                dataset_DOI_UsedOkNotOk.setValue(0)
                dataset_DOI_UsedOkNotOk.setMaximum(100)
                dataset_DOI_UsedOkNotOk.setFormat("")
                font = dataset_DOI_UsedOkNotOk.font()
                font.setPointSize(10)
                font.setBold(False)
                dataset_DOI_UsedOkNotOk.setFont(font)
                dataset_DOI_UsedOkNotOk.setAlignment(Qt.AlignRight)
                palette = QPalette()
                dataset_DOI_UsedOkNotOk.setPalette(palette)
                # Lookup (Dataset): DOI Used,reachabel notOK/None
                dataset_DOI_UsedNotOkReachable = QLineEdit()
                dataset_DOI_UsedNotOkReachable.setFixedWidth(25)
                dataset_DOI_UsedNotOkReachable.setEnabled(True)
                dataset_DOI_UsedNotOkReachable.setFrame(True)
                font = dataset_DOI_UsedNotOkReachable.font()
                font.setPointSize(10)
                font.setBold(False)
                dataset_DOI_UsedNotOkReachable.setFont(font)
                dataset_DOI_UsedNotOkReachable.setAlignment(Qt.AlignCenter)
                dataset_DOI_UsedNotOkReachable.setReadOnly(True)
                # Lookup (Dataset): DOI PANGAEA
                dataset_DOI_Pangaea = QLineEdit()
                dataset_DOI_Pangaea.setFixedWidth(25)
                dataset_DOI_Pangaea.setEnabled(True)
                dataset_DOI_Pangaea.setFrame(True)
                font = dataset_DOI_Pangaea.font()
                font.setPointSize(10)
                font.setBold(False)
                dataset_DOI_Pangaea.setFont(font)
                dataset_DOI_Pangaea.setAlignment(Qt.AlignCenter)
                dataset_DOI_Pangaea.setReadOnly(True)
                # Lookup (Dataset): DOI Not-PANGAEA
                dataset_DOI_PangaeaNot = QLineEdit()
                dataset_DOI_PangaeaNot.setFixedWidth(25)
                dataset_DOI_PangaeaNot.setEnabled(True)
                dataset_DOI_PangaeaNot.setFrame(True)
                font = dataset_DOI_PangaeaNot.font()
                font.setPointSize(10)
                font.setBold(False)
                dataset_DOI_PangaeaNot.setFont(font)
                dataset_DOI_PangaeaNot.setAlignment(Qt.AlignCenter)
                dataset_DOI_PangaeaNot.setReadOnly(True)
                # Lookup (Dataset): Check DOIs
                btnDatasetCheckDoi = QToolButton()
                btnDatasetCheckDoi.setToolTip("Check DOIs in datasets")
                btnDatasetCheckDoi.clicked.connect(self.admin_checkDois)
                btnDatasetCheckDoi.setIcon(QIcon(os.path.dirname(os.path.realpath(__file__)) + "/ressource/icon_go.png"))
                btnDatasetCheckDoi.setAutoRaise(False)
                # Lookup (Dataset): Info Status DOIs
                btnDatasetCheckDoi_status = QToolButton()
                btnDatasetCheckDoi_status.setToolTip("Show report on DOI status in datasets")
                btnDatasetCheckDoi_status.clicked.connect(partial(self.ui_table_stat, "DOI"))
                btnDatasetCheckDoi_status.setIcon(QIcon(os.path.dirname(os.path.realpath(__file__)) + "/ressource/icon_frage.png"))
                btnDatasetCheckDoi_status.setAutoRaise(True)
                # Widgets in Dictionary merken

                self.__widgetStatus_widgets_lookup_dataset_allFreeRequest = dataset_AllFreeRequest
                self.__widgetStatus_widgets_lookup_dataset_allUsedFreeRequest = dataset_AllUsedFreeRequest
                self.__widgetStatus_widgets_lookup_dataset_allUsedNotGoldFreeRequest = dataset_AllUsedNotGoldFreeRequest
                self.__widgetStatus_widgets_lookup_dataset_DOI_OkNotOk = dataset_DOI_OkNotOk
                self.__widgetStatus_widgets_lookup_dataset_DOI_UsedOkNotOk = dataset_DOI_UsedOkNotOk
                self.__widgetStatus_widgets_lookup_dataset_DOI_UsedNotOkReachable = dataset_DOI_UsedNotOkReachable
                self.__widgetStatus_widgets_lookup_dataset_DOI_Pangaea = dataset_DOI_Pangaea
                self.__widgetStatus_widgets_lookup_dataset_DOI_PangaeaNot = dataset_DOI_PangaeaNot
                self.__widgetStatus_widgets_lookup_dataset_checkDois_status = btnDatasetCheckDoi_status

            # Spezielle Widgets in Dictionary merken
            self.__dicStatus_widgets_lookup[sTabName] = ausgabeFeld
            self.__dicStatus_widgets_lookup_special[sTabName] = specialInfoFeld

            # Gold Lookup Widgets in Dictionary merken
            if sTabName != "taxon":
                self.__dicStatus_widgets_lookup_gold[sTabName] = goldFeld
                self.__dicStatus_widgets_lookup_count[sTabName] = usedFeld
                self.__dicStatus_widgets_lookup_usedNotGoldLookup[sTabName] = usedFeld_notgold
                self.__dicStatus_widgets_lookup_notUsedNotGoldLookup[sTabName] = notusedFeld_notgold
            # Zusammenbauen
            gridLookup.addWidget(btnOpen, iRowLookup, iColLookup + 1)
            if sTabName != "taxon":
                gridLookup.addWidget(btnAdd, iRowLookup, iColLookup + 2)
                gridLookup.addWidget(btnPurge, iRowLookup, iColLookup + 9)
            gridLookup.addWidget(ausgabeFeld, iRowLookup, iColLookup + 3)
            if sTabName != "taxon":
                gridLookup.addWidget(usedFeld, iRowLookup, iColLookup + 4)
                gridLookup.addWidget(goldFeld, iRowLookup, iColLookup + 5)
                gridLookup.addWidget(usedFeld_notgold, iRowLookup, iColLookup + 6)
                gridLookup.addWidget(notusedFeld_notgold, iRowLookup, iColLookup + 7)
                gridLookup.addWidget(specialInfoFeld, iRowLookup, iColLookup + 8)
            gridLookup.addWidget(btnStat, iRowLookup, iColLookup + 10)
            if sTabName == "dataset":
                iRowLookup += 1
                gridLookup.addWidget(dataset_AllFreeRequest, iRowLookup, iColLookup + 4)
                gridLookup.addWidget(dataset_AllUsedFreeRequest, iRowLookup, iColLookup + 5)
                gridLookup.addWidget(dataset_AllUsedNotGoldFreeRequest, iRowLookup, iColLookup + 6)
                iRowLookup += 1
                gridLookup.addWidget(dataset_DOI_OkNotOk, iRowLookup, iColLookup + 4)
                gridLookup.addWidget(dataset_DOI_UsedOkNotOk, iRowLookup, iColLookup + 5)
                gridLookup.addWidget(dataset_DOI_UsedNotOkReachable, iRowLookup, iColLookup + 6)
                gridLookup.addWidget(dataset_DOI_Pangaea, iRowLookup, iColLookup + 7)
                gridLookup.addWidget(dataset_DOI_PangaeaNot, iRowLookup, iColLookup + 8)
                gridLookup.addWidget(btnDatasetCheckDoi, iRowLookup, iColLookup + 9)
                gridLookup.addWidget(btnDatasetCheckDoi_status, iRowLookup, iColLookup + 10)
                iWidgets_lookup += 2
            # Nächstes Widget anordnen
            iWidgets_lookup += 1
            if iWidgets_lookup == iWidgetLookup_maxPerCol:
                iColLookup = iWidgetLookup_witdh + 1
                iRowLookup = 1
            else:
                iRowLookup += 1

        # Purge
        btnPurgeAll = QToolButton()
        menuPurgeAll = QMenu()
        btnPurgeAll.setMenu(menuPurgeAll)
        btnPurgeAll.setPopupMode(QToolButton.InstantPopup)
        btnPurgeAll.setToolTip("Purge: all lookups")
        btnPurgeAll.setIcon(QIcon(os.path.dirname(os.path.realpath(__file__)) + "/ressource/icon_trash.png"))
        actionPurgeAll_notUsed = QAction("All lookups: purge hard (all not-used)", self)
        actionPurgeAll_notUsed_notGold = QAction("All lookups: purge soft (all not-used and not-gold)", self)
        actionPurgeAll_notUsed.triggered.connect(partial(self.admin_lookup_purge, full=True))
        actionPurgeAll_notUsed_notGold.triggered.connect(partial(self.admin_lookup_purge, full=False))
        menuPurgeAll.addAction(actionPurgeAll_notUsed)
        menuPurgeAll.addAction(actionPurgeAll_notUsed_notGold)
        btnPurgeAll.setAutoRaise(True)
        gridLookup.addWidget(btnPurgeAll, 1, iColLookup + 23)

        # Import Lookups
        btnImportLookup = QToolButton()
        btnImportLookup.setToolTip("Import lookup data")
        btnImportLookup.clicked.connect(self.import_lookups)
        btnImportLookup.setIcon(QIcon(os.path.dirname(os.path.realpath(__file__)) + "/ressource/icon_import.png"))
        btnImportLookup.setAutoRaise(True)
        gridLookup.addWidget(btnImportLookup, iRowLookup - 2, iColLookup + 23)

        # Export Lookups
        btnExportLookup = QToolButton()
        btnExportLookup.setToolTip("Export lookup data")
        btnExportLookup.clicked.connect(self.admin_lookup_export)
        btnExportLookup.setIcon(QIcon(os.path.dirname(os.path.realpath(__file__)) + "/ressource/icon_export.png"))
        btnExportLookup.setAutoRaise(True)
        gridLookup.addWidget(btnExportLookup, iRowLookup - 1, iColLookup + 23)

        # Baue Grids zusammen (Lookup/Ingest)
        self.groupboxDBStatusLookup.setLayout(gridLookup)
        self.groupboxDBStatusIngest.setLayout(gridIngest)

    def build_browser(self):
        """
        Neuws Browser Widget erzeugen
        """
        class QtWebIntercept(QWebEngineUrlRequestInterceptor):
            def interceptRequest(self, info):
                info.setHttpHeader(b"Accept-Language", b"en-US,en;q=0.9,es;q=0.8,de;q=0.7")

        self.__browser = QWebEngineView()
        # interceptor = QtWebIntercept()
        # self.__browser.page().profile().setUrlRequestInterceptor(interceptor)

    # --- Tools

    def tools_checkCsvEncoding(self, sFile, bPopup=True):

        """
        Teste Datei auf erlaubte Textkodierung (UTF-8/ASCII)

        Args:
            sFile (str): Dateipfad zur zu testenden DAtei
            bPopup (bool): Soll der Testvorgang mit einem Popup gekennzeichnet werden?

        Returns:
            (bool): Ist die Datei "sicher"?

        """
        pn = self.ui_printConsoleNormal
        pv = self.ui_printConsoleVerbose

        self.ui_showProgress()
        self.ui_setWorkflow(top="Checking csv file ", sub="Working")
        pv("* Checking csv file for allowed text encoding (UTF-8/ASCII/IBM852)")
        pv("Datei: "+sFile)

        (sEncoding, fConfidence) = TextEncoding.checkout(sFile, self.__smartprinter)
        pv("Determined encoding type: "+sEncoding.upper() + " (Probability: "+str(round(fConfidence*100))+"%)")


        # Nicht Unterstützte Kodierung gefunden
        if sEncoding.lower() not in ("ascii", "utf-8", "ibm852", "iso-8859-1"):
            sTitle = "Potential problems caused by the type of text encoding may occur"
            sText = "<h3>The text of the CSV file was encoded in an unsupported manner</h3>" + \
                    "<b>Determined Kodierungsart</b>" + \
                    "<ul><li>" + sEncoding.upper()+ "</li></ul>" +\
                    "<b>Following text encodings are supported</b> " + \
                    "<ul><li>ASCII, UTF-8</li></ul>" + \
                    "A sucessful import can not be guaranteed. Do you want to continue anyway?"
            ant = self.popup_generic(title="Encoding", head=sTitle, info=sText, button_yes="Yes", button_no="No")
            self.ui_hideProgress()
            if ant == 1:
                return True
            else:
                return False
            return bAntwort

        self.ui_hideProgress()
        return True

    def getPassword(self):
        """
        Hole DB-Passwort manuell

        """
        pw, ok = QInputDialog.getText(self, 'Connection to server', 'Password:', QLineEdit.Password)
        if ok:
            self.__model.subsystem_getDb().account_single_setPart("pw", pw)

    # --- Popups

    def popup_offline(self):
        self.popup_infoTimer("The system is currently offline ...", "Please connect to server first.")

    def popup_text(self, **kwargs):
        """
         Erzeuge ein Popup-Fenster mit Informationen
         Kwrgs:
             titel (str): Titeltext
             text (str): Fenstertext
             file (bool): Textfile
             html (bool): Inhalt ist HTML
             browser (bool): Browser statt Textwidget
             save (bool): Inhalt abspeicherbar
             title_capitalize (bool): Titel kapitalisieren
        """

        bSave = kwargs["save"] if ("save" in kwargs and kwargs["save"] in [True, False]) else False
        bHtml = kwargs["html"] if ("html" in kwargs and kwargs["html"] in [True, False]) else False
        bBrowser = kwargs["browser"] if ("browser" in kwargs and kwargs["browser"] in [True, False]) else False
        bCapitalizeTitle = kwargs["title_capitalize"] if ("title_capitalize" in kwargs and kwargs["title_capitalize"] in [True, False]) else False
        sTitle = kwargs["title"] if "title" in kwargs else None
        sFile = kwargs["file"] if "file" in kwargs else None
        sText = kwargs["text"] if "text" in kwargs else None

        # Tests
        if bCapitalizeTitle and sTitle is not None:
            sTitle = sTitle.title()
        if sFile not in [None, ""] and sText not in [None, ""]:
            self.ui_printConsoleNormal("Internal error: File and Text given")
        if sFile in [None, ""] and sText in [None, ""]:
            self.ui_printConsoleNormal("Internal error: No File or Text given")

        # Dialog erzeugen
        if bBrowser:
            # *** Browser
            if sFile in ["", None]:
                # Temporäres File mit Text füllen
                pathFile = Path(self.__model.conf_getAppTmpDir()) / Path("browser.html")
                sFile = str(pathFile)
                try:
                    file = open(sFile, "w",  encoding="utf-8")
                    file.write(sText)
                    file.close()
                except:
                    pass
            browser = QWebEngineView()
            # browser.setContextMenuPolicy(Qt.NoContextMenu)
            browser.setUrl(QUrl.fromLocalFile(sFile))
            dia = QDialog()
            dia.setWindowTitle(sTitle)
            dia.setWindowFlags(Qt.Dialog | Qt.CustomizeWindowHint | Qt.WindowTitleHint | Qt.WindowCloseButtonHint)
            # dia.setWindowModality(Qt.ApplicationModal)
            vbox = QVBoxLayout(dia)
            buttonBox = QDialogButtonBox(QDialogButtonBox.Ok, Qt.Horizontal, dia)
            buttonBox.accepted.connect(dia.accept)
            if bSave:
                btnSave = QPushButton()
                btnSave.setText("Save")
                btnSave.setToolTip("Saving to file")
                btnSave.clicked.connect(lambda: QDialog.done(dia, 2))
                buttonBox.addButton(btnSave, QDialogButtonBox.NoRole)
            btnWebbroser = QPushButton()
            btnWebbroser.setText("Browser")
            btnWebbroser.setToolTip("Open in external Webbroser")
            btnWebbroser.clicked.connect(lambda: QDialog.done(dia, 3))
            buttonBox.addButton(btnWebbroser, QDialogButtonBox.NoRole)
            font = QFont()
            font.setFamily("Arial")
            font.setPointSize(10)
            buttonBox.setFont(font)
            vbox.addWidget(browser)
            vbox.addWidget(buttonBox)
            dia.setMinimumSize(1000, 600)
            self.ui_centerWidget(dia)
            dia.setModal(False)
        else:
            # *** Textwidget
            dia = QDialog()
            dia.setWindowTitle(sTitle)
            dia.setWindowModality(Qt.ApplicationModal)
            dia.setWindowFlags(Qt.Dialog | Qt.CustomizeWindowHint | Qt.WindowTitleHint | Qt.WindowCloseButtonHint)
            vbox = QVBoxLayout(dia)
            vbox.setContentsMargins(4, 4, 4, 4)
            vbox.setSpacing(6)
            if sFile not in [None, ""]:
                # Hole Text aus Datei
                try:
                    file = open(sFile, "r", encoding="utf-8")
                    sText = file.read()
                    file.close()
                except Exception as ex:
                    sText = "Internal error: " + str(ex)
            text = QTextEdit()

            if bHtml:
                text.setHtml(sText)
            else:
                text.setPlainText(sText)

            font = QFont()
            font.setFamily("Courier")
            font.setPointSize(10)
            text.setFont(font)
            text.setLineWrapMode(QTextEdit.NoWrap)
            text.setReadOnly(True)
            vbox.addWidget(text)
            buttonBox = QDialogButtonBox()
            if bSave:
                btnSave = QPushButton()
                btnSave.setText("Save")
                btnSave.setToolTip("Saving to file")
                btnSave.clicked.connect(lambda: QDialog.done(dia, 2))
                buttonBox.addButton(btnSave, QDialogButtonBox.NoRole)
            # OK
            btnOk = QPushButton()
            btnOk.setText("OK")
            btnOk.clicked.connect(lambda: QDialog.done(dia, 0))
            buttonBox.accepted.connect(dia.accept)
            buttonBox.addButton(btnOk, QDialogButtonBox.NoRole)
            font = QFont()
            font.setFamily("Arial")
            font.setPointSize(10)
            buttonBox.setFont(font)
            vbox.addWidget(buttonBox)
            dia.resize(800, 600)


        # *** Dialog öffnen
        ant = dia.exec_()
        if bBrowser and ant == 3:
            # Externer Webbrowser
            webbrowser.open("file://"+sFile)
        if bSave and ant == 2:
            # Save
            if sTitle is not None:
                sFilenameDefault = PrettyText.clean_2filename(sTitle)
            else:
                sFilenameDefault = "text"
            sEnd = ".html" if bHtml else ".txt"
            sFilename = QFileDialog.getSaveFileName(self, "Save to file", sFilenameDefault + sEnd)[0]
            if sFilename != "":
                try:
                    with open(sFilename, "w", encoding="utf-8", newline="\r\n") as file:
                        file.write(sText)
                    self.popup_generic(title="Save to file", head="Data Successfully saved to file", info="File: " + sFilename)
                except Exception as ex:
                    self.ui_printConsoleNormal(("Error saving to file (" + sFilename + "): " + str(ex)))

    def popup_info(self, sHead, sInfo, sDetail=None, sShowSpecial=None):
        return self.popup_generic(title="Information", head=sHead, info=sInfo, detail=sDetail, button_ok="OK", button_special=sShowSpecial)

    def popup_infoTimer(self, sHead, sInfo, iTimer=2, sDetail=None):
        self.popup_generic(title="Information", head=sHead, info=sInfo, detail=sDetail, timer=iTimer)

    def popup_error(self, sHead, sInfo, sDetail=None, sShowSpecial=None):
        return self.popup_generic(title="Error", head=sHead, info=sInfo, detail=sDetail, button_special=sShowSpecial)

    # --- Direkte GUI-Aktionen

    def explorer_drawMap(self):
        """
        Draw Map
        """

        if self.sys_isWorkerRunning(popup=True):
            return

        if not self.__model.subsystem_getConnector().isReady():
            self.popup_offline()
            return

        self.explorer_update("draw_map")

    def explorer_reset(self):
        """
        Setze Start Datum auf das Minimum
        """
        self.explorer_reset_visual()

    def explorer_toggleActivateRegion(self):
        """
        Toggle: (De-)Aktivierung Region
        """
        self.explorer_update("toggle_region")

    def explorer_toggleActivateType(self):
        """
        Toggle: (De-)Aktivierung Type
        """
        self.explorer_update("toggle_type")

    def explorer_toggleActivateDataset(self):
        """
        Toggle: (De-)Aktivierung dataset
        """
        self.explorer_update("toggle_dataset")

    def explorer_toggleActivateGear(self):
        """
        Toggle: (De-)Aktivierung Gear
        """
        self.explorer_update("toggle_gear")

    def explorer_toggleActivateTaxcover(self):
        """
        Toggle: (De-)Aktivierung Taxcover
        """
        self.explorer_update("toggle_taxcover")

    def explorer_toggleActivateIngest(self):
        """
        Toggle: (De-)Aktivierung ingest
        """
        self.explorer_update("toggle_ingest")

    def explorer_invActivateType(self):
        """
        Toggle: Invert type
        """
        self.explorer_update("inv_type")

    def explorer_invActivateRegion(self):
        """
        Toggle: Invert Region
        """
        self.explorer_update("inv_region")

    def explorer_invActivateDataset(self):
        """
        Toggle: Invert dataset
        """
        self.explorer_update("inv_dataset")

    def explorer_invActivateGear(self):
        """
        Toggle: Invert Gear
        """
        self.explorer_update("inv_gear")

    def explorer_invActivateTaxcover(self):
        """
        Toggle: Invert Taxcover
        """
        self.explorer_update("inv_taxcover")

    def explorer_invActivateIngest(self):
        """
        Toggle: Invert Ingest
        """
        self.explorer_update("inv_ingest")

    def explorer_setDateStartToMin(self):
        """
        Setze Start Datum auf das Minimum
        """
        self.explorer_update("set_date_start_min")

    def explorer_setDateEndToMax(self):
        """
        Setze End Datum auf das Maximum
        """
        self.explorer_update("set_date_end_max")

    def explorer_showSubset(self, **kwargs):
        """
        Show subset in Dialog Tabelle
        """
        if self.sys_isWorkerRunning(popup=True):
            return

        if not self.__model.subsystem_getConnector().isReady():
            self.popup_offline()
            return

        pn = self.ui_printConsoleNormal
        pv = self.ui_printConsoleVerbose

        # Converter
        sConverter = kwargs.get("converter")

        # R-Friendly
        bRfriendly = kwargs.get("rfriendly")

        # Selector
        sDateStart = str(self.__model.ui_selector_getSelector().getActivEntries("date_start")[0])
        sDateEnd = str(self.__model.ui_selector_getSelector().getActivEntries("date_end")[0])
        lstType = self.__model.ui_selector_getSelector().getActivEntries("type")
        lstDataset = self.__model.ui_selector_getSelector().getActivEntries("dataset")
        lstRegion = self.__model.ui_selector_getSelector().getActivEntries("region")
        lstScope = self.__model.ui_selector_getSelector().getActivEntries("taxcover")
        lstIngest = self.__model.ui_selector_getSelector().getActivEntries("ingest")
        lstGear = self.__model.ui_selector_getSelector().getActivEntries("gear")

        dicSelector = {
            "sample_date_start": sDateStart,
            "sample_date_end": sDateEnd,
            "workflow": lstType,
            "dataset": lstDataset,
            "gear": lstGear,
            "taxcover": lstScope,
            "region": lstRegion,
            "ingest": lstIngest
        }

        # Start Worker
        self.worker_start(Worker_ShowSubset(self.__model, dicSelector, converter=sConverter, rfriendly=bRfriendly), self.worker_finished_showTable)

    def explorer_exportSubset_marlin(self):
        """
        Download subset MARLIN
        """

        if self.sys_isWorkerRunning(popup=True):
            return

        if not self.__model.subsystem_getConnector().isReady():
            self.popup_offline()
            return

        pn = self.ui_printConsoleNormal
        pv = self.ui_printConsoleVerbose

        sFile = QFileDialog.getSaveFileName(self, "Export subset to file", directory=str(Path(self.__model.conf_getWorkingDir() / Path("subset_marlin.xlsx"))), filter="*.xlsx;; *.*")[0]
        if sFile in ("", None):
            return

        # Selector
        sDateStart = str(self.__model.ui_selector_getSelector().getActivEntries("date_start")[0])
        sDateEnd = str(self.__model.ui_selector_getSelector().getActivEntries("date_end")[0])
        lstType = self.__model.ui_selector_getSelector().getActivEntries("type")
        lstDataset = self.__model.ui_selector_getSelector().getActivEntries("dataset")
        lstScope = self.__model.ui_selector_getSelector().getActivEntries("taxcover")
        lstIngest = self.__model.ui_selector_getSelector().getActivEntries("ingest")
        lstRegion = self.__model.ui_selector_getSelector().getActivEntries("region")
        lstGear = self.__model.ui_selector_getSelector().getActivEntries("gear")

        dicSelector = {
            "sample_date_start": sDateStart,
            "sample_date_end": sDateEnd,
            "workflow": lstType,
            "dataset": lstDataset,
            "gear": lstGear,
            "taxcover": lstScope,
            "region": lstRegion,
            "ingest": lstIngest
        }

        # Start Worker
        self.worker_start(Worker_ExportIngest(self.__model, sFile, dicSelector, converter="mar_trans"), self.worker_finished_exportExcel)

    def explorer_exportSubset(self, **kwargs):
        """
        Download subset
        """

        if self.sys_isWorkerRunning(popup=True):
            return

        if not self.__model.subsystem_getConnector().isReady():
            self.popup_offline()
            return

        pn = self.ui_printConsoleNormal
        pv = self.ui_printConsoleVerbose

        # Converter
        sConverter = kwargs.get("converter")

        # R-Friendly
        bRfriendly = False

        # Selector
        sDateStart = str(self.__model.ui_selector_getSelector().getActivEntries("date_start")[0])
        sDateEnd = str(self.__model.ui_selector_getSelector().getActivEntries("date_end")[0])
        lstType = self.__model.ui_selector_getSelector().getActivEntries("type")
        lstDataset = self.__model.ui_selector_getSelector().getActivEntries("dataset")
        lstScope = self.__model.ui_selector_getSelector().getActivEntries("taxcover")
        lstIngest = self.__model.ui_selector_getSelector().getActivEntries("ingest")
        lstRegion = self.__model.ui_selector_getSelector().getActivEntries("region")
        lstGear = self.__model.ui_selector_getSelector().getActivEntries("gear")
        dicSelector = {
            "sample_date_start": sDateStart,
            "sample_date_end": sDateEnd,
            "workflow": lstType,
            "dataset": lstDataset,
            "gear": lstGear,
            "taxcover": lstScope,
            "region": lstRegion,
            "ingest": lstIngest
        }

        if sConverter == "mar_trans":
            # Converter: MARLIN transport
            sFile = QFileDialog.getSaveFileName(self, "Export subset to file", directory=str(
                Path(self.__model.conf_getWorkingDir() / Path("subset_marlin_transport.xlsx"))), filter="*.xlsx;; *.*")[0]
            if sFile in ("", None):
                return
            lstIngest = None
            lstLookup = None
            bPangaea = None
        else:
            # Normal
            diaExportSubset = DialogExportSubset(self.__model)
            self.ui_centerWidget(diaExportSubset)
            ant = diaExportSubset.exec_()
            if ant == 0:
                return
            res = diaExportSubset.getConfig()
            sFile = res.getData("file")
            lstIngest = res.getData("ingest")
            lstLookup = res.getData("lookup")
            bPangaea = res.getData("pangaea")
            bRfriendly = res.getData("rfriendly")

        # Start Worker
        self.worker_start(Worker_ExportIngest(self.__model, sFile, dicSelector, ingest_sheets=lstIngest, lookup_sheets=lstLookup, pangaea=bPangaea, rfriendly=bRfriendly, converter=sConverter), self.worker_finished_exportExcel)

    def explorer_openMapInExternalBrowser(self):
        """
        Open Map in external Browser
        """

        if self.sys_isWorkerRunning(popup=True):
            return

        if not self.__model.subsystem_getConnector().isReady():
            self.popup_offline()
            return

        # Externer Webbrowser
        sUrl = self.__browser.url().toString()
        #sFileHtml = str(Path(self.__model.conf_getAppTmpDir()) / Path("map.html"))
        #webbrowser.open("file://" + sFileHtml)
        webbrowser.open(sUrl)
    def explorer_saveMap(self):
        """
        Save Map
        """

        if self.sys_isWorkerRunning(popup=True):
            return

        if not self.__model.subsystem_getConnector().isReady():
            self.popup_offline()
            return

        sFile = QFileDialog.getSaveFileName(self, "Save map to file", "map.html")[0]
        if sFile == "":
            return

        # Start Worker
        self.worker_start(Worker_SaveMap(self.__model, sFile), self.worker_finished_saveMapToFile)

    def sys_presetChanged(self):
        """
        Verbindungs-Preset wurde verändert
        """

        if self.sys_isWorkerRunning(popup=True, status=True):
            self.status_refresh_connection()  # Combobox-Preset wieder auf aktuellen Wert zurückstellen
            return

        pn = self.ui_printConsoleNormal
        iIndex = self.infoComboDbPreset.currentIndex()  # Index aus Combobox
        if iIndex != self.__model.subsystem_getDb().account_active_getIndex():
            self.__model.subsystem_getDb().account_active_setIndex(iIndex)
            self.sys_server_disconnect()
            self.sys_server_connect()

    def import_ingest(self, sWf):
        """
        Starte Ingest

        Args:
            sWf (str): Name des gewünschten Workflows
        """

        if self.sys_isWorkerRunning(popup=True):
            return

        if not self.__model.subsystem_getConnector().isReady():
            self.popup_offline()
            return

        diaOpenIngest = DialogIngest(self.__model, ingest=sWf, lookup=True)
        self.ui_centerWidget(diaOpenIngest)
        ant = diaOpenIngest.exec_()
        if ant == 0:
            return
        # Dialog auswerten
        res = diaOpenIngest.getConfig()

        sFile = res.getData("file")
        lstIngest = res.getData("ingest")
        lstLookup = res.getData("lookup")
        sExcel = res.getData("lup-excel")
        sMode = res.getData("lup-mode")
        bExtend = res.getData("extend")
        sExtendReact = res.getData("extend-react")
        sLupReuseReact = res.getData("lup-reuse-react")
        bGoldRush = res.getData("lup-goldrush")

        sDb = res.getData("lup-db")
        bTamer = res.getData("tamer")
        bLupSingle = res.getData("ingest-lup-single")

        self.worker_start(Worker_Ingest(self.__model), self.worker_finished_ingestion,
                          wf=sWf, file=sFile, ing=lstIngest,
                          lup=lstLookup, lupExcel=sExcel, lupMode=sMode, lupDb=sDb, lupExcelGoldrush=bGoldRush, lupReuseReact=sLupReuseReact,
                          tamer=bTamer, ingest_lup_single=bLupSingle, extend=bExtend, extend_react=sExtendReact)

    def converter_marlin_doc(self):
        if self.sys_isWorkerRunning(popup=True):
            return
        if not self.__model.subsystem_getConnector().isReady():
            self.popup_offline()
            return
        sDoc = self.__model.subsystem_getConverter().info("bsh", html=True)
        self.popup_text(title="Conversion Engine: CB-MARLIN", text=sDoc, browser=True)

    def import_ingest_template(self, **kwargs):
        """
        Exportiere Excel-Template

        Kwargs:
            ingest (bool): Ingest
            lookup (bool): Lookup
            taxon (bool): Taxon
            wf (str): Kürzel des gewünschten Workflows
        """

        if self.sys_isWorkerRunning(popup=True):
            return

        if not self.__model.subsystem_getConnector().isReady():
            self.popup_offline()
            return

        bIngest = kwargs["ingest"] if ("ingest" in kwargs and kwargs["ingest"] in [True, False]) else False
        bLookup = kwargs["lookup"] if ("lookup" in kwargs and kwargs["lookup"] in [True, False]) else False
        bTaxon = kwargs["taxon"] if ("taxon" in kwargs and kwargs["taxon"] in [True, False]) else False
        sWfName = kwargs["wf"] if "wf" in kwargs else None

        sFile = "template.xlsx"
        if bIngest:
            sWfLong = self.__model.subsystem_getMapper().workflow_short2long(sWfName)
            if sWfLong is None:
                sFile = "template.xlsx"
            else:
                sFile = "template_"+sWfLong+".xlsx"
        if bLookup and bTaxon and not bIngest:
            sFile = "template_lookup_taxonwizard.xlsx"
        if bLookup and not bTaxon and not bIngest:
            sFile = "template_lookup.xlsx"
        if not bLookup and bTaxon and not bIngest:
            sFile = "template_taxonwizard.xlsx"
        sFile = sFile.replace(" ", "_")
        sFile = sFile.replace("(", "")
        sFile = sFile.replace(")", "")
        sFile = sFile.lower()
        excelFile = QFileDialog.getSaveFileName(self, "Saving Excel Template", sFile)[0]
        if excelFile != "":
            self.ui_setWorkflow(top="Building Excel template", sub="Working")
            self.worker_start(Worker_ExportTemplate(self.__model, excelFile, **kwargs), self.worker_finished_exportExcel)

    def converter_templateConverter(self, sWfName):

        if self.sys_isWorkerRunning(popup=True):
            return

        if not self.__model.subsystem_getConnector().isReady():
            self.popup_offline()
            return

        sWfLong = self.__model.subsystem_getMapper().workflow_converter_import_short2long(sWfName)
        if sWfLong is None:
            sFile = "template.xlsx"
        else:
            sFile = "template_" + sWfLong + ".xlsx"
        sFile = sFile.replace(" ", "_")
        sFile = sFile.replace("(", "")
        sFile = sFile.replace(")", "")
        sFile = sFile.lower()
        sExcelFile = QFileDialog.getSaveFileName(self, "Saving Excel Template", sFile)[0]
        if sExcelFile != "":
            self.ui_setWorkflow(top="Building Excel Template", sub="Working")
            self.worker_start(Worker_ExportConverterTemplate(self.__model, sExcelFile, sWfName), self.worker_finished_showPopupInfo)

    def main_batchBot(self):
        """
        BatchBot
        """

        if self.sys_isWorkerRunning(popup=True):
            return

        if not self.__model.subsystem_getConnector().isReady():
            self.popup_offline()
            return


        self.__model.subsystem_getBatchBot().backup()
        dia = DialogBatchBot(self.__model)
        ant = dia.exec_()

        if ant == 1:
            # Batch
            sDirLog = dia.logDir.text()
            sName = dia.comboBox.currentText()
            self.__model.subsystem_getBatchBot().save()
            self.__model.subsystem_getBatchBot().backup_del()
            # Worker
            self.worker_start(Worker_BatchBot(self.__model, sName, sDirLog), self.worker_finished_batchBot)
        elif ant == 3:
            # Export
            self.__model.subsystem_getBatchBot().save()
            self.__model.subsystem_getBatchBot().backup_del()
            sDir = QFileDialog.getExistingDirectory(self, "Select directory export configurations")
            if sDir is not None:
                resExport = Result()
                for sName in self.__model.subsystem_getBatchBot().getConfig_all():
                    sName_clean = PrettyText.clean_2filename(sName)
                    sContent = self.__model.subsystem_getBatchBot().getContent(sName)
                    sPath = Path(sDir, sName_clean + ".txt")
                    FileTools.prepareSaveFile(sPath)
                    try:
                        with open(sPath, "w") as f:
                            f.write(sContent)
                    except Exception as ex:
                        resExport.addErr("IO", add=str(ex), kat="Write configurations", sys="Export")
                if resExport.isErr():
                    self.popup_error("Export configurations", "Errors occured while exporting BatchBot configurations")
                else:
                    self.popup_generic(title="BatchBot", head="Export configurations", info="BatchBot configurations where exported sucessfully")
        elif ant == 2:
            # Apply
            self.__model.subsystem_getBatchBot().save()
            self.__model.subsystem_getBatchBot().backup_del()
        else:
            # Cancel
            self.__model.subsystem_getBatchBot().restore()
            self.__model.subsystem_getBatchBot().backup_del()
        self.ui_hideBatch()
        self.ui_hideProgress()

    def main_optionen(self):
        """
        Starte Optionen
        """

        if self.sys_isWorkerRunning(popup=True):
            return

        pn = self.ui_printConsoleNormal
        status = self.ui_setStatusbarText

        self.__model.subsystem_getDb().account_backup()

        # Erzeuge Options Dialog
        opt = DialogOptionen(self.__model)

        # Merke aktuelle Verbindungsdaten
        iIndexAlt = self.__model.subsystem_getDb().account_active_getIndex()  # Alter (noch aktueller) Index
        sPresetAlt = self.__model.subsystem_getDb().account_single_getPart("account")
        sUserAlt = self.__model.subsystem_getDb().account_single_getPart("user")
        sHostAlt = self.__model.subsystem_getDb().account_single_getPart("host")
        sPwAlt = self.__model.subsystem_getDb().account_single_getPart("pw")
        sPortAlt = self.__model.subsystem_getDb().account_single_getPart("port")
        sNameAlt = self.__model.subsystem_getDb().account_single_getPart("db")
        sSchemaAlt = self.__model.subsystem_getDb().account_single_getPart("schema")

        # Starte Dialog
        self.ui_centerWidget(opt)
        ant = opt.exec_()
        bPwSpeichern = opt.check_pwSpeichern.isChecked()
        if ant == 1 or ant == 2:
            # Schalte Account-Preset um
            iIndexNeu = opt.db_presets.currentIndex() # Index aus Combobox
            self.__model.subsystem_getDb().account_active_setIndex(iIndexNeu) # Index in Modell veraendern
            self.status_refresh_connection()

            # Teste, ob Workflows neu initialisiert werden muessen
            bWfRestart = False
            if sUserAlt != opt.db_user.text() or \
                    sHostAlt != opt.db_host.text() or \
                    sPwAlt != opt.db_pw.text() or \
                    sPortAlt != opt.db_port.text() or \
                    sNameAlt != opt.db_name.text() or \
                    sSchemaAlt != opt.db_schema.text():
                bWfRestart = True

            # Speichere Daten in Model
                # Extension Records
            self.__model.conf_setRecordExtensionAllowed(opt.check_record_extension.isChecked())
            self.__model.conf_setRecordExtensionProblemReactionlevel(opt.combo_record_extension_problem.currentText())
                # Lookup Reuse Differs
            self.__model.conf_setLookupReuseDiffersReactionlevel(opt.combo_record_exsiting_using_differs.currentText())
                # Lookup Import
            if opt.combo_lup_excel.currentIndex() == 1:
                sMode = "not-gold"
            elif opt.combo_lup_excel.currentIndex() == 2:
                sMode = "gold"
            else:
                sMode = "all"
            self.__model.conf_setLookupImport_excel(sMode)
            # Lookup: Goldrush
            self.__model.conf_setLookupImport_excel_goldrush(opt.check_goldrush.isChecked())
            if opt.combo_lup_db.currentIndex() == 0:
                sMode = "new"
            elif opt.combo_lup_db.currentIndex() == 1:
                sMode = "update"
            else:
                sMode = "both"
            self.__model.conf_setLookupImport_db(sMode)
            if opt.combo_lup_db_update.currentIndex() == 0:
                sMode = "not-gold"
            elif opt.combo_lup_db_update.currentIndex() == 1:
                sMode = "gold"
            else:
                sMode = "all"
            self.__model.conf_setLookupImport_db_update(sMode)
                # Geschwaetzigkeit
            self.__model.conf_setVerbosity(False, False)
            if opt.combo_verbosity.currentIndex() == 1:
                self.__model.conf_setVerbose(True)
            elif opt.combo_verbosity.currentIndex() == 2:
                self.__model.conf_setDebug(True)
                # BatchLog
            self.__model.conf_setWorkingDir(opt.dir_working.text())
                # Admin
            self.__model.conf_setAdminPsqlPath(opt.path_psql.text())
            self.__model.conf_setAdminPgdumpPath(opt.path_pgdump.text())
                # Logging
            self.__model.conf_setLogging(opt.check_logging.isChecked())
                # Dry
            self.__model.conf_setDry(opt.check_dry.isChecked())
                # Save PW
            self.__model.conf_setPwSave(opt.check_pwSpeichern.isChecked())

            # Speichere in Konfigurationsdatei
            if ant == 2:
                if bPwSpeichern  == True:
                    status("Saving settings")
                    self.__model.conf_cfgSave()
                else:
                    status("Saving settings. Passwords are not saved ...")
                    self.__model.conf_cfgSave(savePw=False)
            else:
                status("Apply settings")

            # Evaluiere die Pg Admin Tools neu (stumpf immer)
            self.__model.core_evalPgTools()

            # System-Optionen aktualisieren (stumpf immer)
            self.status_refresh_dryMode()
            self.status_refresh_toolbarMini()
            self.status_refresh_system()
            self.status_refresh_verbosity()
            self.status_refresh_adminLock()

            # Workflows neu starten
            if bWfRestart:
                if self.sys_isWorkerRunning():
                    self.popup_infoTimer("There is a process running", "Please connect to the server manually when the process has finished.")
                    return
                pn("Workflows must be reinitilized. Reconnecting to server ...")
                self.sys_server_disconnect()
                self.sys_server_connect()
        else:
            # Änderungen verwerfen
            # Alte Account Konfiguration zurückholen
            self.__model.subsystem_getDb().account_restore()

    def help_help(self):
        """
        Starte Hilfe
        """

        sHelpFile = self.__model.conf_getAppBaseDir() + "/doc/index.html"
        self.popup_text(title="Documentation", file=sHelpFile, browser=True, html=True)

        # browser = QWebEngineView()
        # browser.setContextMenuPolicy(Qt.NoContextMenu)
        # browser.setUrl(QUrl.fromLocalFile(sHelpFile))
        # dia = QDialog()
        # dia.setWindowTitle("Documentation")
        # vbox = QVBoxLayout(dia)
        # buttonBox = QDialogButtonBox(QDialogButtonBox.Ok, Qt.Horizontal, dia)
        # buttonBox.accepted.connect(dia.accept)
        # vbox.addWidget(browser)
        # vbox.addWidget(buttonBox)
        # dia.setMinimumSize(900,600)
        # self.ui_centerWidget(dia)
        # #dia.setModal(False)
        # #dia.setMinimumSize(dia.sizeHint())
        # dia.exec_()

    def help_info(self):
        """
        Starte Info
        """

        info = "<h3>" + self.__model.conf_getInfo("prgname") + " " + self.__model.conf_getInfo("version") + " (" + self.__model.conf_getInfo("dbschema") + ")</h3>"
        info += self.__model.conf_getInfo("comment")
        #info += "<p align =\"left\">Contact: <a href=\"" + self.__model.getInfo("www")+ "\">AWI|FE|Ecosystem Functions</a></p>"
        info += "<p align =\"left\">Licence: BSD-3 license (<a href=\"https://opensource.org/licenses/BSD-3-Clause\">LICENSE</a>)</p>"
        info += u"<p align =\"left\">Contact: <a href=\"" + self.__model.conf_getInfo("www") + "\">AWI | Ecosystem Functions</a></p>"
        QMessageBox.about(self, "About this program", info)

    def help_update(self):
        """
        Starte Update Check
        """
        pn = self.ui_printConsoleNormal

        if self.sys_isWorkerRunning(popup=True):
            return

        pn("* Checking for available updates")

        # Worker starten
        self.worker_start(Worker_CheckUpdateApp(self.__model), self.worker_finished_showUpdateCheck)

    def help_system(self):
        """
        Zeige System Info
        """
        pn = self.ui_printConsoleNormal

        dia = DialogSystemInfo(self.__model)
        dia.exec_()


    def help_appChangelog(self):
        """
        Zeige Changelog App
        """
        if self.sys_isWorkerRunning(popup=True):
            return

        pn = self.ui_printConsoleNormal

        try:
            sChangelogFile = self.__model.conf_getAppBaseDir() + "/" + self.__model.conf_getAppChangelogFile()
            with open(sChangelogFile, 'r') as f:
                sChangelog = f.read().strip()
                self.popup_text(title="Changelog app", text=sChangelog)
        except Exception as ex:
            pn("Error: could not read app changelog file "+sChangelogFile+": "+str(ex))


    def help_datenmodellChangelog(self):
        """
        Zeige Changelog Datenmodell
        """
        if self.sys_isWorkerRunning(popup=True):
            return

        pn = self.ui_printConsoleNormal

        try:
            sChangelogFile = self.__model.conf_getAppBaseDir() + "/" + self.__model.conf_getDatenmodellChangelogFile()
            with open(sChangelogFile, 'r') as f:
                sChangelog = f.read().strip()
                self.popup_text(title="Changelog data model", text=sChangelog)
        except Exception as ex:
            pn("Error: could not read data model changelog file "+sChangelogFile+": "+str(ex))


    def admin_createLookupOrTaxon(self, sTab, sTitel, **kwargs):
        """
        Neuen Datensatz erzeugen

        Args:
            sTab (str): Tabellenname
        Kwargs:
            <werden entgegengenommen und weitergegeben>
        """
        if self.sys_isWorkerRunning(popup=True):
            return
        if not self.__model.subsystem_getConnector().isReady():
            self.popup_offline()
            return
        self.dialog_input_generic(sTitel, sTab, **kwargs)

    def fao_cmd(self, sCmd):
        """
        FAO Regions
        """
        if self.sys_isWorkerRunning(popup=True):
            return
        if not self.__model.subsystem_getConnector().isReady():
            self.popup_offline()
            return
        kwargs = {}
        if sCmd == "reg_clean":
            sTitle = "FAO regions"
            sHead = "Clean FAO region mapping"
            sInfo = "Do you really want to clean the whole mapping?"
            ant = self.popup_generic(title=sTitle, head=sHead, info=sInfo, button_yes="Yes", button_no="No")
            if ant == 0:
                return
        elif sCmd == "import":
            sTitle = "FAO regions"
            sHead = "Import FAO regions"
            sInfo = "Do you really want to import FAO regions from file?"
            ant = self.popup_generic(title=sTitle, head=sHead, info=sInfo, button_yes="Yes", button_no="No")
            if ant == 0:
                return
            sFile = QFileDialog.getOpenFileName(self, "csv file", filter="*.csv;; *.*", directory=self.__model.conf_getUserHomeDir())[0]
            kwargs["file"] = sFile
        elif sCmd == "export":
            sFile = QFileDialog.getSaveFileName(self, "csv file", directory=str(Path(self.__model.conf_getUserHomeDir(), "regions.csv")))[0]
            if sFile in [None, ""]:
                return
            FileTools.prepareSaveFile(str(sFile))
            kwargs["file"] = sFile
        elif sCmd == "reg_create":
            sTitle = "FAO regions"
            sHead = "New FAO regions"
            sInfo = "Do you really want to harvest FAO regions from fao.org?"
            ant = self.popup_generic(title=sTitle, head=sHead, info=sInfo, button_yes="Yes", button_no="No")
            if ant == 0:
                return
            dia = DialogFaoCreate()
            ant = dia.exec_()
            if ant == 0:
                return
            iLevel = dia.slider.value()
            bMirror = dia.awi_mirror.isChecked()
            lstLevel = []
            lstLevel_all = ["major", "subarea", "division", "subdivision"]
            if iLevel == 4:
                lstLevel = "*"
            else:
                for i in range(4):
                    lstLevel.append(lstLevel_all[i])
                    if i == iLevel: break
            kwargs["level_new"] = lstLevel
            kwargs["mirror"] = bMirror
        # Worker starten
        if sCmd == "reg_show":
            self.worker_start(Worker_ShowWholeTable(self.__model, "region"), self.worker_finished_showTable)
        else:
            self.worker_start(Worker_Fao(self.__model, sCmd, **kwargs), self.worker_finished_fao)

    def dataPatrol_cmd(self, sCmd):
        """
        Data Patrol
        """
        if self.sys_isWorkerRunning(popup=True):
            return
        if not self.__model.subsystem_getConnector().isReady():
            self.popup_offline()
            return
        if sCmd == "biota_control":
            # Worker starten
            self.worker_start(Worker_DataPatrol_biotaControl(self.__model.subsystem_getDataPatrol()), self.worker_finished_biotaControl)
        elif sCmd == "biota_control_log":
            resLastBiotaControl = self.__model.subsystem_getDb().dataPatrol_getLastBiotaControl()
            if resLastBiotaControl.isOk():
                sLog = resLastBiotaControl.getData("log")
                if sLog is None:
                    sLog = "DataPatrol has not been executed yet ..."
                self.popup_text(title="Log of last BiotaControl", text=sLog, html=True, browser=True, save=True)

    def virtualAssistant_taxonSpotter(self):
        """
        VA: TaxonSpotter
        """
        if self.sys_isWorkerRunning(popup=True):
            return

        # Dialog
        dia = DialogVirtualAssistantTaxonSpotter(self.__model)
        self.ui_centerWidget(dia)
        ant = dia.exec_()
        if ant == 0:
            return
        res= dia.getConfig()
        sFile = res.getData("file")
        dicParam = {
            "file": res.getData("file"),
            "reportdir": res.getData("reportdir"),
            "sheet": res.getData("sheet"),
            "aid_given": res.getData("aid_given"),
            "sn_given": res.getData("sn_given"),
            "aid": res.getData("aid"),
            "sn": res.getData("sn"),
            "aid_acc": res.getData("aid_acc"),
            "sn_acc": res.getData("sn_acc"),
            "fuzzy": res.getData("fuzzy"),
            "like": res.getData("like"),
            "puffer": res.getData("puffer"),
            "marine_only": res.getData("marine_only"),
            "html": True
        }
        # Worker starten
        if sFile not in ["", None]:
            self.worker_start(Worker_VirtualAssistant_taxaReport(self.__model, dicParam), self.worker_finished_virtualAssistant_taxaReport)

    def virtualAssistent_taxaPolishing(self):
        """
        VA: Polish Taxa
        """

        if self.sys_isWorkerRunning(popup=True):
            return

        # Dialog
        dia = DialogVirtualAssistantPolishBiota(self.__model)
        self.ui_centerWidget(dia)
        ant = dia.exec_()
        if ant == 0:
            return

        # Dialog auswerten
        res = dia.getConfig()
        sFileOriginal = res.getData("file_original")
        sFilePolished = res.getData("file_polished")
        sBiotaSheet = res.getData("sheet_biota")
        sColAid  = res.getData("col_aid")
        sColSn  = res.getData("col_sn")
        bFuzzy = res.getData("fuzzy")
        bLike = res.getData("like")
        bBuffer = res.getData("puffer")
        bMarineOnly = res.getData("marine_only")

        # bMarineOnly = False
        # sFileOriginal = "/Users/paul/_temp/petroliere_2009.xlsx"
        # sFilePolished = "/Users/paul/_temp/petroliere_2009_polished.xlsx"
        # sBiotaSheet = "Biota"
        # sColAid  = "AphiaId"
        # sColSn  = "ScientificName"
        # bMarineOnly = True
        # bFuzzy = False
        # bLike = True
        # bBuffer = True

        # Worker starten
        if sFileOriginal not in ["",None]:
            self.worker_start(Worker_VirtualAssistant_polishTaxa(self.__model, sFileOriginal, sFilePolished, sBiotaSheet, sColAid, sColSn, bLike, bFuzzy, bMarineOnly, bBuffer), self.worker_finished_virtualAssistant_taxaPolish)

    def toolbox_showCsvFile(self):
        """
        Zeige den Inhalt einer CSV-Datei an
        """

        if self.sys_isWorkerRunning(popup=True):
            return

        sFile = QFileDialog.getOpenFileName(self, "csv file", filter="*.csv;; *.*")[0]  # Hole CSV-Datei
        if sFile != "":

            # Teste Encodung der Csv-Datei
            bErg = self.tools_checkCsvEncoding(sFile, False)
            if not bErg:
                return

            # Worker starten
            self.worker_start(Worker_ShowingCsvFile(self.__model.ui_getSmartprinter(), sFile), self.worker_finished_showTable)


    def toolbox_extractNewDataFromCsv(self):
        """
        Extrahiere neue Datensätze aus einer CSV-Datei
        """

        if self.sys_isWorkerRunning(popup=True):
            return

        # Dialog File Select
        dia = DialogCsvExtractNewData()
        self.ui_centerWidget(dia)
        ant = dia.exec_()
        if ant == 0:
            return

        # Hole Files
        sIn1 = dia.text_in1.text()
        sIn2 = dia.text_in2.text()
        sOut = dia.text_out.text()
        sHead = dia.combo_header.currentText()

        # Teste Kodierungsart der Dateien
        bErg = self.tools_checkCsvEncoding(sIn1, False)
        if not bErg:
            return
        bErg = self.tools_checkCsvEncoding(sIn2, False)
        if not bErg:
            return

        # Worker starten
        self.worker_start(Worker_ExctractNewCsvData(self.__model.ui_getSmartprinter(), sHead, sIn1, sIn2, sOut), self.worker_finished)

    def toolbox_versionControl(self):
        """
        Version control
        """

        if self.sys_isWorkerRunning(popup=True):
            return

        # Dialog File Select
        dia = DialogDiff()
        self.ui_centerWidget(dia)
        ant = dia.exec_()
        if ant == 0:
            return

        # Hole Files
        sIn1 = dia.text_in1.text()
        sIn2 = dia.text_in2.text()
        sOut = dia.text_out.text()

        # Irgendwas fehlt
        if sIn1 == "" or sIn2 == "" or sOut == "":
            return

        # Worker starten
        self.worker_start(Worker_VersionControl(self.__model.ui_getSmartprinter(), sIn1, sIn2, sOut),
                          self.worker_finished)

    def core_lock(self):
        """
        Admin Lock clicked
        """
        self.__model.conf_setAdminLock(not self.__model.conf_getAdminLock())
        self.status_refresh_adminLock()
        if self.__model.conf_getAdminLock():
            self.sys_isPsqlCommandLineToolsWorking()


    def core_fetchCredentials(self):
        """
        Admin Fetch
        """
        if self.sys_isWorkerRunning(popup=True):
            return

        sHost = self.__model.subsystem_getDb().account_single_getPart("host")
        sPort = self.__model.subsystem_getDb().account_single_getPart("port")
        sDb = self.__model.subsystem_getDb().account_single_getPart("db")
        sUser = self.__model.subsystem_getDb().account_single_getPart("user")
        sPw = self.__model.subsystem_getDb().account_single_getPart("pw")
        sSchema = self.__model.subsystem_getDb().account_single_getPart("schema")
        self.ui_setGuiAdminAccessData(sHost, sPort, sUser, sPw, sDb, sSchema)
        self.guiAdminAccessDataEdited()

    def core_sqlCmd_single(self):
        """
        SQL-Kommando abschicken
        """

        if self.sys_isWorkerRunning(popup=True):
            return

        # Befehl aufbereiten
        sSqlCmd = str(self.lineSqlCmd.currentText())
        sSqlCmd = sSqlCmd.strip()  # Whitespace entfernen
        if sSqlCmd == "":
            return

        # Hole Credentials
        sHost, sPort, sUser, sPw, sDb, sSchema = self.core_getCredentialsFromGui()
        if "" in (sHost, sPort, sUser, sPw, sDb):
            return
        sConnStr = self.__model.subsystem_getDb().conn_makeConnStr(host=sHost, user=sUser, port=sPort, db=sDb, pw=sPw)

        # Befehl merken in Liste
        self.lineSqlCmd.clearEditText()
        if sSqlCmd != "" and self.lineSqlCmd.findText(sSqlCmd) == -1:
            self.lineSqlCmd.addItem(sSqlCmd)

        # Sicherheit
        lstBadSql = ["insert", "delete", "update", "drop", "create", "grand", "execute", "trigger"]
        for sBad in lstBadSql:
            if sBad in sSqlCmd.lower():
                self.popup_error("SQL Security check",
                                 "The SQL command was not executed because it contains the following illegal parts of words:<ul><li>" + sBad + "</li></ul>")
                return

        # Worker starten
        self.worker_start(
            Worker_SqlCmd(self.__model.ui_getSmartprinter(), sConnStr, sSqlCmd),
            self.worker_finished_showTable)

    def core_showPgServerVersion(self):
        """
        Admin: Show Pg Server Version
        """


        # Hole alle Parameter
        sHost, sPort, sUser, sPw, sDb, sSchema = self.core_getCredentialsFromGui()

        # Test: Serverzugangsdaten vorhanden?
        if "" in [sHost, sPort, sUser, sPw, sDb]:
            sTitle = "Something went wrong"
            sErr = "Missing Host, Port, User or Password"
            self.popup_error(sTitle, sErr)
            return

        self.ui_showProgress()

        # Hole Version
        sConStr = self.__model.subsystem_getDb().conn_makeConnStr(user=sUser, pw=sPw, host=sHost, port=sPort, db=sDb)
        resCursor = self.__model.subsystem_getDb().conn_getConnection(constr=sConStr)
        if resCursor.isErr():
            return
        conn = resCursor.getData("conn")
        cursor = resCursor.getData("cursor")
        sVersion = self.__model.subsystem_getDb().version_getVersionConnectedPostgresServer(cursor=cursor).getData("version") if cursor is not None else None
        self.ui_hideProgress()

        # Anzeigen
        if sVersion is None:
            sVersion = "<error>"
        self.admin_pg_version.setText(str(sVersion))
        self.admin_pg_version.setCursorPosition(0)

    def core_showSchemaVersion(self):
        """
        Admin: Show Pg Server Version
        """
        # Hole alle Parameter
        sHost, sPort, sUser, sPw, sDb, sSchema = self.core_getCredentialsFromGui()

        # Test: Serverzugangsdaten vorhanden?
        if "" in [sHost, sPort, sUser, sPw, sDb]:
            sTitle = "Something went wrong"
            sErr = "Missing Host, Port, User or Password"
            self.popup_error(sTitle, sErr)
            return

        self.ui_showProgress()

        # Hole Version
        sConStr = self.__model.subsystem_getDb().conn_makeConnStr(user=sUser, pw=sPw, host=sHost, port=sPort, db=sDb)
        resCursor = self.__model.subsystem_getDb().conn_getConnection(constr=sConStr)
        if resCursor.isErr():
            return
        conn = resCursor.getData("conn")
        cursor = resCursor.getData("cursor")
        sVersion = self.__model.subsystem_getDb().version_getVersionConnectedDatamodel(cursor=cursor, schema=sSchema).getData("version")
        self.ui_hideProgress()

        # Anzeigen
        if sVersion is None:
            sVersion = "<error>"
        self.admin_schema_version.setText(str(sVersion))
        self.admin_schema_version.setCursorPosition(0)

    def core_startPsqlConsole(self):
        """
        Admin: Start psql tool
        """
        self.core_cmd("start_psql")

    def ui_consoleViewChangeOrientation(self):
        """
        Console View: Change Orientation
        """
        self.splitterConsole.setSizes([1,1])  # Show all
        iOrientation = self.splitterConsole.orientation()
        if iOrientation == Qt.Horizontal:
            self.splitterConsole.setOrientation(Qt.Vertical)
        else:
            self.splitterConsole.setOrientation(Qt.Horizontal)

    def ui_consoleViewChangeVisibility(self):
        """
        Console View: Change Visibility
        """
        lstSizesAlt = list(self.splitterConsole.sizes())
        sizeMain = lstSizesAlt[0]
        sizeConsole = lstSizesAlt[1]
        if sizeConsole == 0:
            lstSizeNeu = [1, 1]
        else:
            lstSizeNeu = [1, 0]
        self.splitterConsole.setSizes(lstSizeNeu)


    def core_snapshot(self):
        """
        Snapshot clicked
        """
        status = self.__model.ui_getSmartprinter().status

        if not self.sys_isPsqlCommandLineToolsWorking():
            return

        # Hole die Zugangsdaten aus aktivem account
        sHost = self.__model.subsystem_getDb().account_single_getPart("host")
        sPort = self.__model.subsystem_getDb().account_single_getPart("port")
        sDb = self.__model.subsystem_getDb().account_single_getPart("db")
        sUser = self.__model.subsystem_getDb().account_single_getPart("user")
        sPw = self.__model.subsystem_getDb().account_single_getPart("pw")
        sSchema = self.__model.subsystem_getDb().account_single_getPart("schema")

        sTitle = "Snapshot"
        sSnapshotFolder = self.__model.conf_getAppSnapshotDirSchema()
        sInfo = '<html><head/><body><p>Create a snapshot of schema<br><b>' + str(sHost) + "." + str(sDb) + "." + str(sSchema) + '</b><br>and save it in <a href="file:///' + str(sSnapshotFolder) + '"><span style=" text-decoration: underline; color:#0000ff;">snapshot folder</span></a></p></body></html>'
        sErr = ""

        # Test: Serverzugangsdaten vorhanden?
        if "" in [sHost, sPort, sUser, sPw, sDb, sSchema]:
            sErr = "Missing host, port, user, database, schema or Password"

        # Error -> Ende
        if sErr != "":
            # Error
            self.popup_error("Error: " + sTitle, sErr)
            return

        # Frage
        ant = self.popup_generic(title="Snapshot", head=sTitle, info=sInfo, button_yes="Yes", button_no="No")
        if ant == 0:
            return

        # Komnando zusammenbauen
        dicCmd = {}
        dicCmd["cmd"] = "snapshot"
        dicCmd["title"] = sTitle
        dicCmd["info"] = sInfo
        dicCmd["db_host"] = sHost
        dicCmd["db_port"] = sPort
        dicCmd["db_name"] = sDb
        dicCmd["db_user"] = sUser
        dicCmd["db_pw"] = sPw
        dicCmd["db_schema"] = sSchema
        dicCmd["popup"] = True

        # Start Worker
        self.worker_start(Worker_AdminCmd(self.__model, dicCmd), self.worker_finished_adminCmd)


    def core_listAllUsers(self):
        """
        Admin List All Users
        """
        self.core_cmd("list_users")

    def core_listAllUsersMarlin(self):
        """
        Admin List All Marlin Users
        """
        self.core_cmd("list_users_marlin")

    def core_listAllSchemas(self):
        """
        Admin List All Schemas
        """
        self.core_cmd("list_schemas")

    def core_listAllDatabases(self):
        """
        Admin List all Databases
        """
        self.core_cmd("list_databases")

    def fao_regions_create(self):
        """
        Admin Create Regions
        """

    def core_sqlCmd_script(self):
        """
        Admin Start SQL Script
        """
        self.core_cmd("sql_script")

    def core_restoreSchema(self):
        """
        Admin Restore Schema
        """

        self.core_cmd("restore_schema")

    def core_backupSchema(self):
        """
        Admin Backup Schema
        """
        self.core_cmd("backup_schema")

    def core_renameSchema(self):
        """
        Admin Rename Schema
        """
        self.core_cmd("rename_schema")

    def core_deleteSchema(self):
        """
        Admin Delete Schema
        """
        self.core_cmd("del_schema")

    def core_CreateNewSchema(self):
        """
        Admin Create new empty Schema
        """

        sErrInfo = "ERROR: function public.st_point(double precision, double precision) does not exist\n"
        sErrInfo += "ERROR: role critterread does not exist\n"
        sErrInfo += "ERROR: role geoserver does not exist"

        #self.popupInfo("Possible Error Messages", "During the installation of an empty CRITTERBASE schema, three types of error messages could occur due to missing users geoserver and critterread (open details for more information). These are only relevant if the schema should be used by the AWI CRITTERBASE webservice. Otherwise, ignore them ...", sErrInfo)

        self.core_cmd("new_schema")

    def core_cmd(self, sCmd):
        """
        Verarbeite den Admin Befehl und führe ihn aus
        """
        if self.sys_isWorkerRunning(popup=True):
            return
        if not self.sys_isPsqlCommandLineToolsWorking():
            return

        # Hole alle Parameter
        sHost, sPort, sUser, sPw, sDb, sSchema = self.core_getCredentialsFromGui()
        sSchemaBackupFile = self.admin_cmd_backup_schema__file.text()
        sSchemaRestoreFile = self.admin_cmd_restore_schema__file.text()
        sSchemaRenameName = self.admin_cmd_rename_schema__schema_new.text()
        sSchemaNew = self.admin_cmd_new_schema__critterbase.text()

        # Werte Befehl aus
        sErr = ""
        sInfo = ""
        sTitle = ""

        # Komnando zusammenbauen (immer alles senden)
        dicCmd = {}
        dicCmd["cmd"] = sCmd
        dicCmd["title"] = None
        dicCmd["info"] = None
        dicCmd["sqlscript"] = None
        dicCmd["db_schema_new"] = sSchemaNew
        dicCmd["db_host"] = sHost
        dicCmd["db_port"] = sPort
        dicCmd["db_user"] = sUser
        dicCmd["db_pw"] = sPw
        dicCmd["db_name"] = sDb
        dicCmd["db_schema"] = sSchema
        dicCmd["db_schema_backup_file"] = sSchemaBackupFile
        dicCmd["db_schema_restore_file"] = sSchemaRestoreFile
        dicCmd["db_schema_rename_name"] = sSchemaRenameName
        dicCmd["popup"] = False

        # Test: Serverzugangsdaten vorhanden?
        if "" in [sHost, sPort, sUser, sPw, sDb]:
            sTitle = "Something went wrong"
            sInfo = "Something went wrong"
            sErr = "Missing Host, Port, User or Password"

        # Werte Befehle aus
        if sErr == "":
            if sCmd == "start_psql":
                # Psql ---
                sTitle = "Start admin tool"
                sInfo = "Start admin tool"
            elif sCmd == "list_databases":
                # List Database ---
                sTitle = "List postres databases"
                sInfo = "List all postgres databases of host<br><b>" + str(sHost)+ ":" +str(sPort) +"</b><br>"
            elif sCmd == "list_schemas":
                # List Schema ---
                sTitle = "List schemas"
                sInfo = "List all schemas in database<br><b>" + str(sDb) + "@" + str(sHost)+ ":" +str(sPort) +"</b><br>"
            elif sCmd == "list_users":
                # List Users ---
                sTitle = "List users"
                sInfo = "List all postgres users of host<br><b>" + str(sHost) + ":" +str(sPort)+"</b><br>"
            elif sCmd == "list_users_marlin":
                # List Users ---
                sTitle = "List all CRITTERBASE users"
                sInfo = "List all CRITTERBASE users in database<br><b/>" + str(sDb) + "@" + str(sHost) + ":" +str(sPort)+"</b><br>"
            elif sCmd == "sql_script":
                # SQL Script ---
                sPathSql = self.__model.conf_getAppSqlScriptsDir()
                sSqlFile = QFileDialog.getOpenFileName(self, "Select SQL file to execute", directory=sPathSql, filter="*.sql;; *.*")[0]
                if sSqlFile == "" or sSqlFile is None:
                    return
                try:
                    f = open(sSqlFile, 'r')
                    sText = f.read()
                except:
                    return
                # Dialog
                sTitel = sSqlFile
                dia = QDialog()
                dia.setWindowTitle(sTitel)
                dia.setWindowModality(Qt.ApplicationModal)
                dia.setWindowFlags(Qt.Dialog | Qt.CustomizeWindowHint | Qt.WindowTitleHint | Qt.WindowCloseButtonHint)
                vbox = QVBoxLayout(dia)
                vbox.setContentsMargins(4, 4, 4, 4)
                vbox.setSpacing(6)
                text = QTextEdit()

                text.setPlainText(sText)
                font = QFont()
                font.setFamily("Courier")
                font.setPointSize(10)
                text.setFont(font)
                text.setLineWrapMode(QTextEdit.NoWrap)
                text.setReadOnly(True)
                vbox.addWidget(text)
                buttonBox = QDialogButtonBox()
                btnOk = QPushButton()
                btnOk.setText("Execute SQL script")
                btnOk.clicked.connect(lambda: QDialog.done(dia, 1))
                btnNo = QPushButton()
                btnNo.setText("Cancel")
                btnNo.clicked.connect(lambda: QDialog.done(dia, 0))
                buttonBox.accepted.connect(dia.accept)
                buttonBox.addButton(btnOk, QDialogButtonBox.NoRole)
                buttonBox.addButton(btnNo, QDialogButtonBox.NoRole)
                vbox.addWidget(buttonBox)
                dia.resize(800, 600)
                self.ui_centerWidget(dia)
                ant = dia.exec()
                # Antwort
                if ant == 0:
                    return
                dicCmd["sqlscript"] = sSqlFile
            elif sCmd == "restore_schema":
                # Restore Schema ---
                sTitle = "Restore schema"
                sInfo = "Restore schema from file<br><b>"+str(FileTools.getFileWithoutPath(sSchemaRestoreFile))+ "</b><br>to database<br><b>" + str(sDb) + "@" + str(sHost)+ ":" + str(sPort)+"</b>"
                if sSchemaRestoreFile == "":
                    sErr = "Missing restore file"
                if sErr == "":
                    ant = self.popup_generic(title="Core", head=sTitle, info=sInfo, button_yes="Yes", button_no="No")
                    if ant == 0:
                        return
            elif sCmd == "backup_schema":
                # Backup Schema ---
                sTitle = "Backup schema"
                sInfo = "Backup schema of database<br><b>" + str(sDb) + "@" + str(sHost)+ ":" +str(sPort) +"</b><br> to file<br><b>" + str(FileTools.getFileWithoutPath(sSchemaBackupFile))+"</b><br>"
                if sSchemaBackupFile == "":
                    sErr = "Missing backup file"
                if sErr == "":
                    ant = self.popup_generic(title="Core", head=sTitle, info=sInfo, button_yes="Yes", button_no="No")
                    if ant == 0:
                        return
            elif sCmd == "rename_schema":
                # Rename Schema ---
                sTitle = "Rename schema"
                sInfo = "Rename schema<br><b>" +str(sSchema)+"."+str(sDb) + "@" + str(sHost)+ ":" +str(sPort) +"</b><br>to<br><b>" + str(sSchemaRenameName)+"</b><br>"
                if sSchemaRenameName == "":
                    sErr = "Missing new name"
                if sErr == "":
                    ant = self.popup_generic(title="Core", head=sTitle, info=sInfo, button_yes="Yes", button_no="No")
                    if ant == 0:
                        return
            elif sCmd == "del_schema":
                # Delete Schema ---
                sTitle = "Delete Schema"
                sInfo = "Delete schema<br><b>" +str(sSchema)+"."+str(sDb) + "@" + str(sHost)+ ":" +str(sPort)+"</b><br>"
                if sErr == "":
                    ant = self.popup_generic(title="Core", head=sTitle, info=sInfo, button_yes="Yes", button_no="No")
                    if ant == 0:
                        return
            elif sCmd == "new_schema":
                # New Schema ---
                sTitle = "New schema"
                sInfo = "Create new schema<br><b>" + str(sSchemaNew) + "."+str(sDb)+"@" + str(sHost) + ":" +str(sPort)+"</b><br>"
                sErr = ""
                if sSchemaNew == "":
                    sErr += "Missing new schema name"
                if sSchemaNew == "cb_empty":
                    sErr += "Schema name \'cb_empty\' is restricted.\n"
                if sErr == "":
                    resCbTemplate = self.__model.subsystem_getDb().check_isSchemaExisting(schema="cb_empty")
                    bCbTemplate = resCbTemplate.getData("exist") if resCbTemplate.isOk() else False
                    resCbNeu = self.__model.subsystem_getDb().check_isSchemaExisting(schema=sSchemaNew)
                    bCbNeu = resCbNeu.getData("exist") if resCbNeu.isOk() else False
                    if True in [bCbTemplate, bCbNeu]:
                        sErr = "Could not create new empty critterbase:\n"
                        if bCbTemplate:
                            sErr += "Schema \'cb_empty\' is existing (this is a restricted schema name - used for temporal purpose).\n"
                        if bCbNeu:
                            sErr += "Schema \'"+str(sSchemaNew)+"\' is already existing.\n"
                        sErr += "Please delete or choose other name."
                if sErr == "":
                    ant = self.popup_generic(title="Core", head=sTitle, info=sInfo, button_yes="Yes", button_no="No")
                    if ant == 0:
                        return
            # elif sCmd == "sql_cmd":
            #     # Exec Single Command ---
            #     sTitle = "Execute SQL command"
            #     sInfo = "Execute SQL command"
            #     if sSchemaRenameName == "":
            #         sErr = "Missing new name"
            #     if sErr == "":
            #         ant = self.popup_frage(sTitle, sInfo)
            #         if ant == 0:
            #             return
            elif sCmd == "create_regions":
                sTitle = "Create Regions"
                sInfo = "Harvest FAO features"
            else:
                # Uuups ---
                sTitle = "Internal error"
                sInfo = "Internal error"
                sErr = "Unknown command:" + str(sCmd)

        # Kommando: Titel/Info eintragen
        dicCmd["title"] = sTitle
        dicCmd["info"] = sInfo

        # Error -> Ende
        if sErr != "":
            # Error
            self.popup_error(sTitle, sErr)
            return
        self.admin_out.setPlainText("WORKING ...")
        # Start Worker
        self.worker_start(Worker_AdminCmd(self.__model, dicCmd), self.worker_finished_adminCmd)

    def taxon_cmd(self, sCmd, **kwargs):
        """
        Starte ein Taxon System Kommando
        Kwargs:
            gold (bool): wird eine Tabelle im (not)gold Modus geöffnet
        """
        pn = self.__smartprinter.normal
        bGold = kwargs["gold"] if ("gold" in kwargs and kwargs["gold"] in (True, False)) else None

        if sCmd == "worms-base-edit":
            self.table_show("wormsbase", gold=bGold, gold_edit=True)
        elif sCmd == "worms-aaid-edit":
            self.table_show("wormsaaid")
        elif sCmd == "worms-private-edit":
            self.table_show("wormsprivate")
        elif sCmd == "worms-colony-edit":
            self.table_show("wormscolony")
        elif sCmd == "noworms-edit":
            self.table_show("noworms")
        elif sCmd == "change-edit":
            self.table_show("wormschange")
        elif sCmd == "update":
            if self.sys_isWorkerRunning(popup=True):
                return
            if not self.__model.subsystem_getConnector().isReady():
                self.popup_offline()
                return
            self.worker_start(Worker_TaxonWizard_update(self.__model.subsystem_getTaxonWizard()), self.worker_finished_updateTaxonList)
        elif sCmd == "puffer-clean":
            if self.sys_isWorkerRunning(popup=True):
                return
            if self.__model.subsystem_getMyWorms().puffer_getSizeTaxa() > 0:
                ant = self.popup_generic(title="MyWoRMS", head="Clean MyWoRMS", info="Wipe my running My WoRMS now?", button_yes="Yes", button_no="No")
                if ant == 0:
                    return
            self.__model.subsystem_getMyWorms().puffer_cleanDisk()
            self.status_refresh_myWorms()
            self.ui_setStatusbarText("WoRMS buffer cleaned")
        elif sCmd == "puffer-clean-taxa":
            if self.sys_isWorkerRunning(popup=True):
                return
            if self.__model.subsystem_getMyWorms().puffer_getSizeTaxa() > 0:
                ant = self.popup_generic(title="MyWoRMS", head="Clean MyWoRMS Taxa", info="Clean all taxa in MyWoRMS now?", button_yes="Yes", button_no="No")
                if ant == 0:
                    return
            self.__model.subsystem_getMyWorms().puffer_cleanDisk(taxa=True)
            self.status_refresh_myWorms()
            self.ui_setStatusbarText("WoRMS buffer cleaned")
        elif sCmd == "puffer-clean-query":
            if self.sys_isWorkerRunning(popup=True):
                return
            if self.__model.subsystem_getMyWorms().puffer_getSizeTaxa() > 0:
                ant = self.popup_generic(title="MyWoRMS", head="Clean MyWoRMS Queries", info="Clean all Queries in MyWoRMS now?", button_yes="Yes", button_no="No")
                if ant == 0:
                    return
            self.__model.subsystem_getMyWorms().puffer_cleanDisk(query=True)
            self.status_refresh_myWorms()
            self.ui_setStatusbarText("WoRMS buffer cleaned")

        elif sCmd == "puffer-reload":
            if self.sys_isWorkerRunning(popup=True):
                return
            # if self.__model.getTaxon().puffer_getSize() > 0:
            #     ant = self.popupFrage("Reload My WoRMS", "Reload My WoRMS buffer now?")
            #     if ant == 0:
            #         return
            resLoad = self.__model.subsystem_getMyWorms().puffer_load()
            if resLoad.isOk():
                sFile = resLoad.getData("file")
                date = resLoad.getData("date")
                pn("* My WoRMS reloaded from disk", fmt=False)
                pn("file: " + str(sFile))
                pn("created: " + str(date))
                self.status_refresh_myWorms()
                self.ui_setStatusbarText("MyWoRMS buffer reloaded from disk")
            else:
                self.status_refresh_myWorms()
                self.popup_error("Loading My WoRMS buffer", "Errors occured while loading My WoRMS buffer", resLoad.getResult(short=True))
        elif sCmd == "puffer-save":
            if self.sys_isWorkerRunning(popup=True):
                return
            resSave = self.__model.subsystem_getMyWorms().puffer_save()
            if resSave.isOk():
                sFile = resSave.getData("file")
                date = resSave.getData("date")
                if resSave.getData("save_not_needed") is True:
                    pn("* MyWoRMS not saved to disk", fmt=False)
                    pn("No changes: no need to save")
                else:
                    pn("* MyWoRMS saved to disk", fmt=False)
                    pn("file: " + str(sFile))
                    pn("created: " + str(date))
                self.status_refresh_myWorms()
                self.ui_setStatusbarText("My WoRMS buffer saved to disk")
            else:
                self.status_refresh_myWorms()
                self.popup_error("Saving My WoRMS buffer", "Errors occured while saving My WoRMS buffer", resSave.getResult(short=True))
        elif sCmd == "puffer-snapshot":
            if self.sys_isWorkerRunning(popup=True):
                return
            resSnap= self.__model.subsystem_getMyWorms().puffer_snapshot()
            if resSnap.isOk():
                sFileName = resSnap.getData("snapshot-filename")
                sFile = resSnap.getData("snapshot-file")
                sFolder= resSnap.getData("snapshot-folder")
                # Snapshot
                sInfo = '<html><head/><body><p>Snapshot<br><b>' + str(
                    sFileName) + '</b><br>was saved in <a href="file:///' + str(
                    sFolder) + '"><span style=" text-decoration: underline; color:#0000ff;">snapshot folder</span></a></p><</body></html>'
                self.popup_generic(title="Snaphot", head="Snapshot of MyWoRMS", info=sInfo)
            else:
                self.popup_error("Snapshot of My WoRMS", "Errors occured while taking a snapshot of My WoRMS", resSnap.getResult(short=True))
        elif sCmd == "puffer-import":
            if self.sys_isWorkerRunning(popup=True):
                return
            #ant = self.popupFrage("Import My WoRMS","Override running My WoRMS now?")
            #if ant == 0:
            #    return
            sFile = QFileDialog.getOpenFileName(self, "Select file", directory=self.__model.conf_getAppSnapshotDirMyWorms(), filter="*.*")[0]
            if sFile == "":
                return
            sFileName = FileTools.getFileWithoutPath(sFile)
            resImport = self.__model.subsystem_getMyWorms().puffer_import(sFile)
            if resImport.isOk():
                # Puffer laden
                resLoad = self.__model.subsystem_getMyWorms().puffer_load()
                if resLoad.isOk():
                    sFile = resLoad.getData("file")
                    date = resLoad.getData("date")
                    pn("* My WoRMS loaded from disk", fmt=False)
                    pn("file: " + str(sFile))
                    pn("created: " + str(date))
                    self.status_refresh_myWorms()
                    sInfo = '<html><head/><body><p>A new instance of My WoRMS was imported from file<br><b>' + str(
                        sFileName) + '<b><br></body></html>'
                    self.popup_generic(title="MyWoRMS", head="Import MyWoRMS", info=sInfo)
                else:
                    self.popup_error("Loading MyWoRMS buffer", "Errors occured while loading MyWoRMS buffer", resLoad.getErrWarnInfoString(short=True))
            else:
                self.popup_error("Import of My WoRMS", "Errors occured while importing My WoRMS", resImport.getResult(short=True))
        elif sCmd == "puffer-export":
            if self.sys_isWorkerRunning(popup=True):
                return
            sFileName = QFileDialog.getSaveFileName(self, "Save copy of My WoRMS to file", "worms.dat")[0]
            if sFileName == "":
                return
            resExport = self.__model.subsystem_getMyWorms().puffer_export(sFileName)
            if resExport.isOk():
                sFileName = resExport.getData("file")
                sFilePath = resExport.getData("ordner")
                sInfo = '<html><head/><body><p>A copy of My WoRMS<br><b>'+str(sFileName)+'</b><br>was exported to <a href="file:///' + str(sFilePath) + '"><span style=" text-decoration: underline; color:#0000ff;">here</span></a></p></body></html>'
                self.popup_generic(title="MyWoRMS", head="Export of MyWoRMS", info=sInfo)
            else:
                self.popup_error("Export of My WoRMS", "Errors occured while exporting My WoRMS", resExport.getResult(short=True))
        elif sCmd == "puffer-update":
            if self.sys_isWorkerRunning(popup=True):
                return
            self.worker_start(Worker_MyWorms_antiage(self.__model.subsystem_getMyWorms()), self.worker_finished_antiageMyWoRMS)
        elif sCmd == "puffer-status":
            if self.sys_isWorkerRunning(popup=True):
                return
            if self.__model.subsystem_getMyWorms().puffer_getSizeTaxa() == 0:
                return
            sStatus = "*** MyWoRMS - Buffered Taxa\n\n"
            sStatus += self.__model.subsystem_getMyWorms().puffer_getStatusString(taxa=True)
            res = Result()
            res.addData("title", "MyWoRMS - Buffered Taxa")
            res.addData("text", sStatus)
            res.addData("title_capitalize", False)
            self.worker_finished_showText(res)
        elif sCmd == "query-status":
            if self.sys_isWorkerRunning(popup=True):
                return
            if self.__model.subsystem_getMyWorms().puffer_getSizeQuery() == 0:
                return
            sStatus = "*** MyWoRMS - Buffered Name Queries\n\n"
            sStatus += self.__model.subsystem_getMyWorms().puffer_getStatusString(search=True)
            res = Result()
            res.addData("title", "MyWoRMS - Buffered Name Queries")
            res.addData("text", sStatus)
            res.addData("title_capitalize", False)
            self.worker_finished_showText(res)
        elif sCmd == "misfits-status":
            if self.sys_isWorkerRunning(popup=True):
                return
            if self.__model.subsystem_getMyWorms().puffer_getSizeProblems() == 0:
                return
            sStatus = "*** MyWoRMS - Misfits\n\n"
            sStatus += self.__model.subsystem_getMyWorms().puffer_getStatusString(misfits=True)
            res = Result()
            res.addData("title", "MyWoRMS - Misfits")
            res.addData("text", sStatus)
            res.addData("title_capitalize", False)
            self.worker_finished_showText(res)
        elif sCmd == "search-aid":
            sAid = self.mywormsSearchAid_input.text()
            bPuffer = self.mywormsSearchAid_buffer.isChecked()
            try:
                iAid = int(sAid)
                self.worker_start(Worker_MyWorms_searchManualByAphiaId(self.__model, iAid, bPuffer), self.worker_finished_showText)
            except:
                pass

        elif sCmd == "search-name":
            sName = self.mywormsSearchName_input.text()
            iFuzzy = self.mywormsSearchName_fuzzy.checkState()
            bPuffer = self.mywormsSearchName_buffer.isChecked()
            bFuzzy = False
            bLike = False
            if iFuzzy == 1:
                bLike = True
            elif iFuzzy == 2:
                bFuzzy = True
            if sName not in (None, ""):
                self.worker_start(Worker_MyWorms_searchManualByName(self.__model, sName, bFuzzy, bLike, bPuffer), self.worker_finished_showText)


    def main_beenden(self, bRestart=False):
        """
        Programm beenden
        """
        pn = self.__smartprinter.normal
        if bRestart:
            print("Restarting App ...")
            pn("Restarting App ...")
            try:
                # Restart: Starte vor Beendigung eine neue Instanz des Programmes
                sOs = self.__model.conf_getOs()
                dicEnv = os.environ
                if sOs == "win":
                    # Windows
                    sCmd = "python " + self.__model.conf_getAppBaseDir() + "/start.pyw"
                    print("Restart (Win): ", sCmd)
                    pn("Restart (Win): " +str(sCmd))
                    subprocess.Popen(sCmd.split(), close_fds=True, shell=True) # TODO: Shell True entfernen
                elif sOs == "mac":
                    # Mac
                    sFront, sParent = FileTools.cutAtPathTail(self.__model.conf_getAppBaseDir())
                    sMacAppContainerBase, sGrandparent = FileTools.cutAtPathTail(sFront)
                    sCmd = 'open -n -a ' + sMacAppContainerBase
                    print("Restart (Mac):", sCmd)
                    pn("Restart (Mac): " + str(sFront)+" "+str(sParent)+" "+str(sGrandparent)+" -> "+str(sCmd))
                    subprocess.run(sCmd.split(), close_fds=True)
                elif sOs == "lin":
                    # Linux
                    sCmd = self.__model.conf_getAppBaseDir() + "/critterbase.py -gui fusion"
                    print("Restart (Linux): ", sCmd)
                    pn("Restart (Linux): " + str(sCmd))
                    subprocess.Popen(sCmd.split(), close_fds=True)
            except Exception as ex:
                # Error
                print("Problems restarting App:")
                print(str(ex))
                pn("Error while restarting: " + str(ex))
                self.popup_infoTimer(sTitle="Restart of program failed", sText="Please restart manually.", iTimer=15)
        # Quit
        self.__model.subsystem_getMyWorms().puffer_save()
        self.close()

    def guiConnectionDry(self):
        """
        Toggle Connection Dry
        """

        self.__model.conf_setDry(not self.__model.conf_isDry())
        self.status_refresh_dryMode()
        self.status_refresh_toolbarMini()
        self.status_refresh_system()

        if self.__model.conf_isDry():
            self.ui_setStatusbarText("Dry mode is acivated. Nothing of what you do has consequences ...")
        else:
            self.ui_setStatusbarText("Dry mode is deacivated. Everything is for real now ...")

    def console_changeVerbosity(self):
        """
        Change verbosity
        """

        status = self.ui_setStatusbarText
        iVb = self.comboVerbosity.currentIndex()
        if iVb == 0:
            self.__model.conf_setVerbosity(False, False)
            status("Verbosity level changed to: normal")
        elif iVb == 1:
            self.__model.conf_setVerbosity(True, False)
            status("Verbosity level changed to: verbose")
        elif iVb == 2:
            #self.__popupGeneric(title="Information flood", text="Too much information for a normal user",textExtra="This level of detail is intended for quality control or troubleshooting.", icon="info")
            self.__model.conf_setVerbosity(False, True)
            status("Verbosity level changed to: debug")

    def console_clear(self):
        """
        Konsole-Widget löschen
        """

        self.textBox.clear()

    def console_save(self):
        """
        Konsolenasugabe abspeichern
        """
        if self.sys_isWorkerRunning(popup=True):
            return

        sText = self.textBox.toPlainText()
        sFilename = QFileDialog.getSaveFileName(self, "Save console output in file", "console_log.txt")[0]
        if sFilename == "":
            return

        # Worker starten
        saveConsole = Worker_SaveDataToFile(self.__model.ui_getSmartprinter(), sText, sFilename)
        self.worker_start(saveConsole, self.worker_finished_saveDataToFile)

    def status_update_batchBot(self):
        self.status_clean_lookups()
        self.status_refresh_ingestion_overview()

    def status_clean_taxonWizard(self):
        """
        Clear Status: Taxon System Tabellen-Inhalte
        """
        for sTabName, (sName, widget) in self.__odicTaxonSystemData.items():
            widget.setText("-")
        iPufferSize = self.__model.subsystem_getMyWorms().puffer_getSizeTaxa()
        sPufferSize = "-" if iPufferSize == 0 else str(iPufferSize)
        self.taxSysPufferInfo.setText(sPufferSize)
        iPufferSizeQuery = self.__model.subsystem_getMyWorms().puffer_getSizeQuery()
        sPufferSizeQuery = "-" if iPufferSizeQuery == 0 else str(iPufferSizeQuery)
        self.taxSysQueryInfo.setText(sPufferSizeQuery)
        self.taxSysBaseProgress.setFormat("")
        self.taxSysBaseProgress.setValue(0)
        self.taxSysBaseProgress.setStyleSheet(CssPresets.progressBar_grey)
        self.taxSysAgeProgress.setFormat("")
        self.taxSysAgeProgress.setValue(0)
        self.taxSysAgeProgress.setToolTip("Days till last update of Taxon List")
        self.taxSysAgeProgress.setStyleSheet(CssPresets.progressBar_grey)
        self.taxSysAgeStatus.setIcon(QIcon(os.path.dirname(os.path.realpath(__file__)) + "/ressource/icon_frage.png"))
        self.taxSysNoTaxList.setText("-")
        self.taxSysWarn.setText("-")
        self.taxSysWarn.setPalette(QPalette())
        self.taxSysWarn.setToolTip("Warnings")

    def status_clean_dataPatrol(self):
        """
        Clear Status: Data Patrol
        """
        self.dataPatrol_biotaControl_lastCheckWarn.setText("")
        self.dataPatrol_biotaControl_lastCheckWarn.setPalette(QPalette())
        self.dataPatrol_biotaControl_lastCheckWarn.setToolTip("Warnings")
        self.dataPatrol_biotaControl_lastCheck.setText("")
        self.dataPatrol_biotaControl_lastCheckLog.setIcon(QIcon(os.path.dirname(os.path.realpath(__file__)) + "/ressource/icon_frage.png"))

    def status_clean_ingest(self):
        for sTabName, widget in self.__dicStatus_widgets_ingest.items():
            widget.setToolTip(None)
            widget.setText("-")

    def status_clean_fao(self):
        self.fao_regions_count.setText("-")
        self.fao_mapping_count.setText("-")
        self.fao_leftover_count.setText("-")
        self.fao_new_count.setText("-")
        widget = self.fao_leftover_proz
        widget.setFormat("")
        widget.setValue(0)
        widget.setToolTip("leftover | all")
        widget = self.fao_new_proz
        widget.setFormat("")
        widget.setValue(0)
        widget.setToolTip("new/unmapped | all")
        self.fao_mapping_stat.setIcon(QIcon(os.path.dirname(os.path.realpath(__file__)) + "/ressource/icon_frage.png"))

    def status_clean_lookups(self):
        for sTabName, widget in self.__dicStatus_widgets_lookup.items():
            widget.setToolTip(None)
            widget.setText("-")
        for sTabName, widget in self.__dicStatus_widgets_lookup_status.items():
            if sTabName != "taxon":
                widget.setIcon(QIcon(os.path.dirname(os.path.realpath(__file__)) + "/ressource/icon_frage.png"))
        for sTabName, widget in self.__dicStatus_widgets_lookup_count.items():
            widget.setFormat("")
            widget.setValue(0)
            widget.setToolTip("used | not used")
        for sTabName, widget in self.__dicStatus_widgets_lookup_gold.items():
            widget.setToolTip("gold | not-gold")
            widget.setFormat("")
            widget.setValue(0)
        for sTabName, widget in self.__dicStatus_widgets_lookup_notUsedNotGoldLookup.items():
            widget.setToolTip("not-used not-gold [purge?]")
            widget.setPalette(QPalette())
            widget.setText("-")
        for sTabName, widget in self.__dicStatus_widgets_lookup_usedNotGoldLookup.items():
            widget.setToolTip("used not-gold [check?]")
            widget.setPalette(QPalette())
            widget.setText("-")
        for sTabName, widget in self.__dicStatus_widgets_lookup_special.items():
            widget.setToolTip("special info")
            widget.setPalette(QPalette())
            widget.setText("-")
        widget = self.__widgetStatus_widgets_lookup_dataset_allFreeRequest
        widget.setToolTip("free/request | restricted")
        widget.setFormat("")
        widget.setValue(0)
        widget = self.__widgetStatus_widgets_lookup_dataset_allUsedFreeRequest
        widget.setToolTip("used and free/request | used and restricted")
        widget.setFormat("")
        widget.setValue(0)
        widget = self.__widgetStatus_widgets_lookup_dataset_allUsedNotGoldFreeRequest
        widget.setToolTip("used not-gold and free/request [check?]")
        widget.setPalette(QPalette())
        self.__widgetStatus_widgets_lookup_dataset_allUsedNotGoldFreeRequest.setText("-")
        widget = self.__widgetStatus_widgets_lookup_dataset_DOI_OkNotOk
        widget.setPalette(QPalette())
        widget.setToolTip("DOI ok | DOI not ok")
        widget.setFormat("")
        widget.setValue(0)
        widget = self.__widgetStatus_widgets_lookup_dataset_DOI_UsedOkNotOk
        widget.setPalette(QPalette())
        widget.setToolTip("used DOI ok | used DOI not ok")
        widget.setFormat("")
        widget.setValue(0)
        widget = self.__widgetStatus_widgets_lookup_dataset_DOI_UsedNotOkReachable
        widget.setToolTip("used free/request DOI not ok [bad doi online!]")
        widget.setPalette(QPalette())
        widget.setText("-")
        widget = self.__widgetStatus_widgets_lookup_dataset_DOI_Pangaea
        widget.setToolTip("DOI PANGAEA")
        widget.setPalette(QPalette())
        widget.setText("-")
        widget = self.__widgetStatus_widgets_lookup_dataset_DOI_PangaeaNot
        widget.setToolTip("DOI not from PANGAEA [really raw data available?]")
        widget.setPalette(QPalette())
        widget.setText("-")
        self.__widgetStatus_widgets_lookup_dataset_checkDois_status.setIcon(QIcon(
            os.path.dirname(
                os.path.realpath(__file__)) + "/ressource/icon_frage.png"))

    def ui_tabWasSwitched(self):
        """
        Tab wurde verändert
        """

        self.__model.conf_setAdminLock(False)
        self.status_refresh_adminLock()

    def core_getCredentialsFromGui(self):
        """
        Hole Admin-Zugangsdaten
        """

        sHost = self.admin_db_host.text()
        sPort = self.admin_db_port.text()
        sUser = self.admin_db_user.text()
        sPw = self.admin_db_pw.text()
        sDb = self.admin_db_name.text()
        sSchema = self.admin_db_schema.text()

        return sHost, sPort, sUser, sPw, sDb, sSchema

    def ui_setGuiAdminAccessData(self, sHost, sPort, sUser, sPw, sDb, sSchema):
        """
        Setze Admin-Zugangsdaten
        """

        lstWidgets = [self.admin_db_host, self.admin_db_port, self.admin_db_user, self.admin_db_pw, self.admin_db_name, self.admin_db_schema]

        for widget in lstWidgets:
            widget.blockSignals(True)

        self.admin_db_host.setText(sHost)
        self.admin_db_port.setText(sPort)
        self.admin_db_user.setText(sUser)
        self.admin_db_pw.setText(sPw)
        self.admin_db_name.setText(sDb)
        self.admin_db_schema.setText(sSchema)

        # Text-Cursor = 0
        self.admin_db_host.setCursorPosition(0)
        self.admin_db_port.setCursorPosition(0)
        self.admin_db_user.setCursorPosition(0)
        self.admin_db_pw.setCursorPosition(0)
        self.admin_db_name.setCursorPosition(0)
        self.admin_db_schema.setCursorPosition(0)

        for widget in lstWidgets:
            widget.blockSignals(False)

        return sHost, sPort, sUser, sPw, sDb, sSchema

    def ui_setRestoreSchemaFile(self):
        """
        Setze Datei für Restore Schema
        """
        if self.sys_isWorkerRunning(popup=True):
            return
        sPath = QFileDialog.getOpenFileName(self, "Select file", directory=self.__model.conf_getAppSnapshotDirSchema(), filter="*.*")[0]
        self.admin_cmd_restore_schema__file.setText(sPath)
        self.admin_cmd_restore_schema__file.setCursorPosition(0)

    def ui_setBackupSchemaFile(self):
        """
        Setze Datei für Backup Schema
        """
        if self.sys_isWorkerRunning(popup=True):
            return

        # Hole alle Parameter
        sHost, sPort, sUser, sPw, sDb, sSchema = self.core_getCredentialsFromGui()

        # Baue vorgefertigten Backup Filenamen zusammen
        sBackupfile = "backup_" + PrettyText.createTimestamp() + ".sql.zip"
        if "" not in [sHost, sDb, sSchema]:

            # Hole Schema Version
            sConStr = self.__model.subsystem_getDb().conn_makeConnStr(user=sUser, pw=sPw, host=sHost, port=sPort, db=sDb)
            resCursor = self.__model.subsystem_getDb().conn_getConnection()
            if resCursor.isErr():
                return
            conn = resCursor.getData("conn")
            cursor = resCursor.getData("cursor")
            sSchemaVersion = self.__model.subsystem_getDb().version_getVersionConnectedDatamodel(cursor=cursor).getData("version")

            # Baue den Filenamen zusammen
            sHostFilename = sHost.replace(".", "_")
            if sSchemaVersion is not None:
                sSchemaVersion = sSchemaVersion.replace(".", "-")
                sSchemaVersion = "_v" + sSchemaVersion
            else:
                sSchemaVersion = ""
            sBackupfileName = sHostFilename + "_" + sDb + "_" + sSchema + sSchemaVersion + "_" + PrettyText.createTimestamp() + ".sql.zip"
            sBackupfile = self.__model.conf_getAppSnapshotDirSchema() + "/" + sBackupfileName

        sPath = QFileDialog.getSaveFileName(self, "Select file", directory=sBackupfile, filter="*.*")[0]
        self.admin_cmd_backup_schema__file.setText(sPath)
        self.admin_cmd_backup_schema__file.setCursorPosition(0)

    def guiAdminAccessDataEdited(self):
        """
        Admin-Zugangsdaten wurden geändert
        """

        sHost, sPort, sUser, sPw, sDb, sSchema = self.core_getCredentialsFromGui()

        # Db
        self.admin_cmd_list_schemas__db.setText(sDb)
        self.admin_cmd_restore_schema__db.setText(sDb)
        self.admin_cmd_backup_schema__db.setText(sDb)
        self.admin_cmd_rename_schema__db.setText(sDb)
        self.admin_cmd_delete_schema__db.setText(sDb)
        self.admin_cmd_new_schema__db.setText(sDb)
        # Schema
        self.admin_cmd_backup_schema__schema.setText(sSchema)
        self.admin_cmd_rename_schema__schema_old.setText(sSchema)
        self.admin_cmd_delete_schema__schema.setText(sSchema)
        #self.admin_cmd_new_schema__critterbase.setText(sSchema)
        # Cursors -> 0
        lstTextFields = [
            self.admin_cmd_list_schemas__db,
            self.admin_cmd_restore_schema__db,
            self.admin_cmd_backup_schema__db,
            self.admin_cmd_rename_schema__db,
            self.admin_cmd_delete_schema__db,
            self.admin_cmd_new_schema__db,
            self.admin_cmd_backup_schema__schema,
            self.admin_cmd_rename_schema__schema_old,
            self.admin_cmd_delete_schema__schema
        ]
        for field in lstTextFields:
            field.setCursorPosition(0)

    # --- GUI Refresh

    def status_refresh_all(self):
        """
        Initiiere einen Refresh aller relevanten Status-Anzeigen im Hauptfenster
        """

        self.status_refresh_connection()
        self.status_refresh_lookup()
        self.status_refresh_ingest()
        self.status_refresh_taxonWizard()
        self.status_refresh_milliometer()
        self.status_refresh_ingestion_overview()
        self.status_refresh_fao()
        self.status_refresh_actionLogTable()
        self.status_refresh_dataPatrol()

        self.status_refresh_dryMode()
        self.status_refresh_verbosity()
        self.status_refresh_toolbarMini()
        self.status_refresh_myWorms()
        self.status_refresh_system()
        self.status_refresh_adminLock()
        self.status_refresh_sysMon()

    def status_refresh_verbosity(self):
        self.comboVerbosity.blockSignals(True)
        if self.__model.conf_isDebug():
            self.comboVerbosity.setCurrentIndex(2)
        elif self.__model.conf_isVerbose():
            self.comboVerbosity.setCurrentIndex(1)
        else:
            self.comboVerbosity.setCurrentIndex(0)
        self.comboVerbosity.blockSignals(False)

    def status_refresh_toolbarMini(self):
        # Server connection

        if self.__model.subsystem_getConnector().db_isConnectionFailed():
            # Error
            #self.buttonConnectionToggle.setIcon(self.style().standardIcon(getattr(QStyle, "SP_BrowserStop")))
            self.buttonConnectionToggle.setIcon(QIcon(os.path.dirname(os.path.realpath(__file__)) + "/ressource/icon_error.png"))
            self.buttonConnectionToggle.setToolTip("Connection is broken (click to disconnect)")
        elif self.__model.subsystem_getConnector().db_isConnected():
            # Online
            #self.buttonConnectionToggle.setIcon(self.style().standardIcon(getattr(QStyle, "SP_DialogYesButton")))
            self.buttonConnectionToggle.setIcon(QIcon(os.path.dirname(os.path.realpath(__file__)) + "/ressource/icon_connected.png"))
            self.buttonConnectionToggle.setToolTip("Server is connected (click to disconnect)")
        else:
            # Offline
            #self.buttonConnectionToggle.setIcon(self.style().standardIcon(getattr(QStyle, "SP_DialogNoButton")))
            self.buttonConnectionToggle.setIcon(QIcon(os.path.dirname(os.path.realpath(__file__)) + "/ressource/icon_disconnected.png"))
            self.buttonConnectionToggle.setToolTip("Server is disconnected (click to connect)")

        # Dry mode
        if self.__model.conf_isDry():
            self.buttonConnectionDry.setIcon(QIcon(os.path.dirname(os.path.realpath(__file__)) + "/ressource/icon_umbrella_open.png"))
            self.buttonConnectionDry.setToolTip("Dry mode is active (click to switch off)")
        else:
            self.buttonConnectionDry.setIcon(
                QIcon(os.path.dirname(os.path.realpath(__file__)) + "/ressource/icon_umbrella_closed.png"))
            self.buttonConnectionDry.setToolTip("Dry mode is not active (click to activate)")

    def status_refresh_connection(self):
        """
        Refresh Status: Connection
        """

        # Combobox DB-Presets
        self.infoComboDbPreset.blockSignals(True)
        self.infoComboDbPreset.clear()
        for sAccount in self.__model.subsystem_getDb().account_all_getNames():
            self.infoComboDbPreset.addItem(sAccount)
        self.infoComboDbPreset.setCurrentIndex(self.__model.subsystem_getDb().account_active_getIndex())
        self.infoComboDbPreset.blockSignals(False)
        # Textfelder
        self.infoTextDbConnect.setPalette(QPalette()) # Standardfarbe Text (deaktiviert)
        if self.__model.subsystem_getConnector().db_isConnectionFailed():
            self.infoTextDbConnect.setText("Connection error!")
            palette = QPalette()
            palette.setColor(QPalette.Text, Qt.red)
            self.infoTextDbConnect.setPalette(palette)
        elif self.__model.subsystem_getConnector().db_isConnected():
            self.infoTextDbConnect.setText("connected")
        else:
            self.infoTextDbConnect.setText("disconnected")

        self.infoTextDbAccount.setText((self.__model.subsystem_getDb().account_single_getPart("account")))
        self.infoTextDbName.setText(self.__model.subsystem_getDb().account_single_getPart("db"))
        self.infoTextSchemaName.setText(self.__model.subsystem_getDb().account_single_getPart("schema"))
        self.infoTextDbHost.setText(self.__model.subsystem_getDb().account_single_getPart("host") + ":" + self.__model.subsystem_getDb().account_single_getPart("port"))
        self.infoTextDbUser.setText(self.__model.subsystem_getDb().account_single_getPart("user"))
        sVersionPgServer = ""
        sVersionSchema = ""
        if self.__model.subsystem_getConnector().isReady():
            sVersionPgServer = self.__model.subsystem_getConnector().status_getData("db").get("pg_version")
            sVersionSchema = self.__model.subsystem_getConnector().status_getData("db").get("dm_version")
            if sVersionSchema is None:
                sVersionSchema = "<error>"
            if sVersionPgServer is None:
                sVersionPgServer = "<error>"
        self.infoTextVersionPg.setText(sVersionPgServer)
        self.infoTextVersionSchema.setText(sVersionSchema)
        # Cursorpostiionen auf Anfang setzen
        self.infoTextDbAccount.setCursorPosition(0)
        self.infoTextDbConnect.setCursorPosition(0)
        self.infoTextDbHost.setCursorPosition(0)
        self.infoTextDbName.setCursorPosition(0)
        self.infoTextSchemaName.setCursorPosition(0)
        self.infoTextDbUser.setCursorPosition(0)

    def status_refresh_system(self):
        """
        Refresh Status: System
        """

        # Textfelder
            # Dry
        if self.__model.conf_isDry():
            self.sysDryRunSwitch.setIcon(QIcon(os.path.dirname(os.path.realpath(__file__)) + "/ressource/icon_switch_on.png"))
            self.sysDryRunSwitch.setToolTip("on")
        else:
            self.sysDryRunSwitch.setIcon(QIcon(os.path.dirname(os.path.realpath(__file__)) + "/ressource/icon_switch_off.png"))
            self.sysDryRunSwitch.setToolTip("off")

    def status_refresh_ingest(self):
        """
        Refresh Status: Tabellen-Inhalte
        """

        pd = self.ui_printConsoleDebug
        pn = self.ui_printConsoleNormal


        if self.__model.subsystem_getConnector().isReady():

            res = Result()
            # Ingest
            dicData = self.__model.subsystem_getConnector().status_getData("ingest")
            if dicData is not None:
                dicIngTableCounts = dicData.get("table_counts")
                for sTableName, widget in self.__dicStatus_widgets_ingest.items():
                    if self.__model.subsystem_getMapper().all_isTableIngest(sTableName):
                        iCount = dicIngTableCounts.get(sTableName)
                        widget.setText(str(iCount) if iCount is not None else "-")

            # Error
            if res.isErr():
                pn("Error getting Status for Igest-Tables")
                pn(res.getErrWarnInfoString(short=True))
        else:
            self.status_clean_ingest()  # Felder loeschen

    def status_refresh_fao(self):
        pn = self.ui_printConsoleNormal
        if self.__model.subsystem_getConnector().isReady():
            sCol_green = Colors.getColor("green")
            sCol_red = Colors.getColor("red")
            reg = self.__model.subsystem_getRegions()
            # Samples
            iSamples_all = reg.map_getSamples_all_nr()
            # Regions (all)
            iRegion_all = reg.regions_getAll_nr(leftover=False)
            widget = self.fao_regions_count
            if iRegion_all == 0 and iSamples_all > 0:
                palette = QPalette(widget.palette())
                palette.setColor(QPalette.Text, QColor(sCol_red))
            else:
                palette = QPalette()
            widget.setPalette(palette)
            self.fao_regions_count.setText(str(iRegion_all))
            # Map
            iLeftover = reg.map_getSamples_leftover_nr()
            iNew = reg.map_getSamples_unmapped_nr()
            iRegion = reg.map_getRegions_nr()
            iNew_proz = reg.map_getSamples_unmapped_proz()
            iLeftover_proz = reg.map_getSamples_leftover_proz()
            # Text
            if iLeftover is not None:
                self.fao_leftover_count.setText(str(iLeftover))
            if iRegion is not None:
                self.fao_mapping_count.setText(str(iRegion))
            if iNew is not None:
                self.fao_new_count.setText(str(iNew))
            # Leftover: Proz
            colGradient = ColorGradient(sCol_green, sCol_red, 100)
            widget = self.fao_leftover_proz
            palette = QPalette(widget.palette())
            iR, iG, iB = colGradient.getRGB(iLeftover_proz)
            palette.setColor(QPalette.ColorRole.Text, QColor(Colors.getColor("black")))
            palette.setColor(QPalette.ColorRole.HighlightedText, QColor(Colors.getColor("black")))
            palette.setColor(QPalette.Highlight, QColor(iR, iG, iB))
            widget.setPalette(palette)
            widget.setFormat(str(iLeftover_proz)+"%")
            widget.setValue(iLeftover_proz)
            widget.setToolTip("leftover | all")
            # New: Proz
            colGradient = ColorGradient(sCol_green, sCol_red, 100)
            widget = self.fao_new_proz
            palette = QPalette(widget.palette())
            palette.setColor(QPalette.ColorRole.Text, QColor(Colors.getColor("black")))
            palette.setColor(QPalette.ColorRole.HighlightedText, QColor(Colors.getColor("black")))
            iR, iG, iB = colGradient.getRGB(iNew_proz)
            palette.setColor(QPalette.Highlight, QColor(iR, iG, iB))
            widget.setPalette(palette)
            widget.setFormat(str(iNew_proz)+"%")
            widget.setValue(iNew_proz)
            widget.setToolTip("new/unmapped | all")
            if iSamples_all == 0:
                self.fao_mapping_stat.setIcon(QIcon(os.path.dirname(os.path.realpath(__file__)) + "/ressource/icon_frage.png"))
            elif iNew > 0:
                self.fao_mapping_stat.setIcon(QIcon(os.path.dirname(os.path.realpath(__file__)) + "/ressource/icon_thumbs_down.png"))
            else:
                self.fao_mapping_stat.setIcon(QIcon(os.path.dirname(os.path.realpath(__file__)) + "/ressource/icon_thumbs_up.png"))
        else:
            self.status_clean_fao()

    def status_refresh_lookup(self):
        """
        Refresh Status: Tabellen-Inhalte
        """

        pd = self.ui_printConsoleDebug
        pn = self.ui_printConsoleNormal

        colorRed = QColor("#ff9494")
        colorBlau = QColor(66,128,180)
        colorGruen = QColor(115,181,115)
        colorHellblau = QColor(136,136,198)

        self.status_clean_lookups()  # Felder loeschen
        if self.__model.subsystem_getConnector().isReady():
            res = Result()
            colGradient = ColorGradient("#ff9494", "#89c496", 100)

            dicStatusData = self.__model.subsystem_getConnector().status_getData("lookup")
            if dicStatusData is None:
                res.addErr("Lookup", add="No data given", sys="Connector", kat="Getting status data")
            if res.isOk():
                for sTableName, widget in self.__dicStatus_widgets_lookup.items():
                    if self.__model.subsystem_getMapper().all_isTableLookup(sTableName) and sTableName != "taxon":
                        resLupStatus = dicStatusData.get(sTableName)
                        res.addResult(resLupStatus, msg_only=True)
                        if resLupStatus is not None and resLupStatus.isOk():
                            iCount = resLupStatus.getData("all")
                            iLookup_gold = resLupStatus.getData("gold")
                            iLookup_gold_proz = resLupStatus.getData("gold-proz")
                            iLookup_used = resLupStatus.getData("used")
                            setLookup_used_id = resLupStatus.getData("used-ids")
                            iLookup_used_proz = resLupStatus.getData("used-proz")
                            iLookup_usedNot = resLupStatus.getData("not-used")
                            iLookup_used_goldNot = resLupStatus.getData("used-not-gold")
                            iLookup_used_gold_proz = resLupStatus.getData("used-gold-proz")
                            iLookup_usedNot_goldNot = resLupStatus.getData("not-used-not-gold")

                            # Special: Dataset
                            iLookup_dataset_accessPossible = resLupStatus.getData("dataset-access-possible")
                            iLookup_dataset_accessPossible_proz = resLupStatus.getData("dataset-access-possible-proz")
                            iLookup_dataset_usedAccessPossible = resLupStatus.getData("dataset-used-access-possible")
                            setLookup_dataset_usedAccessPossible_ids = resLupStatus.getData("dataset-used-access-possible-ids")
                            iLookup_dataset_usedAccessPossible_proz = resLupStatus.getData("dataset-used-access-possible-proz")
                            iLookup_dataset_usedNotGoldAccessPossible = resLupStatus.getData("dataset-used-not-gold-access-possible")
                            # Special: Gear
                            iLookup_gear_doubleConfig = resLupStatus.getData("gear-double")

                            # Status Buttons
                            if iLookup_used_goldNot > 0:
                                self.__dicStatus_widgets_lookup_status[sTableName].setIcon(QIcon(os.path.dirname(os.path.realpath(__file__)) + "/ressource/icon_thumbs_mid.png"))
                            else:
                                self.__dicStatus_widgets_lookup_status[sTableName].setIcon(QIcon(os.path.dirname(os.path.realpath(__file__)) + "/ressource/icon_thumbs_up.png"))
                            # Gold
                            if sTableName in self.__dicStatus_widgets_lookup_gold and iLookup_gold is not None:
                                widget = self.__dicStatus_widgets_lookup_gold[sTableName]
                                if iLookup_gold_proz is None:
                                    sGoldProz = "-"
                                    sGoldInfo = ""
                                else:
                                    palette = QPalette(widget.palette())
                                    palette.setColor(QPalette.ColorRole.Text, QColor(Colors.getColor("black")))
                                    palette.setColor(QPalette.ColorRole.HighlightedText,
                                                     QColor(Colors.getColor("black")))
                                    iR,iG,iB = colGradient.getRGB(iLookup_gold_proz)
                                    palette.setColor(QPalette.Highlight, QColor(iR,iG,iB))
                                    widget.setPalette(palette)
                                    sGoldProz = str(iLookup_gold_proz)+"%"
                                    sGoldInfo = "gold "+str(iLookup_gold) + " | not gold "+str(iCount-iLookup_gold)
                                widget.setFormat(sGoldProz)
                                widget.setValue(iLookup_gold_proz)
                                widget.setToolTip(sGoldInfo)
                            # Used
                            if sTableName in self.__dicStatus_widgets_lookup_count and iLookup_used is not None:
                                widget = self.__dicStatus_widgets_lookup_count[sTableName]
                                if iLookup_used is None:
                                    sProz = "-"
                                    sInfo = ""
                                else:
                                    palette = QPalette(widget.palette())
                                    iR, iG, iB = colGradient.getRGB(iLookup_gold_proz)
                                    #palette.setColor(QPalette.Highlight, QColor(iR, iG, iB))
                                    widget.setPalette(palette)
                                    sProz = str(iLookup_used_proz) + "%"
                                    sInfo = "used " + str(iLookup_used) + " | not used " + str(iCount - iLookup_used)
                                widget.setFormat(sProz)
                                widget.setValue(iLookup_used_proz)
                                widget.setToolTip(sInfo)

                            # Used and not-gold
                            if sTableName in self.__dicStatus_widgets_lookup_usedNotGoldLookup and iLookup_used_goldNot is not None:
                                widget = self.__dicStatus_widgets_lookup_usedNotGoldLookup[sTableName]
                                if iLookup_used_goldNot > 0:
                                    palette = QPalette(widget.palette())
                                    palette.setColor(QPalette.Text, colorRed)
                                    widget.setPalette(palette)
                                    if iLookup_used_goldNot > 99:
                                        widget.setText(">!")
                                        widget.setToolTip(str(iLookup_used_goldNot) + " used not-gold [check?]")
                                    else:
                                        widget.setText(str(iLookup_used_goldNot))
                                else:
                                    widget.setToolTip("used not-gold [check?]")

                            # Not Used and not-gold
                            if sTableName in self.__dicStatus_widgets_lookup_notUsedNotGoldLookup and iLookup_usedNot_goldNot is not None:
                                widget = self.__dicStatus_widgets_lookup_notUsedNotGoldLookup[sTableName]
                                if iLookup_usedNot_goldNot > 0:
                                    palette = QPalette(widget.palette())
                                    palette.setColor(QPalette.Text, colorBlau)
                                    widget.setPalette(palette)
                                    if iLookup_usedNot_goldNot > 99:
                                        widget.setText(">!")
                                        widget.setToolTip(str(iLookup_usedNot_goldNot) + " not-used not-gold entries [purge?]")
                                    else:
                                        widget.setText(str(iLookup_usedNot_goldNot))
                                else:
                                    widget.setToolTip("not-used not-gold [purge?]")
                            # Count
                            self.__dicStatus_widgets_lookup[sTableName].setText(str(iCount) if iCount is not None else "-")

                            # Sepcial: Gear
                            if sTableName == "gear":
                                widget = self.__dicStatus_widgets_lookup_special["gear"]
                                if iLookup_gear_doubleConfig > 0:
                                    widget.setToolTip(str(iLookup_gear_doubleConfig) + " double configurations found")
                                    widget.setText(str(iLookup_gear_doubleConfig))
                                    if iLookup_gear_doubleConfig > 99:
                                        widget.setText(">!")
                                    palette = QPalette(widget.palette())
                                    palette.setColor(QPalette.Text, colorBlau)
                                    widget.setPalette(palette)

                                else:
                                    widget.setToolTip("no double configurations found")

                            # Special: Dataset
                            if sTableName == "dataset":
                                # Free/Request (Data/Product)
                                widget = self.__widgetStatus_widgets_lookup_dataset_allFreeRequest
                                iWert = iLookup_dataset_accessPossible
                                iWertMax = iCount - iLookup_dataset_accessPossible
                                iProz = iLookup_dataset_accessPossible_proz
                                if iProz is None:
                                    sProz = "-"
                                    sInfo = ""
                                else:
                                    palette = QPalette(widget.palette())
                                    palette.setColor(QPalette.ColorRole.Text, QColor(Colors.getColor("black")))
                                    palette.setColor(QPalette.ColorRole.HighlightedText,
                                                     QColor(Colors.getColor("black")))
                                    iR, iG, iB = colGradient.getRGB(iLookup_gold_proz)
                                    # palette.setColor(QPalette.Highlight, QColor(iR, iG, iB))
                                    widget.setPalette(palette)
                                    sProz = str(iProz) + "%"
                                    sInfo = "free/request " + str(iWert) + " | restricted " + str(iWertMax)
                                widget.setFormat(sProz)
                                widget.setValue(iProz)
                                widget.setToolTip(sInfo)
                                # Used Free/Request (Data/Product)
                                widget = self.__widgetStatus_widgets_lookup_dataset_allUsedFreeRequest
                                iWert = iLookup_dataset_usedAccessPossible
                                iProz = iLookup_dataset_usedAccessPossible_proz
                                iWertMax = iCount - iLookup_dataset_usedAccessPossible
                                if iProz is None:
                                    sProz = "-"
                                    sInfo = ""
                                else:
                                    palette = QPalette(widget.palette())
                                    palette.setColor(QPalette.ColorRole.Text, QColor(Colors.getColor("black")))
                                    palette.setColor(QPalette.ColorRole.HighlightedText,QColor(Colors.getColor("black")))
                                    iR, iG, iB = colGradient.getRGB(iLookup_gold_proz)
                                    # palette.setColor(QPalette.Highlight, QColor(iR, iG, iB))
                                    widget.setPalette(palette)
                                    sProz = str(iProz) + "%"
                                    sInfo = "used free/request " + str(iWert) + " | used restricted " + str(iWertMax)
                                widget.setFormat(sProz)
                                widget.setValue(iProz)
                                widget.setToolTip(sInfo)
                                # Used Free/Request and not-gold
                                widget = self.__widgetStatus_widgets_lookup_dataset_allUsedNotGoldFreeRequest
                                iWert = iLookup_dataset_usedNotGoldAccessPossible
                                if iWert > 0:
                                    palette = QPalette(widget.palette())
                                    palette.setColor(QPalette.Text, colorRed)
                                    widget.setPalette(palette)
                                    if iWert > 99:
                                        widget.setText(">!")
                                        widget.setToolTip(str(iWert) + " used not-gold free/request [check?]")
                                        self.__dicStatus_widgets_lookup_status["dataset"].setIcon(QIcon(os.path.dirname(os.path.realpath(__file__)) + "/ressource/icon_thumb_mid.png"))
                                    else:
                                        widget.setText(str(iWert))
                                else:
                                    widget.setToolTip("used not-gold free/request [check?]")

                                # DOI
                                packDoi = self.__model.memory_load("doi")
                                if packDoi is not None:
                                    dicDoi_Ok  = packDoi.get("ok")
                                    dicDoi_Err = packDoi.get("err")
                                    dicDoi_No = packDoi.get("no")
                                    dicDoi_No = packDoi.get("no")
                                    iPangaea = packDoi.get("cluster-pangaea")
                                    iPangaea_proz = packDoi.get("cluster-pangaea-proz")
                                    iNotPangaea = packDoi.get("cluster-not-pangaea")
                                    iNotPangaea_proz = packDoi.get("cluster-not-pangaea-proz")
                                    iPangaea = 0 if iPangaea is None else iPangaea
                                    iPangaea_proz = 0 if iPangaea_proz is None else iPangaea_proz
                                    iNotPangaea = 0 if iNotPangaea is None else iNotPangaea
                                    iNotPangaea_proz = 0 if iNotPangaea_proz is None else iNotPangaea_proz

                                    # Status nur für die DOI-Prüfung
                                    if len(dicDoi_Err) > 0:
                                        self.__widgetStatus_widgets_lookup_dataset_checkDois_status.setIcon(QIcon(
                                            os.path.dirname(
                                                os.path.realpath(__file__)) + "/ressource/icon_thumbs_down.png"))
                                    elif len(dicDoi_No) > 0:
                                        self.__widgetStatus_widgets_lookup_dataset_checkDois_status.setIcon(QIcon(
                                            os.path.dirname(os.path.realpath(__file__)) + "/ressource/icon_thumbs_mid.png"))
                                    else:
                                        self.__widgetStatus_widgets_lookup_dataset_checkDois_status.setIcon(QIcon(
                                            os.path.dirname(os.path.realpath(__file__)) + "/ressource/icon_thumbs_up.png"))

                                    # DOI: OK/Not OK
                                    widget = self.__widgetStatus_widgets_lookup_dataset_DOI_OkNotOk
                                    iWert = len(dicDoi_Ok)
                                    iWert_err = len(dicDoi_Err)
                                    iWert_no = len(dicDoi_No)
                                    iWertMax = iCount
                                    try:
                                        iProz = int((iWert/iWertMax) * 100)
                                    except:
                                        iProz = None
                                    if iProz is None:
                                        sProz = "-"
                                        sInfo = ""
                                    else:
                                        palette = QPalette()
                                        if iWert_err > 0:
                                            palette.setColor(QPalette.Highlight, QColor(Colors.getColor("red")))
                                        else:
                                            pass
                                            palette.setColor(QPalette.Highlight, QColor(Colors.getColor("black")))
                                        widget.setPalette(palette)
                                        if iWert_err > 0 and iWert_no == 0:
                                            sDetail = " [err]"
                                        elif iWert_err == 0 and iWert_no > 0:
                                            sDetail = " [no]"
                                        else:
                                            sDetail = " [" + str(iWert_no) + "no " + str(iWert_err) + "err]"
                                        sProz = str(iProz) + "%"
                                        sInfo = "DOI ok " + str(iWert) + " | DOI not ok " + str(iWertMax-iWert) + sDetail
                                    widget.setFormat(sProz)
                                    widget.setValue(iProz)
                                    widget.setToolTip(sInfo)
                                    # DOI: Used OK/Not OK
                                    widget = self.__widgetStatus_widgets_lookup_dataset_DOI_UsedOkNotOk
                                    iWert = len(set(dicDoi_Ok.keys()).intersection(setLookup_used_id))
                                    iWert_err = len(set(dicDoi_Err.keys()).intersection(setLookup_used_id))
                                    iWert_no = len(set(dicDoi_No.keys()).intersection(setLookup_used_id))
                                    iWertMax = iLookup_used
                                    try:
                                        iProz = int((iWert / iWertMax) * 100)
                                    except:
                                        iProz = None
                                    if iProz is None:
                                        sProz = "-"
                                        sInfo = ""
                                    else:
                                        palette = QPalette()
                                        if iWert_err > 0:
                                            palette.setColor(QPalette.Highlight, QColor(Colors.getColor("red")))
                                        else:
                                            pass
                                            palette.setColor(QPalette.Highlight, QColor(Colors.getColor("black")))
                                        widget.setPalette(palette)
                                        sProz = str(iProz) + "%"
                                        if iWert_err > 0 and iWert_no == 0:
                                            sDetail = " [err]"
                                        elif iWert_err == 0 and iWert_no > 0:
                                            sDetail = " [no]"
                                        else:
                                            sDetail = " [" + str(iWert_no) + "no " + str(iWert_err) + "err]"
                                        sInfo = "used DOI ok " + str(iWert) + " | used DOI not ok " + str(iWertMax - iWert) + sDetail
                                    widget.setFormat(sProz)
                                    widget.setValue(iProz)
                                    widget.setToolTip(sInfo)
                                    # DOI: Used free/request not ok
                                    widget = self.__widgetStatus_widgets_lookup_dataset_DOI_UsedNotOkReachable
                                    iWert_used_free_request_not_ok = len(set(dicDoi_Err.keys()).union(set(dicDoi_No.keys())).intersection(setLookup_dataset_usedAccessPossible_ids))
                                    if iWert_used_free_request_not_ok > 0:
                                        palette = QPalette(widget.palette())
                                        palette.setColor(QPalette.Text, colorRed)
                                        widget.setPalette(palette)
                                        if iWert_used_free_request_not_ok > 99:
                                            widget.setText(">!")
                                            widget.setToolTip(str(iWert_used_free_request_not_ok) + " used free/request DOI not ok [problem!]")
                                        else:
                                            widget.setText(str(iWert_used_free_request_not_ok))
                                    else:
                                        widget.setToolTip("used free/request DOI not ok [problem!]")
                                    # DOI: PANGAEA
                                    widget = self.__widgetStatus_widgets_lookup_dataset_DOI_Pangaea
                                    palette = QPalette(widget.palette())
                                    palette.setColor(QPalette.Text, colorGruen)
                                    widget.setPalette(palette)
                                    if iPangaea > 0:
                                        widget.setToolTip("PANGAEA DOI (" + str(iPangaea_proz)+"%)")
                                        if iPangaea > 99:
                                            widget.setText(">!")
                                            widget.setToolTip("PANGAEA DOI: "+ str(iPangaea) + " (" + str(iPangaea_proz)+"%)")
                                        else:
                                            widget.setText(str(iPangaea))
                                    else:
                                        widget.setToolTip("PANGAEA DOI")
                                    # DOI: PANGAEA
                                    widget = self.__widgetStatus_widgets_lookup_dataset_DOI_PangaeaNot
                                    if iPangaea > 0:
                                        palette = QPalette(widget.palette())
                                        palette.setColor(QPalette.Text, colorRed)
                                        widget.setPalette(palette)
                                        widget.setToolTip("no PANGAEA DOI (" + str(iNotPangaea_proz)+"%) [really raw data available?]")
                                        if iPangaea > 99:
                                            widget.setText(">!")
                                            widget.setToolTip("no PANGAEA DOI: " + str(iNotPangaea) + " (" + str(iNotPangaea_proz)+"%) [really raw data available?]")
                                        else:
                                            widget.setText(str(iNotPangaea))
                                    else:
                                        widget.setToolTip("no PANGAEA DOI [really raw data available?]")
                                    # Dataset Status Button setzen
                                    if iWert_err > 0 or iWert_used_free_request_not_ok > 0:
                                        self.__dicStatus_widgets_lookup_status["dataset"].setIcon(QIcon(
                                            os.path.dirname(os.path.realpath(__file__)) + "/ressource/icon_thumbs_down.png"))
                                        self.__widgetStatus_widgets_lookup_dataset_checkDois_status.setIcon(QIcon(
                                            os.path.dirname(os.path.realpath(__file__)) + "/ressource/icon_thumbs_down.png"))
                                    elif iWert_no > 0:
                                        self.__dicStatus_widgets_lookup_status["dataset"].setIcon(QIcon(
                                            os.path.dirname(os.path.realpath(__file__)) + "/ressource/icon_thumbs_mid.png"))
                                else:
                                    self.__widgetStatus_widgets_lookup_dataset_checkDois_status.setIcon(QIcon(
                                        os.path.dirname(os.path.realpath(__file__)) + "/ressource/icon_frage.png"))

            # Error
            if res.isErr():
                pn("Error getting Status for Admin-Tables")
                pn(res.getErrWarnInfoString(short=True))

    def status_refresh_taxonWizard(self):
        """
        Refresh Status: Tabellen-Inhalte
        """

        pd = self.ui_printConsoleDebug
        pn = self.ui_printConsoleNormal

        colorRed = Colors.getColor("red")
        colorGreen = Colors.getColor("green")

        if self.__model.subsystem_getConnector().isReady():

            res = Result()
            colGradient = ColorGradient(colorRed, colorGreen, 100)


            # Hole Status Infos
            dicWizardStatus = None
            dicTableCounts = None
            iTaxon_count = None
            iTaxbase_count = None
            iTaxbase_gold = None
            resLastTaxListUpdate = None
            try:
                dicWizardStatus = self.__model.subsystem_getConnector().status_getData("wizard")
                dicTableCounts = dicWizardStatus.get("table_counts")
                iTaxon_count = dicWizardStatus.get("taxon_count")
                iTaxbase_count = dicTableCounts.get("wormsbase")
                iTaxbase_gold = dicWizardStatus.get("wormsbase_gold")
                resLastTaxListUpdate = dicWizardStatus.get("taxonlist_lastupdate")
            except Exception as ex:
                res.addErr("Getting TaxonWizard Admin info", add=str(ex), system="TaxonWizard", kat="Status")
            if None in (dicWizardStatus, iTaxbase_gold, iTaxbase_count, iTaxon_count, resLastTaxListUpdate):
                res.addErr("Getting TaxonWizard Admin info", system="TaxonWizard", kat="Status")

            # Admin Tables
            if res.isOk():
                for sTableName, (sName, widget) in self.__odicTaxonSystemData.items():
                    iCount = dicTableCounts.get(sTableName)
                    widget.setText(str(iCount) if iCount is not None else "-")
            # TaxBase
            if res.isOk():
                try:
                    iGoldProz = int((iTaxbase_gold / iTaxbase_count) * 100)
                except:
                    iGoldProz = 0
                if iGoldProz is None:
                    sGoldProz = "-"
                    sGoldInfo = ""
                else:
                    palette = QPalette(self.taxSysBaseProgress.palette())
                    palette.setColor(QPalette.ColorRole.Text, QColor(Colors.getColor("black")))
                    palette.setColor(QPalette.ColorRole.HighlightedText, QColor(Colors.getColor("black")))
                    iR, iG, iB = colGradient.getRGB(iGoldProz)
                    palette.setColor(QPalette.ColorRole.Text, QColor(Colors.getColor("black")))
                    palette.setColor(QPalette.ColorRole.HighlightedText, QColor(Colors.getColor("black")))
                    palette.setColor(QPalette.Highlight, QColor(iR, iG, iB))
                    self.taxSysBaseProgress.setPalette(palette)
                    sGoldProz = str(iGoldProz) + "%"
                    sGoldInfo = "gold " + str(iTaxbase_gold) + " | not-gold " + str(iTaxbase_count - iTaxbase_gold)
                self.taxSysBaseProgress.setFormat(sGoldProz)
                self.taxSysBaseProgress.setValue(iGoldProz)
                self.taxSysBaseProgress.setToolTip(sGoldInfo)

            # Taxon Wizard: Taxon List
            if res.isOk():
                self.__dicStatus_widgets_lookup["taxon"].setText(str(iTaxon_count) if iTaxon_count is not None else "-")
                self.taxSysNoTaxList.setText(str(iTaxon_count) if iTaxon_count is not None else "-")
                if resLastTaxListUpdate.isOk():
                    dateDate = resLastTaxListUpdate.getData("date")
                    iErr= resLastTaxListUpdate.getData("err")
                    iWarn = resLastTaxListUpdate.getData("warn")
                    if None in (dateDate, iErr, iWarn):
                        # Es gab noch kein kein TaxonList Update
                        self.taxSysAgeProgress.setToolTip("No Taxon List update made")
                        self.taxSysAgeProgress.setFormat("-")
                        self.taxSysAgeProgress.setValue(0)
                        self.taxSysWarn.setToolTip("No warnings occurred")
                        self.taxSysWarn.setPalette(QPalette())
                        self.taxSysWarn.setText("")
                        self.taxSysAgeStatus.setIcon(QIcon(os.path.dirname(os.path.realpath(__file__)) + "/ressource/icon_frage.png"))
                    else:
                        # Es gab schon ein TaxonList Update
                        iMaxDaysAllowed = self.__model.subsystem_getMyWorms().puffer_getMaxAgeAllowed()
                        # Days
                        try:
                            iDays = (datetime.datetime.now() - dateDate).days
                        except:
                            iDays = None
                        try:
                            iDaysProz = int((iDays / iMaxDaysAllowed) * 100)
                        except:
                            iDaysProz = None
                            sTlDaysProz = "-"
                        if iDaysProz is None:
                            sTlDaysProz= "-"
                        else:
                            palette = QPalette(self.taxSysAgeProgress.palette())
                            palette.setColor(QPalette.ColorRole.Text, QColor(Colors.getColor("black")))
                            palette.setColor(QPalette.ColorRole.HighlightedText, QColor(Colors.getColor("black")))
                            iR, iG, iB = colGradient.getRGB(100-iDaysProz)
                            palette.setColor(QPalette.Highlight, QColor(iR, iG, iB))
                            self.taxSysAgeProgress.setPalette(palette)
                        if iDays is not None:
                            self.taxSysAgeProgress.setToolTip(str(iDays)+" days till last update of Taxon List (max. allowed "+str(iMaxDaysAllowed)+" days)")
                            self.taxSysAgeProgress.setFormat(str(iDays)+"d")
                            self.taxSysAgeProgress.setValue(iDays)
                        # Warnings
                        if iWarn > 0:
                            palette = QPalette(self.taxSysAgeProgress.palette())
                            palette.setColor(QPalette.Text, QColor(colorRed))
                            self.taxSysWarn.setPalette(palette)
                            self.taxSysWarn.setToolTip(str(iWarn) + " warnings in Taxon list")
                            if iWarn > 99:
                                self.taxSysWarn.setText(">!")
                            else:
                                self.taxSysWarn.setText(str(iWarn))
                        else:
                            self.taxSysWarn.setToolTip("No warnings occurred")
                            self.taxSysWarn.setPalette(QPalette())
                            self.taxSysWarn.setText("ok")
                        # Log/Status Icon
                        if iErr >0 or iDays > iMaxDaysAllowed:
                            self.taxSysAgeStatus.setIcon(QIcon(os.path.dirname(os.path.realpath(__file__)) + "/ressource/icon_thumbs_down.png"))
                        elif iWarn > 0:
                            self.taxSysAgeStatus.setIcon(QIcon(os.path.dirname(os.path.realpath(__file__)) + "/ressource/icon_thumbs_up.png"))
                        else:
                            self.taxSysAgeStatus.setIcon(QIcon(os.path.dirname(os.path.realpath(__file__)) + "/ressource/icon_thumbs_up.png"))
                else:
                    self.taxSysAgeStatus.setIcon(QIcon(os.path.dirname(os.path.realpath(__file__)) + "/ressource/icon_frage.png"))

            # Error
            if res.isErr():
                pn("Error getting Status for Admin-Tables")
                pn(res.getErrWarnInfoString(short=True))
        else:
            self.status_clean_taxonWizard()  # Felder loeschen

    def status_refresh_dataPatrol(self):
        """
        Refresh Status: Data Patrol
        """
        pd = self.ui_printConsoleDebug
        pn = self.ui_printConsoleNormal
        if self.__model.subsystem_getConnector().isReady():
            try:
                # Data Patrol: BiotaControl
                dicStatus = self.__model.subsystem_getConnector().status_getData("datapatrol")
                resLastBiotaControl = dicStatus["biotacontrol"]
                if resLastBiotaControl.isOk():
                    dateDate = resLastBiotaControl.getData("date")
                    iErr = resLastBiotaControl.getData("err")
                    iWarn = resLastBiotaControl.getData("warn")
                    if None in (dateDate, iErr, iWarn):
                        # Es gab noch kein keine DataPatrol
                        self.dataPatrol_biotaControl_lastCheckWarn.setText("-")
                        self.dataPatrol_biotaControl_lastCheckWarn.setToolTip("No warnings occurred")
                        self.dataPatrol_biotaControl_lastCheckWarn.setPalette(QPalette())
                        self.dataPatrol_biotaControl_lastCheckLog.setIcon(
                            QIcon(os.path.dirname(os.path.realpath(__file__)) + "/ressource/icon_frage.png"))
                        self.dataPatrol_biotaControl_lastCheck.setText("-")
                        self.dataPatrol_biotaControl_lastCheckWarn.setText("-")
                    else:
                        # Es gab schon eine DataPatrol
                        iDays = (datetime.datetime.now() - dateDate).days
                        # Date
                        self.dataPatrol_biotaControl_lastCheck.setText(str(iDays))
                        # Warn
                        if iWarn > 0:
                            palette = QPalette(self.dataPatrol_biotaControl_lastCheckWarn.palette())
                            palette.setColor(QPalette.Text, QColor("#ff9494"))
                            self.dataPatrol_biotaControl_lastCheckWarn.setPalette(palette)
                            self.dataPatrol_biotaControl_lastCheckWarn.setToolTip(str(iWarn) + " warnings during last BiotaControl")
                            if iWarn > 9999:
                                self.dataPatrol_biotaControl_lastCheckWarn.setText(">!")
                            else:
                                self.dataPatrol_biotaControl_lastCheckWarn.setText(str(iWarn))
                        else:
                            self.dataPatrol_biotaControl_lastCheckWarn.setText("ok")
                            self.dataPatrol_biotaControl_lastCheckWarn.setToolTip("No warnings occurred")
                            self.dataPatrol_biotaControl_lastCheckWarn.setPalette(QPalette())

                        # Log/Status Icon
                        if iErr > 0:
                            self.dataPatrol_biotaControl_lastCheckLog.setIcon(QIcon(os.path.dirname(os.path.realpath(__file__)) + "/ressource/icon_thumbs_down.png"))
                        elif iWarn > 0:
                            self.dataPatrol_biotaControl_lastCheckLog.setIcon(
                                QIcon(os.path.dirname(os.path.realpath(__file__)) + "/ressource/icon_thumbs_up.png"))
                        else:
                            self.dataPatrol_biotaControl_lastCheckLog.setIcon(
                                QIcon(os.path.dirname(os.path.realpath(__file__)) + "/ressource/icon_thumbs_up.png"))
                else:
                    self.dataPatrol_biotaControl_lastCheckLog.setIcon(QIcon(os.path.dirname(os.path.realpath(__file__)) + "/ressource/icon_frage.png"))
                    self.dataPatrol_biotaControl_lastCheck.setText("?")
                    self.dataPatrol_biotaControl_lastCheckWarn.setText("?")

            except Exception as ex:
                #self.guiClearStatusTables() # Felder loeschen
                pn("Error (Data Patrol Status): During access to database: " + str(ex))
        else:
            self.status_clean_dataPatrol()  # Felder loeschen

    def status_refresh_milliometer(self):
        """
        Refresh Milli-o-Meter
        """
        if self.__model.subsystem_getConnector().isReady():
            try:
                # Hole alle Tabellengroessen
                dicStatusIng = self.__model.subsystem_getConnector().status_getData("ingest")
                dicStatusIng_tablecount = dicStatusIng["table_counts"]
                iMilliometer = 0
                iMilliometer += dicStatusIng_tablecount.get("biota") + dicStatusIng_tablecount.get("sediment") + dicStatusIng_tablecount.get("indmeasure")
                # Maximum einstellen
                self.progressBar.setTextVisible(True)
                iMax = 1000
                sMilestone = "1K"
                self.milliometer.setMinimum(0)
                if iMilliometer > 1000:
                    sMilestone = "10K"
                    iMax = 10000
                if iMilliometer > 10000:
                    sMilestone = "25K"
                    iMax = 25000
                if iMilliometer > 25000:
                    sMilestone = "50K"
                    iMax = 50000
                if iMilliometer > 50000:
                    sMilestone = "75K"
                    iMax = 75000
                if iMilliometer > 50000:
                    sMilestone = "100K"
                    iMax = 100000
                if iMilliometer > 100000:
                    sMilestone = "250K"
                    iMax = 250000
                if iMilliometer > 250000:
                    sMilestone = "500K"
                    iMax = 500000
                if iMilliometer > 500000:
                    sMilestone = "750K"
                    iMax = 750000
                if iMilliometer > 750000:
                    sMilestone = "1M"
                    iMax = 1000000
                if iMilliometer > 1000000:
                    sMilestone = "10M"
                    iMax = 10000000
                self.milliometer.setMaximum(iMax)
                # Milliometer einstellen
                self.milliometer.setValue(iMilliometer)
                self.milliometer.setFormat(str(iMilliometer))
                self.milliometer.setToolTip("Next Milestone "+sMilestone)
                self.milliometer.setEnabled(True)
                # palette = QPalette()
                # palette.setColor(QPalette.Background, QColor(Colors.getColor("white")))
                # self.milliometer.setPalette(palette)
            except Exception as ex:
                self.ui_printConsoleNormal("Error getting Milliometer data: " + str(ex))
                self.milliometer.setValue(0)
                self.milliometer.setFormat("")
        else:
            self.milliometer.setValue(0)
            self.milliometer.setFormat("")

    def status_refresh_dryMode(self):
        """
        Dry-Run-Modus im Titel angeben
        """

        sTitle = self.__model.conf_getInfo("prgname") + " " + self.__model.conf_getInfo("version") + " (" + self.__model.conf_getInfo("dbschema") + ")"
        if self.__model.conf_isDry():
            self.setWindowTitle(sTitle+" <Dry mode is active>")
        else:
            self.setWindowTitle(sTitle)

    def status_refresh_ingestion_overview(self, iIngestFokus=None):
        """
        Regfresh: Ingestions-Tabelle

        Args:
            iIngestFokus (int): Ingest auf das fokussiert werden soll
        """

        pd = self.ui_printConsoleDebug
        pn = self.ui_printConsoleNormal

        if iIngestFokus is not None:
            bIngestFokus = True
        else:
            bIngestFokus = False

        # Helper de/aktiviere Buttons
        def activateButtons():
            self.btnIngestMap.show()
            self.btnIngestTab.show()
            self.btnIngestCsv.show()
            self.btnIngestLookup.show()
            self.btnIngestLog.show()
            self.btnIngestDel.show()

        def deactivateButtons():
            self.btnIngestMap.hide()
            self.btnIngestTab.hide()
            self.btnIngestCsv.hide()
            self.btnIngestLookup.hide()
            self.btnIngestLog.hide()
            self.btnIngestDel.hide()
            self.btnIngestDepShow.hide()
            self.labelIngestDepShowHide.hide()

        # Helper Range wurde verändert
        def rangeChanged():

            # Bestimme angeklickte Bereiche
            lstRanges = self.tableIngest.selectedRanges()
            self.__lstSelectedIngests = []
            for r in lstRanges:
                iRowStart = r.topRow()
                iRowEnde = r.bottomRow()
                for i in range(iRowStart, iRowEnde+1):
                    try:
                        iIngestId = int(self.tableIngest.item(i, 0).text()) # Hole die Ingest Id aus der View-Tabelle
                    except:
                        pass
                    self.__lstSelectedIngests.append(iIngestId)

            # Setze Id in GUI
            sId = "-"
            iId = None
            if len(self.__lstSelectedIngests) == 0:
                sId = ""
            elif len(self.__lstSelectedIngests) == 1:
                iId = self.__lstSelectedIngests[0]
                sId =str(self.__lstSelectedIngests[0])
            elif len(self.__lstSelectedIngests) > 1:
                sId = str(len(self.__lstSelectedIngests)) +"x"
            else:
                sId = "-"

            # De-/Aktiviere Buttons
            if len(self.__lstSelectedIngests) == 0:
                deactivateButtons()
            elif len(self.__lstSelectedIngests) == 1:
                activateButtons()
            elif len(self.__lstSelectedIngests) > 1:
                deactivateButtons()
                self.btnIngestMap.show()
                self.btnIngestLookup.show()
                self.btnIngestDel.show()

            self.infoSelectedIngest.setText(sId)  # Setze Id in GUI-Feld
            pd("Selected Ingests: " + str(self.__lstSelectedIngests))

            # Teste ob angeklicktes Ingest irgendwelche Abhängigkeiten hat -> Gib den Show Dep Button frei
            if iId is not None:
                dicBase = self.__model.subsystem_getConnector().ingest_getIngDepBase()
                dicNeed = self.__model.subsystem_getConnector().ingest_getIngDepNeed()
                if iId in dicBase or iId in dicNeed:
                    self.btnIngestDepShow.show()
                    self.labelIngestDepShowHide.show()
                else:
                    self.btnIngestDepShow.hide()
                    self.labelIngestDepShowHide.hide()

        # Ist der Fokus modus aktiv -> Zeige Button Unfocus
        if bIngestFokus:
            self.btnIngestDepHide.show()
        else:
            self.btnIngestDepHide.hide()

        # Verbunden mit Server?
        if self.__model.subsystem_getConnector().isReady():

            # Ja -> Ingestion-Tabelle aufbauen
            lstIngest = []
            try:
                sModulname = "View: Refresh ingest table"
                dicIng_status = self.__model.subsystem_getConnector().status_getData("ingest")
                lstIngest = dicIng_status["overview"]
            except Exception as ex:
                self.status_clean_lookups()  # Felder loeschen
                pn("Error (" + sModulname + "): Access to database: " + str(ex))

            # Neue Tabelle
            self.tableIngest.clearContents()
            self.tableIngest.setRowCount(0)
            self.tableIngest.setSortingEnabled(False)
            #self.tableIngest.sortItems(0)
            #self.tableIngest.clear()
            #self.tableIngest = QTableWidget(0, 7)
            #self.tableIngest.setHorizontalHeaderLabels(["Id","B","D","User","Date","Type","Info"])
            #self.tableIngest.horizontalHeader().setStretchLastSection(True)

            # Table anklickbar machen
                # Einfacher Klick -> Markiert ganze Row
            self.tableIngest.setSelectionBehavior(QTableView.SelectRows)
            try:
                self.tableIngest.itemSelectionChanged.disconnect()
            except:
                pass
            self.tableIngest.itemSelectionChanged.connect(rangeChanged)

            dicBase = self.__model.subsystem_getConnector().ingest_getIngDepBase()
            dicNeed = self.__model.subsystem_getConnector().ingest_getIngDepNeed()

            # Zeilen (alle Ingestions)
            for row in lstIngest:

                # Ingest-Id
                iIngestId = row[0]
                bBase = iIngestFokus in dicBase and iIngestId in dicBase[iIngestFokus]
                bDepend = iIngestFokus in dicNeed and iIngestId in dicNeed[iIngestFokus]
                bFokus = iIngestId == iIngestFokus

                if not bIngestFokus or (bFokus or bDepend or bBase):
                    # Ingest wird angezeigt wenn: Kein Fokus modus oder Ingest hat Abhängigkeiten

                    rowPosition = self.tableIngest.rowCount()
                    self.tableIngest.insertRow(rowPosition)

                    # Farben bei Fokusmodus
                    if bDepend:
                        color = QColor(159, 223, 159)
                    elif bBase:
                        color = QColor(241, 208, 217)
                    else:
                        color = None

                    # Spalten (eine Ingestion)
                    for i, colInhalt in enumerate(row):
                        if i == 0:
                            # Ingest-Id
                            # a) Id
                            item = QTableWidgetItem()
                            if color is not None:
                                item.setBackground(color)
                            item.setData(Qt.DisplayRole, iIngestId) # Behandle Eintrag als Integer
                            item.setFlags(Qt.ItemIsEnabled | Qt.ItemIsSelectable)
                            self.tableIngest.setItem(rowPosition, 0, item)
                            # b) Base
                            item = QTableWidgetItem()
                            if color is not None:
                                item.setBackground(color)
                            item.setFlags(Qt.ItemIsEnabled | Qt.ItemIsSelectable)
                            lstBase = dicBase[iIngestId] if iIngestId in dicBase else []
                            if len(lstBase) > 0:
                                item.setData(Qt.DisplayRole, len(lstBase)) # Behandle Eintrag als Integer
                            if len(lstBase) > 0:
                                item.setToolTip(PrettyText.lst2strC(lstBase))
                            self.tableIngest.setItem(rowPosition, 1, item)
                            # c) Depend
                            item = QTableWidgetItem()
                            if color is not None:
                                item.setBackground(color)
                            item.setFlags(Qt.ItemIsEnabled | Qt.ItemIsSelectable)
                            lstDep = dicNeed[iIngestId] if iIngestId in dicNeed else []
                            if len(lstDep) > 0:
                                item.setData(Qt.DisplayRole, len(lstDep)) # Behandle Eintrag als Integer
                            if len(lstDep) > 0:
                                item.setToolTip(PrettyText.lst2strC(lstDep))
                            self.tableIngest.setItem(rowPosition, 2, item)
                        else:
                            # Rest
                            if i == 3:
                                colInhalt = self.__map.workflow_short2long(colInhalt)
                                if colInhalt is None:
                                    colInhalr = "?"

                            idx = i + 2
                            item = QTableWidgetItem()
                            if color is not None:
                                item.setBackground(color)
                            try:
                                iColInhalt = int(colInhalt)
                                item.setData(Qt.DisplayRole, iColInhalt)  # Integer
                            except:
                                item.setText(str(colInhalt))  # String
                            item.setFlags(Qt.ItemIsEnabled | Qt.ItemIsSelectable)
                            self.tableIngest.setItem(rowPosition, idx, item)

            # Tabellenbreite auf Optimum einstellen, letzte Spalte stretchen, Sortierung aktivieren
            self.tableIngest.resizeColumnsToContents()
            self.tableIngest.horizontalHeader().setSectionResizeMode(self.tableIngest.columnCount()-1, QHeaderView.Stretch)
            self.tableIngest.setSortingEnabled(True)
        else:
            # Nicht verbunden mit Server -> Ingestion-Tabelle löschen, Buttons deaktivieren
            self.tableIngest.clearContents()
            self.tableIngest.setRowCount(0)
            self.__lstSelectedIngests = []
            deactivateButtons()

    def status_refresh_actionLogTable(self):
        """
        Refresh: Action Log Tabelle
        """

        pd = self.ui_printConsoleDebug
        pn = self.ui_printConsoleNormal

        # Helper de/aktiviere Buttons
        def activateButtons():
            self.btnActionLogDel.show()
            self.btnActionLogDelAll.show()
            self.btnActionLogShow.show()
            self.labelActionLogFilter.show()
            self.comboActionLogFilter.show()

        def deactivateButtons():
            self.btnActionLogDel.hide()
            self.btnActionLogDelAll.hide()
            self.labelActionLogFilter.hide()
            self.comboActionLogFilter.hide()
            self.btnActionLogShow.hide()

        # Helper Range wurde verändert
        def rangeChanged():
            # Bestimme angeklickte Bereiche
            lstRanges = self.tableLog.selectedRanges()
            self.__lstSelectedActionLogs = []
            for r in lstRanges:
                iRowStart = r.topRow()
                iRowEnde = r.bottomRow()
                for i in range(iRowStart, iRowEnde + 1):
                    try:
                        iActionLogId = int(
                            self.tableLog.item(i, 0).text())  # Hole die Ingest Id aus der View-Tabelle
                    except:
                        pass
                    self.__lstSelectedActionLogs.append(iActionLogId)

            # Setze Id in GUI
            iId = None
            if len(self.__lstSelectedActionLogs) == 0:
                sId = ""
            elif len(self.__lstSelectedActionLogs) == 1:
                iId = self.__lstSelectedActionLogs[0]
                sId = str(self.__lstSelectedActionLogs[0])
            elif len(self.__lstSelectedActionLogs) > 1:
                sId = str(len(self.__lstSelectedActionLogs))+"x"
            else:
                sId = "-"

            # De-/Aktiviere Buttons
            if len(self.__lstSelectedActionLogs) == 0:
                deactivateButtons()
            elif len(self.__lstSelectedActionLogs) == 1:
                activateButtons()
            elif len(self.__lstSelectedActionLogs) > 1:
                deactivateButtons()
                self.btnActionLogDel.show()
                self.btnActionLogDelAll.show()
                self.labelActionLogFilter.show()
                self.comboActionLogFilter.show()

            self.infoSelectedActionLogs.setText(sId)  # Setze Id in GUI-Feld
            pd("Selected Action Logs: " + str(self.__lstSelectedActionLogs))

            # Zeige Inhalt an, wenn ein Log angeklickt wurde
            if len(self.__lstSelectedActionLogs) == 1:
                try:
                    iId = int(self.__lstSelectedActionLogs[0])
                    # Start Worker
                    self.worker_start(Worker_ShowActionlog(self.__model, iId), self.worker_finished_showActionLogContent)
                except:
                    self.tableLogContent.setText("")
            else:
                self.tableLogContent.setText("")

        # Verbunden mit Server?
        if self.__model.subsystem_getConnector().isReady():

            # Ja -> Log-Tabelle aufbauen
            lstLogs = []
            try:
                sModulname = "View: Refresh log table"
                pd("* " + sModulname)
                pd("Getting data for log tables from db")
                dicStatus_actionlog = self.__model.subsystem_getConnector().status_getData("actionlog")
                lstLogs = dicStatus_actionlog["table"]
            except Exception as ex:
                self.status_clean_lookups()  # Felder loeschen
                pn("Error (" + sModulname + "): Access to database: " + str(ex))
                return

            # Neue Tabelle
            self.tableLog.clearContents()
            self.tableLog.setRowCount(0)
            self.tableLog.setSortingEnabled(False)
            #self.tableLog.clear()

            sFilter = self.comboActionLogFilter.currentText()

            # Table anklickbar machen
            # Einfacher Klick -> Markiert ganze Row
            self.tableLog.setSelectionBehavior(QTableView.SelectRows)
            try:
                self.tableLog.itemSelectionChanged.disconnect()
            except:
                pass
            self.tableLog.itemSelectionChanged.connect(rangeChanged)

            # Zeilen (alle Ingestions)
            for row in lstLogs:
                iLogId = row[0]
                sFilterInRow = row[2]
                if sFilter == "*" or \
                    (sFilter=="data-patrol*" and "data-patrol" in sFilterInRow) or \
                    (sFilter=="ing*" and "ing" in sFilterInRow) or \
                    (sFilter=="lup*" and "lup" in sFilterInRow) or \
                    (sFilter=="tax*" and "tax" in sFilterInRow) or \
                    (sFilter=="tax-base*" and "tax-base" in sFilterInRow) or \
                    (sFilter == "gold*" and "gold" in sFilterInRow) or \
                    (sFilter == "gold-lup*" and "gold-lup" in sFilterInRow) or \
                    (sFilter == "gold-tax*" and "gold-tax" in sFilterInRow) or \
                    (sFilter == "fao*" and "fao-update" in sFilterInRow) or \
                    (sFilter == "fao*" and "fao-create" in sFilterInRow) or \
                    (sFilter == "fao*" and "fao-del" in sFilterInRow) or \
                    sFilterInRow == sFilter:
                    rowPosition = self.tableLog.rowCount()
                    self.tableLog.insertRow(rowPosition)
                    # Spalten (eine Ingestion)
                    for i, colInhalt in enumerate(row):
                        if i == 0:
                            # Ingest-Id
                            # a) Id
                            item = QTableWidgetItem(str(iLogId))
                            item.setData(Qt.DisplayRole, iLogId)  # Behandle Eintrag als Integer
                            item.setFlags(Qt.ItemIsEnabled | Qt.ItemIsSelectable)
                            #item.setBackground(color)
                            self.tableLog.setItem(rowPosition, 0, item)
                        else:
                            # Rest
                            item = QTableWidgetItem(str(colInhalt))
                            if i in [4, 5, 6]:
                                item.setData(Qt.DisplayRole, colInhalt)  # Behandle Eintrag als Integer
                            item.setFlags(Qt.ItemIsEnabled | Qt.ItemIsSelectable)
                            #item.setBackground(color)
                            self.tableLog.setItem(rowPosition, i, item)

            # Tabellenbreite auf Optimum einstellen, letzte Spalte stretchen, Sortierung aktivieren
            self.tableLog.resizeColumnsToContents()
            self.tableLog.horizontalHeader().setSectionResizeMode(self.tableLog.columnCount() - 1,QHeaderView.Stretch)
            self.tableLog.setSortingEnabled(True)

            # Blende Buttons ein/aus
            deactivateButtons()
            self.btnActionLogDelAll.show()
            self.labelActionLogFilter.show()
            self.comboActionLogFilter.show()
        else:
            # Nicht verbunden mit Server -> Ingestion-Tabelle löschen, Buttons deaktivieren
            self.tableLog.clearContents()
            self.tableLog.setRowCount(0)
            self.__lstSelectedActionLogs = []
            self.infoSelectedActionLogs.setText("-")
            self.tableLogContent.setText("")
            self.comboActionLogFilter.blockSignals(True)
            self.comboActionLogFilter.setCurrentIndex(0)
            self.comboActionLogFilter.blockSignals(False)
            deactivateButtons()

    def status_refresh_myWorms(self):

        # Daten
        colGradient = ColorGradient("#89c496", "#ff9494", 100)

        iTaxaTotal = self.__model.subsystem_getMyWorms().puffer_getSizeTaxa()
        fTaxaAgeMedian = self.__model.subsystem_getMyWorms().puffer_getMeanAge()
        iTaxaAgeOverMax = self.__model.subsystem_getMyWorms().puffer_getOverMaxAge()
        fTaxaAgeOverMaxProz = self.__model.subsystem_getMyWorms().puffer_getOverMaxAgeProzent()
        iAgeMax = self.__model.subsystem_getMyWorms().puffer_getMaxAgeAllowed()

        fTaxaAgeMedian_ration_to_MaxAgeProz = (int(fTaxaAgeMedian) / iAgeMax) * 100

        # Alle Taxa in Puffer
        iPufferSize = self.__model.subsystem_getMyWorms().puffer_getSizeTaxa()
        sPufferSize = "-" if iPufferSize == 0 else str(iPufferSize)
        self.taxSysPufferInfo.setText(sPufferSize)
        iPufferSizeQuery = self.__model.subsystem_getMyWorms().puffer_getSizeQuery()
        sPufferSizeQuery = "-" if iPufferSizeQuery == 0 else str(iPufferSizeQuery)
        self.taxSysQueryInfo.setText(sPufferSizeQuery)

        # Mittleres Alter der Taxa in Puffer
        widget = self.taxSysPufferAge

        iValShow = int(fTaxaAgeMedian_ration_to_MaxAgeProz)
        if iValShow > 100:
            iValShow = 100
        widget.setToolTip("Mean taxa age in MyWorms")
        widget.setFormat(f'{fTaxaAgeMedian:.1f}'+"d")
        widget.setValue(iValShow)
        widget.setStyleSheet(CssPresets.progressBar_grey)
        palette = QPalette(widget.palette())
        iR, iG, iB = colGradient.getRGB(100-iValShow)
        palette.setColor(QPalette.ColorRole.Text, QColor(Colors.getColor("black")))
        palette.setColor(QPalette.ColorRole.HighlightedText, QColor(Colors.getColor("black")))
        palette.setColor(QPalette.Highlight, QColor(iR, iG, iB))
        widget.setPalette(palette)

        # Problematische Taxa
        iMisfits = len(self.__model.subsystem_getMyWorms().puffer_getProblemTaxa_all())
        self.taxSysPufferProblems.setText(str(iMisfits))
        if iMisfits > 0:
            palette = QPalette(self.taxSysPufferProblems.palette())
            palette.setColor(QPalette.Text, QColor("#ff9494"))
            self.taxSysPufferProblems.setPalette(palette)
        else:
            self.taxSysPufferProblems.setPalette(QPalette())

        # Taxa über Max Alter
        widget = self.taxSysPufferOverMaxAgeNumber
        widget.setToolTip("Taxa older than allowed max. age of " + str(iAgeMax) + " days")
        widget.setStyleSheet(CssPresets.progressBar_grey)
        if iTaxaAgeOverMax > 0:
            widget.setText(str(iTaxaAgeOverMax))
            palette = QPalette()
            palette.setColor(QPalette.Text, QColor("#ff9494"))
            widget.setPalette(palette)
            self.taxSysPufferStatus.setIcon(QIcon(os.path.dirname(os.path.realpath(__file__)) + "/ressource/icon_thumbs_down.png"))
        elif iTaxaAgeOverMax == 0 and iTaxaTotal > 0:
            widget.setText("0")
            palette = QPalette()
            self.taxSysPufferStatus.setIcon(QIcon(os.path.dirname(os.path.realpath(__file__)) + "/ressource/icon_thumbs_up.png"))
            #palette.setColor(QPalette.Text, QColor("#89c496"))
            widget.setPalette(palette)
        else:
            widget.setText("0")
            palette = QPalette()
            self.taxSysPufferStatus.setIcon(QIcon(os.path.dirname(os.path.realpath(__file__)) + "/ressource/icon_frage.png"))
            widget.setPalette(palette)

        # Prozent über Max Alter
        sTextTaxaOverAge = "-"
        widget = self.taxSysPufferOverMaxAge
        widget.setToolTip("Ratio of taxa older than allowed max. age of "+str(iAgeMax)+" days")
        widget.setFormat(f'{fTaxaAgeOverMaxProz*100:.1f}' + "%")
        palette = QPalette(widget.palette())
        iValShow = int(fTaxaAgeOverMaxProz*100)
        if iValShow > 100:
            iValShow = 100
        iR, iG, iB = colGradient.getRGB(iValShow)
        palette.setColor(QPalette.ColorRole.Text, QColor(Colors.getColor("black")))
        palette.setColor(QPalette.ColorRole.HighlightedText, QColor(Colors.getColor("black")))
        palette.setColor(QPalette.Highlight, QColor(iR, iG, iB))
        widget.setPalette(palette)
        widget.setValue(iValShow)

    def ui_setBatch(self, iNr, iMax, sInfo):
        """
        Update Batch
        """
        if (iNr, iMax, sInfo) == (-1,-1,"end"):
            self.ui_hideBatch()
        else:
            self.ui_showBatch()
            iProz = 0
            if iMax != 0:
                iProz = int((iNr / iMax) * 100)
            self.batch.setValue(iProz)
            sProz = str(iNr) + "/" + str(iMax)
            self.batch.setFormat(sInfo)

    def status_refresh_sysMon(self):
        """
        Update System Monitor
        """
        try:
            process = psutil.Process(os.getpid())
            colGradient = ColorGradient( "#ff9494", "#89c496", 100)
            iMemUse = process.memory_info().rss / (1024 * 1024)
            iMemUsePercent = process.memory_percent("rss")
            iCpuUsePercent = int(process.cpu_percent())
            if iCpuUsePercent > 100:
                iCpuUsePercent = 100
            sMon = "RAM usage: " + str("{:.0f}".format(iMemUse)) + "MB " + str("{:.2f}".format(iMemUsePercent)) + "%"
            self.sysmon.setValue(iMemUsePercent)
            self.sysmon.setFormat(str("{:.0f}".format(iMemUse)+" MB"))
            self.sysmon.setToolTip(sMon)
            palette = QPalette(self.sysmon.palette())
            iR, iG, iB = colGradient.getRGB(100 - iCpuUsePercent)
            palette.setColor(QPalette.Highlight, QColor(iR, iG, iB))
            palette.setColor(QPalette.ColorRole.Text, QColor(Colors.getColor("black")))
            palette.setColor(QPalette.ColorRole.HighlightedText, QColor(Colors.getColor("black")))
            self.sysmon.setPalette(palette)
        except:
            self.sysmon.setValue(0)
            self.sysmon.setFormat(str("-"))
            self.sysmon.setToolTip("RAM usage: <not available>")
            self.sysmon.setToolTip("RAM usage: <not available>")
            self.sysmon.setPalette(QPalette())

    def status_refresh_adminLock(self):
        """
        Set Gui Admin Lock
        """

        # Admin Lock
        bOpen = self.__model.conf_getAdminLock()
        # Schloss
        if bOpen:
            self.buttonAdminLock.setIcon(QIcon(os.path.dirname(os.path.realpath(__file__)) + "/ressource/icon_lock_open.png"))
            self.labelAdminLock.setText("The core database administration area is open. Dry mode is not working here, be careful what you do ...")
            self.admin_out.setPlainText("READY \u2588")
        else:
            self.buttonAdminLock.setIcon(QIcon(os.path.dirname(os.path.realpath(__file__)) + "/ressource/icon_lock_closed.png"))
            self.labelAdminLock.setText("The core database administration area is locked ...")
            self.admin_out.setPlainText("LOCKED \u2588")
        # Gruppen
        self.adminGroupOutput.setEnabled(bOpen)
        self.adminGroupAccess.setEnabled(bOpen)
        self.adminGroupBasicCommandos.setEnabled(bOpen)
        self.adminGroupCommandosSchema.setEnabled(bOpen)
        self.adminGroupDbInfo.setEnabled(bOpen)
        # Textfelder
        sPsqlVersion = str(self.__model.conf_getAdminPsqlVersion()) if self.__model.core_isAdminPsqlWorking() else "not working"
        sPgDumpVersion = str(self.__model.conf_getAdminPgdumpVersion()) if self.__model.core_isAdminPgdumpWorking() else "not working"
        self.admin_pg_psql_version.setText(sPsqlVersion)
        self.admin_pg_dump_version.setText(sPgDumpVersion)
        sVersionSchema = self.__model.conf_getInfo("dbschema")
        if sVersionSchema is None:
            sVersionSchema = ""
        self.admin_current_schema_version.setText(sVersionSchema)
        self.admin_current_schema_version.setCursorPosition(0)
        #self.admin_cmd_new_schema__marlin_science.setText("marlin_s v."+sVersionSchema)
        #self.admin_cmd_new_schema__marlin_science.setCursorPosition(0)

    # --- GUI Helfer für Visualisierer

    def explorer_reset_visual(self):
        """
        Reset Visual Explorer
        """
        if not self.__model.subsystem_getConnector().isReady():
            return
        self.__model.ui_selector_getSelector().reset()
        self.buttonDrawMap.setStyleSheet(None)
        self.buttonDrawMap.setText("draw")
        self.explorer_setSelector()
        self.build_browser_emptyMap()
        self.status_clean_selector_statistics()
        self.__bMapForceDraw = False

    def explorer_update(self, sCmdSelector=None):
        """
        Update Visual Explorer
        """

        if self.sys_isWorkerRunning(popup=True):
            return

        if not self.__model.subsystem_getConnector().isReady():
            return
        #
        pn = self.ui_printConsoleNormal
        pv = self.ui_printConsoleVerbose
        pd = self.ui_printConsoleDebug

        if self.sys_isWorkerRunning(popup=True):
            return

        if not self.__model.subsystem_getConnector().isReady():
            return

        # --- Hole Selector-Einstellung

        dicInputSelector = {}
        if sCmdSelector == "set_date_start_min":
            # *** date_start auf minimum setzen
            date = self.__model.ui_selector_getSelector().getSpecial("date_start_min")
            if date is None:
                self.ui_hideProgress()
                return
            dicInputSelector = {"date_start": {date: True}}
        elif sCmdSelector == "set_date_end_max":
            # *** date_end auf maximum setzen
            date = self.__model.ui_selector_getSelector().getSpecial("date_end_max")
            if date is None:
                self.ui_hideProgress()
                return
            dicInputSelector = {"date_end": {date: True}}
        elif sCmdSelector == "toggle_type":
            # *** Toggle an/aus: type
            dicInputSelector = {"type": self.vizInType.getEntries(toggle_on_off=True)}
        elif sCmdSelector == "toggle_region":
            # *** Toggle an/aus: region
            dicInputSelector = {"region": self.vizInRegion.getEntries(toggle_on_off=True)}
        elif sCmdSelector == "toggle_dataset":
            # *** Toggle an/aus: dataset
            dicInputSelector = {"dataset": self.vizInDataset.getEntries(toggle_on_off=True)}
        elif sCmdSelector == "toggle_gear":
            # *** Toggle an/aus: gear
            dicInputSelector = {"gear": self.vizInGear.getEntries(toggle_on_off=True)}
        elif sCmdSelector == "toggle_taxcover":
            # *** Toggle an/aus: Taxcover
            dicInputSelector = {"taxcover": self.vizInTaxcover.getEntries(toggle_on_off=True)}
        elif sCmdSelector == "toggle_ingest":
            # *** Toggle an/aus: ingest
            dicInputSelector = {"ingest": self.vizInIngest.getEntries(toggle_on_off=True)}
        elif sCmdSelector == "inv_type":
            # *** Invert: type
            dicInputSelector = {"type": self.vizInType.getEntries(invert=True)}
        elif sCmdSelector == "inv_region":
            # *** Invert: Region
            dicInputSelector = {"region": self.vizInRegion.getEntries(invert=True)}
        elif sCmdSelector == "inv_dataset":
            # *** Invert: dataset
            dicInputSelector = {"dataset": self.vizInDataset.getEntries(invert=True)}
        elif sCmdSelector == "inv_gear":
            # *** Invert: gear
            dicInputSelector = {"gear": self.vizInGear.getEntries(invert=True)}
        elif sCmdSelector == "inv_taxcover":
            # *** Invert: Taxcover
            dicInputSelector = {"taxcover": self.vizInTaxcover.getEntries(invert=True)}
        elif sCmdSelector == "inv_ingest":
            # *** Invert: ingest
            dicInputSelector = {"ingest": self.vizInIngest.getEntries(invert=True)}
        else:
            # *** Normale Eingabe
            if self.vizInDateStart.date() > self.vizInDateEnd.date():
                # Datum vertauscht?
                tmp = self.vizInDateStart.date()
                self.vizInDateStart.setDate(self.vizInDateEnd.date())
                self.vizInDateEnd.setDate(tmp)

            dateStart = self.vizInDateStart.date().toPyDate()
            dateEnd = self.vizInDateEnd.date().toPyDate()
            dicType  = self.vizInType.getEntries()
            dicRegion  = self.vizInRegion.getEntries()
            dicDataset = self.vizInDataset.getEntries()
            dicGear = self.vizInGear.getEntries()
            dicTaxcover = self.vizInTaxcover.getEntries()
            dicIngest = self.vizInIngest.getEntries()

            dicInputSelector = {
                "date_start": {dateStart: True},
                "date_end": {dateEnd: True},
                "type": dicType,
                "region": dicRegion,
                "dataset": dicDataset,
                "gear": dicGear,
                "taxcover": dicTaxcover,
                "ingest": dicIngest
            }

        # --- Spezial: Mappe Type: Long -> Short

        if "type" in dicInputSelector:
            dicType_long = dicInputSelector["type"]
            dicType_short = {}
            for sWf_long, bSwitch in dicType_long.items():
                dicType_short[self.__map.workflow_long2short(sWf_long)] = bSwitch
            dicInputSelector["type"] = dicType_short

        # --- Hole Map Optionen
        sMarkerType = self.vizMarkerType.currentText()
        sProjectionType = self.vizProjectionType.currentText()
        bWorldWrap = self.vizWorldWrap.checkState()

        # --- Verpacke alle nötigen Optionen für Worker
        self.__bMapForceDraw = self.__bMapForceDraw if sCmdSelector == "draw_map" else False
        dicOpt = {
            "selector": dicInputSelector,
            "map_marker_type": sMarkerType,
            "map_projection_type": sProjectionType,
            "force_draw": self.__bMapForceDraw,
            "world_wrap": bWorldWrap
        }

        # Selector deaktivieren
        self.explorer_selectorEnable(False)

        # --- Worker starten
        self.worker_start(Worker_VisUpdate(self.__model, dicOpt), self.worker_finished_updateVis)

    def explorer_selectorEnable(self, bOn=True):
        """
        Aktiviere den Auswahlbereich für die Visualisierung

        Args:
            bOn (bool): De-/aktivieren
        """

        # Combo
        self.vizInDateStart.setEnabled(bOn)
        self.vizInDateEnd.setEnabled(bOn)
        self.vizInDataset.setEnabled(bOn)
        self.vizInType.setEnabled(bOn)
        self.vizInRegion.setEnabled(bOn)
        self.vizInTaxcover.setEnabled(bOn)
        self.vizInIngest.setEnabled(bOn)
        self.vizInGear.setEnabled(bOn)
        self.vizProjectionType.setEnabled(bOn)
        self.vizMarkerType.setEnabled(bOn)

        # Buttons
        self.buttonDrawMap.setEnabled(bOn)
        self.buttonVizInDateStartMin.setEnabled(bOn)
        self.buttonVizInDateEndMax.setEnabled(bOn)
        self.buttonVizInDatasetAll.setEnabled(bOn)
        self.buttonVizInDatasetInv.setEnabled(bOn)
        self.buttonVizInTypeAll.setEnabled(bOn)
        self.buttonVizInTypeInv.setEnabled(bOn)
        self.buttonVizInRegionAll.setEnabled(bOn)
        self.buttonVizInRegionInv.setEnabled(bOn)
        self.buttonVizInGearAll.setEnabled(bOn)
        self.buttonVizInGearInv.setEnabled(bOn)
        self.buttonVizInTaxcoverAll.setEnabled(bOn)
        self.buttonVizInTaxcoverInv.setEnabled(bOn)
        self.buttonVizInIngestAll.setEnabled(bOn)
        self.buttonVizInIngestInv.setEnabled(bOn)
        self.buttonVizInAllRst.setEnabled(bOn)
        self.buttonExportSubset.setEnabled(bOn)
        self.buttonShowSubset.setEnabled(bOn)
        self.buttonSaveMap.setEnabled(bOn)
        self.buttonSaveMap2.setEnabled(bOn)
        self.buttonBrowserMap.setEnabled(bOn)
        self.buttonBrowserMap2.setEnabled(bOn)
        self.buttonBrowserReload.setEnabled(bOn)
        self.buttonBrowserBack.setEnabled(bOn)
        self.buttonBrowserForward.setEnabled(bOn)
        self.vizWorldWrap.setEnabled(bOn)

        # ... die Statistik haengt immer dran
        self.explorer_selectorSatisticsEnable(bOn)

    def explorer_selectorSatisticsEnable(self, bOn=True):
        self.vizStatCruise.setEnabled(bOn)
        self.vizStatStation.setEnabled(bOn)
        self.vizStatSample.setEnabled(bOn)
        self.vizStatSubSample.setEnabled(bOn)
        self.vizStatSubset.setEnabled(bOn)
        self.vizStatBiota.setEnabled(bOn)
        self.vizStatBiotaPresenceAbsence.setEnabled(bOn)
        self.vizStatBiotaCount.setEnabled(bOn)
        self.vizStatBiotaAbundance.setEnabled(bOn)
        self.vizStatIndMeasure.setEnabled(bOn)
        self.vizStatSediment.setEnabled(bOn)
        self.vizStatTaxa.setEnabled(bOn)

    def exporer_selectorClear(self):
        """
        Clear Auswahlbereich für die Visualisierung
        """

        lstWidgets = [self.vizInDateStart, self.vizInDateEnd]
        self.ui_blockSignals(lstWidgets)

        self.vizInDateStart.setDate(QDate.currentDate())
        self.vizInDateEnd.setDate(QDate.currentDate())
        self.vizInRegion.clear()
        self.vizInDataset.clear()
        self.vizInType.clear()
        self.vizInTaxcover.clear()
        self.vizInGear.clear()
        self.vizInIngest.clear()
        self.status_clean_selector_statistics()
        self.__model.ui_selector_getSelector().reset()

        self.ui_deblockSignals(lstWidgets)

    def explorer_setSelector(self):
        """
        Stelle Auswahlbereich für die Visualisierung ein

        Args:
            selector (database.Selector): Selector
        """

        pn = self.ui_printConsoleNormal
        pv = self.ui_printConsoleVerbose
        pd = self.ui_printConsoleDebug

        if not self.__model.subsystem_getConnector().isReady():
            return

        # Ist Selector OK?
        if self.__model.ui_selector_getSelector().isErr():
            pn("* Error in selector configuration")
            pn(self.__model.ui_selector_getSelector().getResult().getErrWarnInfoString(short=True))
            return

        # Info
        pd("* Selector configuration")
        pd(self.__model.ui_selector_getSelector().infoConfigShort())

        # Hilfsmethode: Setze einen Selector
        def setSelector(comboSelectorWidget, sSelectorName):
            if sSelectorName == "type":
                # Workflow
                # Umbiegen der Workflows in Lange Version
                dicTmp1 = self.__model.ui_selector_getSelector().getEntries(sSelectorName)
                dicTmp2 = {}
                for k, v in dicTmp1.items():
                    dicTmp2[self.__map.workflow_short2long(k)] = v
                comboSelectorWidget.setEntries(dicTmp2)
            elif sSelectorName == "region":
                # Region
                comboSelectorWidget.setEntries(self.__model.ui_selector_getSelector().getEntries(sSelectorName))
                # Rot wenn keine Region vorhanden
                iRegions = len(self.__model.ui_selector_getSelector().getEntries(sSelectorName))
                palette = QPalette()
                if len(self.__model.ui_selector_getSelector().getEntries(sSelectorName)) == 0:
                    palette.setColor(QPalette.ButtonText, QColor(Colors.getColor("red")))
                    comboSelectorWidget.setToolTip("No regions available! Please map samples to regions first ...")
                else:
                    comboSelectorWidget.setToolTip(None)
                comboSelectorWidget.setPalette(palette)
            else:
                # Rest
                comboSelectorWidget.setEntries(self.__model.ui_selector_getSelector().getEntries(sSelectorName))

            # Info
            sSelectorInfoText = str(len(self.__model.ui_selector_getSelector().getEffectiveActiveEntries(sSelectorName))) + "/" + str(len(self.__model.ui_selector_getSelector().getEffectiveEntries(sSelectorName)))
            if self.__model.conf_isDebug():
                sSelectorInfoText += " [" + str(len(self.__model.ui_selector_getSelector().getActivEntries(sSelectorName))) + "/" + str(len(self.__model.ui_selector_getSelector().getEntries(sSelectorName))) + "]"
            comboSelectorWidget.setName(sSelectorInfoText)

        # Hole alle Optionen für jeden Selector
        if self.__model.ui_selector_getSelector().isErr():
            pn("* Error in Selector")
            pn(self.__model.ui_selector_getSelector().getResult())
            return

        # Block
        lstWidgets = [self.vizInDateStart, self.vizInDateEnd]
        self.ui_blockSignals(lstWidgets)

        # Lösche alle Selektoren
        self.vizInDateStart.setDate(QDate.currentDate())
        self.vizInDateEnd.setDate(QDate.currentDate())
        self.vizInDataset.clear()
        self.vizInRegion.clear()
        self.vizInType.clear()
        self.vizInTaxcover.clear()
        self.vizInGear.clear()
        self.vizInIngest.clear()

        # Setze alle Selektoren
        dateStart = self.__model.ui_selector_getSelector().getActivEntries("date_start")
        dateEnd = self.__model.ui_selector_getSelector().getActivEntries("date_end")
        if dateStart != []:
            self.vizInDateStart.setDate(dateStart[0])
        if dateEnd != []:
            self.vizInDateEnd.setDate(dateEnd[0])
        setSelector(self.vizInRegion, "region")
        setSelector(self.vizInType, "type")
        setSelector(self.vizInDataset, "dataset")
        setSelector(self.vizInTaxcover, "taxcover")
        setSelector(self.vizInGear, "gear")
        setSelector(self.vizInIngest, "ingest")

        # Deblock
        self.ui_deblockSignals(lstWidgets)

    def status_update_selector_statistics(self):
        """
        Stelle den Selector Statistikbereich ein
        """

        dicStatistics = self.__model.ui_selector_getStatistics()
        if dicStatistics is None:
            return

        iPopAll = dicStatistics["biota"] if "biota" in dicStatistics else 0
        iPopPresenceAbsence = dicStatistics["biota_presence_absence"] if "biota_presence_absence" in dicStatistics else 0
        iPopAbundance = dicStatistics["biota_abundance"] if "biota_abundance" in dicStatistics else 0
        iPopCount = iPopAll - iPopPresenceAbsence - iPopAbundance

        self.vizStatCruise.setText(str(dicStatistics["cruises"]) if "cruises" in dicStatistics else "err")
        self.vizStatStation.setText(str(dicStatistics["stations"]) if "stations" in dicStatistics else "err")
        self.vizStatSample.setText(str(dicStatistics["samples"])  if "samples" in dicStatistics else "err")
        self.vizStatSubSample.setText(str(dicStatistics["subsamples"])  if "subsamples" in dicStatistics else "err")
        self.vizStatSubset.setText(str(dicStatistics["subsets"])  if "subsets" in dicStatistics else "err")
        self.vizStatBiota.setText(str(iPopAll))
        self.vizStatBiotaPresenceAbsence.setText(str(iPopPresenceAbsence))
        self.vizStatBiotaCount.setText(str(iPopCount))
        self.vizStatBiotaAbundance.setText(str(iPopAbundance))
        self.vizStatIndMeasure.setText(str(dicStatistics["indmeasure"]) if "indmeasure" in dicStatistics else "err")
        self.vizStatSediment.setText(str(dicStatistics["sediment"]) if "sediment" in dicStatistics else "err")
        self.vizStatTaxa.setText(str(dicStatistics["taxa"]) if "taxa" in dicStatistics else "err")

        # Multi Subsets vorhanden?
        iSubsets_total = dicStatistics["subsets"]
        if iSubsets_total is None: iSubsets_total = 0
        iSubSample_multiSubsets = dicStatistics.get("subsample_multi_subset")
        iSubSample_noSubsets = dicStatistics.get("subsample_no_subset")
        iSubSample_singleSubsets = dicStatistics.get("subsample_single_subset")
        if iSubSample_multiSubsets is None: iSubSample_multiSubsets = 0
        if iSubSample_singleSubsets is None: iSubSample_singleSubsets = 0
        if iSubSample_noSubsets is None: iSubSample_noSubsets = 0
        iSubsets_multi_total = iSubsets_total - iSubSample_singleSubsets
        if iSubSample_multiSubsets > 0:
            lstTable = [["Type", "Count"]]
            if iSubSample_noSubsets > 0:
                lstTable.append(["SubSamples with 0 Subset attached", iSubSample_noSubsets])
            lstTable.append(["SubSamples with 1 Subset attached", iSubSample_singleSubsets])
            lstTable.append(["SubSamples with 1+ Subsets attached", iSubSample_multiSubsets])
            lstTable.append(["1+ attached Subsets", iSubsets_multi_total])
            sList_ascii = ""
            for iRow, (sType, iCnt) in enumerate(lstTable):
                if iRow > 0:
                    sList_ascii += sType +": " +str(iCnt) +"x\n"
            sTable_ascii = PrettyText.table(lstTable)
            sInfo_ascii = "* Caution: SubSamples with more than one attached Subset found (" + str(
                iSubSample_multiSubsets) + "x)\n"
            sInfo_ascii += sTable_ascii
            self.ui_printConsoleNormal(sInfo_ascii)
            self.popup_generic(title="Caution", head="Multiple Subsets found", icon="warn", info="SubSamples with more than one attached Subset found: "+str(iSubSample_multiSubsets)+"x", detail=sList_ascii)

    def status_clean_selector_statistics(self):
        """
        Clear Statusbereich für die Visualisierung
        """
        
        self.__model.ui_selector_setStatistics(None)
        lstWidgets = [self.vizStatCruise,
                      self.vizStatStation,
                      self.vizStatSample,
                      self.vizStatSubSample,
                      self.vizStatSubset,
                      self.vizStatBiota,
                      self.vizStatBiotaPresenceAbsence,
                      self.vizStatBiotaCount,
                      self.vizStatBiotaAbundance,
                      self.vizStatIndMeasure,
                      self.vizStatSediment,
                      self.vizStatTaxa]
        for widget in lstWidgets:
            #widget.setEnabled(False)
            widget.setText("-")

    def explorer_browser_reloadMap(self):
        """
        Reload map
        """
        self.__browser.stop()
        self.__browser.load(QUrl.fromLocalFile(self.__model.conf_getMap()))

    def status_clean_ingestDepenencies(self):
        """
        CLear Ingest Infos
        """
        self.__lstSelectedIngests = []
        self.infoSelectedIngest.setText("-")
        self.btnIngestDepShow.hide()
        self.btnIngestDepHide.hide()
        self.labelIngestDepShowHide.hide()

    # --- GUI Helper für Datenbearbeitung

    def explorer_modelStatisticsClear(self):
        """
        Lösche Statistiken
        """

        self.__model.ui_selector_setStatistics(None)

    def dialog_input_generic(self, sTitel, sTabelle, lstAttributAusschluss=["id"], **kwargs):
        """
        Erzeuge Dialog zur generischer Eingabe eines neuen Datenstatzes in eine Tabelle
        - Die Attribute werden aus der entsprechenden Tabelle bestimmt

        Args:
            sTitel (str): Titel des Dialoges
            sTabelle (str): Tabellenname
            lstAttributAusschluss (list()): Liste der Attribute, welche ausgeschlossen werden sollen
        """

        pn = self.ui_printConsoleNormal
        pv = self.ui_printConsoleVerbose
        pd = self.ui_printConsoleDebug

        result = Result()
        result.addData("tab", sTabelle)

        # DB: Hole Attribute, Kommentare, Obligatinfo
        try:
            self.ui_showProgress()
            self.ui_setWorkflow(top="Building input dialog", sub="Working")
            conn = psycopg2.connect(self.__model.subsystem_getDb().conn_getConnStrPsql())
            cursor = conn.cursor()
            lstAttr, dicAttrObligat, dicAttrComments = self.input_dialog_generic_getAttr(cursor, sTabelle, lstAttributAusschluss)
            cursor.close()
            conn.close()
            self.ui_hideProgress()
        except Exception as e:
            sErr = "Error (" + sTitel + "): Access to database (getting attribute, ...): " + str(e)
            pn(sErr)
            self.ui_hideProgress()
            return False

        # Wiederhole Eingabe bis alles OK ist ...
        dicDefault = {}
        # Dialog erzeugen
        dia, dicWidgets = self.dialog_input_createGeneric(sTitel, lstAttr, sTabelle, dicAttrObligat, dicAttrComments, showCsv=False, defaultData=dicDefault)

        # Zentrieren zum Hauptfenster
        self.ui_centerWidget(dia)

        # Dialog anzeigen
        iAntwort = dia.exec()

        # DB: Dialog auswerten
        if iAntwort == 1:

            # ---------------------------------------------------------------
            # Einzeleintrag in DB (läuft im Gui-Thread, weil sehr schnell)
            # ---------------------------------------------------------------

            # Hole Daten aus Dialog
            dicVal = self.dialog_input_generic_evalDialog(dicWidgets, sTabelle)
                # Trage in DB ein
            if not all(v == "" for k, v in dicVal.items()):
                # Nur wenn Daten eingetragen wurden
                if sTabelle == "wormsbase":
                    # TaxonWizard: Wormsbase
                    try:
                        iAid = dicVal["aid"]
                        # --- MyWoRMS
                        resMyWorms = self.__model.subsystem_getMyWorms().getTaxon(iAid)
                        if resMyWorms.isErr():
                            self.popup_error("Taxon not added", "Problems asking MyWoRMS", resMyWorms.getErrWarnInfoString(short=True))
                            result.addResult(resMyWorms, msg_only=True)
                        else:
                            taxon = resMyWorms.getData("taxon")
                            if not resMyWorms.getData("found"):
                                self.popup_error("Taxon not found", "Taxon not found in MyWoRMS")
                            else:
                                # Taxon gefunden
                                # --- wormsbase
                                resWormsBase = self.__model.subsystem_getDb().table_isEntryInTable("wormsbase", "aid",iAid)
                                if resWormsBase.isErr():
                                    self.popup_error("Searching failed", "Searching in wormsbase failed", resWormsBase.getErrWarnInfoString(short=True))
                                else:
                                    if resWormsBase.getData("found"):
                                        # Taxon ist bereits in Base list
                                        self.popup_generic(title="New Taxon", head="Taxon already in Base List", info=str(taxon.signature()))
                                    else:
                                        # Add to wormsbase
                                        resAddBase = self.__model.subsystem_getTaxonWizard().addBaseTaxa(taxon)
                                        if resAddBase.isErr():
                                            self.popup_error("Taxon not added", "Taxon could not be added to base list", resAddBase.getErrWarnInfoString(short=True))
                    except Exception as ex:
                        # Probleme bei der MyWoRMS abfrage
                        self.popup_error("Taxon not added", "An unespected problem occurred ...", str(ex))
                        return
                else:
                    # Rest
                    conn = psycopg2.connect(self.__model.subsystem_getDb().conn_getConnStrPsql())
                    cursor = conn.cursor()
                    resAddLookup = self.__model.subsystem_getDb().table_insertIntoTable(sTabelle, dicVal, cursor=cursor)  # Läuft direkt im GUI-Thread weil es sehr schnell geht
                    if resAddLookup.isErr():
                        # Fehler
                        sFehler = "Error (" + sTitel + "): Access to database (create entry): " + resAddLookup.getResult()
                        conn.rollback()
                        cursor.close()
                        conn.close()
                        # Popup
                        sInfo = "<b>Table</b>"
                        sInfo += "<ul><li>" + sTabelle + "</li></ul>"
                        sTitle = "Error Adding Lookup"
                        if self.__model.conf_isDry():
                            sTitle += " in dry mode"
                        self.popup_error(sTitle, sInfo, resAddLookup.getResult())
                    else:
                        # OK
                        iId = resAddLookup.getData("id")
                        if self.__model.conf_isDry():
                            conn.rollback()
                        else:
                            # Eintrag in Action Log
                            # Action Log
                            if sTabelle in self.__map.lookup_getDbTables():
                                sLookupAktion = "lup-add"
                            else:
                                sLookupAktion = "tax-adm-add"
                            sInfoOut = "Table: " + sTabelle + "\n"
                            if "name" in dicVal:
                                sInfoOut += "Name: " + str(dicVal["name"]) + "\n"
                            sInfoOut += "Data:\n"
                            for k, v in dicVal.items():
                                if k != "name":
                                    sInfoOut += str(k) + ": " + str(v) + "\n"
                            result_actionlog = self.__model.subsystem_getDb().actionlog_new(sLookupAktion, sTabelle, sInfoOut, cursor=cursor)
                            result.addResult(result_actionlog)
                            conn.commit()
                        cursor.close()
                        conn.close()
                        # Info
                        sInfo = "<b>Table</b>"
                        sInfo += "<ul><li>" + sTabelle + "</li></ul>"
                        sInfo += "<b>Record ID</b>"
                        sInfo += "<ul><li>" + str(iId) + "</li></ul>"
                        sTitle = "Lookup was added"
                        if self.__model.conf_isDry():
                            sTitle += " in dry mode"
                        self.popup_generic(title="Lookup", head="Add Lookup", info=sInfo)
                # Refresh Gui (Status update leider im UI-Thread)
                if self.__map.all_isTableTaxon(sTabelle):
                    sSystem = "wizard"
                else:
                    sSystem = "lookup"
                # Starte Initialisierung im Thread Container
                self.worker_start(Worker_ServerReconnect(self.__model, start=sSystem, status=sSystem), self.worker_finished_serverConnection,
                                  server=True, comment=False, ingdep=False)
                return True
        elif iAntwort == 2:
            # import
            if self.__map.all_isTableTaxon(sTabelle):
                self.import_taxon(tabs=[sTabelle])
            else:
                self.import_lookups(tabs=[sTabelle])

    def admin_export_taxonList(self):
        """
        Exportiere Taxon
        """

        if not self.__model.subsystem_getConnector().isReady():
            self.popup_offline()
            return

        if self.sys_isWorkerRunning(popup=True):
            return

        diaOpenIngest = DialogExport(self.__model, lookup=False, taxon=True)
        self.ui_centerWidget(diaOpenIngest)
        ant = diaOpenIngest.exec_()
        if ant == 0:
            return
        # Dialog auswerten
        res = diaOpenIngest.getConfig()

        sFile = res.getData("file")
        lstTaxon = res.getData("taxon")

        # Starte Worker
        self.worker_start(Worker_ExportTaxon(self.__model, sFile, lstTaxon), self.worker_finished_exportExcel)

    def import_taxon(self, **kwargs):
        """
        Importiere Taxon
        kwargs:
            tabs (lst): Liste der zu importierenden Taxon Tabellen
        """

        lstTabs = kwargs["tabs"] if "tabs" in kwargs else None

        if not self.__model.subsystem_getConnector().isReady():
            self.popup_offline()
            return

        if self.sys_isWorkerRunning(popup=True):
            return

        diaOpenIngest = DialogIngest(self.__model, taxon=True, taxon_tabs=lstTabs)
        self.ui_centerWidget(diaOpenIngest)
        ant = diaOpenIngest.exec_()
        if ant == 0:
            return
        # Dialog auswerten
        res = diaOpenIngest.getConfig()

        sFile = res.getData("file")
        lstTaxon = res.getData("taxon")
        bUpdate = res.getData("taxon-update")
        bTamer = res.getData("tamer")

        # Starte Worker
        self.worker_start(Worker_ImportTaxon(self.__model, sFile, bTamer, lstTaxon, bUpdate), self.worker_finished_ingestionTaxon)

    def import_lookups(self, **kwargs):
        """
        Importiere Lookups
        kwargs:
            tabs (lst): Liste der zu importierenden Lookups
        """

        lstTabs = kwargs["tabs"] if "tabs" in kwargs else None

        if not self.__model.subsystem_getConnector().isReady():
            self.popup_offline()
            return

        if self.sys_isWorkerRunning(popup=True):
            return

        diaOpenIngest = DialogIngest(self.__model, lookup=True, lookup_tabs=lstTabs)
        self.ui_centerWidget(diaOpenIngest)
        ant = diaOpenIngest.exec_()
        if ant == 0:
            return
        # Dialog auswerten
        res = diaOpenIngest.getConfig()

        sFile = res.getData("file")
        sFile = res.getData("file")
        lstLookup = res.getData("lookup")
        sExcel = res.getData("lup-excel")
        sMode = res.getData("lup-mode")
        sDb = res.getData("lup-db")
        bTamer = res.getData("tamer")
        sLupReuseReact = res.getData("lup-reuse-react")
        bLupSingle = res.getData("ingest-lup-single")

        # Starte Worker
        self.worker_start(Worker_ImportLookup(self.__model, sFile, bTamer, lstLookup, sExcel, sMode, sDb, sLupReuseReact), self.worker_finished_ingestionLookup)

    def admin_lookup_export(self):
        """
        Exportiere Lookups
        """

        if not self.__model.subsystem_getConnector().isReady():
            self.popup_offline()
            return

        if self.sys_isWorkerRunning(popup=True):
            return

        diaOpenIngest = DialogExport(self.__model, lookup=True, taxon=False)
        self.ui_centerWidget(diaOpenIngest)
        ant = diaOpenIngest.exec_()
        if ant == 0:
            return
        # Dialog auswerten
        res = diaOpenIngest.getConfig()

        sFile = res.getData("file")
        lstLookup = res.getData("lookup")
        sGold = res.getData("lookup-gold-export")

        # Starte Worker
        self.worker_start(Worker_ExportLookup(self.__model, sFile, None, lstLookup, sGold), self.worker_finished_exportExcel)

    def admin_checkDois(self):
        if not self.__model.subsystem_getConnector().isReady():
            self.popup_offline()
            return

        if self.sys_isWorkerRunning(popup=True):
            return

        checkDois = Worker_CheckDois(self.__model)
        self.worker_start(checkDois, self.worker_finished_checkDoi)

    def ui_table_stat(self, sTab):
        """
        Zeige Statistik einer Tabelle

        Args:
            sTab (str): Tabellenname
        """

        if not self.__model.subsystem_getConnector().isReady():
            self.popup_offline()
            return

        if self.sys_isWorkerRunning(popup=True):
            return

        pn = self.ui_printConsoleNormal
        pv = self.ui_printConsoleVerbose
        pd = self.ui_printConsoleDebug

        statTable = Worker_StatisticOnTable(self.__model, sTab)
        self.worker_start(statTable, self.worker_finished_showText)

    def table_show(self, sTab, **kwargs):
        """
        Zeige Inhalt einer Tabelle

        Args:
            sTab (str): Tabellenname

        Kwargs:
            id (int): Datensatz Id auf die fokussiert werden soll
            type (str): Art der Tabelle (ingest-raw, ingest, lookup, taxon)
            used (bool): Lookup benutzt
            gold (bool): Nur Gold oder Nicht-Gold Einträge anzeigen
            access (bool): Zugang gewährleistet (irgendwie)
            doi (bool): DOI ok
            doi_pangaea (bool): DOI führt zu PANGAEA
            gear_double (bool): Gear Double Config
        """
        if not self.__model.subsystem_getConnector().isReady():
            self.popup_offline()
            return

        if self.sys_isWorkerRunning(popup=True):
            return

        # Start Worker
        showWholeTable = Worker_ShowWholeTable(self.__model, sTab, **kwargs)
        self.worker_start(showWholeTable, self.worker_finished_showTable)

    def table_entryEdit(self, sTab, iId, **kwargs):
        """
        Ändere einen Eintrag einer Tabelle

        Args:
            sTab (str): Tabellenname
            iId (int): Id des Eintrages

        Kwargs:
            closeWin (PyQt5.QtWidgets.QWidgets): Fenster, welches nach dem Änderungsvorgang geschlossen werden kann
            fokusCol (str): Name des Feldes auf den der Fokus gelegt werden soll
            gold (bool): Die Tabelle war ein Lookup und wurde im (Not)Gold Modus geöffnet -> Wichtig für das Naeuladen
        Returns:
            (bool): Vorgang war erfolgreich?

        """

        # --- Helper
        def closeDb():
            if self.__model.conf_isDry():
                conn.rollback()
            else:
                conn.commit()
            cursor.close()
            conn.close()

        def clearUi():
            self.ui_hideProgress()
            self.ui_setWorkflow()
            self.ui_giveQtTimeToUpdate()

        # ---

        pn = self.ui_printConsoleNormal
        pv = self.ui_printConsoleVerbose
        pd = self.ui_printConsoleDebug

        # Kwargs
        closeWin = None
        sFokusCol = None
        bGoldEditBulk = False
        bGold = False
        if "closeWin" in kwargs:  # Fenster, welches nach Veränderung geschlossen werden soll
            closeWin = kwargs['closeWin']
        if "fokusCol" in kwargs:
            sFokusCol = kwargs['fokusCol']
        if "gold" in kwargs:
            bGold = kwargs['gold']
        if "gold_edit" in kwargs:
            bGoldEditBulk = kwargs['gold_edit']

        # ---

        # Verbunden mit Server?
        if self.__model.subsystem_getConnector().isReady():

            bOk = False

            try:
                sModulname = "Edit a record"
                pv("* " + sModulname)
                pv("Table: " + sTab)
                pv("Id: " + str(iId))

                # DB-Verbindung
                conn = psycopg2.connect(self.__model.subsystem_getDb().conn_getConnStrPsql())
                cursor = conn.cursor()

                # Daten
                dicDataNeu = dict()

                # DB: Hole aktuellen Datensatz
                resTableEntry = self.__model.subsystem_getDb().table_getTableEntries(sTab, "id", iId, cursor=cursor)
                if resTableEntry.isErr():
                    raise NameError("Problems getting entry from Db", sys="Database", kat="Check Record Existance")
                lstHeader = resTableEntry.getData("header")
                try:
                    lstContent = resTableEntry.getData("content")[0]
                except:
                    lstContent = []

                dicDataAktuell = dict(zip(lstHeader, lstContent))

                pd("Actual value: "+str(dicDataAktuell))

                # DB: Hole Attribute, Kommentare, Obligatinfo
                lstAttr, dicAttrObligat, dicAttrComments = self.input_dialog_generic_getAttr(cursor, sTab, ["id"])

                # Edit-Dialog erzeugen
                if self.__map.all_isTableLookup(sTab):
                    sTitle = self.__map.lookup_table2sheet(sTab)
                elif self.__map.all_isTableTaxon(sTab):
                    sTitle = self.__map.taxon_table2sheet(sTab)
                else:
                    sTitle = sTab
                dia, dicWidgets = self.dialog_input_createGeneric(str(sTitle).title() + " (" + str(iId) + ")", lstAttr, sTab, dicAttrObligat, dicAttrComments, defaultData=dicDataAktuell, fokusCol=sFokusCol)

                # Edit-Dialog anzeigen
                iAnt = dia.exec()
                if iAnt == 0:
                    # # Soll noch ein Fenster geschlossen werden
                    # if closeWin is not None:
                    #     closeWin.close()
                    # # Tabelle wieder anzeigen mit der Id des veraenderten Eintrages
                    # self.tableShow(sTab, iId)
                    return True

                # Dialog auswerten
                dicDataNeu = self.dialog_input_generic_evalDialog(dicWidgets, sTab)
                pd("New values: " + str(dicDataNeu))

                # Fenster schliessen
                if closeWin is not None:
                    closeWin.close()

                self.ui_showProgress()
                self.ui_setWorkflow(top="Edit entry", sub="Working")
                self.ui_giveQtTimeToUpdate()

                if sTab == "wormsbase":
                    # *** Spezialfall: BaseTaxa -> Test auf AphiaId-Veränderung -> Protokoll: Del->Neues Taxon
                    try:
                        iAid_alt = int(dicDataAktuell["aid"])
                        iAid_neu = int((dicDataNeu["aid"]))
                    except:
                        # Zurück
                        closeDb()
                        self.popup_error(sTitle, "Sorry", "Something went wrong with format of the AphiaIds")
                        self.table_show(sTab, id=iId, gold=bGold, gold_edit=bGoldEditBulk)
                        return False
                    if iAid_alt != iAid_neu:
                        # --- Edit AphiaId -> Entspricht -> Del und New -> Action Log
                        resTaxAlt = self.__model.subsystem_getMyWorms().getTaxon(iAid_alt)
                        resTaxNeu = self.__model.subsystem_getMyWorms().getTaxon(iAid_neu)
                        if resTaxAlt.isOk() and resTaxNeu.isOk():
                            taxAlt = resTaxAlt.getData("taxon")
                            taxNeu = resTaxNeu.getData("taxon")
                            if not resTaxNeu.getData("found"):
                                # -- Error: Neues Taxon nicht gefunden
                                closeDb()
                                clearUi()
                                self.popup_error("New Taxon not found", "Taxon with AphiaId not found in MyWoRMS:<p><i>" + str(iAid_neu) + "</i>")
                                self.table_show(sTab, id=iId, gold=bGold, gold_edit=bGoldEditBulk)
                                return False
                            else:
                                # --- OK
                                # Update
                                resUpd = self.__model.subsystem_getDb().table_editTableEntry(sTab, "id", iId, dicDataNeu, cursor=cursor)
                                if resUpd.isErr():
                                    # Update-Fehler
                                    self.popup_error("Error changing a Base Taxon", "Details in console ...", resUpd.getErrWarnInfoString(short=True))
                                else:
                                    # Update-Ok
                                    # Log: Alt
                                    sLogPrae = "Base Taxon's AphiaId was edited from " + str(iAid_alt) + " to " + str(iAid_neu) + "\n"
                                    sLogDel = "So this taxon has been deleted form Base List:\n\n"
                                    sLogDel += taxAlt.signature() +"\n"
                                    sLogDelFull = sLogPrae + sLogDel
                                    self.__model.subsystem_getDb().actionlog_new("tax-base-del", "wormsbase", sLogDelFull, cursor=cursor)
                                    # Log: Neu
                                    sLogNew = "So this taxon has been added to Base List:\n\n"
                                    sLogNew += taxNeu.signature() +"\n"
                                    sLogNewFull = sLogPrae + sLogNew
                                    self.__model.subsystem_getDb().actionlog_new("tax-base-new", "wormsbase", sLogNewFull, cursor=cursor)
                                    # Info
                                    sTitle = "Base Taxon switch"
                                    sInfo = "Due to AphiaId change one Base Taxon has been deleted and another has been added.<p>"
                                    sInfo += "Deleted:<br>" + taxAlt.signature() +"<p>" +"New:<br>" +taxNeu.signature()
                                    self.popup_generic(title="AphiaId changed", head=sTitle, info=sInfo)
                                # Zurück
                                clearUi()
                                closeDb()
                                self.table_show(sTab, id=iId, gold=bGold, gold_edit=bGoldEditBulk)
                                return True
                else:
                    # *** Rest (TaxonWizard Admin Tabellen und Lookups)
                    resUpd = self.__model.subsystem_getDb().table_editTableEntry(sTab, "id", iId, dicDataNeu, cursor=cursor)
                    if resUpd.isErr():
                        # Fehler
                        raise NameError(resUpd.getResult())
                    else:
                        # Ok
                        bOk = True
                        pd("SQL: " + str(cursor.query.decode("utf-8")))
                        pd(str(bOk))
                        if not self.__model.conf_isDry():
                            # Action Log
                            if self.__model.subsystem_getMapper().all_isTableTaxon(sTab):
                                sLookupAktion = "tax-adm-edit"
                            else:
                                sLookupAktion = "lup-edit"
                            sInfoOut = "Table: " + sTab + "\n"
                            if "id" in dicDataAktuell:
                                sInfoOut += "Id: " + str(dicDataAktuell["id"]) + "\n"
                            if "name" in dicDataAktuell:
                                sInfoOut += "Name: " + str(dicDataAktuell["name"]) + "\n"
                            sInfoOut += "Changes:\n"
                            for k, v in dicDataAktuell.items():
                                if k != "id" and k in dicDataNeu and str(v) != str(dicDataNeu[k]):
                                    sInfoOut += str(k) + ": " + str(v)+" -> "+str(dicDataNeu[k])+"\n"
                            self.__model.subsystem_getDb().actionlog_new(sLookupAktion, sTab, sInfoOut, cursor=cursor)
                            conn.commit()
                            pn("Entry was edited -> Workflows will be reinitialized ...")
                            self.sys_server_disconnect()
                            self.sys_server_connect()
                        else:
                            pn("<Dry Run> active")
                            pn("Record would have been edited")

            except Exception as ex:
                bOk = False
                print(traceback.print_exc())
                pn("Error (" + sModulname + "): Access to database: " + str(ex) )
                self.popup_error("Error changing a record", "Details in console ...", str(ex))

            # Aufräumen
            closeDb()
            clearUi()

            # Info
            if bOk:
                # Ok
                sInfo = "<b>Table</b>"
                sInfo += "<ul><li>" + sTab + "</li></ul>"
                sInfo += "<b>Record ID</b>"
                sInfo += "<ul><li>" + str(iId) + "</li></ul>"
                sTitle = "Record was edited"
                if self.__model.conf_isDry():
                    sTitle += " in dry mode"
                self.popup_generic(title="Entry edit", head=sTitle, info=sInfo)

            # Tabelle wieder anzeigen mit der Id des veraenderten Eintrages
            self.table_show(sTab, id=iId, gold=bGold, gold_edit=bGoldEditBulk)
            return bOk

    def table_entryEditGold(self, sTab, lstIds, bBool, **kwargs):
        """
        Ändere alle Goldeinträge auf Wert

        Args:
            sTab (str): Tabellenname
            lstIds (list): Ids der Einträge
            bBool (bool): Wert der eingestellt werden soll
        Kwargs:
            closeWin (PyQt5.QtWidgets.QWidgets): Fenster, welches nach dem Löschvorgang geschlossen werden kann

        Returns:
            (bool): Vorgang war erfolgreich?

        """

        if len(lstIds) == 0:
            return

        if self.sys_isWorkerRunning(popup=True):
            return

        if not self.__model.subsystem_getConnector().isReady():
            self.popup_offline()
            return

        # Rückfrage
        sTitle = "Set status to: " + (("gold") if bBool else "not-gold")
        if self.__model.conf_isDry():
            sTitle += " in dry mode"
        mb = QMessageBox.warning(self, " ", "<h3>" + sTitle + "</h3>Set status for selected " + str(
            len(lstIds)) + " records to: " + ("gold" if bBool else "not-gold") + "?", QMessageBox.Yes | QMessageBox.No)
        if mb == QMessageBox.No:
            return True

        # Start Worker
        self.worker_start(Worker_GoldEdit(self.__model, sTab, lstIds, bBool, **kwargs), self.worker_finished_goldEdit)

    def table_entryDel(self, sTab, lstIds, **kwargs):
        """
        Lösche einen Eintrag in der Tabelle

        Args:
            sTab (str): Tabellenname
            lstIds (list): Ids der Einträge
        Kwargs:
            closeWin (PyQt5.QtWidgets.QWidgets): Fenster, welches nach dem Löschvorgang geschlossen werden kann

        Returns:
            (bool): Vorgang war erfolgreich?

        """

        # Nur ein Ingest darf benutzt werden
        if len(lstIds) == 0:
            return

        # Wirklich?
        sTitle = "Delete records"
        mb = QMessageBox.warning(self, " ","<h3>" + sTitle + "</h3>Delete the selected " + str(len(lstIds)) + " records?",QMessageBox.Yes | QMessageBox.No)
        if mb == QMessageBox.No:
            return True

        # Start Worker
        self.worker_start(Worker_DelEntry(self.__model, sTab, lstIds, **kwargs), self.worker_finished_delEntry)


    def actionlog_cmd(self, sCmd):
        """
        Ein Action Log Kommandobutton wurde gedrückt

        Args:
            sCmd (str): Kommando
        """

        if self.sys_isWorkerRunning(popup=True):
            return

        if not self.__model.subsystem_getConnector().isReady():
            self.popup_offline()
            return
        if sCmd == "show":
            self.actionlog_show(self.__lstSelectedActionLogs)
        if sCmd == "win":
            self.actionlog_showInOwnWindow(self.__lstSelectedActionLogs)
        elif sCmd == "del":
            self.actionlog_del(self.__lstSelectedActionLogs)
        elif sCmd == "del-all":
            self.actionlog_delAll()
        elif sCmd == "filter":
            self.status_refresh_actionLogTable()

    def import_ingest_cmd(self, sCmd):
        """
        Ein Ingest-Kommandobutton wurde gedrückt

        Args:
            sCmd (str): Kommando
        """

        if self.sys_isWorkerRunning(popup=True):
            return

        if not self.__model.subsystem_getConnector().isReady():
            self.popup_offline()
            return
        if sCmd == "map":
            self.ingesttable_showOnMap(self.__lstSelectedIngests)
        elif sCmd == "del":
            self.ingesttable_del(self.__lstSelectedIngests)
        elif sCmd == "log":
            self.ingesttable_showLog(self.__lstSelectedIngests)
        elif sCmd == "excel":
            self.ingesttable_save(self.__lstSelectedIngests)
        elif sCmd == "lookup":
            self.ingesttable_lookupSave(self.__lstSelectedIngests)
        elif sCmd == "tab":
            self.ingesttable_showTable(self.__lstSelectedIngests)
        elif sCmd == "dep-show":
            self.ingesttable_showDep(self.__lstSelectedIngests)
        elif sCmd == "dep-hide":
            self.ingesttable_hideDep()

    def ingesttable_showDep(self, lstIngestIds):
        """
        Zeige Abhängigkeiten zu Ingestion

        Args:
            iIngestIds (lst): Ids des Ingests

        """

        # Nur ein Ingest darf benutzt werden
        if len(lstIngestIds) != 1:
            return
        else:
            iIngestId = lstIngestIds[0]

        self.status_refresh_ingestion_overview(iIngestId)

    def ingesttable_hideDep(self):
        """
        Zeige keine Ingest-Abhängigkeiten mehr
        """
        self.status_clean_ingestDepenencies()
        self.status_refresh_ingestion_overview()

    def ingesttable_lookupSave(self, lstIngestIds):
        """
        Lookup-Tableen speichern

        Args:
            lstIngestId (list): Ids
        """

        if len(lstIngestIds) == 0:
            return

        pn = self.ui_printConsoleNormal
        pv = self.ui_printConsoleVerbose
        pd = self.ui_printConsoleDebug

        # Verbunden mit Server?
        if self.__model.subsystem_getConnector().isReady():
            pd("Save lookup tables for ingests with Ids: " + PrettyText.lst2strC(lstIngestIds))

            # Hole Zielordner
            sFileDefault = str(Path.home().joinpath("lookup.xlsx"))
            sFile = QFileDialog.getSaveFileName(self, "Save lookup tables to file", directory=sFileDefault, filter="*.xlsx;; *.*")[0]

            # Start Worker
            if sFile not in [None, ""]:
                # Starte Worker
                self.worker_start(Worker_ExportLookup(self.__model, sFile, lstIngestIds, None), self.worker_finished_exportExcel)

    def ingesttable_save(self, lstIngestIds):
        """
        Ingestion speichern

        Args:
            lstIngestId (lst): Ids
        """

        if len(lstIngestIds) == 0:
            return
        iId = lstIngestIds[0]

        # Selector
        dicSelector = {"ingest": [iId]}

        pn = self.ui_printConsoleNormal
        pv = self.ui_printConsoleVerbose
        pd = self.ui_printConsoleDebug

        # Verbunden mit Server?
        if self.__model.subsystem_getConnector().isReady():
            pd("Save ingest with Id: " + str(iId))

            # Hole Dateinamen aus DB
            resFilename = self.__model.subsystem_getDb().ingest_getIngestFilenames(iId)
            if resFilename.isErr():
                pn(resFilename.getResult())
                return
            sFilename = resFilename.getData("file")

            # Hole Zielordner
            sFileDefault = str(Path.home().joinpath(sFilename))
            sFile = QFileDialog.getSaveFileName(self, "Export ingest to file", directory=sFileDefault, filter="*.xlsx;; *.*")[0]

            # Start Worker
            if sFile not in [None, ""]:
                self.worker_start(Worker_ExportIngest(self.__model, sFile, dicSelector), self.worker_finished_exportExcel)

    def ingesttable_showTable(self, lstIngestIds):
        """
        Ingestion anzeigen

        Args:
            lstIngestId (lst): Ids

        """

        # Nur ein Ingest darf benutzt werden
        if len(lstIngestIds) != 1:
            return
        else:
            iIngestId = lstIngestIds[0]

        # Start Worker
        self.worker_start(Worker_ShowIngest(self.__model, iIngestId), self.worker_finished_showTable)

    def ingesttable_showOnMap(self, lstIngestIds):
        """
        Ingestion plotten

        Args:
            lstIngestIds (lst): Ids
        """

        pn = self.ui_printConsoleNormal
        pv = self.ui_printConsoleVerbose
        pd = self.ui_printConsoleDebug

        if len(lstIngestIds) == 0:
            return

        dicSwitches = {str(iId): True for iId in lstIngestIds}

        if self.sys_isWorkerRunning(popup=True):
            return

        if not self.__model.subsystem_getConnector().isReady():
            self.popup_offline()
            return

        self.tab_switchToViz()
        self.explorer_reset_visual()
        self.__model.ui_selector_getSelector().reset()
        self.__model.ui_selector_getSelector().deactivateAllEntries("ingest")
        dicInput = {"ingest": dicSwitches}
        self.__model.ui_selector_getSelector().update(dicInput)
        self.explorer_setSelector()
        self.explorer_update()

    def ingesttable_showLog(self, lstIngestIds):
        """
        Log zu Ingestion anzeigen

        Args:
            lstIngestIds (list): Ids
        """

        # Nur ein Ingest darf benutzt werden
        if len(lstIngestIds) != 1:
            return
        else:
            iIngestId = lstIngestIds[0]

        pn = self.ui_printConsoleNormal
        pv = self.ui_printConsoleVerbose
        pd = self.ui_printConsoleDebug

        # Start Worker
        self.worker_start(Worker_ShowIngestLog(self.__model, iIngestId), self.worker_finished_showText)

    def admin_lookup_purge(self, **kwargs):
        """
        Purge Lookups
        """
        sLookup = kwargs["lookup"] if "lookup" in kwargs else None
        bFull = kwargs["full"] if "full" in kwargs and kwargs["full"] in [True, False] else False

        if self.sys_isWorkerRunning(popup=True):
            return
        if not self.__model.subsystem_getConnector().isReady():
            self.popup_offline()
            return

        sSheet = self.__model.subsystem_getMapper().lookup_table2sheet(sLookup)

        if bFull:
            sInfo_full = ""
        else:
            sInfo_full = "and not-gold "
        if sSheet is None:
           sInfo_lup = "all lookup tables"
        else:
           sInfo_lup = "lookup table " +sSheet
        sInfo = "Purge all not-used " + sInfo_full +"entires in " + sInfo_lup +"?"

        # Wirklich?
        sTitle = "Purge"
        sHead = "Purge entries"
        sInfo = "Do you really want to purge all unused entries?"
        ant = self.popup_generic(title=sTitle, head=sHead, info=sInfo, button_yes="Yes", button_no="No")
        if ant == 0:
            return True

        # Start Worker
        self.worker_start(Worker_PurgeLookup(self.__model, sLookup, bFull), self.worker_finished_purgeLookups)

    def ingesttable_del(self, lstIngestIds):
        """
        Lösche Ingestion

        Args:
            lstIngestIds (lst): Ids
        """

        # Nur ein Ingest darf benutzt werden
        if len(lstIngestIds) == 0:
            return

        # Wirklich?
        ant = self.popup_generic(title="Delete", head="Delete Ingestion(s)", info="Delete selected "+str(len(lstIngestIds))+ " ingests?", button_yes="Yes", button_no="No")
        if ant == 0:
            return True

        # Start Worker
        self.worker_start(Worker_DelIngest(self.__model, lstIngestIds), self.worker_finished_delIngest)


    def actionlog_del(self, lstActionlogIds):
        """
        Lösche Action Log

        Args:
            lstActionlogIds (lst): Ids
        """
        # Nur ein Ingest darf benutzt werden
        if len(lstActionlogIds) == 0:
            return

        pn = self.ui_printConsoleNormal

        # Wirklich?
        mb = QMessageBox.warning(self," ","<h3>Delete Action Logs</h3>Should the selected action logs ("+str(len(lstActionlogIds))+") really be deleted? ",QMessageBox.Yes | QMessageBox.No)
        if mb == QMessageBox.No:
            return True

        self.ui_showProgress()
        self.ui_setWorkflow(top="Deleting action log", sub="Working")

        # Loeschen
        res = self.__model.subsystem_getDb().actionlog_delSingle(lstActionlogIds)

        self.ui_setWorkflow()
        self.ui_hideProgress()

        # Fehler
        if res.isErr():
            self.popup_error("Error while deleting action log", "Details in console ...", str(res.getResult()))
            return
        # Ergenis anzeigen
        # OK
        if self.__model.conf_isDry():
            sTitle = "Simulation of Action Log Deletion"
            sInfo = "Action logs would have been deleted: "+str(len(lstActionlogIds))
        else:
            sTitle = "Action Logs Deleted"
            sInfo = "Deleted logs: "+str(len(lstActionlogIds))
            pn("Workflows will be reinitialized ...")
            self.sys_server_disconnect()
            self.sys_server_connect()

        self.popup_generic(title="ActionLog", head=sTitle, info=sInfo)

    def actionlog_delAll(self):
        """
        Lösche gesamtes Action Log
        """

        pn = self.ui_printConsoleNormal

        # Wirklich?
        mb = QMessageBox.warning(self," ","<h3>Delete All Action Logs</h3>Should <b>ALL</b> action logs be deleted? ",QMessageBox.Yes | QMessageBox.No)
        if mb == QMessageBox.No:
            return True

        self.ui_showProgress()
        self.ui_setWorkflow(top="Deleting all action log", sub="Working", fmt=False)

        # Loeschen
        res = self.__model.subsystem_getDb().actionlog_delAll()

        self.ui_setWorkflow()
        self.ui_hideProgress()

        # Fehler
        if res.isErr():
            self.popup_error("Error while deleting action log", "Details in console ...", str(res.getResult()))
            return
        # Ergenis anzeigen
        # OK
        if self.__model.conf_isDry():
            sTitle = "Simulation of Action Log Deletion"
            sInfo = "All Action logs would have been deleted: "
        else:
            sTitle = "Delete All Action Logs"
            sInfo = "All Action logs have been deleted: "
            pn("Workflows will be reinitialized ...")
            self.sys_server_disconnect()
            self.sys_server_connect()

        self.popup_generic(title="ActionLog", head=sTitle, info=sInfo)

    def actionlog_show(self, lstActionlogIds):
        """
        Action Log in Texfeld anzeigen

        Args:
            lstActionlogId (lst): Ids

        """

        # Nur ein Ingest darf benutzt werden
        if len(lstActionlogIds) != 1:
            return
        else:
            iActionlogId = lstActionlogIds[0]

        # Start Worker
        self.worker_start(Worker_ShowActionlog(self.__model, iActionlogId),
                          self.worker_finished_showText)

    def actionlog_showInOwnWindow(self, lstActionlogIds):
        """
        Action Log in eigenem Fenster anzeigen

        Args:
            lstActionlogId (lst): Ids
        """

        # Nur ein Ingest darf benutzt werden
        if len(lstActionlogIds) != 1:
            return
        else:
            iActionlogId = lstActionlogIds[0]

        # Start Worker
        self.worker_start(Worker_PopActionLog(self.__model, iActionlogId), self.worker_finished_showText)

class DialogOptionen(QDialog, uic.loadUiType(fileOptionen)[0]):
    """
    Dialog mit Optionen

    """

    def __init__(self, model):
        """
        Args:
            model (gui.model.Model): Model
        """

        # Super
        QDialog.__init__(self)
        self.setupUi(self)
        self.setWindowFlags(Qt.Dialog | Qt.CustomizeWindowHint | Qt.WindowTitleHint | Qt.WindowCloseButtonHint)

        # Instanzvariablen
        self.__model = model

        # Initialisierung
            # DB
        self.buildPreset()
            # Lookup Import
        if self.__model.conf_getLookupImport_excel() == "not-gold":
            self.combo_lup_excel.setCurrentIndex(1)
        elif self.__model.conf_getLookupImport_excel() == "gold":
            self.combo_lup_excel.setCurrentIndex(2)
        else:
            self.combo_lup_excel.setCurrentIndex(0)
        self.check_goldrush.setCheckState(2 if self.__model.conf_getLookupImport_excel_goldrush() == True else 0)
        if self.__model.conf_getLookupImport_db() == "new":
            self.combo_lup_db.setCurrentIndex(0)
        elif self.__model.conf_getLookupImport_db() == "update":
            self.combo_lup_db.setCurrentIndex(1)
        else:
            self.combo_lup_db.setCurrentIndex(2)
        if self.__model.conf_getLookupImport_db_update() == "not-gold":
            self.combo_lup_db_update.setCurrentIndex(0)
        elif self.__model.conf_getLookupImport_db_update() == "gold":
            self.combo_lup_db_update.setCurrentIndex(1)
        else:
            self.combo_lup_db_update.setCurrentIndex(2)
        if self.__model.conf_getLookupReuseDiffersReactionlevel() == "warning":
            self.combo_record_exsiting_using_differs.setCurrentIndex(1)
        elif self.__model.conf_getLookupReuseDiffersReactionlevel() == "info":
            self.combo_record_exsiting_using_differs.setCurrentIndex(2)
        else:
            self.combo_record_exsiting_using_differs.setCurrentIndex(0)
            # Record Extension
        self.check_record_extension.setCheckState(2 if self.__model.conf_isRecordExtensionAllowed() == True else 0)
        if self.__model.conf_getRecordExtensionProblemReactionlevel() == "warning":
            self.combo_record_extension_problem.setCurrentIndex(1)
        elif self.__model.conf_getRecordExtensionProblemReactionlevel() == "info":
            self.combo_record_extension_problem.setCurrentIndex(2)
        else:
            self.combo_record_extension_problem.setCurrentIndex(0)
            # Geschwaetzigkeit
        self.combo_verbosity.setCurrentIndex(0)
        if self.__model.conf_isVerbose():
            self.combo_verbosity.setCurrentIndex(1)
        if self.__model.conf_isDebug():
            self.combo_verbosity.setCurrentIndex(2)
            # Logging
        self.check_logging.setCheckState(2 if self.__model.conf_isLogging() else 0)
            # Dry Run
        self.check_dry.setCheckState(2 if self.__model.conf_isDry() == True else 0)
            # Pw Speichern
        self.check_pwSpeichern.setCheckState(2 if self.__model.conf_isPwSave() else 0)
            # Admin
        self.path_psql.setText(self.__model.conf_getAdminPsqlPath())
        self.path_pgdump.setText(self.__model.conf_getAdminPgdumpPath())
            # Working Dir
        self.dir_working.setText(self.__model.conf_getWorkingDir())
        # Buttons verbinden
            # Dialogbuttons
        self.buttonBox.button(QDialogButtonBox.Apply).clicked.connect(lambda: QDialog.done(self, 1))
        self.buttonBox.button(QDialogButtonBox.Save).clicked.connect(lambda: QDialog.done(self, 2))
        self.buttonBox.button(QDialogButtonBox.Cancel).clicked.connect(lambda: QDialog.done(self, 0))
            # Logging
        self.check_logging.clicked.connect(self.actionLoggingAktivieren)
            # DB-Veränderungen
        self.db_presets.setCompleter(None) # Kein Autocomplete
        self.db_presets.currentIndexChanged.connect(self.actionPresetChanged)
        #self.db_presets.editTextChanged.connect(self.actionDbPresetEdited)
        self.db_host.textEdited.connect(self.actionDbHostEdited)
        self.db_port.textEdited.connect(self.actionDbPortEdited)
        self.db_user.textEdited.connect(self.actionDbUserEdited)
        self.db_pw.textEdited.connect(self.actionDbPwEdited)
        self.db_name.textEdited.connect(self.actionDbNameEdited)
        self.db_schema.textEdited.connect(self.actionDbSchemaEdited)
            # Account Anzahl
        self.db_account_amount_plus.clicked.connect(self.actionDbAccountAmountPlus)
        self.db_account_amount_minus.clicked.connect(self.actionDbAccountAmountMinus)
        self.db_account_rename.clicked.connect(self.actionDbAccountRename)

    # --- GUI tools

        # Buttons verbinden
        self.button_path_psql.clicked.connect(self.selectPsql)
        self.button_path_pgdump.clicked.connect(self.selectPgdump)
        self.button_batchbot_logdir.clicked.connect(self.selectWorkingDir)
        self.check_record_extension.clicked.connect(self.selectRecordExtension)
        self.combo_lup_excel.currentIndexChanged.connect(self.__checkImportLookup)
        self.combo_lup_db.currentIndexChanged.connect(self.__checkImportLookup)

        # Defaults
        self.combo_record_extension_problem.setEnabled(self.__model.conf_isRecordExtensionAllowed())
        self.__checkImportLookup()

    def selectRecordExtension(self):
        """
        Auswahl Record Extension
        """
        self.combo_record_extension_problem.setEnabled(self.check_record_extension.isChecked())

    def __checkImportLookup(self):
        if self.combo_lup_db.currentIndex() == 0:
            self.combo_lup_db_update.hide()
            self.label_lup_db_update.hide()
        else:
            self.combo_lup_db_update.show()
            self.label_lup_db_update.show()
            self.combo_lup_db_update.setEnabled(True)
            if self.combo_lup_excel.currentIndex() == 1:
                self.combo_lup_db_update.setCurrentIndex(0)
                self.combo_lup_db_update.setEnabled(False)

    def selectPsql(self):
        """
        Wähle Path zu psql
        """

        sPath = QFileDialog.getOpenFileName(self, "Select program psql", directory=self.__model.conf_getUserHomeDir(), filter="*.*")[0]
        self.path_psql.setText(sPath)
        self.path_psql.setCursorPosition(0)

    def selectPgdump(self):
        """
        Wähle Path zu pg_dump
        """

        sPath = QFileDialog.getOpenFileName(self, "Select program pg_dump", directory=self.__model.conf_getUserHomeDir(), filter="*.*")[0]
        self.path_pgdump.setText(sPath)
        self.path_pgdump.setCursorPosition(0)

    def selectWorkingDir(self):
        sDir = QFileDialog.getExistingDirectory(self, "Select generic working directory", directory=self.__model.conf_getWorkingDir())
        if sDir not in ["",None]:
            self.dir_working.setText(sDir)
            self.dir_working.setCursorPosition(0)

    def buildPreset(self):
        """
        Preset Combo aufbauen
        """

        self.db_presets.blockSignals(True)
        self.db_presets.clear()
        db = self.__model.subsystem_getDb()
        for sAccount in db.account_all_getNames():
            self.db_presets.addItem(sAccount)

        #self.db_account_amount.setText(str(self.getAccountAmount()))
        sActive_name = db.account_active_getName()
        iActive_index = db.account_active_getIndex()
        iActive_index_combo = self.db_presets.findText(sActive_name)
        self.db_presets.setCurrentIndex(iActive_index_combo)
        self.db_presets.blockSignals(False)
        self.db_user.setText(db.account_single_getPart("user"))
        self.db_pw.setText(db.account_single_getPart("pw"))
        self.db_host.setText(db.account_single_getPart("host"))
        self.db_port.setText(db.account_single_getPart("port"))
        self.db_name.setText(db.account_single_getPart("db"))
        self.db_schema.setText(db.account_single_getPart("schema"))

    def popupFrage(self, sTitel, sText):
        """
        Popup: Frage

        Args:
            sTitel (str): Titel
            sText (str): Text

        Returns:
            (bool): Antwort war ja (True) oder nein (False)

        """
        ant = QMessageBox.question(self, sTitel, sText, QMessageBox.Yes, QMessageBox.No)
        if ant == QMessageBox.Yes:
            return True
        else:
            return False

    def setAccountData(self, sName, sWert, iAccountNr = None):
        """
        Setze Werte in Account

        Args:
            sName (str): Name des zu setzenden Wertes
            sWert (str): zu setzender Wert
            iAccountNr=None (int): Index des Accounts (wenn nichts ausgewählt wird, wird der aktuell aktive Account benutzt)

        Returns:
            (bool): Vorgang war erfolgreich?

        """
        db = self.__model.subsystem_getDb()
        if iAccountNr is None:
            iAccountNr = db.account_active_getIndex()
        db.account_single_setPart(sName, str(sWert), account_index=iAccountNr)


    def actionDbAccountAmountPlus(self):
        """
        Neuen leeren Account hinzufuegen
        """
        db = self.__model.subsystem_getDb()
        dicAccount = {
            "account": "new",
            "host": "",
            "port": "",
            "db": "",
            "schema": "",
            "user": "",
            "pw": ""
        }
        sAccount_new = db.account_single_add(dicAccount)
        db.account_active_setName(sAccount_new)
        self.buildPreset()

    # --- GUI-Actionen

    def actionDbAccountRename(self):
        iIndex = self.db_presets.currentIndex()  # Aktueller Preset-Index aus Combobox
        sPreset_old = self.db_presets.currentText()
        sPreset_new = QInputDialog.getText(self, 'Rename', "Account", QLineEdit.Normal, sPreset_old)[0]
        if sPreset_new in [None, ""]:
            return
        db = self.__model.subsystem_getDb()
        db.account_rename(sPreset_old, sPreset_new)
        self.db_presets.blockSignals(True)
        self.buildPreset()
        self.db_presets.blockSignals(False)

    def actionDbAccountAmountMinus(self):
        """
        Account loeschen
        """

        # Letzen Account nie löschen
        db = self.__model.subsystem_getDb()
        if db.account_all_getNr() <= 1:
            return
        # Hole Index aus Combobox
        iIndex = db.account_active_getIndex()
        sAccount = db.account_active_getName()
        bEmpty = db.account_single_isEmpty(account_index=iIndex)
        if bEmpty is False:
            ant = self.popupFrage("Account not empty","The account is not empty. Should it be deleted anyway?")
            if not ant:
                return
        # Loesche Accountdaten
        db.account_single_del(sAccount)
        # Neuen aktiven Index einstellen
        self.buildPreset()

    def actionLoggingAktivieren(self):
        """
        Logging aktivieren
        """

        bLog = self.check_logging.isChecked()

    def actionDbHostEdited(self, sText):
        """
        Verändere DB-Informationen Host

        Args:
            sText (str): neuer Host
        """

        self.setAccountData("host", self.db_host.text(), self.db_presets.currentIndex())

    def actionDbPortEdited(self, sText):
        """
        Verändere DB-Informationen Port

        Args:
            sText (str): neuer Port
        """

        self.setAccountData("port", self.db_port.text(), self.db_presets.currentIndex())

    def actionDbUserEdited(self, sText):
        """
        Verändere DB-Informationen User

        Args:
            sText (str): neuer User
        """

        self.setAccountData("user", self.db_user.text(), self.db_presets.currentIndex())

    def actionDbPwEdited(self, sText):
        """
        Verändere DB-Informationen Passwort

        Args:
            sText (str): neues Passwort
        """

        self.setAccountData("pw", self.db_pw.text(), self.db_presets.currentIndex())

    def actionDbNameEdited(self, sText):
        """
        Verändere DB-Informationen Name

        Args:
            sText (str): neuer Name
        """

        self.setAccountData("db", self.db_name.text(), self.db_presets.currentIndex())

    def actionDbSchemaEdited(self, sText):
        """
        Verändere DB-Informationen Schema

        Args:
            sText (str): neues Schema
        """

        self.setAccountData("schema", self.db_schema.text(), self.db_presets.currentIndex())

    def actionPresetChanged(self):
        """
        Neues Preset wurde ausgewählt
        """

        iIndex = self.db_presets.currentIndex() # Index aus Combobox
        db = self.__model.subsystem_getDb()
        db.account_active_setIndex(iIndex)
        self.db_user.setText(db.account_single_getPart("user"))
        self.db_pw.setText(db.account_single_getPart("pw"))
        self.db_host.setText(db.account_single_getPart("host"))
        self.db_port.setText(db.account_single_getPart("port"))
        self.db_name.setText(db.account_single_getPart("db"))
        self.db_schema.setText(db.account_single_getPart("schema"))

        # Kleiner Tweak -> Nachdem umgeschaltet wurde wird die Combobox geupdatet um evtl. Änderungen umzusetzen
        # Ich kann das leider nicht direkt bei der Preseteingabe machen, weil QT das nicht hinbekommt
        self.buildPreset()

class DialogCsvExtractNewData(QDialog, uic.loadUiType(fileCsvExtractNewData)[0]):
    """
    Klasse für Dialog: Csv Extract New Data
    """

    def __init__(self):
        # Super
        QDialog.__init__(self)
        # Initialisierung
        self.setupUi(self)
        self.setFixedHeight(self.sizeHint().height())
        self.setWindowModality(Qt.ApplicationModal)
        self.setWindowFlags(Qt.Dialog | Qt.CustomizeWindowHint | Qt.WindowTitleHint | Qt.WindowCloseButtonHint)

        # Buttons verbinden
        self.button_in1.clicked.connect(self.selectIn1)
        self.button_in2.clicked.connect(self.selectIn2)
        self.button_out.clicked.connect(self.selectOut)
        self.text_in1.textChanged.connect(self.__genHeaderCombo)
        self.text_in2.textChanged.connect(self.__genHeaderCombo)

    def selectIn1(self):
        """
        Wähle CSV-Datei ein1 aus
        """

        sIn1 = QFileDialog.getOpenFileName(self, "Select file 1", filter="*.csv;; *.*")[0]
        self.text_in1.setText(sIn1)
        self.text_in1.setCursorPosition(0)

    def selectIn2(self):
        """
        Wähle CSV-Datei ein2 aus
        """

        sIn2 = QFileDialog.getOpenFileName(self, "Select file 2", filter="*.csv;; *.*")[0]
        self.text_in2.setText(sIn2)
        self.text_in2.setCursorPosition(0)

    def selectOut(self):
        """
        Wähle CSV-Datei aus aus
        """

        sOut = QFileDialog.getSaveFileName(self, "Select file 3", filter="*.csv;; *.*")[0]
        self.text_out.setText(sOut)
        self.text_out.setCursorPosition(0)

    def __genHeaderCombo(self):
        """
        Erzeuge Header-Combobox aus vorhandenen Headern in Datei 1 und 2
        """

        self.combo_header.clear()

        # Hole Dateinamen
        sIn1 = self.text_in1.text()
        sIn2 = self.text_in2.text()

        # Teste Dateien
        if sIn1 is None or sIn1 == "" or not os.path.isfile(sIn1):
            return
        if sIn2 is None or sIn2 == "" or not os.path.isfile(sIn2):
            return

        # Hole Header
        lstIn1Header = list()
        lstIn2Header = list()
        try:
            with open(sIn1, newline='') as csvfile:
                csvReader = csv.reader(csvfile, delimiter=',', quotechar='"')
                lstIn1Header = csvReader.__next__()
            with open(sIn2, newline='') as csvfile:
                csvReader = csv.reader(csvfile, delimiter=',', quotechar='"')
                lstIn2Header = csvReader.__next__()
        except Exception as ex:
            return

        # Baue Combobox auf
        for head in lstIn1Header:
            if head in lstIn2Header:
                self.combo_header.addItem(head)

# Dialog: BatchBot
class DialogBatchBot(QDialog, uic.loadUiType(fileBatch)[0]):

    # Konstruktor
    def __init__(self, model):
        # Super
        QDialog.__init__(self)

        # Initialisierung
        self.setupUi(self)
        self.setWindowModality(Qt.ApplicationModal)
        self.setWindowFlags(Qt.Dialog | Qt.CustomizeWindowHint | Qt.WindowTitleHint | Qt.WindowCloseButtonHint)

        # Member
        self.__model = model
        self.__batchbot = self.__model.subsystem_getBatchBot()

        # Default: BatchLog Dir
        self.logDir.setText(str(Path(self.__model.conf_getWorkingDir(), "batchbot")))

        # Button anpassen
        self.buttonBox.button(QDialogButtonBox.Ok).setText("Start")
        self.buttonBox.button(QDialogButtonBox.Ok).clicked.connect(lambda: QDialog.done(self, 1))
        self.buttonBox.button(QDialogButtonBox.Save).setText("Export")
        self.buttonBox.button(QDialogButtonBox.Save).clicked.connect(lambda: QDialog.done(self, 3))
        self.buttonBox.button(QDialogButtonBox.Apply).clicked.connect(lambda: QDialog.done(self, 2))

        # Buttons verbinden
        self.comboBox.currentIndexChanged.connect(self.__config_changed)
        self.textBrowser.textChanged.connect(self.__content_edited)
        self.buttonOpen.clicked.connect(self.__logdir_select)
        self.addConfig.clicked.connect(self.__config_new)
        self.delConfig.clicked.connect(self.__config_del)
        self.renameConfig.clicked.connect(self.__config_rename)

        # Kein Autocomplete
        #self.comboBox.setCompleter(None)
        self.comboBox.setEditable(False)

        # Config laden
        self.__config_reload()
        self.__config_select(special="first")

    def __signalsBlock(self):
        self.comboBox.blockSignals(True)
        self.textBrowser.blockSignals(True)

    def __signalsFree(self):
        self.comboBox.blockSignals(False)
        self.textBrowser.blockSignals(False)

    def __config_reload(self):
        lstConfigNames = self.__batchbot.getConfig_all()
        if lstConfigNames is None:
            lstConfigNames = []
        self.__signalsBlock()
        self.comboBox.clear()
        self.comboBox.addItems(lstConfigNames)
        self.__signalsFree()

    def __content_edited(self):
        self.__signalsBlock()
        sConfig = self.comboBox.currentText()
        sContent = self.textBrowser.toPlainText()
        self.__batchbot.setContent(sConfig, sContent)
        self.__signalsFree()

    def __config_select(self, **kwargs):
        sName = kwargs["name"] if "name" in kwargs else None
        sSpecial = kwargs["special"] if "special" in kwargs else None

        # Hole Daten
        lstConfigs = self.__batchbot.getConfig_all()
        if lstConfigs is None or len(lstConfigs) == 0:
            self.__signalsBlock()
            self.textBrowser.clear()
            self.__signalsFree()
            return
        if sName is None:
            if sSpecial == "first":
                sName = self.__batchbot.getConfig_first()
            elif sSpecial == "last":
                sName = self.__batchbot.getConfig_last()
        # Test
        if sName is None or sName not in lstConfigs:
            return
        # Stelle Combobox ein
        self.__signalsBlock()
        self.textBrowser.setPlainText(self.__batchbot.getContent(sName))
        self.comboBox.setCurrentIndex(self.comboBox.findText(sName))
        self.__signalsFree()

    def __config_rename(self):
        sConfig_old = self.comboBox.currentText()
        sConfig_new = QInputDialog.getText(self, 'Rename', "Configuration", QLineEdit.Normal, sConfig_old)[0]
        if sConfig_new in [None, ""]:
            return
        sConfig_new_real = self.__batchbot.renameConfig(sConfig_old, sConfig_new)
        if sConfig_new_real is not None:
            self.__config_reload()
            self.__config_select(name=sConfig_new_real)

    def __config_changed(self):
        self.__config_select(name=self.comboBox.currentText())

    def __config_new(self):
        sName = self.__batchbot.newConfig()
        self.__config_reload()
        self.__config_select(name=sName)
        #print(self.__batchbot.getContent("Alles"))

    def __config_del(self):
        sName = self.comboBox.currentText()
        self.__batchbot.delConfig(sName)
        self.__config_reload()
        self.__config_select(special="first")

    def __logdir_select(self):
        sDir = QFileDialog.getExistingDirectory(self, "Select directory to log")
        self.logDir.setText(sDir)

# Dialog: CSV Versionskontrolle
class DialogDiff(QDialog, uic.loadUiType(fileDiff)[0]):

    # Konstruktor
    def __init__(self):
        # Super
        QDialog.__init__(self)

        # Initialisierung
        self.setupUi(self)
        self.setFixedHeight(self.sizeHint().height())
        self.setWindowModality(Qt.ApplicationModal)
        self.setWindowFlags(Qt.Dialog | Qt.CustomizeWindowHint | Qt.WindowTitleHint | Qt.WindowCloseButtonHint)

        # Buttons verbinden
        self.button_in1.clicked.connect(self.__selectIn1)
        self.button_in2.clicked.connect(self.__selectIn2)
        self.button_out.clicked.connect(self.__selectOut)

    # In1
    def __selectIn1(self):
        sIn1 = QFileDialog.getOpenFileName(self, "Select file of the original version", filter="*.csv;; *.*")[0]
        self.text_in1.setText(sIn1)
        self.text_in1.setCursorPosition(0)

    # In2
    def __selectIn2(self):
        sIn2 = QFileDialog.getOpenFileName(self, "Select file of the edited version", filter="*.csv;; *.*")[0]
        self.text_in2.setText(sIn2)
        self.text_in2.setCursorPosition(0)

    # Out
    def __selectOut(self):
        sOut = QFileDialog.getSaveFileName(self, "Select file to write the result", filter="*.csv;; *.*")[0]
        self.text_out.setText(sOut)
        self.text_out.setCursorPosition(0)

# Dialog: CSV Versionskontrolle
class DialogFaoCreate(QDialog, uic.loadUiType(fileFaoCreate)[0]):

    # Konstruktor
    def __init__(self):
        # Super
        QDialog.__init__(self)

        # Initialisierung
        self.setupUi(self)
        self.setFixedHeight(self.sizeHint().height())
        self.setWindowModality(Qt.ApplicationModal)
        self.setWindowFlags(Qt.Dialog | Qt.CustomizeWindowHint | Qt.WindowTitleHint | Qt.WindowCloseButtonHint)

# Dialog: Update
class DialogUpdate(QDialog, uic.loadUiType(fileUpdate)[0]):

    # Konstruktor
    def __init__(self):
        # Super
        QDialog.__init__(self)

        # Initialisierung
        self.setupUi(self)
        self.setFixedHeight(self.sizeHint().height())
        self.setWindowModality(Qt.ApplicationModal)
        self.setWindowFlags(Qt.Dialog | Qt.CustomizeWindowHint | Qt.WindowTitleHint | Qt.WindowCloseButtonHint)

# Dialog: Ingest
class DialogIngest(QDialog, uic.loadUiType(fileIngest)[0]):

    # Konstruktor
    def __init__(self, model, **kwargs):
        """

        Args:
            (model): Model
        Kwargs:
            export (bool): Exportmodus
            ingest (str): Workflowkürzel - Importer für Ingest Daten
            lookup (bool): Importer für Lookup Daten
            lookup_tabs (lst): Liste aller zu importierenden Lookup Tabellen
            taxon (bool): Importer für Taxon Daten
            taxon_tabs (lst): Liste aller zu importierenden Taxon Tabellen
        """

        # Super
        QDialog.__init__(self)

        # Instanzvariablen
        self.__model = model

        # Initialisierung
        self.setupUi(self)
        self.setWindowFlags(Qt.Dialog | Qt.CustomizeWindowHint | Qt.WindowTitleHint | Qt.WindowCloseButtonHint)

        # Lookup
        self.__bLookup = kwargs["lookup"] if "lookup" in kwargs else False
        if "lookup_tabs" in kwargs:
            lstTmp = kwargs["lookup_tabs"]
        else:
            lstTmp = list()
        if lstTmp is None:
            lstTmp = list()
        if type(lstTmp) is str:
            lstTmp = [lstTmp]
        self.__lstLookupOnly = []
        for sTab in lstTmp:
            if sTab in self.__model.subsystem_getMapper().lookup_getDbTables():
                self.__lstLookupOnly.append(self.__model.subsystem_getMapper().lookup_table2sheet(sTab))

        # Taxon
        self.__bTaxon = kwargs["taxon"] if "taxon" in kwargs else False
        if "taxon_tabs" in kwargs:
            lstTmp = kwargs["taxon_tabs"]
        else:
            lstTmp = list()
        if lstTmp is None:
            lstTmp = list()
        if type(lstTmp) is str:
            lstTmp = [lstTmp]
        self.__lstTaxonOnly = []
        for sTab in lstTmp:
            if sTab in self.__model.subsystem_getMapper().taxon_getDbTables():
                self.__lstTaxonOnly.append(self.__model.subsystem_getMapper().taxon_table2sheet(sTab))

        # Ingest
        self.__sIngestWf = kwargs["ingest"] if "ingest" in kwargs else None
        self.__bIngest = False
        self.__lstIngestWfSheets = None
        if self.__sIngestWf is not None:
            self.__bIngest = True
            if self.__model.subsystem_getMapper().workflow_converter_import_isExisting(self.__sIngestWf):
                # Converter
                sWfTarget = self.__model.subsystem_getMapper().workflow_converter_import_getTarget(self.__sIngestWf)
                sWfTargetLong = self.__model.subsystem_getMapper().workflow_short2long(sWfTarget)
                self.__lstIngestWfSheets = self.__model.subsystem_getMapper().ingest_getExcelSheets(sWfTarget)
                self.groupIngest.setTitle("Converting to: "+sWfTargetLong)
                self.labelFoundTablesIngest.setText("Converted tables")
            else:
                # Normal
                self.__lstIngestWfSheets = self.__model.subsystem_getMapper().ingest_getExcelSheets(self.__sIngestWf)

        # Daten die nicht beachtet wurden
        self.__lstUnusedExcelSheets = []

        # Voreinstellungen: Show/Hide
        self.groupIngest.hide()
        self.groupLookup.hide()
        self.groupTaxon.hide()
        self.groupStatus.hide()

        # Lookup: Import
        sMode = self.__model.conf_getLookupImport_excel()
        if sMode == "not-gold":
            self.excel_gold.setCurrentIndex(1)
        elif sMode == "gold":
            self.excel_gold.setCurrentIndex(2)
        else:
            self.excel_gold.setCurrentIndex(0)
        sMode = self.__model.conf_getLookupImport_db()
        if sMode == "new":
            self.mode.setCurrentIndex(0)
        elif sMode == "update":
            self.mode.setCurrentIndex(1)
        else:
            self.mode.setCurrentIndex(2)
        sMode = self.__model.conf_getLookupImport_db_update()
        if sMode == "not-gold":
            self.db_gold.setCurrentIndex(0)
        elif sMode == "gold":
            self.db_gold.setCurrentIndex(1)
        else:
            self.db_gold.setCurrentIndex(2)
        sMode = self.__model.conf_getLookupReuseDiffersReactionlevel()
        if sMode == "error":
            self.comboUseExistingRecDiffers.setCurrentIndex(0)
        elif sMode == "warning":
            self.comboUseExistingRecDiffers.setCurrentIndex(1)
        else:
            self.comboUseExistingRecDiffers.setCurrentIndex(2)

        # Ingest: Extension
        if self.__model.conf_isRecordExtensionAllowed():
            self.checkRecExt.setChecked(True)
        sMode = self.__model.conf_getRecordExtensionProblemReactionlevel()
        if sMode == "error":
            self.comboRecExtReact.setCurrentIndex(0)
        elif sMode == "warning":
            self.comboRecExtReact.setCurrentIndex(1)
        else:
            self.comboRecExtReact.setCurrentIndex(2)

        # Mode: Ingest/Lookup/Wizard
        if self.__bIngest:
            sTitle = "Import Data"
            if self.__sIngestWf is not None:
                if self.__model.subsystem_getMapper().workflow_converter_import_isExisting(self.__sIngestWf):
                    # Converter
                    sTitle += ": " + str(self.__model.subsystem_getMapper().workflow_converter_import_short2long(self.__sIngestWf))
                else:
                    # Normal
                    sTitle += ": "+str(self.__model.subsystem_getMapper().workflow_short2long(self.__sIngestWf))
            self.setWindowTitle(sTitle)
            self.groupIngest.show()
            self.groupLookup.show()
            self.groupStatus.show()

        elif self.__bLookup:
            self.setWindowTitle("Import Lookup Data")
            self.groupLookup.show()
            self.groupStatus.show()
        elif self.__bTaxon:
            self.setWindowTitle("Import TaxonWizard Data")
            self.groupTaxon.show()
            self.groupStatus.show()

        # Fenster
        self.setFixedHeight(self.sizeHint().height())
        self.setFixedWidth(self.sizeHint().width())
        self.setWindowModality(Qt.ApplicationModal)

        # Combos
        self.__dicCheckboxIngest = {"Cruise": self.cruise,
                     "Station": self.station,
                     "Sample": self.sample,
                     "SubSample": self.subsample,
                     "Subset": self.subset,
                     "Biota": self.biota,
                     "Sediment": self.sediment,
                     "IndividualMeasure": self.indmeasure,
                                    }
        self.__dicCheckboxLookup = {"Gear": self.gear,
                                  "CRS": self.crs,
                                  "Person": self.person,
                                  "Lifestage": self.lifestage,
                                  "TaxCover": self.taxcover,
                                  "Ship": self.ship,
                                    "PosSys": self.possys,
                                    "Dataset": self.dataset,
                                    "Country": self.country}
        self.__dicCheckboxTaxon = {"TaxBase": self.base,
                                   "TaxPrivate": self.worms_private,
                                   "TaxAaid": self.aaid,
                                   "TaxColony": self.colony,
                                   "TaxOut": self.noworms,
                                   "TaxChange": self.change}

        # Buttons verbinden
        self.open.clicked.connect(self.__open)
        self.select_all_lookup.clicked.connect(self.__selectAllLookup)
        self.deselect_all_lookup.clicked.connect(self.__deselectAllLookup)
        self.select_all_ingest.clicked.connect(self.__selectAllIngest)
        self.deselect_all_ingest.clicked.connect(self.__deselectAllIngest)
        self.select_all_taxon.clicked.connect(self.__selectAllTaxon)
        self.deselect_all_taxon.clicked.connect(self.__deselectAllTaxon)
        self.excel_gold.currentIndexChanged.connect(self.__checkGoldInDbMode)
        self.checkRecExt.clicked.connect(self.__checkExtend)
        self.comboUseExistingRecDiffers.currentIndexChanged.connect(self.__checkGoldInDbMode)
        self.mode.currentIndexChanged.connect(self.__checkGoldInDbMode)
        self.db_gold.currentIndexChanged.connect(self.__checkGoldInDbMode)
        for sName, checkbox in self.__dicCheckboxIngest.items():
            checkbox.stateChanged.connect(self.__check)
        for sName, checkbox in self.__dicCheckboxLookup.items():
            checkbox.stateChanged.connect(self.__check)
        for sName, checkbox in self.__dicCheckboxTaxon.items():
            checkbox.stateChanged.connect(self.__check)
        # Buttonbox
        self.buttonBox.button(QDialogButtonBox.Ok).hide()

        # Check Konsistenz
        self.__checkExtend()
        self.__checkGoldInDbMode()

        # Activate: Sheets
        self.__activateInput()

    # Lookup: Select all
    def __selectAllLookup(self):
        for k, v in self.__dicCheckboxLookup.items():
            #if v.isEnabled():
            v.setChecked(True)

    # Lookup:Deselect all
    def __deselectAllLookup(self):
        for k, v in self.__dicCheckboxLookup.items():
            #if v.isEnabled():
            v.setChecked(False)

    # Ingest: Select all
    def __selectAllIngest(self):
        for k, v in self.__dicCheckboxIngest.items():
            #if v.isEnabled():
            v.setChecked(True)

    # Ingest: Deselect all
    def __deselectAllIngest(self):
        for k, v in self.__dicCheckboxIngest.items():
            #if v.isEnabled():
            v.setChecked(False)

    # Taxon: Select all
    def __selectAllTaxon(self):
        for k, v in self.__dicCheckboxTaxon.items():
            #if v.isEnabled():
            v.setChecked(True)

    # Taxon: Deselect all
    def __deselectAllTaxon(self):
        for k, v in self.__dicCheckboxTaxon.items():
            #if v.isEnabled():
            v.setChecked(False)

    # Aktiviere die Tabellen initial
    def __activateInput(self, **kwargs):
        lstCheckLookuo = kwargs["lookup"] if "lookup" in kwargs else list()
        lstCheckIngest = kwargs["ingest"] if "ingest" in kwargs else list()
        lstCheckTaxon = kwargs["taxon"] if "taxon" in kwargs else list()

        # Signale blockieren
        for sName, checkBox in self.__dicCheckboxIngest.items():
            checkBox.blockSignals(True)
        for sName, checkBox in self.__dicCheckboxLookup.items():
            checkBox.blockSignals(True)
        for sName, checkBox in self.__dicCheckboxTaxon.items():
            checkBox.blockSignals(True)

        # Check Lookup
        for k, v in self.__dicCheckboxLookup.items():
            if k in lstCheckLookuo:
                v.setEnabled(True)
                if (len(self.__lstLookupOnly) > 0 and k in self.__lstLookupOnly) or len(self.__lstLookupOnly) == 0:
                    v.setChecked(True)
                else:
                    v.setChecked(False)
            else:
                v.setEnabled(False)
                v.setChecked(False)

        # Check Ingest
        for k, v in self.__dicCheckboxIngest.items():
            if not self.__bIngest or k not in self.__lstIngestWfSheets:
                v.hide()
            else:
                if k in lstCheckIngest:
                    v.setEnabled(True)
                    v.setChecked(True)
                else:
                    v.setEnabled(False)
                    v.setChecked(False)

        # Check Taxon
        for k, v in self.__dicCheckboxTaxon.items():
            if k in lstCheckTaxon:
                v.setEnabled(True)
                if (len(self.__lstTaxonOnly) > 0 and k in self.__lstTaxonOnly) or len(self.__lstTaxonOnly) == 0:
                    v.setChecked(True)
                else:
                    v.setChecked(False)
            else:
                v.setEnabled(False)
                v.setChecked(False)

        # Signale reaktivieren
        for sName, checkBox in self.__dicCheckboxIngest.items():
            checkBox.blockSignals(False)
        for sName, checkBox in self.__dicCheckboxLookup.items():
            checkBox.blockSignals(False)
        for sName, checkBox in self.__dicCheckboxTaxon.items():
            checkBox.blockSignals(False)

    # Check Eingaben
    def __check(self):

        # Eingabefelder testen
        bFile = False
        bInput = False
        bDep = False

        if self.file.text() != "":
            bFile = True

        if self.__bIngest:
            for k, v in self.__dicCheckboxIngest.items():
                if v.isChecked():
                    bInput = True
                    break
        if self.__bLookup:
            for k, v in self.__dicCheckboxLookup.items():
                if v.isChecked():
                    bInput = True
                    break
        if self.__bTaxon:
            for k, v in self.__dicCheckboxTaxon.items():
                if v.isChecked():
                    bInput = True
                    break

        # Alles Ok?
        sErr = ""
        if not bFile:
            # File
            sErr += "Missing input file ...\n"
        elif not bInput:
            # Auswahl
            sErr += "No sheet selected ...\n"
        else:
            # Abhängigkeiten
            bDep = True
            if self.__bIngest:
                if not self.__dicCheckboxIngest["Subset"].isChecked() and self.__dicCheckboxIngest["Biota"].isChecked():
                        bDep = False
                        sErr += "Biota needs Subset\n"

                if not self.__dicCheckboxIngest["Subset"].isChecked() and self.__dicCheckboxIngest["Sediment"].isChecked():
                    bDep = False
                    sErr += "Sediment needs Subset\n"

                if self.__dicCheckboxIngest["IndividualMeasure"].isChecked() and not self.__dicCheckboxIngest[
                    "Biota"].isChecked():
                    bDep = False
                    sErr += "IndividualMeasure needs Biota\n"
                if len(sErr) > 0:
                    sErr = sErr[:-1]

        # Alles Ok?
        bOk = bFile and bInput and bDep
        if bOk:
            #self.consoleStatusLabel.setText("Status")
            self.consoleStatus.setPlainText("Everything is ok ...")
            self.buttonBox.button(QDialogButtonBox.Ok).show()
        else:
            #self.consoleStatusLabel.setText("Errors occurred")
            self.consoleStatus.setPlainText(sErr)
            self.buttonBox.button(QDialogButtonBox.Ok).hide()

    # Check Lookup Update Modalitäten
    def __checkExtend(self, **kwargs):
        if self.checkRecExt.isChecked():
            self.comboRecExtReact.setEnabled(True)
        else:
            self.comboRecExtReact.setEnabled(False)

    # Check Ingest Extension Modalitäten
    def __checkGoldInDbMode(self, **kwargs):

        # Warning
        if self.comboUseExistingRecDiffers.currentIndex() != 0:
            self.consoleStatus.setPlainText("Inconsistency Warning!\nYou won't get an error message if lookup record is not 100% matching record in database after import")
        else:
            self.consoleStatus.setPlainText("")

        if self.excel_gold.currentIndex() == 0:
            self.excel_goldrush.show()
        else:
            self.excel_goldrush.hide()

        if self.mode.currentIndex() == 0:
            self.db_gold_label_0.hide()
            self.db_gold_label_1.hide()
            self.db_gold.hide()
        else:
            self.db_gold_label_0.show()
            self.db_gold_label_1.show()
            self.db_gold.show()
            self.db_gold.setEnabled(True)
            if self.excel_gold.currentIndex() == 1:
                self.db_gold.setCurrentIndex(0)
                self.db_gold.setEnabled(False)

    # Open
    def __open(self):
        dataFile = QFileDialog.getOpenFileName(self, "Select xlsx file", directory=self.__model.conf_getUserHomeDir(), filter="*.xlsx")[0]
        self.file.setText(dataFile)
        self.file.setCursorPosition(0)

        # --- Lade File
        lstLookUp = []
        lstIngest = []
        lstTaxon = []
        try:
            warnings.simplefilter("ignore")  # Ignoriere Warnungen
            wb = openpyxl.load_workbook(dataFile, read_only=True)
            warnings.simplefilter("default")  # Warnungen wieder aktivieren

            lstLookupNames = self.__model.subsystem_getMapper().lookup_getExcelSheets()
            lstIngestNames = self.__model.subsystem_getMapper().ingest_getExcelSheets(self.__sIngestWf)
            lstTaxonNames = self.__model.subsystem_getMapper().taxon_getExcelSheets()

            lstExcelNames = wb.sheetnames
            self.__lstUnusedExcelSheets = lstExcelNames.copy()
            for sSheet in lstExcelNames:
                if self.__bLookup:
                    if sSheet in lstLookupNames:
                        lstLookUp.append(sSheet)
                        self.__lstUnusedExcelSheets.remove(sSheet)
                if self.__bIngest or self.__bLookup:
                    if sSheet in lstIngestNames:
                        lstIngest.append(sSheet)
                        self.__lstUnusedExcelSheets.remove(sSheet)
                if self.__bTaxon:
                    if sSheet in lstTaxonNames:
                        lstTaxon.append(sSheet)
                        self.__lstUnusedExcelSheets.remove(sSheet)
            if len(self.__lstUnusedExcelSheets) > 0:
                self.consoleStatus.setPlainText("Found surplus sheets:\n"+PrettyText.lst2strC(self.__lstUnusedExcelSheets))

        except Exception as ex:
            pass
        self.__activateInput(lookup=lstLookUp, ingest=lstIngest, taxon=lstTaxon)
        self.__check()

        # Bei Converter alle Ingest sHeets initial aktivieren
        if self.__bIngest and self.__model.subsystem_getMapper().workflow_converter_import_isExisting(self.__sIngestWf):
            self.__selectAllIngest()


    # Konfigurtion auslesen
    def getConfig(self):

        res = Result(name="Ingest Dialog")
        # Auswahl Lookup
        lstLookup = []
        for k, v in self.__dicCheckboxLookup.items():
            if v.isChecked():
                lstLookup.append(k)

        # Auswahl Ingest
        lstIngest = []
        for k, v in self.__dicCheckboxIngest.items():
            if v.isChecked():
                lstIngest.append(k)

        # Auswahl Taxon
        lstTaxon = []
        for k, v in self.__dicCheckboxTaxon.items():
            if v.isChecked():
                lstTaxon.append(k)

        # Update modalitäten
        bExtend = self.checkRecExt.isChecked()
        iExtend = self.comboRecExtReact.currentIndex()
        lstExtendReact = ["error", "warning", "info"]
        sExtendReact = lstExtendReact[iExtend]
        iLookupReuseReact = self.comboUseExistingRecDiffers.currentIndex()
        lstLookupReuseReact = ["error", "warning", "info"]
        sLookupReuseReact = lstLookupReuseReact[iLookupReuseReact]
        iExcel = self.excel_gold.currentIndex()
        iMode = self.mode.currentIndex()
        iDb = self.db_gold.currentIndex()
        lstExcel = ["all", "not-gold", "gold"]
        lstMode = ["new", "update", "both"]
        lstDb = ["not-gold", "gold", "all"]
        sExcel = lstExcel[iExcel]
        sMode = lstMode[iMode]
        if iMode != 0:
            sDb = lstDb[iDb]
        else:
            sDb = None
        res.addData("file", self.file.text())
        res.addData("extend", bExtend)
        res.addData("extend-react", sExtendReact)
        res.addData("ingest", lstIngest)
        res.addData("lookup", lstLookup)
        res.addData("taxon", lstTaxon)
        res.addData("lup-excel", sExcel)
        res.addData("lup-mode", sMode)
        res.addData("lup-db", sDb)
        res.addData("lup-reuse-react", sLookupReuseReact)
        res.addData("tamer", self.checkTamer.isChecked())
        res.addData("ingest-lup-single", self.checkLoadLookupSingle.isChecked())
        res.addData("taxon-update", self.checkTaxonUpdate.isChecked())
        res.addData("lup-goldrush", self.excel_goldrush.isChecked())
        return res

# Dialog: Export
class DialogExport(QDialog, uic.loadUiType(fileExport)[0]):

    # Konstruktor
    def __init__(self, model, **kwargs):
        """

        Args:
            (model): Model
        Kwargs:
            lookup (bool): Exporter für Lookup Daten
            taxon (bool): Exporter für Taxon Daten
        """

        # Super
        QDialog.__init__(self)

        # Instanzvariablen
        self.__model = model

        # Lookup
        self.__bLookup = kwargs["lookup"] if "lookup" in kwargs else False

        # Taxon
        self.__bTaxon = kwargs["taxon"] if "taxon" in kwargs else False

        # Initialisierung
        self.setupUi(self)
        self.setWindowFlags(Qt.Dialog | Qt.CustomizeWindowHint | Qt.WindowTitleHint | Qt.WindowCloseButtonHint)

        # Voreinstellungen: Show/Hide
        if self.__bLookup:
            self.setWindowTitle("Export Lookup Data")
            self.file.setText(str(Path(self.__model.conf_getUserHomeDir()) / "lookup.xlsx"))
            self.groupLookup.show()
            self.groupTaxon.hide()
            self.labelFile.hide()
        elif self.__bTaxon:
            self.setWindowTitle("Export TaxonWizard Data")
            self.file.setText(str(Path(self.__model.conf_getUserHomeDir()) / "taxon.xlsx"))
            self.groupTaxon.show()
            self.groupLookup.hide()
            self.comboGold.hide()
            self.labelGold1.hide()
            self.labelGold2.hide()

        # Fenster
        self.setFixedHeight(self.sizeHint().height())
        self.setFixedWidth(self.sizeHint().width())
        self.setWindowModality(Qt.ApplicationModal)
        self.buttonBox.button(QDialogButtonBox.Ok).hide()

        # Combos
        self.__dicCheckboxLookup = {"Gear": self.gear,
                                  "CRS": self.crs,
                                  "Person": self.person,
                                  "Lifestage": self.lifestage,
                                  "TaxCover": self.taxcover,
                                  "Taxon": self.taxon,
                                  "Ship": self.ship,
                                    "PosSys": self.possys,
                                    "Dataset": self.dataset,
                                    "Country": self.country}
        self.__dicCheckboxTaxon = {"TaxBase": self.base,
                                   "TaxPrivate": self.worms_private,
                                   "TaxAaid": self.aaid,
                                   "TaxColony": self.colony,
                                   "TaxOut": self.noworms,
                                   "TaxChange": self.change}

        # Buttons verbinden
        self.open.clicked.connect(self.__save)
        self.select_all_lookup.clicked.connect(self.__selectAllLookup)
        self.deselect_all_lookup.clicked.connect(self.__deselectAllLookup)
        self.select_all_taxon.clicked.connect(self.__selectAllTaxon)
        self.deselect_all_taxon.clicked.connect(self.__deselectAllTaxon)
        for sName, checkbox in self.__dicCheckboxLookup.items():
            checkbox.stateChanged.connect(self.__check)
        for sName, checkbox in self.__dicCheckboxTaxon.items():
            checkbox.stateChanged.connect(self.__check)
        self.file.textChanged.connect(self.__check)
        # Checkboxen aktivieren
        self.__selectAllLookup()
        self.__selectAllTaxon()

    # Lookup: Select all
    def __selectAllLookup(self):
        for k, v in self.__dicCheckboxLookup.items():
            v.setChecked(True)

    # Lookup:Deselect all
    def __deselectAllLookup(self):
        for k, v in self.__dicCheckboxLookup.items():
            v.setChecked(False)

    # Taxon: Select all
    def __selectAllTaxon(self):
        for k, v in self.__dicCheckboxTaxon.items():
            v.setChecked(True)

    # Taxon: Deselect all
    def __deselectAllTaxon(self):
        for k, v in self.__dicCheckboxTaxon.items():
            v.setChecked(False)

    # Check Input Config
    def __check(self):
        bLookupOk = False
        bTaxonOk = False
        bFileOk = False
        if self.__bLookup:
            for k, v in self.__dicCheckboxLookup.items():
                if v.isChecked():
                    bLookupOk = True
                    break
        if self.__bTaxon:
            for k, v in self.__dicCheckboxTaxon.items():
                if v.isChecked():
                    bTaxonOk = True
                    break
        if self.file.text() != "":
            bFileOk = True
        if self.__bLookup and bLookupOk and bFileOk:
            self.buttonBox.button(QDialogButtonBox.Ok).show()
        elif self.__bTaxon and bTaxonOk and bFileOk:
            self.buttonBox.button(QDialogButtonBox.Ok).show()
        else:
            self.buttonBox.button(QDialogButtonBox.Ok).hide()

    # Save
    def __save(self):
        if self.__bLookup:
            sFile = "lookup.xlsx"
        else:
            sFile= "taxsys.xlsx"
        dataFile = QFileDialog.getSaveFileName(self, "Select xlsx file", directory=self.__model.conf_getUserHomeDir(), filter="*.xlsx")[0]
        self.file.setText(dataFile)
        self.file.setCursorPosition(0)

    # Konfigurtion auslesen
    def getConfig(self):

        res = Result(name="Result Data Import")
        # Auswahl Lookup
        lstLookup = []
        for k, v in self.__dicCheckboxLookup.items():
            if v.isChecked():
                lstLookup.append(k)

        # Auswahl Taxon
        lstTaxon = []
        for k, v in self.__dicCheckboxTaxon.items():
            if v.isChecked():
                lstTaxon.append(k)

        # Gold
        iGold = self.comboGold.currentIndex()
        lstGold = ["all", "not-gold", "gold"]
        sGold = lstGold[iGold]

        res.addData("file", self.file.text())
        if self.__bLookup:
            res.addData("lookup", lstLookup)
            res.addData("lookup-gold-export", sGold)
        if self.__bTaxon:
            res.addData("taxon", lstTaxon)

        return res

# Dialog: ExportSubset
class DialogExportSubset(QDialog, uic.loadUiType(fileExportSubset)[0]):

    # Konstruktor
    def __init__(self, model):
        """

        Args:
            (model): Model
        """

        # Super
        QDialog.__init__(self)

        # Instanzvariablen
        self.__model = model

        # Initialisierung
        self.setupUi(self)
        self.setWindowFlags(Qt.Dialog | Qt.CustomizeWindowHint | Qt.WindowTitleHint | Qt.WindowCloseButtonHint)

        sTitle = "Export subset"

        # Fenster
        self.setFixedHeight(self.sizeHint().height())
        self.setFixedWidth(self.sizeHint().width())
        self.setWindowModality(Qt.ApplicationModal)
        self.setWindowTitle(sTitle)

        # Combos
        self.__dicCheckboxIngest = {"Cruise": self.cruise,
                     "Station": self.station,
                     "Sample": self.sample,
                     "SubSample": self.subsample,
                     "Subset": self.subset,
                     "Biota": self.biota,
                     "Sediment": self.sediment,
                     "IndividualMeasure": self.indmeasure,
                                    }
        self.__dicCheckboxLookup = {"Gear": self.gear,
                                    "CRS": self.crs,
                                    "Person": self.person,
                                    "Lifestage": self.lifestage,
                                    "TaxCover": self.taxcover,
                                    "Taxon": self.taxon,
                                    "Ship": self.ship,
                                    "PosSys": self.possys,
                                    "Dataset": self.dataset,
                                    "Country": self.country}

        # Buttons verbinden
        self.open.clicked.connect(self.__open)
        self.select_all_ingest.clicked.connect(self.__selectAllIngest)
        self.deselect_all_ingest.clicked.connect(self.__deselectAllIngest)
        self.select_all_lookup.clicked.connect(self.__selectAllLookup)
        self.deselect_all_lookup.clicked.connect(self.__deselectAllLookup)
        # Check
        for sName, checkbox in self.__dicCheckboxIngest.items():
            checkbox.stateChanged.connect(self.__check)
        for sName, checkbox in self.__dicCheckboxLookup.items():
            checkbox.stateChanged.connect(self.__check)
        self.file.textChanged.connect(self.__check)
        # Buttonbox
        self.buttonBox.button(QDialogButtonBox.Ok).hide()
        # Alle Checkboxen aktivieren
        self.__selectAllIngest()
        self.__selectAllLookup()

    # Ingest: Select all
    def __selectAllIngest(self):
        for k, v in self.__dicCheckboxIngest.items():
            if v.isEnabled():
                v.setChecked(True)

    # Ingest: Deselect all
    def __deselectAllIngest(self):
        for k, v in self.__dicCheckboxIngest.items():
            if v.isEnabled():
                v.setChecked(False)

    # Lookup: Select all
    def __selectAllLookup(self):
        for k, v in self.__dicCheckboxLookup.items():
            if v.isEnabled():
                v.setChecked(True)

    # Lookup: Deselect all
    def __deselectAllLookup(self):
        for k, v in self.__dicCheckboxLookup.items():
            if v.isEnabled():
                v.setChecked(False)

    # Check Input Config
    def __check(self):
        bIngestOk = False
        bLookupOk = False
        bFileOk = False
        for k, v in self.__dicCheckboxIngest.items():
            if v.isChecked():
                bIngestOk = True
                break
        for k, v in self.__dicCheckboxLookup.items():
            if v.isChecked():
                bLookupOk = True
                break
        if self.file.text() != "":
            bFileOk = True
        if (bIngestOk or bLookupOk) and bFileOk:
            self.buttonBox.button(QDialogButtonBox.Ok).show()
        else:
            self.buttonBox.button(QDialogButtonBox.Ok).hide()

    # Open
    def __open(self):
        sFileDefault = str(Path.home().joinpath("subset.xlsx"))
        dataFile = QFileDialog.getSaveFileName(self, "Export subset to file", directory=str(Path(self.__model.conf_getWorkingDir() / Path("subset.xlsx"))), filter="*.xlsx;; *.*")[0]
        self.file.setText(dataFile)
        self.file.setCursorPosition(0)

    # Konfigurtion auslesen
    def getConfig(self):
        res = Result(name="Result Data Import")
        # Auswahl Ingest
        lstIngest = []
        for k, v in self.__dicCheckboxIngest.items():
            if v.isChecked():
                lstIngest.append(k)
        # Auswahl Lookup
        lstLookup = []
        for k, v in self.__dicCheckboxLookup.items():
            if v.isChecked():
                lstLookup.append(k)
        res.addData("file", self.file.text())
        res.addData("ingest", lstIngest)
        res.addData("lookup", lstLookup)
        res.addData("pangaea", self.checkPangaea.isChecked())
        res.addData("rfriendly", self.checkRfriendly.isChecked())
        return res

# Dialog: System Info
class DialogSystemInfo(QDialog, uic.loadUiType(fileSysInfo)[0]):

    # Konstruktor
    def __init__(self, model):
        """

        Args:
            (model): Model
        """

        # Super
        QDialog.__init__(self)

        # Instanzvariablen
        self.__model = model

        # Initialisierung
        self.setupUi(self)
        self.setWindowFlags(Qt.Dialog | Qt.CustomizeWindowHint | Qt.WindowTitleHint | Qt.WindowCloseButtonHint)

        # Buttons verbinden
        self.mapper.clicked.connect(self.__mapper)
        self.wizard.clicked.connect(self.__wizard)
        self.myworms.clicked.connect(self.__myworms)
        self.base.clicked.connect(self.__base)
        self.database.clicked.connect(self.__database)
        self.exporter.clicked.connect(self.__exporter)
        self.converter.clicked.connect(self.__converter)
        self.feeder.clicked.connect(self.__feeder)
        self.ingestor.clicked.connect(self.__ingestor)
        self.importer.clicked.connect(self.__importer)
        self.validator.clicked.connect(self.__validator)
        self.tamer.clicked.connect(self.__tamer)
        self.comments.clicked.connect(self.__comments)
        self.lookup.clicked.connect(self.__lookup)
        self.connector.clicked.connect(self.__connector)
        self.virtualAssistant.clicked.connect(self.__virtualAssistant)
        self.dataPatrol.clicked.connect(self.__dataPatrol)
        self.batchbot.clicked.connect(self.__batchBot)
        self.regions.clicked.connect(self.__regions)

    # Infos
    def __mapper(self):
        self.textBox.setPlainText(self.__model.subsystem_getMapper().info())

    def __wizard(self):
        self.textBox.setPlainText(self.__model.subsystem_getTaxonWizard().info())

    def __base(self):
        self.textBox.setPlainText(self.__model.info())

    def __myworms(self):
        self.textBox.setPlainText(self.__model.subsystem_getMyWorms().info())

    def __database(self):
        self.textBox.setPlainText(self.__model.subsystem_getDb().info())

    def __exporter(self):
        self.textBox.setPlainText(self.__model.subsystem_getExporter().info())

    def __converter(self):
        self.textBox.setPlainText(self.__model.subsystem_getConverter().info())

    def __ingestor(self):
        self.textBox.setPlainText(self.__model.subsystem_getIngestor().info())

    def __feeder(self):
        self.textBox.setPlainText(self.__model.subsystem_getFeeder().info())

    def __importer(self):
        self.textBox.setPlainText(self.__model.subsystem_getImporter().info())

    def __validator(self):
        self.textBox.setPlainText(self.__model.subsystem_getValidator().info())

    def __tamer(self):
        self.textBox.setPlainText(self.__model.subsystem_getTamer().info())

    def __comments(self):
        self.textBox.setPlainText(self.__model.conf_getDbSchema().info())

    def __lookup(self):
        self.textBox.setPlainText(self.__model.subsystem_getLookup().info())

    def __connector(self):
        self.textBox.setPlainText(self.__model.subsystem_getConnector().info())

    def __virtualAssistant(self):
        self.textBox.setPlainText(self.__model.subsystem_getVirtualAssistant().info())

    def __dataPatrol(self):
        self.textBox.setPlainText(self.__model.subsystem_getDataPatrol().info())

    def __batchBot(self):
        self.textBox.setPlainText(self.__model.subsystem_getBatchBot().info())

    def __regions(self):
        self.textBox.setPlainText(self.__model.subsystem_getRegions().info())

# Dialog: TaxaDetective
class  DialogVirtualAssistantTaxonSpotter(QDialog, uic.loadUiType(fileVaTaxonSpotter)[0]):

    # Konstruktor
    def __init__(self, model, **kwargs):
        """
        Args:
            (model): Model
        """

        # Super
        QDialog.__init__(self)

        # Instanzvariablen
        self.__model = model
        self.__odicWb = OrderedDict()
        self.__sSheetBiota = "Taxa" # Std: Sheet in dem sich die Taxa Daten befinden
        self.__sColAid_user = "AphiaId given" # Std: Spalte in der sich die Aphia Ids befinden
        self.__sColSn_user = "ScientificName given" # Std: Spalte in der sich die ScientificNamen befinden
        self.__sColAid = "AphiaId" # Std: Spalte in der sich die Aphia Ids befinden
        self.__sColSn = "ScientificName" # Std: Spalte in der sich die ScientificNamen befinden
        self.__sColAid_acc = "AphiaId accepted"  # Std: Spalte in der sich die Aphia Ids befinden
        self.__sColSn_acc = "ScientificName accepted"  # Std: Spalte in der sich die ScientificNamen befinden

        # Initialisierung
        self.setupUi(self)
        self.setWindowModality(Qt.ApplicationModal)
        self.setWindowFlags(Qt.Dialog | Qt.CustomizeWindowHint | Qt.WindowTitleHint | Qt.WindowCloseButtonHint)

        # Verbinden
        self.open_original.clicked.connect(self.__openFile)
        self.open_dir_working.clicked.connect(self.__openWorkDir)
        self.comboSheet.currentIndexChanged.connect(self.__changedSheet)

        # Report Working Dir Std
        sReportDirStd = str(Path(self.__model.conf_getWorkingDir(), "report"))
        self.dir_working.setText(sReportDirStd)

        # Buttonbox
        self.buttonBox.button(QDialogButtonBox.Ok).hide()

        # Combos
        self.__combos = OrderedDict([
            ("aid_given", self.comboColAid_given),
            ("sn_given", self.comboColSn_given),
            ("aid", self.comboColAid),
            ("sn", self.comboColSn),
            ("aid_acc", self.comboColAid_acc),
            ("sn_acc", self.comboColSn_acc)
        ])

        # Groeße fixieren
        self.setMaximumSize(self.sizeHint())

    # Aus Combo-Box mit dem Biota-Sheet wurde ein neues Sheet ausgewählt
    def __changedSheet(self):
        self.__setupCombos(str(self.comboSheet.currentText()))

    # Fülle die Combo-Boxen mit Inhalt
    def __setupCombos(self, sSheet=None):
        # Initaler Aufbau nach Laden
        # Signale aus
        self.comboSheet.blockSignals(True)
        for sName, combo in self.__combos.items():
            combo.blockSignals(True)

        # Sheet
        try:
            lstSheets = list(self.__odicWb.keys())
        except:
            lstSheets = []
        self.comboSheet.clear()
        # Fülle Combo
        self.comboSheet.addItems(lstSheets)
        # Ermittle Voreinstellung
        if sSheet is None:
            for iIdx, sSheet_tmp in enumerate(lstSheets):
                sSheet_tmp_lower = sSheet_tmp.lower()
                if ("taxon" in sSheet_tmp_lower) or ("taxa" in sSheet_tmp_lower) or ("biota" in sSheet_tmp_lower) or ("species" in sSheet_tmp_lower):
                    sSheet = sSheet_tmp
                    self.comboSheet.setCurrentIndex(iIdx)
                    break
        else:
            iIdx = self.comboSheet.findText(sSheet)
            if iIdx >= 0:
                self.comboSheet.setCurrentIndex(iIdx)

        # Columns
        for sName, combo in self.__combos.items():
            combo.clear()
            if sSheet in self.__odicWb:
                # Fülle Combo mit Columns
                combo.addItem(None)
                combo.addItems(self.__odicWb[sSheet])
                # Ermittle Voreinstellung
                for iIdx, sCol in enumerate(self.__odicWb[sSheet]):
                    sCol_lower = sCol.lower()
                    if sName == "aid_given":
                        # AphiaId given
                        if ("given" in sCol_lower) and ("aphia" in sCol_lower or "aid" in sCol_lower):
                            combo.setCurrentIndex(iIdx+1)
                            break
                    elif sName == "sn_given":
                        # Sn given
                        if ("given" in sCol_lower) and ("scientific" in sCol_lower):
                            combo.setCurrentIndex(iIdx+1)
                            break
                    elif sName == "aid":
                        # Aid
                        if ("given" not in sCol_lower) and ("acc" not in sCol_lower) and ("aphia" in sCol_lower or "aid" in sCol_lower):
                            combo.setCurrentIndex(iIdx+1)
                            break
                    elif sName == "sn":
                        # Sn
                        if ("given" not in sCol_lower) and ("acc" not in sCol_lower) and ("scientific" in sCol_lower):
                            combo.setCurrentIndex(iIdx+1)
                            break
                    elif sName == "aid_acc":
                        # AphiaId accepted
                        if ("acc" in sCol_lower) and ("aphia" in sCol_lower or "aid" in sCol_lower):
                            combo.setCurrentIndex(iIdx+1)
                            break
                    elif sName == "sn_acc":
                        # Sn accepted
                        if ("acc" in sCol_lower) and ("scientific" in sCol_lower):
                            combo.setCurrentIndex(iIdx+1)
                            break

        # Given ist leer - Accepted nicht -> Normal zu Given holen
        if self.comboColAid_given.currentIndex() == 0 and self.comboColAid.currentIndex() != 0:
            self.comboColAid_given.setCurrentIndex(self.comboColAid.currentIndex())
            self.comboColAid.setCurrentIndex(0)
        if self.comboColSn_given.currentIndex() == 0 and self.comboColSn.currentIndex() != 0:
            self.comboColSn_given.setCurrentIndex(self.comboColSn.currentIndex())
            self.comboColSn.setCurrentIndex(0)
        # Given ist leer - Accepted nicht -> Accepted zu Given holen
        if self.comboColAid_given.currentIndex() == 0 and self.comboColAid_acc.currentIndex() != 0:
            self.comboColAid_given.setCurrentIndex(self.comboColAid_acc.currentIndex())
            self.comboColAid_acc.setCurrentIndex(0)
        if self.comboColSn_given.currentIndex() == 0 and self.comboColSn_acc.currentIndex() != 0:
            self.comboColSn_given.setCurrentIndex(self.comboColSn_acc.currentIndex())
            self.comboColSn_acc.setCurrentIndex(0)

        # Signale an
        self.comboSheet.blockSignals(False)
        for sName, combo in self.__combos.items():
            combo.blockSignals(False)

        # Sinnvolle Auswahl möglich -> OK Button freigeben
        bOk = True
        for sName, combo in self.__combos.items():
            if combo.count() == 0:
                bOk = False

        if bOk:
            self.buttonBox.button(QDialogButtonBox.Ok).show()
        else:
            self.buttonBox.button(QDialogButtonBox.Ok).hide()

        # Groeße fixieren
        self.setMaximumSize(self.sizeHint())

    def __openWorkDir(self):
        sReportDirStd = str(Path(self.__model.conf_getWorkingDir(), "report"))
        sDir = QFileDialog.getExistingDirectory(self, "Select directory to save report", directory=sReportDirStd)
        if sDir not in ["", None]:
            self.dir_working.setText(sDir)
            self.dir_working.setCursorPosition(0)
            
    def __openFile(self):
        dataFile = QFileDialog.getOpenFileName(self, "Select EXCEL file", directory=self.__model.conf_getWorkingDir(), filter="*.xlsx")[0]
        self.file_xlsx.setText(dataFile)
        self.file_xlsx.setCursorPosition(0)
        try:
            warnings.simplefilter("ignore")  # Ignoriere Warnungen
            wb = openpyxl.load_workbook(dataFile, read_only=True)
            warnings.simplefilter("default")  # Warnungen wieder aktivieren
            # Hole Wb
            resWb = ExcelTools.getWb(dataFile, rw=False)
            if resWb.isOk():
                wb = resWb.getData("wb")
                # Hole alle Sheets mit Columns
                self.__odicWb = OrderedDict()
                iIdxColAid = None
                iIdxColScientificname = None
                for sheet in wb:
                    # Sheets
                    sSheet = sheet.title
                    if sSheet not in self.__odicWb:
                        self.__odicWb[sSheet] = []
                        # Header
                        try:
                            header = wb[sSheet][1]
                            for cell in header:
                                if cell.value is not None:
                                    sCol = str(cell.value)
                                    self.__odicWb[sSheet].append(sCol)
                                    if sCol == self.__sColAid:
                                        iIdxColAid = cell.column - 1
                                    elif sCol == self.__sColSn:
                                        iIdxColScientificname = cell.column - 1
                        except:
                            # Tabelle ist leer
                            pass
                # Setup Combos
                self.__setupCombos()
        except Exception as ex:
            self.__model.ui_getSmartprinter().normal("Internal Problem: " + str(ex))


    # Konfigutaion auslesen und zurückgeben
    def getConfig(self):
        res = Result(name="Detective")
        res.addData("file", self.file_xlsx.text())
        res.addData("reportdir", self.dir_working.text())
        res.addData("sheet", self.comboSheet.currentText())
        res.addData("fuzzy", self.checkFuzzy.isChecked())
        res.addData("like", self.checkLike.isChecked())
        res.addData("puffer", self.checkBuffer.isChecked())
        res.addData("marine_only", self.checkMarineOnly.isChecked())
        for sName, combo in self.__combos.items():
            res.addData(sName, combo.currentText())
        return res

# Dialog: TaxaPolishing
class DialogVirtualAssistantPolishBiota(QDialog, uic.loadUiType(fileVaBiotaPolish)[0]):

    # Konstruktor
    def __init__(self, model, **kwargs):
        """
        Args:
            (model): Model
        """

        # Super
        QDialog.__init__(self)

        # Instanzvariablen
        self.__model = model
        self.__odicWb = OrderedDict()

        # Initialisierung
        self.setupUi(self)
        self.setWindowModality(Qt.ApplicationModal)
        self.setWindowFlags(Qt.Dialog | Qt.CustomizeWindowHint | Qt.WindowTitleHint | Qt.WindowCloseButtonHint)

        # Verbinden
        self.open_original.clicked.connect(self.__openOriginal)
        self.open_polished.clicked.connect(self.__openPolished)
        self.comboSheetBiota.currentIndexChanged.connect(self.__changedSheet)

        # Buttonbox
        self.buttonBox.button(QDialogButtonBox.Ok).hide()

        # Groeße fixieren
        self.setMaximumSize(self.sizeHint())

    # Aus Combo-Box mit dem Biota-Sheet wurde ein neues Sheet ausgewählt
    def __changedSheet(self):
        self.__setupCombos(str(self.comboSheetBiota.currentText()))

    # Fülle die Combo-Boxen mit Inhalt
    def __setupCombos(self, sSheet=None):
        self.comboSheetBiota.blockSignals(True)
        self.comboColAid.blockSignals(True)
        self.comboColSn.blockSignals(True)

        # Initaler Aufbau nach Laden
        bSheetBiotaFound = False

        # Sheets
        self.comboSheetBiota.clear()
        self.comboSheetBiota.addItems(self.__odicWb.keys())
        if sSheet is None:
            for iIdx, sSheet_tmp in enumerate(self.__odicWb.keys()):
                sSheet_lower = sSheet_tmp.lower()
                if ("biota" in sSheet_lower or "species" in sSheet_lower or "taxa" in sSheet_lower or "taxon" in sSheet_lower ):
                    self.comboSheetBiota.setCurrentIndex(iIdx)
                    bSheetBiotaFound = True
                    sSheet = sSheet_tmp
                    break
        else:
            iIdx = self.comboSheetBiota.findText(sSheet)
            if iIdx >= 0:
                self.comboSheetBiota.setCurrentIndex(iIdx)
                bSheetBiotaFound = True

        # Col: Aphia Ids
        self.comboColAid.clear()
        if bSheetBiotaFound:
            # Biota-Sheet gefunden -> Hole Vordefinierte Scientificname Spalte
            self.comboColAid.addItem(None)
            self.comboColAid.addItems(self.__odicWb[sSheet])
            for iIdx, sCol in enumerate(self.__odicWb[sSheet]):
                sCol_lower = sCol.lower()
                if ("aphia" in sCol_lower or "aid" in sCol_lower):
                    self.comboColAid.setCurrentIndex(iIdx + 1)
                    break

        # Col: ScientificNames
        self.comboColSn.clear()
        if bSheetBiotaFound:
            self.comboColSn.addItem(None)
            self.comboColSn.addItems(self.__odicWb[sSheet])
            for iIdx, sCol in enumerate(self.__odicWb[sSheet]):
                sCol_lower = sCol.lower()
                if ("scientific" in sCol_lower or "name" in sCol_lower):
                    self.comboColSn.setCurrentIndex(iIdx + 1)
                    break

        self.comboSheetBiota.blockSignals(False)
        self.comboColAid.blockSignals(False)
        self.comboColSn.blockSignals(False)

        # Sinnvolle Auswahl möglich -> OK Button freigeben
        if self.comboSheetBiota.count() > 0 and self.comboColAid.count() > 0 and self.comboColSn.count() > 0:
            self.buttonBox.button(QDialogButtonBox.Ok).show()
        else:
            self.buttonBox.button(QDialogButtonBox.Ok).hide()
        # Groeße fixieren
        self.setMaximumSize(self.sizeHint())

    # Open Original
    def __openOriginal(self):
        dataFile = QFileDialog.getOpenFileName(self, "Select EXCEL file", directory=self.__model.conf_getUserHomeDir(), filter="*.xlsx")[0]
        self.file_original.setText(dataFile)
        self.file_original.setCursorPosition(0)
        try:
            warnings.simplefilter("ignore")  # Ignoriere Warnungen
            wb = openpyxl.load_workbook(dataFile, read_only=True)
            warnings.simplefilter("default")  # Warnungen wieder aktivieren
            # Hole Wb
            resWb = ExcelTools.getWb(dataFile, rw=False)
            if resWb.isOk():
                wb = resWb.getData("wb")
                # Hole alle Sheets mit Columns
                self.__odicWb = OrderedDict()
                for sheet in wb:
                    # Sheets
                    sSheet = sheet.title
                    if sSheet not in self.__odicWb:
                        self.__odicWb[sSheet] = []
                        # Header
                        try:
                            header = wb[sSheet][1]
                            for cell in header:
                                if cell.value is not None:
                                    sCol = str(cell.value)
                                    self.__odicWb[sSheet].append(sCol)
                        except:
                            # Tabelle ist leer
                            pass
                # Setup Combos
                self.__setupCombos()
        except Exception as ex:
            self.__model.ui_getSmartprinter().normal("Internal Problem: " + str(ex))

    # Open Polished
    def __openPolished(self):
        sFileIn = self.file_original.text()
        if sFileIn not in ["", None]:
            sFileName = FileTools.getFileName(sFileIn)
            sFilePath = FileTools.getFilePath(sFileIn)
            sFilePolished = "/"+sFileName+"_polished.xlsx"
        else:
            sFilePolished = ""
        dataFile = QFileDialog.getSaveFileName(self, "Select polished Excel filename", directory=self.__model.conf_getUserHomeDir() + sFilePolished, filter="*.xlsx")[0]
        self.file_polished.setText(dataFile)
        self.file_polished.setCursorPosition(0)


    # Konfigutaion auslesen und zurückgeben
    def getConfig(self):
        res = Result(name="Virtiual Assistant Biota Polish Dialog")
        res.addData("file_original", self.file_original.text())
        res.addData("file_polished", self.file_polished.text())
        res.addData("sheet_biota", self.comboSheetBiota.currentText())
        res.addData("col_aid", self.comboColAid.currentText())
        res.addData("col_sn", self.comboColSn.currentText())
        res.addData("fuzzy", self.checkFuzzy.isChecked())
        res.addData("like", self.checkLike.isChecked())
        res.addData("puffer", self.checkBuffer.isChecked())
        res.addData("marine_only", self.checkMarineOnly.isChecked())
        return res

# Dialog: Ingest
class DialogGenericComboChoice(QDialog, uic.loadUiType(fileGenericComboChoice)[0]):

    # Konstruktor
    def __init__(self, model, sTitle, odicChoices):
        """
        Args:
            (model): Model
        """

        # Super
        QDialog.__init__(self)

        # Instanzvariablen
        self.__model = model
        self.__sTitle = sTitle
        self.__odicChoices = odicChoices
        self.__dicCombos = {}

        # Initialisierung
        self.setupUi(self)
        self.setWindowFlags(Qt.Dialog | Qt.CustomizeWindowHint | Qt.WindowTitleHint | Qt.WindowCloseButtonHint)

        # Build Comboboxen
        self.buildCombos()

        # Größe anpassen
        self.adjustSize()
        #self.setFixedWidth(self.width()+30)
        self.setMaximumHeight(800)

    def buildCombos(self):
        self.setWindowTitle(self.__sTitle)
        widget = QWidget()
        grid = QGridLayout()
        for iRow, (sSn, lstChoices) in enumerate(self.__odicChoices.items()):
            combo = QComboBox()
            combo.addItems(lstChoices)
            self.__dicCombos[sSn] = combo
            grid.addWidget(QLabel(sSn), iRow, 1)
            grid.addWidget(combo, iRow, 2)
        widget.setLayout(grid)
        self.scroll.setWidget(widget)

    # Konfigutaion auslesen und zurückgeben
    def getSelection(self):
        dicSelection = {}
        for sSn, combo in self.__dicCombos.items():
            dicSelection[sSn] = combo.currentIndex()
        return dicSelection