"""
Modul mit Hilfsklasse für Datenbankzugriffe
"""

# coding=utf-8
import psycopg2, re, sys, datetime, math, getpass, platform, socket, requests, json, time, datetime, subprocess, os, zipfile
from psycopg2 import sql
from collections import OrderedDict
from tools.printing import SmartPrinter
from tools.coding import Result, DataTools
from tools.fileio import FileTools
from tools.text import PrettyText
from helper.selector import Selector
from logik.regions import Regions
from pathlib import Path

class Database:
    """
    DIESE KLASSE BIETET FUNKTIONALITÄTEN FÜR DEN ZUGRIFF AUF DIE DATENBANK
    """
    def __init__(self, model):
        """
        Args:
            model (Model): Model
        """

        # Instanzvariablen
        self.__model = model
        self.__prnt = model.ui_getSmartprinter()

        # DB-Login
        self.__iConnetionTimeout = 4  # Timeout für Connection in Sekunden
        self.__lstAccounts = []

        # Accounts
        self.__dicAccounts_backup = {}
        self.__dicAccounts = {}
        self.__sAccount_active_backup = None
        self.__sAccount_active = None
        self.account_all_set()

        # ------------------------------------------------------------------------------------------------------------

    # ToString
    def __str__(self):
        return self.info()

    # Info Init
    def info(self):
        sOut = "*** Status of Database System\n\n"
        sOut += "Connection Timeout: " + str(self.__iConnetionTimeout) +"\n"
        sOut += "Number of Accounts: " + str(self.account_all_getNr()) + "\n"
        sOut += "Active Account (" + str(self.account_active_getIndex()) + "): "+str(self.account_active_getName()) + "\n"
        if self.account_all_getNr() > 0:
            sOut += "Accounts:\n"
            for iAccountIndex, sAccountName in enumerate(self.account_all_getNames()):

                sOut += str(iAccountIndex)+" ["+str(sAccountName)+"] "+self.conn_getConnectionInfo(sAccountName) + "\n"
        return sOut

    # ------------------------------------------------------------------------------------------------------------

    # --- Connection

    def conn_makeConnStr(self, **kwargs):
        """
        Erzeuge einen DB Connection-String fuer PSQL

        Returns:
            str: Connection-String fuer PSQL
        """
        # Kwargs
        sHost = kwargs["host"] if "host" in kwargs else None
        sPort = kwargs["port"] if "port" in kwargs else None
        sDb = kwargs["db"] if "db" in kwargs else None
        sUser = kwargs["user"] if "user" in kwargs else None
        sPw = kwargs["pw"] if "pw" in kwargs else None
        return "connect_timeout=" + str(self.__iConnetionTimeout) + " dbname='" + str(sDb) + "' user='" + str(sUser) + "' host='" + str(sHost) + "' port='" + str(sPort) + "' password='" + str(sPw) + "'"

    def conn_getConnStrPsql(self):
        """
        Hole DB Connection-String fuer PSQL
        Returns:
            str: Connection-String fuer PSQL
        """
        return "connect_timeout=" + str(self.__iConnetionTimeout) +" dbname='" + str(self.account_single_getPart("db")) + "' user='" + str(self.account_single_getPart("user")) + "' password='" + str(self.account_single_getPart("pw")) + "' host='" + str(self.account_single_getPart("host")) + "' port='" + str(self.account_single_getPart("port")) + "'"  # Verbindungs-String

    def conn_getConnection(self, **kwargs):
        """
        Hole Connection
        Kwargs
            constr (str): Connection String
        Return
            Result:
                connection
                cursor
        """
        res = Result("Get DB Connection")
        sConString  = kwargs["constr"] if "constr" in kwargs else self.conn_getConnStrPsql()
        cursor = None
        conn = None
        try:
            conn = psycopg2.connect(sConString)
            cursor = conn.cursor()
            res.addData("connection", conn)
            res.addData("cursor", cursor)
        except Exception as e:
            res.addErr("Problems establishing connection", add=str(e), sys="Database", kat="Get Connection")
        return res

    def conn_getConnectionInfo(self, sAccount=None):
        """
        Hole Info String zu Verbindung
        """
        if sAccount is None:
            sAccount = self.account_active_getName()
        sConnection = str(self.account_single_getPart("user", account_name=sAccount)) + "@" + str(self.account_single_getPart("schema", account_name=sAccount)) + "." + str(self.account_single_getPart("db", account_name=sAccount)) + "("+ str(self.account_single_getPart("host", account_name=sAccount)) + ":" + str(self.account_single_getPart("port", account_name=sAccount))+")"
        return sConnection

    # --- Check & Test

    def check_isLeafExisting(self, **kwargs):
        """
        Test ob Leaf exisitert

        FUNKTION

        a) Teste sukzessive ob und wie weit ein Ast bereits in der Datenbank exisitert Cruise[-Station[-Sample[-Subset]]]
        b) Bestimme dann die Id und die das beinhaltende Ingest (Nummer)
        c) Das Ingest (Nummer) stellt die direkte Abhängigkeit dar
        d) Bestimme die Datasets, die die selben Teilbäume referenzieren

        EINGABE
        namen Cruise,Station,Sample,Subset: Namen des zu testenden Teilbaumes

        ERGEBNIS
        level:      Länge des Teilbaums: Cruise, Cruise/Station, Cruise/Station/Sample
        ids:        Die Ids der weiterbenutzten/erweiterten Teilbaumes
        ingests:    Es bestehen direkte Abhängigkeit von diesen Ingests
        datasets:   Datasets, die die selben Abhängigkeiten haben wie erweiterter Teilbaum

        BSP: DATASETS-ABHÄNGIGKEIT
        Ein zu testender Teilbaum kann nicht von einem Dataset abhängig sein
                subset1 dataset 1
        cruise
                subset2 dataset 2
        Subset2 kann nicht von Subset1 abhängig sein.
        Sie haben immer eine Gemeinsamkeit:
        Beide sind vom selber Cruise abhängig (egal wo sie sich treffen: Sample/Station/Cruise)
        Mehr kann man über Datasets nicht sagen

        Kwargs:
            cursor: Cursor
            row (row): Row
            sCruise_name (str): Cruise
            sStation_name (str): Station
            sSample_name (str): Sample
            sSubSample_name (str): SubSample
            sSubset_name (str): Subset
        Returns:
            tools.coding.Result: Ergebnis
                leaf (str): Ast
                leaf_verbose (str): Ast mit mehr Informationen
                level (str): Leaf Länge
                cruise_id (int): Id Cruise
                station_id (int): Id Station von Kombination (Cruise-Station)
                sample_id (int): Id Sample von Kombination (Cruise-Station-Sample)
                subsample_id (int): of SubSample von Kombination (Cruise-Station-Sample-SubSample)
                subset_id (int): Id Subset von Kombination (Cruise-Station-Sample-SubSample-Subset)
                cruise_ingest_id (int): IngestId Cruise
                station_ingest_id (int): IngestId Station von Kombination (Cruise-Station)
                sample_ingest_id (int): IngestId Sample von Kombination (Cruise-Station-Sample)
                subsample_ingest_id (int): IngestId SubSample von Kombination (Cruise-Station-Sample-SubSample)
                subset_ingest_id (int): IngestId Sample von Kombination (Cruise-Station-Sample-SubSampleSubset)
                ingests (set): Liste der Ingests von denen eine direkte Abhängigkeit besteht
                sql (dic): SQL Befehle
        """

        res = Result(name="Testing for existance of leaf")

        # Kwargs
        cursor = kwargs["cursor"] if "cursor" in kwargs else None
        row = kwargs["row"] if "row" in kwargs else None
        sCruise_name = str(kwargs["cruise"]) if "cruise" in kwargs and (kwargs["cruise"] not in ("", None)) else None
        sStation_name = str(kwargs["station"]) if "station" in kwargs and (kwargs["station"] not in ("", None)) else None
        sSample_name = str(kwargs["sample"]) if "sample" in kwargs and (kwargs["sample"] not in ("", None)) else None
        sSubSample_name = str(kwargs["subsample"]) if "subsample" in kwargs and (kwargs["subsample"] not in ("", None)) else None
        sSubset_name = str(kwargs["subset"]) if "subset" in kwargs and (kwargs["subset"] not in ("", None)) else None

        # DB
        if cursor is None:
            resCursor = self.__model.subsystem_getDb().conn_getConnection()
            if resCursor.isErr():
                res.addResult(resCursor)
                return res
            conn = resCursor.getData("connection")
            cursor = resCursor.getData("cursor")

        # Test Parameter
        # Row überschreibt Einzelparameter
        if row is not None:
            try:
                sCruise_name = row.get("Cruise")
                sStation_name = row.get("Station")
                sSample_name = row.get("Sample")
                sSubSample_name = row.get("SubSample")
                sSubset_name = row.get("Subset")
            except:
                res.addErr("Wrong configuration of leaf: Cruise[[[-Station[-Sample-[Subset]]]", add="Row object is corrupt", sys="Database", kat="Check Leaf Exsistance")
                return res

        # Teste ob Ast wirklich vollständig ist: Cruise[-Station[-Sample[-Subset]]]
        bConfigErr = False
        if sCruise_name is None:
            bConfigErr = True
        else:
            bConfigSetToNone = False
            for sLevel in [sStation_name, sSample_name, sSubset_name]:
                if sLevel is not None:
                    if bConfigSetToNone:
                        bConfigErr = True
                else:
                    bConfigSetToNone = True
        if bConfigErr:
            res.addErr("Wrong configuration of leaf: Cruise[[[-Station[-Sample-[SubSample-[Subset]]]]", add=str(sCruise_name)+"|"+str(sStation_name)+"|"+str(sSample_name)+"|"+str(sSubSample_name)+"|"+str(sSubset_name), sys="Database", kat="Check Leaf Exsistance")
            return res

        # Daten
        sSep = "|"

        iCruise_id = None
        iStation_id = None
        iSample_id = None
        iSubSample_id = None
        iSubset_id = None

        iCruise_ingestId = None
        iStation_ingestId = None
        iSample_ingestId = None
        iSubSample_ingestId = None
        iSubset_IngestId = None

        lstSubsetNamesUsed = None
        setBaseIngests = set()
        setAffectedDatasets = set()

        dicSqlCmds = {}

        # Astllevel
        sLeaf_level = None
        for sLevel, sName in [("cruise", sCruise_name), ("station", sStation_name), ("sample", sSample_name),("subsample", sSubSample_name),("subset", sSubset_name)]:
            if sName is None:
                break
            sLeaf_level = sLevel

        # Ast Info String
        sLeaf = PrettyText.lst2str([sPart for sPart in (sCruise_name,sStation_name,sSample_name, sSubSample_name, sSubset_name) if sPart is not None], sSep)

        # Ast ist korrekt definiert -> weiter
        if res.isOk():
            # *** Bestimme ob Cruise[-Station[-Sample[-SubSample[-Subset]]]] in DB existiert
            # - hole die Ids und das beinhaltende Ingest (ingest_id)
            # - das darf jeweils nur einmal exisiteren, da die Einmaligkeit für Cruise-Station-Sample-SubSample-Subset gilt

            # Cruise
            # TODO: Hier hab ich das ingest_id hinzugefuegt, ist das ein Problem?
            try:
                # Cmd
                sqlCmd = sql.SQL(
                    "SELECT id, ingest_id FROM {schema}.cruise cru WHERE cru.name={cruise_name}").format(
                    schema=sql.Identifier(self.account_active_getSchema()),
                    cruise_name=sql.Literal(sCruise_name))
                cursor.execute(sqlCmd)
                sSqlQuery = cursor.query.decode("utf-8")
                dicSqlCmds["cruise"] = sSqlQuery
                # Ergebnisse
                lstContent = cursor.fetchall()
                # Speichern: wenn gefunden
                if len(lstContent) > 0:
                    iCruise_id = lstContent[0][0]
                    iCruise_ingestId = lstContent[0][1]
                    setBaseIngests.add(iCruise_ingestId)
            except Exception as ex:
                res.addErr("Internal Problem checking Cruise:", add=sLeaf+" - "+str(ex), sys="Database", kat="Check Leaf Exsistance")

            # Station
            if iCruise_id is not None:
                try:
                    # Cmd
                    sqlCmd = sql.SQL(
                        "SELECT id, ingest_id FROM {schema}.station sta WHERE sta.name={station_name} AND sta.cruise_id={cruise_id}").format(
                        schema=sql.Identifier(self.account_active_getSchema()),
                        cruise_id=sql.Literal(iCruise_id),
                        station_name=sql.Literal(sStation_name))
                    cursor.execute(sqlCmd)
                    sSqlQuery = cursor.query.decode("utf-8")
                    dicSqlCmds["station"] = sSqlQuery
                    # Ergebnisse
                    lstContent = cursor.fetchall()
                    # DB-Konsistenz: Existieren mehr als eine Station mit selbem Namen zur selben Cruise -> SCHWERWIEGENDER-FEHLER!
                    if lstContent is not None and len(lstContent) > 1:
                        lstIdProblems = []
                        try:
                            lstIdProblems = [elem[0] for elem in lstContent]
                        except:
                            pass
                        res.addErr("SEVERE INTERNAL PROBLEM IN DATABASE!",
                                   add="More than one (" + str(len(lstIdProblems)) + ") " +
                                    "station with the same name (" + sStation_name + ") " +
                                    "pointing to same cruise (" + sCruise_name + "): " +
                                    "StationIds= " + PrettyText.lst2strC(lstIdProblems),
                                   sys="Database", kat="Check Leaf Exsistance")
                    # Speichern: wenn gefunden
                    if len(lstContent) > 0:
                        iStation_id = lstContent[0][0]
                        iStation_ingestId = lstContent[0][1]
                        setBaseIngests.add(iStation_ingestId)
                except Exception as ex:
                    res.addErr("Internal Problem checking Cruise-Station", add=sLeaf+" - "+str(ex), sys="Database", kat="Check Leaf Exsistance")

            # Sample
            if iCruise_id is not None and iStation_id is not None:
                try:
                    # Cmd
                    sqlCmd = sql.SQL(
                        "SELECT id, ingest_id FROM {schema}.sample smp WHERE smp.name={sample_name} AND smp.station_id={station_id}").format(
                        schema=sql.Identifier(self.account_active_getSchema()),
                        station_id=sql.Literal(iStation_id),
                        sample_name=sql.Literal(sSample_name))
                    cursor.execute(sqlCmd)
                    sSqlQuery = cursor.query.decode("utf-8")
                    dicSqlCmds["sample"] = sSqlQuery
                    # Ergebnis
                    lstContent = cursor.fetchall()
                    # # DB-Konsistenz: Existieren mehr als ein Sample mit selbem Namen zur selben Station -> SCHWERWIEGENDER-FEHLER!
                    if lstContent is not None and len(lstContent) > 1:
                        lstIdProblems = []
                        try:
                            lstIdProblems = [elem[0] for elem in lstContent]
                        except:
                            pass
                        res.addErr("SEVERE INTERNAL PROBLEM IN DATABASE!",
                                   add="More than one (" + str(len(lstIdProblems)) + ") " +
                                       "sample with same name (" + sSample_name + ") " +
                                       "pointing to same station (" + sStation_name + ") " +
                                       "in same cruise (" + sCruise_name + "): " +
                                       "SampleIds= " + PrettyText.lst2strC(lstIdProblems),
                                   sys="Database",
                                   kat="Check Leaf Exsistance")
                    # Speichern: Wenn gefunden
                    if len(lstContent) > 0:
                        iSample_id = lstContent[0][0]
                        iSample_ingestId = lstContent[0][1]
                        setBaseIngests.add(iSample_ingestId)
                except Exception as ex:
                    res.addErr("Internal Problem checking Cruise-Station-Sample", add=sLeaf + " - " + str(ex),
                               sys="Database", kat="Check Leaf Exsistance")

            # SubSample
            if iCruise_id is not None and iStation_id is not None and iSample_id is not None:
                try:
                    # Cmd
                    sqlCmd = sql.SQL(
                        "SELECT id, ingest_id FROM {schema}.subsample ssp WHERE ssp.name={subsample_name} AND ssp.sample_id={sample_id}").format(
                        schema=sql.Identifier(self.account_active_getSchema()),
                        sample_id=sql.Literal(iSample_id),
                        subsample_name=sql.Literal(sSubSample_name))
                    cursor.execute(sqlCmd)
                    sSqlQuery = cursor.query.decode("utf-8")
                    dicSqlCmds["subsample"] = sSqlQuery
                    # Ergebnis
                    lstContent = cursor.fetchall()
                    # # DB-Konsistenz: Existieren mehr als ein SubSample mit selbem Namen zum selben Sample -> SCHWERWIEGENDER-FEHLER!
                    if lstContent is not None and len(lstContent) > 1:
                        lstIdProblems = []
                        try:
                            lstIdProblems = [elem[0] for elem in lstContent]
                        except:
                            pass
                        res.addErr("SEVERE INTERNAL PROBLEM IN DATABASE!",
                                   add="More than one (" + str(len(lstIdProblems)) + ") " +
                                       "subsample with same name (" + sSubSample_name+ ") " +
                                       "pointing to same sample (" + sSample_name+ ") " +
                                       "pointing to same station (" + sStation_name+ "): " +
                                       "in same cruise (" + sCruise_name + "): " +
                                       "SampleIds= " + PrettyText.lst2strC(lstIdProblems),
                                   sys="Database",
                                   kat="Check Leaf Exsistance")
                    # Speichern: Wenn gefunden
                    if len(lstContent) > 0:
                        iSubSample_id = lstContent[0][0]
                        iSubSample_ingestId = lstContent[0][1]
                        setBaseIngests.add(iSubSample_ingestId)
                except Exception as ex:
                    res.addErr("Internal Problem checking Cruise-Station-Sample-SubSample", add=sLeaf + " - " + str(ex),
                               sys="Database", kat="Check Leaf Exsistance")

            # Subset
            if iCruise_id is not None and iStation_id is not None and iSample_id is not None:
                try:
                    # Cmd
                    sqlCmd = sql.SQL(
                        "SELECT id, ingest_id FROM {schema}.subset sub WHERE sub.name={subset_name} AND sub.subsample_id={subsample_id}").format(
                        schema=sql.Identifier(self.account_active_getSchema()),
                        subset_name=sql.Literal(sSubset_name),
                        subsample_id=sql.Literal(iSubSample_id))
                    cursor.execute(sqlCmd)
                    sSqlQuery = cursor.query.decode("utf-8")
                    dicSqlCmds["subset"] = sSqlQuery
                    # Ergebnis
                    lstContent = cursor.fetchall()
                    # Bestimme bereits benutzte Subset Namen
                    # Das hilft Benutzern, die sich an ein Sample hängen wollen und immer wieder einen bereits benutzten Namen benutzen
                    if lstContent is not None:
                        # Cmd
                        sqlCmd = sql.SQL(
                            "SELECT name FROM {schema}.subset sub WHERE sub.subsample_id={subsample_id}").format(
                            schema=sql.Identifier(self.account_active_getSchema()),
                            subsample_id=sql.Literal(iSubSample_id))
                        cursor.execute(sqlCmd)
                        sSqlQuery = cursor.query.decode("utf-8")
                        dicSqlCmds["subset_used"] = sSqlQuery
                        # Ergebnis
                        lstSubsetNamesUsedTmp = cursor.fetchall()
                        sSqlQuery = cursor.query.decode("utf-8")
                        try:
                            lstSubsetNamesUsed = [v[0] for v in lstSubsetNamesUsedTmp] # Immer nur das erste Element nehmen
                        except:
                            pass
                    # DB-Konsistenz: Existieren mehr als ein Subset mit selbem Namen zu selbem Sample -> SCHWERWIEGENDER-FEHLER!
                    if lstContent is not None and len(lstContent) > 1:
                        lstIdProblems = []
                        try:
                            lstIdProblems = [elem[0] for elem in lstContent]
                        except:
                            pass
                        res.addErr("Severe internal problem in database",
                                   add="More than one (" + str(len(lstIdProblems)) + ") "+
                                    "subset with same name (" + sSubset_name + ") " +
                                    "pointing to same subsample (" + sSubSample_name+ "): " +
                                    "pointing to same sample (" + sSample_name+ "): " +
                                    "pointing to same station (" + sStation_name+ ") " +
                                    "in same cruise (" + sCruise_name + "): " +
                                    "SubsetIds= " + PrettyText.lst2strC(lstIdProblems), sys="Database", kat="Check Leaf Exsistance")
                    # Speichern: wenn gefunden
                    if len(lstContent) > 0:
                        iSubset_id = lstContent[0][0]
                        iSubset_IngestId = lstContent[0][1]
                        setBaseIngests.add(iSubset_IngestId)
                except Exception as ex:
                    res.addErr("Internal Problem checking Cruise-Station-Sample-SubSample-Subset", add=sLeaf+" - "+str(ex), sys="Database", kat="Check Leaf Exsistance")

            # Bestimme alle Datasets die auf die selbe Cruise zeigen
            # Dafür ist lediglich die Cruise nötig -> Station, Sample, Subset schränken das Ergebnis nur noch ein
            if iCruise_id is not None:
                # Kommando
                sqlCmd = sql.SQL("SELECT DISTINCT cru.ingest_id, dat.name " \
                    "FROM {schema}.dataset dat " \
                    "INNER JOIN {schema}.subset sub ON sub.dataset_id=dat.id " \
                    "INNER JOIN {schema}.subsample ssp ON sub.subsample_id=ssp.id " \
                    "INNER JOIN {schema}.sample smp ON ssp.sample_id=smp.id " \
                    "INNER JOIN {schema}.station sta ON smp.station_id=sta.id " \
                    "INNER JOIN {schema}.cruise cru ON sta.cruise_id=cru.id " \
                    "WHERE cru.name={cruise_name}").format(
                     schema=sql.Identifier(self.account_active_getSchema()),
                     cruise_name=sql.Literal(sCruise_name))
                cursor.execute(sqlCmd)
                sSqlQuery = cursor.query.decode("utf-8")
                dicSqlCmds["datasets_affected"] = sSqlQuery
                # Speichern
                lstContent = cursor.fetchall()
                try:
                    for elem in lstContent:
                        setBaseIngests.add(elem[0])
                        setAffectedDatasets.add(elem[1])
                except Exception as ex:
                    res.addErr("Getting datasets and ingest ids", add=str(ex), sys="Database", kat="Check Leaf Exsistance")

        # Ast Info String (mit allen Informationen)
        sLeaf_verbose = PrettyText.lst2str([sPartName + "{ing:" + str(iPartId) + "}" for (sPartName, iPartId) in (
            (sCruise_name, iCruise_ingestId), (sStation_name, iStation_ingestId), (sSample_name, iSample_ingestId),
            (sSubset_name, iSubset_IngestId)) if sPartName is not None], sSep)

        # Ergebnis
        res.addData("leaf_level", sLeaf_level)
        res.addData("leaf", sLeaf)
        res.addData("leaf_verbose", sLeaf_verbose)
        res.addData("cruise_id", iCruise_id)
        res.addData("station_id", iStation_id)
        res.addData("sample_id", iSample_id)
        res.addData("subsample_id", iSubSample_id)
        res.addData("subset_id", iSubset_id)
        res.addData("subset_names_used", lstSubsetNamesUsed)
        res.addData("datasets", setAffectedDatasets)
        res.addData("ingests", setBaseIngests)
        res.addData("cruise_ingest_id", iCruise_ingestId)
        res.addData("station_ingest_id", iStation_ingestId)
        res.addData("sample_ingest_id", iSample_ingestId)
        res.addData("subsample_ingest_id", iSubSample_ingestId)
        res.addData("subset_ingest_id", iSubset_IngestId)
        res.addData("sql", dicSqlCmds)
        return res

    def check_isRecordExistingInTable(self, sTab, sAttrNameSearch, dicSearch, bLazy=False, **kwargs):
        """
        Teste ob ein Datensatz bereits in einer Tabelle existiert

        Args:
            Table
            Suchatribut
            Search (dic)
            Lazy (bool)
        kwargs:
            cursor
        Return
            Result

        Entfernung der ID:
        Bei der DB-Antwort wird das Attribut id immer aus den Antwortdaten entfernt, ein vergleich damit ist ja sinnlos, da sich die ids immer unterscheiden
        Ausnahme: Es wird anhand der Id gesucht, dann wird diese in den Antwortdaten belassen

         Args:
            cursor (psycopg2.cursor.Cursor): Postgres Cursor
            sTab (str): Tabellenname
            name (str): Attributname nachdem gesucht werden soll (wenn es mehr als einen in db gibt nimm die id des ersten treffers)
            dicIn (dict): Testdatensatz mit Dictionary der Attributnamen und entsprechenden Werten
            bLazy (bool): True -  Es werden nur die Attribute des Testdatensatzes mit den entsprechenden eines Eintrages in der DB-Tabelle verglichen
                          False - Alle Attribute eines Eintrages der DB-Tabelle werden für den Vergleich herangezogen
        Returns:
            tools.coding.Result: Ergebnis
                search (dic): Nochmal die Suchanfrage ohne id zurückgeben
                found (bool): gefunden
                equal (bool): gefunden und alle einträge gleich
                answer (dic): Eintrag in DB
                diff: (set): Unterschiede
        """
        res = Result(name="Testing existance of entry in in table: "+str(sTab))

        cursor = kwargs["cursor"] if "cursor" in kwargs else None
        if cursor is None:
            resCursor = self.conn_getConnection()
            if resCursor.isErr():
                res.addResult(resCursor)
                return res
            conn = resCursor.getData("connection")
            cursor = resCursor.getData("cursor")


        # Bei Float-Werten Gleichheit bis zu einer Toleranzabweichung zulassen
        fTolerance = 0.001

        try:
            if sAttrNameSearch not in dicSearch:
                res.addErr("Search Problems", add="Search attribute ("+str(sAttrNameSearch) +") not found in search data: "+str(dicSearch), sys="Database", kat="Check Record Exsistance")
                return res

            sAttrName = sAttrNameSearch
            sAttrWert = dicSearch[sAttrNameSearch]

            resTableEntry = self.table_getTableEntries(sTab, sAttrName, sAttrWert, cursor=cursor)
            if resTableEntry.isErr():
                res.addResult(resTableEntry)
                raise NameError("Problems getting entry from Db", sys="Database", kat="Check Record Existance")
            lstHead = resTableEntry.getData("header")
            try:
                lstContent = resTableEntry.getData("content")[0]
            except:
                lstContent = None

                # Nichts gefunden
            if lstContent is None or len(lstContent) == 0:
                res.addData("found", False)
                res.addData("equal", False)
                return res
            else:
                res.addData("found", True)

            # Voller gefundener Datensatz aus der Datenbank
            dicDbFull = dict(zip(lstHead, lstContent))

            # TODO: Ist nicht sooo schön
            # Wenn Id nicht das Suchattribut ist -> Id des Datensatzes sichern und aus Daten entfernen
            if "id" in dicDbFull and sAttrName != "id":
                res.addData("id", dicDbFull["id"])
                dicDbFull.pop("id")

            # Im Lazy-Modus: Umfang des DB-Datensatzes auf die des Test-Datensatzes reduzieren
            if bLazy:
                dicDb = {k: v for k, v in dicDbFull.items() if k in dicSearch.keys()}
            else:
                dicDb = dicDbFull

            # Daten sichern
            res.addData("answer-full", dicDbFull)
            res.addData("answer", dicDb)

            # Harmonisiere die unterchiedlichen Ergebnisformate aus Suche und DB (alles wird String)
                # Search
            for k, v in dicSearch.items():
                if v is None:
                    # None -> ""
                    dicSearch[k] = ""
                else:
                    # Rest wird zu String
                    dicSearch[k] = str(v)
                # Db
            for k, v in dicDb.items():
                if v is None:
                    # None -> ""
                    dicDb[k] = ""
                else:
                    # Rest wird zu String
                    dicDb[k] = str(v)

            # Tolerantes Diff
            dicDiff = {}
            for kDb, vDb in dicDb.items():
                bUngleich = False
                if kDb not in dicSearch:
                    # KeyDb Nicht in Suchmenge
                    dicDiff[kDb] = vDb
                else:
                    # KeyDb in Suchmenge
                    vSearch = dicSearch[kDb]
                    if vDb != vSearch:
                        # ValDb ungleich
                        bUngleich = True
                        # *** Spezialfälle behandeln
                        # --- Float: Sind es floats und sind sie fast gleich?
                        try:
                            fvDb = float(vDb)
                            fvSearch = float(vSearch)
                            if math.isclose(fvDb,fvSearch,abs_tol=fTolerance):
                                # Fast gleich -> nicht als Ungleich markieren
                                bUngleich = False
                        except ValueError:
                            # Nein
                            pass
                        # --- Bool: Ziemlich gewagt: Jedes gefundene true/false wird als gleich zu 0/1 gewertet
                        if vDb.lower() in ["true","false"] or vSearch.lower() in ["true","false"]:
                            vDb = "True" if vDb.lower() in ["true", "1"] else vDb
                            vDb = "False" if vDb.lower() in ["false", "0"] else vDb
                            vSearch = "True" if vSearch.lower() in ["true", "1"] else vSearch
                            vSearch = "False" if vSearch.lower() in ["false", "0"] else vSearch
                            if vDb == vSearch:
                                bUngleich = False
                        if bUngleich:
                            # Wenn ungleich -> merken
                            if vDb in [None,""] and vSearch not in [None, ""]:
                                # Spezialmeldung: Achtung Fehlende Information in DB blockiert Information in Eingabe
                                pass
                                res.addWarn("Blocking Data", add="NULL-Value in DB for attribute "+str(kDb)+" is blocking input value: "+str(vSearch), sys="Database", kat="Check Record Exsistance")
                            dicDiff[kDb] = vDb

            setSearchErr = set(dicSearch.keys()) - set(dicDb.keys())

            # Suchfelder wurden nicht in DB gefunden
            if setSearchErr != set():
                res.addErr("Attribute not found", add=str(setSearchErr), sys="Database", kat="Check Record Exsistance")

            # Gleich oder nicht?
            if not len(dicDiff) != 0:
                res.addData("equal", True)
            else:
                res.addData("equal", False)
                res.addData("diff", dicDiff)

            # Suche nochmal zurückgeben
            res.addData("search", dicSearch)

        except Exception as ex:
            res.addErr("Unexcepted Error", add=str(ex), sys="Database", kat="Check Record Existance")
        return res

    def check_isSchemaExisting(self, **kwargs):
        """
        Check Datenbankverbindung
        Args:
            cursor: Cursor
        Returns:
            Result
        """
        res = Result("Check Schema")
        cursor = kwargs["cursor"] if "cursor" in kwargs else None
        sSchema = kwargs["schema"] if "schema" in kwargs else  self.account_single_getPart("schema")
        if cursor is None:
            resCursor = self.conn_getConnection()
            if resCursor.isErr():
                res.addResult(resCursor)
                return res
            conn = resCursor.getData("connection")
            cursor = resCursor.getData("cursor")
        try:
            cursor.execute("SELECT schema_name FROM information_schema.schemata WHERE schema_name = '" +sSchema + "'")
            if cursor.fetchone() is None:
                bExist = False
                raise NameError("not found")
            else:
                bExist = True
        except Exception as e:
            res.addErr("Schema not existing: "+ sSchema, add=str(e), sys="Database", kat="Check Schema")
            bExist = False
        res.addData("exist", bExist)
        return res

    def check_checkServerCompatibility(self):
        """
        Teste DAten Model auf Server
        Args:
        Kwargs:
        Return:
            Result:
        """
        # Evaluiere Versionsstring
        def evalVersion(sVersion):
            if sVersion is None:
                return None, None, None
            # Whitespaces entfernen
            sVersion = re.sub(r'[\ \t\n]', "", sVersion)
            # Teste auf unerlaubte Zeichen
            sNotAllowedChars = re.sub(r'[0-9.]', "", sVersion)
            if sNotAllowedChars != "":
                return None, None, None, None
            iPatch = None
            iMin = None
            iMaj = None
            try:
                # Check String-Format (major.minor.patch-build)
                erg = re.fullmatch(r'^[0-9]+\.[0-9]+\.[0-9]+', sVersion)
                if erg is not None:
                    # Extrahiere Einzelwerte
                    lst2 = sVersion.split(".")
                    iMaj = int(lst2[0])
                    iMin = int(lst2[1])
                    iPatch = int(lst2[2])
            except:
                pass
            return iMaj, iMin, iPatch

        # ---

        res = Result("Check Server Compatibility")
        status = self.__prnt.status

        sConnection = self.conn_getConnectionInfo()
        if self.account_single_isEmpty():
            res.addErr("Account is empty", kat="User Account", sys="Database")
            return res

        resCheck = self.check_isSchemaExisting()
        bCheckSchema = resCheck.getData("exist") if resCheck.isOk() else False
        if not bCheckSchema:
            res.addErr("Connection failed", add=sConnection, kat="Server Connection", sys="Database")
            return res

        # ---- CB Version
        # CB Schema: App
        sCbVersion_app = self.__model.conf_getInfo("dbschema")
        if sCbVersion_app is None:
            res.addErr("Db version string is corrupt", add=str(sCbVersion_app), kat="Check Server Compatibility", sys="Database")
            return res

        # CB Schema: DB
        sCbVersion_db = self.version_getVersionConnectedDatamodel().getData("version")
        if sCbVersion_db is None:
            res.addErr("Db version string is corrupt", add=str(sCbVersion_db), kat="Check Server Compatibility", sys="Database")
            return res

        # CB Schema: Versionskontrolle (App vs Db)
        sVersionWarn = None
        iMajApp, iMinApp, iPatchApp = evalVersion(sCbVersion_app)
        iMajDb, iMinDb, iPatchDb =  evalVersion(sCbVersion_db)
        if None in [iMajApp, iMinApp, iPatchApp]:
            res.addErr("App version string is corrupt", add=str(sCbVersion_app), kat="Check Server Compatibility", sys="Database")
            return res
        if None in [iMajDb, iMinDb, iPatchDb]:
            res.addErr("Db version string is corrupt", add=str(sCbVersion_db), kat="Check Server Compatibility", sys="Database")
            return res
        bAppGreater = True
        bDbGreater = True
        for iApp,iDb in ((iPatchApp,iPatchDb),(iMinApp,iMinDb),(iMajApp,iMajDb)):
            if iApp > iDb:
                bAppGreater = True
                bDbGreater = False
            elif iApp < iDb:
                bAppGreater = False
                bDbGreater = True
        if bDbGreater != bAppGreater:
            if bDbGreater:
                sAdd = "Database schema (" + sCbVersion_db + ") is newer than app schema (" + sCbVersion_app + ") -> Problems could arise"
                sVersionWarn = sAdd
                res.addWarn("DB Schema version conflict", add=sAdd, kat="Check Server Compatibility", sys="Database")
            elif bAppGreater:
                sAdd = "Database schema ("+sCbVersion_db+") is older than app schema ("+sCbVersion_app+") -> Problems could arise"
                sVersionWarn = sAdd
                res.addWarn("DB Schema version conflict", add=sAdd, kat="Check Server Compatibility", sys="Database")

        # --- Alle Attribute vergleichen
        # CB Attribute DB

        # Tabellen
        lstInit = [
            ("ingest", self.__model.subsystem_getMapper().ingest_getDbTables, self.__model.subsystem_getMapper().ingest_getDbAttributes),
            ("lookup", self.__model.subsystem_getMapper().lookup_getDbTables, self.__model.subsystem_getMapper().lookup_getDbAttributes),
            ("taxon", self.__model.subsystem_getMapper().taxon_getDbTables, self.__model.subsystem_getMapper().taxon_getDbAttributes),
            ("admin", self.__model.subsystem_getMapper().admin_getDbTables, self.__model.subsystem_getMapper().admin_getDbAttributes)
        ]
        iTablesTotal = 0
        iTable = 0
        odicTables = OrderedDict()
        for sSys, fktTable, fktAttrib in lstInit:
            dic = {}
            for sTable in fktTable():
                dic[sTable] = fktAttrib(sTable)
            odicTables[sSys] = dic
            iTablesTotal += 1

        # Check
        lstDbSurplus = []
        lstDbMissing = []
        lstDbDouble = []

        resCursor = self.conn_getConnection()
        if resCursor.isErr():
            res.addResult(resCursor)
            return res
        conn = resCursor.getData("connection")
        cursor = resCursor.getData("cursor")

        for sSys, dicSys in odicTables.items():
            for sTabApp, lstAttribApp in dicSys.items():
                resAttr = self.admin_getAttributNames(sTabApp, cursor=cursor)
                if resAttr.isErr():
                    res.addResult(resAttr, msg_only=True)
                else:
                    lstAttribDb = resAttr.getData("header")
                    if lstAttribDb is None:
                        res.addErr("Missing table in database", sheet=sTabApp, kat="Check Server Compatibility", sys="Database")
                    else:
                        # Entferne Einträge
                        # id
                        try:
                            lstAttribDb.remove("id")
                        except:
                            pass
                        # ingest_id
                        try:
                            lstAttribDb.remove("ingest_id")
                        except:
                            pass
                        resCmp = DataTools.cmpLists(lstAttribDb, lstAttribApp)
                        lstIstPlus = resCmp.getData("istPlus")
                        lstIstFehlt = resCmp.getData("istFehlt")
                        lstIstDoppelt = resCmp.getData("istDoppelt")
                        if lstIstFehlt is not None:
                            res.addErr("Missing attributes in Db table", sheet=sTabApp, add=PrettyText.lst2strC(lstIstFehlt), kat="Check Server Compatibility", sys="Database")
                            if sTabApp not in lstDbMissing:
                                lstDbMissing.append(sTabApp+": "+PrettyText.lst2strC(lstIstFehlt))
                        if lstIstDoppelt:
                            res.addErr("Double attributes in Db table", sheet=sTabApp, add=PrettyText.lst2strC(lstIstDoppelt), kat="Check Server Compatibility", sys="Database")
                            if sTabApp not in lstDbMissing:
                                lstDbMissing.append(sTabApp+": "+PrettyText.lst2strC(lstIstDoppelt))
                        if lstIstPlus:
                            res.addWarn("Surplus attributes in Db table", sheet=sTabApp, add=PrettyText.lst2strC(lstIstPlus), kat="Check Server Compatibility", sys="Database")
                            if sTabApp not in lstDbSurplus:
                                lstDbSurplus.append(sTabApp+ ": "+PrettyText.lst2strC(lstIstPlus))
                        iTable += 1
                        #progress(iTable, iTablesTotal, 50, proz=False, cli=False)
                        status("Checking table: " + str(sTable))
        sCbVersion_db = self.version_getVersionConnectedDatamodel().getData("version")

        # Antwort
        res.addData("cb_app", sCbVersion_app)
        res.addData("cb_db", sCbVersion_db)
        res.addData("cb_version_warn", sVersionWarn)

        if len(lstDbMissing) > 0:
            res.addData("missing", lstDbMissing)
        if len(lstDbDouble) > 0:
            res.addData("double", lstDbDouble)
        if len(lstDbSurplus) > 0:
            res.addData("surplus", lstDbSurplus)
        return res

    # --- Get Tech Info: Versionen

    def version_getVersionConnectedPostgresServer(self, **kwargs):
        """
        Hole Version des Postgres Servers
        Kwargs:
            cursor (psycopg2.cursor.Cursor): Postgres Cursor
        Returns:
            str: Version
        """
        res = Result("Get Version Connected to Postgres Server")
        cursor = kwargs["cursor"] if "cursor" in kwargs else None
        sVersion=None
        if cursor is None:
            resCursor = self.__model.subsystem_getDb().conn_getConnection()
            if resCursor.isErr():
                res.addResult(resCursor)
                return res
            conn = resCursor.getData("connection")
            cursor = resCursor.getData("cursor")
        sSqlCmd = "SHOW server_version;"
        cursor.execute(sSqlCmd)
        sVersion = cursor.fetchone()[0]  # Antwort
        sVersion = sVersion.split()[0] # Nimm wenn nötig und möglich nur den ersten Teil
        res.addData("version", str(sVersion))
        return res

    def version_getVersionConnectedDatamodel(self, **kwargs):
        """
        Hole Version des Schemas
        Kwargs:
            cursor (psycopg2.cursor.Cursor): Postgres Cursor
            schema (str): Schema
        Returns:
            str: Version
        """
        res = Result("Getting Version of Db Schema")
        cursor = kwargs["cursor"] if "cursor" in kwargs else None
        sSchema = kwargs["schema"] if "schema" in kwargs else self.account_active_getSchema()
        if cursor is None:
            resCursor = self.__model.subsystem_getDb().conn_getConnection()
            if resCursor.isErr():
                res.addResult(resCursor)
                return res
            conn = resCursor.getData("connection")
            cursor = resCursor.getData("cursor")
        try:
            sSqlCmd = "SELECT value FROM "+sSchema+".admin WHERE key='version';"
            cursor.execute(sSqlCmd)
            sVersion = cursor.fetchone()[0]  # Antwort
            res.addData("version", str(sVersion))
        except Exception as ex:
            res.addErr("Problems getting Version of connected Data model", kat="Version", sys="Database")
        return res

    def taxonwizard_getLastTaxonListUpdateInfo(self, **kwargs):
        """
        Hole alle DAten zum letzen TaxonWizard Taxon List Update
        Kwargs:
            cursor (psycopg2.cursor.Cursor): Postgres Cursor
            schema (str): Schema
        Returns:
            result
                date (str)
                err (int)
                warn (warn)
                log (str)
        """
        res = Result("Getting Data of last TaxonWizard Update")
        cursor = kwargs["cursor"] if "cursor" in kwargs else None
        sSchema = kwargs["schema"] if "schema" in kwargs else self.account_active_getSchema()
        if cursor is None:
            resCursor = self.__model.subsystem_getDb().conn_getConnection()
            if resCursor.isErr():
                res.addResult(resCursor)
                return res
            conn = resCursor.getData("connection")
            cursor = resCursor.getData("cursor")
        try:
            dateDate = None
            iErr = None
            iWarn = None
            sLog = None
            # Datum
            sSqlCmd = "SELECT value FROM "+sSchema+".admin WHERE key='taxon_wizard_last_update_date';"
            cursor.execute(sSqlCmd)
            tpAnswer = cursor.fetchone()
            if tpAnswer is not None:
                sDate = tpAnswer[0]
                dateDate = datetime.datetime.strptime(sDate, '%Y-%m-%d %H:%M:%S.%f')
            # Error
            sSqlCmd = "SELECT value FROM " + sSchema + ".admin WHERE key='taxon_wizard_last_update_err';"
            cursor.execute(sSqlCmd)
            tpAnswer = cursor.fetchone()
            if tpAnswer is not None:
                iErr = int(tpAnswer[0])
            # Warning
            sSqlCmd = "SELECT value FROM " + sSchema + ".admin WHERE key='taxon_wizard_last_update_warn';"
            cursor.execute(sSqlCmd)
            tpAnswer = cursor.fetchone()
            if tpAnswer is not None:
                iWarn = int(tpAnswer[0])
            # Log
            sSqlCmd = "SELECT value FROM " + sSchema + ".admin WHERE key='taxon_wizard_last_update_log';"
            cursor.execute(sSqlCmd)
            tpAnswer = cursor.fetchone()
            if tpAnswer is not None:
                sLog = str(tpAnswer[0])
            # Anwort
            res.addData("date", dateDate)
            res.addData("err", iErr)
            res.addData("warn", iWarn)
            res.addData("log", sLog)
        except Exception as ex:
            res.addErr("Problems getting Data of last TaxonWizard Update", add=str(ex), kat="TaxonWizard Admin", sys="Database")
        return res

    def dataPatrol_biotaControl_getAllBiota(self, **kwargs):
        """
        Hole alle Daten für die Biota Kontrolle
        Kwargs:
            cursor (psycopg2.cursor.Cursor): Postgres Cursor
            schema (str): Schema
        Returns:
            result
                biota (lst)
        """
        res = Result("Getting Data of BiotaControl")
        cursor = kwargs["cursor"] if "cursor" in kwargs else None
        sSchema = kwargs["schema"] if "schema" in kwargs else self.account_active_getSchema()
        if cursor is None:
            resCursor = self.__model.subsystem_getDb().conn_getConnection()
            if resCursor.isErr():
                res.addResult(resCursor)
                return res
            conn = resCursor.getData("connection")
            cursor = resCursor.getData("cursor")
        try:
            sSql = """
select distinct tax.aid, given_taxon_name, dat.name, ing.file 
from {schema}.biota bio 
join {schema}.taxon tax on bio.taxon_id=tax.id 
join {schema}.subset sub on bio.subset_id=sub.id 
join {schema}.dataset dat on sub.dataset_id=dat.id 
join {schema}.ingest ing on bio.ingest_id=ing.id 
"""
            sSqlCmd = sql.SQL(sSql).format(
                schema=sql.Identifier(sSchema)
            )
            cursor.execute(sSqlCmd)
            sSqlQuery = cursor.query.decode("utf-8")
            lstBiota = cursor.fetchall()
            cursor.execute(sSqlCmd)
            lstBiota = cursor.fetchall()
            res.addData("biota", lstBiota)
        except Exception as ex:
            res.addErr("Problems getting Data for BiotaControl",add=str(ex), kat="BiotaControl", sys="Database")
        return res

    def dataPatrol_getLastBiotaControl(self, **kwargs):
        """
        Hole alle Daten zur letzen Biota Kontrolle
        Kwargs:
            cursor (psycopg2.cursor.Cursor): Postgres Cursor
            schema (str): Schema
        Returns:
            result
                date (str)
                err (int)
                warn (warn)
                log (str)
        """
        res = Result("Getting Data of BiotaControl")
        cursor = kwargs["cursor"] if "cursor" in kwargs else None
        sSchema = kwargs["schema"] if "schema" in kwargs else self.account_active_getSchema()
        if cursor is None:
            resCursor = self.__model.subsystem_getDb().conn_getConnection()
            if resCursor.isErr():
                res.addResult(resCursor)
                return res
            conn = resCursor.getData("connection")
            cursor = resCursor.getData("cursor")
        try:
            dateDate = None
            iErr = None
            iWarn = None
            sLog = None
            # Datum
            sSqlCmd = "SELECT value FROM "+sSchema+".admin WHERE key='last_biota_control_date'"
            cursor.execute(sSqlCmd)
            tpAnswer = cursor.fetchone()
            if tpAnswer is not None:
                sDate = str(tpAnswer[0])
                dateDate = datetime.datetime.strptime(sDate, '%Y-%m-%d %H:%M:%S.%f')
            # Errors
            sSqlCmd = "SELECT value FROM " + sSchema + ".admin WHERE key='last_biota_control_err'"
            cursor.execute(sSqlCmd)
            tpAnswer = cursor.fetchone()
            if tpAnswer is not None:
                iErr = int(tpAnswer[0])
            # Warnings
            sSqlCmd = "SELECT value FROM " + sSchema + ".admin WHERE key='last_biota_control_warn'"
            cursor.execute(sSqlCmd)
            tpAnswer = cursor.fetchone()
            if tpAnswer is not None:
                iWarn = int(tpAnswer[0])
            # Log
            sSqlCmd = "SELECT value FROM " + sSchema + ".admin WHERE key='last_biota_control_log'"
            cursor.execute(sSqlCmd)
            tpAnswer = cursor.fetchone()
            if tpAnswer is not None:
                sLog = str(tpAnswer[0])
            # Antwort
            res.addData("date", dateDate)
            res.addData("err", iErr)
            res.addData("warn", iWarn)
            res.addData("log", sLog)
        except Exception as ex:
            res.addErr("Problems getting Data of last BiotaControl", add=str(ex), kat="BiotaControl", sys="Database")
        return res

    # --- PostGis FAO
    def fao_regions_getShape(self, sRegion, **kwargs):
        """
        Hole eine Geometry für eine Region
        kwargs
            schema (str): Schema
        """
        sSchema = kwargs["schema"] if "schema" in kwargs else self.account_active_getSchema()
        res = Result("Getting shape for FAO regions")
        resCursor = self.__model.subsystem_getDb().conn_getConnection()
        if resCursor.isErr():
            res.addResult(resCursor)
            return res
        conn = resCursor.getData("connection")
        cursor = resCursor.getData("cursor")
        try:
            sSql = "select ST_AsGeoJSON(geometry) from {schema}.region where name={region} "
            sqlCmd = sql.SQL(sSql).format(
                schema=sql.Identifier(self.__model.subsystem_getDb().account_active_getSchema()),
                region=sql.Literal(sRegion),
            )
            cursor.execute(sqlCmd)
            sSqlQuery = cursor.query.decode("utf-8")
            lstContent = cursor.fetchall()
            try:
                sShape = lstContent[0][0]
                iShape = len(sShape)
            except:
                sShape = None
                iShape = 0
            res.addData("sql", sSqlQuery)
            res.addData("shape", sShape)
            res.addData("shape-size", iShape)
        except Exception as ex:
            res.addErr("Unexpected error", add=str(ex), sys="Database", kat="FAO Harvest")
        return res

    def fao_regions_getRegions(self, **kwargs):
        """
        Hole alle Regions
        kwargs
            schema (str): Schema
        return:
            result
                sql (str)
                level (set)
                level-nr (int)
                regions (odic)
                regions-nr (int)
        """
        sSchema = kwargs["schema"] if "schema" in kwargs else self.account_active_getSchema()
        res = Result("Getting all FAO regions")
        resCursor = self.__model.subsystem_getDb().conn_getConnection()
        if resCursor.isErr():
            res.addResult(resCursor)
            return res
        conn = resCursor.getData("connection")
        cursor = resCursor.getData("cursor")
        try:
            sSql = "select distinct id, code, name, level, ST_NPoints(geometry) as points from {schema}.region order by points desc, name asc"
            sqlCmd = sql.SQL(sSql).format(schema=sql.Identifier(sSchema))
            cursor.execute(sqlCmd)
            sSqlQuery = cursor.query.decode("utf-8")
            lstContent = cursor.fetchall()
            odicRegions = OrderedDict()
            setLevels = set()
            try:
                for id, code, name, level, points in lstContent:
                    if level is not None:
                        setLevels.add(level)
                    odicRegions[name] = (id, code, level, points)
            except:
                lstLevel = []
            odicRegions = OrderedDict(sorted(odicRegions.items(), key=lambda x: x[0]))
            res.addData("sql", sSqlQuery)
            res.addData("level", setLevels)
            res.addData("level-nr", len(setLevels))
            res.addData("regions", odicRegions)
            res.addData("regions-nr", len(odicRegions))
        except Exception as ex:
            res.addErr("Unexpected error", add=str(ex), sys="Database", kat="FAO Harvest")
        return res

    def fao_regions_imExport(self, sFile, sDirection, **kwargs):
        """
        Im-/Export von FAO Daten von/in Datei
        Args:
            file (str)
            direction (str): import/export
        Kwargs:
            file (str)
        """
        pn = self.__prnt.normal
        pv = self.__prnt.verbose
        progress = self.__prnt.progress
        workflow = self.__prnt.workflow
        status = self.__prnt.status

        # Account
        sSchema = self.account_single_getPart("schema")
        sPw = self.account_single_getPart("pw")
        sUser = self.account_single_getPart("user")
        sDb = self.account_single_getPart("db")
        sHost = self.account_single_getPart("host")
        sPort = self.account_single_getPart("port")

        progress("on-pulse")
        workflow(top="FAO Features", sub="Export to file", fmt=False)
        res = Result("Export FAO regions to fike")
        if sDirection not in ["import", "export"]:
            res.addErr("Intern: direction unknown")
            return res
        resCursor = self.__model.subsystem_getDb().conn_getConnection()
        if resCursor.isErr():
            res.addResult(resCursor)
            return res
        conn = resCursor.getData("connection")
        cursor = resCursor.getData("cursor")

        # Lösche alle Regionen
        if sDirection == "import":
            try:
                pn("* Deleting FAO regions in CB", fmt=False)
                workflow(top="FAO regions", sub="Deleting regions in CB", fmt=False)
                sSql = "DELETE FROM {schema}.region"
                sqlCmd = sql.SQL(sSql).format(schema=sql.Identifier(sSchema))
                cursor.execute(sqlCmd)
                conn.commit()
            except Exception as ex:
                res.addErr("Problems deleting regions", add=str(ex), kat="FAO regions", sys="Database")
                return res

        # PSQL-Einstellungen
        sPsqlBin = "psql" if self.__model.conf_getAdminPsqlPath() in [None,""] else self.__model.conf_getAdminPsqlPath()
        dicEnv = os.environ
        dicEnv["PGPASSWORD"] = self.account_single_getPart("pw")
        sPgConsoleLogin = "-U " + sUser + " -d " + sDb + " -h " + sHost + " -p " + sPort
        sPgConsoleCmd = sPsqlBin + " " + sPgConsoleLogin + " -c"
        lstPgConsoleCmd = sPgConsoleCmd.split()
        if sDirection == "export":
            # FAO: Export to file
            workflow(top="FAO regions", sub="Export to file", fmt=False)
            pn("* Exporting FAO regions to file", fmt=False)
            sSql = "\copy (SELECT * FROM <schema>.region) to <file> with csv"
        else:
            # FAO: Import from file
            workflow(top="FAO regions", sub="Import from file", fmt=False)
            pn("* Importing FAO regions from file", fmt=False)
            sSql = "\copy <schema>.region from <file> with csv"
        sSql = sSql.replace("<schema>", sSchema)
        sSql = sSql.replace("<file>", sFile)
        lstPgConsoleCmd.append(sSql)
        try:
            # Windows-only: Startup-Info -> verändern, damit die Konsole verborgen bleibt (würde auch über shell=True funktionieren)
            if self.__model.conf_getOs() == "win":
                startup_info = subprocess.STARTUPINFO()
                startup_info.dwFlags |= subprocess.STARTF_USESHOWWINDOW
            else:
                startup_info = None
            sPgConsoleCmd = " ".join(lstPgConsoleCmd)
            # Warten mit Prozess: StdOut Realtime Ausgabe
            p = subprocess.Popen(lstPgConsoleCmd, stdout=subprocess.PIPE, stderr=subprocess.STDOUT,
                                 startupinfo=startup_info, env=dicEnv)
            sStdOut = ""
            sStdErr = ""
            iStdErr = 0
            for byteLine in p.stdout:
                sLine = str(byteLine.decode("utf-8", errors="ignore"))
                pn(sLine, "")
                if "error" in sLine.lower() or "fail" in sLine.lower():
                    sStdErr += sLine
                    iStdErr += 1
                sStdOut += sLine
            # Command error
            if p.returncode not in [0, None]:
                raise NameError("Error executing SQL command")
            if sDirection == "import":
                sLog = "Imported FAO regions from file: "+str(sFile)+"\n"
            else:
                sLog = "Exported FAO regions to file: "+str(sFile)+"\n"
            res.addData("log", sLog)
        except Exception as ex:
            res.addErr("Problems exporting regions", add=str(ex), kat="FAO regions", sys="Database")
        if res.isOk():
            conn.commit()
            conn.close()
        return res

    def fao_regions_create(self, **kwargs):
        """
        Hole alle FAO Daten und update die Tabelle region
        Kwargs:
            schema (str): Schema
            leftover (str): Override Leftover name (Name, Code:Name.lower())
            level (str/lst): Levels to harvest
        """

        # Helper
        def storeNonFao(cursor, sSchema, sLeftover_name):
            sSql = "INSERT INTO {schema}.region (code, name, level, geometry) values({leftover_key}, {leftover_value}, NULL, NULL);"
            sqlCmd = sql.SQL(sSql).format(
                schema=sql.Identifier(sSchema),
                leftover_key=sql.Literal(sLeftover_name.lower()),
                leftover_value=sql.Literal(sLeftover_name)
            )
            cursor.execute(sqlCmd)

        # ---

        pn = self.__prnt.normal
        pv = self.__prnt.verbose
        progress = self.__prnt.progress
        workflow = self.__prnt.workflow
        status = self.__prnt.status
        sSchema = kwargs["schema"] if "schema" in kwargs else self.account_active_getSchema()
        sLeftover_name_std = self.__model.subsystem_getRegions().conf_getLeftover_name()
        sLeftover_name = kwargs["leftover"] if "leftover" in kwargs else sLeftover_name_std
        bMirror = kwargs.get("mirror")
        if bMirror is None:
            bMirror = False

        # Log
        lstRegions_log_table = []

        # Level
        lstLevel_in = kwargs["level"] if "level" in kwargs else None
        lstLevel_std = ['major', 'subarea']
        if lstLevel_in in ["all", "*"]:
            lstLevel = None
        elif lstLevel_in is None:
            lstLevel = lstLevel_std
        elif type(lstLevel_in) is str:
            lstLevel = [lstLevel_in]
        else:
            lstLevel = lstLevel_in
        progress("on-pulse")
        workflow(top="FAO Features", sub="Retrieving data", fmt=False)
        res = Result("Getting FAO Data")

        resCursor = self.__model.subsystem_getDb().conn_getConnection()
        if resCursor.isErr():
            res.addResult(resCursor)
            return res
        conn = resCursor.getData("connection")
        cursor = resCursor.getData("cursor")

        # Lösche alle Regionen
        try:
            pn("* Deleting FAO regions in CB", fmt=False)
            workflow(top="FAO regions", sub="Deleting regions in CB", fmt=False)
            sSql = "DELETE FROM {schema}.region"
            sqlCmd = sql.SQL(sSql).format(schema=sql.Identifier(sSchema))
            cursor.execute(sqlCmd)
        except Exception as ex:
            res.addErr("Problems deleting regions", add=str(ex), kat="FAO regions", sys="Database")

        # Hole alle FAO Definitonen
        if res.isOk():
            sFao = ""
            try:
                pn("* Harvesting all regions from fao.org", fmt=False)
                workflow(top="FAO regions", sub="Harvesting fao.org", fmt=False)
                if bMirror:
                    # Benutze AWI mirror
                    pn("* Using AWI mirror", fmt=False)
                    workflow(top="FAO regions", sub="Download from AWI mirror", fmt=False)

                    # FAO-Mirror
                    sFAO_mirror_zipped = self.__model.conf_getFAO_mirror_zipped()
                    sFao_mirror_unzipped = self.__model.conf_getFAO_mirror_unzipped()
                    sFAO_mirror_url = self.__model.conf_getUpdateServer() + "/" + sFAO_mirror_zipped

                    # Tmp
                    pn("Mirror: " + str(sFAO_mirror_url), fmt=False)
                    sPath_tmp = self.__model.conf_getAppTmpDir()
                    pn("Saving to: " + str(sPath_tmp), fmt=False)
                    sFAO_tmp_zipped = str(Path(sPath_tmp, "fao.zip"))
                    sFAO_tmp_unzipped = str(Path(sPath_tmp, "fao.txt"))
                    res = Result()
                    try:
                        # Download fao.zip
                        r = requests.get(sFAO_mirror_url, allow_redirects=True, timeout=1,
                                         auth=('critter', 'P0lych4eta'))  # TWEAK für Preview-Version
                        if r.status_code == 200 and r.headers['Content-Type'] == 'application/zip':
                            with open(sFAO_tmp_zipped, 'wb') as f:
                                f.write(r.content)
                            res.addData("length", int(r.headers['Content-Length']))
                        else:
                            res.addErr("Download failed: Server reachable but data not available - are you connectred to AWI intranet?", sys="Update",
                                       kat="Download")
                    except Exception as ex:
                        res.addErr("Download failed: " + str(ex))
                    if res.isOk():
                        # Download: ok
                        # statinfo = os.stat(sFAO_tmp_zipped)
                        # dt = datetime.datetime.fromtimestamp(statinfo.st_ctime)
                        # dt_delta = datetime.datetime.now() - dt
                        # sDate = str(dt.year) + "-" + str(dt.month) + "-" + str(dt.day)
                        # pn("Mirror created on: "+sDate)
                        # pn("Mirror age: "+str(dt_delta.days)+" days")
                        pn("Unzip file: " + sFAO_mirror_zipped + " -> " +sFao_mirror_unzipped)
                        # Unzip: fao.zip
                        zipfile.ZipFile(sFAO_tmp_zipped).extract(sFao_mirror_unzipped, path=sPath_tmp)
                        with open(sFAO_tmp_unzipped, "r", encoding='utf-8') as f:
                            sFao = f.read()
                else:
                    # Harvest from FAO.org
                    workflow(top="FAO regions", sub="Harvesting fao.org", fmt=False)
                    r = requests.get(self.__model.conf_getFAO_server())
                    sFao = r.text

                # FAO: JSON -> dic
                dicFao = json.loads(sFao)

            except Exception as ex:
                # Fehler bei FAO-Download (Server down)
                sLog = "************************************\n"
                sLog += "Create new FAO regions\n"
                sLog += "************************************\n\n"
                sLog += "FAO Server is down or has problems!\n"
                sLog += "Workaround activated: Only "+str(sLeftover_name)+" category has been made\n"
                sLog += "The server answered this:\n\n"
                sLog += "------------------------------------\n\n"
                sLog += sFao[:10000] +"\n ... truncated ...\n"
                sLog += str(ex)

                # TWEAK: Mach das System zumindest lauffähig
                res.addData("fao-msg",sFao)
                res.addData("log",sLog)
                res.addWarn("Problems harvesting regions", add=str(ex), kat="FAO regions", sys="Database")
                # Speicher zumindest Non-FAO
                try:
                    storeNonFao(cursor,sSchema, sLeftover_name)
                except Exception as ex:
                    res.addErr("Problems Writing non-FAO category to CB", add=str(ex), kat="FAO regions", sys="Database")
                workflow()
                progress()
                conn.commit()
                conn.close()
                return res

        # Speicher alle Regionen in CB
        if res.isOk():
            try:
                pn("* Importing regions to CB", fmt=False)
                workflow(top="FAO regions", sub="Importing regions to CB", fmt=False)
                if lstLevel_in == "all":
                    pn("Import all possible levels ")
                elif lstLevel_in is None:
                    pn("Import only this std levels: " + PrettyText.lst2strC(lstLevel_std))
                else:
                    pn("Import only this levels: " + PrettyText.lst2strC(lstLevel_in))
                iNr_max = 0
                iNr = 0
                for feature in dicFao['features']:
                    properties = feature['properties']
                    sLevel = properties['F_LEVEL'].lower()
                    sName = properties['NAME_EN']
                    if sName is not None:
                        if lstLevel is None:
                            iNr_max += 1
                        elif sLevel in lstLevel:
                            iNr_max += 1
                for feature in dicFao['features']:
                    progress(iNr, iNr_max, proz=True, abs=True, cli=True)
                    properties = feature['properties']
                    iCode = properties['F_CODE']
                    sName = properties['NAME_EN']
                    sLevel = properties['F_LEVEL'].lower()
                    geom = feature['geometry']
                    geometry = json.dumps(geom)
                    if lstLevel is not None and sLevel not in lstLevel:
                        continue
                    if not sName:
                        continue
                    iNr += 1
                    pn(str(sLevel) + ": "+str(sName), fmt=False)
                    status(sName)
                    lstRegions_log_table.append((sName, sLevel, iCode))
                    sSql = "INSERT INTO {schema}.region (code, name, level, geometry) VALUES ({code}, {name}, {level}, ST_AsText(ST_GeomFromGeoJSON({geometry})))"
                    sqlCmd = sql.SQL(sSql).format(
                        schema=sql.Identifier(sSchema),
                        code=sql.Literal(iCode),
                        name=sql.Literal(sName),
                        level=sql.Literal(sLevel),
                        geometry=sql.Literal(geometry)
                    )
                    #sSql_final = sqlCmd.as_string(cursor)
                    cursor.execute(sqlCmd)
                # No Region
                storeNonFao(cursor,sSchema, sLeftover_name)
                pn("* Finished")
                pn("Created regions: "+str(iNr))
            except Exception as ex:
                res.addErr("Problems putting FAO regions to CB", add=str(ex), kat="FAO regions", sys="Database")

        # ***  Log
        sLog = "************************************\n"
        sLog += "Create new FAO regions\n"
        sLog += "************************************\n\n"
        sLog += "Created FAO regions: " + str(len(lstRegions_log_table)) + " x\n"
        if len(lstRegions_log_table) > 0:
            lstRegions_log_table.insert(0, ("Region", "Level", "Code"))
            sLog += PrettyText.table(lstRegions_log_table)
            res.addData("log", sLog)

        # Alles OK
        if res.isOk():
            # *** DB: Commit
            conn.commit()
            conn.close()

        # *** Action Log
        if not self.__model.conf_isDry() and res.isOk():
            resActionLog = self.__model.subsystem_getDb().actionlog_new("fao-create", "regions", sLog)
            res.addResult(resActionLog)

        workflow()
        progress()
        return res

    def fao_map_clean(self, **kwargs):
        """
        Entfernen Samples aus FAO-Zuordnung
        Kwargs:
            schema (str): Schema
            samples (lst): Sample Ids
        """
        pn = self.__prnt.normal
        pv = self.__prnt.verbose
        progress = self.__prnt.progress
        workflow = self.__prnt.workflow
        status = self.__prnt.status

        sSchema = kwargs["schema"] if "schema" in kwargs else self.account_active_getSchema()
        lstSamples = kwargs["samples"] if "samples" in kwargs else None
        iIng = kwargs["ing"] if "ing" in kwargs else None

        progress("on-pulse")
        workflow(top="FAO Regions", sub="Clean mapping", fmt=False)
        res = Result("Remove samples from regions")
        resCursor = self.__model.subsystem_getDb().conn_getConnection()
        if resCursor.isErr():
            res.addResult(resCursor)
            return res
        conn = resCursor.getData("connection")
        cursor = resCursor.getData("cursor")

        # *** Enferne Samples
        pv("* Remove samples from mapping to FAO regions", fmt=False)
        try:
            workflow(top="FAO regions", sub="Remove samples", fmt=False)
            if iIng is not None:
                pv("Remove samples of specific ingest from mapping to FAO regions", fmt=False)
                pv("ingest: "+str(iIng), fmt=False)
                sSql = "WITH deleted AS (DELETE FROM {schema}.region2sample WHERE sample_id IN (select id from {schema}.sample where ingest_id = {ing}) RETURNING *) SELECT count(*) FROM deleted"
                sqlCmd = sql.SQL(sSql).format(
                    schema=sql.Identifier(sSchema),
                    ing=sql.Literal(iIng)
                )
            elif lstSamples is None:
                pv("Remove all mapped samples", fmt=False)
                # Alle Samples
                # Liste von Samples
                sSql = "WITH deleted AS (DELETE FROM {schema}.region2sample RETURNING *) SELECT count(*) FROM deleted"
                sqlCmd = sql.SQL(sSql).format(
                    schema=sql.Identifier(sSchema)
                )
            else:
                pv("* Remove samples from mapping to FAO regions", fmt=False)
                pv("Remove subset of samples", fmt=False)
                sSql = "WITH deleted AS (DELETE FROM {schema}.region2sample where sample_id IN ({ids}) RETURNING *) SELECT count(*) FROM deleted"
                sqlCmd = sql.SQL(sSql).format(
                    schema=sql.Identifier(sSchema),
                    ids=sql.SQL(', ').join(map(sql.Literal, lstSamples))
                )
            cursor.execute(sqlCmd)
            sSqlQuery = cursor.query.decode("utf-8")
            iDel = cursor.fetchone()[0]  # Antwort
            pv("removed samples: "+str(iDel))
            res.addData("sql", sSqlQuery)
            res.addData("del", iDel)
        except Exception as ex:
            res.addErr("Problems removing samples from mapping to FAO regions", add=str(ex), kat="FAO regions", sys="Database")
        if res.isOk():
            conn.commit()
            conn.close()

        # *** Action Log
        if iDel > 0:
            sLog = "Removed samples: " + str(iDel) + "\n"
            if not self.__model.conf_isDry() and res.isOk():
                resActionLog = self.__model.subsystem_getDb().actionlog_new("fao-del", "map", sLog)
                res.addResult(resActionLog)

        return res

    def fao_map_getData(self, **kwargs):
        """
        Hole Statistik zu FAO Mapping
        Kwargs:
            schema (str): Schema
            leftover (str): Override Leftover name (Name, Code:Name.lower())
        """

        pn = self.__prnt.normal
        pv = self.__prnt.verbose
        sLeftover_name_std = self.__model.subsystem_getRegions().conf_getLeftover_name()

        sSchema = kwargs["schema"] if "schema" in kwargs else self.account_active_getSchema()
        sLeftover_name = kwargs["leftover"] if "leftover" in kwargs else sLeftover_name_std
        bShort = kwargs["short"] if "short" in kwargs else False

        res = Result("FAO stats")
        resCursor = self.__model.subsystem_getDb().conn_getConnection()
        if resCursor.isErr():
            res.addResult(resCursor)
            return res
        conn = resCursor.getData("connection")
        cursor = resCursor.getData("cursor")

        # Daten
        iNew = 0
        iLeftover = 0
        iRegions = 0

        # Log (Basis)
        sLog = ""
        sLog_leftover = ""
        sLog_new = ""
        sLog_region_overview = ""
        sLog_region_datasets = ""
        sLog_region = ""

        # *** Hole Anzahl aller Samples
        try:
            sSql = "SELECT count(id) FROM {schema}.sample smp"
            sqlCmd = sql.SQL(sSql).format(
                schema=sql.Identifier(sSchema)
            )
            cursor.execute(sqlCmd)
            sSqlQuery = cursor.query.decode("utf-8")
            iSmp = cursor.fetchone()[0]
            res.addData("sample_nr", iSmp)
        except Exception as ex:
            res.addErr("Problems getting new unmapped samples", add=str(ex), kat="FAO regions", sys="Database")

        # *** Hole Anzahl aller Leftover
        try:
            sSql = \
"""
SELECT count(DISTINCT reg2smp.sample_id) 
FROM {schema}.region2sample reg2smp 
WHERE reg2smp.region_id = (SELECT id FROM {schema}.region WHERE code={leftover_code})
"""
            sqlCmd = sql.SQL(sSql).format(
                schema=sql.Identifier(sSchema),
                leftover_code=sql.Literal(sLeftover_name.lower())
            )
            cursor.execute(sqlCmd)
            sSqlQuery = cursor.query.decode("utf-8")
            iLeftover = cursor.fetchone()[0]
            res.addData("leftover_nr", iLeftover)
        except Exception as ex:
            res.addErr("Problems getting new leftover samples", add=str(ex), kat="FAO mapping", sys="Database")

        # *** Hole Anzahl aller noch nicht verteilten Samples
        try:
            sSql = \
"""
SELECT count(DISTINCT smp.id) 
FROM {schema}.sample smp 
WHERE smp.id NOT IN (SELECT DISTINCT sample_id FROM {schema}.region2sample)
"""
            sqlCmd = sql.SQL(sSql).format(
                schema=sql.Identifier(sSchema)
            )
            cursor.execute(sqlCmd)
            sSqlQuery = cursor.query.decode("utf-8")
            iNew = cursor.fetchone()[0]
            res.addData("unmapped_nr", iNew)
        except Exception as ex:
            res.addErr("Problems getting new unmapped samples", add=str(ex), kat="FAO mapping", sys="Database")

        # *** Hole Datasets mit allen noch nicht verteilten Samples
        if not bShort:
            try:
                sSql = \
"""
SELECT dat.name, count(DISTINCT smp.id) 
FROM {schema}.subset sub 
JOIN {schema}.subsample ssmp ON ssmp.id=sub.subsample_id 
JOIN {schema}.dataset dat ON dat.id=sub.dataset_id
JOIN {schema}.sample smp ON smp.id=ssmp.sample_id 
WHERE smp.id NOT IN (select DISTINCT sample_id FROM {schema}.region2sample)
GROUP BY dat.name
"""
                sqlCmd = sql.SQL(sSql).format(
                    schema=sql.Identifier(sSchema)
                )
                cursor.execute(sqlCmd)
                sSqlQuery = cursor.query.decode("utf-8")
                lstNew = cursor.fetchall()
                res.addData("unmapped", lstNew)
            except Exception as ex:
                res.addErr("Problems getting new unmapped samples", add=str(ex), kat="FAO regions", sys="Database")

        # *** Hole Region/Dataset Anzahl aller verteilten Samples pro Dataset
        try:
            sSql = \
"""
select reg.id "id", reg.code as "code",  reg.level as "level", reg.name as "region", ST_NPoints(geometry) as "points", dat.name as "dataset", count(smp.lat_start) as "count" 
from {schema}.subset sub 
join {schema}.subsample ssmp on ssmp.id=sub.subsample_id 
join {schema}.sample smp on smp.id=ssmp.sample_id 
join {schema}.region2sample r2s on r2s.sample_id=smp.id
join {schema}.region reg on reg.id=r2s.region_id
join {schema}.dataset dat on dat.id=sub.dataset_id
group by reg.id, reg.name, dat.name
order by points desc, reg.name asc, dat.name asc
"""
            sqlCmd = sql.SQL(sSql).format(
                schema=sql.Identifier(sSchema)
            )
            cursor.execute(sqlCmd)
            sSqlQuery = cursor.query.decode("utf-8")
            lstContent = cursor.fetchall()

            dicReg_name2id = {}
            dicReg = OrderedDict()
            setReg_names = set()
            for iId, sCode, sLevel, sReg, iPoints, sDat, iCnt in lstContent:
                dicReg_name2id[sReg] = iId
                setReg_names.add(sReg)
                if iId not in dicReg:
                    dicReg[iId] = (sReg, sCode, sLevel, iPoints, [(sDat, iCnt)], iCnt)
                else:
                    sReg_tmp, sCode_tmp, sLevel_tmp, iPoints_tmp, lstDat_tmp, iCnt_tmp = dicReg[iId]
                    lstDat_tmp.append((sDat, iCnt))
                    iCnt_tmp += iCnt
                    dicReg[iId] = (sReg_tmp, sCode_tmp, sLevel_tmp, iPoints_tmp, lstDat_tmp, iCnt_tmp)
            res.addData("region_dic_name2id", dicReg_name2id)
            res.addData("region_dic", dicReg)
        except Exception as ex:
            res.addErr("Problems getting region statistics", add=str(ex), kat="FAO regions", sys="Database")

        if res.isOk():
            conn.commit()
            conn.close()
        return res

    def fao_mapping_stats_old(self, **kwargs):
        """
        Hole Statistik zu FAO Mapping
        Kwargs:
            schema (str): Schema
            short (bool): Short
            leftover (str): Override Leftover name (Name, Code:Name.lower())
        """

        pn = self.__prnt.normal
        pv = self.__prnt.verbose
        progress = self.__prnt.progress
        workflow = self.__prnt.workflow
        status = self.__prnt.status
        sLeftover_name_std = self.__model.subsystem_getRegions().conf_getLeftover_name()

        sSchema = kwargs["schema"] if "schema" in kwargs else self.account_active_getSchema()
        sLeftover_name = kwargs["leftover"] if "leftover" in kwargs else sLeftover_name_std
        bShort = kwargs["short"] if "short" in kwargs else False

        progress("on-pulse")
        workflow(top="FAO Regions", sub="Get statistics", fmt=False)
        res = Result("FAO stats")
        resCursor = self.__model.subsystem_getDb().conn_getConnection()
        if resCursor.isErr():
            res.addResult(resCursor)
            return res
        conn = resCursor.getData("connection")
        cursor = resCursor.getData("cursor")

        # Daten
        iNew = 0
        iLeftover = 0
        iRegions = 0

        # Log (Basis)
        sLog = ""
        sLog_leftover = ""
        sLog_new = ""
        sLog_region_overview = ""
        sLog_region_datasets = ""
        sLog_region = ""

        # *** Hole Anzahl aller Samples
        try:
            sSql = "SELECT count(id) FROM {schema}.sample smp"
            sqlCmd = sql.SQL(sSql).format(
                schema=sql.Identifier(sSchema)
            )
            cursor.execute(sqlCmd)
            sSqlQuery = cursor.query.decode("utf-8")
            iSmp = cursor.fetchone()[0]
            res.addData("sample_nr", iSmp)
        except Exception as ex:
            res.addErr("Problems getting new unmapped samples", add=str(ex), kat="FAO regions", sys="Database")

        # *** Hole Anzahl aller noch nicht verteilten Samples
        sSql = \
"""
SELECT count(DISTINCT smp.id) 
FROM {schema}.sample smp 
WHERE smp.id NOT IN (SELECT DISTINCT sample_id FROM {schema}.region2sample)
"""
        sqlCmd = sql.SQL(sSql).format(
            schema=sql.Identifier(sSchema)
        )
        cursor.execute(sqlCmd)
        sSqlQuery = cursor.query.decode("utf-8")
        iNew = cursor.fetchone()[0]
        res.addData("new_nr", iNew)

        # *** Hole Datasets mit allen noch nicht verteilten Samples
        if not bShort:
            try:
                sSql = \
"""
SELECT dat.name, count(DISTINCT smp.id) 
FROM {schema}.subset sub 
JOIN {schema}.subsample ssmp ON ssmp.id=sub.subsample_id 
JOIN {schema}.dataset dat ON dat.id=sub.dataset_id
JOIN {schema}.sample smp ON smp.id=ssmp.sample_id 
WHERE smp.id NOT IN (select DISTINCT sample_id FROM {schema}.region2sample)
GROUP BY dat.name
"""
                sqlCmd = sql.SQL(sSql).format(
                    schema=sql.Identifier(sSchema)
                )
                cursor.execute(sqlCmd)
                sSqlQuery = cursor.query.decode("utf-8")
                lstNew = cursor.fetchall()
                res.addData("new", lstNew)
            except Exception as ex:
                res.addErr("Problems getting new unmapped samples", add=str(ex), kat="FAO regions", sys="Database")

        # *** Hole Region/Dataset Anzahl aller verteilten Samples pro Dataset
        if bShort:
            # --- KURZ
            try:
                sSql = \
                    """
                    select reg.id "id", reg.code as "code", reg.level as "level", ST_NPoints(geometry) as "points", reg.name as "region", count(smp.id) as "count" 
                    from {schema}.sample smp 
                    join {schema}.region2sample r2s on r2s.sample_id=smp.id
                    join {schema}.region reg on reg.id=r2s.region_id
                    group by reg.id, reg.name
                    order by points desc, reg.name asc
                    """
                sqlCmd = sql.SQL(sSql).format(
                    schema=sql.Identifier(sSchema)
                )
                cursor.execute(sqlCmd)
                sSqlQuery = cursor.query.decode("utf-8")
                lstContent = cursor.fetchall()

                dicReg_name2id = {}
                dicReg = OrderedDict()
                setReg_names = set()
                iLeftover = 0
                for iId, sCode, sLevel, iPoints, sReg, iCnt in lstContent:
                    dicReg_name2id[sReg] = iId
                    setReg_names.add(sReg)
                    iRegions += 1
                    if sCode.lower() == sLeftover_name.lower():
                        iLeftover += iCnt
                    dicReg[iId] = (sReg, sCode, sLevel, iPoints, iCnt)
                lstRegions = sorted(list(setReg_names))
                res.addData("leftover_nr", iLeftover)
                res.addData("region_nr", len(lstRegions))
                res.addData("region_name", lstRegions)
                res.addData("region_dic_name2id", dicReg_name2id)
                res.addData("region_dic", dicReg)
            except Exception as ex:
                res.addErr("Problems getting region statistics", add=str(ex), kat="FAO regions", sys="Database")
        else:
            # *** LANG
            try:
                sSql = \
"""
select reg.id "id", reg.code as "code",  reg.level as "level", reg.name as "region", ST_NPoints(geometry) as "points", dat.name as "dataset", count(smp.lat_start) as "count" 
from {schema}.subset sub 
join {schema}.subsample ssmp on ssmp.id=sub.subsample_id 
join {schema}.sample smp on smp.id=ssmp.sample_id 
join {schema}.region2sample r2s on r2s.sample_id=smp.id
join {schema}.region reg on reg.id=r2s.region_id
join {schema}.dataset dat on dat.id=sub.dataset_id
group by reg.id, reg.name, dat.name
order by points desc, reg.name asc, dat.name asc
"""
                sqlCmd = sql.SQL(sSql).format(
                    schema=sql.Identifier(sSchema)
                )
                cursor.execute(sqlCmd)
                sSqlQuery = cursor.query.decode("utf-8")
                lstContent = cursor.fetchall()

                dicReg_name2id = {}
                dicReg = OrderedDict()
                setReg_names = set()
                iLeftover = 0
                for iId, sCode, sLevel, sReg, iPoints, sDat, iCnt in lstContent:
                    dicReg_name2id[sReg] = iId
                    setReg_names.add(sReg)
                    if sCode.lower() == sLeftover_name.lower():
                        iLeftover += iCnt
                    if iId not in dicReg:
                        dicReg[iId] = (sReg, sCode, sLevel, iPoints, [(sDat, iCnt)], iCnt)
                    else:
                        sReg_tmp, sCode_tmp, sLevel_tmp, iPoints_tmp, lstDat_tmp, iCnt_tmp = dicReg[iId]
                        lstDat_tmp.append((sDat,iCnt))
                        iCnt_tmp += iCnt
                        dicReg[iId] = (sReg_tmp, sCode_tmp, sLevel_tmp, iPoints_tmp, lstDat_tmp, iCnt_tmp)
                iRegions = len(setReg_names)
                if sLeftover_name in setReg_names:
                    iRegions -= 1
                lstRegions = sorted(list(setReg_names))
                res.addData("leftover_nr", iLeftover)
                res.addData("region_nr", len(lstRegions))
                res.addData("region_name", lstRegions)
                res.addData("region_dic_name2id", dicReg_name2id)
                res.addData("region_dic", dicReg)
            except Exception as ex:
                res.addErr("Problems getting region statistics", add=str(ex), kat="FAO regions", sys="Database")

        # *** Proz
        fProz_new = 0
        fProz_leftover = 0
        try:
            fProz_new = iNew/iSmp*100
            if fProz_new >= 1:
                fProz_new = int(fProz_new)
            else:
                fProz_new = round(fProz_new,2)
            fProz_leftover = iLeftover/iSmp*100
            if fProz_leftover >= 1:
                fProz_leftover = int(fProz_leftover)
            else:
                fProz_leftover = round(fProz_leftover,2)
        except:
            pass
        res.addData("leftover_proz", fProz_leftover)
        res.addData("new_proz", fProz_new)

        # *** Log
        sLog_region += "------------------------------------\n"
        sLog_region += "Regions (" + str(iRegions) + "x)\n"
        sLog_region += "------------------------------------\n\n"
        sLog_leftover += "------------------------------------\n"
        sLog_leftover += "Leftover (" + str(iLeftover) + "x " + str(fProz_leftover) + "%)\n"
        sLog_leftover += "------------------------------------\n\n"
        sLog_new += "------------------------------------\n"
        sLog_new += "New/Unmapped (" + str(iNew) + "x " + str(fProz_new) + "%)\n"
        sLog_new += "------------------------------------\n\n"
        if bShort:
            # *** Short
            lstTable_leftover = []
            lstTable_regions = []
            lstTable_new = []
            # Mapped Samples
            for iId, (sReg, sCode, sLevel, iPoints, iCnt) in dicReg.items():
                if sCode.lower() != sLeftover_name.lower():
                    # Normale Region
                    lstTable_regions.append((sReg, iPoints, iCnt))
                    iRegions += 1
                else:
                    # Leftover separieren
                    lstTable_leftover.append((iLeftover, str(fProz_leftover)+"%"))
                    iLeftover += 1
            # Regions
            if iRegions > 0:
                lstTable_regions.insert(0,("Region", "Size (pts)", "Samples"))
                sLog_region += PrettyText.table(lstTable_regions) + "\n\n"
            # Leftover
            if iLeftover > 0:
                lstTable_leftover.insert(0, ("Samples", "Ratio"))
                sLog_leftover += PrettyText.table(lstTable_leftover) + "\n\n"
            # New Samples
            if iNew > 0:
                lstTable_new.insert(0, ("Samples", "Ratio"))
                lstTable_new.append((iNew, str(fProz_new) + "%"))
                sLog_new += PrettyText.table(lstTable_new) + "\n\n"

        else:
            # *** Long
            # Mapped Samples
            lstTable_regions_overview =  [("Dataset", "Size (pts)", "Samples")]
            for iId, (sReg, sCode, sLevel, iPoints, lstDat, iCnt) in dicReg.items():
                if sCode != sLeftover_name.lower():
                    lstTable_regions_overview.append((sReg, iPoints, iCnt))
                    # Normale Region
                    sLog_region_datasets += "* " + str(sReg)  + "\n"
                    lstTable_regions = [("Dataset", "Samples")]
                    lstTable_regions += lstDat
                    sLog_region_datasets += PrettyText.table(lstTable_regions) + "\n\n"
                else:
                    # Leftover
                    lstTable_leftover = [("Dataset", "Samples")]
                    lstTable_leftover += lstDat
                    sLog_leftover += PrettyText.table(lstTable_leftover) + "\n\n"
            # Overview
            sLog_region_overview += PrettyText.table(lstTable_regions_overview) + "\n\n"

            if iNew > 0:
                # New Samples
                lstTable_new = [("Dataset", "Samples")]
                lstTable_new += lstNew
                sLog_new += PrettyText.table(lstTable_new) + "\n\n"

        # *** Log zusammenbauen
        if iNew > 0:
            sLog += sLog_new
        if iLeftover > 0:
            sLog += sLog_leftover
        if iRegions > 0:
            sLog_region += sLog_region_overview
            sLog_region += sLog_region_datasets
            sLog += sLog_region
        res.addData("log", sLog)

        workflow()
        progress("off")
        if res.isOk():
            conn.commit()
            conn.close()
        return res

    def fao_map_update(self, **kwargs):
        """
        Ordne alle neuen Samples den FAO Regionen zu
        Kwargs:
            leftover (str): Override Leftover name (Name, Code:Name.lower())
            schema (str): Schema
        """

        pn = self.__prnt.normal
        pv = self.__prnt.verbose
        pd = self.__prnt.debug
        progress = self.__prnt.progress
        workflow = self.__prnt.workflow
        status = self.__prnt.status
        sLeftover_name_std = self.__model.subsystem_getRegions().conf_getLeftover_name()

        sSchema = kwargs["schema"] if "schema" in kwargs else self.account_active_getSchema()
        sLeftover_name = kwargs["leftover"] if "leftover" in kwargs else sLeftover_name_std
        cursor = kwargs["cursor"] if "cursor" in kwargs else None

        progress("on-pulse")
        workflow(top="FAO Regions", sub="Update mapping", fmt=False)
        res = Result("Update regions mapping")
        resCursor = self.__model.subsystem_getDb().conn_getConnection()
        if resCursor.isErr():
            res.addResult(resCursor)
            return res
        conn = resCursor.getData("connection")
        cursor = resCursor.getData("cursor")

        # Startzeit
        iTime_start_total = time.time()
        sTime_start_total = str(datetime.datetime.now().strftime("%Y-%m-%d %H:%M"))

        pn("* Update mapping of samples to FAO regions", fmt=False)

        # *** Sammle ungemappte Samples ein
        iNewSamples = 0
        try:
            workflow(top="FAO regions", sub="Collect new samples", fmt=False)
            pn("* Collecting new/unmapped samples", fmt=False)
            # Alle Samples
            # Liste von Samples
            sSql = "CREATE OR REPLACE VIEW {schema}.tmpsmpnew AS SELECT smp.id from {schema}.sample smp where smp.id not in (select distinct sample_id from {schema}.region2sample); SELECT count(*) FROM {schema}.tmpsmpnew;"
            sqlCmd = sql.SQL(sSql).format(
                schema=sql.Identifier(sSchema)
            )
            cursor.execute(sqlCmd)
            iNewSamples = cursor.fetchone()[0]  # Antwort
            sSqlQuery = cursor.query.decode("utf-8")
            pd(sSqlQuery)
        except Exception as ex:
            res.addErr("Problems updating FAO mapping", add=str(ex), kat="FAO regions", sys="Database")
        res.addData("new_samples", iNewSamples)
        if iNewSamples == 0:
            pn("No new/unmapped samples found. Nothing to do ...")
            res.addData("log", "No new/unmapped samples found. Nothing to do ...")
            res.addData("idle", True)
            return res

        # *** Hole alle Regionen
        dicRegions = None
        iId_noRegion = None
        resReg =self.fao_regions_getRegions(schema=sSchema)
        if resReg.isOk():
            dicRegions = resReg.getData("regions")
            if len(dicRegions) == 0:
                sLog = "************************************\n"
                sLog += "Mapping samples to FAO regions \n"
                sLog += "************************************\n\n"
                sLog += "New/unmapped samples: " + str(iNewSamples) + "\n"
                sLog += "\n"
                sLog += "No FAO regions available. Please create regions first ..."
                res.addData("idle", True)
                res.addData("log", sLog)
                return res
        else:
            res.addResult(resReg, msg_only=True)

        # *** Check für alle neuen Samples ob sie in eine Region gehören
        if res.isOk():
            workflow(top="FAO Regions", sub="Update mapping", fmt=False)
            # Bestimme alle Punkte gegen die Verglichen werden müssen
            iNr_regions = len(dicRegions) -1
            iPoints_total = 0
            for iNr, (sName, (iId, sCode, sLevel, iPoints)) in enumerate(dicRegions.items()):
                if iPoints is not None:
                    iPoints_total += iPoints
            pn("Number of unmapped samples: "+ str(iNewSamples))
            pn("Number of regions: "+ str(iNr_regions))
            pn("Number of region points: "+ str(iPoints_total))
            pn("* Update every region", fmt=False)
            # *** Loop: Regions
            if iNr_regions < 0: iNr_regions = 0
            # Time
            iRegions_total = len(dicRegions)
            iTime_end_region = time.time()
            iTime_duration_now = 0
            iTime_duration_region = 0
            iRegions_left = iRegions_total
            iTime_duration_left_estimated = 0
            iPoints_done = 0
            iPoints_left = iPoints_total
            lstMappedSamples = []
            for iNr, (sName, (iId, sCode, sLevel, iPoints)) in enumerate(dicRegions.items()):
                # No Region rausfiltern
                if sCode.lower() == sLeftover_name.lower():
                    iId_noRegion = iId
                    continue
                # Regionen left
                iRegion_now = iNr + 1
                iRegions_left =  iRegions_total - iRegion_now
                # Points left
                if iPoints is not None:
                    iPoints_done += iPoints
                iPoints_left = iPoints_total - iPoints_done
                # Progres (GUI)
                sInfoProgress_bar = str(sName)
                sInfoProgress_bar_time = None
                if iTime_duration_left_estimated != 0:
                    sInfoProgress_bar_time = PrettyText.duration(iTime_duration_left_estimated) + " left …"

                # Progress (CLI)
                pv(str(iRegion_now) + "/" + str(iRegions_total) + ": " + str(sName))
                pv("   size: "+str(sLevel)+" with "+str(iPoints)+ " points")
                if iTime_duration_left_estimated != 0:
                    eta = datetime.datetime.now() + datetime.timedelta(seconds=iTime_duration_left_estimated)
                    sEta = eta.strftime("%X")
                    pv("   eta: " + str(sEta) + " (~" + PrettyText.duration(iTime_duration_left_estimated) + " remaining)")

                progress(iNr, iNr_regions, 100, cli=True, proz=True, abs=True, err=res.getErrNo(), warn=res.getWarnNo(), info=res.getInfoNo(), text=sInfoProgress_bar, time=sInfoProgress_bar_time, fmt=False)
                # Teste Region
                iTime_start_region = time.time()
                iMapped = 0
                try:
                    sSql = \
"""
WITH mapped AS (
INSERT INTO {schema}.region2sample (sample_id, region_id)
SELECT tmpwork.id as sample_id, reg.id AS region_id
FROM (

SELECT smp.id, smp.lon_start AS lon_start, smp.lat_start AS lat_start, smp.precision AS precision
FROM {schema}.sample smp 
WHERE smp.id IN 
(SELECT distinct id FROM {schema}.tmpsmpnew)
GROUP BY smp.id, smp.lon_start, smp.lat_start

) tmpwork, {schema}.region reg 
WHERE 
st_intersects(reg.geometry, 
ST_POINT(
trunc(tmpwork.lon_start::numeric, tmpwork.precision::integer), 
trunc(tmpwork.lat_start::numeric, tmpwork.precision::integer))
)		  
AND reg.id={id} RETURNING *
) 
SELECT COUNT(*) FROM mapped
"""
                    sqlCmd = sql.SQL(sSql).format(
                        schema=sql.Identifier(sSchema),
                        id=sql.Literal(iId)
                    )
                    cursor.execute(sqlCmd)
                    sSqlQuery = cursor.query.decode("utf-8")
                    pd(sSqlQuery)
                    iMapped = cursor.fetchone()[0]  # Antwort
                except Exception as ex:
                    res.addErr("Problems updating FAO mapping", add=str(ex), kat="FAO regions", sys="Database")

                if iMapped > 0:
                    pv("   mapped samples: "+str(iMapped))
                    lstMappedSamples.append((sName, iMapped))
                iTime_end_region = time.time()
                iTime_duration_region = iTime_end_region - iTime_start_region
                iTime_duration_now = iTime_end_region - iTime_start_total
                iTime_duration_left_estimated = (iTime_duration_now / iPoints_done) * iPoints_left
                pv("   regions left: "+str(iRegions_left)+" with "+str(iPoints_left)+" points in total")
                pv("   duration: "+ PrettyText.duration(iTime_duration_region) + " (total elapsed time: "+ PrettyText.duration(iTime_duration_now)+")")

        # *** Abschluss: Sammle alle Samples ein die nicht zugeordnet werden konnten
        if res.isOk():
            try:
                workflow(top="FAO Regions", sub="Collect leftovers", fmt=False)
                pn("* Collecting leftovers", fmt=False)
                sSql = \
"""
DELETE FROM {schema}.region2sample r2s WHERE r2s.region_id={id};
INSERT INTO {schema}.region2sample (sample_id, region_id)
SELECT smp.id as sample_id, {id} AS region_id FROM {schema}.sample smp 
WHERE smp.id NOT IN (select distinct sample_id FROM {schema}.region2sample)
"""
                sqlCmd = sql.SQL(sSql).format(
                    schema=sql.Identifier(sSchema),
                    id=sql.Literal(iId_noRegion)
                )
                cursor.execute(sqlCmd)
                sSqlQuery = cursor.query.decode("utf-8")
            except Exception as ex:
                res.addErr("Problems collection of not mapped samples", add=str(ex), kat="FAO regions", sys="Database")

        # Endzeit
        sTime_end_total = str(datetime.datetime.now().strftime("%Y-%m-%d %H:%M"))
        iSec_end_total = time.time() - iTime_start_total

        if res.isOk():
            conn.commit()
            conn.close()

        # *** Log
        if res.isOk():
            sLog = "************************************\n"
            sLog += "Mapping samples to FAO regions \n"
            sLog += "Duration: " + PrettyText.duration(iSec_end_total)+"\n"
            sLog += "************************************\n\n"
            sLog += "New/unmapped samples: " + str(iNewSamples) +"\n"
            sLog += "FAO regions to check: " + str(iNr_regions) +"\n"
            sLog += "Region points: " + str(iPoints_total) +"\n"
            sLog += "\n"
            if len(lstMappedSamples) > 0:
                sLog += "------------------------------------\n"
                sLog += "Mapped samples in this run\n"
                sLog += "------------------------------------\n\n"
                lstMappedSamples.insert(0,("Region", "Samples"))
                sLog += PrettyText.table(lstMappedSamples) +"\n\n"
            res.addData("log", sLog)

        # *** Action Log
        if not self.__model.conf_isDry() and res.isOk():
            resActionLog = self.__model.subsystem_getDb().actionlog_new("fao-update", "map", sLog)
            res.addResult(resActionLog)
        res.addData("idle", False)
        workflow()
        progress()
        return res

    # --- Tools

    def tools_isSqlUnsafe(self, eingabe):
        """
        Check SQL-Strings
        Erlaube nur nicht leere alphanumerische Strings
        Benutzung für Tabellen/Attributnamen
        Werte werden durch Framework getestet
        Args:
            eingabe (str): Zu testender String
        Returns:
            bool: Ist der SQL-String unsicher
        """
        # Nur ein Element uebergeben -> Liste
        if not isinstance(eingabe, (list, tuple)):
            lstText = [eingabe]
        else:
            lstText = eingabe
        # Jedes Listenelement pruefen (erlaube nur nicht-leere alphanumerische Strings)
        for sText in lstText:
            if not (re.match(r'^[A-Za-z0-9_-]+$', sText)):
                self.__prnt.normal("Error: SQL string contains forbidden characters: \"" + str(sText) + "\"")
                return True
        return False

    def tools_polishSql(self, sSqlCmd):
        """
        Bereite ein SQL-Kommando auf:
        Entferne Whitespaces und ersetze ein Schema-Tag
        Args:
            sSqlCmd (String): SQL-Kommando

        Returns:
            str: SQL-Kommando
        """
        # Bereite die Kommandos auf
        sSchema = self.__model.subsystem_getDb().account_single_getPart("schema")
        sSqlCmd = sSqlCmd.replace("<schema>", sSchema)  # Schema einfügen
        sSqlCmd = re.sub(' +', ' ', sSqlCmd)  # Mehrfache Spaces
        sSqlCmd = re.sub('\t+', ' ', sSqlCmd)  # Mehrfache Tabs
        return sSqlCmd

    # --- Account
    def account_backup(self):
        if self.__dicAccounts is not None:

            self.__sAccount_active_backup = self.__sAccount_active
            self.__dicAccounts_backup = self.__dicAccounts.copy()
            return True
        else:
            return False

    def account_restore(self):
        if self.__dicAccounts_backup is not None:
            self.__sAccount_active = self.__sAccount_active_backup
            self.__dicAccounts = self.__dicAccounts_backup.copy()
            return True
        else:
            return False

    def account_all_get(self):
        """
        Hole Kopie der gesamten Account-Einstellung
        Returns:
            (tuple) dbAccount: Alle Account-Einstellungen mit folgendem Aufbau:
                Aktiver Account (str), Accounts (odic))
        """
        return self.__dicAccounts.copy()

    def account_all_set(self, dbAccount=None):
        """
        Setze gesamte Account-Einstellung neu
        Args:
            (tuple) dbAccount: Alle Account-Einstellungen mit folgendem Aufbau:
                Aktiver Account (str), Accounts (odic))
        """
        if dbAccount is None:
            dbAccount = (None, OrderedDict())
        self.__sAccount_active = dbAccount[0]
        self.__dicAccounts = dbAccount[1]

    def account_all_getNr(self):
        return len(self.__dicAccounts)

    def account_all_getNames(self):
        return list(sorted(self.__dicAccounts.keys(), key=lambda i: i.lower()))

    def account_active_getName(self):
        return self.__sAccount_active

    def account_active_setName(self, sPreset):
        if sPreset in self.__dicAccounts:
            self.__sAccount_active = sPreset
            return True
        else:
            return False

    def account_active_getIndex(self):
        try:
            return list(self.__dicAccounts.keys()).index(self.__sAccount_active)
        except:
            return None

    def account_active_setIndex(self, iIdx):
        if iIdx < 0 or iIdx >= self.account_all_getNr() or iIdx == None:
            return False
        else:
            try:
                self.__sAccount_active = list(self.__dicAccounts.keys())[iIdx]
                return True
            except:
                return False

    def account_single_isEmpty(self, **kwargs):
        iAccountNr = kwargs.get("account_index")
        sAccountName = kwargs.get("account")
        if sAccountName == iAccountNr == None:
            sAccountName = self.account_active_getName()
        elif sAccountName is None and iAccountNr is not None:
            try:
                sAccountName = list(self.__dicAccounts.keys())[iAccountNr]
            except:
                return None
        try:
            sAcc_dat = self.__dicAccounts[sAccountName]
            sHost = sAcc_dat.get("host")
            sPort = sAcc_dat.get("port")
            sDatabase = sAcc_dat.get("db")
            sSchema = sAcc_dat.get("schema")
            sUser = sAcc_dat.get("user")
            sPw = sAcc_dat.get("pw")
            for sContent in [sHost, sPw, sPort, sDatabase, sSchema, sUser, sPw]:
                if sContent not in [None, ""]:
                    return False
        except:
            return None
        return True

    def account_single_del(self, sAccount):
        if sAccount in self.__dicAccounts:
            if sAccount == self.account_active_getName():
                iIndex = self.account_active_getIndex()
                if iIndex == 0:
                    iIndex_active_new = 0
                else:
                    iIndex_active_new = iIndex -1
            self.__dicAccounts.pop(sAccount)
            self.account_active_setIndex(iIndex_active_new)
            return True
        else:
            return False

    def account_single_add(self, dicAccount):
        try:
            sPreset = dicAccount.get("account")
            sHost = dicAccount.get("host")
            sPort = dicAccount.get("port")
            sDatabase = dicAccount.get("db")
            sSchema = dicAccount.get("schema")
            sUser = dicAccount.get("user")
            sPw = dicAccount.get("pw")
            if [None, ""] in [sPreset, sHost, sPw, sPort, sDatabase, sSchema, sUser, sPw]:
                return False
            sPreset = PrettyText.versionize(sPreset, list(self.__dicAccounts.keys())) # Versionieren falls notwendig
            self.__dicAccounts[sPreset] = {
                "host": sHost,
                "port": sPort,
                "db": sDatabase,
                "schema": sSchema,
                "user": sUser,
                "pw": sPw
            }
            return sPreset
        except:
            return None

    def account_single_getPart(self, sName, **kwargs):
        """
        Hole bestimmten Wert von einem Account
        Werte-Typen: preset, host, port, name, schema, user, pw
        Wird kein Account-Index übergeben wird der sich aktuell
        in Benutzung befindliche Account gewählt.
        Args:
            sName (str): Gewünschter Wert des Accounts
        Kwargs
            account_index (int):
            account (str):
        Returns:
            str: gewählter Wert des Accounts
        """
        iAccountNr = kwargs.get("account_index")
        sAccountName = kwargs.get("account")

        if sAccountName == iAccountNr == None:
            sAccountName = self.account_active_getName()
        elif sAccountName is None and iAccountNr is not None:
            sAccountName = list(self.__dicAccounts.keys())[iAccountNr] if iAccountNr < len(self.__dicAccounts.keys()) else None
        try:
            if sName == "account": return sAccountName
            if sName == "host": return self.__dicAccounts[sAccountName].get("host")
            if sName == "port": return self.__dicAccounts[sAccountName].get("port")
            if sName == "db": return self.__dicAccounts[sAccountName].get("db")
            if sName == "schema": return self.__dicAccounts[sAccountName].get("schema")
            if sName == "user": return self.__dicAccounts[sAccountName].get("user")
            if sName == "pw": return self.__dicAccounts[sAccountName].get("pw")
        except Exception as ex:
            pass
        return None

    def account_single_setPart(self, sName, sWert, **kwargs):
        """
        Setze bestimmten Wert in einem Account
        Wird kein Account-Index übergeben wird der sich aktuell
        in Benutzung befindliche Account gewählt.
        Args:
            sName (str): Gewünschter Wert des Accounts
        Kwargs
            account_index (int):
            account (str):
        Returns:
            bool: War die Aktion erfolgreich oder nicht?
        """
        iAccountNr = kwargs.get("account_index")
        sAccountName = kwargs.get("account")

        if sAccountName == iAccountNr == None:
            sAccountName = self.account_active_getName()
        elif sAccountName is None and iAccountNr is not None:
            try:
                sAccountName = list(self.__dicAccounts.keys())[iAccountNr]
            except:
                return None
        try:
            if sName == "host": self.__dicAccounts[sAccountName]["host"] = sWert
            elif sName == "port": self.__dicAccounts[sAccountName]["port"] = sWert
            elif sName == "db": self.__dicAccounts[sAccountName]["db"] = sWert
            elif sName == "schema": self.__dicAccounts[sAccountName]["schema"] = sWert
            elif sName == "user": self.__dicAccounts[sAccountName]["user"] = sWert
            elif sName == "pw": self.__dicAccounts[sAccountName]["pw"] = sWert
            else: return None
        except Exception as ex:
            return None
        return True

    def account_rename(self, sAccount_old, sAccount_new):
        if sAccount_old not in self.__dicAccounts:
            return None
        elif sAccount_old == sAccount_new:
            return True
        elif sAccount_new in self.__dicAccounts:
            # Neuen Namen gibt es bereits
            sAccount_new_version = PrettyText.versionize(sAccount_new, list(
                self.__dicAccounts.keys()))  # Versionieren falls notwendig
            sAccount_new = sAccount_new_version
        # War der alter Account aktiv?
        if sAccount_old == self.account_active_getName():
            bForge = True
        else:
            bForge = False
        # Account umbenennen
        self.__dicAccounts[sAccount_new] = self.__dicAccounts.pop(sAccount_old)
        self.__dicAccounts[sAccount_new]["account"] = sAccount_new
        # Aktiven Account neu setzen
        if bForge:
            self.account_active_setName(sAccount_new)

    def account_active_getSchema(self):
        try:
            return self.__dicAccounts[self.__sAccount_active].get("schema")
        except:
            self.__prnt.normal("Problems getting schema for Account: " + str(self.__sAccount_active))
            return None

    def account_active_setSchema(self, sSchema):
        if self.tools_isSqlUnsafe(sSchema):
            self.__prnt.normal("Db schema (" + sSchema + ") will not be saved in model.")
            return False
        else:
            try:
                self.__dicAccounts[self.__sAccount_active]["schema"] = sSchema
            except:
                self.__prnt.normal("Problems saving schema: " + sSchema)
                return False
            return True

    # --- Get: Admin Data

    def admin_getObligat(self, sTab, sCol, **kwargs):
        """
        Hole Obligatinfo zu einem Tabellen-Attribut
        Args:
            sTab (str): Tabellenname
            sCol (str): Spaltenname
        Kwargs:
            cursor (psycopg2.cursor.Cursor): Postgres Cursor
        Returns:
            bool: Ist das Attribut (Spalte) obligat?
        """
        res = Result("Get obligat Info")
        cursor = kwargs["cursor"] if "cursor" in kwargs else None
        resCursor = self.conn_getConnection()
        if resCursor.isErr():
            res.addResult(resCursor)
            return res
        conn = resCursor.getData("connection")
        cursor = resCursor.getData("cursor")
        try:
            cursor.execute("SELECT is_nullable FROM information_schema.columns where table_name = '" + sTab + "' and table_schema = '" + self.account_active_getSchema() + "' and column_name = '" + sCol + "'")
            sNullable = str(cursor.fetchone()[0])
            if sNullable.lower() == "no":
                bObligat = True
            else:
                bObligat = False
        except Exception as ex:
            res.addErr("Problems getting obligate information to attribute", add=str(ex), kat="Obligate", sys="Database")
            bObligat = False
        res.addData("obl", bObligat)
        return res

    def admin_getAllComments(self, **kwargs):
        """
        Hole Alle Commentare zu einem Schema
        Kwargs:
            cursor (psycopg2.cursor.Cursor): Postgres Cursor
        Returns:
            Result
        """
        res = Result("Get all Commentaries")
        cursor = kwargs["cursor"] if "cursor" in kwargs else None
        resCursor = self.conn_getConnection()
        if resCursor.isErr():
            res.addResult(resCursor)
            return res
        conn = resCursor.getData("connection")
        cursor = resCursor.getData("cursor")
        try:
            sSql = \
"""select c.table_name, c.column_name,c.udt_name,c.is_nullable,d.description
from information_schema.columns c
inner join pg_class c1
on c.table_name=c1.relname
inner join pg_catalog.pg_namespace n
on c.table_schema=n.nspname
and c1.relnamespace=n.oid
left join pg_catalog.pg_description d
on d.objsubid=c.ordinal_position
and d.objoid=c1.oid
where c.table_schema='<schema>'"""
            sSql = sSql.replace("<schema>", self.account_active_getSchema())
            cursor.execute(sSql)
            ant = cursor.fetchall()
            dic = {}
            setTest = set()
            for (sTable, sCol, sDataType, sNullable, sComment) in ant:
                if (sTable, sCol) not in setTest:
                    setTest.add((sTable, sCol))
                else:
                    res.addErr("Problems getting commentaries to schema", add="Double data found: "+str(sTable)+"."+str(sCol), kat="Commentary", sys="Database")
                if sTable not in dic:
                    dic[sTable] = {}
                if sNullable.lower() == "no":
                    bObl = True
                elif sNullable.lower() == "yes":
                    bObl = False
                else:
                    raise NameError("Error in nullable definition of: "+str(sTable)+"."+str(sCol)+" -> Value: "+str(sNullable))
                dic[sTable][sCol] =  (bObl, sDataType, sComment)
            res.addData("comments", dic)
        except Exception as ex:
            res.addErr("Problems getting commentaries to schema", add=str(ex), kat="Commentary", sys="Database")
        return res

    def admin_getComment(self, sTab, sCol, **kwargs):
        """
        Hole Kommentar zu einem Tabellenattribut
        Args:
            sTab (str): Tabellenname
            sCol (str): Spaltenname
        Kwargs:
            cursor (psycopg2.cursor.Cursor): Postgres Cursor
        Returns:
            str: Kommentar zu Tabellenattribut
        """
        res = Result("Get Comment")
        cursor = kwargs["cursor"] if "cursor" in kwargs else None
        resCursor = self.conn_getConnection()
        if resCursor.isErr():
            res.addResult(resCursor)
            return res
        conn = resCursor.getData("connection")
        cursor = resCursor.getData("cursor")
        if None in (sTab, sCol):
            res.addData("comment", None)
            return res
        try:
            cursor.execute("SELECT oid FROM pg_catalog.pg_namespace WHERE nspname = '" + self.account_active_getSchema() + "'")
            sRelnamespace = str(cursor.fetchone()[0])
            cursor.execute("SELECT oid FROM pg_class WHERE relname = '" + sTab + "' and relnamespace='" + sRelnamespace + "';")
            sObjoid = str(cursor.fetchone()[0])
            cursor.execute("SELECT ordinal_position FROM information_schema.columns WHERE table_schema = '" + self.account_active_getSchema() + "' AND table_name = '" + sTab + "' and column_name = '" + sCol + "'")
            sObjsubid = str(cursor.fetchone()[0])
            cursor.execute("SELECT description FROM pg_catalog.pg_description WHERE objsubid ='" + sObjsubid + "' AND objoid ='" + sObjoid + "'")
            sCommentColumn = str(cursor.fetchone()[0])
        except:
            sCommentColumn = ""
        res.addData("comment", sCommentColumn)
        return res

    def admin_getAttributNames(self, sTab, **kwargs):
        """
        Hole alle Attribute zu einer Tabelle
        Args:
            sTab (str): Tabellenname
        Kwargs:
            cursor (psycopg2.cursor.Cursor): Postgres Cursor
        Returns:
            Result
                sql (str): SQL Befehl
                atr (lst): Liste mit Attributen einer Tabelle
        """
        res = Result("Get attribute names of table")
        cursor = kwargs["cursor"] if "cursor" in kwargs else None
        resCursor = self.conn_getConnection()
        if resCursor.isErr():
            res.addResult(resCursor)
            return res
        conn = resCursor.getData("connection")
        cursor = resCursor.getData("cursor")
        try:
            sqlCmd = sql.SQL(
                "select column_name from information_schema.columns where table_schema = {schema} and table_name = {table}").format(
                schema=sql.Literal(self.account_active_getSchema()),
                table=sql.Literal(sTab)
            )
            cursor.execute(sqlCmd)
            sSqlQuery = cursor.query.decode("utf-8")
            lstSqlAttr = cursor.fetchall()
            lstAttr = list()
            for i in lstSqlAttr:
                lstAttr.append(i[0])
            res.addData("sql", sSqlQuery)
            res.addData("header", lstAttr)
        except Exception as ex:
            res.addErr("Unexcepted Error", add=str(ex), sys="Database", kat="Get Attribute Names")
        return res

    # --- Admin Ingest

    def ingest_getIngestDependencies(self, **kwargs):
        """
        Bestimme alle Abhängigkeiten der Ingests
        Args:
        Returns:
            tools.coding.Result: Ergebnis
                depend (dict): Depends on  ... Ingests von denen key abhängig ist
                base (dict): Foundation for ... Ingests, die von key abhängig sind
        """
        res = Result(name="Testing dependencies between ingests")
        cursor = kwargs["cursor"] if "cursor" in kwargs else None
        if cursor is None:
            resCursor = self.conn_getConnection()
            if resCursor.isErr():
                res.addResult(resCursor)
                return res
            conn = resCursor.getData("connection")
            cursor = resCursor.getData("cursor")
        try:
            # Helper: Füge ein Tupel (key,value) in Dictionary ein (v kann ein wert oder eine liste von werten sein)
            def store(k, v, dic):
                if k not in dic:
                    # Neuer Key
                    if type(v) is list:
                        # Value ist Liste
                        dic[k] = v
                    else:
                        # Value ist Element
                        dic[k] = [v]
                else:
                    # Key bereits vorhanden
                    if type(v) is list:
                        # Value ist Liste
                        for item in v:
                            # Füege nur neue Elemente hinzu
                            if item not in dic[k]:
                                dic[k].append(item)
                    else:
                        # Value ist Element
                        dic[k].append(v)

            # Helper: Invertiere ein Dictionary (key<->values)
            def invert(dic):
                inv = {}
                for k, lstV in dic.items():
                    for v in lstV:
                        if v not in inv:
                            inv[v] = [k]
                        else:
                            inv[v].append(k)
                return inv

            # Dictionaries
            dicNeed__sample2station = {}
            dicNeed__station2cruise = {}
            dicNeed__subsample2sample = {}
            dicNeed = {}
            dicBase = {}

            # # --- Depend: SubSample -> Sample
            sSqlCmd = \
                """
                SELECT DISTINCT
                subsmp.ingest_id AS subsample,
                smp.ingest_id AS sample
                FROM
                <schema>.subsample subsmp
                JOIN
                <schema>.sample smp
                ON
                subsmp.sample_id=smp.id
                WHERE
                smp.ingest_id!=subsmp.ingest_id;
                """
            sSqlCmd = self.tools_polishSql(sSqlCmd)

            cursor.execute(sSqlCmd)
            lstContent = cursor.fetchall()
            for (iSubSample, iSample) in lstContent:
                store(iSubSample, iSample, dicNeed__subsample2sample)

            # --- Depend: Sample -> Station
            sSqlCmd = \
                """
                SELECT DISTINCT
                smp.ingest_id AS sample,
                sta.ingest_id AS station
                FROM
                <schema>.sample smp
                JOIN
                <schema>.station sta
                ON
                smp.station_id=sta.id
                WHERE
                smp.ingest_id!=sta.ingest_id;
                """
            sSqlCmd = self.tools_polishSql(sSqlCmd)

            cursor.execute(sSqlCmd)
            lstContent = cursor.fetchall()
            for (iSample, iStation) in lstContent:
                store(iSample, iStation, dicNeed__sample2station)

            # --- Depend: Station -> Cruise
            sSqlCmd = \
                """
                SELECT DISTINCT
                sta.ingest_id AS station,
                cru.ingest_id AS cruise
                FROM
                <schema>.station sta
                JOIN
                <schema>.cruise cru
                ON
                sta.cruise_id=cru.id
                WHERE
                sta.ingest_id!=cru.ingest_id;
                """
            sSqlCmd = self.tools_polishSql(sSqlCmd)
            resCursor = self.__model.subsystem_getDb().conn_getConnection()
            if resCursor.isErr():
                res.addResult(resCursor)
                return res
            conn = resCursor.getData("connection")
            cursor = resCursor.getData("cursor")
            cursor.execute(sSqlCmd)
            lstContent = cursor.fetchall()
            for (iStation, iCruise) in lstContent:
                store(iStation, iCruise, dicNeed__station2cruise)

            # --- Ergebnis
            # 1) Depend
            # Sammle IngIds ein: SubSample->Sample(->Station(->Cruise))
            for iSubSample, lstSample in dicNeed__subsample2sample.items():
                store(iSubSample, lstSample, dicNeed) # Save: SubSample->Sample
                for iSample in lstSample:
                    # Füege evtl. noch die Station des Samples hinzu
                    if iSample in dicNeed__sample2station:
                        store(iSubSample, dicNeed__sample2station[iSample], dicNeed) # Save: SubSample(->Sample)->Station
                        for iStation in dicNeed__sample2station[iSample]:
                            # Füege evtl. noch die Cruises der Stations hinzu
                            if iStation in dicNeed__station2cruise:
                                store(iSubSample, dicNeed__station2cruise[iStation], dicNeed) # Save: SubSample(->Sample)(->Station)->Cruise

            # Sammle IngIds ein: Sample->Station(->Cruise)
            for iSample, lstStation in dicNeed__sample2station.items():
                store(iSample, lstStation, dicNeed) # Save: Sample->Station
                for iStation in lstStation:
                    # Füege evtl. noch die Cruises der Stationen hinzu
                    if iStation in dicNeed__station2cruise:
                        store(iSample, dicNeed__station2cruise[iStation], dicNeed) # Save: Sample(->Station)->Cruise

            # Sammle IngIds ein: Station->Cruise
            for iStation, lstCruise in dicNeed__station2cruise.items():
                store(iStation, lstCruise, dicNeed) # Save!

            # 2) Base
            dicBase__cruise2station = invert(dicNeed__station2cruise)
            dicBase__station2sample = invert(dicNeed__sample2station)
            dicBase__sample2subsample = invert(dicNeed__subsample2sample)
            # Sammle IngIds ein: Cruise->Station(->Sample(->SubSample))
            for iCruise, lstStation in dicBase__cruise2station.items():
                store(iCruise, lstStation, dicBase) # Save
                for iStation in lstStation:
                    if iStation in dicBase__station2sample:
                        store(iCruise, dicBase__station2sample[iStation], dicBase) # Save
                        for iSample in dicBase__station2sample[iStation]:
                            if iSample in dicBase__sample2subsample:
                                store(iCruise, dicBase__sample2subsample[iSample], dicBase) # Save
            # Sammle IngIds ein: Station->Sample(->SubSample)
            for iStation, lstSample in dicBase__station2sample.items():
                store(iStation, lstSample, dicBase)
                for iSample in lstSample:
                    if iSample in dicBase__sample2subsample:
                        store(iStation, dicBase__sample2subsample[iStation], dicBase) # Save
            # Sammle IngIds ein: Sample->SubSample
            for iSample, lstSubSample in dicBase__sample2subsample.items():
                store(iSample, lstSubSample, dicBase) # Save

            #print("depend (smp2sta):", dicNeed__sample2station)
            #print("depend (sta2cru)", dicNeed__station2cruise)
            #print("depend (sta2cru)", dicNeed__station2cruise)

            # --- Antwort verpacken
            res.addData("need", dicNeed)
            res.addData("base", dicBase)

        except Exception as ex:
            res.addErr("Unexcepted Error", add=str(ex), sys="Database", kat="Get Ingest Dependencies")
        return res

    def ingest_getWorkflowOfIngest(self, iIngestId, **kwargs):
        """
        Hole Workflow zu Ingest
        Args:
            iIngestId (int): Ingest Id
        Kwargs:
            cursor (psycopg2.cursor.Cursor): Postgres Cursor
        Returns:
            Result:
                wf (str): Workflow
        """
        res = Result("Get Workflow of Ingest")
        cursor = kwargs["cursor"] if "cursor" in kwargs else None
        if cursor is None:
            resCursor = self.conn_getConnection()
            if resCursor.isErr():
                res.addResult(resCursor)
                return res
            conn = resCursor.getData("connection")
            cursor = resCursor.getData("cursor")
        try:
            sqlCmd = sql.SQL(
                "SELECT workflow FROM {schema}.ingest where id=%s").format(
                schema=sql.Identifier(self.__model.subsystem_getDb().account_active_getSchema()),)
            cursor.execute(sqlCmd, [str(iIngestId)])
            sWf = cursor.fetchone()[0]
            res.addData("workflow", sWf)
        except Exception as ex:
            res.addErr("Unexcepted Error", add=str(ex), sys="Database", kat="Get Workflow of Ingest")
        return res

    def ingest_getIngestFilenames(self, lstIds, **kwargs):
        """
        Hole Namen zu Ingests
        Args:
            lstIds (list): Ids
        Kwargs:
            cursor (cursor): Cursor
        Returns:
            Result:
                files (list): Filenames
        """
        res = Result("Get Ingest Filename")
        if type(lstIds) in (str,int):
            lstIds = [lstIds]
        cursor = kwargs["cursor"] if "cursor" in kwargs else None
        bOwnConn = False
        if cursor is None:
            bOwnConn = True
            resConn = self.conn_getConnection()
            if resConn.isErr():
                return resConn
            cursor = resConn.getData("cursor")
            conn = resConn.getData("connection")
        try:
            # TODO: Baue SQL (sicher)
            sqlCmd = sql.SQL(
                "SELECT id, file FROM {schema}.ingest where id in ({ids})").format(
                schema=sql.Identifier(self.__model.subsystem_getDb().account_active_getSchema()),
                ids=sql.SQL(', ').join(map(sql.Literal, lstIds))
            )
            cursor.execute(sqlCmd)
            sSqlQuery = cursor.query.decode("utf-8")
            lstIdFiles = cursor.fetchall()
            setFiles = set()
            dicFiles = {}
            for iId, sFile in lstIdFiles:
                setFiles.add(sFile)
                dicFiles[iId] = sFile
            if len(setFiles) == 1:
                res.addData("file", list(setFiles)[0])
            res.addData("files", list(setFiles))
            res.addData("dic_files", dicFiles)
            res.addData("sql", sSqlQuery)
        except Exception as ex:
            res.addErr("Unexcepted Error", add=str(ex), sys="Database", kat="Get Ingest Filename")

        if bOwnConn:
            cursor.close()
            conn.commit()
            conn.close()

        return res

    def ingest_getLookupDataForIngests(self, sLookup, lstIngests, **kwargs):
        """
        Hole alle Lookupdaten für Ingests
        Args:
            sTab (str): Lookup-Tabelle
            lstIngests (lst): Ingests
        Kwargs:
            cursor (psycopg2.cursor.Cursor): Postgres Cursor
        Returns:
            (tools.coding.Result): Lookupdaten
        """
        res = Result("Get Lookup Data For Ingest")
        cursor = kwargs["cursor"] if "cursor" in kwargs else None
        if cursor is None:
            resCursor = self.conn_getConnection()
            if resCursor.isErr():
                res.addResult(resCursor)
                return res
            conn = resCursor.getData("connection")
            cursor = resCursor.getData("cursor")

        tpIngests = tuple(lstIngests)
        if sLookup == "gear":
            # Gear
            sSqlCmd = "select distinct gea.* from <schema>.gear gea, <schema>.sample smp where smp.ingest_id in %(ingests)s and smp.gear_id=gea.id"
        elif sLookup == "taxon":
            # Taxon
            sSqlCmd = "select distinct tax.* from <schema>.taxon tax, <schema>.biota bio where bio.ingest_id in %(ingests)s and bio.taxon_id=tax.id"
        elif sLookup == "dataset":
            # Dataset
            sSqlCmd = "select distinct dat.* from <schema>.dataset dat, <schema>.subset sub where sub.ingest_id in %(ingests)s and sub.dataset_id=dat.id"
        elif sLookup == "crs":
            # CRS
            sSqlCmd = "select distinct crs.* from <schema>.crs crs, <schema>.cruise cru where cru.ingest_id in %(ingests)s and cru.crs_id=crs.id"
        elif sLookup == "ship":
            # ship
            sSqlCmd = "select distinct shp.* from <schema>.ship shp, <schema>.cruise cru where cru.ingest_id in %(ingests)s and cru.ship_id=shp.id"
        elif sLookup == "possys":
            # possys
            sSqlCmd = "select distinct pos.* from <schema>.possys pos, <schema>.cruise cru where cru.ingest_id in %(ingests)s and cru.positioningsystem_id=pos.id"
        elif sLookup == "taxcover":
            # taxcover
            sSqlCmd = "select distinct cov.* from <schema>.taxcover cov, <schema>.subset sub where sub.ingest_id in %(ingests)s and sub.taxcover_id=cov.id"
        elif sLookup == "lifestage":
            # lifestage
            sSqlCmd = "select distinct lif.* from <schema>.lifestage lif, <schema>.biota bio where bio.ingest_id in %(ingests)s and bio.lifestage_id=lif.id"
        elif sLookup == "country":
            # country
            sSqlCmd = "select distinct cnt.* from <schema>.country cnt, <schema>.dataset dat, <schema>.subset sub where sub.ingest_id in %(ingests)s and sub.dataset_id=dat.id and dat.owner_country_id=cnt.id"
        elif sLookup == "person":
            # person
            sSqlCmd = \
                """
                select distinct * from ( 
                select distinct per.* from <schema>.person per, <schema>.cruise cru
                where (cru.ingest_id in %(ingests)s and cru.lead_person_id=per.id)
                union all
                select distinct per.* from <schema>.person per, <schema>.station st
                where (st.ingest_id in %(ingests)s and st.responsible_person_id=per.id)
                union all
                select distinct per.* from <schema>.person per, <schema>.sample smp
                where (smp.ingest_id in %(ingests)s and smp.responsible_person_id=per.id)
                union all
                select distinct per.* from <schema>.person per, <schema>.subset sub, <schema>.dataset dat 
                where (sub.ingest_id in %(ingests)s and sub.dataset_id=dat.id and dat.contact_person_id=per.id)
                ) p order by name
                """
        else:
            res.addErr("Unknown lookup", add=str(sLookup), sys="Database", kat="Get lookup data for ingest")
            return res
        sSqlCmd = sSqlCmd.replace("<schema>", self.account_single_getPart("schema"))
        sSqlCmd = re.sub(' +', ' ', sSqlCmd)  # Mehrfache Spaces
        sSqlCmd = re.sub('\t+', ' ', sSqlCmd)  # Mehrfache Tabs
        try:
            cursor.execute(sSqlCmd, {"ingests": tpIngests})
            sSqlQuery = cursor.query.decode("utf-8")
            lstHeader = [head[0] for head in cursor.description]
            lstContent = cursor.fetchall()
            res.addData("header", lstHeader)
            res.addData("content", lstContent)
        except Exception as ex:
            res.addErr("Unexpected error", add=str(ex), sys="Database", kat="Get lookup data for ingest")
        return res

    def lookup_purge(self, sLookup=None, bFull=False):
        """
        Purge Lookup
        Args:
            Args:
                sLookup (str): Lookup Table
                bFull (bool): alle unbenutzen Lookupeinträge (ansonsten nur die unbenutzten nicht goldenen)
        Kwargs:
            cursor (psycopg2.cursor.Cursor): Postgres Cursor
        Returns:
            (dict,str): Dictionary mit Übersicht über gelösche Datensätze, Fehlermeldung
        """

        progress = self.__prnt.progress
        status = self.__prnt.status
        workflow = self.__prnt.workflow
        pv = self.__prnt.verbose

        result = Result(name="Purges Lookup entries")

        # DB-Verbindung
        resCursor = self.conn_getConnection()
        if resCursor.isErr():
            result.addResult(resCursor)
            return result
        conn = resCursor.getData("connection")
        cursor = resCursor.getData("cursor")

        # Daten
        progress("pulse-on")
        lstLookup = []
        if sLookup is None:
            for sLup in self.__model.subsystem_getMapper().lookup_getDbTables():
                if sLup not in lstLookup and sLup not in ["taxon", "person", "country"]:
                    lstLookup.append(sLup)
            lstLookup.append("person")
            lstLookup.append("country")
        else:
            lstLookup.append(sLookup)

        sMode = "unused" if bFull else "unused not-gold"

        dicPurged = {}

        pv("* Purge Lookups (" +sMode+ ")", fmt=False)
        for iNr, sLookup in enumerate(lstLookup):
            sSheet = self.__model.subsystem_getMapper().lookup_table2sheet(sLookup)
            # Status
            workflow(top="Purge Lookup", sub=str(sSheet) + " - getting stats from db")
            resLupStatus = self.__model.subsystem_getDb().lookup_getStatus(sLookup)
            result.addResult(resLupStatus, msg_only=True)
            if resLupStatus.isOk():
                dicNames = resLupStatus.getData("all-names")
                if bFull:
                    setIds = resLupStatus.getData("not-used-ids")
                else:
                    setIds = resLupStatus.getData("not-used-not-gold-ids")
                # Purge
                iIds = len(setIds)
                lstIds = list(setIds)
                if iIds > 0:
                    try:
                        workflow(top="Purge Lookup", sub=str(sSheet), fmt=False)
                        dicErg = OrderedDict()
                        sqlCmd = sql.SQL(
                            "DELETE FROM {schema}.{table} WHERE id IN ({ids})").format(
                            schema=sql.Identifier(self.__model.subsystem_getDb().account_active_getSchema()),
                            table=sql.Identifier(sLookup),
                            ids=sql.SQL(', ').join(map(sql.Literal, lstIds))
                        )
                        cursor.execute(sqlCmd)
                        sSqlQuery = cursor.query.decode("utf-8")
                        pv(sSheet +": "+str(iIds)+"x", fmt=False)
                        dicPurged[sSheet] = iIds

                        # Eintrag in Action Log
                        sPurgeLog = "--- lookup: "+str(sLookup) +"\n"
                        sPurgeLog += "purged records ("+str(iIds)+"):\n"
                        for i in setIds:
                            if i in dicNames:
                                sPurgeLog += str(dicNames[i])+"\n"
                        sPurgeLog = sPurgeLog[:-1]
                        resultActionLog = self.__model.subsystem_getDb().actionlog_new("lup-del", str(sLookup), sPurgeLog, cursor=cursor)
                        if resultActionLog.isErr():
                            result.addResult(resultActionLog, err=True)

                    except Exception as ex:
                        conn.rollback()
                        result.addErr("Purge: "+str(sLookup), add=str(ex), sys="Database", kat="Purge lookup")

            # Progress
            progress(iNr, len(lstLookup), 100, proz=True, abs=True, err=result.getErrNo(), warn=result.getWarnNo(),info=result.getInfoNo())
            conn.commit()

        # DB-Verbindung schliessen
        cursor.close()
        conn.close()

        # UI sauber
        progress("off")
        workflow()

        # Antwort
        result.addData("purged", dicPurged)
        return result

    def ingest_getOverview(self):
        res = Result("Get ingest overview")
        resCursor = self.conn_getConnection()
        if resCursor.isErr():
            res.addResult(resCursor)
            return res
        conn = resCursor.getData("connection")
        cursor = resCursor.getData("cursor")
        try:
            sqlCmd = sql.SQL("SELECT id, name, created_on, workflow, rows, warn, info, file FROM {schema}.ingest;").format(
                schema=sql.Identifier(self.account_active_getSchema())
            )
            cursor.execute(sqlCmd)
            sSqlQuery = cursor.query.decode("utf-8")
            lstHeader = [head[0] for head in cursor.description]
            lstContent = cursor.fetchall()
            res.addData("sql", sSqlQuery)
            res.addData("header", lstHeader)
            res.addData("content", lstContent)
        except Exception as ex:
            res.addErr("Unexpected error", add=str(ex), sys="Database", kat="Get entries from table")
        cursor.close()
        conn.close()
        return res

    def ingest_delIngest(self, lstIngestIds):
        """
        Lösche ein gesamtes Ingest aus Datenbank (Manuell)
        Dry Mode funktioniert hier nicht

        Args:
            lstIngestId (int): Liste mit Ingest-ID

        Returns:
            (dict,str): Dictionary mit Übersicht über gelösche Datensätze, Fehlermeldung
        """

        progress = self.__prnt.progress
        status = self.__prnt.status
        workflow = self.__prnt.workflow
        pn = self.__prnt.normal
        pv = self.__prnt.verbose

        result = Result(name="Delete ingests")

        # DB-Verbindung
        resCursor = self.conn_getConnection()
        if resCursor.isErr():
            result.addResult(resCursor)
            return result
        conn = resCursor.getData("connection")
        cursor = resCursor.getData("cursor")

        # Daten
        iIngests = len(lstIngestIds)
        iIngest = 1
        iRun = 0
        lstDelOk = []
        lstIngestIdNow = lstIngestIds.copy()
        lstIngestIdLast = []


        progress("pulse-on")
        # Look Ingest Ids immer wieder bis Liste vollständig geeleert ist oder nichts weiter geht
        while (len(lstIngestIdNow)>0):
            iRun += 1

            if lstIngestIdNow == lstIngestIdLast: # Es kann ncihts mehr gelöscht werden -> Ende
                break
            lstIngestIdTmp = lstIngestIdNow.copy()
            lstIngestIdLast = lstIngestIdNow.copy()

            for iIngestId in lstIngestIdTmp:
                try:
                    dicErg = OrderedDict()

                    # Filename
                    sSqlCmd = "select file from " + self.__model.subsystem_getDb().account_single_getPart("schema") + ".ingest where id=%s;"
                    cursor.execute(sSqlCmd, [str(iIngestId)])
                    sFilename = str(cursor.fetchone()[0])
                    pv("* Trying to delete ingest ("+str(iIngestId)+"): "+str(sFilename), fmt=False)
                    # Vorarbeit
                    # *** Entferne Samples aus FAO region mapping
                    resFaoClean = self.fao_map_clean(ing=iIngestId)
                    if resFaoClean.isErr():
                        result.addResult(resFaoClean, msg_only=True)

                    workflow(top="Delete Ingests", sub="[run:" + str(iRun) + " del:" + str(len(lstDelOk)) + " left:" + str(len(lstIngestIdNow)) +  "]")

                    # *** Manuell: Korrekte Reihenfolge der Löschungen
                    # indmeasure
                    sSqlCmd = "delete from " + self.__model.subsystem_getDb().account_single_getPart(
                        "schema") + ".indmeasure mea where mea.ingest_id=%s;"
                    cursor.execute(sSqlCmd, [str(iIngestId)])
                    if cursor.rowcount > 0:
                        dicErg["indmeasure"] = cursor.rowcount
                        # Sediment
                    sSqlCmd = "delete from " + self.__model.subsystem_getDb().account_single_getPart(
                        "schema") + ".sediment sedi where sedi.ingest_id=%s;"
                    cursor.execute(sSqlCmd, [str(iIngestId)])
                    if cursor.rowcount > 0:
                        dicErg["sediment"] = cursor.rowcount
                        # Population
                    sSqlCmd = "delete from " + self.__model.subsystem_getDb().account_single_getPart("schema") + ".biota bio where bio.ingest_id=%s;"
                    cursor.execute(sSqlCmd, [str(iIngestId)])
                    if cursor.rowcount > 0:
                        dicErg["biota"] = cursor.rowcount
                        # Subset
                    sSqlCmd = "delete from " + self.__model.subsystem_getDb().account_single_getPart("schema") + ".subset sub where sub.ingest_id=%s;"
                    cursor.execute(sSqlCmd, [str(iIngestId)])
                    if cursor.rowcount > 0:
                        dicErg["subset"] = cursor.rowcount
                        # SubSample
                    sSqlCmd = "delete from " + self.__model.subsystem_getDb().account_single_getPart(
                        "schema") + ".subsample subsmp where subsmp.ingest_id=%s;"
                    cursor.execute(sSqlCmd, [str(iIngestId)])
                    if cursor.rowcount > 0:
                        dicErg["subsample"] = cursor.rowcount
                        # Sample:
                    sSqlCmd = "delete from " + self.__model.subsystem_getDb().account_single_getPart("schema") + ".sample smp where smp.ingest_id=%s;"
                    cursor.execute(sSqlCmd, [str(iIngestId)])
                    if cursor.rowcount > 0:
                        dicErg["sample"] = cursor.rowcount
                        # Station
                    sSqlCmd = "delete from " + self.__model.subsystem_getDb().account_single_getPart("schema") + ".station st where st.ingest_id=%s;"
                    cursor.execute(sSqlCmd, [str(iIngestId)])
                    if cursor.rowcount > 0:
                        dicErg["station"] = cursor.rowcount
                        # Cruise
                    sSqlCmd = "delete from " + self.__model.subsystem_getDb().account_single_getPart("schema") + ".cruise cr where cr.ingest_id=%s;"
                    cursor.execute(sSqlCmd, [str(iIngestId)])
                    if cursor.rowcount > 0:
                        dicErg["cruise"] = cursor.rowcount
                        # Ingest
                    sSqlCmd = "delete from " + self.__model.subsystem_getDb().account_single_getPart("schema") + ".ingest ing where ing.id=%s;"
                    cursor.execute(sSqlCmd, [str(iIngestId)])
                    if cursor.rowcount > 0:
                        dicErg["ingest"] = cursor.rowcount

                    # Eintrag in Action Log
                    sDelLog = "--- id: "+str(iIngestId) +"\n"
                    sDelLog += "Deleted records:\n"
                    for k,v in dicErg.items():
                        sDelLog += str(k) +": "+str(v) + "\n"
                    sDelLog = sDelLog[:-1]
                    resultActionLog = self.__model.subsystem_getDb().actionlog_new("ing-del", str(FileTools.getFileWithoutPath(sFilename)), sDelLog, cursor=cursor)
                    if resultActionLog.isErr():
                        result.addResult(resultActionLog, err=True)

                    # Ok
                    pv("ingest successfully deleted")
                    status("Deleted ingest: "+str(iIngestId))
                    lstDelOk.append(iIngestId)
                    lstIngestIdNow.remove(iIngestId)
                    iIngest += 1
                    conn.commit()

                except Exception as ex:
                    pv("problem deleting ingest ... better luck next run ...")
                    result.addErr("Problem", add=str(ex), kat="Delete Ingest", system="Database")
                    conn.commit()

                # Progress
                progress(iIngest, iIngests, 100, proz=True, abs=True, err=result.getErrNo(), warn=result.getWarnNo(),info=result.getInfoNo())

        # DB-Verbindung schliessen
        cursor.close()
        conn.close()

        # UI sauber
        progress("off")
        workflow()

        # Antwort
        result.addData("del-total", lstIngestIds)
        result.addData("del-ok", lstDelOk)
        result.addData("del-err", lstIngestIdNow)

        return result


    # --- Lookup
    def lookup_getStatus(self, sTab, **kwargs):
        """
        Hole Informationen zu Lookup
            used, gold, ...
        Args:
            sTab (str): Tabellenname
        Kwargs:
            cursor (psycopg2.cursor.Cursor): Postgres Cursor
        Returns:
           Result
                table (str)
                all-names (dict)
                all (int)

                gold (int)
                gold-proz (int)
                gold-ids (set)

                not-gold (int)
                not-gold-proz (int)
                not-gold-ids (set)

                used (int)
                used-proz (int)
                used-ids (set)

                not-used-ids (set)
                not-used-proz (int)
                not-used (int)

                used-ing (int)
                used-ing-ids (set)
                used-lup (int)
                used-lup-ids (set)
                not-used-ing (int)
                not-used-ing-ids (set)
                not-used-lup (int)
                not-used-lup-ids (set)

                used-gold (int)
                used-gold-proz (int)
                used-gold-ids (set)


                used-not-gold (int)
                used-not-gold-ids (set)
                used-not-gold-proz (int)

                not-used-gold (int)
                not-used-gold-proz (int)
                not-used-gold-ids (set)
                
                not-used-not-gold (int)
                not-used-not-gold-proz (int)
                not-used-not-gold-ids (set) 

                Dataset (only)
                    dataset-files (dic)
                    dataset-access-data-free (int)
                    dataset-access-data-free-ids (lst)
                    dataset-access-data-request (int)
                    dataset-access-data-request-ids (lst)
                    dataset-access-data-confidential (int)
                    dataset-access-data-confidential-ids (lst)
                    dataset-access-product-free (int)
                    dataset-access-product-free-ids (lst)
                    dataset-access-product-request (int)
                    dataset-access-product-request-ids (lst)
                    dataset-access-product-confidential (int)
                    dataset-access-product-confidential-ids (lst)
                    dataset-access-possible (int)
                    dataset-access-possible-ids (lst)
                    dataset-access-possible-proz (int)
                    dataset-access-restricted (int)
                    dataset-access-restricted-ids (lst)
                    dataset-access-restricted-proz (int)

                Gear (only)
                    gear-double-header (lst, lst)
                    gear-double-groups (int)
                    gear-double-groups-set (set)
                    gear-double-dic (dic)
                    gear-double (int)
                    gear-double-ids (set)
        """
        res = Result("Get gold Lookups")
        cursor = kwargs["cursor"] if "cursor" in kwargs else None

        if cursor is None:
            resCursor = self.conn_getConnection()
            if resCursor.isErr():
                res.addResult(resCursor)
                return res
            conn = resCursor.getData("connection")
            cursor = resCursor.getData("cursor")

        if sTab == "taxon" or not self.__model.subsystem_getMapper().all_isTableLookup(sTab):
            res.addErr("Table is not allowed", add=str(sTab), sys="Database", kat="Getting status of lookup table")
            return res

        # ---------------------------------------------------------------------
        # Allgemein
        # ---------------------------------------------------------------------

        setAll = set()
        dicAll_names = {}

        setGold = set()
        setNotGold = set()

        setUsed = set()
        setUsed_ing = set()
        setUsed_lup = set()
        dicUsed_ing_sourcefiles = {}
        dicUsed_lup_sourcelup = {}
        setUsed_Gold = set()
        setUsed_NotGold = set()

        setNotUsed = set()
        setNotUsed_Gold = set()
        setNotUsed_NotGold = set()

        iGoldProz = 0
        iUsedProz = 0
        iUsed_Gold_Proz = 0
        iNotUsed_NotGold_Proz = 0
        iUsed_NotGold_Proz = 0
        iNotUsed_Gold_Proz = 0

        # All
        try:
            sqlCmd = sql.SQL(
                "SELECT id, name from {schema}.{table}").format(
                schema=sql.Identifier(self.account_active_getSchema()),
                table=sql.Identifier(sTab)
            )
            cursor.execute(sqlCmd)
            lstAll = cursor.fetchall()  # Antwort
            dicAll_names = {tp[0] : tp[1] for tp in lstAll}
            setAll = set(dicAll_names.keys())
        except Exception as ex:
            res.addErr("Unexpected error", add=str(ex), sys="Database", kat="Lookup: Get all entries")
        res.addData("all-names", dicAll_names)
        res.addData("all-ids", setAll)
        res.addData("all", len(setAll))

        # Not-Gold
        try:
            sqlCmd = sql.SQL(
                "SELECT id from {schema}.{table} WHERE gold=False").format(
                schema=sql.Identifier(self.account_active_getSchema()),
                table=sql.Identifier(sTab)
            )
            cursor.execute(sqlCmd)
            lstNotGold = cursor.fetchall()  # Antwort
            setNotGold = {tp[0] for tp in lstNotGold}
        except Exception as ex:
            res.addErr("Unexpected error", add=str(ex), sys="Database", kat="Lookup: Get not-gold entries")
        res.addData("not-gold-ids", setNotGold)
        res.addData("not-gold", len(setNotGold))

        # Gold
        try:
            sqlCmd = sql.SQL(
                "SELECT id from {schema}.{table} WHERE gold=True").format(
                schema=sql.Identifier(self.account_active_getSchema()),
                table=sql.Identifier(sTab)
            )
            cursor.execute(sqlCmd)
            lstGold = cursor.fetchall()  # Antwort
            setGold = {tp[0] for tp in lstGold}
        except Exception as ex:
            res.addErr("Unexpected error", add=str(ex), sys="Database", kat="Lookup: Get gold entries")
        # %
        try:
            iGoldProz = int((len(setGold) / len(setAll)) * 100)
        except:
            pass
        res.addData("gold-ids", setGold)
        res.addData("gold", len(setGold))
        res.addData("gold-proz", iGoldProz)
        res.addData("not-gold-proz", 100-iGoldProz)

        # Used: In Ingest Tables
        lstUse_ing = self.__model.subsystem_getMapper().lookup_getAllUsages_ingest(sTab)
        if lstUse_ing is not None and len(lstUse_ing) > 0:
            for sIngTab, sIngAtr in lstUse_ing:
                try:
                    sqlCmd = sql.SQL(
                        "SELECT DISTINCT {atr}, {schema}.ingest.file FROM {schema}.{table} JOIN {schema}.ingest ON {schema}.ingest.id={schema}.{table}.ingest_id WHERE {atr} is not NULL").format(
                        schema=sql.Identifier(self.account_active_getSchema()),
                        table=sql.Identifier(sIngTab),
                        atr=sql.Identifier(sIngAtr)
                    )
                    cursor.execute(sqlCmd)
                    sSqlQuery = cursor.query.decode("utf-8")
                    lstIngIds_tab = cursor.fetchall()  # Antwort
                    setIngIds_tab = {tp[0] for tp in lstIngIds_tab}
                    for iId, sFile in lstIngIds_tab:
                        if iId not in dicUsed_ing_sourcefiles:
                            dicUsed_ing_sourcefiles[iId] = set()
                        dicUsed_ing_sourcefiles[iId].add(sFile)
                    setIngFiles_tab = {tp[1] for tp in lstIngIds_tab}
                    setUsed_ing = setUsed_ing.union(setIngIds_tab)
                except Exception as ex:
                    res.addErr("Unexpected error", add=str(ex), sys="Database", kat="Lookup: Get used-in-ing entries")
        res.addData("used-ing-ids", setUsed_ing)
        res.addData("used-ing-sourcefiles-dic", dicUsed_ing_sourcefiles)
        res.addData("not-used-ing-ids", setAll.difference(setUsed_ing))
        res.addData("used-ing-sourcefiles", len(dicUsed_ing_sourcefiles))
        res.addData("used-ing", len(setUsed_ing))
        res.addData("not-used-ing", len(setAll.difference(setUsed_ing)))

        # Used: In other Lookup Tables
        lstUse_lup = self.__model.subsystem_getMapper().lookup_getAllUsages_lookup(sTab)
        if lstUse_lup is not None and len(lstUse_lup) > 0:
            for sLupTab, sLupAtr in lstUse_lup:
                try:
                    sLupKey_name = self.__model.subsystem_getMapper().lookup_getKey(sLupTab)
                    sqlCmd = sql.SQL(
                        "SELECT DISTINCT {atr}, {lupkey} FROM {schema}.{table} WHERE {atr} is not NULL").format(
                        schema=sql.Identifier(self.account_active_getSchema()),
                        table=sql.Identifier(sLupTab),
                        atr=sql.Identifier(sLupAtr),
                        lupkey=sql.Identifier(sLupKey_name)
                    )
                    cursor.execute(sqlCmd)
                    sSqlQuery = cursor.query.decode("utf-8")
                    lstLupIds_tab = cursor.fetchall()  # Antwort
                    setLupIds_tab = {tp[0] for tp in lstLupIds_tab}
                    setUsed_lup = setUsed_ing.union(setLupIds_tab)
                    for iId, sLupSource in lstLupIds_tab:
                        if iId not in dicUsed_lup_sourcelup:
                            dicUsed_lup_sourcelup[iId] = {}
                        if sLupTab not in dicUsed_lup_sourcelup[iId]:
                            dicUsed_lup_sourcelup[iId][sLupTab] = set()
                        dicUsed_lup_sourcelup[iId][sLupTab].add(sLupSource)
                except Exception as ex:
                    res.addErr("Unexpected error", add=str(ex), sys="Database", kat="Lookup: Get used-in-lup entries")
        res.addData("used-lup-ids", setUsed_lup)
        res.addData("used-lup-sourcelup-dic", dicUsed_lup_sourcelup)
        res.addData("not-used-lup-ids", setAll.difference(setUsed_lup))
        res.addData("used-lup", len(setUsed_lup))
        res.addData("used-lup-sourcelup", len(dicUsed_lup_sourcelup))
        res.addData("not-used-lup", len(setAll.difference(setUsed_lup)))

        # Used/Not-used total
        setUsed = setUsed_ing.union(setUsed_lup)
        setNotUsed = setAll.difference(setUsed)
        res.addData("used-ids", setUsed)
        res.addData("used", len(setUsed))
        res.addData("not-used-ids", setNotUsed)
        res.addData("not-used", len(setNotUsed))
        try:
            iUsedProz = int((len(setUsed) / len(setAll)) * 100)
        except:
            pass
        res.addData("used-proz", iUsedProz)
        res.addData("not-used-proz", 100 - iUsedProz)

        # Used and gold
        setUsed_Gold = setUsed.intersection(setGold)
        res.addData("used-gold-ids", setUsed_Gold)
        res.addData("used-gold", len(setUsed_Gold))
        try:
            iUsed_Gold_Proz = int((len(setUsed_Gold) / len(setUsed)) * 100)
        except:
            pass
        res.addData("used-gold-proz", iUsed_Gold_Proz)

        # Used and Not-gold (potentiell zu checken und auf gold zu setzen)
        setUsed_NotGold = setUsed.intersection(setNotGold)
        res.addData("used-not-gold-ids", setUsed_NotGold)
        res.addData("used-not-gold", len(setUsed_NotGold))
        # %
        try:
            iUsed_NotGold_Proz = int((len(setUsed_NotGold) / len(setUsed)) * 100)
        except:
            pass
        res.addData("used-not-gold-proz", iUsed_NotGold_Proz)

        # Not-used and gold
        setNotUsed_Gold = setNotUsed.intersection(setGold)
        res.addData("not-used-gold-ids", setNotUsed_Gold)
        res.addData("not-used-gold", len(setNotUsed_Gold))
        # %
        try:
            iNotUsed_Gold_Proz = 100 -int((len(setNotUsed_Gold) / len(setNotUsed)) * 100)
        except:
            pass
        res.addData("not-used-gold-proz", iNotUsed_Gold_Proz)

        # Not-used and Not-gold (können potentiell gelöscht werden)
        setNotUsed_NotGold = setNotUsed.intersection(setNotGold)
        res.addData("not-used-not-gold-ids", setNotUsed_NotGold)
        res.addData("not-used-not-gold", len(setNotUsed_NotGold))
        # %
        try:
            iNotUsed_NotGold_Proz = int((len(setNotUsed_NotGold) / len(setNotUsed)) * 100)
        except:
            pass
        res.addData("not-used-not-gold-proz", iNotUsed_NotGold_Proz)

        # ---------------------------------------------------------------------
        # Special
        # ---------------------------------------------------------------------

        # --- Dataset

        if sTab == "dataset":
            # Datasets / Filenames
            lstDatasetFilename = []
            dicDatasetFilename = {}
            try:
                sqlCmd = sql.SQL(
                    "SELECT distinct dat.id, ing.file from {schema}.subset sub join {schema}.dataset dat on dat.id = sub.dataset_id join {schema}.ingest ing on ing.id=sub.ingest_id").format(
                    schema=sql.Identifier(self.account_active_getSchema())
                )
                cursor.execute(sqlCmd)
                sSqlQuery = cursor.query.decode("utf-8")
                lstDatasetFilename = cursor.fetchall()
                # Antwort
                for iId, sFile in lstDatasetFilename:
                    if iId not in dicDatasetFilename:
                        dicDatasetFilename[iId] = []
                    dicDatasetFilename[iId].append(sFile)
                res.addData("dataset-files", dicDatasetFilename)
            except Exception as ex:
                res.addErr("Unexpected error", add=str(ex), sys="Database",kat="Lookup: Dataset: Getting file names")

            # Access Level
            setLevel_data_free = set()
            setLevel_data_request = set()
            setLevel_data_confidential = set()
            setLevel_product_free = set()
            setLevel_product_request = set()
            setLevel_product_confidential = set()
            try:
                sqlCmd = sql.SQL(
                    "SELECT id, access_level_data, access_level_data_product from {schema}.dataset").format(
                    schema=sql.Identifier(self.account_active_getSchema())
                )
                cursor.execute(sqlCmd)
                lstLevel = cursor.fetchall()  # Antwort
                for iId, sData, sProduct in lstLevel:
                    # Level: Data
                    if sData == "free":
                        setLevel_data_free.add(iId)
                    elif sData == "request":
                        setLevel_data_request.add(iId)
                    elif sData == "confidential":
                        setLevel_data_confidential.add(iId)
                    # Level: Product
                    if sProduct == "free":
                        setLevel_product_free.add(iId)
                    elif sProduct == "request":
                        setLevel_product_request.add(iId)
                    elif sProduct == "confidential":
                        setLevel_product_confidential.add(iId)

                # Access Level
                setAccessPossible_all = setLevel_data_free.union(setLevel_data_request).union(setLevel_product_free).union(setLevel_product_request)
                setAccessRestricted_all = setAll.difference(setAccessPossible_all)
                res.addData("dataset-access-data-free-ids", setLevel_data_free)
                res.addData("dataset-access-data-request-ids", setLevel_data_request)
                res.addData("dataset-access-data-confidential-ids", setLevel_data_confidential)
                res.addData("dataset-access-product-free-ids", setLevel_product_free)
                res.addData("dataset-access-product-request-ids", setLevel_product_request)
                res.addData("dataset-access-product-confidential-ids", setLevel_product_confidential)
                res.addData("dataset-access-possible-ids", setAccessPossible_all)
                res.addData("dataset-access-restricted-ids", setAccessRestricted_all)
                res.addData("dataset-access-data-free", len(setLevel_data_free))
                res.addData("dataset-access-data-request", len(setLevel_data_request))
                res.addData("dataset-access-data-confidential", len(setLevel_data_confidential))
                res.addData("dataset-access-product-free", len(setLevel_product_free))
                res.addData("dataset-access-product-request", len(setLevel_product_request))
                res.addData("dataset-access-product-confidential", len(setLevel_product_confidential))
                res.addData("dataset-access-possible", len(setAccessPossible_all))
                res.addData("dataset-access-restricted", len(setAccessRestricted_all))
                # %
                try:
                    iProz = int((len(setLevel_data_free) / len(setAll)) * 100)
                    res.addData("dataset-access-data-free-proz", iProz)
                except:
                    res.addData("dataset-access-data-free-proz", 0)
                try:
                    iProz = int((len(setLevel_data_request) / len(setAll)) * 100)
                    res.addData("dataset-access-data-request-proz", iProz)
                except:
                    res.addData("dataset-access-data-request-proz", 0)
                try:
                    iProz = int((len(setLevel_data_confidential) / len(setAll)) * 100)
                    res.addData("dataset-access-data-confidential-proz", iProz)
                except:
                    res.addData("dataset-access-data-confidential-proz", 0)
                try:
                    iProz = int((len(setLevel_product_free) / len(setAll)) * 100)
                    res.addData("dataset-access-product-free-proz", iProz)
                except:
                    res.addData("dataset-access-product-free-proz", 0)
                try:
                    iProz = int((len(setLevel_product_request) / len(setAll)) * 100)
                    res.addData("dataset-access-product-request-proz", iProz)
                except:
                    res.addData("dataset-access-product-request-proz", 0)
                try:
                    iProz = int((len(setLevel_product_confidential) / len(setAll)) * 100)
                    res.addData("dataset-access-product-confidential-proz", iProz)
                except:
                    res.addData("dataset-access-product-confidential-proz", 0)
                try:
                    iAccessPossible_proz = int((len(setAccessPossible_all) / len(setAll)) * 100)
                    res.addData("dataset-access-possible-proz", iAccessPossible_proz)
                except:
                    res.addData("dataset-access-possible-proz", 0)
                try:
                    iAccessRestricted_proz = int((len(setAccessRestricted_all) / len(setAll)) * 100)
                    res.addData("dataset-access-restricted-proz", iAccessRestricted_proz)
                except:
                    res.addData("dataset-access-restricted-proz", 0)

                # Access Level: Used
                setAccessPossible_all = setLevel_data_free.union(setLevel_data_request).union(
                    setLevel_product_free).union(setLevel_product_request)
                res.addData("dataset-used-access-data-free-ids", setLevel_data_free.intersection(setUsed))
                res.addData("dataset-used-access-data-request-ids", setLevel_data_request.intersection(setUsed))
                res.addData("dataset-used-access-data-confidential-ids", setLevel_data_confidential.intersection(setUsed))
                res.addData("dataset-used-access-product-free-ids", setLevel_product_free.intersection(setUsed))
                res.addData("dataset-used-access-product-request-ids", setLevel_product_request.intersection(setUsed))
                res.addData("dataset-used-access-product-confidential-ids", setLevel_product_confidential.intersection(setUsed))
                res.addData("dataset-used-access-data-free", len(setLevel_data_free.intersection(setUsed)))
                res.addData("dataset-used-access-data-request", len(setLevel_data_request.intersection(setUsed)))
                res.addData("dataset-used-access-data-confidential", len(setLevel_data_confidential.intersection(setUsed)))
                res.addData("dataset-used-access-product-free", len(setLevel_product_free.intersection(setUsed)))
                res.addData("dataset-used-access-product-request", len(setLevel_product_request.intersection(setUsed)))
                res.addData("dataset-used-access-product-confidential", len(setLevel_product_confidential.intersection(setUsed)))
                # %
                try:
                    iProz = int((len(setLevel_data_free.intersection(setUsed)) / len(setAll)) * 100)
                    res.addData("dataset-used-access-data-free-proz", iProz)
                except:
                    res.addData("dataset-used-access-data-free-proz", 0)
                try:
                    iProz = int((len(setLevel_data_request.intersection(setUsed)) / len(setAll)) * 100)
                    res.addData("dataset-used-access-data-request-proz", iProz)
                except:
                    res.addData("dataset-used-access-data-request-proz", 0)
                try:
                    iProz = int((len(setLevel_data_confidential.intersection(setUsed)) / len(setAll)) * 100)
                    res.addData("dataset-used-access-data-confidential-proz", iProz)
                except:
                    res.addData("dataset-used-access-data-confidential-proz", 0)
                try:
                    iProz = int((len(setLevel_product_free.intersection(setUsed)) / len(setAll)) * 100)
                    res.addData("dataset-used-access-product-free-proz", iProz)
                except:
                    res.addData("dataset-used-access-product-free-proz", 0)
                try:
                    iProz = int((len(setLevel_product_request.intersection(setUsed)) / len(setAll)) * 100)
                    res.addData("dataset-used-access-product-request-proz", iProz)
                except:
                    res.addData("dataset-used-access-product-request-proz", 0)
                try:
                    iProz = int((len(setLevel_product_confidential.intersection(setUsed)) / len(setAll)) * 100)
                    res.addData("dataset-used-access-product-confidential-proz", iProz)
                except:
                    res.addData("dataset-used-access-product-confidential-proz", 0)

                # Used and Access Possibble
                setUsed_AccessPossible_all = setUsed.intersection(
                    setLevel_data_free.union(setLevel_data_request).union(setLevel_product_free).union(
                        setLevel_product_request))
                res.addData("dataset-used-access-possible-ids", setUsed_AccessPossible_all)
                res.addData("dataset-used-access-restricted-ids", setUsed.difference(setUsed_AccessPossible_all))
                res.addData("dataset-used-access-possible", len(setUsed_AccessPossible_all))
                res.addData("dataset-used-access-restricted", len(setUsed.difference(setUsed_AccessPossible_all)))
                try:
                    iUsedAccessPossible_proz = int((len(setUsed_AccessPossible_all) / len(setAll)) * 100)
                    res.addData("dataset-used-access-possible-proz", iUsedAccessPossible_proz)
                except:
                    res.addData("dataset-used-access-possible-proz", 0)
                try:
                    iUsedAccessRestricted_proz = int((len(setUsed.difference(setUsed_AccessPossible_all)) / len(setUsed)) * 100)
                    res.addData("dataset-used-access-restricted-proz", iUsedAccessRestricted_proz)
                except:
                    res.addData("dataset-used-access-restricted-proz", 0)

                # Used and Not-Gold and Access possible
                setUsed_AccessPossible_NotGold_all = setUsed_NotGold.intersection(setLevel_data_free.union(setLevel_data_request).union(setLevel_product_free).union(
                        setLevel_product_request))

                res.addData("dataset-used-not-gold-access-data-free-ids", setUsed_NotGold.intersection(setLevel_data_free))
                res.addData("dataset-used-not-gold-access-data-request-ids", setUsed_NotGold.intersection(setLevel_data_request))
                res.addData("dataset-used-not-gold-access-product-free-ids", setUsed_NotGold.intersection(setLevel_product_free))
                res.addData("dataset-used-not-gold-access-product-request-ids", setUsed_NotGold.intersection(setLevel_product_request))
                res.addData("dataset-used-not-gold-access-possible-ids", setUsed_AccessPossible_NotGold_all)
                res.addData("dataset-used-not-gold-access-data-free", len(setUsed_NotGold.intersection(setLevel_data_free)))
                res.addData("dataset-used-not-gold-access-data-request", len(setUsed_NotGold.intersection(setLevel_data_request)))
                res.addData("dataset-used-not-gold-access-product-free", len(setUsed_NotGold.intersection(setLevel_product_free)))
                res.addData("dataset-used-not-gold-access-product-request", len(setUsed_NotGold.intersection(setLevel_product_request)))
                res.addData("dataset-used-not-gold-access-possible", len(setUsed_AccessPossible_NotGold_all))

            except Exception as ex:
                    res.addErr("Unexpected error", add=str(ex), sys="Database", kat="Lookup: Dataset: Getting Access Levels")

        # --- Gear

        elif sTab == "gear":

            # Gears mit gleichen Werten aber unterschiedlichen Namen
            lstEqual_base = ("id", "name")
            lstEqual_check_config = ("width", "height", "length", "area", "mass", "meshsize")
            lstEqual_check_group = ("category", "type", "smpunit")

            resGear = self.table_getWholeTable("gear")

            dicEqual_check_erg = {}
            setEqual_check_ids = set()
            setGroups_all = set()

            if resGear.isErr():
                res.addResult(resGear, sys="Status on Lookup Table", kat="Get whole table for lookup gear")
            else:
                lstHeader = resGear.getData("header")
                lstContent = resGear.getData("content")
                setContent = set(lstContent)
                setMissing = set(lstEqual_base + lstEqual_check_config + lstEqual_check_group) - set(lstHeader)
                if len(setMissing) > 0:
                    res.addErr("Missing attributes", add=str(PrettyText.lst2strC(setMissing)), sys="Status on Lookup Table", kat="Incompatable Table")
                else:
                    # Test
                    for tpContent in lstContent:
                        dicContent = dict(zip(lstHeader, tpContent))
                        # Name
                        sTestName = dicContent["name"]
                        iTestId = dicContent["id"]

                        # Gruppe
                        lstTestGroup = []
                        for sCol in lstEqual_check_group:
                            lstTestGroup.append(dicContent[sCol])
                        tpTestGroup = tuple(lstTestGroup)
                        # Config
                        lstTestConfig = []
                        bEmptyConfig = True
                        for sCol in lstEqual_check_config:
                            if dicContent[sCol] is not None:
                                bEmptyConfig = False
                            lstTestConfig.append(dicContent[sCol])
                        tpTestConfig = tuple(lstTestConfig)
                        # Eine Nicht leere Configuration zu einer Gruppe merken
                        setGroups_all.add(tpTestGroup)

                        if not bEmptyConfig:
                            if tpTestGroup not in dicEqual_check_erg:
                                dicEqual_check_erg[tpTestGroup] = {}
                            if tpTestConfig not in dicEqual_check_erg[tpTestGroup]:
                                dicEqual_check_erg[tpTestGroup][tpTestConfig] = set()
                            dicEqual_check_erg[tpTestGroup][tpTestConfig].add((sTestName, iTestId))

                # Filtere alle doppelten Namen aus den Gruppen heraus
                dicEqual_check_erg_final = {}
                iDouble = 0
                for tpGroup, dicConf in dicEqual_check_erg.items():
                    for tpConf, lstNamesIds in  dicConf.items():
                        if len(lstNamesIds) > 1:
                            if tpGroup not in dicEqual_check_erg_final:
                                dicEqual_check_erg_final[tpGroup] = {}
                            if tpConf not in dicEqual_check_erg_final[tpGroup]:
                                lstNames = [sName for sName, iId in lstNamesIds]
                                dicEqual_check_erg_final[tpGroup][tpConf] = lstNames
                                for sName, iId in lstNamesIds:
                                    setEqual_check_ids.add(iId)
                                iDouble += len(lstNamesIds)

                # Hole die Sheet Names
                lstEqual_check_group_excel = []
                for sAttr in lstEqual_check_group:
                    sSheet, sCol = self.__model.subsystem_getMapper().lookup_db2excel("gear", sAttr)
                    lstEqual_check_group_excel.append(sCol)
                lstEqual_check_config_excel = []
                for sAttr in lstEqual_check_config:
                    sSheet, sCol = self.__model.subsystem_getMapper().lookup_db2excel("gear", sAttr)
                    lstEqual_check_config_excel.append(sCol)
                # Übersichts summe
                res.addData("gear-double-header", (lstEqual_check_group_excel, lstEqual_check_config_excel))
                res.addData("gear-double-groups", len(setGroups_all))
                res.addData("gear-double-groups-set", setGroups_all)
                res.addData("gear-double-dic", dicEqual_check_erg_final)
                res.addData("gear-double", iDouble)
                res.addData("gear-double-ids", setEqual_check_ids)

        # Tests
        if len(setAll) != len(setGold) + len(setNotGold):
            res.addErr("Unexpected error",
                       add="all:" + str(len(setAll)) + " not-gold:" + str(len(setNotGold)) + " gold:" + str(
                           len(setGold)), sys="Database", kat="Lookup: inkonsistent gold/not-gold entries")
        if len(setAll) != len(setUsed) + len(setNotUsed):
            res.addErr("Unexpected error",
                       add="all:" + str(len(setAll)) + " not-used:" + str(len(setNotUsed)) + " used:" + str(
                           len(setUsed)), sys="Database", kat="Lookup: inkonsistent used/not-used entries")
        if len(setAll) != len(setUsed) + len(setNotUsed):
            res.addErr("Unexpected error",
                       add="all:" + str(len(setAll)) + " not-used:" + str(len(setNotUsed)) + " used:" + str(
                           len(setUsed)), sys="Database", kat="Lookup: inkonsistent used/not-used entries")

        # Result
        res.addData("table", sTab)
        return res

    def lookup_getGoldLookups(self, sTab, **kwargs):
        """
        Hole Anzahl aller Gold Lookup Eintraege
        Args:
            sTab (str): Tabellenname
        KWargs:
            cursor (psycopg2.cursor.Cursor): Postgres Cursor
        Returns:
           Resulst
        """
        res = Result("Get gold Lookups")
        cursor = kwargs["cursor"] if "cursor" in kwargs else None
        if cursor is None:
            resCursor = self.conn_getConnection()
            if resCursor.isErr():
                res.addResult(resCursor)
                return res
            conn = resCursor.getData("connection")
            cursor = resCursor.getData("cursor")
        try:
            sSqlCmd = "SELECT count(id) from " + self.account_active_getSchema() + "." + sTab + " where gold = True"
            cursor.execute(sSqlCmd)
            iCount = cursor.fetchone()[0]  # Antwort
            res.addData("count", iCount)
        except Exception as ex:
            res.addErr("Unexpected error", add=str(ex), sys="Database", kat="Get gold lookups")
        return res

    def lookup_dataset_checkDois(self, **kwargs):
        """
        Teste DOIs aller Datasets
        Kwargs:
            cursor (psycopg2.cursor.Cursor): Postgres Cursor
        Returns:
           Result
        """
        pn = self.__prnt.normal
        #pv = pn
        pv = self.__prnt.verbose
        pd = self.__prnt.debug

        workflow = self.__prnt.workflow
        progress = self.__prnt.progress
        status = self.__prnt.status

        res = Result("Check DOIS")
        cursor = kwargs["cursor"] if "cursor" in kwargs else None
        if cursor is None:
            resCursor = self.conn_getConnection()
            if resCursor.isErr():
                res.addResult(resCursor)
                return res
            conn = resCursor.getData("connection")
            cursor = resCursor.getData("cursor")

        # *** Get DOIs
        try:
            sqlCmd = sql.SQL(
                "SELECT id, name, doi FROM {schema}.dataset").format(
                schema=sql.Identifier(self.account_active_getSchema()),
            )
            cursor.execute(sqlCmd)
            lstDois = cursor.fetchall()  # Antwort
        except Exception as ex:
            res.addErr("Unexpected Error", add=str(ex), sys="Database", kat="Lookup: Checking DOI")

        # *** Teste Erreichbarkeit

        dicDoi_ok = {}
        dicDoi_none = {}
        dicDoi_fail = {}

        if res.isOk():
            progress("on-pulse")
            workflow(top="DOI", sub="Checking reachability", fmt=False)
            for iNr, (iId, sName, sDoi) in enumerate(lstDois):
                if sDoi is None:
                    dicDoi_none[iId] = sName
                    pv(sName +": no DOI given")
                else:
                    try:
                        iErr = len(dicDoi_fail)
                        iWarn = len(dicDoi_none)
                        iInfo = 0

                        progress(iNr, len(lstDois), 100, proz=True, abs=True, cli=True, info=iInfo, warn=iWarn, err=iErr, status=str(sName))
                        req = requests.head(sDoi)
                        bRedirect = req.is_redirect
                        iStatus = req.status_code
                        sRedirect = req.headers["Location"] if "Location" in req.headers else None
                        bOk = iStatus in [200, 302]
                        status("Checking DOI: "+sDoi, fmt=False)
                        if bOk:
                            dicDoi_ok[iId] = (sName, sDoi, sRedirect)
                            sOut = sName + ": ok [" + sDoi
                            if bRedirect:
                                sOut += " -redirected to-> " +str(sRedirect)
                            sOut += "]"
                            pv(sOut)
                        else:
                            dicDoi_fail[iId] = (sName, sDoi, str(iStatus))
                            sOut = sName + ": fail ("+str(iStatus)+") [" + sDoi
                            if bRedirect:
                                sOut += " -redirected-> [" + str(sRedirect)
                            sOut += "]"
                            pv(sOut)
                    except Exception as ex:
                        dicDoi_fail[iId] = (sName, sDoi, str(ex))
                        sErr = PrettyText.clean_ret2space(str(ex))
                        pv(sName + " [" + sDoi+"]: fail ("+sErr+")")

        res.addData("ok", dicDoi_ok)
        res.addData("err", dicDoi_fail)
        res.addData("no", dicDoi_none)

        # *** Gruppiere DOIs nach Journals
        progress("on-pulse")
        workflow(top="DOI", sub="Clustering DOIs", fmt=False)

        def getDoiEffetive(sDoi, sDoi_redirect):
            if sDoi_redirect not in [None, ""]:
                return sDoi_redirect
            else:
                return sDoi

        # Cluster by Publisher
        dicDOIs_clustered_byPublisher = {}
        dicDatasets_clustered_byDoi = {}
        dicJournals_known = {
            "pangaea.de": "PANGAEA",
            "plos.org": "PLOS",
            "springer.com": "Springer",
            "springer.de": "Springer",
            "elsevier.com": "ELSEVIER",
            "int-res.com": "Inter-Research Science Publisher"}

        bHtml = False
        for iId, (sName, sDoi, sDoi_redirect) in dicDoi_ok.items():
            bFound = False
            for sJournal_linkpart, sJournal_name in dicJournals_known.items():
                if not bFound:
                    sDoi_effective = sDoi_redirect if sDoi_redirect not in [None, ""] else sDoi
                    if sDoi_effective not in [None, ""]:
                        if sJournal_linkpart in str(sDoi_effective).lower():
                            if sJournal_name not in dicDOIs_clustered_byPublisher:
                                dicDOIs_clustered_byPublisher[sJournal_name] = {}
                            dicDOIs_clustered_byPublisher[sJournal_name][iId] = (sName, sDoi, sDoi_redirect)
                            bFound = True
            if not bFound:
                if "Others" not in dicDOIs_clustered_byPublisher:
                    dicDOIs_clustered_byPublisher["Others"] = {}
                dicDOIs_clustered_byPublisher["Others"][iId] = (sName, sDoi, sDoi_redirect)

        # *** Datasets gruppiert nach selber DOIs

        for sJournal_name, dicDois_journal in dicDOIs_clustered_byPublisher.items():
            for iId, (sName, sDoi, sDoi_redirect) in dicDois_journal.items():
                sDoi_effective = getDoiEffetive(sDoi, sDoi_redirect)
                if sJournal_name not in dicDatasets_clustered_byDoi:
                    dicDatasets_clustered_byDoi[sJournal_name] = {}
                if sDoi_effective not in dicDatasets_clustered_byDoi[sJournal_name]:
                    dicDatasets_clustered_byDoi[sJournal_name][sDoi_effective] = []
                dicDatasets_clustered_byDoi[sJournal_name][sDoi_effective].append([sName])

        # *** PANGAEA / Nicht-PANGAEA-Repos
        iDataset_gesamt = len(dicDoi_ok)
        setPangaea_ids = set()
        setNotPangaea_ids = set()
        for sPublisher, dicPublisher in dicDOIs_clustered_byPublisher.items():
            setPublisher_ids = set(dicPublisher.keys())
            if sPublisher == "PANGAEA":
                setPangaea_ids = setPangaea_ids.union(setPublisher_ids)
            else:
                setNotPangaea_ids = setNotPangaea_ids.union(setPublisher_ids)
        # Pangaea
        iPangaea_nr = len(setPangaea_ids)
        try:
            iPangaea_proz = int(((iPangaea_nr / iDataset_gesamt) * 100))
        except:
            iPangaea_proz = 0
        # Not Pangaea
        iNotPangaea_nr = iDataset_gesamt - iPangaea_nr
        try:
            iNotPangaea_proz = int((iNotPangaea_nr / iDataset_gesamt) * 100)
        except:
            iNotPangaea_proz = 0

        # Ergebnisse
        res.addData("cluster-journal", dicDOIs_clustered_byPublisher)
        res.addData("cluster-doi", dicDatasets_clustered_byDoi)
        res.addData("cluster-pangaea", iPangaea_nr)
        res.addData("cluster-pangaea-ids", setPangaea_ids)
        res.addData("cluster-pangaea-proz", iPangaea_proz)
        res.addData("cluster-not-pangaea-ids", setNotPangaea_ids)
        res.addData("cluster-not-pangaea", iNotPangaea_nr)
        res.addData("cluster-not-pangaea-proz", iNotPangaea_proz)
        return res

    # --- Taxon

    def taxon_getAllAidFromTaxonList(self, **kwargs):
        """
        Hole aller Aid von Taxon Liste
        KWargs:
            cursor: Cursor
        Returns:
            result: Liste mit allen Taxon Aid
        """
        res=Result("Get Aid from Taxon List")
        cursor = kwargs["cursor"] if "cursor" in kwargs else None
        if cursor is None:
            resCursor = self.conn_getConnection()
            if resCursor.isErr():
                res.addResult(resCursor)
                return res
            conn = resCursor.getData("connection")
            cursor = resCursor.getData("cursor")
        try:
            sSqlCmd = "select distinct aid from " + self.account_active_getSchema() + ".taxon;"
            cursor.execute(sSqlCmd)
            lstContent = sorted([tp[0] for tp in cursor.fetchall()])
            res.addData("aid",lstContent)
        except Exception as ex:
            res.addErr("Unexcepted Error", add=str(ex), sys="Database", kat="Get all Aphia Ids from Taxon List")
        return res

    def taxon_getAllAaidFromTaxonList(self, **kwargs):
        """
        Hole aller Aaid von Taxon Liste
        Kwargs:
            cursor: Cursor
        Returns:
            result: Liste mit allen Taxon Aaid
        """
        res = Result("Get AphiaIds from Taxon List")
        cursor = kwargs["cursor"] if "cursor" in kwargs else None
        if cursor is None:
            resCursor = self.conn_getConnection()
            if resCursor.isErr():
                res.addResult(resCursor)
                return res
            conn = resCursor.getData("connection")
            cursor = resCursor.getData("cursor")

        try:
            sSqlCmd = "select distinct aaid from " + self.account_active_getSchema() + ".taxon where aaid is not null"
            cursor.execute(sSqlCmd)
            lstContent = sorted([tp[0] for tp in cursor.fetchall()])
            res.addData("aaid", lstContent)
        except Exception as ex:
            res.addErr("Unexcepted Error", add=str(ex), sys="Database", kat="Get ingest filename")
        return res

    # --- Generic Table

    def table_getCountEntries(self, sTab, **kwargs):
        """
        Hole Anzahl aller Eintraege zu einer Tabelle
        Args:
            sTab (str): Tabellenname
        Kwargs:
            cursor (psycopg2.cursor.Cursor): Postgres Cursor
        Returns:
            int: Anzahl aller Attribute der Tabelle
        """
        res = Result("Get count of entries")
        cursor = kwargs["cursor"] if "cursor" in kwargs else None
        if cursor is None:
            resCursor = self.conn_getConnection()
            if resCursor.isErr():
                res.addResult(resCursor)
                return res
            conn = resCursor.getData("connection")
            cursor = resCursor.getData("cursor")
        try:
            sSqlCmd = "SELECT count(*) from " + self.account_active_getSchema() + "." + sTab + ";"
            cursor.execute(sSqlCmd)
            iCount = cursor.fetchone()[0]  # Antwort
            res.addData("count", iCount)
        except Exception as ex:
            res.addErr("Unexpected error", add=str(ex), sys="Database", kat="Get lookup data for ingest")
        return res

    def table_getWholeTable(self, sTab, **kwargs):
        """
        Hole aller Eintraege zu einer Tabelle
        Args:
            sTab (str): Tabellenname
            order (str): Attribut nach dem sortiert werden soll
            cursor (psycopg2.cursor.Cursor): Postgres Cursor
            gold (bool): Einschränken nach Gold Atribut
            ids (lst): Liste mit Ids
        Returns:
            (list,list): Tupel mit Liste mit Headern und Liste mit Zeilen (Liste mit Attributen)
        """
        res = Result("Get all Table of entries")
        cursor = kwargs["cursor"] if "cursor" in kwargs else None
        bGold = kwargs["gold"] if "gold" in kwargs else None
        lstIds = kwargs["ids"] if "ids" in kwargs else None
        sOrderBy = kwargs["order"] if "order" in kwargs else None

        if bGold is not None:
            bGold is True if bGold is True else False

        if cursor is None:
            resCursor = self.conn_getConnection()
            if resCursor.isErr():
                res.addResult(resCursor)
                return res
            conn = resCursor.getData("connection")
            cursor = resCursor.getData("cursor")
        try:
            lstSqlWhere = []
            if bGold is not None:
                lstSqlWhere.append("gold="+str(bGold))
            if lstIds is not None:
                if len(lstIds) == 0:
                    lstSqlWhere.append("FALSE") # Leere Tabelle zurückgeben
                elif len(lstIds) == 1:
                    lstSqlWhere.append("id="+str(lstIds[0]))
                else:
                    lstSqlWhere.append("id in ("+PrettyText.lst2strC(lstIds)+")")
            if len(lstSqlWhere) > 0:
                sSqlWhere = " WHERE "+PrettyText.lst2str(lstSqlWhere, " AND ")
            else:
                sSqlWhere = ""

            # TODO: TWEAK region (geometry nie mitladen -> zu groß)
            if sTab == "region":
                sSqlSelect = "id, name, code, level"
            else:
                sSqlSelect = "*"
            sSqlOrderBy = " order by " + str(sOrderBy) if sOrderBy is not None else ""
            sSql = "SELECT " + sSqlSelect +" FROM " + self.account_active_getSchema() + "." + sTab + sSqlWhere + sSqlOrderBy
            cursor.execute(sSql)
            sSqlQuery = cursor.query.decode("utf-8")
            lstHeader = [head[0] for head in cursor.description]
            lstContent = cursor.fetchall()
            res.addData("header", lstHeader)
            res.addData("content", lstContent)
            res.addData("sql", sSqlQuery)

        except Exception as ex:
            res.addErr("Unexcepted Error", add=str(ex), sys="Database", kat="Get Whole Table")
        return res

    def table_isEntryInTable(self, sTab, sAtr, sVal, **kwargs):
        """
        Tete ob Eintrag in eine Tabelle ist
        Args:
            sTab (str): Tabellenname
            sAtr (str): Attribut
            val (str): Wert
        Kwargs:
            cursor (psycopg2.cursor.Cursor): Postgres Cursor
        Returns:
            found (bool)
        """
        res = Result("Get count of entries")
        cursor = kwargs["cursor"] if "cursor" in kwargs else None
        if cursor is None:
            resCursor = self.conn_getConnection()
            if resCursor.isErr():
                res.addResult(resCursor)
                return res
            conn = resCursor.getData("connection")
            cursor = resCursor.getData("cursor")
        try:
            # Daten
            sqlCmd = sql.SQL(
                "SELECT count (*) from {schema}.{table} WHERE {atr}={val}").format(
                schema=sql.Identifier(self.account_active_getSchema()),
                table=sql.Identifier(sTab),
                atr=sql.Identifier(sAtr),
                val=sql.Literal(sVal)
            )
            cursor.execute(sqlCmd)
            sSqlQuery = cursor.query.decode("utf-8")
            iCount = cursor.fetchone()[0]
            bFound = True if iCount > 0 else False
            res.addData("sql", sSqlQuery)
            res.addData("found", bFound)
            res.addData("count", iCount)
        except Exception as ex:
            res.addErr("Unexpected error", add=str(ex), sys="Database", kat="Is entry in table")
        return res

    def table_getTableEntries(self, sTab, sAtr, sVal, **kwargs):
        """
        Hole einen Eintrag einer Tabelle
        Args:
            sTab (str): Tabelle
            sAtr (str): Attribut
            sVal (int): Wert
        KWargs:
            cursor (psycopg2.cursor.Cursor): Postgres Cursor
        Returns:
            Result
        """
        res = Result("Get entries in Table")
        cursor = kwargs["cursor"] if "cursor" in kwargs else None
        bOwnConnection = False
        if cursor is None:
            bOwnConnection = True
            resCursor = self.conn_getConnection()
            if resCursor.isErr():
                res.addResult(resCursor)
                return res
            conn = resCursor.getData("connection")
            cursor = resCursor.getData("cursor")
        try:
            sqlCmd = sql.SQL(
                "SELECT * from {schema}.{table} WHERE {atr}={val}").format(
                schema=sql.Identifier(self.account_active_getSchema()),
                table=sql.Identifier(sTab),
                atr=sql.Identifier(sAtr),
                val=sql.Literal(sVal)
            )
            cursor.execute(sqlCmd)
            sSqlQuery = cursor.query.decode("utf-8")
            lstHeader = [head[0] for head in cursor.description]
            lstContent = cursor.fetchall()
            res.addData("sql", sSqlQuery)
            res.addData("header", lstHeader)
            res.addData("content", lstContent)
        except Exception as ex:
            res.addErr("Unexpected error", add=str(ex), sys="Database", kat="Get entries from table")
        if bOwnConnection:
            cursor.close()
            conn.close()
        return res

    def table_delTableEntry(self, sTab, sAtr, sVal, **kwargs):
        """
        Lösche einen Eintrag in einer Tabelle
        Args:
            sTab (str): Tabelle
            sAtr (str): Attribut
            sVal (int): Wert
        KWargs:
            cursor (psycopg2.cursor.Cursor): Postgres Cursor
        Returns:
            Result
        """
        res = Result("Del entries in Table")
        cursor = kwargs["cursor"] if "cursor" in kwargs else None
        bOwnConnection = False
        if cursor is None:
            bOwnConnection = True
            resCursor = self.conn_getConnection()
            if resCursor.isErr():
                res.addResult(resCursor)
                return res
            conn = resCursor.getData("connection")
            cursor = resCursor.getData("cursor")
        try:
            sqlCmd = sql.SQL(
                "DELETE FROM {schema}.{table} WHERE {atr}={val}").format(
                schema=sql.Identifier(self.account_active_getSchema()),
                table=sql.Identifier(sTab),
                atr=sql.Identifier(sAtr),
                val=sql.Literal(sVal)
            )
            cursor.execute(sqlCmd)
            sSqlQuery = cursor.query.decode("utf-8")
            res.addData("sql", sSqlQuery)
        except Exception as ex:
            res.addErr("Unexpected error", add=str(ex), sys="Database", kat="Delete Table Entries")
        if bOwnConnection:
            cursor.close()
            conn.close()
        return res

    def table_insertIntoTable(self, sTab, dicIn, **kwargs):
        """
        Eintrag in eine Tabelle vornehmen
        Args:
            sTab (str): Tabellenname
            dicIn (dict): Dictionary mit Attributnamen und entsprechenden Werten
        Kwargs:
            none (bool): None-Attribute benutzen oder entferenen (std: benutzen)
            cursor (psycopg2.cursor.Cursor): Postgres Cursor
        Returns:
            Result:
                sql (str): Kommandostring
        """
        res = Result("Create new record in database")
        cursor = kwargs["cursor"] if "cursor" in kwargs else None
        bUpdate = kwargs["update"] if ("update" in kwargs and kwargs["update"] in (True, False)) else False
        bNone = kwargs["none"] if ("none" in kwargs and kwargs["none"] in (True, False)) else True
        if cursor is None:
            resCursor = self.__model.subsystem_getDb().conn_getConnection()
            if resCursor.isErr():
                res.addResult(resCursor)
                return res
            conn = resCursor.getData("connection")
            cursor = resCursor.getData("cursor")

        if not bNone:
            for k in list(dicIn.keys()):
                if dicIn[k] in [None, ""]:
                    del(dicIn[k])
        try:
            sqlCmd = sql.SQL(
                "INSERT INTO {schema}.{table} ({attr}) VALUES ({val}) RETURNING id").format(
                schema=sql.Identifier(self.account_active_getSchema()),
                table=sql.Identifier(sTab),
                attr=sql.SQL(', ').join(map(sql.Identifier, dicIn.keys())),
                val=sql.SQL(', ').join(map(sql.Placeholder, dicIn.keys()))
            )
            cursor.execute(sqlCmd, dicIn)
            sSqlQuery = cursor.query.decode("utf-8")
            iId = cursor.fetchone()[0]
            res.addData("sql", sSqlQuery)
            res.addData("id", iId)
        except Exception as ex:
            res.addErr("Creation failed", add=str(ex), kat="Create entry in table", sys="Database")
        return res

    def table_updateInTable(self, sTab, iId, dicIn, **kwargs):
        """
        Update eines Eintrages in Tabelle
        Args:
            sTab (str): Tabellenname
            iId (int): Id des Eintrages
            dicIn (dict): Dictionary mit Attributnamen und entsprechenden Werten
        kwargs:
            none (bool): None-Attribute benutzen oder entferenen (std: benutzen)
            cursor (psycopg2.cursor.Cursor): Postgres Cursor
        Returns:
            Result:
                sql (str): Kommandostring
        """
        res = Result("Update record in database")
        cursor = kwargs["cursor"] if "cursor" in kwargs else None
        bNone = bool(kwargs["none"]) if "none" in kwargs else True
        if cursor is None:
            resCursor = self.__model.subsystem_getDb().conn_getConnection()
            if resCursor.isErr():
                res.addResult(resCursor)
                return res
            conn = resCursor.getData("connection")
            cursor = resCursor.getData("cursor")

        if not bNone:
            for k in list(dicIn.keys()):
                if dicIn[k] in [None, ""]:
                    del (dicIn[k])
        try:
            if len(dicIn) == 1:
                # Einzelwert
                sSql = "update {schema}.{table} set {attr} = {val} where id = {id}"
            else:
                # Mehrere Werte
                sSql = "update {schema}.{table} set ({attr}) = ({val}) where id = {id}"
            sqlCmd = sql.SQL(sSql).format(
                schema=sql.Identifier(self.account_active_getSchema()),
                table=sql.Identifier(sTab),
                id=sql.Literal(iId),
                attr=sql.SQL(', ').join(map(sql.Identifier, dicIn.keys())),
                val=sql.SQL(', ').join(map(sql.Placeholder, dicIn.keys()))
            )
            cursor.execute(sqlCmd, dicIn)
            sSqlQuery = cursor.query.decode("utf-8")
            res.addData("sql", sSqlQuery)
        except Exception as ex:
            res.addErr("Update failed", add=str(ex), sys="Update entry in table", kat="Database")
        return res

    def table_editTableEntry(self, sTab, sIdName, iId, dicEdit, **kwargs):
        """
        Verändere einen Eintrag in einer Tabelle

        Args:
            cursor (psycopg2.cursor.Cursor): Postgres Cursor
            sTab (str): Tabellenname
            sIdName (str): Name des ID-Feldes
            iId (int): Wert des ID-Feldes
            dicEdit (dict): Dictionary mit Attributnamen und entsprechenden Werten

        Returns:
            bool: War der Eintrag erfolgreich?
        """
        cursor = kwargs["cursor"] if "cursor" in kwargs else None
        res = Result("Edit record in database")
        if self.tools_isSqlUnsafe([sTab, sIdName]):
            res.addErr("Sql is unsafe")
            return res
        if sIdName in dicEdit: # Id-Attribut darf nicht mit verändert werden können
            res.addErr("Error: Id attribute cannot be edited: \"" + str(sIdName) + "\"")
            return res
        sSqlCmdTeil = ""
        sSqlVal = []
        # Durchlaufe alle Attribute und Werte
        for sAttribut, sWert in dicEdit.items():
            if self.tools_isSqlUnsafe(sAttribut):
                res.addErr("Sql is unsafe")
                return res
            if sWert == "":
                sWert = None
            sSqlCmdTeil += sAttribut + "=%s,"
            sSqlVal.append(sWert)
        # Sql-Cmd
        try:
            sSqlCmd = "update " + self.account_active_getSchema() + "." + sTab + " set " + sSqlCmdTeil[:-1] + " where " + sIdName + " = " + str(iId) + ";"
            cursor.execute(sSqlCmd, sSqlVal)
            res.addData("sql", cursor.query.decode("utf-8"))
        except Exception as ex:
            res.addErr(str(ex))
        return res

    def table_setGold(self, sTab, iId, bBool, **kwargs):
        """
        Setze den Gold Eintrag eines Eintrages
        Args:
            sTab (str): Tabellenname
            iId (int): Wert des ID-Feldes (das ID-Feld trägt immer den Namen "id")
            bBool (bool): Wert
        KWargs:
            cursor (psycopg2.cursor.Cursor): Postgres Cursor
        Returns:
            Result
        """
        res = Result("Set gold status for record")
        cursor = kwargs["cursor"] if "cursor" in kwargs else None
        if cursor is None:
            resCursor = self.conn_getConnection()
            if resCursor.isErr():
                res.addResult(resCursor)
                return res
            conn = resCursor.getData("connection")
            cursor = resCursor.getData("cursor")
        try:
            sSqlCmd = "UPDATE from " + self.account_active_getSchema() + "." + sTab + " where id=%s;"
            cursor.execute(sSqlCmd, [str(iId)])
        except Exception as ex:
            res.addErr("Unexcepted Error", add=str(ex), sys="Database", kat="Delete Table Entry")
        return res

    # --- Log

    def actionlog_new(self, sAction, sObject, sLog, **kwargs):
        """
        Eintrag in Action Log Tabelle vornehmen
        Args:
            sAction (str): Aktion
            sLog (str): Logging Text
        Kwargs:
            cursor (psycopg2.cursor.Cursor): Postgres Cursor
        Returns:
            Result
        """
        result = Result(name="Entry in action log")
        cursor = kwargs["cursor"] if "cursor" in kwargs else None

        # Db
        bOwnConnection = False
        if cursor is None:
            bOwnConnection = True
            resCursor = self.__model.subsystem_getDb().conn_getConnection()
            if resCursor.isErr():
                result.addResult(resCursor)
                return result
            conn = resCursor.getData("connection")
            cursor = resCursor.getData("cursor")

        # Check Action und Titel
        dicActions = {
            "ing-add": "Ingest Data",
            "ing-del": "Delete Ingest",
            "lup-add": "Lookup: Add Entries",
            "lup-del": "Lookup: Delete Entries",
            "lup-edit": "Lookup: Edit Entry",
            "tax-adm-add": "TaxonWizard: Add Admin List Entries",
            "tax-adm-del": "TaxonWizard: Delete Admin List Entries",
            "tax-adm-edit": "TaxonWizard: Edit Admin List Entiy",
            "tax-list-update": "TaxonWizard: Update Taxon List",
            "tax-base-new": "TaxonWizard: New Base Taxa",
            "tax-base-del": "TaxonWizard: Deleted Base Taxa",
            "data-patrol-biota-control": "Data Patrol: BiotaControl",
            "gold-lup-yes": "Lookup: Entries set to gold",
            "gold-lup-no": "Lookup: Entries set to not-gold",
            "gold-tax-yes": "TaxonWizard: Base Taxa set to gold",
            "gold-tax-no": "TaxonWizard: Base Taxa set to not-gold",
            "fao-update": "FAO regions: Update mapping",
            "fao-create": "FAO regions: Harvest new regions",
            "fao-del": "FAO regions: Remove samples from mapping"
        }
        if sAction in dicActions:
            sLog = "*** " + dicActions[sAction] + " ***\n\n" + sLog
        else:
            result.addErr("Creating entry in action log: Action unkwnown", add=str(sAction))
            return result

        # Eintrag in ActionLog-Tabelle
        try:
            sSqlCmd = "INSERT INTO " + self.__model.subsystem_getDb().account_single_getPart(
                "schema") + ".actionlog (date, username, host, ip, os, action, object, log) VALUES (%s,%s,%s,%s,%s,%s,%s,%s);"
            sUser = str(getpass.getuser())
            sDatum = datetime.datetime.now().strftime("%Y-%m-%d %H:%M:%S")
            sHost = str(socket.gethostname())
            sIp = str(socket.gethostbyname(sHost))
            sOs = str(platform.system()) + " " + str(platform.release())
            cursor.execute(sSqlCmd, (sDatum, sUser, sHost, sIp, sOs, sAction, sObject, sLog))
        except Exception as ex:
            # Error
            result.addErr("Creating entry in action log failed", add=str(ex))

        # Db
        if bOwnConnection:
            conn.commit()
            cursor.close()
            conn.close()

        return result

    def actionlog_getOverview(self):
        """
        Overview
        """
        res = Result("Get actionlog overview")
        resCursor = self.conn_getConnection()
        if resCursor.isErr():
            res.addResult(resCursor)
            return res
        conn = resCursor.getData("connection")
        cursor = resCursor.getData("cursor")
        try:
            sqlCmd = sql.SQL(
                "SELECT id, date, action, object, username, host, ip, os FROM {schema}.actionlog;"
                ).format(
                schema=sql.Identifier(self.account_active_getSchema())
            )
            cursor.execute(sqlCmd)
            sSqlQuery = cursor.query.decode("utf-8")
            lstHeader = [head[0] for head in cursor.description]
            lstContent = cursor.fetchall()
            res.addData("sql", sSqlQuery)
            res.addData("header", lstHeader)
            res.addData("content", lstContent)
        except Exception as ex:
            res.addErr("Unexpected error", add=str(ex), sys="Database", kat="Get entries from table")
        cursor.close()
        conn.close()
        return res

    def actionlog_delAll(self, **kwargs):
        """
        Lösche alle Action Log Einträge
        Kwargs:
            cursor:  Cursor
        Returns:
            Result
        """
        pn = self.__prnt.normal
        result = Result(name="Deleting all action logs")
        cursor = kwargs["cursor"] if "cursor" in kwargs else None
        resCursor = self.__model.subsystem_getDb().conn_getConnection()
        if resCursor.isErr():
            result.addResult(resCursor)
            return result
        conn = resCursor.getData("connection")
        cursor = resCursor.getData("cursor")
        try:
            sSqlCmd = "delete from " + self.__model.subsystem_getDb().account_single_getPart("schema") + ".actionlog"
            cursor.execute(sSqlCmd)
            if not self.__model.conf_isDry():
                conn.commit()
                pn("All action logs  deleted")
            else:
                pn("<Dry Run> activated")
                pn("Nothing was deleted")
                conn.rollback()
        except Exception as ex:
            result.addErr("Delete all action logs", add=str(ex), kat = "Database")
        cursor.close()
        conn.close()
        return result

    def actionlog_delSingle(self, lstActionLogIds):
        """
        Lösche Action Logs aus Datenbank
        Args:
            lstActionLogIds (lst): Ingest-IDs
        Returns:
            Result
        """
        pn = self.__prnt.normal
        result = Result(name="Deleting action logs")
        tpLstActionLogIds = tuple(lstActionLogIds)
        resCursor = self.__model.subsystem_getDb().conn_getConnection()
        if resCursor.isErr():
            result.addResult(resCursor)
            return result
        conn = resCursor.getData("connection")
        cursor = resCursor.getData("cursor")
        try:
            sSqlCmd = "delete from " + self.__model.subsystem_getDb().account_single_getPart("schema") + ".actionlog where id in %s"
            cursor.execute(sSqlCmd, (tpLstActionLogIds,))
            if not self.__model.conf_isDry():
                conn.commit()
                pn("Action logs deleted")
            else:
                pn("<Dry Run> activated")
                pn("Nothing was deleted")
                conn.rollback()
        except Exception as ex:
            result.addErr("Delete one action log", add=str(ex), kat = "Database")
        cursor.close()
        conn.close()
        return result


    def inglog_new(self, **kwargs):
        """
        Neuer leerer Eintrag in Ingest Log
        Kwargs:
           cursor
           schema
           wf (str): Workflow
        Returns:
           sql
           id
        """
        res = Result("Create Ingest Entry")
        cursor = kwargs["cursor"] if "cursor" in kwargs else None
        sSchema = kwargs["schema"] if "schema" in kwargs else self.account_active_getSchema()
        sWf = kwargs["wf"] if "wf" in kwargs else None
        if cursor is None:
            resCursor = self.__model.subsystem_getDb().conn_getConnection()
            if resCursor.isErr():
                res.addResult(resCursor)
                return res
            conn = resCursor.getData("connection")
            cursor = resCursor.getData("cursor")
        try:
            sName = str(getpass.getuser())
            sDatum = datetime.datetime.now().strftime("%Y-%m-%d %H:%M:%S")
            sqlCmd = sql.SQL(
                "INSERT INTO {schema}.ingest (name, workflow, created_on) VALUES ({name}, {wf}, {created_on}) RETURNING id").format(
                schema=sql.Identifier(sSchema),
                name=sql.Literal(sName),
                wf=sql.Literal(sWf),
                created_on=sql.Literal(sDatum)
            )
            cursor.execute(sqlCmd)
            sSqlQuery = cursor.query.decode("utf-8")
            iIngestId = cursor.fetchone()[0]  # Antwort
            res.addData("sql", sSqlQuery)
            res.addData("id", iIngestId)
        except Exception as ex:
            res.addErr("Create new entry in ingestion log", add=str(ex), sys="Database", kat="Ingestion log")
        return res


    def inglog_update(self, iIngestId, **kwargs):
        """
        Update Eintrag in Ingest Log
        Args:
           id: Ingest Id
        Kwargs:
           cursor
           schema
           wf
           rows
           warn
           info
           log
        """
        res = Result("Update Ingest Entry")
        cursor = kwargs["cursor"] if "cursor" in kwargs else None
        sSchema = kwargs["schema"] if "schema" in kwargs else self.account_active_getSchema()
        sWf = kwargs["wf"] if "wf" in kwargs else None
        sFile = kwargs["file"] if "file" in kwargs else None
        iRow = kwargs["rows"] if "rows" in kwargs else None
        iWarn= kwargs["warn"] if "warn" in kwargs else None
        iInfo = kwargs["info"] if "info" in kwargs else None
        sLog = kwargs["log"] if "log" in kwargs else None

        try:
            # Daten
            sqlCmd = sql.SQL(
                "UPDATE {schema}.ingest set rows={rows}, warn={warn}, info={info}, log={log}, workflow={wf}, file={file} where id={id}").format(
                schema=sql.Identifier(sSchema),
                wf = sql.Literal(sWf),
                file = sql.Literal(sFile),
                rows=sql.Literal(iRow),
                warn=sql.Literal(iWarn),
                info=sql.Literal(iInfo),
                log=sql.Literal(sLog),
                id=sql.Literal(iIngestId)
            )
            cursor.execute(sqlCmd)
            sSqlQuery = cursor.query.decode("utf-8")
            res.addData("sql", sSqlQuery)
        except Exception as ex:
            res.addErr("Update entry in ingestion log", add=str(ex), sys="Database", kat="Ingestion log")
        return res

    # --- Selector

    def selector_getSubsetStatsFromDb(self):
        """
        Hole eine Statistik über das Subset
        Returns:
            (tools.coding.Result): result
        """

        pn = self.__prnt.normal
        pv = self.__prnt.verbose
        pd = self.__prnt.debug

        result = Result(name="Get statistics of subset from database")

        if self.__model.ui_selector_getSelector().isErr():
            result.addErr("Selector has an error", kat="Selector")
            return result

        # --- Sql Abfragen für die Tabellen

        odicSqlCmds = OrderedDict()

        odicSqlCmds["selection"] = \
            """
            DROP VIEW IF EXISTS selection CASCADE;
            CREATE OR REPLACE TEMP VIEW selection AS
            SELECT DISTINCT sub.id
            FROM <schema>.sample smp 
            JOIN <schema>.subsample subsmp ON subsmp.sample_id = smp.id  
            JOIN <schema>.subset sub ON sub.subsample_id = subsmp.id  
            JOIN <schema>.station sta ON sta.id = smp.station_id
            JOIN <schema>.cruise cru ON cru.id = sta.cruise_id
            JOIN <schema>.ingest ing ON ing.id = smp.ingest_id
            JOIN <schema>.dataset dat ON dat.id = sub.dataset_id
            JOIN <schema>.gear gea ON gea.id = smp.gear_id
            JOIN <schema>.taxcover cov ON cov.id = sub.taxcover_id
            JOIN <schema>.region2sample reg2smp ON reg2smp.sample_id = smp.id
            JOIN <schema>.region reg ON reg2smp.region_id = reg.id
            WHERE 
            smp.date_start BETWEEN %s AND %s AND
            ing.workflow IN %s AND
            dat.name IN %s AND
            gea.category IN %s AND
            cov.name IN %s AND
            sub.ingest_id in %s AND
            reg.name in %s
            ;
            """

        odicSqlCmds["cruises"] = \
            """
            SELECT 
            COUNT(DISTINCT cru.id)
            FROM <schema>.subset sub
            JOIN <schema>.subsample subsmp ON subsmp.id = sub.subsample_id
            JOIN <schema>.sample smp ON smp.id = subsmp.sample_id
            JOIN <schema>.station sta ON sta.id = smp.station_id
            JOIN <schema>.cruise cru ON cru.id = sta.cruise_id
            WHERE sub.id IN (SELECT id FROM selection)
            ;
            """

        odicSqlCmds["stations"] = \
            """
            SELECT 
            COUNT(DISTINCT sta.id)
            FROM <schema>.subset sub
            JOIN <schema>.subsample subsmp ON subsmp.id = sub.subsample_id
            JOIN <schema>.sample smp ON smp.id = subsmp.sample_id
            JOIN <schema>.station sta ON sta.id = smp.station_id
            JOIN <schema>.cruise cru ON cru.id = sta.cruise_id
            WHERE sub.id IN (SELECT id FROM selection)
            ;
            """

        odicSqlCmds["samples"] = \
            """
            SELECT
            COUNT(DISTINCT smp.id)
            FROM <schema>.subset sub
            JOIN <schema>.subsample subsmp ON subsmp.id = sub.subsample_id
            JOIN <schema>.sample smp ON smp.id = subsmp.sample_id
            JOIN <schema>.station sta ON sta.id = smp.station_id
            JOIN <schema>.cruise cru ON cru.id = sta.cruise_id
            WHERE sub.id IN (SELECT id FROM selection)
            ;
            """

        odicSqlCmds["subsamples"] = \
            """
            SELECT
            COUNT(DISTINCT subsmp.id)
            FROM <schema>.subset sub
            JOIN <schema>.subsample subsmp ON subsmp.id = sub.subsample_id
            JOIN <schema>.sample smp ON smp.id = subsmp.sample_id
            JOIN <schema>.station sta ON sta.id = smp.station_id
            JOIN <schema>.cruise cru ON cru.id = sta.cruise_id
            WHERE sub.id IN (SELECT id FROM selection)
            ;
            """

        odicSqlCmds["subsets"] = \
            """
            SELECT
            COUNT(DISTINCT sub.id)
            FROM 
            <schema>.subset sub
            JOIN <schema>.subsample subsmp ON subsmp.id = sub.subsample_id
            JOIN <schema>.sample smp ON smp.id = subsmp.sample_id
            JOIN <schema>.station sta ON sta.id = smp.station_id
            JOIN <schema>.cruise cru ON cru.id = sta.cruise_id
            WHERE sub.id IN (SELECT id FROM selection)
            ;
            """

        odicSqlCmds["biota"] = \
            """
            SELECT
            COUNT(DISTINCT bio.id)
            FROM 
            <schema>.biota bio
            JOIN <schema>.subset sub ON sub.id = bio.subset_id
            JOIN <schema>.subsample subsmp ON subsmp.id = sub.subsample_id
            JOIN <schema>.sample smp ON smp.id = subsmp.sample_id
            JOIN <schema>.station sta ON sta.id = smp.station_id
            JOIN <schema>.cruise cru ON cru.id = sta.cruise_id
            WHERE sub.id IN (SELECT id FROM selection)
            ;
            """

        odicSqlCmds["biota_presence_absence"] = \
            """
            SELECT
            COUNT(DISTINCT bio.id)
            FROM 
            <schema>.biota bio
            JOIN <schema>.subset sub ON sub.id = bio.subset_id
            JOIN <schema>.subsample subsmp ON subsmp.id = sub.subsample_id
            JOIN <schema>.sample smp ON smp.id = subsmp.sample_id
            JOIN <schema>.station sta ON sta.id = smp.station_id
            JOIN <schema>.cruise cru ON cru.id = sta.cruise_id
            WHERE sub.id IN (SELECT id FROM selection) AND bio.presence_absence = 1
            ;
            """

        odicSqlCmds["biota_abundance"] = \
            """
            SELECT
            COUNT(DISTINCT bio.id)
            FROM 
            <schema>.biota bio
            JOIN <schema>.subset sub ON sub.id = bio.subset_id
            JOIN <schema>.subsample subsmp ON subsmp.id = sub.subsample_id
            JOIN <schema>.sample smp ON smp.id = subsmp.sample_id
            JOIN <schema>.station sta ON sta.id = smp.station_id
            JOIN <schema>.cruise cru ON cru.id = sta.cruise_id
            WHERE sub.id IN (SELECT id FROM selection) AND bio.abundance is not null
            ;
            """

        odicSqlCmds["taxa"] = \
            """
            SELECT
            COUNT (DISTINCT bio.taxon_id)
            FROM 
            <schema>.biota bio
            JOIN <schema>.subset sub ON sub.id = bio.subset_id
            JOIN <schema>.subsample subsmp ON subsmp.id = sub.subsample_id
            JOIN <schema>.sample smp ON smp.id = subsmp.sample_id
            JOIN <schema>.station sta ON sta.id = smp.station_id
            JOIN <schema>.cruise cru ON cru.id = sta.cruise_id
            WHERE sub.id IN (SELECT id FROM selection)
            ;
            """

        odicSqlCmds["sediment"] = \
            """
            SELECT 
            COUNT(DISTINCT sed.id)
            FROM 
            <schema>.sediment sed
            JOIN <schema>.subset sub ON sub.id = sed.subset_id
            JOIN <schema>.subsample subsmp ON subsmp.id = sub.subsample_id
            JOIN <schema>.sample smp ON smp.id = subsmp.sample_id
            JOIN <schema>.station sta ON sta.id = smp.station_id
            JOIN <schema>.cruise cru ON cru.id = sta.cruise_id
            WHERE sub.id IN (SELECT id FROM selection)
            ;
            """

        odicSqlCmds["indmeasure"] = \
            """
            SELECT 
            COUNT(DISTINCT aut.id)
            FROM 
            <schema>.indmeasure aut
            JOIN <schema>.biota bio ON bio.id = aut.biota_id
            JOIN <schema>.subset sub ON sub.id = bio.subset_id
            JOIN <schema>.subsample subsmp ON subsmp.id = sub.subsample_id
            JOIN <schema>.sample smp ON smp.id = subsmp.sample_id
            JOIN <schema>.station sta ON sta.id = smp.station_id
            JOIN <schema>.cruise cru ON cru.id = sta.cruise_id
            WHERE sub.id IN (SELECT id FROM selection)
            ;
            """

        odicSqlCmds["subsample_multi_subset"] = \
            """
            SELECT COUNT(*) FROM (
            SELECT
            subsmp.id AS subsample_id
            FROM 
            critterbase.subset sub
            JOIN <schema>.subsample subsmp ON subsmp.id = sub.subsample_id
            WHERE 
            sub.id IN (SELECT id FROM selection) AND
            subsmp.id = sub.subsample_id
            GROUP BY subsmp.id
            HAVING count(sub.id) > 1
            ) t1
            """

        odicSqlCmds["subsample_single_subset"] = \
            """
            SELECT COUNT(*) FROM (
            SELECT
            subsmp.id AS subsample_id
            FROM 
            critterbase.subset sub
            JOIN <schema>.subsample subsmp ON subsmp.id = sub.subsample_id
            WHERE 
            sub.id IN (SELECT id FROM selection) AND
            subsmp.id = sub.subsample_id
            GROUP BY subsmp.id
            HAVING count(sub.id) = 1
            ) t1
            """

        odicSqlCmds["subsample_no_subset"] = \
            """
            SELECT COUNT(*) FROM (
            SELECT
            subsmp.id AS subsample_id
            FROM 
            critterbase.subset sub
            JOIN <schema>.subsample subsmp ON subsmp.id = sub.subsample_id
            WHERE 
            sub.id IN (SELECT id FROM selection) AND
            subsmp.id = sub.subsample_id
            GROUP BY subsmp.id
            HAVING count(sub.id) = 0
            ) t1
            """

        # Bereite die Kommandos auf
        sSchema = self.__model.subsystem_getDb().account_single_getPart("schema")
        for sName, sSqlCmd in odicSqlCmds.items():
            sSqlCmd = sSqlCmd.replace("<schema>", sSchema)  # Schema einfügen
            sSqlCmd = re.sub(' +', ' ', sSqlCmd)  # Mehrfache Spaces
            sSqlCmd = re.sub('\t+', ' ', sSqlCmd)  # Mehrfache Tabs
            odicSqlCmds[sName] = sSqlCmd

        # --- Selectionsparameter

        # ... muss für DB in Tupel verpackt werden
        tpDateStart = tuple(self.__model.ui_selector_getSelector().getActivEntries("date_start"))
        tpDateEnd = tuple(self.__model.ui_selector_getSelector().getActivEntries("date_end"))
        tpRegion = tuple(self.__model.ui_selector_getSelector().getActivEntries("region"))
        tpType = tuple(self.__model.ui_selector_getSelector().getActivEntries("type"))
        tpDataset = tuple(self.__model.ui_selector_getSelector().getActivEntries("dataset"))
        tpTaxcover = tuple(self.__model.ui_selector_getSelector().getActivEntries("taxcover"))
        tpIngest = tuple(self.__model.ui_selector_getSelector().getActivEntries("ingest"))
        tpGear = tuple(self.__model.ui_selector_getSelector().getActivEntries("gear"))

        # ... leere Tuple -> Tuple mit None
        if tpDateStart == ():
            tpDateStart = (None,)
        if tpDateEnd == ():
            tpDateEnd = (None,)
        if tpRegion == ():
            tpRegion = (None,)
        if tpGear == ():
            tpGear = (None,)
        if tpType == ():
            tpType = (None,)
        if tpDataset == ():
            tpDataset = (None,)
        if tpIngest == ():
            tpIngest = (None,)
        if tpTaxcover == ():
            tpTaxcover = (None,)
        try:

            pd("* Fetching statistics on subset data")

            dicStat = {}

            # DB
            conn = psycopg2.connect(self.__model.subsystem_getDb().conn_getConnStrPsql())
            cursor = conn.cursor()

            # Loop Queries
            for sName, sSqlCmd in odicSqlCmds.items():
                if sName == "selection":
                    # Selection
                    cursor.execute(sSqlCmd, (tpDateStart, tpDateEnd, tpType, tpDataset, tpGear, tpTaxcover, tpIngest, tpRegion))
                    sSqlQuery = cursor.query.decode("utf-8")
                else:
                    # Eigentliche Daten
                    cursor.execute(sSqlCmd)
                    sSqlQuery = cursor.query.decode("utf-8")

                pd("Table: " + sName + "\n" + sSqlQuery)

                # Ergebnis
                if sName != "selection":
                    iCount = cursor.fetchone()[0]
                    dicStat[sName] = iCount

            # DB-Verbindung schliessen
            cursor.close()
            conn.close()
            result.addData("stat", dicStat)

        except Exception as ex:
            exc_type, exc_obj, exc_tb = sys.exc_info()
            iZeileCode = exc_tb.tb_lineno  # Falls ein Fehler entsteht -> Zeile
            result.addErr("Access to database", kat= "Database", add=str(ex), row=iZeileCode)
        return result

    def selector_getGeoJsonMap(self):
        """
        Baue ein GeoJson String mit den Map-Daten anhand des aktuellen Selectors zusammen

        Returns:
            (tools.coding.Result): result
        """

        pn = self.__prnt.normal
        pv = self.__prnt.verbose
        pd = self.__prnt.debug

        result = Result(name="Get GeoJson for map from database")

        if self.__model.ui_selector_getSelector().isErr():
            result.addErr("Selector has an error", kat="Selector")
            return result

        # --- Hole die Sql Abfragen für die Tabellen
        odicSqlCmds = self.__selector_getGeoJsonMap_sqlCmds()

        # --- Selectionsparameter

        # ... muss für DB in Tupel verpackt werdn
        tpDateStart = tuple(self.__model.ui_selector_getSelector().getActivEntries("date_start"))
        tpDateEnd = tuple(self.__model.ui_selector_getSelector().getActivEntries("date_end"))
        tpRegion = tuple(self.__model.ui_selector_getSelector().getActivEntries("region"))
        tpType = tuple(self.__model.ui_selector_getSelector().getActivEntries("type"))
        tpDataset = tuple(self.__model.ui_selector_getSelector().getActivEntries("dataset"))
        tpTaxcover = tuple(self.__model.ui_selector_getSelector().getActivEntries("taxcover"))
        tpIngest = tuple(self.__model.ui_selector_getSelector().getActivEntries("ingest"))
        tpGear = tuple(self.__model.ui_selector_getSelector().getActivEntries("gear"))

        # ... leere Tuple -> Tuple mit None
        if tpDateStart == ():
            tpDateStart = (None,)
        if tpDateEnd == ():
            tpDateEnd = (None,)
        if tpRegion == ():
            tpRegion = (None,)
        if tpGear == ():
            tpGear = (None,)
        if tpType == ():
            tpType = (None,)
        if tpDataset == ():
            tpDataset = (None,)
        if tpIngest == ():
            tpIngest = (None,)
        if tpTaxcover == ():
            tpTaxcover = (None,)

        try:

            # DB
            conn = psycopg2.connect(self.__model.subsystem_getDb().conn_getConnStrPsql())
            cursor = conn.cursor()

            # Selection
            sSchema = self.__model.subsystem_getDb().account_single_getPart("schema")
            sSqlCmdSelection = odicSqlCmds["selection"]
            sSqlCmdSelection = sSqlCmdSelection.replace("<schema>", sSchema)  # Schema einfügen
            sSqlCmdSelection = re.sub(' +', ' ', sSqlCmdSelection)  # Mehrfache Spaces
            sSqlCmdSelection = re.sub('\t+', ' ', sSqlCmdSelection)  # Mehrfache Tabs
            cursor.execute(sSqlCmdSelection, (tpDateStart, tpDateEnd, tpType, tpDataset, tpGear, tpTaxcover, tpIngest, tpRegion))
            sSqlQuery = cursor.query.decode("utf-8")
            result.addData("sql_sel", sSqlQuery)

            # GeoJson
            sSqlCmdGeoJson = """
                                SELECT 
                                json_build_object('type', 'FeatureCollection','features', 
                                    json_agg(
                                        json_build_object('type', 'Feature','id', p.sample_id,'geometry', 
                                            json_build_object('type', 'Point','coordinates', 
                                                json_build_array(p.sample_start_lon, p.sample_start_lat)),'properties', 
                                                    json_build_object('subset_name', p.subset_name, 'sample_name', p.sample_name, 'subsample_name', p.subsample_name, 'sample_date_start', p.sample_date_start, 'cruise_name', p.cruise_name,'station_name', p.station_name,'dataset_name', p.dataset_name, 'precision', p.precision, 'dataset_doi', p.doi, 'bio_image', p.bio_image)
                                        )
                                    )
                                )

                                FROM (
                                SELECT DISTINCT smp.id AS sample_id, 
                                cr.name AS cruise_name, 
                                st.name AS station_name, 
                                smp.name AS sample_name, 
                                subsmp.name AS subsample_name,
                                sub.name AS subset_name,
                                smp.ingest_id AS ingest_id, 
                                dat.name AS dataset_name, 
                                smp.precision AS precision,
                                dat.doi AS doi,
                                smp.lat_start AS sample_start_lat, 
                                smp.lon_start AS sample_start_lon,
                                smp.date_start AS sample_date_start,
                                bio.image AS bio_image  
                                
                                FROM <schema>.sample smp
                                JOIN <schema>.subsample subsmp ON subsmp.sample_id=smp.id 
                                JOIN <schema>.subset sub ON sub.subsample_id=subsmp.id
                                JOIN <schema>.dataset dat ON sub.dataset_id=dat.id 
                                JOIN <schema>.station st ON smp.station_id=st.id 
                                JOIN <schema>.cruise cr ON st.cruise_id=cr.id 
                                JOIN <schema>.gear ON smp.gear_id=gear.id 
                                JOIN <schema>.taxcover ON sub.taxcover_id=taxcover.id 
                                JOIN <schema>.biota bio ON bio.subset_id=sub.id 
                                WHERE sub.id IN (SELECT id FROM selection)
                                ) p;
                                """
            # sSqlCmdGeoJson = """
            #         SELECT
            #         json_build_object('type', 'FeatureCollection','features',
            #             json_agg(
            #                 json_build_object('type', 'Feature','id', p.sample_id,'geometry',
            #                     json_build_object('type', 'Point','coordinates',
            #                         json_build_array(p.sample_start_lon, p.sample_start_lat)),'properties',
            #                             json_build_object('subset_name', p.subset_name, 'sample_name', p.sample_name, 'subsample_name', p.subsample_name, 'sample_date_start', p.sample_date_start, 'cruise_name', p.cruise_name,'station_name', p.station_name,'dataset_name', p.dataset_name, 'spatial_precision', p.spatial_precision)
            #                 )
            #             )
            #         )
            #
            #         FROM (
            #         SELECT smp.id AS sample_id,
            #         cr.name AS cruise_name,
            #         st.name AS station_name,
            #         smp.name AS sample_name,
            #         subsmp.name AS subsample_name,
            #         sub.name AS subset_name,
            #         smp.ingest_id AS ingest_id,
            #         dat.name AS dataset_name,
            #         dat.spatial_precision AS spatial_precision,
            #         smp.lat_start AS sample_start_lat,
            #         smp.lon_start AS sample_start_lon,
            #         smp.date_start AS sample_date_start
            #
            #         FROM <schema>.sample smp
            #         JOIN <schema>.subsample subsmp ON subsmp.sample_id=smp.id
            #         JOIN <schema>.subset sub ON sub.subsample_id=subsmp.id
            #         JOIN <schema>.dataset dat ON sub.dataset_id=dat.id
            #         JOIN <schema>.station st ON smp.station_id=st.id
            #         JOIN <schema>.cruise cr ON st.cruise_id=cr.id
            #         JOIN <schema>.gear ON smp.gear_id=gear.id
            #         JOIN <schema>.taxcover ON sub.taxcover_id=taxcover.id
            #
            #         WHERE sub.id IN (SELECT id FROM selection)
            #         ) p;
            #         """
            sSqlCmdGeoJson = sSqlCmdGeoJson.replace("<schema>", sSchema)  # Schema einfügen
            sSqlCmdGeoJson = re.sub(' +', ' ', sSqlCmdGeoJson)  # Mehrfache Spaces
            sSqlCmdGeoJson = re.sub('\t+', ' ', sSqlCmdGeoJson)  # Mehrfache Tabs
            cursor.execute(sSqlCmdGeoJson)
            sSqlQuery = cursor.query.decode("utf-8")
            geoJson = cursor.fetchone()[0]

            # DB-Verbindung schliessen
            cursor.close()
            conn.close()
            result.addData("geojson", geoJson)
            result.addData("sql_json", sSqlQuery)

        except Exception as ex:
            exc_type, exc_obj, exc_tb = sys.exc_info()
            iZeileCode = exc_tb.tb_lineno  # Falls ein Fehler entsteht -> Zeile
            result.addErr("Access to database", kat="Database", add=str(ex) , row=iZeileCode, sys="Create JSON map")

        return result

    def __selector_getGeoJsonMap_sqlCmds(self, dicExcludeHeaders=None):
        """
        Passe die Sql-Kommandos für den Export von subsets an:

        a) Veraendere das aktuelle DB-Schema
        b) Benuzte nur bestimmte Header

        Args:
            odicColExclude (OrderedDictionary): Dictionary mit den Auszulassenden Columns der einzelnen Tabellen
        Returns:
            (OrderedDictionary): Befehle (die Werte für den Selector müssen noch in die %s-Felder eingefügt werden)
        """

        if dicExcludeHeaders is None:
            dicExcludeHeaders = OrderedDict()

        lstColNotAllowedToDelete = ("cruise_name", "station_name", "sample_name", "subsample_name", "subset_name", "biota_name")

        # Bereite die Kommandos auf
        sSchema = self.__model.subsystem_getDb().account_single_getPart("schema")
        odicSql = OrderedDict()
        for sTableName, (odicSqlCmdColumns, sSqlCmd) in self.__selector_getGeoJsonMap_sqlCmds_templateCmdsExportData().items():
            # Baue das Schema ein
            sSqlCmd = sSqlCmd.replace("<schema>", sSchema)  # Schema einfügen
            # Entferne Spaltenköpfe die nicht zum Download gewünscht sind
            lstSqlCmdColumns = []
            for sColName, sCmd in odicSqlCmdColumns.items():
                if sColName not in lstColNotAllowedToDelete and sTableName in dicExcludeHeaders and sColName in dicExcludeHeaders[sTableName]:
                    # Benutze Header nicht
                    pass
                else:
                    # Benutzer Header
                    lstSqlCmdColumns.append(sCmd)
            sSqlCmdColumns = ",\n".join(lstSqlCmdColumns)
            sSqlCmd = sSqlCmd.replace("<columns>", sSqlCmdColumns)  # Ausgewählte Spalten
            # Entferne Whitespaces
            sSqlCmd = re.sub(' +', ' ', sSqlCmd)  # Mehrfache Spaces
            sSqlCmd = re.sub('\t+', ' ', sSqlCmd)  # Mehrfache Tabs
            odicSql[sTableName] = sSqlCmd
        return odicSql

    def __selector_getGeoJsonMap_sqlCmds_templateCmdsExportData(self):
        """
        Baue die Vorlagen für die Befehle für den Export von subsets aus der Datenbank zusammen
        """

        odicSqlCmds = OrderedDict()
        odicSqlCmds["selection"] = (
            OrderedDict([
                ("id", "sub.id")
            ]),
            """
            DROP VIEW IF EXISTS selection CASCADE;
            CREATE OR REPLACE TEMP VIEW selection AS
            SELECT DISTINCT
            <columns>
            FROM <schema>.subset sub
            JOIN <schema>.subsample subsmp ON subsmp.id = sub.subsample_id
            JOIN <schema>.sample smp ON smp.id = subsmp.sample_id
            JOIN <schema>.station sta ON sta.id = smp.station_id
            JOIN <schema>.cruise cru ON cru.id = sta.cruise_id
            JOIN <schema>.ingest ing ON ing.id = smp.ingest_id
            JOIN <schema>.dataset dat ON dat.id = sub.dataset_id
            JOIN <schema>.gear gea ON gea.id = smp.gear_id
            JOIN <schema>.taxcover cov ON cov.id = sub.taxcover_id
            JOIN <schema>.region2sample reg2smp ON reg2smp.sample_id = smp.id
            JOIN <schema>.region reg ON reg2smp.region_id = reg.id
            WHERE 
            smp.date_start BETWEEN %s AND %s AND
            ing.workflow IN %s AND
            dat.name IN %s AND
            gea.category IN %s AND
            cov.name IN %s AND
            smp.ingest_id in %s AND
            reg.name in %s
            ;
            """)

        odicSqlCmds["cruise"] = (
            OrderedDict([
                ("cruise_name", "cru.name AS cruise_name"),
                ("lead_person", "per.full_name AS lead_person"),
                ("ship_name", "shp.name AS ship_name"),
                ("cruise_region", "cru.region AS cruise_region"),
                ("date_start", "cru.date_start"),
                ("date_end", "cru.date_end"),
                ("crs_name", "crs.name AS crs_name"),
                ("positioningsystem", "pos.name AS positioningsystem"),
                ("remark", "cru.remark")

            ]),
            """
            SELECT DISTINCT
            <columns>
            FROM <schema>.subset sub
            JOIN <schema>.subsample subsmp ON subsmp.id = sub.subsample_id
            JOIN <schema>.sample smp ON smp.id = subsmp.sample_id
            JOIN <schema>.station sta ON sta.id = smp.station_id
            JOIN <schema>.cruise cru ON cru.id = sta.cruise_id
            LEFT JOIN <schema>.person per ON cru.lead_person_id = per.id
            LEFT JOIN <schema>.crs crs ON cru.crs_id = crs.id
            LEFT JOIN <schema>.possys pos ON cru.positioningsystem_id = pos.id
            LEFT JOIN <schema>.ship shp ON cru.ship_id = shp.id
            WHERE sub.id IN (SELECT id FROM selection)
            ORDER BY cruise_name
            ;
            """)

        odicSqlCmds["station"] = (
            OrderedDict([
                ("cruise_name", "cru.name AS cruise_name"),
                ("station_name", "sta.name AS station_name"),
                ("start_lon", "sta.lon_start"),
                ("start_lat", "sta.lat_start"),
                ("date_start", "sta.date_start"),
                ("start_time", "sta.time_start"),
                ("start_depth", "sta.depth_start"),
                ("end_lon", "sta.lon_end"),
                ("end_lat", "sta.lat_end"),
                ("end_on", "sta.date_end"),
                ("end_time", "sta.time_end"),
                ("end_depth", "sta.depth_end"),
                ("location", "sta.location"),
                ("target_lon", "sta.lon_target"),
                ("target_lat", "sta.lat_target"),
                ("replicates", "sta.replicates"),
                ("temperature_air", "sta.temperature_air"),
                ("temperature_surfacewater", "sta.temperature_surfacewater"),
                ("temperature_bottomwater", "sta.temperature_bottomwater"),
                ("salinity_surfacewater", "sta.salinity_surfacewater"),
                ("salinity_bottomwater", "sta.salinity_bottomwater"),
                ("oxygen_bottomwater", "sta.oxygen_bottomwater"),
                ("windspeed_beaufort", "sta.windspeed_beaufort"),
                ("wind_direction", "sta.wind_direction"),
                ("remark", "sta.remark")
            ]),
            """
            SELECT * FROM 
            (
            SELECT DISTINCT  
            <columns>
            FROM <schema>.subset sub
            JOIN <schema>.subsample subsmp ON subsmp.id = sub.subsample_id
            JOIN <schema>.sample smp ON smp.id = subsmp.sample_id
            JOIN <schema>.station sta ON sta.id = smp.station_id
            JOIN <schema>.cruise cru ON cru.id = sta.cruise_id
            WHERE sub.id IN (SELECT id FROM selection)
            ) t
            ORDER BY length(cruise_name), cruise_name, length(station_name), station_name
            ;
            """)

        odicSqlCmds["sample"] = (
            OrderedDict([
                ("cruise_name", "cru.name AS cruise_name"),
                ("station_name", "sta.name AS station_name"),
                ("sample_name", "smp.name AS sample_name"),
                ("responsible_person", "per.full_name AS responsible_person"),
                ("gear_name", "gea.name AS gear_name"),
                ("sampled_area", "smp.sampled_area"),
                ("reference_area", "smp.reference_area"),
                ("start_lon", "smp.lon_start"),
                ("start_lat", "smp.lat_start"),
                ("date_start", "smp.date_start"),
                ("start_time", "smp.time_start"),
                ("date_start_error", "smp.date_start_error"),
                ("start_depth", "smp.depth_start"),
                ("end_lon", "smp.lon_end"),
                ("end_lat", "smp.lat_end"),
                ("end_on", "smp.date_end"),
                ("end_time", "smp.time_end"),
                ("end_depth", "smp.depth_end"),
                ("replicates", "smp.replicates"),
                ("sample_mass", "smp.sample_mass"),
                ("subsample_mass", "smp.subsample_mass"),
                ("subsample_share", "smp.subsample_share"),
                ("sampling_distance", "smp.sampling_distance"),
                ("method_biomass", "smp.method_biomass"),
                ("method_biomass_dry", "smp.method_biomass_dry"),
                ("method_biomass_afdm", "smp.method_biomass_afdm"),
                ("method_conservation", "smp.method_conservation"),
                ("temperature_surfacewater_start", "smp.temperature_surfacewater_start"),
                ("temperature_surfacewater_end", "smp.temperature_surfacewater_end"),
                ("salinity_surfacewater_start", "smp.salinity_surfacewater_start"),
                ("salinity_surfacewater_end", "smp.salinity_surfacewater_end"),
                ("sediment_properties", "smp.sediment_properties"),
                ("remark", "smp.remark")
            ]),
            """
            SELECT 
            <columns>
            FROM <schema>.subset sub
            JOIN <schema>.subsample subsmp ON subsmp.id = sub.subsample_id
            JOIN <schema>.sample smp ON smp.id = subsmp.sample_id
            JOIN <schema>.station sta ON sta.id = smp.station_id
            JOIN <schema>.cruise cru ON cru.id = sta.cruise_id
            JOIN <schema>.gear gea ON smp.gear_id = gea.id
            LEFT JOIN <schema>.person per ON smp.responsible_person_id = per.id
            WHERE sub.id IN (SELECT id FROM selection)
            ORDER BY length(cru.name), cruise_name, length(sta.name), station_name, length(smp.name), sample_name 
            ;
            """)

        odicSqlCmds["subset"] = (
            OrderedDict([
                ("cruise_name", "cru.name AS cruise_name"),
                ("station_name", "sta.name AS station_name"),
                ("sample_name", "smp.name AS sample_name"),
                ("subset_name", "sub.name AS subset_name"),
                ("dataset_name", "dat.name AS dataset_name"),
                ("taxonomic_coverage", "cov.name AS taxonomic_coverage"),
                ("taxonomic_resolution", "sub.taxres AS taxonomic_resolution"),
                ("spatial_level", "sub.spatiallevel AS spatial_level")
            ]),
            """
            SELECT 
            <columns>
            FROM <schema>.subset sub
            JOIN <schema>.subsample subsmp ON subsmp.id = sub.subsample_id
            JOIN <schema>.sample smp ON smp.id = subsmp.sample_id
            JOIN <schema>.station sta ON sta.id = smp.station_id
            JOIN <schema>.cruise cru ON cru.id = sta.cruise_id
            JOIN <schema>.taxcover cov ON cov.id = sub.taxcover_id
            JOIN <schema>.dataset dat ON dat.id = sub.dataset_id
            WHERE sub.id IN (SELECT id FROM selection)
            ORDER BY length(cru.name), cruise_name, length(sta.name), station_name, length(smp.name), sample_name, length(sub.name), subset_name 
            ;
            """)

        odicSqlCmds["biota"] = (
            OrderedDict([
                ("cruise_name", "cru.name AS cruise_name"),
                ("station_name", "sta.name AS station_name"),
                ("sample_name", "smp.name AS sample_name"),
                ("subset_name", "sub.name AS subset_name"),
                ("biota_name", "bio.name AS biota_name"),
                ("count", "bio.number"),
                ("abundance", "bio.abundance"),
                ("presence_absence", "bio.presence_absence"),
                ("biomass_wet", "bio.biomass_wet"),
                ("biomass_dry", "bio.biomass_dry"),
                ("biomass_afdm", "bio.biomass_afdm"),
                ("remark", "bio.remark"),
                ("lifestage_name", "lif.name as lifestage_name"),
                ("sieve_size", "siv.size as sieve_size"),
                ("sieve_description", "siv.description as sieve_description"),
                ("tax_aaid", "tax.aaid as tax_aaid"),
                ("tax_is_colony", "tax.is_colony as tax_is_colony"),
                ("tax_aid", "tax.aid as tax_aid"),
                ("tax_url", "tax.url as tax_url"),
                ("tax_scientificname", "tax.scientificname as tax_scientificname"),
                ("tax_authority", "tax.authority as tax_authority"),
                ("tax_status", "tax.status as tax_status"),
                ("tax_unacceptreason", "tax.unacceptreason as tax_unacceptreason"),
                ("tax_rank", "tax.rank as tax_rank"),
                ("tax_vaid", "tax.vaid as tax_vaid"),
                ("tax_valid_name", "tax.valid_name as tax_valid_name"),
                ("tax_valid_authority", "tax.valid_authority as tax_valid_authority"),
                ("tax_kingdom", "tax.kingdom as tax_kingdom"),
                ("tax_phylum", "tax.phylum as tax_phylum"),
                ("tax_class", "tax.class as tax_class"),
                ("tax_order", "tax.order as tax_order"),
                ("tax_family", "tax.family as tax_family"),
                ("tax_genus", "tax.genus as tax_genus"),
                ("tax_citation", "tax.citation as tax_citation"),
                ("tax_lsid", "tax.lsid as tax_lsid"),
                ("tax_ismarine", "tax.ismarine as tax_ismarine"),
                ("tax_isbrackish", "tax.isbrackish as tax_isbrackish"),
                ("tax_isfreshwater", "tax.isfreshwater as tax_isfreshwater"),
                ("tax_isterrestrial", "tax.isterrestrial as tax_isterrestrial"),
                ("tax_isextinct", "tax.isextinct as tax_isextinct"),
                ("tax_match_type", "tax.match_type as tax_match_type"),
                ("tax_modified", "tax.modified as tax_modified")
            ]),
            """
            SELECT  
            <columns>
            FROM 
            <schema>.biota bio
            JOIN <schema>.subset sub ON sub.id = bio.subset_id
            JOIN <schema>.subsample subsmp ON subsmp.id = sub.subsample_id
            JOIN <schema>.sample smp ON smp.id = subsmp.sample_id
            JOIN <schema>.station sta ON sta.id = smp.station_id
            JOIN <schema>.cruise cru ON cru.id = sta.cruise_id
            JOIN <schema>.taxon tax ON bio.taxon_id = tax.id
            LEFT JOIN <schema>.sieve siv ON bio.sieve_id = siv.id
            LEFT JOIN <schema>.lifestage lif ON bio.lifestage_id = lif.id
            WHERE sub.id IN (SELECT id FROM selection)
            ORDER BY length(cru.name), cruise_name, length(sta.name), station_name, length(smp.name), sample_name, length(bio.name), biota_name;
            ;
            """)

        odicSqlCmds["sediment"] = (
            OrderedDict([
                ("cruise_name", "cru.name AS cruise_name"),
                ("station_name", "sta.name AS station_name"),
                ("sample_name", "smp.name AS sample_name"),
                ("subset_name", "sub.name AS subset_name"),
                #("type", "sed.type"),
                ("residue1", "sed.residue1"),
                ("meshsize1", "sed.meshsize1"),
                ("residue2", "sed.residue2"),
                ("meshsize2", "sed.meshsize2"),
                ("residue3", "sed.residue3"),
                ("meshsize3", "sed.meshsize3"),
                ("residue4", "sed.residue4"),
                ("meshsize4", "sed.meshsize4"),
                ("residue5", "sed.residue5"),
                ("meshsize5", "sed.meshsize5"),
                ("residue6","sed.residue6"),
                ("meshsize6", "sed.meshsize6"),
                ("residue7", "sed.residue7"),
                ("meshsize7", "sed.meshsize7"),
                ("residue8", "sed.residue8"),
                ("meshsize8", "sed.meshsize8"),
                ("residue9", "sed.residue9"),
                ("meshsize9", "sed.meshsize9"),
                ("residue10", "sed.residue10"),
                ("meshsize10", "sed.meshsize10"),
                ("median", "sed.median"),
                ("gsd", "sed.gsd"),
                ("weight", "sed.weight"),
                #("grab_pd", "sed.grab_pd"),
                ("method_grainsize", "sed.method_grainsize"),
                ("temperature", "sed.temperature"),
                ("c_org", "sed.c_org"),
                ("c_total", "sed.c_total"),
                ("loss_on_ignition", "sed.loss_on_ignition"),
                ("nitrogen", "sed.nitrogen"),
                ("sulfur", "sed.sulfur"),
                ("oxygen_pd", "sed.oxygen_pd"),
                #("smell", "sed.smell"),
                ("remark", "sed.remark"),
                ("fuz_silt", "sed.fuz_silt"),
                ("fuz_vfsand", "sed.fuz_vfsand"),
                ("fuz_fsand", "sed.fuz_fsand"),
                ("fuz_msand", "sed.fuz_msand"),
                ("fuz_csand", "sed.fuz_csand"),
                ("fuz_vcsand", "sed.fuz_vcsand"),
                ("fuz_gravel", "sed.fuz_gravel"),
                ("fuz_stones", "sed.fuz_stones"),
                ("fuz_shell", "sed.fuz_shell"),
                ("fuz_clay", "sed.fuz_clay"),
            ]),
            """
            SELECT 
            <columns>
            FROM 
            <schema>.sediment sed
            JOIN <schema>.subset sub ON sub.id = sed.subset_id
            JOIN <schema>.subsample subsmp ON subsmp.id = sub.subsample_id
            JOIN <schema>.sample smp ON smp.id = subsmp.sample_id
            JOIN <schema>.station sta ON sta.id = smp.station_id
            JOIN <schema>.cruise cru ON cru.id = sta.cruise_id
            WHERE sub.id IN (SELECT id FROM selection)
            ORDER BY length(cru.name), cruise_name, length(sta.name), station_name, length(smp.name), sample_name
            ;
            """)

        odicSqlCmds["indmeasure"] = (
            OrderedDict([
                ("cruise_name", "cru.name AS cruise_name"),
                ("station_name", "sta.name AS station_name"),
                ("sample_name", "smp.name AS sample_name"),
                ("subset_name", "sub.name AS subset_name"),
                ("biota_name", "bio.name AS biota_name"),
                ("sex", "ind.sex"),
                ("length_total", "ind.length_total"),
                ("length_st", "ind.length_st"),
                ("mass_total", "ind.mass_total"),
                ("mass_sh", "ind.mass_sh"),
                ("remark", "ind.remark")
            ]),
            """
            SELECT 
            <columns>
            FROM 
            <schema>.indmeasure ind
            JOIN <schema>.biota bio ON bio.id = ind.biota_id
            JOIN <schema>.subset sub ON sub.id = bio.subset_id
            JOIN <schema>.subsample subsmp ON subsmp.id = sub.subsample_id
            JOIN <schema>.sample smp ON smp.id = subsmp.sample_id
            JOIN <schema>.station sta ON sta.id = smp.station_id
            JOIN <schema>.cruise cru ON cru.id = sta.cruise_id
            WHERE sub.id IN (SELECT id FROM selection)
            ORDER BY length(cru.name), cruise_name, length(sta.name), station_name, length(smp.name), sample_name
            ;
            """)

        return odicSqlCmds


